Package: qcaERT
Title: Enhanced Robustness Tests for Qualitative Comparative Analysis
Version: 0.1.0
Authors@R: 
    person("Breno A. H.", "Marisguia", email = "marisguiabreno@gmail.com",
           role = c("aut", "cre", "cph"))
Description: Provides functions for assessing and visualizing robustness in
    Qualitative Comparative Analysis (QCA) workflows built with the 'QCA'
    package, including calibration thresholds, inclusion cutoffs, frequency
    cutoffs, case influence, subsample stability, alternative analysis settings,
    theory-specific condition sets, cluster-specific patterns, and solution
    summaries. Methods build on Dusa (2019)
    <doi:10.1007/978-3-319-75668-4> and Ragin (2014, ISBN:9780520280038).
License: MIT + file LICENSE
Encoding: UTF-8
RoxygenNote: 7.3.3
VignetteBuilder: knitr, rmarkdown
Imports: methods, QCA, stats, utils
Suggests: ggplot2, knitr, rmarkdown, testthat (>= 3.0.0)
Config/testthat/edition: 3
NeedsCompilation: no
Packaged: 2026-05-19 15:11:56 UTC; breno
Author: Breno A. H. Marisguia [aut, cre, cph]
Maintainer: Breno A. H. Marisguia <marisguiabreno@gmail.com>
Repository: CRAN
Date/Publication: 2026-05-27 10:00:09 UTC
Built: R 4.5.2; ; 2026-05-27 11:13:05 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method(as.data.frame,altset_test)
S3method(as.data.frame,calib_test)
S3method(as.data.frame,cluster_test)
S3method(as.data.frame,incl_test)
S3method(as.data.frame,loo_test)
S3method(as.data.frame,ncut_test)
S3method(as.data.frame,subsample_test)
S3method(as.data.frame,theory_test)
S3method(plot,calib_test)
S3method(plot,incl_test)
S3method(plot,theory_test)
S3method(print,altset_test)
S3method(print,calib_test)
S3method(print,cluster_test)
S3method(print,incl_test)
S3method(print,loo_test)
S3method(print,ncut_test)
S3method(print,subsample_test)
S3method(print,theory_test)
export(altset.test)
export(calib.test)
export(cluster.test)
export(incl.test)
export(loo.test)
export(ncut.test)
export(sol.chart)
export(sol.df)
export(subsample.test)
export(theory.test)
importFrom(stats,setNames)
importFrom(stats,xtabs)

# qcaERT

qcaERT provides enhanced robustness tests for Qualitative Comparative Analysis
(QCA). It is designed for the moment after calibration, truth table
construction, and minimization, when the question becomes:

> How much does this QCA solution depend on my thresholds, cutoffs, cases,
> sample, or grouping structure?

The package is built as a family of sibling tools. They share common argument
conventions, returned-object structure, print behavior, and data-frame
coercion.

The strongest coverage is the `QCA` package workflow built around
`QCA::calibrate()`, `QCA::truthTable()`, `QCA::minimize()`, and
`QCA::findRows()`, especially crisp-set and fuzzy-set analyses. qcaERT is not
intended to cover every possible QCA variant.

## What qcaERT helps with

| Concern | Use |
| --- | --- |
| Calibration thresholds may be fragile | `calib.test()` |
| Inclusion cutoff may be fragile | `incl.test()` |
| Frequency cutoff may be fragile | `ncut.test()` |
| Individual cases may drive the solution | `loo.test()` |
| Sample composition may matter | `subsample.test()` |
| Several analytic choices may vary together | `altset.test()` |
| Theoretical condition-set specifications should be compared | `theory.test()` |
| Results may differ across groups or clusters | `cluster.test()` |
| QCA minimization output needs a clean table or chart | `sol.df()`, `sol.chart()` |

Inside R, the same map is available with:

```r
?qcaERT_tests
```

## Installation

qcaERT is currently in development. From a local source checkout, install it
with:

```r
install.packages("devtools")
devtools::install()
```

Then load it normally:

```r
library(qcaERT)
library(QCA)
```

Plotting is optional and uses `ggplot2`:

```r
install.packages("ggplot2")
```

## A compact LR workflow

This example uses the `LR` data from the `QCA` package.

```r
library(QCA)
library(qcaERT)

data(LR)

conditions <- c("DEV", "URB", "LIT", "IND", "STB")
outcome <- "SURV"
dir_exp <- rep("1", length(conditions))

thresholds <- list(
  DEV = findTh(LR$DEV, groups = 7),
  URB = findTh(LR$URB, groups = 4),
  LIT = findTh(LR$LIT, groups = 4),
  IND = findTh(LR$IND, groups = 4),
  STB = findTh(LR$STB, groups = 4),
  SURV = findTh(LR$SURV, groups = 4)
)

dat <- LR
dat$DEV <- calibrate(LR$DEV, type = "fuzzy", thresholds = thresholds$DEV)
dat$URB <- calibrate(LR$URB, type = "fuzzy", thresholds = thresholds$URB)
dat$LIT <- calibrate(LR$LIT, type = "fuzzy", thresholds = thresholds$LIT)
dat$IND <- calibrate(LR$IND, type = "fuzzy", thresholds = thresholds$IND)
dat$STB <- calibrate(LR$STB, type = "fuzzy", thresholds = thresholds$STB)
dat$SURV <- calibrate(LR$SURV, type = "fuzzy", thresholds = thresholds$SURV)
```

Run a regular QCA analysis:

```r
tt <- truthTable(
  dat,
  outcome = outcome,
  conditions = conditions,
  incl.cut = 0.8,
  n.cut = 1,
  complete = TRUE,
  show.cases = TRUE
)

sol <- minimize(tt, include = "", details = TRUE, show.cases = FALSE)
solution_table <- sol.df(conservative = sol, solution = "conservative")
solution_table
sol.chart(solution_table)
```

Then check robustness.

```r
incl_out <- incl.test(
  data = dat,
  outcome = outcome,
  conditions = conditions,
  incl.cut = 0.8,
  step = 0.05,
  max_steps = 4,
  n.cut = 1,
  solution = "all",
  dir.exp = dir_exp,
  progress = TRUE
)

incl_out
as.data.frame(incl_out)
incl_out$diagnostics
```

Compare theoretically motivated condition sets under the same analytic
settings:

```r
theories <- list(
  development = c("DEV", "URB", "LIT"),
  industrial = c("DEV", "URB", "IND"),
  broad = c("DEV", "URB", "LIT", "IND", "STB")
)

dir_exp_theories <- list(
  development = c("1", "1", "1"),
  industrial = c("1", "1", "1"),
  broad = c("1", "1", "1", "1", "1")
)

theory_out <- theory.test(
  data = dat,
  outcome = outcome,
  theories = theories,
  incl.cut = 0.8,
  n.cut = 1,
  solution = "all",
  dir.exp = dir_exp_theories,
  progress = TRUE
)

theory_out
as.data.frame(theory_out)
theory_out$results$solutions
theory_out$results$pairwise
```

For calibration robustness, define `calib_spec` and let qcaERT compute
scale-aware perturbation steps. `calib_spec` records the raw column,
calibration type, method, thresholds, and any extra `QCA::calibrate()`
arguments in one place.

```r
calib_spec <- list(
  DEV = list(raw = "DEV", type = "fuzzy", method = "direct", thresholds = thresholds$DEV),
  URB = list(raw = "URB", type = "fuzzy", method = "direct", thresholds = thresholds$URB),
  LIT = list(raw = "LIT", type = "fuzzy", method = "direct", thresholds = thresholds$LIT),
  IND = list(raw = "IND", type = "fuzzy", method = "direct", thresholds = thresholds$IND),
  STB = list(raw = "STB", type = "fuzzy", method = "direct", thresholds = thresholds$STB)
)

calib_spec_outcome <- calib_spec
calib_spec_outcome$SURV <- list(
  raw = "SURV",
  type = "fuzzy",
  method = "direct",
  thresholds = thresholds$SURV
)

calib_out <- calib.test(
  raw.data = LR,
  calib.data = dat,
  outcome = outcome,
  conditions = conditions,
  calib_spec = calib_spec,
  test.conditions = c("DEV", "URB"),
  unit_step = NULL,
  unit_step_divisor = 10,
  max_steps = 5,
  incl.cut = 0.8,
  n.cut = 1,
  solution = "all",
  dir.exp = dir_exp,
  progress = TRUE
)

calib_out
as.data.frame(calib_out)
calib_out$bounds
```

Here, `conditions` defines the full QCA model, while `test.conditions` selects
which calibrated conditions are perturbed. If `test.conditions` is omitted, all
model conditions are tested.

To test the outcome calibration, keep the outcome out of `conditions` and ask
for it explicitly:

```r
calib_outcome <- calib.test(
  raw.data = LR,
  calib.data = dat,
  outcome = outcome,
  conditions = conditions,
  calib_spec = calib_spec_outcome,
  test.conditions = NULL,
  test.outcome = TRUE,
  unit_step = NULL,
  unit_step_divisor = 10,
  max_steps = 5,
  incl.cut = 0.8,
  n.cut = 1,
  solution = "all",
  dir.exp = dir_exp,
  progress = TRUE
)
```

## Reading qcaERT results

Most qcaERT robustness functions return an S3 object with:

- `diagnostics`: detailed/internal results
- `results`: clean results
- `settings`: the settings used to run the analysis
- supporting components such as `baseline`, `bounds`, `by_direction`,
  `by_case`, `by_run`, `by_draw`, or `summary`

Use:

```r
print(incl_out)
as.data.frame(incl_out)
incl_out$diagnostics
```

For `incl.test()`, `ncut.test()`, and `calib.test()`, `result_shape` controls
the layout of the clean table when `solution = "all"`. The default `"wide"`
layout keeps one row per tested path with solution-type-specific columns;
`"long"` returns one row per tested path and solution type, with a
`solution_type` column.

`cluster.test()` and `theory.test()` are the deliberate structured-result
exceptions. `cluster.test()` contains three tables:

- `overview`
- `clusters`
- `units`

For `cluster_test` objects, `as.data.frame()` returns `results$overview`.

`theory.test()` contains:

- `models`
- `solutions`
- `pairwise`

For `theory_test` objects, `as.data.frame()` returns `results$models`.

## Plotting

If `ggplot2` is installed, `calib.test()`, `incl.test()`, and `theory.test()`
results can be plotted directly.

```r
plot(incl_out, solution_type = "conservative")
plot(incl_out, solution_type = "conservative", type = "trace", direction = "lower")

plot(calib_out, solution_type = "conservative")
plot(calib_out, solution_type = "conservative", type = "heatmap")
plot(calib_out, solution_type = "conservative", type = "trace", set = "DEV", anchor = "E1", direction = "lower")

plot(theory_out, solution_type = "conservative")
```

For direct six-threshold fuzzy calibration, anchors are `E1`, `C1`, `I1`,
`I2`, `C2`, and `E2`. For indirect calibration, anchors are `T1`, `T2`, and so
on.

See:

```r
?qcaERT_plots
```

## Learn more

```r
?qcaERT
?qcaERT_tests
?qcaERT_conventions
?qcaERT_plots
vignette("qcaERT-overview", package = "qcaERT")
vignette("qcaERT-result-objects", package = "qcaERT")
vignette("qcaERT-calibration", package = "qcaERT")
news(package = "qcaERT")
citation(package = "qcaERT")
```

## Status

qcaERT is release-ready.

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = FALSE
)

## ----core-rule----------------------------------------------------------------
# print(x)
# as.data.frame(x)
# x$results
# x$diagnostics

## ----boundary-layout----------------------------------------------------------
# as.data.frame(incl_out)
# incl_out$diagnostics
# incl_out$bounds
# 
# as.data.frame(ncut_out)
# ncut_out$diagnostics
# ncut_out$bounds
# 
# as.data.frame(calib_out)
# calib_out$diagnostics
# calib_out$bounds

## ----case-sample-layout-------------------------------------------------------
# as.data.frame(loo_out)
# loo_out$diagnostics
# loo_out$by_case
# 
# as.data.frame(subsample_out)
# subsample_out$diagnostics
# subsample_out$by_run
# subsample_out$summary

## ----altset-layout------------------------------------------------------------
# as.data.frame(altset_out)
# altset_out$diagnostics
# altset_out$by_draw
# altset_out$summary

## ----cluster-layout-----------------------------------------------------------
# cluster_out$results$overview
# cluster_out$results$clusters
# cluster_out$results$units
# 
# as.data.frame(cluster_out)
# cluster_out$diagnostics

## ----theory-layout------------------------------------------------------------
# as.data.frame(theory_out)
# theory_out$results$models
# theory_out$results$solutions
# theory_out$results$pairwise
# theory_out$diagnostics
# theory_out$by_theory

## ----sol-df-layout------------------------------------------------------------
# tab <- sol.df(intermediate = sol, solution = "intermediate")
# tab
# is.data.frame(tab)

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = FALSE
)

## ----setup-lr-----------------------------------------------------------------
# library(QCA)
# library(qcaERT)
# 
# data(LR)
# 
# conditions <- c("DEV", "URB", "LIT", "IND", "STB")
# outcome <- "SURV"
# dir_exp <- rep("1", length(conditions))
# 
# thresholds <- list(
#   DEV = findTh(LR$DEV, groups = 7),
#   URB = findTh(LR$URB, groups = 4),
#   LIT = findTh(LR$LIT, groups = 4),
#   IND = findTh(LR$IND, groups = 4),
#   STB = findTh(LR$STB, groups = 4),
#   SURV = findTh(LR$SURV, groups = 4)
# )
# 
# dat <- LR
# dat$DEV <- calibrate(LR$DEV, type = "fuzzy", thresholds = thresholds$DEV)
# dat$URB <- calibrate(LR$URB, type = "fuzzy", thresholds = thresholds$URB)
# dat$LIT <- calibrate(LR$LIT, type = "fuzzy", thresholds = thresholds$LIT)
# dat$IND <- calibrate(LR$IND, type = "fuzzy", thresholds = thresholds$IND)
# dat$STB <- calibrate(LR$STB, type = "fuzzy", thresholds = thresholds$STB)
# dat$SURV <- calibrate(LR$SURV, type = "fuzzy", thresholds = thresholds$SURV)
# 
# tt <- truthTable(
#   data = dat,
#   outcome = outcome,
#   conditions = conditions,
#   incl.cut = 0.8,
#   n.cut = 1
# )
# 
# sol <- minimize(tt, include = "?", dir.exp = dir_exp)
# sol.df(intermediate = sol, solution = "intermediate")

## ----incl-ncut----------------------------------------------------------------
# incl_out <- incl.test(
#   data = dat,
#   outcome = outcome,
#   conditions = conditions,
#   incl.cut = 0.8,
#   n.cut = 1,
#   step = 0.02,
#   max_steps = 5,
#   solution = "all",
#   dir.exp = dir_exp,
#   progress = TRUE
# )
# 
# incl_out
# as.data.frame(incl_out)
# incl_out$diagnostics
# 
# ncut_out <- ncut.test(
#   data = dat,
#   outcome = outcome,
#   conditions = conditions,
#   n.cut = 1,
#   incl.cut = 0.8,
#   step = 1,
#   max_steps = 3,
#   solution = "all",
#   dir.exp = dir_exp,
#   progress = TRUE
# )
# 
# ncut_out
# as.data.frame(ncut_out)

## ----calib--------------------------------------------------------------------
# calib_spec <- list(
#   DEV = list(raw = "DEV", type = "fuzzy", thresholds = thresholds$DEV),
#   URB = list(raw = "URB", type = "fuzzy", thresholds = thresholds$URB),
#   LIT = list(raw = "LIT", type = "fuzzy", thresholds = thresholds$LIT),
#   IND = list(raw = "IND", type = "fuzzy", thresholds = thresholds$IND),
#   STB = list(raw = "STB", type = "fuzzy", thresholds = thresholds$STB)
# )
# 
# calib_spec_outcome <- calib_spec
# calib_spec_outcome$SURV <- list(
#   raw = "SURV",
#   type = "fuzzy",
#   thresholds = thresholds$SURV
# )
# 
# calib_out <- calib.test(
#   raw.data = LR,
#   calib.data = dat,
#   outcome = outcome,
#   conditions = conditions,
#   calib_spec = calib_spec,
#   test.conditions = c("DEV", "URB"),
#   unit_step = NULL,
#   unit_step_divisor = 10,
#   max_steps = 5,
#   incl.cut = 0.8,
#   n.cut = 1,
#   solution = "all",
#   dir.exp = dir_exp,
#   progress = TRUE
# )
# 
# calib_out
# as.data.frame(calib_out)
# calib_out$bounds

## ----calib-outcome------------------------------------------------------------
# calib_outcome <- calib.test(
#   raw.data = LR,
#   calib.data = dat,
#   outcome = outcome,
#   conditions = conditions,
#   calib_spec = calib_spec_outcome,
#   test.conditions = NULL,
#   test.outcome = TRUE,
#   unit_step = NULL,
#   unit_step_divisor = 10,
#   max_steps = 5,
#   incl.cut = 0.8,
#   n.cut = 1,
#   solution = "all",
#   dir.exp = dir_exp,
#   progress = TRUE
# )

## ----cases-samples------------------------------------------------------------
# loo_out <- loo.test(
#   data = dat,
#   outcome = outcome,
#   conditions = conditions,
#   cases = 1:5,
#   incl.cut = 0.8,
#   n.cut = 1,
#   solution = "all",
#   dir.exp = dir_exp,
#   progress = TRUE
# )
# 
# loo_out
# as.data.frame(loo_out)
# 
# subsample_out <- subsample.test(
#   data = dat,
#   outcome = outcome,
#   conditions = conditions,
#   sample_prop = 0.8,
#   reps = 25,
#   seed = 123,
#   incl.cut = 0.8,
#   n.cut = 1,
#   solution = "all",
#   dir.exp = dir_exp,
#   progress = TRUE
# )
# 
# subsample_out
# subsample_out$summary

## ----altsets------------------------------------------------------------------
# altset_out <- altset.test(
#   raw.data = LR,
#   calib.data = dat,
#   outcome = outcome,
#   conditions = conditions,
#   calib_spec = calib_spec,
#   test.conditions = c("DEV", "URB"),
#   unit_step = NULL,
#   unit_step_divisor = 10,
#   calib_max_steps = 5,
#   incl.cut = 0.8,
#   incl_step = 0.02,
#   incl_max_steps = 5,
#   n.cut = 1,
#   ncut_step = 1,
#   ncut_max_steps = 2,
#   n_draws = 50,
#   seed = 123,
#   solution = "all",
#   dir.exp = dir_exp,
#   progress = TRUE
# )
# 
# altset_out
# as.data.frame(altset_out)
# altset_out$summary

## ----theory-------------------------------------------------------------------
# theories <- list(
#   development = c("DEV", "URB", "LIT"),
#   industrial = c("DEV", "URB", "IND"),
#   broad = c("DEV", "URB", "LIT", "IND", "STB")
# )
# 
# dir_exp_theories <- list(
#   development = c("1", "1", "1"),
#   industrial = c("1", "1", "1"),
#   broad = c("1", "1", "1", "1", "1")
# )
# 
# theory_out <- theory.test(
#   data = dat,
#   outcome = outcome,
#   theories = theories,
#   incl.cut = 0.8,
#   n.cut = 1,
#   solution = "all",
#   dir.exp = dir_exp_theories,
#   progress = TRUE
# )
# 
# theory_out
# as.data.frame(theory_out)
# theory_out$results$solutions
# theory_out$results$pairwise

## ----clusters-----------------------------------------------------------------
# cluster_data <- dat
# cluster_data$region <- ifelse(seq_len(nrow(cluster_data)) %% 2 == 0, "A", "B")
# cluster_data$unit <- rownames(cluster_data)
# 
# cluster_out <- cluster.test(
#   data = cluster_data,
#   tt = tt,
#   cluster_id = "region",
#   unit_id = "unit",
#   solution = "all",
#   dir.exp = dir_exp,
#   progress = TRUE
# )
# 
# cluster_out
# as.data.frame(cluster_out)
# cluster_out$results$clusters
# cluster_out$results$units

## ----plots--------------------------------------------------------------------
# plot(incl_out, solution_type = "conservative")
# plot(calib_out, solution_type = "conservative")
# plot(calib_out, solution_type = "conservative", type = "heatmap")
# plot(calib_out, solution_type = "conservative", type = "trace", set = "DEV", anchor = "E1", direction = "lower")
# plot(theory_out, solution_type = "conservative")

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = FALSE
)

## ----setup-lr-----------------------------------------------------------------
# library(QCA)
# library(qcaERT)
# 
# data(LR)
# 
# conditions <- c("DEV", "URB", "LIT", "IND", "STB")
# outcome <- "SURV"
# dir_exp <- rep("1", length(conditions))
# 
# thresholds <- list(
#   DEV = findTh(LR$DEV, groups = 7),
#   URB = findTh(LR$URB, groups = 4),
#   LIT = findTh(LR$LIT, groups = 4),
#   IND = findTh(LR$IND, groups = 4),
#   STB = findTh(LR$STB, groups = 4),
#   SURV = findTh(LR$SURV, groups = 4)
# )
# 
# dat <- LR
# dat$DEV <- calibrate(LR$DEV, type = "fuzzy", thresholds = thresholds$DEV)
# dat$URB <- calibrate(LR$URB, type = "fuzzy", thresholds = thresholds$URB)
# dat$LIT <- calibrate(LR$LIT, type = "fuzzy", thresholds = thresholds$LIT)
# dat$IND <- calibrate(LR$IND, type = "fuzzy", thresholds = thresholds$IND)
# dat$STB <- calibrate(LR$STB, type = "fuzzy", thresholds = thresholds$STB)
# dat$SURV <- calibrate(LR$SURV, type = "fuzzy", thresholds = thresholds$SURV)

## ----calib-spec---------------------------------------------------------------
# calib_spec <- list(
#   DEV = list(
#     raw = "DEV",
#     type = "fuzzy",
#     method = "direct",
#     thresholds = thresholds$DEV
#   ),
#   URB = list(
#     raw = "URB",
#     type = "fuzzy",
#     thresholds = thresholds$URB
#   ),
#   LIT = list(
#     raw = "LIT",
#     type = "fuzzy",
#     thresholds = thresholds$LIT
#   ),
#   IND = list(
#     raw = "IND",
#     type = "fuzzy",
#     thresholds = thresholds$IND
#   ),
#   STB = list(
#     raw = "STB",
#     type = "fuzzy",
#     thresholds = thresholds$STB
#   )
# )
# 
# calib_spec_outcome <- calib_spec
# calib_spec_outcome$SURV <- list(
#   raw = "SURV",
#   type = "fuzzy",
#   thresholds = thresholds$SURV
# )

## ----conditions---------------------------------------------------------------
# calib.test(
#   raw.data = LR,
#   calib.data = dat,
#   outcome = "SURV",
#   conditions = c("DEV", "URB", "LIT", "IND", "STB"),
#   calib_spec = calib_spec,
#   test.conditions = c("DEV", "URB"),
#   unit_step = NULL,
#   unit_step_divisor = 10,
#   max_steps = 5,
#   incl.cut = 0.8,
#   n.cut = 1,
#   solution = "all",
#   dir.exp = rep("1", 5),
#   progress = TRUE
# )

## ----outcome-calibration------------------------------------------------------
# calib_outcome <- calib.test(
#   raw.data = LR,
#   calib.data = dat,
#   outcome = outcome,
#   conditions = conditions,
#   calib_spec = calib_spec_outcome,
#   test.conditions = NULL,
#   test.outcome = TRUE,
#   unit_step = NULL,
#   unit_step_divisor = 10,
#   max_steps = 5,
#   incl.cut = 0.8,
#   n.cut = 1,
#   solution = "all",
#   dir.exp = dir_exp,
#   progress = TRUE
# )
# 
# calib_outcome
# as.data.frame(calib_outcome)

## ----anchors------------------------------------------------------------------
# calib.test(
#   raw.data = LR,
#   calib.data = dat,
#   outcome = "SURV",
#   conditions = conditions,
#   calib_spec = calib_spec,
#   test.conditions = "DEV",
#   anchors_to_test = c("E1", "C1", "I1", "I2", "C2", "E2"),
#   unit_step = NULL,
#   unit_step_divisor = 10,
#   max_steps = 5,
#   incl.cut = 0.8,
#   n.cut = 1,
#   solution = "all",
#   dir.exp = dir_exp,
#   progress = TRUE
# )

## ----scale-aware--------------------------------------------------------------
# calib_out <- calib.test(
#   raw.data = LR,
#   calib.data = dat,
#   outcome = outcome,
#   conditions = conditions,
#   calib_spec = calib_spec,
#   test.conditions = c("DEV", "URB"),
#   unit_step = NULL,
#   unit_step_divisor = 10,
#   max_steps = 5,
#   incl.cut = 0.8,
#   n.cut = 1,
#   solution = "all",
#   dir.exp = dir_exp,
#   progress = TRUE
# )

## ----indirect-----------------------------------------------------------------
# calib_spec_indirect <- list(
#   A = list(
#     raw = "A_raw",
#     type = "fuzzy",
#     method = "indirect",
#     thresholds = c(10, 20, 30, 40)
#   )
# )

## ----altset-------------------------------------------------------------------
# altset_out <- altset.test(
#   raw.data = LR,
#   calib.data = dat,
#   outcome = outcome,
#   conditions = conditions,
#   calib_spec = calib_spec,
#   test.conditions = c("DEV", "URB"),
#   unit_step = NULL,
#   unit_step_divisor = 10,
#   calib_max_steps = 5,
#   incl.cut = 0.8,
#   incl_step = 0.02,
#   incl_max_steps = 5,
#   n.cut = 1,
#   ncut_step = 1,
#   ncut_max_steps = 2,
#   n_draws = 50,
#   seed = 123,
#   solution = "all",
#   dir.exp = dir_exp,
#   progress = TRUE
# )
# 
# altset_out
# as.data.frame(altset_out)
# altset_out$summary

## ----altset-outcome-----------------------------------------------------------
# altset_outcome <- altset.test(
#   raw.data = LR,
#   calib.data = dat,
#   outcome = outcome,
#   conditions = conditions,
#   calib_spec = calib_spec_outcome,
#   test.conditions = c("DEV", "URB"),
#   test.outcome = TRUE,
#   unit_step = NULL,
#   unit_step_divisor = 10,
#   calib_max_steps = 5,
#   incl.cut = 0.8,
#   incl_step = 0.02,
#   incl_max_steps = 5,
#   n.cut = 1,
#   ncut_step = 1,
#   ncut_max_steps = 2,
#   n_draws = 50,
#   seed = 456,
#   solution = "all",
#   dir.exp = dir_exp,
#   progress = TRUE
# )
# 
# as.data.frame(altset_outcome)

## ----calib-plots--------------------------------------------------------------
# plot(calib_out, solution_type = "conservative")
# plot(calib_out, solution_type = "conservative", type = "heatmap")
# plot(calib_out, solution_type = "conservative", type = "trace", set = "DEV", anchor = "E1", direction = "lower")

#' Cluster heterogeneity diagnostics for QCA configurations
#'
#' Evaluates how the consistency and coverage of selected QCA configurations
#' vary across clusters and, when repeated units are available, across units
#' observed in more than one cluster. The function starts from an existing
#' truth table, minimizes it under the requested solution type, extracts the
#' selected baseline configurations, and then compares pooled fit values with
#' cluster-specific and within-unit fit values.
#'
#' @param data A data frame object containing the columns used to build
#'   `tt` plus the cluster and, optionally, unit identifiers. `data`
#'   must have the same number of rows as `tt$recoded.data`.
#' @param tt A truth table object of class `"QCA_tt"`, typically created with
#'   [QCA::truthTable()].
#' @param cluster_id Name of the column in `data` identifying clusters. This
#'   must be a single non-empty character string, and the column must contain
#'   at least two distinct non-missing cluster values.
#' @param unit_id Optional name of the column in `data` identifying units that
#'   may appear in more than one cluster. If `NULL`, units are treated as row
#'   positions and within-unit diagnostics are not available across clusters.
#' @param solution Solution type to evaluate. Accepted values are `"all"`,
#'   `"con"` or `"conservative"`, `"par"` or `"parsimonious"`, and `"int"` or
#'   `"intermediate"`.
#' @param include Optional minimization include setting. Currently, this
#'   argument accepts only `NULL`, `""`, or `"?"`.
#' @param dir.exp Directional expectations used when the monitored solution is
#'   intermediate. When `solution = "all"`, supplying `dir.exp` adds
#'   intermediate solutions to the monitored set.
#' @param exclude Optional exclusion specification passed to [QCA::minimize()]
#'   for parsimonious and intermediate minimization.
#' @param which_M Positive integer giving which solution alternative to use
#'   when minimization returns multiple models.
#' @param i_mode Character string controlling intermediate-solution selection.
#'   Accepted values are `"all"` and `"C1P1"`.
#' @param necessity Logical; if `TRUE`, compute fit as necessity consistency
#'   and coverage. If `FALSE`, compute fit as sufficiency consistency and
#'   coverage.
#' @param progress Logical; if `TRUE` and the session is interactive, show a
#'   text progress bar while minimizing the monitored solution types.
#' @param x A `cluster_test` object returned by [cluster.test()].
#' @param row.names Logical; passed to [print.data.frame()] by
#'   [print.cluster_test()].
#' @param ... Additional arguments. In [cluster.test()], these are filtered and
#'   forwarded to [QCA::minimize()]. The function also looks in `...` for
#'   `include`, `dir.exp` or `direxp`, and `exclude` or `omit` if those
#'   arguments were not supplied explicitly. In [print.cluster_test()], `...`
#'   is passed to [print.data.frame()]. In
#'   [as.data.frame.cluster_test()], `...` is ignored.
#'
#' @returns An object of class `cluster_test` with the following components:
#'   \describe{
#'     \item{`diagnostics`}{A detailed data frame with one row per selected
#'     configuration. It records the solution type, configuration key, component count,
#'     status, pooled consistency and coverage, maximum and mean absolute
#'     cluster-level deltas, the clusters with the worst consistency and
#'     coverage values, whether within-unit diagnostics were available, the
#'     number of repeated units, and error information when a configuration could not
#'     be evaluated.}
#'     \item{`results`}{A named list with three tables:
#'       \describe{
#'         \item{`overview`}{One row per selected configuration, summarizing pooled fit,
#'         maximum cluster deltas, worst clusters, and within-unit
#'         availability.}
#'         \item{`clusters`}{One row per configuration-component-cluster combination,
#'         giving cluster-specific consistency, coverage, and deltas from the
#'         pooled configuration values. The whole configuration is stored as
#'         `component = "solution"`; separate term-level rows are included only
#'         for multi-term solutions.}
#'         \item{`units`}{One row per configuration-component-unit combination for
#'         units observed in more than one cluster, giving within-unit
#'         consistency, coverage, and deltas from the pooled configuration values.
#'         Separate term-level rows are included only for multi-term solutions.}
#'       }}
#'     \item{`baseline`}{A list containing the input truth table, minimization
#'     results by solution type, selected solution terms used for comparison,
#'     metadata, and the extracted configuration definitions used for the
#'     heterogeneity diagnostics.}
#'     \item{`by_cluster`}{A named list of detailed cluster-level diagnostics
#'     for each selected configuration.}
#'     \item{`by_unit`}{A named list of detailed unit-level diagnostics for
#'     each selected configuration when repeated units are available, or `NULL`
#'     otherwise.}
#'     \item{`settings`}{A list containing the analysis settings used to build
#'     the result object.}
#'   }
#'
#'   `print.cluster_test()` prints a compact overview and, when the object has
#'   one displayed configuration, cluster and within-unit details.
#'   `as.data.frame.cluster_test()` returns `results$overview`.
#'
#' @details
#' The function recovers the outcome membership from `tt$recoded.data`, runs
#' minimization for the requested solution type or solution types, extracts the
#' selected baseline configurations, and computes pooled fit values for each
#' configuration and configuration component.
#'
#' Cluster-level diagnostics compare the pooled fit of each selected
#' configuration with the fit obtained inside each cluster defined by
#' `cluster_id`.
#'
#' If `unit_id` identifies units that appear in more than one cluster, the
#' function also computes within-unit diagnostics by comparing the same unit
#' across clusters.
#'
#' Configuration extraction uses the data-frame `pims` component produced by
#' [QCA::minimize()]. The shared solution-control, QCA solution-object, and
#' returned-object conventions are described in `?qcaERT_conventions`.
#'
#' @examples
#' \donttest{
#' library(QCA)
#' data(LR)
#'
#' conditions <- c("DEV", "URB", "LIT", "IND", "STB")
#' outcome <- "SURV"
#' dir_exp <- rep("1", length(conditions))
#'
#' thresholds <- list(
#'   DEV = findTh(LR$DEV, groups = 7),
#'   URB = findTh(LR$URB, groups = 4),
#'   LIT = findTh(LR$LIT, groups = 4),
#'   IND = findTh(LR$IND, groups = 4),
#'   STB = findTh(LR$STB, groups = 4),
#'   SURV = findTh(LR$SURV, groups = 4)
#' )
#'
#' dat <- LR
#' dat$DEV <- calibrate(LR$DEV, type = "fuzzy", thresholds = thresholds$DEV)
#' dat$URB <- calibrate(LR$URB, type = "fuzzy", thresholds = thresholds$URB)
#' dat$LIT <- calibrate(LR$LIT, type = "fuzzy", thresholds = thresholds$LIT)
#' dat$IND <- calibrate(LR$IND, type = "fuzzy", thresholds = thresholds$IND)
#' dat$STB <- calibrate(LR$STB, type = "fuzzy", thresholds = thresholds$STB)
#' dat$SURV <- calibrate(LR$SURV, type = "fuzzy", thresholds = thresholds$SURV)
#' dat$development_group <- ifelse(
#'   LR$DEV >= median(LR$DEV, na.rm = TRUE),
#'   "higher development",
#'   "lower development"
#' )
#' dat$case_id <- rownames(dat)
#'
#' tt <- truthTable(
#'   data = dat,
#'   outcome = outcome,
#'   conditions = conditions,
#'   incl.cut = 0.8,
#'   n.cut = 1,
#'   complete = TRUE,
#'   show.cases = TRUE
#' )
#' enhanced <- findRows(tt, type = 2)
#'
#' out <- cluster.test(
#'   data = dat,
#'   tt = tt,
#'   cluster_id = "development_group",
#'   unit_id = "case_id",
#'   solution = "intermediate",
#'   dir.exp = dir_exp,
#'   exclude = enhanced,
#'   which_M = 1,
#'   necessity = FALSE,
#'   progress = TRUE
#' )
#'
#' out
#' as.data.frame(out)
#' out$results$clusters
#' out$results$units
#' }
#'
#' @seealso [calib.test()], [incl.test()], [ncut.test()], [loo.test()],
#'   [subsample.test()], [altset.test()], [theory.test()], [sol.df()]
#' @export
cluster.test <- function(
    data,
    tt,
    cluster_id,
    unit_id = NULL,
    solution = "all",
    include = NULL,
    dir.exp = NULL,
    exclude = NULL,
    which_M = 1,
    i_mode = c("all", "C1P1"),
    necessity = FALSE,
    progress = TRUE,
    ...
) {
  .require_qca()

  selection_controls <- .normalize_i_mode(i_mode)
  i_mode <- selection_controls$i_mode
  dots_raw <- list(...)

  if (is.null(dim(data)) || is.null(nrow(data)) || nrow(data) < 1L) {
    stop("`data` must be a non-empty data frame object with at least one row.")
  }

  if (!methods::is(tt, "QCA_tt")) {
    stop("`tt` must be a truth table object of class 'QCA_tt'.")
  }

  if (!is.character(cluster_id) || length(cluster_id) != 1L || !nzchar(cluster_id)) {
    stop("`cluster_id` must be a single non-empty character string.")
  }
  if (!cluster_id %in% colnames(data)) {
    stop("`cluster_id` must name a column in `data`.")
  }

  if (!is.null(unit_id)) {
    if (!is.character(unit_id) || length(unit_id) != 1L || !nzchar(unit_id)) {
      stop("`unit_id` must be NULL or a single non-empty character string.")
    }
    if (!unit_id %in% colnames(data)) {
      stop("`unit_id` must name a column in `data`.")
    }
  }

  if (!is.logical(necessity) || length(necessity) != 1L || is.na(necessity)) {
    stop("`necessity` must be a single TRUE/FALSE value.")
  }

  if (is.null(tt$recoded.data) || nrow(tt$recoded.data) != nrow(data)) {
    stop("`data` must have the same number of rows as `tt$recoded.data`.")
  }

  .err_msg <- function(e) {
    if (!inherits(e, "error")) return(NA_character_)
    tryCatch(conditionMessage(e), error = function(...) as.character(e))
  }

  .fit_pair <- function(x, y, necessity = FALSE) {
    if (length(x) != length(y)) stop("Internal error: x and y must have equal length.")

    ok <- !(is.na(x) | is.na(y))
    x <- as.numeric(x[ok])
    y <- as.numeric(y[ok])

    if (length(x) == 0L) {
      return(c(consistency = NA_real_, coverage = NA_real_))
    }

    num <- sum(pmin(x, y))
    den_cons <- if (necessity) sum(y) else sum(x)
    den_cov  <- if (necessity) sum(x) else sum(y)

    cons <- if (den_cons > 0) num / den_cons else NA_real_
    cov  <- if (den_cov  > 0) num / den_cov  else NA_real_

    c(consistency = cons, coverage = cov)
  }

  .summarize_delta <- function(d) {
    d <- as.numeric(d)
    d <- d[!is.na(d)]
    if (length(d) == 0L) {
      return(c(max_abs = NA_real_, mean_abs = NA_real_))
    }
    c(max_abs = max(abs(d)), mean_abs = mean(abs(d)))
  }

  .diag_component_tables <- function(x, y, cluster_vec, unit_vec, necessity) {
    pooled <- .fit_pair(x, y, necessity = necessity)

    ok <- !(is.na(x) | is.na(y) | is.na(cluster_vec) | is.na(unit_vec))
    x <- as.numeric(x[ok])
    y <- as.numeric(y[ok])
    cluster_vec <- as.character(cluster_vec[ok])
    unit_vec <- as.character(unit_vec[ok])

    X <- xtabs(as.numeric(x) ~ unit_vec + cluster_vec)
    Y <- xtabs(as.numeric(y) ~ unit_vec + cluster_vec)

    cls <- colnames(X)
    cls_sizes <- table(cluster_vec)

    by_cluster <- lapply(seq_along(cls), function(j) {
      fitj <- .fit_pair(X[, j], Y[, j], necessity = necessity)
      data.frame(
        cluster_id = cls[j],
        cluster_size = as.integer(cls_sizes[cls[j]]),
        consistency = unname(fitj["consistency"]),
        coverage = unname(fitj["coverage"]),
        delta_consistency = unname(fitj["consistency"] - pooled["consistency"]),
        delta_coverage = unname(fitj["coverage"] - pooled["coverage"]),
        stringsAsFactors = FALSE
      )
    })
    by_cluster <- do.call(rbind, by_cluster)

    unit_clusters <- tapply(cluster_vec, unit_vec, function(z) length(unique(as.character(z))))
    repeated_units <- names(unit_clusters)[!is.na(unit_clusters) & unit_clusters > 1L]

    by_unit <- NULL
    if (length(repeated_units) > 0L) {
      row_keep <- rownames(X) %in% repeated_units
      Xu <- X[row_keep, , drop = FALSE]
      Yu <- Y[row_keep, , drop = FALSE]

      by_unit <- lapply(seq_len(nrow(Xu)), function(i) {
        fiti <- .fit_pair(Xu[i, ], Yu[i, ], necessity = necessity)
        data.frame(
          unit_id = rownames(Xu)[i],
          n_clusters = sum((Xu[i, ] > 0) | (Yu[i, ] > 0)),
          consistency = unname(fiti["consistency"]),
          coverage = unname(fiti["coverage"]),
          delta_consistency = unname(fiti["consistency"] - pooled["consistency"]),
          delta_coverage = unname(fiti["coverage"] - pooled["coverage"]),
          stringsAsFactors = FALSE
        )
      })
      by_unit <- do.call(rbind, by_unit)
    }

    list(
      pooled = pooled,
      by_cluster = by_cluster,
      by_unit = by_unit
    )
  }

  .run_solution_type <- function(tt_obj, solution_type, exclude, dir.exp, dots_filtered, which_M, i_mode) {
    args <- c(
      list(
        input = tt_obj,
        include = if (solution_type == "conservative") "" else "?"
      ),
      dots_filtered
    )

    if (solution_type == "intermediate") {
      args$dir.exp <- dir.exp
    }

    if (solution_type %in% c("parsimonious", "intermediate") && !is.null(exclude)) {
      args$exclude <- exclude
    }

    res <- tryCatch(
      suppressWarnings(do.call(QCA::minimize, args)),
      error = function(e) e
    )

    if (inherits(res, "error")) {
      return(list(
        status = "minimize_error",
        error_source = solution_type,
        error_message = .err_msg(res),
        res = NULL,
        sig = NULL,
        selected_solution_missing = NA,
        meta = NULL
      ))
    }

    sig_info <- .extract_solution_type_sig(
      res,
      solution_type = solution_type,
      which_M = which_M,
      i_mode = i_mode
    )

    list(
      status = "ok",
      error_source = NA_character_,
      error_message = NA_character_,
      res = res,
      sig = sig_info$sig,
      selected_solution_missing = isTRUE(sig_info$selected_solution_missing),
      meta = sig_info$meta
    )
  }

  conditions <- tt$options$conditions
  if (is.null(conditions) || length(conditions) < 1L) {
    stop("Could not infer conditions from `tt`.")
  }
  if (!is.null(tt$options$outcome)) {
    .validate_outcome_conditions_distinct(tt$options$outcome, conditions)
  }

  include_user <- include
  dir_user <- dir.exp
  exclude_user <- exclude
  solution_user <- solution

  if (is.null(include)) {
    inc <- .dot_get(dots_raw, "include")
    if (!is.null(inc)) include <- inc
  }

  if (is.null(dir.exp)) {
    de <- .dot_get(dots_raw, "dir.exp")
    if (is.null(de)) de <- .dot_get(dots_raw, "direxp")
    if (!is.null(de)) dir.exp <- de
  }
  dir.exp <- .normalize_dir_exp_generic(
    dir.exp,
    conditions = conditions,
    endpoint_phrase = "the truth table conditions"
  )

  if (is.null(exclude)) {
    ex <- .dot_get(dots_raw, "exclude")
    if (is.null(ex)) ex <- .dot_get(dots_raw, "omit")
    if (!is.null(ex)) exclude <- ex
  }

  solution_controls <- .resolve_solution_controls(
    solution = solution,
    include = include,
    dir.exp = dir.exp,
    caller = "cluster.test",
    style = "std"
  )

  solution <- solution_controls$solution
  include <- solution_controls$include
  monitored_solutions <- solution_controls$monitored
  which_M <- .coerce_which_M(which_M)
  dots_filtered <- .filter_dots_for_minimize(dots_raw)

  progress_state <- .new_qcaert_progress(
    total = 2L * length(monitored_solutions),
    progress = progress
  )
  on.exit(progress_state$close(), add = TRUE)
  .bump_pb <- progress_state$tick

  cluster_nonmiss <- stats::na.omit(as.character(data[[cluster_id]]))
  if (length(unique(cluster_nonmiss)) < 2L) {
    stop("`cluster_id` must identify at least two distinct non-missing clusters.")
  }

  outcome_name <- tt$options$outcome
  if (length(outcome_name) > 1L) outcome_name <- outcome_name[1L]
  outcome_name <- gsub("^~", "", trimws(as.character(outcome_name)))

  if (!outcome_name %in% colnames(tt$recoded.data)) {
    stop("Could not recover the outcome membership from `tt$recoded.data`.")
  }

  recoded_outcome <- as.numeric(tt$recoded.data[, outcome_name])
  cluster_vec <- as.character(data[[cluster_id]])
  if (is.null(unit_id)) {
    unit_vec <- as.character(seq_len(nrow(data)))
  } else {
    unit_vec <- as.character(data[[unit_id]])
  }

  baseline_res <- list(conservative = NULL, parsimonious = NULL, intermediate = NULL)
  baseline_sig <- list(conservative = NULL, parsimonious = NULL, intermediate = NULL)
  baseline_meta <- list(conservative = NULL, parsimonious = NULL, intermediate = NULL)
  baseline_status <- c(conservative = NA_character_, parsimonious = NA_character_, intermediate = NA_character_)
  baseline_error_source <- c(conservative = NA_character_, parsimonious = NA_character_, intermediate = NA_character_)
  baseline_error_message <- c(conservative = NA_character_, parsimonious = NA_character_, intermediate = NA_character_)

  for (rg in monitored_solutions) {
    rr <- .run_solution_type(
      tt_obj = tt,
      solution_type = rg,
      exclude = exclude,
      dir.exp = dir.exp,
      dots_filtered = dots_filtered,
      which_M = which_M,
      i_mode = i_mode
    )

    baseline_status[[rg]] <- rr$status
    baseline_error_source[[rg]] <- rr$error_source
    baseline_error_message[[rg]] <- rr$error_message
    baseline_res[[rg]] <- rr$res
    baseline_sig[[rg]] <- rr$sig
    baseline_meta[[rg]] <- rr$meta

    .bump_pb()
  }

  targets <- list()
  diagnostics_rows <- list()
  by_cluster <- list()
  by_unit <- list()

  for (rg in monitored_solutions) {
    if (!isTRUE(baseline_status[[rg]] == "ok")) {
      target_key <- paste0("<", toupper(substr(rg, 1L, 3L)), "_ERROR>")
      diagnostics_rows[[length(diagnostics_rows) + 1L]] <- data.frame(
        solution_type = rg,
        i_mode = if (identical(rg, "intermediate")) i_mode else NA_character_,
        which_M = which_M,
        necessity = necessity,
        configuration_key = target_key,
        component_count = NA_integer_,
        status = "minimize_error",
        error_source = baseline_error_source[[rg]],
        error_message = baseline_error_message[[rg]],
        pooled_consistency = NA_real_,
        pooled_coverage = NA_real_,
        n_clusters = length(unique(stats::na.omit(cluster_vec))),
        max_abs_delta_consistency = NA_real_,
        max_abs_delta_coverage = NA_real_,
        mean_abs_delta_consistency = NA_real_,
        mean_abs_delta_coverage = NA_real_,
        worst_cluster_consistency_id = NA_character_,
        worst_cluster_coverage_id = NA_character_,
        worst_cluster_consistency = NA_real_,
        worst_cluster_coverage = NA_real_,
        within_available = FALSE,
        n_units_repeated = 0L,
        max_abs_within_delta_consistency = NA_real_,
        max_abs_within_delta_coverage = NA_real_,
        mean_abs_within_delta_consistency = NA_real_,
        mean_abs_within_delta_coverage = NA_real_,
        stringsAsFactors = FALSE
      )

      .bump_pb()
      next
    }

    tar_rg <- .qca_extract_targets_for_solution_type(
      res = baseline_res[[rg]],
      solution_type = rg,
      which_M = which_M,
      i_mode = i_mode
    )

    for (ttg in tar_rg) {
      target_key <- ttg$target_key
      component_count <- length(ttg$terms)
      term_components <- if (length(ttg$terms) > 1L) ttg$terms else character(0)

      if (!identical(ttg$status, "ok")) {
        diagnostics_rows[[length(diagnostics_rows) + 1L]] <- data.frame(
          solution_type = rg,
          i_mode = if (identical(rg, "intermediate")) i_mode else NA_character_,
          which_M = which_M,
          necessity = necessity,
          configuration_key = target_key,
          component_count = component_count,
          status = ttg$status,
          error_source = ttg$error_source,
          error_message = ttg$error_message,
          pooled_consistency = NA_real_,
          pooled_coverage = NA_real_,
          n_clusters = length(unique(stats::na.omit(cluster_vec))),
          max_abs_delta_consistency = NA_real_,
          max_abs_delta_coverage = NA_real_,
          mean_abs_delta_consistency = NA_real_,
          mean_abs_delta_coverage = NA_real_,
          worst_cluster_consistency_id = NA_character_,
          worst_cluster_coverage_id = NA_character_,
          worst_cluster_consistency = NA_real_,
          worst_cluster_coverage = NA_real_,
          within_available = FALSE,
          n_units_repeated = 0L,
          max_abs_within_delta_consistency = NA_real_,
          max_abs_within_delta_coverage = NA_real_,
          mean_abs_within_delta_consistency = NA_real_,
          mean_abs_within_delta_coverage = NA_real_,
          stringsAsFactors = FALSE
        )

        targets[[target_key]] <- c(
          ttg,
          list(
            outcome_membership = recoded_outcome,
            pooled_consistency = NA_real_,
            pooled_coverage = NA_real_
          )
        )
        next
      }

      comp_names <- c("solution", term_components)
      comp_vectors <- list(solution = as.numeric(ttg$solution_membership))

      if (length(term_components) > 0L &&
          !is.null(ttg$term_memberships) &&
          ncol(ttg$term_memberships) > 0L) {
        for (j in seq_along(ttg$terms)) {
          term_name <- ttg$terms[j]
          if (term_name %in% term_components) {
            comp_vectors[[term_name]] <- as.numeric(ttg$term_memberships[[j]])
          }
        }
      }

      pooled_terms <- vector("list", length(comp_names))
      names(pooled_terms) <- comp_names

      cluster_terms <- vector("list", length(comp_names))
      names(cluster_terms) <- comp_names

      unit_terms <- vector("list", length(comp_names))
      names(unit_terms) <- comp_names

      for (nm in comp_names) {
        dd <- .diag_component_tables(
          x = comp_vectors[[nm]],
          y = recoded_outcome,
          cluster_vec = cluster_vec,
          unit_vec = unit_vec,
          necessity = necessity
        )

        pooled_terms[[nm]] <- dd$pooled

        tmpc <- dd$by_cluster
        tmpc$component <- nm
        cluster_terms[[nm]] <- tmpc[, c(
          "cluster_id",
          "cluster_size",
          "component",
          "consistency",
          "coverage",
          "delta_consistency",
          "delta_coverage"
        )]

        if (!is.null(dd$by_unit)) {
          tmpu <- dd$by_unit
          tmpu$component <- nm
          unit_terms[[nm]] <- tmpu[, c(
            "unit_id",
            "n_clusters",
            "component",
            "consistency",
            "coverage",
            "delta_consistency",
            "delta_coverage"
          )]
        } else {
          unit_terms[[nm]] <- NULL
        }
      }

      clusters_df <- do.call(rbind, cluster_terms)
      units_df <- if (all(vapply(unit_terms, is.null, logical(1)))) {
        NULL
      } else {
        do.call(rbind, unit_terms[!vapply(unit_terms, is.null, logical(1))])
      }

      sol_cluster <- subset(clusters_df, component == "solution")
      sum_cl_cons <- .summarize_delta(sol_cluster$delta_consistency)
      sum_cl_cov  <- .summarize_delta(sol_cluster$delta_coverage)

      worst_cons_idx <- if (nrow(sol_cluster) > 0L && any(!is.na(sol_cluster$delta_consistency))) {
        which.max(abs(sol_cluster$delta_consistency))
      } else {
        NA_integer_
      }
      worst_cov_idx <- if (nrow(sol_cluster) > 0L && any(!is.na(sol_cluster$delta_coverage))) {
        which.max(abs(sol_cluster$delta_coverage))
      } else {
        NA_integer_
      }

      pooled_solution_df <- data.frame(
        consistency = unname(pooled_terms[["solution"]]["consistency"]),
        coverage = unname(pooled_terms[["solution"]]["coverage"]),
        stringsAsFactors = FALSE
      )

      term_only <- term_components
      pooled_terms_df <- if (length(term_only) > 0L) {
        data.frame(
          component = term_only,
          consistency = vapply(term_only, function(nm) unname(pooled_terms[[nm]]["consistency"]), numeric(1)),
          coverage = vapply(term_only, function(nm) unname(pooled_terms[[nm]]["coverage"]), numeric(1)),
          stringsAsFactors = FALSE
        )
      } else {
        data.frame(
          component = character(0),
          consistency = numeric(0),
          coverage = numeric(0),
          stringsAsFactors = FALSE
        )
      }

      term_hetero <- if (length(term_only) > 0L) {
        do.call(rbind, lapply(term_only, function(nm) {
          dfx <- subset(clusters_df, component == nm)
          sc <- .summarize_delta(dfx$delta_consistency)
          sv <- .summarize_delta(dfx$delta_coverage)
          data.frame(
            component = nm,
            max_abs_delta_consistency = unname(sc["max_abs"]),
            max_abs_delta_coverage = unname(sv["max_abs"]),
            mean_abs_delta_consistency = unname(sc["mean_abs"]),
            mean_abs_delta_coverage = unname(sv["mean_abs"]),
            stringsAsFactors = FALSE
          )
        }))
      } else {
        data.frame(
          component = character(0),
          max_abs_delta_consistency = numeric(0),
          max_abs_delta_coverage = numeric(0),
          mean_abs_delta_consistency = numeric(0),
          mean_abs_delta_coverage = numeric(0),
          stringsAsFactors = FALSE
        )
      }

      by_cluster[[target_key]] <- list(
        solution_type = rg,
        configuration_key = target_key,
        solution_expression = ttg$signature,
        terms = ttg$terms,
        pooled = list(
          solution = pooled_solution_df,
          terms = pooled_terms_df
        ),
        clusters = clusters_df,
        heterogeneity = list(
          solution = data.frame(
            max_abs_delta_consistency = unname(sum_cl_cons["max_abs"]),
            max_abs_delta_coverage = unname(sum_cl_cov["max_abs"]),
            mean_abs_delta_consistency = unname(sum_cl_cons["mean_abs"]),
            mean_abs_delta_coverage = unname(sum_cl_cov["mean_abs"]),
            stringsAsFactors = FALSE
          ),
          terms = term_hetero
        )
      )

      within_available <- !is.null(units_df)
      n_units_repeated <- if (within_available) length(unique(units_df$unit_id)) else 0L

      if (within_available) {
        within_sol_df <- subset(units_df, component == "solution")
        sum_wi_cons <- .summarize_delta(within_sol_df$delta_consistency)
        sum_wi_cov  <- .summarize_delta(within_sol_df$delta_coverage)

        term_hetero_u <- if (length(term_only) > 0L) {
          do.call(rbind, lapply(term_only, function(nm) {
            dfx <- subset(units_df, component == nm)
            sc <- .summarize_delta(dfx$delta_consistency)
            sv <- .summarize_delta(dfx$delta_coverage)
            data.frame(
              component = nm,
              max_abs_delta_consistency = unname(sc["max_abs"]),
              max_abs_delta_coverage = unname(sv["max_abs"]),
              mean_abs_delta_consistency = unname(sc["mean_abs"]),
              mean_abs_delta_coverage = unname(sv["mean_abs"]),
              stringsAsFactors = FALSE
            )
          }))
        } else {
          data.frame(
            component = character(0),
            max_abs_delta_consistency = numeric(0),
            max_abs_delta_coverage = numeric(0),
            mean_abs_delta_consistency = numeric(0),
            mean_abs_delta_coverage = numeric(0),
            stringsAsFactors = FALSE
          )
        }

        by_unit[[target_key]] <- list(
          solution_type = rg,
          configuration_key = target_key,
          solution_expression = ttg$signature,
          terms = ttg$terms,
          units = units_df,
          heterogeneity = list(
            solution = data.frame(
              max_abs_delta_consistency = unname(sum_wi_cons["max_abs"]),
              max_abs_delta_coverage = unname(sum_wi_cov["max_abs"]),
              mean_abs_delta_consistency = unname(sum_wi_cons["mean_abs"]),
              mean_abs_delta_coverage = unname(sum_wi_cov["mean_abs"]),
              stringsAsFactors = FALSE
            ),
            terms = term_hetero_u
          )
        )
      } else {
        sum_wi_cons <- c(max_abs = NA_real_, mean_abs = NA_real_)
        sum_wi_cov  <- c(max_abs = NA_real_, mean_abs = NA_real_)
      }

      diagnostics_rows[[length(diagnostics_rows) + 1L]] <- data.frame(
        solution_type = rg,
        i_mode = if (identical(rg, "intermediate")) i_mode else NA_character_,
        which_M = which_M,
        necessity = necessity,
        configuration_key = target_key,
        component_count = component_count,
        status = "ok",
        error_source = NA_character_,
        error_message = NA_character_,
        pooled_consistency = unname(pooled_terms[["solution"]]["consistency"]),
        pooled_coverage = unname(pooled_terms[["solution"]]["coverage"]),
        n_clusters = length(unique(stats::na.omit(cluster_vec))),
        max_abs_delta_consistency = unname(sum_cl_cons["max_abs"]),
        max_abs_delta_coverage = unname(sum_cl_cov["max_abs"]),
        mean_abs_delta_consistency = unname(sum_cl_cons["mean_abs"]),
        mean_abs_delta_coverage = unname(sum_cl_cov["mean_abs"]),
        worst_cluster_consistency_id = if (is.na(worst_cons_idx)) NA_character_ else as.character(sol_cluster$cluster_id[worst_cons_idx]),
        worst_cluster_coverage_id = if (is.na(worst_cov_idx)) NA_character_ else as.character(sol_cluster$cluster_id[worst_cov_idx]),
        worst_cluster_consistency = if (is.na(worst_cons_idx)) NA_real_ else as.numeric(sol_cluster$consistency[worst_cons_idx]),
        worst_cluster_coverage = if (is.na(worst_cov_idx)) NA_real_ else as.numeric(sol_cluster$coverage[worst_cov_idx]),
        within_available = within_available,
        n_units_repeated = n_units_repeated,
        max_abs_within_delta_consistency = unname(sum_wi_cons["max_abs"]),
        max_abs_within_delta_coverage = unname(sum_wi_cov["max_abs"]),
        mean_abs_within_delta_consistency = unname(sum_wi_cons["mean_abs"]),
        mean_abs_within_delta_coverage = unname(sum_wi_cov["mean_abs"]),
        stringsAsFactors = FALSE
      )

      targets[[target_key]] <- c(
        ttg,
        list(
          outcome_membership = recoded_outcome,
          pooled_consistency = unname(pooled_terms[["solution"]]["consistency"]),
          pooled_coverage = unname(pooled_terms[["solution"]]["coverage"])
        )
      )
    }

    .bump_pb()
  }

  diagnostics <- .bind_rows_result(diagnostics_rows)
  if (length(by_unit) == 0L) by_unit <- NULL
  baseline_configurations <- lapply(targets, function(x) {
    out <- x
    if ("target_key" %in% names(out)) {
      out$configuration_key <- out$target_key
      out$target_key <- NULL
    }
    if ("signature" %in% names(out)) {
      out$solution_expression <- out$signature
      out$signature <- NULL
    }
    out
  })

  .new_result_object(
    "cluster_test",
    diagnostics = diagnostics,
    results = .make_cluster_results(diagnostics, by_cluster, by_unit),
    baseline = list(
      tt = tt,
      res = baseline_res,
      solution_terms = baseline_sig,
      meta = baseline_meta,
      configurations = baseline_configurations
    ),
    by_cluster = by_cluster,
    by_unit = by_unit,
    settings = list(
      outcome = outcome_name,
      conditions = conditions,
      cluster_id = cluster_id,
      unit_id = unit_id,
      solution = solution,
      monitored_solutions = monitored_solutions,
      include = include,
      dir.exp = dir.exp,
      exclude = exclude,
      which_M = which_M,
      i_mode = i_mode,
      necessity = necessity,
      progress = progress
    )
  )
}

.cluster_print_number <- function(x, digits = 3L) {
  out <- rep(NA_character_, length(x))
  ok <- !is.na(x)
  out[ok] <- formatC(as.numeric(x[ok]), format = "f", digits = digits)
  out
}

.cluster_print_delta <- function(x, digits = 3L) {
  out <- rep(NA_character_, length(x))
  ok <- !is.na(x)
  out[ok] <- sprintf(paste0("%+.", digits, "f"), as.numeric(x[ok]))
  out
}

.cluster_print_has_columns <- function(x, cols) {
  is.data.frame(x) && all(cols %in% names(x))
}

.cluster_overview_for_print <- function(overview) {
  needed <- c(
    "solution_type",
    "configuration",
    "status",
    "pooled_consistency",
    "pooled_coverage",
    "worst_cluster_consistency_id",
    "max_abs_cluster_delta_consistency",
    "worst_cluster_coverage_id",
    "max_abs_cluster_delta_coverage",
    "within_available",
    "n_units_repeated"
  )
  if (!.cluster_print_has_columns(overview, needed)) {
    return(overview)
  }

  within_units <- ifelse(
    !is.na(overview$within_available) & overview$within_available,
    paste0("available (", overview$n_units_repeated, " units)"),
    "not available"
  )

  data.frame(
    solution_type = overview$solution_type,
    configuration = overview$configuration,
    status = overview$status,
    pooled_cons = .cluster_print_number(overview$pooled_consistency),
    pooled_cov = .cluster_print_number(overview$pooled_coverage),
    worst_consistency_cluster = overview$worst_cluster_consistency_id,
    max_delta_cons = .cluster_print_number(overview$max_abs_cluster_delta_consistency),
    worst_coverage_cluster = overview$worst_cluster_coverage_id,
    max_delta_cov = .cluster_print_number(overview$max_abs_cluster_delta_coverage),
    within_units = within_units,
    stringsAsFactors = FALSE
  )
}

.cluster_details_for_print <- function(results, overview) {
  clusters <- results$clusters
  needed <- c(
    "solution_type",
    "configuration",
    "component",
    "cluster_id",
    "cluster_size",
    "consistency",
    "delta_consistency",
    "coverage",
    "delta_coverage"
  )
  if (!.cluster_print_has_columns(overview, c("solution_type", "configuration")) ||
      nrow(overview) != 1L ||
      !.cluster_print_has_columns(clusters, needed)) {
    return(NULL)
  }

  keep <- clusters$solution_type == overview$solution_type[1L] &
    clusters$configuration == overview$configuration[1L] &
    clusters$component == "solution"
  clusters <- clusters[keep, , drop = FALSE]
  if (nrow(clusters) == 0L) {
    return(NULL)
  }

  data.frame(
    cluster_id = clusters$cluster_id,
    n = clusters$cluster_size,
    consistency = .cluster_print_number(clusters$consistency),
    delta_consistency = .cluster_print_delta(clusters$delta_consistency),
    coverage = .cluster_print_number(clusters$coverage),
    delta_coverage = .cluster_print_delta(clusters$delta_coverage),
    stringsAsFactors = FALSE
  )
}

.cluster_units_for_print <- function(results, overview) {
  units <- results$units
  needed <- c(
    "solution_type",
    "configuration",
    "component",
    "unit_id",
    "n_clusters",
    "consistency",
    "delta_consistency",
    "coverage",
    "delta_coverage"
  )
  if (!.cluster_print_has_columns(overview, c("solution_type", "configuration")) ||
      nrow(overview) != 1L ||
      !.cluster_print_has_columns(units, needed)) {
    return(NULL)
  }

  keep <- units$solution_type == overview$solution_type[1L] &
    units$configuration == overview$configuration[1L] &
    units$component == "solution"
  units <- units[keep, , drop = FALSE]
  if (nrow(units) == 0L) {
    return(NULL)
  }

  data.frame(
    unit_id = units$unit_id,
    n_clusters = units$n_clusters,
    consistency = .cluster_print_number(units$consistency),
    delta_consistency = .cluster_print_delta(units$delta_consistency),
    coverage = .cluster_print_number(units$coverage),
    delta_coverage = .cluster_print_delta(units$delta_coverage),
    stringsAsFactors = FALSE
  )
}

.cluster_within_unavailable_message <- function(overview, settings) {
  if (is.null(settings$unit_id) || is.na(settings$unit_id)) {
    return("Not available: no unit id was supplied.")
  }

  if (.cluster_print_has_columns(overview, "n_units_repeated") && nrow(overview) == 1L) {
    n_units <- overview$n_units_repeated[1L]
    return(paste0(
      "Not available: ",
      n_units,
      " units are observed in more than one cluster."
    ))
  }

  "Not available."
}

#' Print a `cluster_test` object
#'
#' @rdname cluster.test
#' @export
print.cluster_test <- function(x, row.names = FALSE, ...) {
  results <- x$results
  overview <- results$overview
  settings <- x$settings

  .print_qcaert_heading("cluster_test", "cluster heterogeneity", settings)
  if (!is.null(settings$cluster_id)) {
    cat("Cluster id: ", settings$cluster_id, "\n", sep = "")
  }
  if (!is.null(settings$unit_id) && !is.na(settings$unit_id)) {
    cat("Unit id: ", settings$unit_id, "\n", sep = "")
  }
  if (!is.null(settings$necessity)) {
    cat("Necessity mode: ", settings$necessity, "\n", sep = "")
  }

  .print_qcaert_table(.cluster_overview_for_print(overview), "Overview", row.names = row.names, ...)

  if (is.data.frame(overview) && nrow(overview) == 1L) {
    cluster_details <- .cluster_details_for_print(results, overview)
    if (!is.null(cluster_details)) {
      .print_qcaert_table(cluster_details, "Cluster details", row.names = row.names, ...)
    }

    cat("\nWithin-unit diagnostics\n")
    unit_details <- .cluster_units_for_print(results, overview)
    if (!is.null(unit_details)) {
      print(unit_details, row.names = row.names, ...)
    } else {
      cat(" ", .cluster_within_unavailable_message(overview, settings), "\n", sep = "")
    }
  }

  cat("\nSummary\n")
  if (is.data.frame(overview) && nrow(overview) > 0L) {
    n_ok <- sum(overview$status == "ok", na.rm = TRUE)
    n_errors <- sum(!is.na(overview$status) & overview$status != "ok", na.rm = TRUE)

    cat(" Configurations evaluated: ", nrow(overview), "\n", sep = "")
    cat(" Successful configurations: ", n_ok, "\n", sep = "")
    cat(" Non-ok configurations: ", n_errors, "\n", sep = "")
  }

  detailed_tables <- "x$results$overview"
  if (!is.null(results$clusters)) {
    detailed_tables <- c(detailed_tables, "x$results$clusters")
  }
  if (!is.null(results$units)) {
    detailed_tables <- c(detailed_tables, "x$results$units")
  }
  cat(" Detailed tables: ", paste(detailed_tables, collapse = ", "), "\n", sep = "")

  invisible(x)
}

#' Return the overview table from a `cluster_test` object
#'
#' @rdname cluster.test
#' @export
as.data.frame.cluster_test <- function(x, ...) {
  .as.data.frame_cluster_overview(x, ...)
}

test_that("public functions reject the outcome as a condition", {
  skip_if_not_installed("QCA")

  fixture <- qcaert_fixture_direct6()
  bad_conditions <- c(fixture$conditions, fixture$outcome)
  msg <- "`outcome` must not be included in `conditions`"

  expect_error(
    incl.test(
      data = fixture$calib,
      outcome = fixture$outcome,
      conditions = bad_conditions,
      progress = FALSE
    ),
    msg,
    fixed = TRUE
  )

  expect_error(
    ncut.test(
      data = fixture$calib,
      outcome = fixture$outcome,
      conditions = bad_conditions,
      progress = FALSE
    ),
    msg,
    fixed = TRUE
  )

  expect_error(
    calib.test(
      raw.data = fixture$raw,
      calib.data = fixture$calib,
      outcome = fixture$outcome,
      conditions = bad_conditions,
      progress = FALSE
    ),
    msg,
    fixed = TRUE
  )

  expect_error(
    altset.test(
      raw.data = fixture$raw,
      calib.data = fixture$calib,
      outcome = fixture$outcome,
      conditions = bad_conditions,
      progress = FALSE
    ),
    msg,
    fixed = TRUE
  )

  expect_error(
    loo.test(
      data = fixture$calib,
      outcome = fixture$outcome,
      conditions = bad_conditions,
      progress = FALSE
    ),
    msg,
    fixed = TRUE
  )

  expect_error(
    subsample.test(
      data = fixture$calib,
      outcome = fixture$outcome,
      conditions = bad_conditions,
      sample_prop = 0.8,
      progress = FALSE
    ),
    msg,
    fixed = TRUE
  )
})

test_that("cluster.test rejects truth tables whose outcome is also a condition", {
  skip_if_not_installed("QCA")

  fixture <- qcaert_fixture_cluster()
  bad_tt <- fixture$truth_table
  bad_tt$options$conditions <- c(fixture$conditions, fixture$outcome)

  expect_error(
    cluster.test(
      data = fixture$data,
      tt = bad_tt,
      cluster_id = fixture$cluster_id,
      unit_id = fixture$unit_id,
      progress = FALSE
    ),
    "`outcome` must not be included in `conditions`",
    fixed = TRUE
  )
})

test_that("boundary robustness result schemas are frozen", {
  skip_if_not_installed("QCA")

  dat <- qcaert_schema_calib3()

  boundary_diag <- c(
    "direction",
    "solution",
    "monitored_solutions",
    "i_mode",
    "which_M",
    "incl.cut_start",
    "incl.cut_last_safe",
    "incl.cut_first_failing",
    "number_of_steps",
    "total_delta",
    "stop_reason",
    "changed_types",
    "change_kind",
    "error_source",
    "error_message",
    "n_exclude_baseline",
    "n_exclude_last_safe",
    "n_exclude_first_failing",
    "exclude_baseline",
    "exclude_last_safe",
    "exclude_first_failing"
  )
  boundary_results <- c(
    "direction",
    "start",
    "last_safe",
    "first_failing",
    "steps",
    "total_delta",
    "reason"
  )

  incl <- suppressWarnings(incl.test(
    data = dat,
    outcome = "Y",
    conditions = c("A", "B"),
    incl.cut = 0.75,
    step = 0.1,
    max_steps = 1,
    n.cut = 1,
    solution = "conservative",
    progress = FALSE
  ))

  expect_qcaert_schema(
    incl,
    class = "incl_test",
    top = c("diagnostics", "results", "bounds", "baseline", "by_direction", "settings"),
    diagnostics = boundary_diag,
    results = boundary_results,
    settings = c(
      "outcome",
      "conditions",
      "incl.cut",
      "step",
      "max_steps",
      "n.cut",
      "solution",
      "monitored_solutions",
      "include",
      "dir.exp",
      "which_M",
      "i_mode",
      "exclude_mode",
      "exclude_recompute",
      "exclude_static",
      "result_shape"
    )
  )
  expect_identical(names(incl$by_direction), c("lower", "upper"))
  expect_identical(incl$results$direction, c("lower", "upper"))
  expect_identical(nrow(incl$diagnostics), 2L)

  ncut_diag <- boundary_diag
  ncut_diag[6:8] <- c("n.cut_start", "n.cut_last_safe", "n.cut_first_failing")
  ncut_diag <- c(ncut_diag, "upper_limit")

  ncut <- suppressWarnings(ncut.test(
    data = dat,
    outcome = "Y",
    conditions = c("A", "B"),
    n.cut = 1,
    step = 1,
    max_steps = 1,
    incl.cut = 0.75,
    solution = "conservative",
    progress = FALSE
  ))

  expect_qcaert_schema(
    ncut,
    class = "ncut_test",
    top = c("diagnostics", "results", "bounds", "baseline", "by_direction", "settings"),
    diagnostics = ncut_diag,
    results = boundary_results,
    settings = c(
      "outcome",
      "conditions",
      "n.cut",
      "step",
      "max_steps",
      "incl.cut",
      "solution",
      "monitored_solutions",
      "include",
      "dir.exp",
      "which_M",
      "i_mode",
      "exclude_mode",
      "exclude_recompute",
      "exclude_static",
      "result_shape",
      "upper_limit"
    )
  )
  expect_identical(names(ncut$by_direction), c("lower", "upper"))
  expect_identical(ncut$results$direction, c("lower", "upper"))
  expect_identical(nrow(ncut$diagnostics), 2L)
})

test_that("calibration robustness result schemas are frozen", {
  skip_if_not_installed("QCA")

  raw <- qcaert_schema_raw()
  calib6 <- qcaert_schema_calib6(raw)
  spec6 <- qcaert_schema_calib6_spec()

  calib_diag <- c(
    "set",
    "role",
    "raw",
    "type",
    "method",
    "anchor",
    "direction",
    "solution",
    "monitored_solutions",
    "which_M_tested",
    "n_M_conservative_baseline",
    "n_M_parsimonious_baseline",
    "n_M_intermediate_min_baseline",
    "step_unit_used",
    "start_value",
    "last_safe_value",
    "failing_value",
    "number_of_steps",
    "total_delta_units",
    "delta_as_pct_of_raw_range",
    "stop_reason",
    "changed_types",
    "change_kind",
    "error_source",
    "error_message"
  )
  calib_results <- c(
    "set",
    "role",
    "raw",
    "type",
    "method",
    "anchor",
    "direction",
    "start",
    "last_safe",
    "first_failing",
    "step_unit",
    "steps",
    "total_delta",
    "pct_raw_range",
    "reason"
  )
  calib_settings <- c(
    "outcome",
    "conditions",
    "test.conditions",
    "test.outcome",
    "calib_spec",
    "anchors_to_test",
    "solution",
    "monitored_solutions",
    "include",
    "which_M",
    "unit_step",
    "unit_step_divisor",
    "max_steps",
    "incl.cut",
    "n.cut",
    "dir.exp",
    "i_mode",
    "exclude_mode",
    "exclude_recompute",
    "exclude_static",
    "result_shape"
  )

  calib <- suppressWarnings(calib.test(
    raw.data = raw,
    calib.data = calib6,
    outcome = "Y",
    conditions = c("A", "B"),
    calib_spec = spec6,
    test.conditions = "A",
    unit_step = 1,
    max_steps = 1,
    incl.cut = 0.75,
    n.cut = 1,
    solution = "conservative",
    progress = FALSE
  ))

  expect_qcaert_schema(
    calib,
    class = "calib_test",
    top = c("diagnostics", "results", "bounds", "baseline", "by_set", "settings"),
    diagnostics = calib_diag,
    results = calib_results,
    settings = calib_settings
  )
  expect_identical(unique(calib$results$anchor), c("E1", "C1", "I1", "I2", "C2", "E2"))
  expect_identical(unique(calib$results$role), "condition")
  expect_identical(unique(calib$results$method), "direct")
  expect_identical(calib$baseline$status, "ok")
  expect_identical(names(calib$baseline$by_solution_type), "conservative")
  expect_identical(nrow(calib$results), 12L)

  indirect <- suppressWarnings(calib.test(
    raw.data = raw,
    calib.data = qcaert_schema_calib3(raw),
    outcome = "Y",
    conditions = c("A", "B"),
    calib_spec = qcaert_schema_indirect_spec(),
    test.conditions = "A",
    unit_step = 1,
    max_steps = 1,
    incl.cut = 0.75,
    n.cut = 1,
    solution = "conservative",
    progress = FALSE
  ))

  expect_identical(unique(indirect$results$anchor), c("T1", "T2", "T3"))
  expect_identical(unique(indirect$results$method), "indirect")
  expect_identical(nrow(indirect$results), 6L)
})

test_that("alternative-set robustness result schemas are frozen", {
  skip_if_not_installed("QCA")

  raw <- qcaert_schema_raw()
  calib6 <- qcaert_schema_calib6(raw)

  alt <- suppressWarnings(altset.test(
    raw.data = raw,
    calib.data = calib6,
    outcome = "Y",
    conditions = c("A", "B"),
    calib_spec = qcaert_schema_calib6_spec(),
    test.conditions = "A",
    unit_step = 1,
    calib_max_steps = 1,
    incl.cut = 0.75,
    incl_step = 0.1,
    incl_max_steps = 1,
    n.cut = 1,
    ncut_step = 1,
    ncut_max_steps = 1,
    n_draws = 2,
    seed = 1,
    solution = "conservative",
    progress = FALSE
  ))

  expect_qcaert_schema(
    alt,
    class = "altset_test",
    top = c("diagnostics", "results", "summary", "baseline", "by_draw", "settings"),
    diagnostics = c(
      "draw",
      "incl.cut",
      "n.cut",
      "status",
      "changed",
      "solution_changed",
      "changed_types",
      "change_kind",
      "fit_compared",
      "fit_changed",
      "fit_changed_types",
      "n_fit_deltas",
      "max_abs_fit_delta",
      "error_source",
      "error_message",
      "changed_sets",
      "changed_roles",
      "n_changed_sets",
      "exclude_used",
      "calibration",
      "status_conservative",
      "error_source_conservative",
      "error_message_conservative",
      "solution_changed_conservative",
      "fit_compared_conservative",
      "fit_changed_conservative",
      "max_abs_fit_delta_conservative",
      "matched_fit_keys_conservative"
    ),
    results = c(
      "draw",
      "incl.cut",
      "n.cut",
      "status",
      "n_changed_sets",
      "changed_sets",
      "changed_roles",
      "solution_change",
      "fit_changed_types",
      "n_fit_deltas",
      "max_abs_fit_delta"
    ),
    settings = c(
      "outcome",
      "conditions",
      "test.conditions",
      "test.outcome",
      "calib_spec",
      "anchors_to_test",
      "solution",
      "monitored_solutions",
      "include",
      "which_M",
      "unit_step",
      "unit_step_divisor",
      "calib_max_steps",
      "incl.cut",
      "incl_step",
      "incl_max_steps",
      "n.cut",
      "ncut_step",
      "ncut_max_steps",
      "dir.exp",
      "i_mode",
      "exclude_mode",
      "exclude_recompute",
      "exclude_static",
      "n_draws",
      "fit_tol",
      "seed"
    )
  )
  expect_identical(nrow(alt$diagnostics), 2L)
  expect_identical(names(alt$by_draw), NULL)
  expect_identical(names(alt$baseline$draw_meta$calibration$A$delta_steps), c("E1", "C1", "I1", "I2", "C2", "E2"))
  expect_identical(names(alt$by_draw[[1]]$result$draw_meta$calibration$A$delta_steps), c("E1", "C1", "I1", "I2", "C2", "E2"))
})

test_that("case-deletion and subsample result schemas are frozen", {
  skip_if_not_installed("QCA")

  dat <- qcaert_schema_calib3()
  case_diag <- c(
    "row_index",
    "case_label",
    "status",
    "changed",
    "solution_changed",
    "changed_types",
    "change_kind",
    "fit_changed",
    "fit_changed_types",
    "n_fit_deltas",
    "max_abs_fit_delta",
    "error_source",
    "error_message"
  )
  case_results <- c(
    "row_index",
    "case_label",
    "status",
    "solution_change",
    "fit_changed_types",
    "n_fit_deltas",
    "max_abs_fit_delta"
  )

  loo <- suppressWarnings(loo.test(
    data = dat,
    outcome = "Y",
    conditions = c("A", "B"),
    cases = 1:3,
    incl.cut = 0.75,
    n.cut = 1,
    solution = "conservative",
    progress = FALSE
  ))

  expect_qcaert_schema(
    loo,
    class = "loo_test",
    top = c("diagnostics", "results", "baseline", "by_case", "settings"),
    diagnostics = case_diag,
    results = case_results,
    settings = c(
      "outcome",
      "conditions",
      "cases",
      "case_labels",
      "calib",
      "n_cases_tested",
      "incl.cut",
      "n.cut",
      "solution",
      "monitored_solutions",
      "include",
      "dir.exp",
      "which_M",
      "i_mode",
      "exclude_mode",
      "exclude_recompute",
      "exclude_static",
      "fit_measures",
      "fit_tol"
    )
  )
  expect_identical(nrow(loo$results), 3L)
  expect_identical(names(loo$by_case), c("1:1", "2:2", "3:3"))

  subsample <- suppressWarnings(subsample.test(
    data = dat,
    outcome = "Y",
    conditions = c("A", "B"),
    sample_n = 8,
    reps = 2,
    seed = 1,
    incl.cut = 0.75,
    n.cut = 1,
    solution = "conservative",
    progress = FALSE
  ))

  expect_qcaert_schema(
    subsample,
    class = "subsample_test",
    top = c("diagnostics", "results", "summary", "baseline", "by_run", "settings"),
    diagnostics = c(
      "rep",
      "status",
      "n_sample",
      "n_holdout",
      "changed",
      "solution_changed",
      "changed_types",
      "change_kind",
      "fit_changed",
      "fit_changed_types",
      "n_fit_deltas",
      "max_abs_fit_delta",
      "exact_match_baseline",
      "term_jaccard_baseline",
      "error_source",
      "error_message"
    ),
    results = c(
      "rep",
      "n_sample",
      "n_holdout",
      "status",
      "exact_match_baseline",
      "term_jaccard_baseline",
      "solution_change",
      "fit_changed_types",
      "n_fit_deltas",
      "max_abs_fit_delta"
    ),
    settings = c(
      "outcome",
      "conditions",
      "calib",
      "sample_n",
      "sample_prop_requested",
      "sample_prop_realized",
      "reps",
      "stratify",
      "incl.cut",
      "n.cut",
      "solution",
      "monitored_solutions",
      "include",
      "dir.exp",
      "which_M",
      "i_mode",
      "exclude_mode",
      "exclude_recompute",
      "exclude_static",
      "fit_measures",
      "fit_tol",
      "seed"
    )
  )
  expect_identical(nrow(subsample$results), 2L)
  expect_identical(names(subsample$by_run), c("R1", "R2"))
  expect_setequal(
    names(subsample$summary),
    c("exact_solution", "term_stability", "fit_stability", "similarity", "calibration")
  )
})

test_that("cluster and sol.df result schemas are frozen", {
  skip_if_not_installed("QCA")

  raw <- qcaert_schema_raw()
  dat <- qcaert_schema_calib6(raw)
  tt <- suppressWarnings(QCA::truthTable(
    dat,
    outcome = "Y",
    conditions = c("A", "B"),
    incl.cut = 0.75,
    n.cut = 1
  ))

  cluster <- suppressWarnings(cluster.test(
    data = dat,
    tt = tt,
    cluster_id = "cluster",
    unit_id = "unit",
    solution = "conservative",
    progress = FALSE
  ))

  expect_qcaert_schema(
    cluster,
    class = "cluster_test",
    top = c("diagnostics", "results", "baseline", "by_cluster", "by_unit", "settings"),
    diagnostics = c(
      "solution_type",
      "i_mode",
      "which_M",
      "necessity",
      "configuration_key",
      "component_count",
      "status",
      "error_source",
      "error_message",
      "pooled_consistency",
      "pooled_coverage",
      "n_clusters",
      "max_abs_delta_consistency",
      "max_abs_delta_coverage",
      "mean_abs_delta_consistency",
      "mean_abs_delta_coverage",
      "worst_cluster_consistency_id",
      "worst_cluster_coverage_id",
      "worst_cluster_consistency",
      "worst_cluster_coverage",
      "within_available",
      "n_units_repeated",
      "max_abs_within_delta_consistency",
      "max_abs_within_delta_coverage",
      "mean_abs_within_delta_consistency",
      "mean_abs_within_delta_coverage"
    ),
    results = c("overview", "clusters", "units"),
    settings = c(
      "outcome",
      "conditions",
      "cluster_id",
      "unit_id",
      "solution",
      "monitored_solutions",
      "include",
      "dir.exp",
      "exclude",
      "which_M",
      "i_mode",
      "necessity",
      "progress"
    )
  )
  expect_identical(names(as.data.frame(cluster)), names(cluster$results$overview))
  expect_identical(names(cluster$results), c("overview", "clusters", "units"))

  sol <- suppressWarnings(QCA::minimize(tt, include = ""))
  sol_table <- sol.df(conservative = sol, solution = "conservative")

  expect_identical(class(sol_table), "data.frame")
  expect_identical(
    names(sol_table),
    c(
      "Solution",
      "Model",
      "Intermediate_CnPn",
      "Prime_Implicants",
      "Consistency_PI",
      "PRI_PI",
      "Raw_Coverage_PI",
      "Unique_Coverage_PI",
      "Solution_Consistency",
      "Solution_PRI",
      "Solution_Coverage",
      "Cases"
    )
  )
  expect_true(nrow(sol_table) >= 1L)
})

test_that("regular invariant helper validates the common returned-object structure", {
  skip_if_not_installed("QCA")

  dat <- qcaert_schema_calib3()
  out <- suppressWarnings(incl.test(
    data = dat,
    outcome = "Y",
    conditions = c("A", "B"),
    incl.cut = 0.75,
    step = 0.1,
    max_steps = 1,
    n.cut = 1,
    solution = "conservative",
    progress = FALSE
  ))

  expect_qcaert_result_structure(
    out,
    class = "incl_test",
    diagnostics = c("direction", "solution", "stop_reason", "change_kind"),
    results = c("direction", "start", "last_safe", "reason"),
    settings = c("outcome", "conditions", "solution", "exclude_recompute", "exclude_static"),
    top = c("bounds", "baseline", "by_direction", "settings")
  )
})

test_that("cluster invariant helper validates the explicit cluster exception", {
  skip_if_not_installed("QCA")

  fixture <- qcaert_fixture_cluster()
  out <- suppressWarnings(cluster.test(
    data = fixture$data,
    tt = fixture$truth_table,
    cluster_id = fixture$cluster_id,
    unit_id = fixture$unit_id,
    solution = "conservative",
    progress = FALSE
  ))

  expect_qcaert_cluster_structure(
    out,
    diagnostics = c("solution_type", "configuration_key", "status", "within_available"),
    overview = c("solution_type", "configuration", "status", "within_available"),
    clusters = c("solution_type", "configuration", "component", "cluster_id"),
    units = c("solution_type", "configuration", "component", "unit_id"),
    settings = c("outcome", "conditions", "solution", "cluster_id", "unit_id"),
    top = c("baseline", "by_cluster", "by_unit", "settings")
  )
})

test_that("exclusion-control formals use the common recompute/static conventions", {
  shared <- c(
    "incl.test",
    "ncut.test",
    "calib.test",
    "loo.test",
    "subsample.test",
    "altset.test",
    "theory.test"
  )

  for (fn in shared) {
    args <- names(formals(get(fn, envir = asNamespace("qcaERT"))))
    expect_true(all(c("exclude_mode", "exclude_recompute", "exclude_static") %in% args), info = fn)
    expect_false("exclude" %in% args, info = fn)
    expect_false("exclude_spec" %in% args, info = fn)
  }

  cluster_args <- names(formals(cluster.test))
  expect_true("exclude" %in% cluster_args)
  expect_false("exclude_mode" %in% cluster_args)
  expect_false("exclude_recompute" %in% cluster_args)
  expect_false("exclude_static" %in% cluster_args)
})

test_that("shared exclusion-control validator rejects crossed wires", {
  expect_error(
    .validate_exclusion_controls(
      exclude_mode = "none",
      exclude_recompute = list(type = 2),
      exclude_static = NULL,
      exclude_recompute_supplied = TRUE,
      monitored_solutions = "parsimonious"
    ),
    "cannot be combined",
    fixed = TRUE
  )

  expect_error(
    .validate_exclusion_controls(
      exclude_mode = "static",
      exclude_recompute = list(type = 2),
      exclude_static = 1:2,
      exclude_recompute_supplied = TRUE,
      monitored_solutions = "parsimonious"
    ),
    "uses `exclude_static`, not `exclude_recompute`",
    fixed = TRUE
  )

  expect_error(
    .validate_exclusion_controls(
      exclude_mode = "recompute",
      exclude_recompute = list(type = 2),
      exclude_static = 1:2,
      monitored_solutions = "parsimonious"
    ),
    "uses `exclude_recompute`, not `exclude_static`",
    fixed = TRUE
  )
})

test_that("incl.test runs end to end on the shared direct-six fixture", {
  skip_if_not_installed("QCA")

  fixture <- qcaert_fixture_direct6()
  out <- suppressWarnings(incl.test(
    data = fixture$calib,
    outcome = fixture$outcome,
    conditions = fixture$conditions,
    incl.cut = 0.75,
    step = 0.1,
    max_steps = 1,
    n.cut = 1,
    solution = "all",
    dir.exp = fixture$dir.exp,
    progress = FALSE
  ))

  expect_qcaert_result_structure(
    out,
    class = "incl_test",
    diagnostics = c("direction", "solution_type", "monitored_solutions", "change_kind", "stop_reason"),
    results = c("direction", "start", "con_last_safe", "par_last_safe", "int_last_safe"),
    settings = c("outcome", "conditions", "solution", "monitored_solutions"),
    top = c("bounds", "baseline", "by_direction", "settings")
  )
  expect_identical(out$results$direction, c("lower", "upper"))
  expect_setequal(
    out$settings$monitored_solutions,
    c("conservative", "parsimonious", "intermediate")
  )
})

test_that("ncut.test runs end to end on the shared direct-six fixture", {
  skip_if_not_installed("QCA")

  fixture <- qcaert_fixture_direct6()
  out <- suppressWarnings(ncut.test(
    data = fixture$calib,
    outcome = fixture$outcome,
    conditions = fixture$conditions,
    n.cut = 1,
    step = 1,
    max_steps = 1,
    incl.cut = 0.75,
    solution = "all",
    dir.exp = fixture$dir.exp,
    progress = FALSE
  ))

  expect_qcaert_result_structure(
    out,
    class = "ncut_test",
    diagnostics = c("direction", "solution_type", "monitored_solutions", "change_kind", "stop_reason", "upper_limit"),
    results = c("direction", "start", "con_last_safe", "par_last_safe", "int_last_safe"),
    settings = c("outcome", "conditions", "solution", "monitored_solutions", "upper_limit"),
    top = c("bounds", "baseline", "by_direction", "settings")
  )
  expect_identical(out$results$direction, c("lower", "upper"))
  expect_true(out$settings$upper_limit >= out$settings$n.cut)
})

test_that("calib.test runs end to end for indirect calibration", {
  skip_if_not_installed("QCA")

  fixture <- qcaert_fixture_indirect()
  out <- suppressWarnings(calib.test(
    raw.data = fixture$raw,
    calib.data = fixture$calib,
    outcome = fixture$outcome,
    conditions = fixture$conditions,
    calib_spec = fixture$calib_spec,
    test.conditions = fixture$test.conditions,
    unit_step = 1,
    max_steps = 1,
    incl.cut = 0.75,
    n.cut = 1,
    solution = "conservative",
    progress = FALSE
  ))

  expect_qcaert_result_structure(
    out,
    class = "calib_test",
    diagnostics = c("set", "role", "method", "anchor", "direction", "change_kind"),
    results = c("set", "role", "method", "anchor", "direction", "start", "reason"),
    settings = c("outcome", "conditions", "calib_spec", "solution", "test.outcome"),
    top = c("bounds", "baseline", "by_set", "settings")
  )
  expect_identical(unique(out$results$method), "indirect")
  expect_identical(unique(out$results$anchor), c("T1", "T2", "T3"))
})

test_that("altset.test runs end to end with deterministic draws", {
  skip_if_not_installed("QCA")

  fixture <- qcaert_fixture_direct6()
  out <- suppressWarnings(altset.test(
    raw.data = fixture$raw,
    calib.data = fixture$calib,
    outcome = fixture$outcome,
    conditions = fixture$conditions,
    calib_spec = fixture$calib_spec,
    test.conditions = fixture$test.conditions,
    anchors_to_test = c("E1", "C1", "I1"),
    unit_step = 1,
    calib_max_steps = 1,
    incl.cut = 0.75,
    incl_step = 0.1,
    incl_max_steps = 1,
    n.cut = 1,
    ncut_step = 1,
    ncut_max_steps = 1,
    n_draws = 2,
    seed = 101,
    solution = "conservative",
    progress = FALSE
  ))

  expect_qcaert_result_structure(
    out,
    class = "altset_test",
    diagnostics = c("draw", "status", "changed_sets", "changed_roles", "fit_changed_types"),
    results = c("draw", "status", "changed_sets", "changed_roles", "fit_changed_types"),
    settings = c("outcome", "conditions", "test.outcome", "n_draws", "seed", "solution"),
    top = c("summary", "baseline", "by_draw", "settings")
  )
  expect_identical(nrow(out$results), 2L)
  expect_identical(length(out$by_draw), 2L)
  expect_equal(out$settings$seed, 101)
})

test_that("loo.test runs end to end on selected cases", {
  skip_if_not_installed("QCA")

  fixture <- qcaert_fixture_direct6()
  out <- suppressWarnings(loo.test(
    data = fixture$calib,
    outcome = fixture$outcome,
    conditions = fixture$conditions,
    cases = 1:3,
    incl.cut = 0.75,
    n.cut = 1,
    solution = "all",
    dir.exp = fixture$dir.exp,
    progress = FALSE
  ))

  expect_qcaert_result_structure(
    out,
    class = "loo_test",
    diagnostics = c("row_index", "case_label", "solution_changed", "fit_changed_types"),
    results = c("row_index", "case_label", "solution_change", "fit_changed_types"),
    settings = c("outcome", "conditions", "cases", "monitored_solutions"),
    top = c("baseline", "by_case", "settings")
  )
  expect_identical(nrow(out$results), 3L)
  expect_identical(names(out$by_case), c("1:1", "2:2", "3:3"))
})

test_that("subsample.test runs end to end with deterministic fixed calibration", {
  skip_if_not_installed("QCA")

  fixture <- qcaert_fixture_direct6()
  out <- suppressWarnings(subsample.test(
    data = fixture$calib,
    outcome = fixture$outcome,
    conditions = fixture$conditions,
    sample_n = 8,
    reps = 2,
    seed = 202,
    incl.cut = 0.75,
    n.cut = 1,
    solution = "conservative",
    progress = FALSE
  ))

  expect_qcaert_result_structure(
    out,
    class = "subsample_test",
    diagnostics = c("rep", "status", "n_sample", "n_holdout", "fit_changed_types"),
    results = c("rep", "n_sample", "n_holdout", "status", "fit_changed_types"),
    settings = c("outcome", "conditions", "sample_n", "reps", "seed"),
    top = c("summary", "baseline", "by_run", "settings")
  )
  expect_identical(nrow(out$results), 2L)
  expect_identical(names(out$by_run), c("R1", "R2"))
  expect_identical(out$settings$seed, 202L)
})

test_that("cluster.test runs end to end on repeated-unit cluster data", {
  skip_if_not_installed("QCA")

  fixture <- qcaert_fixture_cluster()
  out <- suppressWarnings(cluster.test(
    data = fixture$data,
    tt = fixture$truth_table,
    cluster_id = fixture$cluster_id,
    unit_id = fixture$unit_id,
    solution = "all",
    dir.exp = fixture$dir.exp,
    progress = FALSE
  ))

  expect_qcaert_cluster_structure(
    out,
    diagnostics = c("solution_type", "configuration_key", "status", "within_available"),
    overview = c("solution_type", "configuration", "status", "within_available"),
    clusters = c("solution_type", "configuration", "component", "cluster_id"),
    units = c("solution_type", "configuration", "component", "unit_id"),
    settings = c("outcome", "conditions", "solution", "monitored_solutions"),
    top = c("baseline", "by_cluster", "by_unit", "settings")
  )
  expect_setequal(
    out$settings$monitored_solutions,
    c("conservative", "parsimonious", "intermediate")
  )
})

test_that("sol.df extracts a clean table from QCA minimization objects", {
  skip_if_not_installed("QCA")

  fixture <- qcaert_fixture_direct6()
  tt <- qcaert_truth_table(
    data = fixture$calib,
    outcome = fixture$outcome,
    conditions = fixture$conditions,
    incl.cut = 0.75,
    n.cut = 1
  )
  conservative <- suppressWarnings(QCA::minimize(tt, include = ""))
  intermediate <- suppressWarnings(QCA::minimize(
    tt,
    include = "?",
    dir.exp = fixture$dir.exp
  ))
  parsimonious <- suppressWarnings(QCA::minimize(tt, include = "?"))

  out <- sol.df(
    conservative = conservative,
    intermediate = intermediate,
    parsimonious = parsimonious,
    solution = "all",
    which_M = 1,
    include_cases = TRUE,
    digits = 3
  )

  qcaert_expect_table(
    out,
    "sol.df output",
    required = c(
      "Solution",
      "Model",
      "Prime_Implicants",
      "Solution_Consistency",
      "Solution_Coverage",
      "Cases"
    )
  )
  expect_true(nrow(out) >= 1L)
  expect_setequal(unique(out$Solution), c("Conservative", "Intermediate", "Parsimonious"))
})

test_that("stress: LR QCA solution workflow feeds sol.df", {
  skip_qcaert_stress()
  skip_if_not_installed("QCA")

  lr <- qcaert_fixture_lr_solutions()
  out <- sol.df(
    conservative = lr$conservative,
    intermediate = lr$intermediate,
    parsimonious = lr$parsimonious,
    solution = "all",
    which_M = 1,
    include_cases = TRUE,
    digits = 3
  )

  qcaert_expect_table(
    out,
    "LR sol.df output",
    required = c(
      "Solution",
      "Model",
      "Prime_Implicants",
      "Solution_Consistency",
      "Solution_Coverage",
      "Cases"
    )
  )
  expect_true(nrow(out) >= 3L)
  expect_setequal(unique(out$Solution), c("Conservative", "Intermediate", "Parsimonious"))
  expect_false(any(is.na(out$Prime_Implicants)))
})

test_that("stress: LR boundary and leave-one-out workflows preserve the family structure", {
  skip_qcaert_stress()
  skip_if_not_installed("QCA")

  lr <- qcaert_fixture_lr()

  incl <- suppressWarnings(incl.test(
    data = lr$calib,
    outcome = lr$outcome,
    conditions = lr$conditions,
    incl.cut = 0.8,
    step = 0.05,
    max_steps = 2,
    n.cut = 1,
    solution = "all",
    dir.exp = lr$dir.exp,
    progress = FALSE
  ))
  expect_qcaert_result_structure(
    incl,
    class = "incl_test",
    diagnostics = c("direction", "solution_type", "monitored_solutions", "change_kind"),
    results = c("direction", "start", "con_last_safe", "par_last_safe"),
    settings = c("outcome", "conditions", "solution", "monitored_solutions"),
    top = c("bounds", "baseline", "by_direction", "settings")
  )

  ncut <- suppressWarnings(ncut.test(
    data = lr$calib,
    outcome = lr$outcome,
    conditions = lr$conditions,
    n.cut = 1,
    step = 1,
    max_steps = 2,
    incl.cut = 0.8,
    solution = "all",
    dir.exp = lr$dir.exp,
    progress = FALSE
  ))
  expect_qcaert_result_structure(
    ncut,
    class = "ncut_test",
    diagnostics = c("direction", "solution_type", "monitored_solutions", "change_kind", "upper_limit"),
    results = c("direction", "start", "con_last_safe", "par_last_safe"),
    settings = c("outcome", "conditions", "solution", "monitored_solutions"),
    top = c("bounds", "baseline", "by_direction", "settings")
  )

  loo <- suppressWarnings(loo.test(
    data = lr$calib,
    outcome = lr$outcome,
    conditions = lr$conditions,
    cases = 1:6,
    incl.cut = 0.8,
    n.cut = 1,
    solution = "all",
    dir.exp = lr$dir.exp,
    progress = FALSE
  ))
  expect_qcaert_result_structure(
    loo,
    class = "loo_test",
    diagnostics = c("row_index", "case_label", "solution_changed", "fit_changed_types"),
    results = c("row_index", "case_label", "solution_change", "fit_changed_types"),
    settings = c("outcome", "conditions", "cases", "monitored_solutions"),
    top = c("baseline", "by_case", "settings")
  )
  expect_identical(nrow(loo$results), 6L)
})

test_that("stress: calibration workflows cover direct-six and indirect anchors", {
  skip_qcaert_stress()
  skip_if_not_installed("QCA")

  direct <- qcaert_fixture_direct6()
  direct_out <- suppressWarnings(calib.test(
    raw.data = direct$raw,
    calib.data = direct$calib,
    outcome = direct$outcome,
    conditions = direct$conditions,
    calib_spec = direct$calib_spec,
    test.conditions = direct$test.conditions,
    unit_step = 1,
    max_steps = 2,
    incl.cut = 0.75,
    n.cut = 1,
    solution = "all",
    dir.exp = direct$dir.exp,
    progress = FALSE
  ))
  expect_qcaert_result_structure(
    direct_out,
    class = "calib_test",
    diagnostics = c("set", "role", "method", "anchor", "direction", "monitored_solutions"),
    results = c("set", "role", "method", "anchor", "direction", "con_reason"),
    settings = c("outcome", "conditions", "calib_spec", "monitored_solutions", "test.outcome"),
    top = c("bounds", "baseline", "by_set", "settings")
  )
  expect_identical(unique(direct_out$results$anchor), c("E1", "C1", "I1", "I2", "C2", "E2"))

  indirect <- qcaert_fixture_indirect()
  indirect_out <- suppressWarnings(calib.test(
    raw.data = indirect$raw,
    calib.data = indirect$calib,
    outcome = indirect$outcome,
    conditions = indirect$conditions,
    calib_spec = indirect$calib_spec,
    test.conditions = indirect$test.conditions,
    unit_step = 1,
    max_steps = 2,
    incl.cut = 0.75,
    n.cut = 1,
    solution = "all",
    dir.exp = indirect$dir.exp,
    progress = FALSE
  ))
  expect_qcaert_result_structure(
    indirect_out,
    class = "calib_test",
    diagnostics = c("set", "role", "method", "anchor", "direction", "monitored_solutions"),
    results = c("set", "role", "method", "anchor", "direction", "con_reason"),
    settings = c("outcome", "conditions", "calib_spec", "monitored_solutions", "test.outcome"),
    top = c("bounds", "baseline", "by_set", "settings")
  )
  expect_identical(unique(indirect_out$results$anchor), c("T1", "T2", "T3"))
})

test_that("stress: randomized workflows are reproducible with fixed seeds", {
  skip_qcaert_stress()
  skip_if_not_installed("QCA")

  fixture <- qcaert_fixture_direct6()

  alt_one <- suppressWarnings(altset.test(
    raw.data = fixture$raw,
    calib.data = fixture$calib,
    outcome = fixture$outcome,
    conditions = fixture$conditions,
    calib_spec = fixture$calib_spec,
    test.conditions = fixture$test.conditions,
    unit_step = 1,
    calib_max_steps = 2,
    incl.cut = 0.75,
    incl_step = 0.05,
    incl_max_steps = 2,
    n.cut = 1,
    ncut_step = 1,
    ncut_max_steps = 2,
    n_draws = 5,
    seed = 303,
    solution = "all",
    dir.exp = fixture$dir.exp,
    progress = FALSE
  ))
  alt_two <- suppressWarnings(altset.test(
    raw.data = fixture$raw,
    calib.data = fixture$calib,
    outcome = fixture$outcome,
    conditions = fixture$conditions,
    calib_spec = fixture$calib_spec,
    test.conditions = fixture$test.conditions,
    unit_step = 1,
    calib_max_steps = 2,
    incl.cut = 0.75,
    incl_step = 0.05,
    incl_max_steps = 2,
    n.cut = 1,
    ncut_step = 1,
    ncut_max_steps = 2,
    n_draws = 5,
    seed = 303,
    solution = "all",
    dir.exp = fixture$dir.exp,
    progress = FALSE
  ))
  expect_qcaert_result_structure(
    alt_one,
    class = "altset_test",
    diagnostics = c("draw", "status", "changed_sets", "changed_roles", "fit_changed_types"),
    results = c("draw", "status", "changed_sets", "changed_roles", "fit_changed_types"),
    settings = c("outcome", "conditions", "test.outcome", "n_draws", "seed"),
    top = c("summary", "baseline", "by_draw", "settings")
  )
  expect_equal(alt_one$results, alt_two$results)
  expect_equal(alt_one$diagnostics, alt_two$diagnostics)

  subsample_one <- suppressWarnings(subsample.test(
    data = fixture$calib,
    outcome = fixture$outcome,
    conditions = fixture$conditions,
    sample_prop = 0.8,
    reps = 5,
    stratify = "outcome",
    seed = 404,
    incl.cut = 0.75,
    n.cut = 1,
    solution = "all",
    dir.exp = fixture$dir.exp,
    progress = FALSE
  ))
  subsample_two <- suppressWarnings(subsample.test(
    data = fixture$calib,
    outcome = fixture$outcome,
    conditions = fixture$conditions,
    sample_prop = 0.8,
    reps = 5,
    stratify = "outcome",
    seed = 404,
    incl.cut = 0.75,
    n.cut = 1,
    solution = "all",
    dir.exp = fixture$dir.exp,
    progress = FALSE
  ))
  expect_qcaert_result_structure(
    subsample_one,
    class = "subsample_test",
    diagnostics = c("rep", "status", "n_sample", "n_holdout", "fit_changed_types"),
    results = c("rep", "n_sample", "n_holdout", "status", "fit_changed_types"),
    settings = c("outcome", "conditions", "sample_prop_requested", "stratify", "seed"),
    top = c("summary", "baseline", "by_run", "settings")
  )
  expect_equal(subsample_one$results, subsample_two$results)
  expect_equal(subsample_one$diagnostics, subsample_two$diagnostics)
})

test_that("stress: cluster diagnostics handle all solution types with repeated units", {
  skip_qcaert_stress()
  skip_if_not_installed("QCA")

  fixture <- qcaert_fixture_cluster()
  out <- suppressWarnings(cluster.test(
    data = fixture$data,
    tt = fixture$truth_table,
    cluster_id = fixture$cluster_id,
    unit_id = fixture$unit_id,
    solution = "all",
    dir.exp = fixture$dir.exp,
    progress = FALSE
  ))

  expect_qcaert_cluster_structure(
    out,
    diagnostics = c("solution_type", "configuration_key", "status", "within_available", "n_units_repeated"),
    overview = c("solution_type", "configuration", "status", "within_available", "n_units_repeated"),
    clusters = c("solution_type", "configuration", "component", "cluster_id"),
    units = c("solution_type", "configuration", "component", "unit_id"),
    settings = c("outcome", "conditions", "solution", "monitored_solutions"),
    top = c("baseline", "by_cluster", "by_unit", "settings")
  )
  expect_setequal(out$settings$monitored_solutions, c("conservative", "parsimonious", "intermediate"))
  expect_true(any(out$results$overview$within_available))
})

#' Plot and chart qcaERT results
#'
#' Plot methods provide visual inspection helpers for qcaERT result objects.
#' They are simplified views of the existing result object: the
#' detailed `diagnostics` table supplies the interval and heatmap views, while
#' the path-level trace components supply trace views. [sol.chart()] provides a
#' matching visual display for [sol.df()] solution tables.
#'
#' Plotting is optional. The qcaERT analysis functions do not require
#' `ggplot2`, but these methods do.
#'
#' @param x An `incl_test` object returned by [incl.test()], a `calib_test`
#'   object returned by [calib.test()], or a `theory_test` object returned by
#'   [theory.test()].
#' @param type Plot type. `incl_test` supports `"interval"` and `"trace"`.
#'   `calib_test` supports `"interval"`, `"heatmap"`, and `"trace"`.
#'   `theory_test` has one consistency-coverage plot and does not use `type`.
#' @param solution_type Solution type to plot. Required when more than one
#'   solution type is present. Accepted values follow the common qcaERT
#'   solution-type conventions, excluding `"all"`.
#' @param intermediate_branch For `theory_test` intermediate plots, the
#'   intermediate branch to plot when more than one branch is present.
#' @param directions Character vector selecting lower and/or upper searches.
#' @param stop_reason Optional character vector used to filter diagnostic rows
#'   by stop reason before plotting.
#' @param changed_types Optional solution-type filter applied to comma-coded
#'   `changed_types` diagnostic values. Accepted tokens follow the common
#'   qcaERT solution-type conventions: `"con"`/`"conservative"`,
#'   `"par"`/`"parsimonious"`, and `"int"`/`"intermediate"`.
#' @param solution Reserved. Plot methods do not accept `solution`; use
#'   `solution_type`.
#' @param monitored_solutions Reserved. Plot methods do not accept
#'   `monitored_solutions`; use `solution_type`.
#' @param i_mode For `incl_test`, optional filter applied to the `i_mode`
#'   diagnostic column. Accepted values are `"all"` and `"C1P1"`.
#' @param direction For `type = "trace"`, the single search direction to plot:
#'   `"lower"` or `"upper"`.
#' @param show_stop Logical. If `TRUE`, draw reference lines for baseline and,
#'   when present, first-failing values.
#' @param legend Logical. If `FALSE`, hide the plot legend.
#' @param theme A ggplot2 theme object added to the plot.
#' @param sets For `calib_test`, optional character vector selecting calibrated
#'   sets.
#' @param roles For `calib_test`, optional character vector selecting set
#'   roles. Supported values are `"condition"` and `"outcome"`.
#' @param anchors For `calib_test`, optional character vector selecting
#'   calibration anchors. Supported anchors are taken from the result object and
#'   may include crisp (`"T"`), fuzzy direct three-threshold (`"E"`, `"C"`,
#'   `"I"`), fuzzy direct six-threshold (`"E1"`, `"C1"`, `"I1"`, `"I2"`,
#'   `"C2"`, `"E2"`), or fuzzy indirect (`"T1"`, `"T2"`, ...) anchors.
#' @param metric For `calib_test` heatmaps, the quantity used for the fill
#'   color: `"raw"`, `"pct"`, or `"steps"`.
#' @param value For `calib_test` heatmaps with `metric = "raw"`, the raw
#'   threshold value to show: `"delta"`, `"last_safe"`, or `"failing"`.
#' @param abs_delta Logical. If `TRUE`, heatmaps using raw or percentage deltas
#'   display absolute values.
#' @param cell For `calib_test` heatmaps, whether cells are split by
#'   anchor-direction path (`"anchor_direction"`) or by anchor only
#'   (`"anchor"`).
#' @param show_text Logical. If `TRUE`, add rounded heatmap values as text.
#' @param show_labels For `theory_test`, logical. If `TRUE`, print theory names
#'   next to their consistency/coverage points.
#' @param label_line For `theory_test`, logical. If `TRUE`, draw a line from
#'   each consistency/coverage point to its theory-name label.
#' @param label_line_alpha For `theory_test`, transparency of label connector
#'   lines.
#' @param label_line_width For `theory_test`, width of label connector lines.
#' @param point_size For `theory_test`, point size.
#' @param text_size For `theory_test`, theory-name label size.
#' @param label_nudge_x For `theory_test`, horizontal nudge for theory-name
#'   labels.
#' @param label_nudge_y For `theory_test`, vertical nudge for theory-name
#'   labels.
#' @param set For `calib_test` trace plots, the single calibrated set to plot.
#' @param anchor For `calib_test` trace plots, the single calibration anchor to
#'   plot.
#' @param order_sets For `calib_test`, set ordering in interval and heatmap
#'   views. `"input"` preserves result order; `"most_sensitive"` and
#'   `"least_sensitive"` order by the smallest solution-changing percentage
#'   delta when available.
#' @param ... Additional graphical arguments forwarded to the primary ggplot2
#'   geometry.
#'
#' @returns A ggplot object.
#'
#' @section Plot Types:
#' For `incl_test`, `"interval"` shows the baseline inclusion cutoff and the
#' lower/upper last-safe values. `"trace"` shows the stepwise path for one
#' search direction.
#'
#' For `calib_test`, `"interval"` shows baseline and lower/upper last-safe
#' threshold values by set and anchor. `"heatmap"` summarizes sensitivity
#' across anchor paths. `"trace"` shows the stepwise path for one set,
#' anchor, and direction.
#'
#' For `theory_test`, the plot compares theory-specific selected models in
#' consistency/coverage space for one selected solution type. Theory names are shown
#' near the points, and the legend pairs each theory with its selected solution
#' terms.
#'
#' For [sol.df()] tables, [sol.chart()] draws a table-like chart of sufficient
#' configurations, using filled and open circles for condition presence and
#' absence.
#'
#' @section Family Consistency:
#' These methods follow the common qcaERT output conventions described in
#' `?qcaERT_conventions`: plots read from `diagnostics` and path-level supporting
#' components, while `print()` and `as.data.frame()` keep their existing
#' concise and tabular behavior. [sol.chart()] follows the same division of
#' labor by reading from the already-extracted [sol.df()] table.
#'
#' @examples
#' \donttest{
#' library(QCA)
#' library(ggplot2)
#'
#' data(LR)
#'
#' conditions <- c("DEV", "URB", "LIT", "IND", "STB")
#' outcome <- "SURV"
#' dir_exp <- rep("1", length(conditions))
#'
#' thresholds <- list(
#'   DEV = findTh(LR$DEV, groups = 7),
#'   URB = findTh(LR$URB, groups = 4),
#'   LIT = findTh(LR$LIT, groups = 4),
#'   IND = findTh(LR$IND, groups = 4),
#'   STB = findTh(LR$STB, groups = 4),
#'   SURV = findTh(LR$SURV, groups = 4)
#' )
#'
#' dat <- LR
#' dat$DEV <- calibrate(LR$DEV, type = "fuzzy", thresholds = thresholds$DEV)
#' dat$URB <- calibrate(LR$URB, type = "fuzzy", thresholds = thresholds$URB)
#' dat$LIT <- calibrate(LR$LIT, type = "fuzzy", thresholds = thresholds$LIT)
#' dat$IND <- calibrate(LR$IND, type = "fuzzy", thresholds = thresholds$IND)
#' dat$STB <- calibrate(LR$STB, type = "fuzzy", thresholds = thresholds$STB)
#' dat$SURV <- calibrate(LR$SURV, type = "fuzzy", thresholds = thresholds$SURV)
#'
#' incl_out <- incl.test(
#'   data = dat,
#'   outcome = outcome,
#'   conditions = conditions,
#'   incl.cut = 0.8,
#'   step = 0.05,
#'   max_steps = 5,
#'   n.cut = 1,
#'   solution = "all",
#'   dir.exp = dir_exp,
#'   progress = TRUE
#' )
#'
#' calib_spec <- list(
#'   DEV = list(raw = "DEV", type = "fuzzy", method = "direct", thresholds = thresholds$DEV),
#'   URB = list(raw = "URB", type = "fuzzy", method = "direct", thresholds = thresholds$URB),
#'   LIT = list(raw = "LIT", type = "fuzzy", method = "direct", thresholds = thresholds$LIT),
#'   IND = list(raw = "IND", type = "fuzzy", method = "direct", thresholds = thresholds$IND),
#'   STB = list(raw = "STB", type = "fuzzy", method = "direct", thresholds = thresholds$STB)
#' )
#'
#' calib_out <- calib.test(
#'   raw.data = LR,
#'   calib.data = dat,
#'   outcome = outcome,
#'   conditions = conditions,
#'   calib_spec = calib_spec,
#'   test.conditions = c("DEV", "URB"),
#'   unit_step = NULL,
#'   unit_step_divisor = 10,
#'   max_steps = 5,
#'   incl.cut = 0.8,
#'   n.cut = 1,
#'   solution = "all",
#'   dir.exp = dir_exp,
#'   progress = TRUE
#' )
#'
#' theories <- list(
#'   development = c("DEV", "URB", "LIT"),
#'   industrial = c("DEV", "URB", "IND"),
#'   broad = c("DEV", "URB", "LIT", "IND", "STB")
#' )
#'
#' theory_out <- theory.test(
#'   data = dat,
#'   outcome = outcome,
#'   theories = theories,
#'   incl.cut = 0.8,
#'   n.cut = 1,
#'   solution = "all",
#'   dir.exp = list(
#'     development = c("1", "1", "1"),
#'     industrial = c("1", "1", "1"),
#'     broad = c("1", "1", "1", "1", "1")
#'   ),
#'   progress = TRUE
#' )
#'
#' plot(incl_out, solution_type = "conservative")
#' plot(incl_out, solution_type = "conservative", type = "trace", direction = "lower")
#' plot(calib_out, solution_type = "conservative")
#' plot(calib_out, solution_type = "conservative", type = "heatmap")
#' plot(
#'   calib_out,
#'   solution_type = "conservative",
#'   type = "trace",
#'   set = "DEV",
#'   anchor = "E1",
#'   direction = "lower"
#' )
#' plot(theory_out, solution_type = "conservative")
#' }
#'
#' @name qcaERT_plots
NULL

.empty_change_trace <- function(value_col, value_type = c("numeric", "integer")) {
  value_type <- match.arg(value_type)

  value_vec <- switch(
    value_type,
    numeric = numeric(0),
    integer = integer(0)
  )

  out <- data.frame(
    step = integer(0),
    changed = logical(0),
    status = character(0),
    change_kind = character(0),
    stringsAsFactors = FALSE
  )

  out[[value_col]] <- value_vec
  out <- out[, c("step", value_col, "changed", "status", "change_kind"), drop = FALSE]
  out
}

.append_change_trace <- function(trace, step, value_col, value, changed, status, change_kind = NA_character_) {
  row <- data.frame(
    step = step,
    changed = changed,
    status = status,
    change_kind = change_kind,
    stringsAsFactors = FALSE
  )

  row[[value_col]] <- value
  row <- row[, c("step", value_col, "changed", "status", "change_kind"), drop = FALSE]

  rbind(trace, row)
}

.empty_same_trace <- function() {
  data.frame(
    step = integer(0),
    value = numeric(0),
    same = logical(0),
    stringsAsFactors = FALSE
  )
}

.append_same_trace <- function(trace, step, value, same) {
  rbind(
    trace,
    data.frame(
      step = step,
      value = value,
      same = same,
      stringsAsFactors = FALSE
    )
  )
}

#' @rdname qcaERT_plots
#' @name qcaERT_plots
#' @method plot calib_test
#' @export
plot.calib_test <- function(
    x,
    type = c("interval", "heatmap", "trace"),
    sets = NULL,
    roles = NULL,
    anchors = NULL,
    directions = c("lower", "upper"),
    stop_reason = NULL,
    changed_types = NULL,
    solution_type = NULL,
    solution = NULL,
    monitored_solutions = NULL,
    metric = c("raw", "pct", "steps"),
    value = c("delta", "last_safe", "failing"),
    abs_delta = FALSE,
    cell = c("anchor_direction", "anchor"),
    show_text = FALSE,
    set = NULL,
    anchor = NULL,
    direction = NULL,
    show_stop = TRUE,
    order_sets = c("input", "most_sensitive", "least_sensitive"),
    legend = TRUE,
    theme = .plot_theme(),
    ...
) {
  .plot_require_ggplot2()

  type <- match.arg(type)
  dots_plot <- list(...)
  .plot_reject_solution_args(solution, monitored_solutions, "plot.calib_test")
  .plot_reject_solution_dots(dots_plot, "plot.calib_test")
  metric <- match.arg(metric)
  value <- match.arg(value)
  cell <- match.arg(cell)
  order_sets <- match.arg(order_sets)

  if (is.null(x$diagnostics) || !is.data.frame(x$diagnostics)) {
    stop("x must contain a data frame in x$diagnostics.", call. = FALSE)
  }

  required <- c(
    "set",
    "role",
    "raw",
    "type",
    "method",
    "anchor",
    "direction",
    "solution",
    "monitored_solutions",
    "start_value",
    "last_safe_value",
    "failing_value",
    "number_of_steps",
    "total_delta_units",
    "delta_as_pct_of_raw_range",
    "stop_reason",
    "changed_types"
  )
  missing <- setdiff(required, names(x$diagnostics))
  if (length(missing)) {
    stop(
      "x$diagnostics is missing required column(s): ",
      paste(missing, collapse = ", "),
      call. = FALSE
    )
  }

  df0 <- x$diagnostics
  df0$anchor <- toupper(df0$anchor)

  stop_no_rows <- function() {
    lines <- c("No rows left to plot after filtering.")

    if (!is.null(sets)) lines <- c(lines, paste0("Requested sets: ", paste(sets, collapse = ", ")))
    if (!is.null(roles)) lines <- c(lines, paste0("Requested roles: ", paste(roles, collapse = ", ")))
    if (!is.null(anchors)) lines <- c(lines, paste0("Requested anchors: ", paste(anchors, collapse = ", ")))
    if (!is.null(directions)) lines <- c(lines, paste0("Requested directions: ", paste(directions, collapse = ", ")))
    if (!is.null(stop_reason)) lines <- c(lines, paste0("Requested stop_reason: ", paste(stop_reason, collapse = ", ")))
    if (!is.null(changed_types)) lines <- c(lines, paste0("Requested changed_types: ", paste(changed_types, collapse = ", ")))
    if (!is.null(solution_type)) lines <- c(lines, paste0("Requested solution_type: ", paste(solution_type, collapse = ", ")))

    available_df <- df0
    available_df$anchor <- toupper(available_df$anchor)
    for (column in c("set", "role", "anchor", "direction", "solution_type", "stop_reason", "changed_types")) {
      available <- sort(table(.plot_as_chr0(available_df[[column]])), decreasing = TRUE)
      available <- available[nzchar(names(available))]
      label <- paste0("Available ", column, ": ")
      if (length(available)) {
        lines <- c(lines, paste0(label, .plot_fmt_counts(available)))
      } else {
        lines <- c(lines, paste0(label, "<none>"))
      }
    }

    stop(paste(lines, collapse = "\n"), call. = FALSE)
  }

  directions <- .plot_normalize_directions(directions)
  if (!is.null(anchors)) {
    anchors <- toupper(trimws(as.character(anchors)))
    anchors <- anchors[nzchar(anchors)]
  }
  selected_solution <- .plot_select_solution_type(df0, solution_type)
  df0 <- selected_solution$data
  df0$anchor <- toupper(df0$anchor)
  df <- df0
  solution_type <- selected_solution$solution_type

  if (!is.null(sets)) df <- df[df$set %in% sets, , drop = FALSE]
  if (!is.null(roles)) df <- df[df$role %in% roles, , drop = FALSE]
  if (!is.null(anchors)) df <- df[df$anchor %in% anchors, , drop = FALSE]
  if (!is.null(directions)) df <- df[df$direction %in% directions, , drop = FALSE]
  if (!is.null(stop_reason)) df <- df[df$stop_reason %in% stop_reason, , drop = FALSE]
  if (!is.null(changed_types)) {
    keep <- .plot_match_solution_tokens(df$changed_types, changed_types, arg = "changed_types")
    df <- df[keep, , drop = FALSE]
  }

  if (!nrow(df)) stop_no_rows()

  set_levels <- unique(df$set)
  if (order_sets != "input") {
    changed <- df[df$stop_reason == "solution_change", , drop = FALSE]
    if (nrow(changed)) {
      sensitivity <- tapply(
        abs(changed$delta_as_pct_of_raw_range),
        changed$set,
        function(value) suppressWarnings(min(value, na.rm = TRUE))
      )
      sensitivity[is.na(sensitivity)] <- Inf

      all_sets <- unique(df$set)
      ordered <- rep(Inf, length(all_sets))
      names(ordered) <- all_sets
      ordered[names(sensitivity)] <- sensitivity

      order_index <- if (order_sets == "most_sensitive") {
        order(ordered, na.last = TRUE)
      } else {
        order(ordered, decreasing = TRUE, na.last = TRUE)
      }
      set_levels <- all_sets[order_index]
    }
  }
  anchor_levels <- unique(df$anchor)

  df$set <- factor(df$set, levels = set_levels)
  df$anchor <- factor(df$anchor, levels = anchor_levels)

  if (type == "trace") {
    if (is.null(set) || is.null(anchor) || is.null(direction)) {
      stop("For type = 'trace', supply `set`, `anchor`, and `direction`.", call. = FALSE)
    }

    anchor <- toupper(trimws(as.character(anchor)))
    direction <- .plot_normalize_directions(direction, arg = "direction")
    if (length(anchor) != 1L || !nzchar(anchor)) {
      stop("For type = 'trace', `anchor` must be a single anchor name.", call. = FALSE)
    }
    if (length(direction) != 1L) {
      stop("For type = 'trace', `direction` must be a single value.", call. = FALSE)
    }

    trace_row <- df[
      df$set == set & df$anchor == anchor & df$direction == direction,
      ,
      drop = FALSE
    ]
    if (!nrow(trace_row)) {
      available <- df[df$set == set, c("anchor", "direction"), drop = FALSE]
      if (nrow(available)) {
        available <- unique(available)
        available_paths <- paste0(available$anchor, "/", available$direction)
        available_msg <- paste(available_paths, collapse = ", ")
      } else {
        available_msg <- "<none for this set after filtering>"
      }

      stop(
        "No trace path found for set = '", set,
        "', anchor = '", anchor,
        "', direction = '", direction,
        "'. Available anchor/direction paths for this set: ",
        available_msg,
        call. = FALSE
      )
    }

    if (is.null(x$by_set) || is.null(x$by_set[[set]])) {
      stop("set not found in x$by_set.", call. = FALSE)
    }

    step_key <- paste(anchor, direction, sep = "_")
    path <- x$by_set[[set]]$steps[[step_key]]
    if (!is.null(path$by_solution_type)) {
      row_solution_type <- as.character(trace_row$solution_type[1L])
      if (!is.null(path$by_solution_type[[row_solution_type]])) {
        path <- path$by_solution_type[[row_solution_type]]
      }
    }
    if (is.null(path) || is.null(path$trace) || !is.data.frame(path$trace)) {
      stop("Trace not found for this set/anchor/direction.", call. = FALSE)
    }

    trace <- path$trace
    trace_required <- c("step", "value", "changed", "status")
    missing_trace <- setdiff(trace_required, names(trace))
    if (length(missing_trace)) {
      stop(
        "Trace is missing required column(s): ",
        paste(missing_trace, collapse = ", "),
        call. = FALSE
      )
    }

    trace$preserved <- ifelse(is.na(trace$changed), "Unknown", ifelse(trace$changed, "No", "Yes"))
    trace$preserved <- factor(trace$preserved, levels = c("Yes", "No", "Unknown"))

    trace_row <- trace_row[1L, , drop = FALSE]
    start_value <- trace_row$start_value
    failing_value <- trace_row$failing_value

    fail_point <- NULL
    if (is.finite(failing_value) && !is.na(failing_value)) {
      next_step <- if (nrow(trace) == 0L) 1L else max(trace$step, na.rm = TRUE) + 1L
      fail_point <- data.frame(
        step = next_step,
        value = failing_value,
        preserved = factor("No", levels = levels(trace$preserved))
      )
    }

    line_df <- trace[, c("step", "value"), drop = FALSE]
    if (!is.null(fail_point)) {
      present <- nrow(line_df) && any(line_df$step == fail_point$step & line_df$value == fail_point$value)
      if (!present) line_df <- rbind(line_df, fail_point[, c("step", "value"), drop = FALSE])
      line_df <- line_df[order(line_df$step), , drop = FALSE]
    }

    p <- ggplot2::ggplot() +
      ggplot2::geom_line(
        data = line_df,
        ggplot2::aes(x = .data[["step"]], y = .data[["value"]]),
        ...
      ) +
      ggplot2::geom_point(
        data = trace,
        ggplot2::aes(
          x = .data[["step"]],
          y = .data[["value"]],
          shape = .data[["preserved"]],
          colour = .data[["preserved"]]
        ),
        size = 4,
        alpha = 0.8,
        ...
      ) +
      ggplot2::scale_shape_manual(name = "Preserved", values = c(Yes = 19, No = 15, Unknown = 17)) +
      ggplot2::scale_colour_manual(name = "Preserved", values = .plot_preserved_colors()) +
      ggplot2::labs(
        title = paste0("Calibration trace: ", set, " | ", anchor, " | ", direction),
        x = "Step",
        y = "Threshold value"
      ) +
      theme

    if (!is.null(fail_point)) {
      p <- p + ggplot2::geom_point(
        data = fail_point,
        ggplot2::aes(
          x = .data[["step"]],
          y = .data[["value"]],
          shape = .data[["preserved"]],
          colour = .data[["preserved"]]
        ),
        inherit.aes = FALSE,
        size = 4,
        alpha = 0.8
      )
    }
    if (show_stop && is.finite(start_value) && !is.na(start_value)) {
      p <- p + ggplot2::geom_hline(yintercept = start_value, linetype = "dashed")
    }
    if (show_stop && is.finite(failing_value) && !is.na(failing_value)) {
      p <- p + ggplot2::geom_hline(yintercept = failing_value, linetype = "dotdash")
    }
    if (!legend) p <- p + ggplot2::theme(legend.position = "none")

    return(p)
  }

  if (type == "heatmap") {
    df2 <- df

    fill <- switch(
      metric,
      raw = switch(
        value,
        delta = df2$total_delta_units,
        last_safe = df2$last_safe_value,
        failing = df2$failing_value
      ),
      pct = df2$delta_as_pct_of_raw_range,
      steps = df2$number_of_steps
    )
    if (abs_delta && metric %in% c("raw", "pct")) fill <- abs(fill)

    df2$fill <- fill
    df2$cell <- if (cell == "anchor_direction") {
      paste(df2$anchor, df2$direction, sep = "_")
    } else {
      as.character(df2$anchor)
    }
    df2$cell <- factor(df2$cell, levels = unique(df2$cell))

    fill_label <- if (metric == "pct") {
      "% raw range"
    } else if (metric == "steps") {
      "Steps"
    } else {
      switch(value, delta = "Delta", last_safe = "Last safe", failing = "First failing")
    }

    p <- ggplot2::ggplot(
      df2,
      ggplot2::aes(x = .data[["cell"]], y = .data[["set"]], fill = .data[["fill"]])
    ) +
      ggplot2::geom_tile(...) +
      ggplot2::labs(x = NULL, y = NULL, fill = fill_label, title = NULL) +
      theme

    if (metric %in% c("raw", "pct") && identical(value, "delta") && !isTRUE(abs_delta)) {
      p <- p + ggplot2::scale_fill_gradient2(
        low = .plot_colors()[["orange"]],
        mid = "white",
        high = .plot_colors()[["blue"]],
        midpoint = 0,
        na.value = .plot_colors()[["light_grey"]]
      )
    } else {
      p <- p + ggplot2::scale_fill_gradient(
        low = "white",
        high = .plot_colors()[["blue"]],
        na.value = .plot_colors()[["light_grey"]]
      )
    }

    if (show_text) {
      p <- p + ggplot2::geom_text(ggplot2::aes(label = round(.data[["fill"]], 3)), ...)
    }
    if (!legend) p <- p + ggplot2::theme(legend.position = "none")

    return(p)
  }

  lower <- df[df$direction == "lower", , drop = FALSE]
  upper <- df[df$direction == "upper", , drop = FALSE]

  lower <- lower[!duplicated(lower[c("set", "anchor")]), , drop = FALSE]
  upper <- upper[!duplicated(upper[c("set", "anchor")]), , drop = FALSE]

  merged <- merge(
    lower[, c("set", "anchor", "start_value", "last_safe_value"), drop = FALSE],
    upper[, c("set", "anchor", "start_value", "last_safe_value"), drop = FALSE],
    by = c("set", "anchor"),
    all = TRUE,
    suffixes = c("_lower", "_upper")
  )

  key_order <- unique(paste(as.character(df$set), as.character(df$anchor), sep = "\r"))
  merged_key <- paste(as.character(merged$set), as.character(merged$anchor), sep = "\r")
  merged <- merged[order(match(merged_key, key_order)), , drop = FALSE]

  merged$start_value <- merged$start_value_lower
  use_upper_start <- is.na(merged$start_value)
  merged$start_value[use_upper_start] <- merged$start_value_upper[use_upper_start]

  merged$set <- factor(merged$set, levels = set_levels)
  merged$anchor <- factor(merged$anchor, levels = anchor_levels)

  points <- rbind(
    data.frame(
      set = merged$set,
      anchor = merged$anchor,
      point = "Baseline",
      value = merged$start_value,
      stringsAsFactors = FALSE
    ),
    data.frame(
      set = merged$set,
      anchor = merged$anchor,
      point = "Lower last safe",
      value = merged$last_safe_value_lower,
      stringsAsFactors = FALSE
    ),
    data.frame(
      set = merged$set,
      anchor = merged$anchor,
      point = "Upper last safe",
      value = merged$last_safe_value_upper,
      stringsAsFactors = FALSE
    )
  )
  points <- points[is.finite(points$value) & !is.na(points$value), , drop = FALSE]
  if (!nrow(points)) {
    stop("No finite values to plot after filtering.", call. = FALSE)
  }

  points$set <- factor(points$set, levels = set_levels)
  points$anchor <- factor(points$anchor, levels = anchor_levels)
  points$point <- factor(points$point, levels = c("Baseline", "Lower last safe", "Upper last safe"))

  segment <- merged[
    is.finite(merged$last_safe_value_lower) &
      !is.na(merged$last_safe_value_lower) &
    is.finite(merged$last_safe_value_upper) &
      !is.na(merged$last_safe_value_upper),
    c("set", "anchor", "last_safe_value_lower", "last_safe_value_upper"),
    drop = FALSE
  ]

  dots_point <- list(...)
  pt_size <- if (!is.null(dots_point$size)) dots_point$size else 3
  pt_alpha <- if (!is.null(dots_point$alpha)) dots_point$alpha else 0.8

  dots_point$size <- NULL
  dots_point$alpha <- NULL
  dots_point$mapping <- NULL
  dots_point$data <- NULL
  dots_point$inherit.aes <- NULL

  point_layer <- do.call(
    ggplot2::geom_point,
    c(
      list(
        mapping = ggplot2::aes(
          x = .data[["value"]],
          y = .data[["anchor"]],
          shape = .data[["point"]],
          colour = .data[["point"]],
          fill = .data[["point"]]
        ),
        size = pt_size,
        alpha = pt_alpha
      ),
      dots_point
    )
  )

  p <- ggplot2::ggplot(points)

  if (nrow(segment)) {
    p <- p + ggplot2::geom_segment(
      data = segment,
      ggplot2::aes(
        x = .data[["last_safe_value_lower"]],
        xend = .data[["last_safe_value_upper"]],
        y = .data[["anchor"]],
        yend = .data[["anchor"]]
      ),
      inherit.aes = FALSE,
      linewidth = 0.8
    )
  }

  p <- p +
    point_layer +
    ggplot2::scale_shape_manual(name = "Point", values = c("Baseline" = 16, "Lower last safe" = 25, "Upper last safe" = 24)) +
    ggplot2::scale_colour_manual(name = "Point", values = .plot_boundary_colors()) +
    ggplot2::scale_fill_manual(name = "Point", values = .plot_boundary_colors()) +
    ggplot2::labs(x = "Threshold value", y = NULL, title = NULL) +
    theme +
    ggplot2::facet_wrap(ggplot2::vars(.data[["set"]]), scales = "free_x")

  if (!legend) p <- p + ggplot2::theme(legend.position = "none")

  p
}

#' Build a compact solution table from QCA minimization output
#'
#' Extracts prime implicants, fit statistics, model identifiers, intermediate
#' branch labels, and optional case membership strings from supplied
#' [QCA::minimize()] result objects and returns them as one combined data
#' frame.
#'
#' @param conservative Optional conservative minimization result of class
#'   `"QCA_min"`.
#' @param intermediate Optional intermediate minimization result of class
#'   `"QCA_min"`.
#' @param parsimonious Optional parsimonious minimization result of class
#'   `"QCA_min"`.
#' @param solution Which solution type to extract. Accepted values are
#'   `"all"`, `"con"` or `"conservative"`, `"par"` or `"parsimonious"`, and
#'   `"int"` or `"intermediate"`. When `solution = "all"`, the function
#'   extracts every supplied object among `conservative`, `parsimonious`, and
#'   `intermediate`.
#' @param which_M Positive integer giving which model to extract when a
#'   supplied minimization object contains multiple models.
#' @param i_mode Character string controlling which intermediate branches to
#'   extract when `intermediate` is supplied and `branches = NULL`. Accepted
#'   values are `"all"` and `"C1P1"`.
#' @param branches Optional character vector of intermediate branch names to
#'   extract from `intermediate$i.sol`. When supplied, this overrides `i_mode`.
#' @param include_cases Logical; if `TRUE`, include a `Cases` column using the
#'   case information stored in the minimization object when available. Case
#'   reconstruction expects the data-frame `pims` component produced by
#'   [QCA::minimize()].
#' @param pi_incl_cut Optional lower cutoff applied to the prime-implicant
#'   consistency column. Rows with `Consistency_PI < pi_incl_cut` are removed.
#' @param digits Optional non-negative integer used to round numeric columns in
#'   the returned table.
#' @param na_string Single character string used to replace missing values in
#'   non-numeric output fields. The default is `"-"`.
#' @param verbose Logical; if `TRUE`, print a short progress message while
#'   processing each requested solution type.
#'
#' @returns A data frame with one row per extracted prime implicant and the
#'   following columns:
#'   \describe{
#'     \item{`Solution`}{Solution type label: `"Conservative"`,
#'     `"Parsimonious"`, or `"Intermediate"`.}
#'     \item{`Model`}{Selected model number when model-specific output is
#'     available.}
#'     \item{`Intermediate_CnPn`}{Intermediate branch label when rows come from
#'     `intermediate$i.sol`.}
#'     \item{`Prime_Implicants`}{Prime implicant name.}
#'     \item{`Consistency_PI`}{Prime-implicant consistency (`inclS`).}
#'     \item{`PRI_PI`}{Prime-implicant PRI.}
#'     \item{`Raw_Coverage_PI`}{Prime-implicant raw coverage (`covS`).}
#'     \item{`Unique_Coverage_PI`}{Prime-implicant unique coverage (`covU`).}
#'     \item{`Solution_Consistency`}{Solution-level consistency.}
#'     \item{`Solution_PRI`}{Solution-level PRI.}
#'     \item{`Solution_Coverage`}{Solution-level coverage.}
#'     \item{`Cases`}{Case labels or case strings when available and requested.}
#'   }
#'
#'   The function returns a regular data frame, not a custom result object.
#'
#' @details
#' Supply at least one of `conservative`, `parsimonious`, or `intermediate`.
#'
#' For conservative and parsimonious solutions, the function extracts rows from
#' the `IC` component of the supplied minimization object. For intermediate
#' solutions, it extracts rows from the selected `i.sol` branches.
#'
#' If a supplied minimization object contains multiple models in
#' `IC$individual`, `which_M` selects the model position to extract. If the
#' requested model does not exist, the function stops with an error.
#'
#' When `include_cases = TRUE`, the function first uses any `cases` column
#' already stored in the relevant `incl.cov` table. If that is unavailable, it
#' tries to reconstruct case strings from the corresponding `pims` data frame
#' stored by [QCA::minimize()].
#'
#' The QCA solution-object conventions are described in `?qcaERT_conventions`.
#'
#' @examples
#' library(QCA)
#' data(LC)
#'
#' conditions <- c("DEV", "URB", "LIT", "IND", "STB")
#' dir_exp <- rep("1", length(conditions))
#'
#' tt <- truthTable(
#'   data = LC,
#'   outcome = "SURV",
#'   conditions = conditions,
#'   incl.cut = 0.8,
#'   n.cut = 1
#' )
#'
#' enhanced <- findRows(tt, type = 2)
#'
#' con <- minimize(
#'   input = tt,
#'   include = "",
#'   details = TRUE,
#'   show.cases = FALSE
#' )
#'
#' par <- minimize(
#'   input = tt,
#'   include = "?",
#'   exclude = enhanced,
#'   details = TRUE,
#'   show.cases = FALSE
#' )
#'
#' int <- minimize(
#'   input = tt,
#'   include = "?",
#'   dir.exp = dir_exp,
#'   exclude = enhanced,
#'   details = TRUE,
#'   show.cases = FALSE
#' )
#'
#' sol.df(
#'   conservative = con,
#'   parsimonious = par,
#'   intermediate = int,
#'   solution = "all",
#'   which_M = 1,
#'   i_mode = "C1P1",
#'   include_cases = FALSE,
#'   digits = 2
#' )
#'
#' \donttest{
#' library(QCA)
#' data(LR)
#'
#' conditions <- c("DEV", "URB", "LIT", "IND", "STB")
#' outcome <- "SURV"
#' dir_exp <- rep("1", length(conditions))
#'
#' thresholds <- list(
#'   DEV = findTh(LR$DEV, groups = 7),
#'   URB = findTh(LR$URB, groups = 4),
#'   LIT = findTh(LR$LIT, groups = 4),
#'   IND = findTh(LR$IND, groups = 4),
#'   STB = findTh(LR$STB, groups = 4),
#'   SURV = findTh(LR$SURV, groups = 4)
#' )
#'
#' dat <- LR
#' dat$DEV <- calibrate(LR$DEV, type = "fuzzy", thresholds = thresholds$DEV)
#' dat$URB <- calibrate(LR$URB, type = "fuzzy", thresholds = thresholds$URB)
#' dat$LIT <- calibrate(LR$LIT, type = "fuzzy", thresholds = thresholds$LIT)
#' dat$IND <- calibrate(LR$IND, type = "fuzzy", thresholds = thresholds$IND)
#' dat$STB <- calibrate(LR$STB, type = "fuzzy", thresholds = thresholds$STB)
#' dat$SURV <- calibrate(LR$SURV, type = "fuzzy", thresholds = thresholds$SURV)
#'
#' tt <- truthTable(
#'   data = dat,
#'   outcome = outcome,
#'   conditions = conditions,
#'   incl.cut = 0.8,
#'   n.cut = 1,
#'   complete = TRUE,
#'   show.cases = TRUE
#' )
#' enhanced <- findRows(tt, type = 2)
#'
#' con <- minimize(tt, include = "", details = TRUE, show.cases = FALSE)
#' par <- minimize(tt, include = "?", exclude = enhanced, details = TRUE, show.cases = FALSE)
#' int <- minimize(
#'   tt,
#'   include = "?",
#'   dir.exp = dir_exp,
#'   exclude = enhanced,
#'   details = TRUE,
#'   show.cases = FALSE
#' )
#'
#' sol.df(
#'   conservative = con,
#'   parsimonious = par,
#'   intermediate = int,
#'   solution = "all",
#'   which_M = 1,
#'   i_mode = "C1P1",
#'   include_cases = TRUE,
#'   digits = 3
#' )
#' }
#'
#' @seealso [sol.chart()], [calib.test()], [incl.test()], [ncut.test()],
#'   [loo.test()], [subsample.test()], [altset.test()], [theory.test()],
#'   [cluster.test()]
#' @export
sol.df <- function(
    conservative = NULL,
    intermediate = NULL,
    parsimonious = NULL,
    solution = "all",
    which_M = 1,
    i_mode = c("all", "C1P1"),
    branches = NULL,
    include_cases = TRUE,
    pi_incl_cut = NULL,
    digits = NULL,
    na_string = "-",
    verbose = FALSE
) {
  selection_controls <- .normalize_i_mode(i_mode)
  i_mode <- selection_controls$i_mode

  solution <- .normalize_solution_std(solution)

  which_M <- .as_integerish_scalar(which_M, "which_M", min = 1L)

  if (!is.null(pi_incl_cut)) {
    if (!is.numeric(pi_incl_cut) || length(pi_incl_cut) != 1L || !is.finite(pi_incl_cut)) {
      stop("`pi_incl_cut` must be NULL or a single finite numeric value.")
    }
  }

  if (!is.null(digits)) {
    if (!is.numeric(digits) || length(digits) != 1L || !is.finite(digits) || digits < 0) {
      stop("`digits` must be NULL or a single non-negative integer.")
    }
    digits <- as.integer(digits)
  }

  if (!is.character(na_string) || length(na_string) != 1L || is.na(na_string)) {
    stop("`na_string` must be a single character string.")
  }

  .validate_qca_min(conservative, "conservative")
  .validate_qca_min(intermediate, "intermediate")
  .validate_qca_min(parsimonious, "parsimonious")

  objects <- .qca_solution_objects(
    conservative = conservative,
    intermediate = intermediate,
    parsimonious = parsimonious
  )

  if (length(objects) == 0L) {
    stop("Provide at least one of `conservative`, `parsimonious`, or `intermediate`.")
  }

  requested <- .qca_requested_solutions(solution, objects)

  .build_rows <- function(
    solution_label,
    incl_cov,
    sol_incl_cov,
    cases,
    model = NA_integer_,
    branch = NA_character_
  ) {
    if (is.null(incl_cov) || !is.data.frame(incl_cov) || nrow(incl_cov) == 0L) {
      return(data.frame(
        Solution = character(0),
        Model = integer(0),
        Intermediate_CnPn = character(0),
        Prime_Implicants = character(0),
        Consistency_PI = numeric(0),
        PRI_PI = numeric(0),
        Raw_Coverage_PI = numeric(0),
        Unique_Coverage_PI = numeric(0),
        Solution_Consistency = numeric(0),
        Solution_PRI = numeric(0),
        Solution_Coverage = numeric(0),
        Cases = character(0),
        stringsAsFactors = FALSE
      ))
    }

    pi_names <- rownames(incl_cov)
    if (is.null(pi_names)) {
      pi_names <- as.character(seq_len(nrow(incl_cov)))
    }

    sol_row <- 1L
    if (!is.null(sol_incl_cov) && is.data.frame(sol_incl_cov) && nrow(sol_incl_cov) >= 1L) {
      sol_row <- 1L
    }

    out <- data.frame(
      Solution = rep(solution_label, nrow(incl_cov)),
      Model = rep(model, nrow(incl_cov)),
      Intermediate_CnPn = rep(branch, nrow(incl_cov)),
      Prime_Implicants = pi_names,
      Consistency_PI = if ("inclS" %in% colnames(incl_cov)) as.numeric(incl_cov[["inclS"]]) else NA_real_,
      PRI_PI = if ("PRI" %in% colnames(incl_cov)) as.numeric(incl_cov[["PRI"]]) else NA_real_,
      Raw_Coverage_PI = if ("covS" %in% colnames(incl_cov)) as.numeric(incl_cov[["covS"]]) else NA_real_,
      Unique_Coverage_PI = if ("covU" %in% colnames(incl_cov)) as.numeric(incl_cov[["covU"]]) else NA_real_,
      Solution_Consistency = if (!is.null(sol_incl_cov) && "inclS" %in% colnames(sol_incl_cov)) rep(as.numeric(sol_incl_cov[sol_row, "inclS"]), nrow(incl_cov)) else NA_real_,
      Solution_PRI = if (!is.null(sol_incl_cov) && "PRI" %in% colnames(sol_incl_cov)) rep(as.numeric(sol_incl_cov[sol_row, "PRI"]), nrow(incl_cov)) else NA_real_,
      Solution_Coverage = if (!is.null(sol_incl_cov) && "covS" %in% colnames(sol_incl_cov)) rep(as.numeric(sol_incl_cov[sol_row, "covS"]), nrow(incl_cov)) else NA_real_,
      Cases = as.character(cases),
      stringsAsFactors = FALSE
    )

    if (!is.null(pi_incl_cut)) {
      out <- out[is.na(out$Consistency_PI) | out$Consistency_PI >= pi_incl_cut, , drop = FALSE]
    }

    rownames(out) <- NULL
    out
  }

  .extract_standard <- function(obj, solution_type, which_M) {
    if (isTRUE(verbose)) {
      message("Processing ", solution_type, " solution...")
    }

    solution_label <- .qca_solution_label(solution_type)
    ic <- obj[["IC"]]
    selected <- .qca_select_ic_model(ic, which_M = which_M, context = solution_type)
    icm <- selected$ic
    incl_cov <- icm[["incl.cov"]]
    sol_incl_cov <- icm[["sol.incl.cov"]]
    pims <- if ("pims" %in% names(icm)) icm[["pims"]] else obj[["pims"]]
    cases <- .qca_cases_from_incl_cov(incl_cov, pims, include_cases = include_cases)

    .build_rows(
      solution_label = solution_label,
      incl_cov = incl_cov,
      sol_incl_cov = sol_incl_cov,
      cases = cases,
      model = selected$model,
      branch = NA_character_
    )
  }

  .extract_intermediate <- function(obj, which_M, i_mode, branches) {
    if (isTRUE(verbose)) {
      message("Processing intermediate solution...")
    }

    take_branches <- .resolve_branches(obj, i_mode = i_mode, branches = branches)

    out <- list()

    for (br in take_branches) {
      current <- obj[["i.sol"]][[br]]
      ic <- current[["IC"]]
      selected <- .qca_select_ic_model(
        ic,
        which_M = which_M,
        context = paste0("intermediate branch ", br)
      )

      icm <- selected$ic
      incl_cov <- icm[["incl.cov"]]
      sol_incl_cov <- icm[["sol.incl.cov"]]
      pims <- if ("pims" %in% names(icm)) icm[["pims"]] else current[["pims"]]
      cases <- .qca_cases_from_incl_cov(incl_cov, pims, include_cases = include_cases)

      out[[length(out) + 1L]] <- .build_rows(
        solution_label = "Intermediate",
        incl_cov = incl_cov,
        sol_incl_cov = sol_incl_cov,
        cases = cases,
        model = selected$model,
        branch = br
      )
    }

    if (length(out) == 0L) {
      return(.build_rows(
        solution_label = "Intermediate",
        incl_cov = NULL,
        sol_incl_cov = NULL,
        cases = character(0)
      ))
    }

    do.call(rbind, out)
  }

  rows <- list()

  if ("conservative" %in% requested) {
    rows[[length(rows) + 1L]] <- .extract_standard(
      obj = objects$conservative,
      solution_type = "conservative",
      which_M = which_M
    )
  }

  if ("parsimonious" %in% requested) {
    rows[[length(rows) + 1L]] <- .extract_standard(
      obj = objects$parsimonious,
      solution_type = "parsimonious",
      which_M = which_M
    )
  }

  if ("intermediate" %in% requested) {
    rows[[length(rows) + 1L]] <- .extract_intermediate(
      obj = objects$intermediate,
      which_M = which_M,
      i_mode = i_mode,
      branches = branches
    )
  }

  out <- .bind_rows_result(rows)

  if (!is.null(digits)) {
    numeric_cols <- vapply(out, is.numeric, logical(1))
    out[numeric_cols] <- lapply(out[numeric_cols], round, digits = digits)
  }

  if (!is.null(na_string)) {
    out$Model[is.na(out$Model)] <- na_string
    out$Intermediate_CnPn[is.na(out$Intermediate_CnPn)] <- na_string
    out$Cases[is.na(out$Cases)] <- na_string

    for (nm in names(out)) {
      if (!is.numeric(out[[nm]]) && !is.integer(out[[nm]])) {
        out[[nm]][is.na(out[[nm]])] <- na_string
      }
    }
  }

  rownames(out) <- NULL
  out
}

test_that("ncut.test rejects non-integer which_M", {
  skip_if_not_installed("QCA")

  expect_error(
    ncut.test(
      data = tiny_qca,
      outcome = "Y",
      conditions = c("A", "B"),
      which_M = 1.9,
      progress = FALSE
    ),
    "`which_M` must be integer-like.",
    fixed = TRUE
  )
})

test_that("sol.df rejects non-QCA objects", {
  expect_error(
    sol.df(conservative = mtcars),
    "`conservative` must be a `QCA_min` object or NULL.",
    fixed = TRUE
  )
})

#' @rdname qcaERT_plots
#' @name qcaERT_plots
#' @method plot theory_test
#' @export
plot.theory_test <- function(
    x,
    solution_type = NULL,
    intermediate_branch = NULL,
    show_labels = TRUE,
    label_line = TRUE,
    label_line_alpha = 0.45,
    label_line_width = 0.35,
    point_size = 3.5,
    text_size = 3.5,
    label_nudge_x = 0.008,
    label_nudge_y = 0.006,
    legend = TRUE,
    theme = .plot_theme(),
    ...
) {
  .plot_require_ggplot2()
  dots_point <- list(...)
  .plot_reject_solution_dots(dots_point, "plot.theory_test")

  if (is.null(x$results) || is.null(x$results$models) || !is.data.frame(x$results$models)) {
    stop("x must contain a data frame in x$results$models.", call. = FALSE)
  }
  if (is.null(x$results$solutions) || !is.data.frame(x$results$solutions)) {
    stop("x must contain a data frame in x$results$solutions.", call. = FALSE)
  }

  models <- x$results$models
  required_models <- c("theory", "solution_type", "intermediate_branch", "status", "inclS", "covS")
  missing_models <- setdiff(required_models, names(models))
  if (length(missing_models)) {
    stop(
      "x$results$models is missing required column(s): ",
      paste(missing_models, collapse = ", "),
      call. = FALSE
    )
  }

  selected_solution_type <- .theory_plot_select_solution_type(models, solution_type)
  selected_branch <- .theory_plot_select_branch(models, selected_solution_type, intermediate_branch)

  df <- .theory_plot_models(models, selected_solution_type, selected_branch)
  if (!nrow(df)) {
    stop("No successful model rows with finite consistency and coverage are available for the selected solution type.", call. = FALSE)
  }

  solutions <- .theory_plot_solution_labels(
    solutions = x$results$solutions,
    solution_type = selected_solution_type,
    intermediate_branch = selected_branch,
    theories = as.character(df$theory)
  )

  df$legend_label <- unname(solutions[as.character(df$theory)])
  df$legend_label[is.na(df$legend_label) | !nzchar(df$legend_label)] <- paste0(as.character(df$theory), ": <no selected solution>")
  df <- .theory_plot_label_positions(df, label_nudge_x, label_nudge_y)

  theory_names <- as.character(df$theory)
  colors <- .plot_theory_colors(theory_names)
  legend_labels <- stats::setNames(
    .theory_plot_wrap_label(df$legend_label),
    theory_names
  )

  dots_point$mapping <- NULL
  dots_point$data <- NULL
  dots_point$inherit.aes <- NULL
  if (is.null(dots_point$size)) {
    dots_point$size <- point_size
  }
  if (is.null(dots_point$alpha)) {
    dots_point$alpha <- 0.9
  }

  point_layer <- do.call(
    ggplot2::geom_point,
    c(
      list(
        mapping = ggplot2::aes(
          x = .data[["inclS"]],
          y = .data[["covS"]],
          colour = .data[["theory"]]
        )
      ),
      dots_point
    )
  )

  p <- ggplot2::ggplot(df) +
    point_layer

  if (isTRUE(show_labels)) {
    if (isTRUE(label_line)) {
      p <- p + ggplot2::geom_segment(
        ggplot2::aes(
          x = .data[["inclS"]],
          y = .data[["covS"]],
          xend = .data[["label_x"]],
          yend = .data[["label_y"]],
          colour = .data[["theory"]]
        ),
        linewidth = label_line_width,
        alpha = label_line_alpha,
        lineend = "round",
        show.legend = FALSE
      )
    }

    p <- p + ggplot2::geom_text(
      ggplot2::aes(
        x = .data[["label_x"]],
        y = .data[["label_y"]],
        label = .data[["theory"]],
        colour = .data[["theory"]],
        hjust = .data[["label_hjust"]],
        vjust = .data[["label_vjust"]]
      ),
      size = text_size,
      show.legend = FALSE
    )
  }

  title <- paste0("Theory-specification comparison: ", .compact_solution_type_names(selected_solution_type))
  if (identical(selected_solution_type, "intermediate") && !is.na(selected_branch)) {
    title <- paste0(title, " | ", selected_branch)
  }

  p <- p +
    ggplot2::scale_colour_manual(
      name = NULL,
      values = colors,
      breaks = theory_names,
      labels = legend_labels
    ) +
    ggplot2::scale_x_continuous(breaks = seq(0, 1, by = 0.25)) +
    ggplot2::scale_y_continuous(breaks = seq(0, 1, by = 0.25)) +
    ggplot2::coord_cartesian(xlim = c(0, 1), ylim = c(0, 1), clip = "off") +
    ggplot2::labs(
      title = title,
      x = "Consistency",
      y = "Coverage"
    ) +
    ggplot2::guides(
      colour = ggplot2::guide_legend(
        ncol = 1,
        byrow = TRUE,
        override.aes = list(size = point_size, alpha = 1)
      )
    ) +
    theme +
    ggplot2::theme(
      legend.position = if (isTRUE(legend)) "bottom" else "none",
      legend.text = ggplot2::element_text(hjust = 0),
      plot.margin = ggplot2::margin(5.5, 20, 5.5, 5.5)
    )

  p
}

.theory_plot_select_solution_type <- function(models, solution_type) {
  available <- unique(as.character(models$solution_type))
  available <- available[!is.na(available) & nzchar(available)]

  if (is.null(solution_type)) {
    if (length(available) == 1L) {
      return(available)
    }
    stop(
      "`solution_type` must be supplied when more than one solution type is present. Available solution types: ",
      paste(available, collapse = ", "),
      call. = FALSE
    )
  }

  solution_type <- .plot_solution_filter(solution_type, allow_all = FALSE, arg = "solution_type")
  if (length(solution_type) != 1L) {
    stop("`solution_type` must be a single solution type.", call. = FALSE)
  }
  if (!solution_type %in% available) {
    stop(
      "`solution_type` is not available in this object. Available solution types: ",
      paste(available, collapse = ", "),
      call. = FALSE
    )
  }

  solution_type
}

.theory_plot_select_branch <- function(models, solution_type, intermediate_branch) {
  if (!identical(solution_type, "intermediate")) {
    if (!is.null(intermediate_branch)) {
      stop("`intermediate_branch` is only used when solution_type = 'intermediate'.", call. = FALSE)
    }
    return(NA_character_)
  }

  branch_values <- unique(as.character(models$intermediate_branch[models$solution_type == "intermediate"]))
  branch_values <- branch_values[!is.na(branch_values) & nzchar(branch_values)]

  if (length(branch_values) == 0L) {
    return(NA_character_)
  }

  if (is.null(intermediate_branch)) {
    if (length(branch_values) == 1L) {
      return(branch_values)
    }
    stop(
      "`intermediate_branch` must be supplied when more than one intermediate branch is present. Available branches: ",
      paste(branch_values, collapse = ", "),
      call. = FALSE
    )
  }

  intermediate_branch <- trimws(as.character(intermediate_branch))
  intermediate_branch <- intermediate_branch[nzchar(intermediate_branch)]
  if (length(intermediate_branch) != 1L) {
    stop("`intermediate_branch` must be a single branch name.", call. = FALSE)
  }
  if (!intermediate_branch %in% branch_values) {
    stop(
      "`intermediate_branch` is not available in this object. Available branches: ",
      paste(branch_values, collapse = ", "),
      call. = FALSE
    )
  }

  intermediate_branch
}

.theory_plot_models <- function(models, solution_type, intermediate_branch) {
  keep <- models$solution_type == solution_type
  if (identical(solution_type, "intermediate")) {
    if (is.na(intermediate_branch)) {
      keep <- keep & is.na(models$intermediate_branch)
    } else {
      keep <- keep & !is.na(models$intermediate_branch) & models$intermediate_branch == intermediate_branch
    }
  }

  df <- models[keep, , drop = FALSE]
  df <- df[df$status == "ok" & is.finite(df$inclS) & is.finite(df$covS), , drop = FALSE]
  if (!nrow(df)) {
    return(df)
  }

  theory_levels <- unique(as.character(df$theory))
  df$theory <- factor(as.character(df$theory), levels = theory_levels)
  df
}

.theory_plot_label_positions <- function(df, label_nudge_x, label_nudge_y) {
  label_width <- nchar(as.character(df$theory), type = "width") * 0.006
  right_end <- df$inclS + label_nudge_x + label_width
  left_start <- df$inclS - label_nudge_x - label_width
  needs_left <- right_end > 0.985
  needs_right <- left_start < 0.015
  x_sign <- ifelse(needs_left & !needs_right, -1, 1)

  df$label_x <- pmin(pmax(df$inclS + x_sign * label_nudge_x, 0.01), 0.99)
  df$label_y <- df$covS
  df$label_hjust <- ifelse(x_sign > 0, 0, 1)
  df$label_vjust <- 0.5

  min_sep <- max(0.04, 6 * label_nudge_y)
  for (side in unique(x_sign)) {
    idx <- which(x_sign == side)
    df$label_y[idx] <- .theory_plot_spread_label_y(df$label_y[idx], min_sep = min_sep)
  }

  df$label_y <- pmin(pmax(df$label_y, 0.01), 0.99)
  df
}

.theory_plot_spread_label_y <- function(y, min_sep = 0.04) {
  if (length(y) <= 1L) {
    return(y)
  }

  order_idx <- order(y)
  sorted <- y[order_idx]
  placed <- sorted

  for (i in seq_along(placed)[-1L]) {
    placed[i] <- max(placed[i], placed[i - 1L] + min_sep)
  }

  upper_overflow <- max(placed, na.rm = TRUE) - 0.98
  if (is.finite(upper_overflow) && upper_overflow > 0) {
    placed <- placed - upper_overflow
  }

  lower_overflow <- 0.02 - min(placed, na.rm = TRUE)
  if (is.finite(lower_overflow) && lower_overflow > 0) {
    placed <- placed + lower_overflow
  }

  out <- y
  out[order_idx] <- placed
  out
}

.theory_plot_solution_labels <- function(solutions, solution_type, intermediate_branch, theories) {
  required <- c("theory", "solution_type", "intermediate_branch", "prime_implicant")
  missing <- setdiff(required, names(solutions))
  if (length(missing)) {
    stop(
      "x$results$solutions is missing required column(s): ",
      paste(missing, collapse = ", "),
      call. = FALSE
    )
  }

  keep <- solutions$solution_type == solution_type
  if (identical(solution_type, "intermediate")) {
    if (is.na(intermediate_branch)) {
      keep <- keep & is.na(solutions$intermediate_branch)
    } else {
      keep <- keep & !is.na(solutions$intermediate_branch) & solutions$intermediate_branch == intermediate_branch
    }
  }

  sol <- solutions[keep, , drop = FALSE]
  labels <- stats::setNames(rep(NA_character_, length(theories)), theories)
  if (!nrow(sol)) {
    return(labels)
  }

  for (theory in theories) {
    terms <- unique(as.character(sol$prime_implicant[sol$theory == theory]))
    terms <- terms[!is.na(terms) & nzchar(terms)]
    if (length(terms)) {
      labels[[theory]] <- paste0(theory, ": ", paste(terms, collapse = " + "))
    }
  }

  labels
}

.theory_plot_wrap_label <- function(x, width = 70L) {
  vapply(
    x,
    function(value) {
      value <- gsub(" \\+ ", " + ", value, fixed = TRUE)
      paste(strwrap(value, width = width), collapse = "\n")
    },
    character(1)
  )
}

.dot_get <- function(dots, key) {
  if (length(dots) == 0L) return(NULL)
  nms <- names(dots)
  if (is.null(nms)) return(NULL)
  i <- which(!is.na(pmatch(nms, key)))
  if (length(i) == 0L) return(NULL)
  dots[[i[1L]]]
}

.require_qca <- function() {
  if (!requireNamespace("QCA", quietly = TRUE)) {
    stop("Package 'QCA' is required.")
  }

  invisible(TRUE)
}

.safe_eval_expr <- function(expr, envir = parent.frame()) {
  tryCatch(eval(expr, envir = envir), error = function(e) e)
}

.expr_to_chrvec <- function(expr) {
  if (is.null(expr)) return(NULL)
  if (is.character(expr)) return(as.character(expr))
  if (is.name(expr)) return(as.character(expr))
  if (is.call(expr) && identical(expr[[1L]], as.name("c"))) {
    args <- as.list(expr)[-1L]
    out <- unlist(lapply(args, .expr_to_chrvec), use.names = FALSE)
    return(out)
  }
  gsub("\\s+", "", paste(deparse(expr), collapse = ""))
}

.filter_named_list <- function(x, reserved) {
  if (length(x) == 0L) return(x)
  nms <- names(x)
  if (is.null(nms)) return(x)
  m <- pmatch(nms, reserved, duplicates.ok = TRUE)
  x[is.na(m)]
}

.minimize_reserved_args <- function() {
  c(
    "input", "data", "tt",
    "outcome", "conditions",
    "incl.cut", "n.cut",
    "include", "dir.exp", "direxp",
    "exclude", "omit"
  )
}

.filter_dots_for_minimize <- function(dots) {
  .filter_named_list(dots, .minimize_reserved_args())
}

.truth_table_forwarded_args <- function() {
  setdiff(
    names(formals(QCA::truthTable)),
    c("data", "outcome", "conditions", "incl.cut", "n.cut", "show.cases", "use.labels")
  )
}

.split_truth_table_minimize_dots <- function(dots) {
  if (length(dots) == 0L) {
    return(list(tt = dots, min = dots))
  }

  nms <- names(dots)
  if (is.null(nms)) {
    return(list(tt = list(), min = dots))
  }

  is_tt <- !is.na(match(nms, .truth_table_forwarded_args()))
  dots_tt <- dots[is_tt]
  dots_rest <- dots[!is_tt]
  dots_min <- .filter_dots_for_minimize(dots_rest)

  list(tt = dots_tt, min = dots_min)
}

.reject_exclusion_controls_in_dots <- function(dots, caller) {
  if (length(dots) == 0L) return(invisible(TRUE))
  nms <- names(dots)
  if (is.null(nms)) return(invisible(TRUE))

  blocked <- intersect(nms, c("exclude", "omit", "exclude_spec"))
  if (length(blocked) > 0L) {
    stop(
      "`", caller, "()` does not accept `",
      blocked[1L],
      "` through `...`. Use `exclude_static` for an already computed exclusion object ",
      "or `exclude_recompute` for a `QCA::findRows()` specification."
    )
  }

  invisible(TRUE)
}

.reject_calibration_inputs_in_dots <- function(dots, caller) {
  if (length(dots) == 0L) return(invisible(TRUE))
  nms <- names(dots)
  if (is.null(nms)) return(invisible(TRUE))

  blocked <- intersect(
    nms,
    c(
      "thresholds",
      "type",
      "raw_conditions",
      "raw.conditions",
      "raw_condition",
      "raw.condition",
      "raw_outcome",
      "raw.outcome",
      "calibrate_args",
      "calibrate.args",
      "calib_args",
      "calib.args"
    )
  )
  if (length(blocked) > 0L) {
    stop(
      "`", caller, "()` does not accept `",
      blocked[1L],
      "` through `...`. Calibration inputs must be supplied through `calib_spec`."
    )
  }

  invisible(TRUE)
}

.qcaert_compute_exclude <- function(tt_obj, exclude_mode, exclude_recompute, exclude_static) {
  if (exclude_mode == "none") {
    return(NULL)
  }

  if (exclude_mode == "static") {
    return(exclude_static)
  }

  fr_args <- exclude_recompute
  fr_args$obj <- tt_obj

  tryCatch(
    suppressWarnings(do.call(QCA::findRows, fr_args)),
    error = function(e) e
  )
}

.validate_exclusion_controls <- function(
    exclude_mode,
    exclude_recompute,
    exclude_static,
    exclude_recompute_supplied = FALSE,
    exclude_static_supplied = FALSE,
    monitored_solutions = NULL,
    solution = NULL,
    style = c("std", "plain")
) {
  style <- match.arg(style)

  if (!is.null(monitored_solutions)) {
    needs_exclusion <- any(monitored_solutions %in% c("parsimonious", "intermediate"))
  } else if (!is.null(solution)) {
    needs_exclusion <- solution %in% c("all", "parsimonious", "intermediate")
  } else {
    needs_exclusion <- TRUE
  }

  if (identical(exclude_mode, "none")) {
    if (isTRUE(exclude_recompute_supplied) || isTRUE(exclude_static_supplied) || !is.null(exclude_static)) {
      stop("`exclude_mode = \"none\"` cannot be combined with `exclude_recompute` or `exclude_static`.")
    }
    return(invisible(TRUE))
  }

  if (identical(exclude_mode, "static")) {
    if (isTRUE(exclude_recompute_supplied)) {
      stop("`exclude_mode = \"static\"` uses `exclude_static`, not `exclude_recompute`.")
    }
    if (isTRUE(needs_exclusion) && is.null(exclude_static)) {
      stop("`exclude_mode = \"static\"` requires `exclude_static` when parsimonious or intermediate solutions are monitored.")
    }
    return(invisible(TRUE))
  }

  if (!identical(exclude_mode, "recompute")) {
    return(invisible(TRUE))
  }

  if (!is.null(exclude_static)) {
    stop("`exclude_mode = \"recompute\"` uses `exclude_recompute`, not `exclude_static`.")
  }

  if (!needs_exclusion) return(invisible(TRUE))

  if (style == "std") {
    list_msg <- "`exclude_mode = \"recompute\"` requires `exclude_recompute` as a list of `QCA::findRows()` arguments."
    expr_msg <- "For `exclude_recompute$type` 0 or 1, `exclude_recompute$expression` must be supplied."
  } else {
    list_msg <- "exclude_mode='recompute' requires exclude_recompute (a list of arguments for QCA::findRows)."
    expr_msg <- "exclude_recompute$expression is required for findRows type 0 or 1."
  }

  if (is.null(exclude_recompute) || !is.list(exclude_recompute)) {
    stop(list_msg)
  }

  if (!is.null(exclude_recompute$type) &&
      is.numeric(exclude_recompute$type) &&
      any(exclude_recompute$type %in% c(0, 1)) &&
      is.null(exclude_recompute$expression)) {
    stop(expr_msg)
  }

  invisible(TRUE)
}

test_that("truthTable/minimize dot split preserves minimize-only arguments", {
  dots <- list(all.sol = TRUE, details = TRUE)

  split <- qcaERT:::.split_truth_table_minimize_dots(dots)

  expect_length(split$tt, 0)
  expect_identical(split$min, dots)
})

test_that("truthTable/minimize dot split separates truthTable and minimize arguments", {
  dots <- list(pri.cut = 0.7, all.sol = TRUE, include = "?")

  split <- qcaERT:::.split_truth_table_minimize_dots(dots)

  expect_identical(split$tt, list(pri.cut = 0.7))
  expect_identical(split$min, list(all.sol = TRUE))
})

test_that("boundary and sampled siblings forward truthTable dots", {
  skip_if_not_installed("QCA")

  fixture <- qcaert_fixture_direct6()

  incl <- suppressWarnings(incl.test(
    data = fixture$calib,
    outcome = fixture$outcome,
    conditions = fixture$conditions,
    incl.cut = 0.75,
    step = 0.1,
    max_steps = 1,
    n.cut = 1,
    solution = "conservative",
    complete = TRUE,
    pri.cut = 0.6,
    all.sol = TRUE,
    progress = FALSE
  ))

  expect_true(incl$baseline$tt$options$complete)
  expect_equal(incl$baseline$tt$options$pri.cut, 0.6)

  ncut <- suppressWarnings(ncut.test(
    data = fixture$calib,
    outcome = fixture$outcome,
    conditions = fixture$conditions,
    n.cut = 1,
    step = 1,
    max_steps = 1,
    incl.cut = 0.75,
    solution = "conservative",
    complete = TRUE,
    pri.cut = 0.6,
    all.sol = TRUE,
    progress = FALSE
  ))

  expect_true(ncut$baseline$tt$options$complete)
  expect_equal(ncut$baseline$tt$options$pri.cut, 0.6)

  altset <- suppressWarnings(altset.test(
    raw.data = fixture$raw,
    calib.data = fixture$calib,
    outcome = fixture$outcome,
    conditions = fixture$conditions,
    calib_spec = fixture$calib_spec,
    test.conditions = fixture$test.conditions,
    anchors_to_test = "E1",
    unit_step = 1,
    calib_max_steps = 1,
    incl.cut = 0.75,
    incl_step = 0.1,
    incl_max_steps = 1,
    n.cut = 1,
    ncut_step = 1,
    ncut_max_steps = 1,
    n_draws = 1,
    seed = 101,
    solution = "conservative",
    complete = TRUE,
    pri.cut = 0.6,
    all.sol = TRUE,
    progress = FALSE
  ))

  expect_true(altset$baseline$result$tt$options$complete)
  expect_equal(altset$baseline$result$tt$options$pri.cut, 0.6)
})

test_that("calib.test accepts split truthTable and minimize dots", {
  skip_if_not_installed("QCA")

  fixture <- qcaert_fixture_direct6()

  out <- suppressWarnings(calib.test(
    raw.data = fixture$raw,
    calib.data = fixture$calib,
    outcome = fixture$outcome,
    conditions = fixture$conditions,
    calib_spec = fixture$calib_spec,
    test.conditions = fixture$test.conditions,
    anchors_to_test = "E1",
    unit_step = 1,
    max_steps = 1,
    incl.cut = 0.75,
    n.cut = 1,
    solution = "conservative",
    complete = TRUE,
    pri.cut = 0.6,
    all.sol = TRUE,
    progress = FALSE
  ))

  expect_s3_class(out, "calib_test")
  expect_identical(out$baseline$status, "ok")
})

test_that("calibration robustness functions expose only calib_spec for calibration inputs", {
  fns <- c("calib.test", "altset.test", "loo.test", "subsample.test")
  blocked <- c(
    "thresholds",
    "type",
    "raw_conditions",
    "raw.conditions",
    "raw_condition",
    "raw.condition",
    "raw_outcome",
    "raw.outcome",
    "calibrate_args",
    "calibrate.args",
    "calib_args",
    "calib.args"
  )

  for (fn in fns) {
    args <- names(formals(get(fn, envir = asNamespace("qcaERT"))))
    expect_true("calib_spec" %in% args, info = fn)
    expect_false(any(blocked %in% args), info = fn)
  }
})

test_that("top-level calibration inputs are rejected through dots", {
  expect_error(
    .reject_calibration_inputs_in_dots(
      list(thresholds = list(A = c(1, 2, 3))),
      "calib.test"
    ),
    "Calibration inputs must be supplied through `calib_spec`.",
    fixed = TRUE
  )

  expect_error(
    .reject_calibration_inputs_in_dots(
      list(calibrate_args = list(logistic = FALSE)),
      "altset.test"
    ),
    "Calibration inputs must be supplied through `calib_spec`.",
    fixed = TRUE
  )
})

test_that("public calibration functions reject top-level calibration inputs before analysis", {
  skip_if_not_installed("QCA")

  fixture <- qcaert_fixture_direct6()
  msg <- "Calibration inputs must be supplied through `calib_spec`."

  expect_error(
    calib.test(
      raw.data = fixture$raw,
      calib.data = fixture$calib,
      outcome = fixture$outcome,
      conditions = fixture$conditions,
      calib_spec = fixture$calib_spec,
      test.conditions = fixture$test.conditions,
      solution = "conservative",
      thresholds = list(A = c(10, 20, 30)),
      progress = FALSE
    ),
    msg,
    fixed = TRUE
  )

  expect_error(
    altset.test(
      raw.data = fixture$raw,
      calib.data = fixture$calib,
      outcome = fixture$outcome,
      conditions = fixture$conditions,
      calib_spec = fixture$calib_spec,
      test.conditions = fixture$test.conditions,
      solution = "conservative",
      n_draws = 1,
      type = "fuzzy",
      progress = FALSE
    ),
    msg,
    fixed = TRUE
  )

  expect_error(
    loo.test(
      data = fixture$calib,
      outcome = fixture$outcome,
      conditions = fixture$conditions,
      cases = 1,
      calib = "fixed",
      solution = "conservative",
      calibrate_args = list(logistic = FALSE),
      progress = FALSE
    ),
    msg,
    fixed = TRUE
  )

  expect_error(
    subsample.test(
      data = fixture$calib,
      outcome = fixture$outcome,
      conditions = fixture$conditions,
      calib = "fixed",
      sample_n = 8,
      reps = 1,
      solution = "conservative",
      raw_conditions = c(A = "A_raw"),
      progress = FALSE
    ),
    msg,
    fixed = TRUE
  )
})

test_that("old exclusion spellings are rejected through dots", {
  expect_error(
    .reject_exclusion_controls_in_dots(list(exclude_spec = list(type = 2)), "incl.test"),
    "Use `exclude_static`",
    fixed = TRUE
  )

  expect_error(
    .reject_exclusion_controls_in_dots(list(exclude = 1:2), "ncut.test"),
    "Use `exclude_static`",
    fixed = TRUE
  )

  expect_error(
    .reject_exclusion_controls_in_dots(list(omit = 1:2), "theory.test"),
    "Use `exclude_static`",
    fixed = TRUE
  )
})

test_that("plot solution selectors are reserved for solution_type", {
  skip_if_not_installed("QCA")
  skip_if_not_installed("ggplot2")

  fixture <- qcaert_fixture_direct6()
  out <- suppressWarnings(incl.test(
    data = fixture$calib,
    outcome = fixture$outcome,
    conditions = fixture$conditions,
    incl.cut = 0.75,
    step = 0.1,
    max_steps = 1,
    n.cut = 1,
    solution = "all",
    dir.exp = fixture$dir.exp,
    progress = FALSE
  ))

  expect_error(
    plot(out, solution = "conservative"),
    "Use `solution_type`",
    fixed = TRUE
  )
  expect_error(
    plot(out, monitored_solutions = "conservative"),
    "Use `solution_type`",
    fixed = TRUE
  )
})

.boundary_error_message <- function(e) {
  if (!inherits(e, "error")) return(NA_character_)
  tryCatch(conditionMessage(e), error = function(...) as.character(e))
}

.boundary_compute_exclude <- function(tt_obj, exclude_mode, exclude_recompute, exclude_static) {
  .qcaert_compute_exclude(
    tt_obj = tt_obj,
    exclude_mode = exclude_mode,
    exclude_recompute = exclude_recompute,
    exclude_static = exclude_static
  )
}

.boundary_run_solution_type <- function(
    tt_obj,
    solution_type,
    ex,
    dots_filtered,
    dir.exp,
    which_M,
    i_mode
) {
  args <- c(
    list(
      input = tt_obj,
      include = if (solution_type == "conservative") "" else "?"
    ),
    dots_filtered
  )

  if (solution_type == "intermediate") {
    args$dir.exp <- dir.exp
  }

  if (solution_type %in% c("parsimonious", "intermediate") && !is.null(ex)) {
    args$exclude <- ex
  }

  res <- tryCatch(
    suppressWarnings(do.call(QCA::minimize, args)),
    error = function(e) e
  )

  if (inherits(res, "error")) {
    return(list(
      error = res,
      res = NULL,
      sig = NULL,
      selected_solution_missing = NA,
      meta = NULL
    ))
  }

  sig_info <- .extract_solution_type_sig(
    res,
    solution_type = solution_type,
    which_M = which_M,
    i_mode = i_mode
  )

  list(
    error = NULL,
    res = res,
    sig = sig_info$sig,
    selected_solution_missing = isTRUE(sig_info$selected_solution_missing),
    meta = sig_info$meta
  )
}

.boundary_empty_solution_type_list <- function() {
  list(conservative = NULL, parsimonious = NULL, intermediate = NULL)
}

.boundary_empty_selected_solution_missing <- function() {
  c(conservative = NA, parsimonious = NA, intermediate = NA)
}

.boundary_run_once <- function(
    value,
    build_truth_table,
    monitored_solutions,
    exclude_mode,
    exclude_recompute,
    exclude_static,
    dots_filtered,
    dir.exp,
    which_M,
    i_mode
) {
  tt_obj <- tryCatch(
    suppressWarnings(build_truth_table(value)),
    error = function(e) e
  )

  if (inherits(tt_obj, "error")) {
    return(list(
      status = "truth_table_error",
      tt = NULL,
      exclude_used = NULL,
      res = .boundary_empty_solution_type_list(),
      sig = .boundary_empty_solution_type_list(),
      selected_solution_missing = .boundary_empty_selected_solution_missing(),
      meta = .boundary_empty_solution_type_list(),
      error_source = "truthTable",
      error_message = .boundary_error_message(tt_obj)
    ))
  }

  need_exclude <- any(monitored_solutions %in% c("parsimonious", "intermediate")) &&
    exclude_mode != "none"

  ex <- NULL
  if (need_exclude) {
    ex <- .boundary_compute_exclude(
      tt_obj = tt_obj,
      exclude_mode = exclude_mode,
      exclude_recompute = exclude_recompute,
      exclude_static = exclude_static
    )

    if (inherits(ex, "error")) {
      return(list(
        status = "exclude_error",
        tt = tt_obj,
        exclude_used = NULL,
        res = .boundary_empty_solution_type_list(),
        sig = .boundary_empty_solution_type_list(),
        selected_solution_missing = .boundary_empty_selected_solution_missing(),
        meta = .boundary_empty_solution_type_list(),
        error_source = "exclude",
        error_message = .boundary_error_message(ex)
      ))
    }
  }

  res_out <- .boundary_empty_solution_type_list()
  sig_out <- .boundary_empty_solution_type_list()
  selected_solution_out <- .boundary_empty_selected_solution_missing()
  meta_out <- .boundary_empty_solution_type_list()

  for (solution_type in monitored_solutions) {
    rr <- .boundary_run_solution_type(
      tt_obj = tt_obj,
      solution_type = solution_type,
      ex = ex,
      dots_filtered = dots_filtered,
      dir.exp = dir.exp,
      which_M = which_M,
      i_mode = i_mode
    )

    if (!is.null(rr$error)) {
      return(list(
        status = "minimize_error",
        tt = tt_obj,
        exclude_used = ex,
        res = res_out,
        sig = sig_out,
        selected_solution_missing = selected_solution_out,
        meta = meta_out,
        error_source = solution_type,
        error_message = .boundary_error_message(rr$error)
      ))
    }

    res_out[[solution_type]] <- rr$res
    sig_out[[solution_type]] <- rr$sig
    selected_solution_out[[solution_type]] <- isTRUE(rr$selected_solution_missing)
    meta_out[[solution_type]] <- rr$meta
  }

  list(
    status = "ok",
    tt = tt_obj,
    exclude_used = ex,
    res = res_out,
    sig = sig_out,
    selected_solution_missing = selected_solution_out,
    meta = meta_out,
    error_source = NA_character_,
    error_message = NA_character_
  )
}

.boundary_solution_change_info <- function(base, cur, monitored_solutions) {
  changed_flags <- vapply(
    monitored_solutions,
    function(solution_type) .sig_changed(base$sig[[solution_type]], cur$sig[[solution_type]]),
    logical(1)
  )

  changed <- any(changed_flags)

  if (!changed) {
    return(list(
      changed = FALSE,
      changed_types = NA_character_,
      change_kind = NA_character_
    ))
  }

  changed_types <- paste(monitored_solutions[changed_flags], collapse = ",")
  change_kind <- paste(
    vapply(
      monitored_solutions[changed_flags],
      function(solution_type) {
        paste0(solution_type, ":", .change_kind_sig(base$sig[[solution_type]], cur$sig[[solution_type]]))
      },
      character(1)
    ),
    collapse = ","
  )

  list(
    changed = TRUE,
    changed_types = changed_types,
    change_kind = change_kind
  )
}

.boundary_exclude_count <- function(x) {
  if (is.null(x)) return(0L)
  length(x)
}

.boundary_exclude_chr <- function(x) {
  if (is.null(x)) return(NA_character_)
  if (length(x) == 0L) return("")
  paste(as.character(x), collapse = ",")
}

.boundary_stop_reason_from_status <- function(status, direction = NULL, baseline = FALSE) {
  out <- switch(
    status,
    truth_table_error = "truth_table_build_error",
    exclude_error = "exclude_recompute_error",
    minimize_error = "requested_minimize_error",
    boundary = if (!is.null(direction)) paste0("search_boundary_", direction) else "search_boundary",
    run_budget_exhausted = "search_budget_exhausted",
    status
  )

  if (isTRUE(baseline)) {
    out <- paste0("baseline_", out)
  }

  out
}

.boundary_cast_value <- function(value, value_type) {
  value_type <- match.arg(value_type, c("numeric", "integer"))

  if (length(value) == 0L || is.na(value)) {
    return(switch(value_type, numeric = NA_real_, integer = NA_integer_))
  }

  switch(
    value_type,
    numeric = as.numeric(value),
    integer = as.integer(value)
  )
}

.boundary_make_search_row <- function(
    direction,
    solution,
    solution_types,
    solution_type,
    i_mode,
    which_M,
    value_name,
    value_type,
    start_value,
    last_safe_value,
    failing_value,
    steps_done,
    total_delta,
    stop_reason,
    changed_types,
    change_kind,
    error_source,
    error_message,
    baseline_exclude,
    last_safe_exclude,
    failing_exclude,
    extra_row
) {
  value_na <- .boundary_cast_value(NA, value_type)

  row_fields <- list(
    direction = direction,
    solution = solution,
    monitored_solutions = paste(solution_types, collapse = ","),
    i_mode = if ("intermediate" %in% solution_types) i_mode else NA_character_,
    which_M = which_M
  )
  row_fields[[paste0(value_name, "_start")]] <- .boundary_cast_value(start_value, value_type)
  row_fields[[paste0(value_name, "_last_safe")]] <- if (is.null(last_safe_value)) value_na else .boundary_cast_value(last_safe_value, value_type)
  row_fields[[paste0(value_name, "_first_failing")]] <- if (is.null(failing_value)) value_na else .boundary_cast_value(failing_value, value_type)
  row_fields$number_of_steps <- steps_done
  row_fields$total_delta <- if (is.null(total_delta)) value_na else .boundary_cast_value(total_delta, value_type)
  row_fields$stop_reason <- stop_reason
  row_fields$changed_types <- changed_types
  row_fields$change_kind <- change_kind
  row_fields$error_source <- error_source
  row_fields$error_message <- error_message
  row_fields$n_exclude_baseline <- .boundary_exclude_count(baseline_exclude)
  row_fields$n_exclude_last_safe <- if (is.null(last_safe_value)) NA_integer_ else .boundary_exclude_count(last_safe_exclude)
  row_fields$n_exclude_first_failing <- if (is.null(failing_value)) NA_integer_ else .boundary_exclude_count(failing_exclude)
  row_fields$exclude_baseline <- .boundary_exclude_chr(baseline_exclude)
  row_fields$exclude_last_safe <- if (is.null(last_safe_value)) NA_character_ else .boundary_exclude_chr(last_safe_exclude)
  row_fields$exclude_first_failing <- if (is.null(failing_value)) NA_character_ else .boundary_exclude_chr(failing_exclude)
  row_fields <- c(row_fields, extra_row)

  row <- do.call(data.frame, c(row_fields, list(stringsAsFactors = FALSE)))

  if (!is.null(solution_type)) {
    row$solution_type <- solution_type
  }

  row
}

.boundary_search_direction <- function(
    direction,
    value_name,
    value_type,
    start_value,
    step,
    max_steps,
    lower_limit,
    upper_limit,
    run_once,
    solution,
    solution_types,
    baseline,
    solution_type,
    i_mode,
    which_M,
    extra_row = list(),
    extra_path = list()
) {
  value_type <- match.arg(value_type, c("numeric", "integer"))
  value_na <- .boundary_cast_value(NA, value_type)
  cur_value <- .boundary_cast_value(start_value, value_type)
  steps_done <- 0L
  last_safe_value <- .boundary_cast_value(start_value, value_type)
  last_safe_exclude <- baseline$exclude_used

  failing_value <- value_na
  failing_exclude <- NULL

  stop_reason <- NA_character_
  changed_types <- NA_character_
  change_kind <- NA_character_
  error_source <- NA_character_
  error_message <- NA_character_

  trace <- .empty_change_trace(value_name, value_type = value_type)

  baseline_missing <- FALSE
  if (baseline$status == "ok") {
    baseline_missing <- any(unname(baseline$selected_solution_missing[solution_types]) %in% TRUE)
  }

  baseline_invalid <- baseline$status != "ok" || baseline_missing

  if (baseline_invalid) {
    stop_reason <- if (baseline$status != "ok") {
      .boundary_stop_reason_from_status(baseline$status, baseline = TRUE)
    } else {
      "baseline_selected_solution_missing"
    }
    error_source <- baseline$error_source
    error_message <- baseline$error_message

    row <- .boundary_make_search_row(
      direction = direction,
      solution = solution,
      solution_types = solution_types,
      solution_type = solution_type,
      i_mode = i_mode,
      which_M = which_M,
      value_name = value_name,
      value_type = value_type,
      start_value = start_value,
      last_safe_value = NULL,
      failing_value = NULL,
      steps_done = NA_integer_,
      total_delta = NULL,
      stop_reason = stop_reason,
      changed_types = NA_character_,
      change_kind = NA_character_,
      error_source = error_source,
      error_message = error_message,
      baseline_exclude = baseline$exclude_used,
      last_safe_exclude = NULL,
      failing_exclude = NULL,
      extra_row = extra_row
    )

    path <- c(
      list(
        direction = direction,
        solution = solution,
        monitored_solutions = solution_types,
        solution_type = solution_type,
        trace = trace,
        stop_reason = stop_reason,
        changed_types = NA_character_,
        change_kind = NA_character_,
        last_safe_value = value_na,
        failing_value = value_na,
        baseline = baseline,
        last_safe_exclude = NULL,
        failing_exclude = NULL
      ),
      extra_path
    )

    return(list(row = row, path = path))
  }

  for (k in seq_len(max_steps)) {
    next_value <- if (direction == "lower") cur_value - step else cur_value + step

    if (next_value < lower_limit || next_value > upper_limit) {
      stop_reason <- .boundary_stop_reason_from_status("boundary", direction = direction)
      failing_value <- .boundary_cast_value(next_value, value_type)
      break
    }

    cur <- run_once(next_value, solution_types = solution_types)

    if (cur$status != "ok") {
      stop_reason <- .boundary_stop_reason_from_status(cur$status, direction = direction)
      failing_value <- .boundary_cast_value(next_value, value_type)
      failing_exclude <- cur$exclude_used
      error_source <- cur$error_source
      error_message <- cur$error_message

      trace <- .append_change_trace(
        trace = trace,
        step = k,
        value_col = value_name,
        value = .boundary_cast_value(next_value, value_type),
        changed = NA,
        status = cur$status,
        change_kind = NA_character_
      )
      break
    }

    change_info <- .boundary_solution_change_info(
      base = baseline,
      cur = cur,
      monitored_solutions = solution_types
    )

    changed <- isTRUE(change_info$changed)

    trace <- .append_change_trace(
      trace = trace,
      step = k,
      value_col = value_name,
      value = .boundary_cast_value(next_value, value_type),
      changed = changed,
      status = cur$status,
      change_kind = if (changed) change_info$change_kind else NA_character_
    )

    if (changed) {
      stop_reason <- "solution_change"
      changed_types <- change_info$changed_types
      change_kind <- change_info$change_kind
      failing_value <- .boundary_cast_value(next_value, value_type)
      failing_exclude <- cur$exclude_used
      break
    }

    cur_value <- .boundary_cast_value(next_value, value_type)
    last_safe_value <- .boundary_cast_value(next_value, value_type)
    last_safe_exclude <- cur$exclude_used
    steps_done <- steps_done + 1L
  }

  if (is.na(stop_reason)) {
    stop_reason <- .boundary_stop_reason_from_status("run_budget_exhausted")
  }

  row <- .boundary_make_search_row(
    direction = direction,
    solution = solution,
    solution_types = solution_types,
    solution_type = solution_type,
    i_mode = i_mode,
    which_M = which_M,
    value_name = value_name,
    value_type = value_type,
    start_value = start_value,
    last_safe_value = last_safe_value,
    failing_value = failing_value,
    steps_done = as.integer(steps_done),
    total_delta = last_safe_value - start_value,
    stop_reason = stop_reason,
    changed_types = changed_types,
    change_kind = change_kind,
    error_source = error_source,
    error_message = error_message,
    baseline_exclude = baseline$exclude_used,
    last_safe_exclude = last_safe_exclude,
    failing_exclude = failing_exclude,
    extra_row = extra_row
  )

  path <- c(
    list(
      direction = direction,
      solution = solution,
      monitored_solutions = solution_types,
      solution_type = solution_type,
      trace = trace,
      stop_reason = stop_reason,
      changed_types = changed_types,
      change_kind = change_kind,
      last_safe_value = .boundary_cast_value(last_safe_value, value_type),
      failing_value = .boundary_cast_value(failing_value, value_type),
      baseline = baseline,
      last_safe_exclude = last_safe_exclude,
      failing_exclude = failing_exclude
    ),
    extra_path
  )

  list(row = row, path = path)
}

.boundary_search <- function(
    value_name,
    value_type,
    start_value,
    step,
    max_steps,
    lower_limit,
    upper_limit,
    run_once,
    solution,
    monitored_solutions,
    i_mode,
    which_M,
    progress_tick,
    extra_row = list(),
    extra_path = list()
) {
  diag_rows <- list()
  by_direction <- list()

  if (solution == "all") {
    baseline_by_solution_type <- setNames(
      lapply(monitored_solutions, function(solution_type) run_once(start_value, solution_types = solution_type)),
      monitored_solutions
    )

    baseline <- list(
      by_solution_type = baseline_by_solution_type,
      status = vapply(baseline_by_solution_type, function(x) x$status, character(1)),
      exclude_used = lapply(baseline_by_solution_type, function(x) x$exclude_used),
      sig = lapply(baseline_by_solution_type, function(x) x$sig),
      selected_solution_missing = lapply(baseline_by_solution_type, function(x) x$selected_solution_missing),
      meta = lapply(baseline_by_solution_type, function(x) x$meta)
    )

    for (direction in c("lower", "upper")) {
      paths <- list()
      for (solution_type in monitored_solutions) {
        searched <- .boundary_search_direction(
          direction = direction,
          value_name = value_name,
          value_type = value_type,
          start_value = start_value,
          step = step,
          max_steps = max_steps,
          lower_limit = lower_limit,
          upper_limit = upper_limit,
          run_once = run_once,
          solution = solution,
          solution_types = solution_type,
          baseline = baseline_by_solution_type[[solution_type]],
          solution_type = solution_type,
          i_mode = i_mode,
          which_M = which_M,
          extra_row = extra_row,
          extra_path = extra_path
        )
        diag_rows[[length(diag_rows) + 1L]] <- searched$row
        paths[[solution_type]] <- searched$path
        progress_tick()
      }

      first_path <- paths[[monitored_solutions[1L]]]
      by_direction[[direction]] <- c(
        list(
          direction = direction,
          solution = solution,
          monitored_solutions = monitored_solutions,
          by_solution_type = paths
        ),
        first_path[setdiff(names(first_path), c("direction", "solution", "monitored_solutions"))]
      )
    }
  } else {
    baseline <- run_once(start_value)

    for (direction in c("lower", "upper")) {
      searched <- .boundary_search_direction(
        direction = direction,
        value_name = value_name,
        value_type = value_type,
        start_value = start_value,
        step = step,
        max_steps = max_steps,
        lower_limit = lower_limit,
        upper_limit = upper_limit,
        run_once = run_once,
        solution = solution,
        solution_types = monitored_solutions,
        baseline = baseline,
        solution_type = NULL,
        i_mode = i_mode,
        which_M = which_M,
        extra_row = extra_row,
        extra_path = extra_path
      )
      diag_rows[[length(diag_rows) + 1L]] <- searched$row
      by_direction[[direction]] <- searched$path
      progress_tick()
    }
  }

  list(
    diagnostics = .bind_rows_result(diag_rows),
    baseline = baseline,
    by_direction = by_direction
  )
}

.fit_get <- function(x, nm) {
  if (is.null(x) || length(x) == 0L) return(NULL)
  nms <- names(x)
  if (is.null(nms) || !nm %in% nms) return(NULL)
  x[[match(nm, nms)]]
}

.fit_exceeds_tol <- function(delta, tol = 0) {
  if (is.na(delta)) return(NA)
  abs(delta) > (tol + sqrt(.Machine$double.eps))
}

.fit_changed <- function(base, cur, tol = 0) {
  if (is.null(base) && is.null(cur)) return(FALSE)
  if (xor(is.null(base), is.null(cur))) return(TRUE)

  nms <- union(names(base), names(cur))
  if (length(nms) == 0L) return(FALSE)

  for (nm in nms) {
    b <- .fit_get(base, nm)
    c <- .fit_get(cur, nm)

    if (is.null(b) && is.null(c)) next
    if (is.null(b) || is.null(c)) return(TRUE)
    if (xor(is.na(b), is.na(c))) return(TRUE)
    if (is.na(b) && is.na(c)) next
    if (isTRUE(.fit_exceeds_tol(c - b, tol = tol))) return(TRUE)
    }

  FALSE
}

.fit_change_names <- function(base, cur, tol = 0) {
  if (is.null(base) && is.null(cur)) return(character(0))
  if (xor(is.null(base), is.null(cur))) {
    return(union(names(base), names(cur)))
  }

  nms <- union(names(base), names(cur))
  out <- character(0)

  for (nm in nms) {
    b <- .fit_get(base, nm)
    c <- .fit_get(cur, nm)

    if (is.null(b) && is.null(c)) {
      next
    } else if (is.null(b) || is.null(c)) {
      out <- c(out, nm)
    } else if (xor(is.na(b), is.na(c))) {
      out <- c(out, nm)
    } else if (!(is.na(b) && is.na(c)) && isTRUE(.fit_exceeds_tol(c - b, tol = tol))) {
      out <- c(out, nm)
    }
  }

  sort(unique(out))
}

.fit_delta_vec <- function(base, cur) {
  if (is.null(base) && is.null(cur)) return(numeric(0))

  nms <- union(names(base), names(cur))
  out <- numeric(0)

  for (nm in nms) {
    b <- .fit_get(base, nm)
    c <- .fit_get(cur, nm)

    val <- if (is.null(b) && is.null(c)) {
      NA_real_
    } else if (is.null(b) || is.null(c) || xor(is.na(b), is.na(c))) {
      NA_real_
    } else if (is.na(b) && is.na(c)) {
      NA_real_
    } else {
      c - b
    }

    out <- c(out, val)
    names(out)[length(out)] <- nm
  }

  out
}

.fit_abs_max <- function(x) {
  if (is.null(x) || length(x) == 0L) return(NA_real_)
  x <- x[is.finite(x)]
  if (length(x) == 0L) return(NA_real_)
  max(abs(x))
}

.fit_vec_from_ic <- function(ic_obj, which_m) {
  if (is.null(ic_obj) || !is.list(ic_obj)) return(NULL)

  if ("individual" %in% names(ic_obj) &&
      length(ic_obj$individual) >= which_m &&
      !is.null(ic_obj$individual[[which_m]]) &&
      is.list(ic_obj$individual[[which_m]]) &&
      "sol.incl.cov" %in% names(ic_obj$individual[[which_m]])) {
    sic <- ic_obj$individual[[which_m]]$sol.incl.cov
    if (is.data.frame(sic)) sic <- as.matrix(sic)
    if (is.matrix(sic)) {
      if (nrow(sic) < 1L) return(NULL)
      vec <- as.numeric(sic[1L, , drop = TRUE])
      names(vec) <- colnames(sic)
      return(vec)
    }
    vec <- as.numeric(sic)
    names(vec) <- names(sic)
    return(vec)
  }

  if (!("sol.incl.cov" %in% names(ic_obj))) return(NULL)
  sic <- ic_obj$sol.incl.cov
  if (is.data.frame(sic)) sic <- as.matrix(sic)

  if (is.matrix(sic)) {
    if (nrow(sic) < which_m) return(NULL)
    vec <- as.numeric(sic[which_m, , drop = TRUE])
    names(vec) <- colnames(sic)
    return(vec)
  }

  if (which_m != 1L) return(NULL)
  vec <- as.numeric(sic)
  names(vec) <- names(sic)
  vec
}

.compare_named_fit_vectors <- function(base_vec, cur_vec, tol) {
  if (is.null(base_vec) || is.null(cur_vec)) {
    return(list(same_fit = NA, max_abs_delta = NA_real_, deltas = NULL))
  }
  if (is.null(names(base_vec)) || is.null(names(cur_vec))) {
    return(list(same_fit = NA, max_abs_delta = NA_real_, deltas = NULL))
  }
  common <- intersect(names(base_vec), names(cur_vec))
  if (length(common) == 0L) {
    return(list(same_fit = NA, max_abs_delta = NA_real_, deltas = NULL))
  }
  d <- cur_vec[common] - base_vec[common]
  ok <- !is.na(d)
  if (!any(ok)) {
    return(list(same_fit = NA, max_abs_delta = NA_real_, deltas = d))
  }
  mad <- max(abs(d[ok]))
  list(
    same_fit = isTRUE(mad <= (tol + sqrt(.Machine$double.eps))),
    max_abs_delta = mad,
    deltas = d
  )
  }

.flatten_fit_map <- function(fit_map, solution_type) {
  if (is.null(fit_map) || length(fit_map) == 0L) {
    return(setNames(numeric(0), character(0)))
  }

  out <- numeric(0)

  for (sol_name in names(fit_map)) {
    vec_raw <- fit_map[[sol_name]]

    if (is.null(vec_raw) || length(vec_raw) == 0L) {
      next
    }

    vec <- as.numeric(vec_raw)
    measure_names <- names(vec_raw)

    if (is.null(measure_names) || any(measure_names == "")) {
      measure_names <- paste0("V", seq_along(vec))
    }

    names(vec) <- paste(solution_type, sol_name, measure_names, sep = "::")
    out <- c(out, vec)
  }

  out
}

.common_fit_keys <- function(base_map, cur_map) {
  intersect(names(base_map), names(cur_map))
}

.group_fit_delta_by_solution <- function(delta_vec, solution_type) {
  if (is.null(delta_vec) || length(delta_vec) == 0L) {
    return(list())
  }

  nms <- names(delta_vec)

  if (is.null(nms) || length(nms) == 0L) {
    return(list())
  }

  prefix <- paste0(solution_type, "::")
  keep <- startsWith(nms, prefix)

  if (!any(keep)) {
    return(list())
  }

  vals <- delta_vec[keep]
  labs <- nms[keep]
  rest <- substring(labs, nchar(prefix) + 1L)

  parts <- strsplit(rest, "::", fixed = TRUE)

  sol_names <- vapply(
    parts,
    function(x) if (length(x) >= 1L) x[1L] else "",
    character(1)
  )

  measure_names <- vapply(
    parts,
    function(x) if (length(x) >= 2L) paste(x[-1L], collapse = "::") else "",
    character(1)
  )

  out <- split(vals, sol_names)

  for (sol_name in names(out)) {
    idx <- sol_names == sol_name
    names(out[[sol_name]]) <- measure_names[idx]
  }

  out
}

.fit_details_from_delta <- function(delta_vec, solution_type, tol = 0) {
  grouped <- .group_fit_delta_by_solution(delta_vec, solution_type)

  if (length(grouped) == 0L) {
    return(NULL)
  }

  lapply(
    grouped,
    function(d) {
      mad <- .fit_abs_max(d)

      list(
        same_fit = if (all(is.na(d))) NA else isTRUE(mad <= (tol + sqrt(.Machine$double.eps))),
        max_abs_delta = mad,
        deltas = d
      )
    }
  )
}

.compare_fit_maps_hybrid <- function(base_map, cur_map, solution_type, tol = 0) {
  common_keys <- .common_fit_keys(base_map, cur_map)

  if (length(common_keys) == 0L) {
    return(list(
      fit_compared = FALSE,
      same_fit = NA,
      max_abs_delta = NA_real_,
      matched_keys = character(0),
      details = NULL,
      changed_names = character(0),
      delta = numeric(0)
    ))
  }

  base_flat <- .flatten_fit_map(base_map[common_keys], solution_type)
  cur_flat  <- .flatten_fit_map(cur_map[common_keys], solution_type)

  if (length(base_flat) == 0L || length(cur_flat) == 0L) {
    return(list(
      fit_compared = FALSE,
      same_fit = NA,
      max_abs_delta = NA_real_,
      matched_keys = common_keys,
      details = NULL,
      changed_names = character(0),
      delta = numeric(0)
    ))
  }

  delta_vec <- .fit_delta_vec(base_flat, cur_flat)

  list(
    fit_compared = TRUE,
    same_fit = !.fit_changed(base_flat, cur_flat, tol = tol),
    max_abs_delta = .fit_abs_max(delta_vec),
    matched_keys = common_keys,
    details = .fit_details_from_delta(delta_vec, solution_type = solution_type, tol = tol),
    changed_names = .fit_change_names(base_flat, cur_flat, tol = tol),
    delta = delta_vec
  )
}

test_that("hybrid fit comparison uses all matched solution names", {
  base_map <- list(
    "M1:A*B + C" = c(inclS = 0.84, PRI = 0.79, covS = 0.52),
    "M2:A + ~B" = c(inclS = 0.81, PRI = 0.75, covS = 0.49)
  )

  cur_map <- list(
    "M1:A*B + C" = c(inclS = 0.85, PRI = 0.79, covS = 0.50),
    "M2:A + ~B" = c(inclS = 0.84, PRI = 0.76, covS = 0.48)
  )

  cmp <- qcaERT:::.compare_fit_maps_hybrid(
    base_map = base_map,
    cur_map = cur_map,
    solution_type = "conservative",
    tol = 0.02
  )

  expect_true(cmp$fit_compared)
  expect_false(cmp$same_fit)

  expect_identical(
    cmp$matched_keys,
    c("M1:A*B + C", "M2:A + ~B")
  )

  expect_equal(cmp$max_abs_delta, 0.03, tolerance = 1e-12)

  expect_setequal(
    names(cmp$details),
    c("M1:A*B + C", "M2:A + ~B")
  )

  expect_identical(
    cmp$changed_names,
    "conservative::M2:A + ~B::inclS"
  )

  expect_equal(
    unname(cmp$details[["M1:A*B + C"]][["deltas"]]["covS"]),
    -0.02,
    tolerance = 1e-12
  )

  expect_equal(
    unname(cmp$details[["M2:A + ~B"]][["deltas"]]["inclS"]),
    0.03,
    tolerance = 1e-12
  )
})

test_that("hybrid fit comparison returns not compared when no solution names match", {
  base_map <- list(
    "M1:A*B + C" = c(inclS = 0.84, PRI = 0.79)
  )

  cur_map <- list(
    "M2:A + ~B" = c(inclS = 0.84, PRI = 0.79)
  )

  cmp <- qcaERT:::.compare_fit_maps_hybrid(
    base_map = base_map,
    cur_map = cur_map,
    solution_type = "conservative",
    tol = 0.02
  )

  expect_false(cmp$fit_compared)
  expect_true(is.na(cmp$same_fit))
  expect_identical(cmp$matched_keys, character(0))
  expect_null(cmp$details)
})

#' Truth table frequency-cutoff robustness test for QCA solutions
#'
#' Perturbs the truth table frequency cutoff used in the analysis and checks
#' when the monitored QCA solution changes. Starting from a baseline `n.cut`,
#' the function searches downward and upward in fixed integer steps and records
#' the last value that preserved the baseline solution, the first value that
#' changed the solution or triggered an error, and the reason the search
#' stopped in each direction.
#'
#' @param data A non-empty data frame containing the outcome and condition
#'   columns used in the QCA analysis.
#' @param outcome Name of the outcome. This must be a single
#'   non-empty character string.
#' @param conditions Optional character vector of condition names. If `NULL`,
#'   the condition set is left to [QCA::truthTable()].
#' @param n.cut Baseline truth table frequency cutoff passed to
#'   [QCA::truthTable()]. This must be an integer-like value of at least `1`.
#' @param step Positive integer step size used to move `n.cut` downward and
#'   upward from the baseline value.
#' @param max_steps Maximum number of stepwise moves to attempt in each
#'   direction.
#' @param incl.cut Inclusion cutoff passed to [QCA::truthTable()]. This must
#'   be a single finite number in `[0, 1]`.
#' @param solution Solution type to monitor. Accepted values are `"all"`,
#'   `"con"` or `"conservative"`, `"par"` or `"parsimonious"`, and `"int"` or
#'   `"intermediate"`.
#' @param include Optional minimization include setting. Currently, this
#'   argument accepts only `NULL`, `""`, or `"?"`.
#' @param dir.exp Directional expectations used when the monitored solution is
#'   intermediate.
#' @param which_M Positive integer giving which solution alternative to use
#'   when minimization returns multiple models.
#' @param i_mode Character string controlling intermediate-solution selection.
#'   Accepted values are `"all"` and `"C1P1"`.
#' @param exclude_mode Character string controlling how excluded rows are
#'   handled for parsimonious and intermediate minimization. `"recompute"`
#'   recalculates exclusions from each truth table, `"static"` reuses
#'   `exclude_static`, and `"none"` does not use exclusions.
#' @param exclude_recompute Named list of arguments passed to [QCA::findRows()]
#'   when `exclude_mode = "recompute"`.
#' @param exclude_static Already computed exclusion object reused when
#'   `exclude_mode = "static"`.
#' @param result_shape Layout of the clean `results` table when
#'   `solution = "all"`. `"wide"` keeps one row per direction with
#'   solution-type-specific columns such as `con_last_safe` and `par_reason`.
#'   `"long"` returns one row per direction and solution type, with a `solution_type`
#'   column. Single-solution-type calls keep the compact single-solution-type
#'   layout.
#' @param progress Logical; if `TRUE` and the session is interactive, show a
#'   text progress bar.
#' @param x An `ncut_test` object returned by [ncut.test()].
#' @param row.names Logical; passed to [print.data.frame()] by
#'   [print.ncut_test()].
#' @param ... Additional arguments routed through the QCA workflow. Arguments
#'   matching [QCA::truthTable()] are forwarded to truth table construction;
#'   remaining minimization arguments are filtered and forwarded to
#'   [QCA::minimize()]. The function also looks in `...` for `include`,
#'   `dir.exp`, or `direxp` if those arguments were not supplied explicitly.
#'   In [print.ncut_test()], `...` is passed to [print.data.frame()]. In
#'   [as.data.frame.ncut_test()], `...` is ignored.
#'
#' @returns An object of class `ncut_test` with the following components:
#'   \describe{
#'     \item{`diagnostics`}{A detailed data frame with one row for the lower
#'     search and one row for the upper search. When `solution = "all"`, each
#'     monitored solution type is searched independently, so diagnostics has
#'     one row per direction and solution type. It includes the baseline
#'     `n.cut`, the last safe value, the first failing value, the number of
#'     successful steps, stop reason, change classification, exclusion
#'     information for the baseline, last safe, and first failing runs, and
#'     the computed upper limit.}
#'     \item{`results`}{A compact data frame with the columns
#'     `direction`, `start`, `last_safe`, `first_failing`, `steps`,
#'     `total_delta`, and `reason`. When `solution = "all"` and
#'     `result_shape = "wide"`, the row unit remains `direction`, but the
#'     boundary and reason columns are solution-type-specific, using prefixes `con_`,
#'     `par_`, and, when available, `int_`. When `result_shape = "long"`,
#'     `results` has one row per direction and solution type.}
#'     \item{`bounds`}{A named integer vector with `Lower` and `Upper`
#'     elements taken from the last safe values in each search direction. When
#'     `solution = "all"`, this is a lower/upper matrix with one column per
#'     monitored solution type.}
#'     \item{`baseline`}{A list containing the baseline truth table,
#'     minimization results, selected solution terms used for comparison,
#'     exclusion set used, and
#'     status information.}
#'     \item{`by_direction`}{A named list with one entry for the lower search
#'     and one for the upper search, each containing the search trace and
#'     stopping information.}
#'     \item{`settings`}{A list containing the analysis settings used to build
#'     the result object, including the computed upper limit.}
#'   }
#'
#'   `print.ncut_test()` prints a concise summary and the `results` table.
#'   `as.data.frame.ncut_test()` returns the `results` table.
#'
#' @details
#' The baseline analysis is built with [QCA::truthTable()] using the supplied
#' `incl.cut` and `n.cut`, followed by minimization under the requested
#' solution type. The function then tests lower and upper values of `n.cut`
#' one step at a time. When `solution = "all"`, conservative, parsimonious,
#' and, when requested, intermediate paths are searched independently.
#'
#' The search range is bounded below by `1` and above by the number of rows in
#' `data`.
#'
#' A directional search stops when one of the following occurs:
#' \itemize{
#'   \item the next tested value falls outside the allowed range,
#'   \item truth table construction fails,
#'   \item exclusion recomputation fails,
#'   \item minimization fails, or
#'   \item the monitored solution changes relative to the baseline.
#' }
#'
#' The shared solution-control, exclusion, and returned-object conventions are
#' described in `?qcaERT_conventions`.
#'
#' @examples
#' \donttest{
#' library(QCA)
#' data(LR)
#'
#' conditions <- c("DEV", "URB", "LIT", "IND", "STB")
#' outcome <- "SURV"
#' dir_exp <- rep("1", length(conditions))
#'
#' thresholds <- list(
#'   DEV = findTh(LR$DEV, groups = 7),
#'   URB = findTh(LR$URB, groups = 4),
#'   LIT = findTh(LR$LIT, groups = 4),
#'   IND = findTh(LR$IND, groups = 4),
#'   STB = findTh(LR$STB, groups = 4),
#'   SURV = findTh(LR$SURV, groups = 4)
#' )
#'
#' dat <- LR
#' dat$DEV <- calibrate(LR$DEV, type = "fuzzy", thresholds = thresholds$DEV)
#' dat$URB <- calibrate(LR$URB, type = "fuzzy", thresholds = thresholds$URB)
#' dat$LIT <- calibrate(LR$LIT, type = "fuzzy", thresholds = thresholds$LIT)
#' dat$IND <- calibrate(LR$IND, type = "fuzzy", thresholds = thresholds$IND)
#' dat$STB <- calibrate(LR$STB, type = "fuzzy", thresholds = thresholds$STB)
#' dat$SURV <- calibrate(LR$SURV, type = "fuzzy", thresholds = thresholds$SURV)
#'
#' out <- ncut.test(
#'   data = dat,
#'   outcome = outcome,
#'   conditions = conditions,
#'   n.cut = 1,
#'   step = 1,
#'   max_steps = 4,
#'   incl.cut = 0.8,
#'   solution = "all",
#'   dir.exp = dir_exp,
#'   result_shape = "long",
#'   progress = TRUE
#' )
#'
#' out
#' as.data.frame(out)
#' }
#'
#' @seealso [calib.test()], [incl.test()], [loo.test()], [subsample.test()],
#'   [altset.test()], [theory.test()], [cluster.test()], [sol.df()]
#' @export
ncut.test <- function(
    data,
    outcome,
    conditions = NULL,
    n.cut = 1,
    step = 1,
    max_steps = 20,
    incl.cut = 1,
    solution = "all",
    include = NULL,
    dir.exp = NULL,
    which_M = 1,
    i_mode = c("all", "C1P1"),
    exclude_mode = c("recompute", "static", "none"),
    exclude_recompute = list(type = 2),
    exclude_static = NULL,
    result_shape = c("wide", "long"),
    progress = TRUE,
    ...
) {
  .require_qca()

  selection_controls <- .normalize_i_mode(i_mode)
  i_mode <- selection_controls$i_mode
  exclude_mode <- match.arg(exclude_mode)
  result_shape <- match.arg(result_shape)

  if (is.null(dim(data)) || is.null(nrow(data)) || nrow(data) < 1L) {
    stop("`data` must be a non-empty data frame object with at least one row.")
  }

  if (!is.character(outcome) || length(outcome) != 1L || !nzchar(outcome)) {
    stop("`outcome` must be a single non-empty character string.")
  }

  if (!is.null(conditions)) {
    if (!is.character(conditions) || length(conditions) < 1L) {
      stop("`conditions` must be NULL or a non-empty character vector.")
    }
  }
  .validate_outcome_conditions_distinct(outcome, conditions)

  if (!is.numeric(incl.cut) || length(incl.cut) != 1L || !is.finite(incl.cut) ||
      incl.cut < 0 || incl.cut > 1) {
    stop("`incl.cut` must be a single finite number in [0, 1].")
  }

  n.cut <- .as_integerish_scalar(n.cut, "n.cut", min = 1L)
  step <- .as_integerish_scalar(step, "step", min = 1L)
  max_steps <- .as_integerish_scalar(max_steps, "max_steps", min = 1L)
  which_M <- .as_integerish_scalar(which_M, "which_M", min = 1L)

  mc <- match.call(expand.dots = FALSE)
  dots_raw <- list(...)
  .reject_exclusion_controls_in_dots(dots_raw, "ncut.test")

  if (is.null(dir.exp)) {
    de <- .dot_get(dots_raw, "dir.exp")
    if (is.null(de)) de <- .dot_get(dots_raw, "direxp")
    if (!is.null(de)) dir.exp <- de
  }

  if (is.null(include)) {
    inc <- .dot_get(dots_raw, "include")
    if (!is.null(inc)) include <- inc
  }

  solution_controls <- .resolve_solution_controls(
    solution = solution,
    include = include,
    dir.exp = dir.exp,
    caller = "ncut.test",
    style = "std"
  )

  solution <- solution_controls$solution
  include <- solution_controls$include
  monitored_solutions <- solution_controls$monitored

  .validate_exclusion_controls(
    exclude_mode = exclude_mode,
    exclude_recompute = exclude_recompute,
    exclude_static = exclude_static,
    exclude_recompute_supplied = !is.null(mc$exclude_recompute),
    exclude_static_supplied = !is.null(mc$exclude_static),
    monitored_solutions = monitored_solutions,
    style = "std"
  )

  dots_split <- .split_truth_table_minimize_dots(dots_raw)
  dots_tt <- dots_split$tt
  dots_min <- dots_split$min

  .build_truth_table <- function(cutoff) {
    args <- c(
      list(
        data = data,
        outcome = outcome,
        incl.cut = incl.cut,
        n.cut = cutoff
      ),
      dots_tt
    )

    if (!is.null(conditions)) {
      args$conditions <- conditions
    }

    do.call(QCA::truthTable, args)
  }

  .run_once <- function(cutoff, solution_types = monitored_solutions) {
    .boundary_run_once(
      value = cutoff,
      build_truth_table = .build_truth_table,
      monitored_solutions = solution_types,
      exclude_mode = exclude_mode,
      exclude_recompute = exclude_recompute,
      exclude_static = exclude_static,
      dots_filtered = dots_min,
      dir.exp = dir.exp,
      which_M = which_M,
      i_mode = i_mode
    )
  }

  upper_limit <- as.integer(nrow(data))

  progress_total <- if (solution == "all") 2L * length(monitored_solutions) else 2L
  progress_state <- .new_qcaert_progress(total = progress_total, progress = progress)
  on.exit(progress_state$close(), add = TRUE)
  .bump_pb <- progress_state$tick

  searched <- .boundary_search(
    value_name = "n.cut",
    value_type = "integer",
    start_value = n.cut,
    step = step,
    max_steps = max_steps,
    lower_limit = 1L,
    upper_limit = upper_limit,
    run_once = .run_once,
    solution = solution,
    monitored_solutions = monitored_solutions,
    i_mode = i_mode,
    which_M = which_M,
    progress_tick = .bump_pb,
    extra_row = list(upper_limit = upper_limit),
    extra_path = list(upper_limit = upper_limit)
  )
  diagnostics <- searched$diagnostics
  baseline <- searched$baseline
  by_direction <- searched$by_direction

  bounds <- .boundary_bounds(diagnostics, "n.cut")

  .new_result_object(
    "ncut_test",
    diagnostics = diagnostics,
    results = .make_boundary_results(diagnostics, "n.cut", result_shape = result_shape),
    bounds = bounds,
    baseline = baseline,
    by_direction = by_direction,
    settings = list(
      outcome = outcome,
      conditions = conditions,
      n.cut = n.cut,
      step = step,
      max_steps = max_steps,
      incl.cut = incl.cut,
      solution = solution,
      monitored_solutions = monitored_solutions,
      include = include,
      dir.exp = dir.exp,
      which_M = which_M,
      i_mode = i_mode,
      exclude_mode = exclude_mode,
      exclude_recompute = exclude_recompute,
      exclude_static = exclude_static,
      result_shape = result_shape,
      upper_limit = upper_limit
    )
  )
}

#' Print an `ncut_test` object
#'
#' @rdname ncut.test
#' @export
print.ncut_test <- function(x, row.names = FALSE, ...) {
  results <- x$results
  settings <- x$settings

  .print_qcaert_heading("ncut_test", "n.cut robustness", settings)
  if (is.data.frame(results) && nrow(results) > 0L) {
    cat("Starting n.cut: ", format(results$start[1L], trim = TRUE), "\n", sep = "")
  }
  if (!is.null(settings$step)) {
    cat("Step size: ", format(settings$step, trim = TRUE), "\n", sep = "")
  }
  if (!is.null(settings$max_steps)) {
    cat("Max steps: ", settings$max_steps, "\n", sep = "")
  }
  if (!is.null(settings$upper_limit)) {
    cat("Upper limit: ", settings$upper_limit, "\n", sep = "")
  }

  .print_qcaert_table(results, "Result", row.names = row.names, ...)
  .print_boundary_summary(x)

  invisible(x)
}

#' Return the main results table from an `ncut_test` object
#'
#' @rdname ncut.test
#' @export
as.data.frame.ncut_test <- function(x, ...) {
  .as.data.frame_results(x, ...)
}

test_that("boundary tests report staged baseline QCA failures", {
  skip_if_not_installed("QCA")

  fixture <- qcaert_fixture_direct6()

  incl_truth_table <- suppressWarnings(incl.test(
    data = fixture$calib,
    outcome = fixture$outcome,
    conditions = fixture$conditions,
    incl.cut = 0.75,
    step = 0.1,
    max_steps = 1,
    n.cut = 1,
    solution = "conservative",
    sort.by = "bad",
    progress = FALSE
  ))
  expect_true(all(incl_truth_table$diagnostics$stop_reason == "baseline_truth_table_build_error"))
  expect_true(all(incl_truth_table$diagnostics$error_source == "truthTable"))

  incl_minimize <- suppressWarnings(incl.test(
    data = fixture$calib,
    outcome = fixture$outcome,
    conditions = fixture$conditions,
    incl.cut = 0.75,
    step = 0.1,
    max_steps = 1,
    n.cut = 1,
    solution = "conservative",
    show.cases = "bad",
    progress = FALSE
  ))
  expect_true(all(incl_minimize$diagnostics$stop_reason == "baseline_requested_minimize_error"))
  expect_true(all(incl_minimize$diagnostics$error_source == "conservative"))

  ncut_truth_table <- suppressWarnings(ncut.test(
    data = fixture$calib,
    outcome = fixture$outcome,
    conditions = fixture$conditions,
    n.cut = 1,
    step = 1,
    max_steps = 1,
    incl.cut = 0.75,
    solution = "conservative",
    sort.by = "bad",
    progress = FALSE
  ))
  expect_true(all(ncut_truth_table$diagnostics$stop_reason == "baseline_truth_table_build_error"))
  expect_true(all(ncut_truth_table$diagnostics$error_source == "truthTable"))
})

test_that("calib.test reports staged baseline QCA failures and missing selected solutions", {
  skip_if_not_installed("QCA")

  fixture <- qcaert_fixture_direct6()

  missing_solution <- suppressWarnings(calib.test(
    raw.data = fixture$raw,
    calib.data = fixture$calib,
    outcome = fixture$outcome,
    conditions = fixture$conditions,
    calib_spec = fixture$calib_spec,
    test.conditions = fixture$test.conditions[1L],
    anchors_to_test = "E1",
    unit_step = 1,
    max_steps = 1,
    incl.cut = 0.75,
    n.cut = 1,
    solution = "conservative",
    which_M = 99,
    progress = FALSE
  ))
  expect_identical(missing_solution$baseline$status, "baseline_error")
  expect_true(all(missing_solution$diagnostics$stop_reason == "baseline_selected_solution_missing"))
  expect_true(all(missing_solution$diagnostics$error_source == "conservative"))
  expect_true(all(missing_solution$results$reason == "baseline_selected_solution_missing"))

  truth_table_failure <- suppressWarnings(calib.test(
    raw.data = fixture$raw,
    calib.data = fixture$calib,
    outcome = fixture$outcome,
    conditions = fixture$conditions,
    calib_spec = fixture$calib_spec,
    test.conditions = fixture$test.conditions[1L],
    anchors_to_test = "E1",
    unit_step = 1,
    max_steps = 1,
    incl.cut = 0.75,
    n.cut = 1,
    solution = "conservative",
    sort.by = "bad",
    progress = FALSE
  ))
  expect_true(all(truth_table_failure$diagnostics$stop_reason == "baseline_truth_table_build_error"))
  expect_true(all(truth_table_failure$diagnostics$error_source == "truthTable"))

  minimize_failure <- suppressWarnings(calib.test(
    raw.data = fixture$raw,
    calib.data = fixture$calib,
    outcome = fixture$outcome,
    conditions = fixture$conditions,
    calib_spec = fixture$calib_spec,
    test.conditions = fixture$test.conditions[1L],
    anchors_to_test = "E1",
    unit_step = 1,
    max_steps = 1,
    incl.cut = 0.75,
    n.cut = 1,
    solution = "conservative",
    show.cases = "bad",
    progress = FALSE
  ))
  expect_true(all(minimize_failure$diagnostics$stop_reason == "baseline_requested_minimize_error"))
  expect_true(all(minimize_failure$diagnostics$error_source == "conservative"))

  exclude_failure <- suppressWarnings(calib.test(
    raw.data = fixture$raw,
    calib.data = fixture$calib,
    outcome = fixture$outcome,
    conditions = fixture$conditions,
    calib_spec = fixture$calib_spec,
    test.conditions = fixture$test.conditions[1L],
    anchors_to_test = "E1",
    unit_step = 1,
    max_steps = 1,
    incl.cut = 0.75,
    n.cut = 1,
    solution = "parsimonious",
    exclude_mode = "recompute",
    exclude_recompute = list(type = 2, bogus = TRUE),
    progress = FALSE
  ))
  expect_true(all(exclude_failure$diagnostics$stop_reason == "baseline_exclude_recompute_error"))
  expect_true(all(exclude_failure$diagnostics$error_source == "exclude"))
})

test_that("plot methods explain zero-row filters", {
  skip_if_not_installed("QCA")
  skip_if_not_installed("ggplot2")

  fixture <- qcaert_fixture_direct6()

  incl_out <- suppressWarnings(incl.test(
    data = fixture$calib,
    outcome = fixture$outcome,
    conditions = fixture$conditions,
    incl.cut = 0.75,
    step = 0.1,
    max_steps = 1,
    n.cut = 1,
    solution = "conservative",
    progress = FALSE
  ))
  expect_error(
    plot(incl_out, stop_reason = "not_a_recorded_reason"),
    "No rows left to plot after filtering.",
    fixed = TRUE
  )

  calib_out <- suppressWarnings(calib.test(
    raw.data = fixture$raw,
    calib.data = fixture$calib,
    outcome = fixture$outcome,
    conditions = fixture$conditions,
    calib_spec = fixture$calib_spec,
    test.conditions = fixture$test.conditions[1L],
    unit_step = 1,
    max_steps = 1,
    incl.cut = 0.75,
    n.cut = 1,
    solution = "conservative",
    progress = FALSE
  ))
  expect_error(
    plot(calib_out, sets = "missing_set"),
    "No rows left to plot after filtering.",
    fixed = TRUE
  )
})

test_that("incl.test solution = all keeps independent solution_type paths", {
  skip_if_not_installed("QCA")

  fixture <- qcaert_fixture_direct6()

  all <- suppressWarnings(incl.test(
    data = fixture$calib,
    outcome = fixture$outcome,
    conditions = fixture$conditions,
    incl.cut = 0.75,
    step = 0.1,
    max_steps = 2,
    n.cut = 1,
    solution = "all",
    dir.exp = fixture$dir.exp,
    progress = FALSE
  ))

  con <- suppressWarnings(incl.test(
    data = fixture$calib,
    outcome = fixture$outcome,
    conditions = fixture$conditions,
    incl.cut = 0.75,
    step = 0.1,
    max_steps = 2,
    n.cut = 1,
    solution = "conservative",
    progress = FALSE
  ))

  expect_false("solution" %in% names(all$results))
  expect_false("reason" %in% names(all$results))
  expect_setequal(all$diagnostics$solution_type, all$settings$monitored_solutions)
  expect_equal(all$results$con_last_safe, con$results$last_safe)
  expect_equal(all$results$con_first_failing, con$results$first_failing)
  expect_equal(all$results$con_steps, con$results$steps)
  expect_true(is.matrix(all$bounds))

  long <- suppressWarnings(incl.test(
    data = fixture$calib,
    outcome = fixture$outcome,
    conditions = fixture$conditions,
    incl.cut = 0.75,
    step = 0.1,
    max_steps = 2,
    n.cut = 1,
    solution = "all",
    dir.exp = fixture$dir.exp,
    result_shape = "long",
    progress = FALSE
  ))

  expect_true("solution_type" %in% names(long$results))
  expect_false("solution" %in% names(long$results))
  expect_equal(nrow(long$results), nrow(con$results) * length(long$settings$monitored_solutions))
  expect_equal(long$results$last_safe[long$results$solution_type == "conservative"], con$results$last_safe)
  expect_equal(long$results$first_failing[long$results$solution_type == "conservative"], con$results$first_failing)
  expect_equal(long$results$steps[long$results$solution_type == "conservative"], con$results$steps)
})

test_that("ncut.test solution = all keeps independent solution_type paths", {
  skip_if_not_installed("QCA")

  fixture <- qcaert_fixture_direct6()

  all <- suppressWarnings(ncut.test(
    data = fixture$calib,
    outcome = fixture$outcome,
    conditions = fixture$conditions,
    n.cut = 1,
    step = 1,
    max_steps = 2,
    incl.cut = 0.75,
    solution = "all",
    dir.exp = fixture$dir.exp,
    progress = FALSE
  ))

  con <- suppressWarnings(ncut.test(
    data = fixture$calib,
    outcome = fixture$outcome,
    conditions = fixture$conditions,
    n.cut = 1,
    step = 1,
    max_steps = 2,
    incl.cut = 0.75,
    solution = "conservative",
    progress = FALSE
  ))

  expect_false("solution" %in% names(all$results))
  expect_false("reason" %in% names(all$results))
  expect_setequal(all$diagnostics$solution_type, all$settings$monitored_solutions)
  expect_equal(all$results$con_last_safe, con$results$last_safe)
  expect_equal(all$results$con_first_failing, con$results$first_failing)
  expect_equal(all$results$con_steps, con$results$steps)
  expect_true(is.matrix(all$bounds))

  long <- suppressWarnings(ncut.test(
    data = fixture$calib,
    outcome = fixture$outcome,
    conditions = fixture$conditions,
    n.cut = 1,
    step = 1,
    max_steps = 2,
    incl.cut = 0.75,
    solution = "all",
    dir.exp = fixture$dir.exp,
    result_shape = "long",
    progress = FALSE
  ))

  expect_true("solution_type" %in% names(long$results))
  expect_false("solution" %in% names(long$results))
  expect_equal(nrow(long$results), nrow(con$results) * length(long$settings$monitored_solutions))
  expect_equal(long$results$last_safe[long$results$solution_type == "conservative"], con$results$last_safe)
  expect_equal(long$results$first_failing[long$results$solution_type == "conservative"], con$results$first_failing)
  expect_equal(long$results$steps[long$results$solution_type == "conservative"], con$results$steps)
})

test_that("calib.test solution = all keeps independent solution_type paths", {
  skip_if_not_installed("QCA")

  fixture <- qcaert_fixture_direct6()

  all <- suppressWarnings(calib.test(
    raw.data = fixture$raw,
    calib.data = fixture$calib,
    outcome = fixture$outcome,
    conditions = fixture$conditions,
    calib_spec = fixture$calib_spec,
    test.conditions = fixture$test.conditions,
    unit_step = 1,
    max_steps = 1,
    incl.cut = 0.75,
    n.cut = 1,
    solution = "all",
    dir.exp = fixture$dir.exp,
    progress = FALSE
  ))

  con <- suppressWarnings(calib.test(
    raw.data = fixture$raw,
    calib.data = fixture$calib,
    outcome = fixture$outcome,
    conditions = fixture$conditions,
    calib_spec = fixture$calib_spec,
    test.conditions = fixture$test.conditions,
    unit_step = 1,
    max_steps = 1,
    incl.cut = 0.75,
    n.cut = 1,
    solution = "conservative",
    progress = FALSE
  ))

  expect_false("solution" %in% names(all$results))
  expect_false("reason" %in% names(all$results))
  expect_setequal(all$diagnostics$solution_type, all$settings$monitored_solutions)
  expect_equal(nrow(all$diagnostics), nrow(con$diagnostics) * length(all$settings$monitored_solutions))
  expect_equal(all$results$con_last_safe, con$results$last_safe)
  expect_equal(all$results$con_first_failing, con$results$first_failing)
  expect_equal(all$results$con_steps, con$results$steps)
  expect_true(is.list(all$bounds[[fixture$test.conditions[1L]]]))

  long <- suppressWarnings(calib.test(
    raw.data = fixture$raw,
    calib.data = fixture$calib,
    outcome = fixture$outcome,
    conditions = fixture$conditions,
    calib_spec = fixture$calib_spec,
    test.conditions = fixture$test.conditions,
    unit_step = 1,
    max_steps = 1,
    incl.cut = 0.75,
    n.cut = 1,
    solution = "all",
    dir.exp = fixture$dir.exp,
    result_shape = "long",
    progress = FALSE
  ))

  expect_true("solution_type" %in% names(long$results))
  expect_false("solution" %in% names(long$results))
  expect_equal(nrow(long$results), nrow(con$results) * length(long$settings$monitored_solutions))
  expect_equal(long$results$last_safe[long$results$solution_type == "conservative"], con$results$last_safe)
  expect_equal(long$results$first_failing[long$results$solution_type == "conservative"], con$results$first_failing)
  expect_equal(long$results$steps[long$results$solution_type == "conservative"], con$results$steps)
})

test_that("calib.test exposes intermediate branch selection through i_mode", {
  skip_if_not_installed("QCA")

  fixture <- qcaert_fixture_direct6()

  out <- suppressWarnings(calib.test(
    raw.data = fixture$raw,
    calib.data = fixture$calib,
    outcome = fixture$outcome,
    conditions = fixture$conditions,
    calib_spec = fixture$calib_spec,
    test.conditions = fixture$test.conditions[1L],
    anchors_to_test = "E1",
    unit_step = 1,
    max_steps = 1,
    incl.cut = 0.75,
    n.cut = 1,
    solution = "intermediate",
    dir.exp = fixture$dir.exp,
    i_mode = "C1P1",
    progress = FALSE
  ))

  expect_identical(out$settings$i_mode, "C1P1")
  expect_identical(out$settings$monitored_solutions, "intermediate")
  expect_true(all(grepl("^C1P1:", out$baseline$sig$intermediate)))
})

#' qcaERT argument and output conventions
#'
#' The qcaERT robustness functions are siblings. They share a common public
#' structure for solution controls, exclusion handling, returned objects,
#' printing, and data-frame coercion. Individual function pages describe the
#' function-specific workflow; this page describes the package-wide conventions
#' that should be read consistently across the function family.
#'
#' @section Solution Controls:
#' Most functions accept `solution`, `include`, `dir.exp`, `which_M`, and
#' `i_mode`.
#'
#' `solution` may be `"all"`, `"con"`/`"conservative"`,
#' `"par"`/`"parsimonious"`, or `"int"`/`"intermediate"`. For single-solution-type
#' calls, `include` is optional and is resolved from `solution`: conservative
#' uses `include = ""`, while parsimonious and intermediate use
#' `include = "?"`. Intermediate solutions require `dir.exp`.
#'
#' When `solution = "all"`, do not supply `include`. The monitored set contains
#' conservative and parsimonious solutions by default. If `dir.exp` is supplied,
#' the intermediate solution is also monitored. In stepwise boundary searches
#' such as [incl.test()], [ncut.test()], and [calib.test()], these solution
#' types are searched independently; one solution type changing does not stop
#' the search for another solution type. When a baseline or step fails in one
#' monitored solution type, inspect the solution-type-specific rows in
#' `diagnostics`, `results`, or supporting `by_*` components before interpreting
#' an all-solution run as a single combined failure.
#'
#' `which_M` selects the model or solution alternative to use when QCA
#' minimization returns multiple alternatives. `i_mode` controls which
#' intermediate branches are included; accepted values are `"all"` and
#' `"C1P1"`. Most robustness functions default to allowing all intermediate
#' branches. `theory.test()` defaults to `"C1P1"` because
#' comparative theory testing needs one comparable intermediate branch per
#' theory by default.
#'
#' @section Exclusions:
#' Functions that can monitor parsimonious or intermediate solutions use a
#' common exclusion convention. `exclude_mode = "recompute"` recalculates excluded
#' rows for each perturbed, reduced, or sampled truth table using
#' [QCA::findRows()] and `exclude_recompute`. `exclude_mode = "static"`
#' reuses the supplied `exclude_static` object. `exclude_mode = "none"` avoids
#' exclusion handling.
#'
#' The family default for recomputation is
#' `exclude_recompute = list(type = 2)`. [incl.test()], [ncut.test()],
#' [calib.test()], [loo.test()], [subsample.test()], [altset.test()], and
#' [theory.test()] share this `exclude_mode`, `exclude_recompute`, and
#' `exclude_static` convention. `cluster.test()` starts from an existing truth
#' table and therefore accepts `exclude` directly instead of exposing
#' `exclude_mode`.
#'
#' @section Calibration Specifications:
#' Calibration-family functions use `calib_spec` for calibration information.
#' A `calib_spec` entry is named by calibrated set, contains the raw source name
#' in `raw`, the calibration `type`, and `thresholds`, and may contain `method`
#' and a `calibrate` list forwarded to [QCA::calibrate()].
#'
#' For condition-only calibration tests, `calib_spec` is named by
#' `conditions`. When `calib.test()` or `altset.test()` is called with
#' `test.outcome = TRUE`, `calib_spec` is named by `c(conditions, outcome)`.
#' The outcome must not be included in
#' `conditions`; outcome calibration testing is requested explicitly with
#' `test.outcome = TRUE`.
#'
#' Crisp conditions use one threshold. Fuzzy direct calibration supports three
#' thresholds in `c(E, C, I)` order or six thresholds in
#' `c(E1, C1, I1, I2, C2, E2)` order. Fuzzy indirect calibration treats its
#' thresholds as ordered cutpoints.
#'
#' These calibration-perturbation routines are designed for QCA workflows using
#' crisp and fuzzy sets. They are not multi-value calibration tools.
#' In [loo.test()] and [subsample.test()] with `calib = "recompute"`, a
#' `calib_spec` entry may additionally contain `findTh`, a named list of
#' [QCA::findTh()] arguments, to re-estimate thresholds on each reduced or
#' subsampled raw dataset. Without `findTh`, the baseline `thresholds` stored in
#' `calib_spec` are reused.
#'
#' @section Result Objects:
#' Most robustness functions return an S3 object with `diagnostics`, `results`,
#' and `settings`, plus supporting components such as `baseline`, `bounds`,
#' `by_direction`, `by_case`, `by_run`, `by_draw`, or `summary`.
#'
#' `diagnostics` is the detailed internal table. `results` is the clean
#'  table. The `print()` method shows a concise summary plus the
#'  results, and `as.data.frame()` returns the clean `results` table.
#'
#' For [incl.test()], [ncut.test()], and [calib.test()], `result_shape`
#' controls the layout of the clean `results` table when `solution = "all"`.
#' The default `"wide"` layout keeps one row per tested path and uses
#' solution-type-prefixed columns such as `con_last_safe` and `par_reason`. The
#' `"long"` layout uses one row per tested path and solution type, with a
#' `solution_type` column. The raw `diagnostics` table is unchanged.
#'
#' `cluster.test()` and `theory.test()` use structured `results` lists because
#' their clean output has more than one natural table. `cluster.test()` returns
#' `overview`, `clusters`, and `units`; `as.data.frame()` returns
#' `results$overview`. `theory.test()` returns `models`, `pairwise`, and
#' `solutions`; `as.data.frame()` returns `results$models`.
#'
#' @section Plotting:
#' Plotting requires `ggplot2`. `plot()` methods are provided
#' for `incl_test`, `calib_test`, and `theory_test` objects. Interval and trace
#' views map directly to stored boundary-search diagnostics; theory plots map
#' selected theory-specific models into consistency/coverage space.
#' [sol.chart()] renders [sol.df()] tables as solution charts. See
#' `?qcaERT_plots` and `?sol.chart`.
#'
#' @section QCA Solution Objects:
#' `cluster.test()` and [sol.df()] consume QCA minimization objects. [sol.df()]
#' extracts a compact data frame, and [sol.chart()] gives a visual display of
#' that extracted table. Case and solution-expression reconstruction use the
#' data-frame `pims` component produced by [QCA::minimize()].
#'
#' @name qcaERT_conventions
NULL

#' Draw a chart from a sol.df table
#'
#' Turns the solution table returned by [sol.df()] into a visual chart
#' of sufficient configurations. The chart uses one column per aligned prime
#' implicant and shows condition presence, condition absence, prime implicant
#' fit, solution fit, and optionally cases.
#'
#' @param x A data frame returned by [sol.df()].
#' @param conditions Optional character vector giving the condition order. If
#'   `NULL`, conditions are inferred from `Prime_Implicants` in order of first
#'   appearance.
#' @param solution_types Solution types to display, in plotting order. Accepted
#'   values follow the common qcaERT solution-type conventions, excluding `"all"`.
#'   `"complex"` is accepted as an alias for `"conservative"` for display use.
#' @param model Optional positive integer selecting which model to display when
#'   `x` contains more than one model.
#' @param intermediate_branch Optional intermediate branch to display when `x`
#'   contains more than one intermediate branch.
#' @param colors Named character vector with colors for `"conservative"`,
#'   `"intermediate"`, and `"parsimonious"`.
#' @param show_cases Logical; if `TRUE`, include a cases row. Cases are taken
#'   from the most detailed available configuration in each column.
#' @param show_pi_fit Logical; if `TRUE`, include prime-implicant consistency,
#'   raw coverage, unique coverage, and PRI rows.
#' @param show_solution_fit Logical; if `TRUE`, include solution-level
#'   consistency, coverage, and PRI rows.
#' @param digits Non-negative integer used to format fit statistics.
#' @param title Optional plot title. If `NULL`, a default title is used.
#' @param note Logical; if `TRUE`, include a caption explaining symbols and
#'   colors.
#' @param legend Logical; if `TRUE`, include ggplot legends for solution type
#'   and condition state.
#' @param point_size Size of presence/absence symbols.
#' @param text_size Size of table text.
#' @param theme A ggplot2 theme object added to the chart.
#'
#' @returns A ggplot object.
#'
#' @details
#' `sol.chart()` is a visual display for [sol.df()] tables. It does not extract
#' solutions from QCA minimization objects. Use [sol.df()] first, then pass the
#' resulting table to `sol.chart()`.
#'
#' Prime implicant columns are aligned according to the following hierarchy:
#' conservative/complex, intermediate, and parsimonious. When several
#' detailed configurations simplify into the same intermediate or parsimonious
#' prime implicant, the simpler prime implicant is repeated across the relevant
#' columns so that the detailed empirical configurations remain visible.
#'
#' Filled circles denote condition presence. Open circles denote condition
#' absence. Empty cells denote conditions that are irrelevant to that prime
#' implicant at the displayed solution type.
#'
#' @examples
#' library(QCA)
#' data(LC)
#'
#' conditions <- c("DEV", "URB", "LIT", "IND", "STB")
#' dir_exp <- rep("1", length(conditions))
#'
#' tt <- truthTable(
#'   data = LC,
#'   outcome = "SURV",
#'   conditions = conditions,
#'   incl.cut = 0.8,
#'   n.cut = 1
#' )
#'
#' enhanced <- findRows(tt, type = 2)
#'
#' con <- minimize(
#'   input = tt,
#'   include = "",
#'   details = TRUE,
#'   show.cases = FALSE
#' )
#'
#' par <- minimize(
#'   input = tt,
#'   include = "?",
#'   exclude = enhanced,
#'   details = TRUE,
#'   show.cases = FALSE
#' )
#'
#' int <- minimize(
#'   input = tt,
#'   include = "?",
#'   dir.exp = dir_exp,
#'   exclude = enhanced,
#'   details = TRUE,
#'   show.cases = FALSE
#' )
#'
#' solution_table <- sol.df(
#'   conservative = con,
#'   parsimonious = par,
#'   intermediate = int,
#'   solution = "all",
#'   which_M = 1,
#'   i_mode = "C1P1",
#'   include_cases = FALSE,
#'   digits = 2
#' )
#'
#' if (requireNamespace("ggplot2", quietly = TRUE)) {
#'   library(ggplot2)
#'
#'   sol.chart(
#'     solution_table,
#'     conditions = conditions,
#'     solution_types = c("conservative", "intermediate", "parsimonious"),
#'     show_cases = FALSE
#'   )
#' }
#'
#' \donttest{
#' library(QCA)
#' library(ggplot2)
#' data(LR)
#'
#' conditions <- c("DEV", "URB", "LIT", "IND", "STB")
#' outcome <- "SURV"
#' dir_exp <- rep("1", length(conditions))
#'
#' thresholds <- list(
#'   DEV = findTh(LR$DEV, groups = 7),
#'   URB = findTh(LR$URB, groups = 4),
#'   LIT = findTh(LR$LIT, groups = 4),
#'   IND = findTh(LR$IND, groups = 4),
#'   STB = findTh(LR$STB, groups = 4),
#'   SURV = findTh(LR$SURV, groups = 4)
#' )
#'
#' dat <- LR
#' dat$DEV <- calibrate(LR$DEV, type = "fuzzy", thresholds = thresholds$DEV)
#' dat$URB <- calibrate(LR$URB, type = "fuzzy", thresholds = thresholds$URB)
#' dat$LIT <- calibrate(LR$LIT, type = "fuzzy", thresholds = thresholds$LIT)
#' dat$IND <- calibrate(LR$IND, type = "fuzzy", thresholds = thresholds$IND)
#' dat$STB <- calibrate(LR$STB, type = "fuzzy", thresholds = thresholds$STB)
#' dat$SURV <- calibrate(LR$SURV, type = "fuzzy", thresholds = thresholds$SURV)
#'
#' tt <- truthTable(
#'   data = dat,
#'   outcome = outcome,
#'   conditions = conditions,
#'   incl.cut = 0.8,
#'   n.cut = 1,
#'   complete = TRUE,
#'   show.cases = TRUE
#' )
#' enhanced <- findRows(tt, type = 2)
#'
#' con <- minimize(tt, include = "", details = TRUE, show.cases = FALSE)
#' par <- minimize(tt, include = "?", exclude = enhanced, details = TRUE, show.cases = FALSE)
#' int <- minimize(
#'   tt,
#'   include = "?",
#'   dir.exp = dir_exp,
#'   exclude = enhanced,
#'   details = TRUE,
#'   show.cases = FALSE
#' )
#'
#' solution_table <- sol.df(
#'   conservative = con,
#'   intermediate = int,
#'   parsimonious = par,
#'   solution = "all",
#'   which_M = 1,
#'   i_mode = "C1P1",
#'   include_cases = TRUE,
#'   digits = 2
#' )
#'
#' sol.chart(solution_table)
#' }
#'
#' @seealso [sol.df()], [qcaERT_plots]
#' @export
sol.chart <- function(
    x,
    conditions = NULL,
    solution_types = c("conservative", "intermediate", "parsimonious"),
    model = NULL,
    intermediate_branch = NULL,
    colors = c(
      conservative = "#222222",
      intermediate = "#E69F00",
      parsimonious = "#0072B2"
    ),
    show_cases = TRUE,
    show_pi_fit = TRUE,
    show_solution_fit = TRUE,
    digits = 2,
    title = NULL,
    note = TRUE,
    legend = FALSE,
    point_size = 3.2,
    text_size = 3.3,
    theme = .plot_theme()
) {
  .plot_require_ggplot2()

  chart <- .sol_chart_prepare(
    x = x,
    conditions = conditions,
    solution_types = solution_types,
    model = model,
    intermediate_branch = intermediate_branch,
    colors = colors,
    show_cases = show_cases,
    show_pi_fit = show_pi_fit,
    show_solution_fit = show_solution_fit,
    digits = digits
  )

  if (!nrow(chart$grid)) {
    stop("No chart cells are available to draw.", call. = FALSE)
  }

  caption <- NULL
  if (isTRUE(note)) {
    caption <- paste(
      "Filled circle = presence of condition; open circle = absence of condition; empty cell = irrelevant condition.",
      "Black = conservative/complex; orange = intermediate; blue = parsimonious."
    )
  }

  p <- ggplot2::ggplot(chart$grid, ggplot2::aes(x = .data[["x"]], y = .data[["y"]])) +
    ggplot2::geom_tile(
      ggplot2::aes(width = .data[["width"]], height = .data[["height"]]),
      fill = "white",
      colour = "#222222",
      linewidth = 0.25
    ) +
    ggplot2::geom_text(
      data = chart$labels,
      ggplot2::aes(
        x = .data[["x"]],
        y = .data[["y"]],
        label = .data[["label"]]
      ),
      size = text_size,
      inherit.aes = FALSE
    )

  if (nrow(chart$states)) {
    p <- p +
      ggplot2::geom_point(
        data = chart$states,
        ggplot2::aes(
          x = .data[["x"]],
          y = .data[["y"]],
          colour = .data[["solution_type"]],
          shape = .data[["state"]]
        ),
        size = point_size,
        stroke = 0.8,
        inherit.aes = FALSE
      )
  }

  if (nrow(chart$text)) {
    p <- p +
      ggplot2::geom_text(
        data = chart$text,
        ggplot2::aes(
          x = .data[["x"]],
          y = .data[["y"]],
          label = .data[["label"]],
          colour = .data[["solution_type"]]
        ),
        size = text_size,
        inherit.aes = FALSE
      )
  }

  if (nrow(chart$plain_text)) {
    p <- p +
      ggplot2::geom_text(
        data = chart$plain_text,
        ggplot2::aes(
          x = .data[["x"]],
          y = .data[["y"]],
          label = .data[["label"]]
        ),
        size = text_size,
        inherit.aes = FALSE
      )
  }

  title <- if (is.null(title)) "Sufficient configurations" else as.character(title)[1L]

  p +
    ggplot2::scale_colour_manual(
      name = "Solution type",
      values = chart$colors,
      breaks = names(chart$colors),
      labels = .sol_chart_solution_labels(names(chart$colors)),
      drop = FALSE
    ) +
    ggplot2::scale_shape_manual(
      name = "Condition",
      values = c(present = 16, absent = 1),
      labels = c(present = "Presence", absent = "Absence"),
      drop = FALSE
    ) +
    ggplot2::scale_x_continuous(
      limits = c(-0.5, chart$n_prime_implicants + 0.5),
      breaks = NULL,
      expand = c(0, 0)
    ) +
    ggplot2::scale_y_continuous(
      limits = c(0.5, chart$n_rows + 0.5),
      breaks = NULL,
      expand = c(0, 0)
    ) +
    ggplot2::labs(
      title = title,
      x = NULL,
      y = NULL,
      caption = caption
    ) +
    ggplot2::coord_cartesian(clip = "off") +
    theme +
    ggplot2::theme(
      axis.text = ggplot2::element_blank(),
      axis.ticks = ggplot2::element_blank(),
      axis.title = ggplot2::element_blank(),
      panel.grid = ggplot2::element_blank(),
      panel.border = ggplot2::element_blank(),
      legend.position = if (isTRUE(legend)) "bottom" else "none",
      plot.caption = ggplot2::element_text(hjust = 0, size = max(text_size * 2.6, 8)),
      plot.margin = ggplot2::margin(5.5, 12, 5.5, 5.5)
    )
}

.sol_chart_prepare <- function(
    x,
    conditions = NULL,
    solution_types = c("conservative", "intermediate", "parsimonious"),
    model = NULL,
    intermediate_branch = NULL,
    colors = c(
      conservative = "#222222",
      intermediate = "#E69F00",
      parsimonious = "#0072B2"
    ),
    show_cases = TRUE,
    show_pi_fit = TRUE,
    show_solution_fit = TRUE,
    digits = 2
) {
  .sol_chart_validate_input(x)
  solution_types <- .sol_chart_normalize_solution_types(solution_types)
  colors <- .sol_chart_colors(colors)
  digits <- .sol_chart_digits(digits)

  df <- .sol_chart_terms(x)
  df <- df[df$solution_type %in% solution_types, , drop = FALSE]

  if (!nrow(df)) {
    stop("No rows in `x` match the requested `solution_types`.", call. = FALSE)
  }

  df <- .sol_chart_select_model(df, model)

  selected_branch <- .sol_chart_select_intermediate_branch(df, intermediate_branch)
  if (!is.na(selected_branch)) {
    keep_intermediate <- df$solution_type != "intermediate" |
      (!is.na(df$intermediate_branch) & df$intermediate_branch == selected_branch)
    df <- df[keep_intermediate, , drop = FALSE]
  }

  if (!nrow(df)) {
    stop("No solution rows remain after filtering.", call. = FALSE)
  }

  if (is.null(conditions)) {
    conditions <- .sol_chart_infer_conditions(df$states)
  } else {
    conditions <- trimws(as.character(conditions))
    conditions <- conditions[nzchar(conditions)]
    if (!length(conditions)) {
      stop("`conditions` must contain at least one condition name.", call. = FALSE)
    }
    conditions <- unique(conditions)
  }

  aligned_implicants <- .sol_chart_align_prime_implicants(df)
  if (!length(aligned_implicants)) {
    stop("No prime implicants are available to chart.", call. = FALSE)
  }

  used_types <- .sol_chart_used_solution_types(df, aligned_implicants)
  solution_types <- solution_types[solution_types %in% used_types]
  colors <- colors[solution_types]

  structure <- .sol_chart_rows(
    conditions = conditions,
    show_cases = show_cases,
    show_pi_fit = show_pi_fit,
    show_solution_fit = show_solution_fit
  )

  n_prime_implicants <- length(aligned_implicants)
  n_rows <- nrow(structure)
  grid <- .sol_chart_grid(structure, n_prime_implicants)
  grid$y <- n_rows - grid$row + 1

  labels <- .sol_chart_cell_labels(structure, n_prime_implicants)
  labels$y <- n_rows - labels$row + 1

  offsets <- .sol_chart_offsets(solution_types)

  list(
    grid = grid,
    labels = labels,
    states = .sol_chart_state_points(df, aligned_implicants, conditions, structure, offsets),
    text = .sol_chart_value_text(df, aligned_implicants, structure, offsets, solution_types, digits),
    plain_text = .sol_chart_plain_text(df, aligned_implicants, structure),
    colors = colors,
    n_prime_implicants = n_prime_implicants,
    n_rows = n_rows,
    aligned_prime_implicants = aligned_implicants,
    terms = df,
    row_structure = structure
  )
}

.sol_chart_validate_input <- function(x) {
  if (!is.data.frame(x)) {
    stop("`x` must be a data frame returned by sol.df().", call. = FALSE)
  }

  required <- c(
    "Solution",
    "Model",
    "Intermediate_CnPn",
    "Prime_Implicants",
    "Consistency_PI",
    "PRI_PI",
    "Raw_Coverage_PI",
    "Unique_Coverage_PI",
    "Solution_Consistency",
    "Solution_PRI",
    "Solution_Coverage",
    "Cases"
  )
  missing <- setdiff(required, names(x))
  if (length(missing)) {
    stop(
      "`x` must be a table produced by sol.df(). Missing column(s): ",
      paste(missing, collapse = ", "),
      call. = FALSE
    )
  }

  invisible(TRUE)
}

.sol_chart_normalize_solution_types <- function(solution_types) {
  if (is.null(solution_types)) {
    stop("`solution_types` cannot be NULL.", call. = FALSE)
  }

  out <- vapply(solution_types, .sol_chart_normalize_solution_type, character(1))
  out <- out[nzchar(out)]
  out <- unique(out)
  if (!length(out)) {
    stop("`solution_types` must contain at least one solution type.", call. = FALSE)
  }
  if (any(out == "all")) {
    stop("`solution_types` cannot include 'all'.", call. = FALSE)
  }

  out
}

.sol_chart_normalize_solution_type <- function(x) {
  x <- tolower(trimws(as.character(x)))
  if (identical(x, "complex")) {
    x <- "conservative"
  }
  .normalize_solution_std(x)
}

.sol_chart_colors <- function(colors) {
  defaults <- c(
    conservative = "#222222",
    intermediate = "#E69F00",
    parsimonious = "#0072B2"
  )
  if (is.null(colors)) {
    return(defaults)
  }
  if (!is.character(colors) || is.null(names(colors))) {
    stop("`colors` must be a named character vector.", call. = FALSE)
  }

  colors <- colors[!is.na(names(colors)) & nzchar(names(colors))]
  names(colors) <- vapply(names(colors), .sol_chart_normalize_solution_type, character(1))
  defaults[names(colors)] <- colors
  defaults
}

.sol_chart_digits <- function(digits) {
  if (!is.numeric(digits) || length(digits) != 1L || !is.finite(digits) || digits < 0) {
    stop("`digits` must be a single non-negative integer.", call. = FALSE)
  }
  as.integer(digits)
}

.sol_chart_select_model <- function(df, model) {
  models <- suppressWarnings(as.integer(as.character(df$Model)))
  available <- sort(unique(models[!is.na(models)]))

  if (!length(available)) {
    if (!is.null(model)) {
      stop("`model` was supplied, but no model identifiers are available in `x`.", call. = FALSE)
    }
    return(df)
  }

  if (is.null(model)) {
    if (length(available) == 1L) {
      return(df[models == available[[1L]], , drop = FALSE])
    }
    stop(
      "`model` must be supplied when more than one model is present. Available models: ",
      paste(available, collapse = ", "),
      call. = FALSE
    )
  }

  model <- .as_integerish_scalar(model, "model", min = 1L)
  if (!model %in% available) {
    stop(
      "`model` is not available in `x`. Available models: ",
      paste(available, collapse = ", "),
      call. = FALSE
    )
  }

  df[models == model, , drop = FALSE]
}

.sol_chart_terms <- function(x) {
  out <- x
  out$solution_type <- vapply(out$Solution, .sol_chart_normalize_solution_type, character(1))
  out$intermediate_branch <- .sol_chart_clean_text(out$Intermediate_CnPn)
  out$prime_implicant <- trimws(as.character(out$Prime_Implicants))
  out$term_id <- seq_len(nrow(out))
  out$states <- lapply(out$prime_implicant, .sol_chart_parse_term)
  out$n_literals <- vapply(out$states, length, integer(1))
  out
}

.sol_chart_clean_text <- function(x) {
  x <- trimws(as.character(x))
  x[is.na(x) | x %in% c("", "-", "NA", "<NA>")] <- NA_character_
  x
}

.sol_chart_select_intermediate_branch <- function(df, intermediate_branch) {
  branches <- unique(df$intermediate_branch[df$solution_type == "intermediate"])
  branches <- branches[!is.na(branches) & nzchar(branches)]

  if (!length(branches)) {
    if (!is.null(intermediate_branch)) {
      stop("`intermediate_branch` was supplied, but no intermediate branch is available in `x`.", call. = FALSE)
    }
    return(NA_character_)
  }

  if (is.null(intermediate_branch)) {
    if (length(branches) == 1L) {
      return(branches)
    }
    stop(
      "`intermediate_branch` must be supplied when more than one intermediate branch is present. Available branches: ",
      paste(branches, collapse = ", "),
      call. = FALSE
    )
  }

  intermediate_branch <- trimws(as.character(intermediate_branch))
  intermediate_branch <- intermediate_branch[nzchar(intermediate_branch)]
  if (length(intermediate_branch) != 1L) {
    stop("`intermediate_branch` must be a single branch name.", call. = FALSE)
  }
  if (!intermediate_branch %in% branches) {
    stop(
      "`intermediate_branch` is not available in `x`. Available branches: ",
      paste(branches, collapse = ", "),
      call. = FALSE
    )
  }

  intermediate_branch
}

.sol_chart_parse_term <- function(term) {
  term <- trimws(as.character(term)[1L])
  if (is.na(term) || !nzchar(term) || term %in% c("-", "<no selected solution>")) {
    return(stats::setNames(character(0), character(0)))
  }

  tokens <- unlist(strsplit(term, "*", fixed = TRUE), use.names = FALSE)
  tokens <- trimws(tokens)
  tokens <- tokens[nzchar(tokens)]
  if (!length(tokens)) {
    return(stats::setNames(character(0), character(0)))
  }

  states <- character(length(tokens))
  names <- character(length(tokens))
  for (i in seq_along(tokens)) {
    token <- gsub("\\s+", "", tokens[[i]])
    if (startsWith(token, "~")) {
      states[[i]] <- "absent"
      names[[i]] <- substring(token, 2L)
    } else {
      states[[i]] <- "present"
      names[[i]] <- token
    }
  }

  keep <- nzchar(names)
  stats::setNames(states[keep], names[keep])
}

.sol_chart_infer_conditions <- function(states) {
  out <- character(0)
  for (state in states) {
    nms <- names(state)
    out <- c(out, nms[!nms %in% out])
  }
  if (!length(out)) {
    stop("Could not infer condition names from `Prime_Implicants`; supply `conditions`.", call. = FALSE)
  }
  out
}

.sol_chart_align_prime_implicants <- function(df) {
  order_types <- c("conservative", "intermediate", "parsimonious")
  available <- order_types[order_types %in% unique(df$solution_type)]
  if (!length(available)) {
    return(list())
  }

  by_type <- lapply(available, function(tp) df[df$solution_type == tp, , drop = FALSE])
  names(by_type) <- available

  anchor_type <- available[[1L]]
  aligned <- vector("list", nrow(by_type[[anchor_type]]))
  for (i in seq_len(nrow(by_type[[anchor_type]]))) {
    row <- by_type[[anchor_type]][i, , drop = FALSE]
    column <- stats::setNames(rep(NA_integer_, length(available)), available)
    column[[anchor_type]] <- row$term_id
    base_states <- row$states[[1L]]

    for (tp in .sol_chart_later_types(available, anchor_type)) {
      candidate <- .sol_chart_best_match(by_type[[tp]], base_states)
      if (!is.na(candidate)) {
        column[[tp]] <- candidate
        base_states <- df$states[[match(candidate, df$term_id)]]
      }
    }

    aligned[[i]] <- column
  }

  aligned <- .sol_chart_add_unmatched_prime_implicants(aligned, df, by_type, available)
  names(aligned) <- paste0("PI ", seq_along(aligned))
  aligned
}

.sol_chart_add_unmatched_prime_implicants <- function(aligned, df, by_type, available) {
  used <- unlist(aligned, use.names = FALSE)
  used <- used[!is.na(used)]

  for (tp in available) {
    current <- by_type[[tp]]
    unused <- current$term_id[!current$term_id %in% used]
    if (!length(unused)) {
      next
    }

    later <- .sol_chart_later_types(available, tp)
    for (term_id in unused) {
      column <- stats::setNames(rep(NA_integer_, length(available)), available)
      column[[tp]] <- term_id
      base_states <- df$states[[match(term_id, df$term_id)]]

      for (later_type in later) {
        candidate <- .sol_chart_best_match(by_type[[later_type]], base_states)
        if (!is.na(candidate)) {
          column[[later_type]] <- candidate
          base_states <- df$states[[match(candidate, df$term_id)]]
        }
      }

      aligned[[length(aligned) + 1L]] <- column
      used <- c(used, term_id, column[!is.na(column)])
    }
  }

  aligned
}

.sol_chart_later_types <- function(available, solution_type) {
  idx <- match(solution_type, available)
  if (is.na(idx) || idx >= length(available)) {
    return(character(0))
  }
  available[(idx + 1L):length(available)]
}

.sol_chart_best_match <- function(candidates, detailed_states) {
  if (!nrow(candidates)) {
    return(NA_integer_)
  }

  ok <- vapply(candidates$states, .sol_chart_covers, logical(1), detailed = detailed_states)
  if (!any(ok)) {
    return(NA_integer_)
  }

  candidate_rows <- candidates[ok, , drop = FALSE]
  candidate_rows <- candidate_rows[order(-candidate_rows$n_literals, candidate_rows$term_id), , drop = FALSE]
  candidate_rows$term_id[[1L]]
}

.sol_chart_covers <- function(simple, detailed) {
  simple_names <- names(simple)
  if (!length(simple_names)) {
    return(TRUE)
  }
  if (!all(simple_names %in% names(detailed))) {
    return(FALSE)
  }
  all(unname(simple) == unname(detailed[simple_names]))
}

.sol_chart_used_solution_types <- function(df, aligned_implicants) {
  ids <- unlist(aligned_implicants, use.names = FALSE)
  ids <- ids[!is.na(ids)]
  unique(df$solution_type[df$term_id %in% ids])
}

.sol_chart_rows <- function(conditions, show_cases, show_pi_fit, show_solution_fit) {
  rows <- data.frame(
    row = integer(0),
    key = character(0),
    label = character(0),
    kind = character(0),
    stringsAsFactors = FALSE
  )

  add <- function(key, label, kind) {
    rows <<- rbind(
      rows,
      data.frame(
        row = nrow(rows) + 1L,
        key = key,
        label = label,
        kind = kind,
        stringsAsFactors = FALSE
      )
    )
  }

  add("header", "Conditions", "header")
  for (condition in conditions) {
    add(condition, condition, "condition")
  }
  if (isTRUE(show_cases)) {
    add("cases", "Cases", "cases")
  }
  if (isTRUE(show_pi_fit)) {
    add("Consistency_PI", "PI consistency", "pi_fit")
    add("Raw_Coverage_PI", "PI raw coverage", "pi_fit")
    add("Unique_Coverage_PI", "PI unique coverage", "pi_fit")
    add("PRI_PI", "PI PRI", "pi_fit")
  }
  if (isTRUE(show_solution_fit)) {
    add("Solution_Consistency", "Solution consistency", "solution_fit")
    add("Solution_Coverage", "Solution coverage", "solution_fit")
    add("Solution_PRI", "Solution PRI", "solution_fit")
  }

  rows
}

.sol_chart_cell_labels <- function(structure, n_prime_implicants) {
  labels <- data.frame(
    row = integer(0),
    x = numeric(0),
    label = character(0),
    stringsAsFactors = FALSE
  )

  labels <- rbind(
    labels,
    data.frame(row = 1L, x = 0, label = structure$label[[1L]], stringsAsFactors = FALSE)
  )
  labels <- rbind(
    labels,
    data.frame(row = 1L, x = seq_len(n_prime_implicants), label = paste0("PI ", seq_len(n_prime_implicants)), stringsAsFactors = FALSE)
  )

  row_labels <- structure[structure$kind != "header", , drop = FALSE]
  labels <- rbind(
    labels,
    data.frame(row = row_labels$row, x = 0, label = row_labels$label, stringsAsFactors = FALSE)
  )

  labels
}

.sol_chart_grid <- function(structure, n_prime_implicants) {
  rows <- list()

  for (i in seq_len(nrow(structure))) {
    row <- structure$row[[i]]
    kind <- structure$kind[[i]]

    rows[[length(rows) + 1L]] <- data.frame(
      row = row,
      col = 0,
      x = 0,
      width = 1,
      height = 1,
      stringsAsFactors = FALSE
    )

    if (identical(kind, "solution_fit")) {
      rows[[length(rows) + 1L]] <- data.frame(
        row = row,
        col = NA_integer_,
        x = mean(seq_len(n_prime_implicants)),
        width = n_prime_implicants,
        height = 1,
        stringsAsFactors = FALSE
      )
    } else {
      rows[[length(rows) + 1L]] <- data.frame(
        row = row,
        col = seq_len(n_prime_implicants),
        x = seq_len(n_prime_implicants),
        width = 1,
        height = 1,
        stringsAsFactors = FALSE
      )
    }
  }

  out <- do.call(rbind, rows)
  rownames(out) <- NULL
  out
}

.sol_chart_offsets <- function(solution_types) {
  if (length(solution_types) == 1L) {
    return(stats::setNames(0, solution_types))
  }
  stats::setNames(seq(-0.16, 0.16, length.out = length(solution_types)), solution_types)
}

.sol_chart_state_points <- function(df, aligned_implicants, conditions, structure, offsets) {
  rows <- list()
  condition_rows <- structure[structure$kind == "condition", , drop = FALSE]

  for (pi_i in seq_along(aligned_implicants)) {
    column <- aligned_implicants[[pi_i]]
    for (solution_type in names(column)) {
      term_id <- column[[solution_type]]
      if (is.na(term_id) || !solution_type %in% names(offsets)) {
        next
      }
      term <- df[df$term_id == term_id, , drop = FALSE]
      states <- term$states[[1L]]

      for (condition in conditions) {
        if (!condition %in% names(states)) {
          next
        }
        row_number <- condition_rows$row[condition_rows$key == condition]
        rows[[length(rows) + 1L]] <- data.frame(
          x = pi_i + offsets[[solution_type]],
          y = nrow(structure) - row_number + 1,
          solution_type = solution_type,
          state = states[[condition]],
          stringsAsFactors = FALSE
        )
      }
    }
  }

  .sol_chart_bind_or_empty(rows, c("x", "y", "solution_type", "state"))
}

.sol_chart_value_text <- function(df, aligned_implicants, structure, offsets, solution_types, digits) {
  rows <- list()
  pi_rows <- structure[structure$kind == "pi_fit", , drop = FALSE]

  for (pi_i in seq_along(aligned_implicants)) {
    column <- aligned_implicants[[pi_i]]
    for (solution_type in names(column)) {
      term_id <- column[[solution_type]]
      if (is.na(term_id) || !solution_type %in% names(offsets)) {
        next
      }
      term <- df[df$term_id == term_id, , drop = FALSE]

      for (i in seq_len(nrow(pi_rows))) {
        value <- .sol_chart_format_number(term[[pi_rows$key[[i]]]], digits)
        if (!nzchar(value)) {
          next
        }
        rows[[length(rows) + 1L]] <- data.frame(
          x = pi_i + offsets[[solution_type]],
          y = nrow(structure) - pi_rows$row[[i]] + 1,
          label = value,
          solution_type = solution_type,
          stringsAsFactors = FALSE
        )
      }
    }
  }

  solution_rows <- structure[structure$kind == "solution_fit", , drop = FALSE]
  if (nrow(solution_rows)) {
    center <- mean(seq_along(aligned_implicants))
    for (solution_type in solution_types) {
      one <- df[df$solution_type == solution_type, , drop = FALSE]
      if (!nrow(one) || !solution_type %in% names(offsets)) {
        next
      }
      for (i in seq_len(nrow(solution_rows))) {
        value <- .sol_chart_first_number(one[[solution_rows$key[[i]]]], digits)
        if (!nzchar(value)) {
          next
        }
        rows[[length(rows) + 1L]] <- data.frame(
          x = center + offsets[[solution_type]],
          y = nrow(structure) - solution_rows$row[[i]] + 1,
          label = value,
          solution_type = solution_type,
          stringsAsFactors = FALSE
        )
      }
    }
  }

  .sol_chart_bind_or_empty(rows, c("x", "y", "label", "solution_type"))
}

.sol_chart_plain_text <- function(df, aligned_implicants, structure) {
  cases_row <- structure[structure$kind == "cases", , drop = FALSE]
  if (!nrow(cases_row)) {
    return(.sol_chart_bind_or_empty(list(), c("x", "y", "label")))
  }

  rows <- list()
  for (pi_i in seq_along(aligned_implicants)) {
    term_id <- .sol_chart_anchor_term_id(aligned_implicants[[pi_i]])
    if (is.na(term_id)) {
      next
    }
    term <- df[df$term_id == term_id, , drop = FALSE]
    label <- .sol_chart_clean_text(term$Cases)
    if (length(label) != 1L || is.na(label) || !nzchar(label)) {
      next
    }

    rows[[length(rows) + 1L]] <- data.frame(
      x = pi_i,
      y = nrow(structure) - cases_row$row[[1L]] + 1,
      label = .sol_chart_wrap(label),
      stringsAsFactors = FALSE
    )
  }

  .sol_chart_bind_or_empty(rows, c("x", "y", "label"))
}

.sol_chart_anchor_term_id <- function(aligned_implicant) {
  ids <- aligned_implicant[!is.na(aligned_implicant)]
  if (!length(ids)) {
    return(NA_integer_)
  }
  ids[[1L]]
}

.sol_chart_format_number <- function(x, digits) {
  x <- suppressWarnings(as.numeric(x))[1L]
  if (is.na(x) || !is.finite(x)) {
    return("")
  }
  formatC(x, format = "f", digits = digits)
}

.sol_chart_first_number <- function(x, digits) {
  x <- suppressWarnings(as.numeric(x))
  x <- x[is.finite(x)]
  if (!length(x)) {
    return("")
  }
  .sol_chart_format_number(x[[1L]], digits)
}

.sol_chart_wrap <- function(x, width = 22L) {
  paste(strwrap(as.character(x), width = width), collapse = "\n")
}

.sol_chart_bind_or_empty <- function(rows, cols) {
  if (!length(rows)) {
    out <- as.data.frame(stats::setNames(rep(list(logical(0)), length(cols)), cols))
    for (nm in cols) {
      if (nm %in% c("x", "y")) out[[nm]] <- numeric(0)
      else out[[nm]] <- character(0)
    }
    return(out)
  }

  out <- do.call(rbind, rows)
  rownames(out) <- NULL
  out
}

.sol_chart_solution_labels <- function(solution_types) {
  labels <- c(
    conservative = "Conservative/complex",
    intermediate = "Intermediate",
    parsimonious = "Parsimonious"
  )
  unname(labels[solution_types])
}

sol_chart_fixture <- function(branches = "C1P1") {
  data.frame(
    Solution = c("Conservative", "Conservative", rep("Intermediate", length(branches)), "Parsimonious"),
    Model = 1,
    Intermediate_CnPn = c("-", "-", branches, "-"),
    Prime_Implicants = c("A*B*C", "A*B*~C", rep("A*B", length(branches)), "A"),
    Consistency_PI = c(0.96, 0.95, rep(0.97, length(branches)), 0.98),
    PRI_PI = c(0.91, 0.9, rep(0.92, length(branches)), 0.93),
    Raw_Coverage_PI = c(0.24, 0.23, rep(0.3, length(branches)), 0.4),
    Unique_Coverage_PI = c(0.12, 0.11, rep(0.2, length(branches)), 0.31),
    Solution_Consistency = c(0.97, 0.97, rep(0.98, length(branches)), 0.99),
    Solution_PRI = c(0.92, 0.92, rep(0.93, length(branches)), 0.94),
    Solution_Coverage = c(0.45, 0.45, rep(0.5, length(branches)), 0.53),
    Cases = c("case1, case2", "case3, case4", rep("case1, case2, case3", length(branches)), "case1, case2, case3, case4"),
    stringsAsFactors = FALSE
  )
}

test_that("sol.chart parses prime implicants into condition states", {
  parsed <- .sol_chart_parse_term("A*~B*C")

  expect_identical(
    parsed,
    c(A = "present", B = "absent", C = "present")
  )
})

test_that("sol.chart validates sol.df-style input", {
  expect_error(
    sol.chart(data.frame(Solution = "Conservative")),
    "Missing column\\(s\\):",
    fixed = FALSE
  )
})

test_that("sol.chart aligns detailed prime implicants with simpler prime implicants", {
  chart <- .sol_chart_prepare(sol_chart_fixture(), digits = 2)

  expect_length(chart$aligned_prime_implicants, 2L)
  expect_identical(unname(chart$aligned_prime_implicants[[1]][["conservative"]]), 1L)
  expect_identical(unname(chart$aligned_prime_implicants[[1]][["intermediate"]]), 3L)
  expect_identical(unname(chart$aligned_prime_implicants[[1]][["parsimonious"]]), 4L)
  expect_identical(unname(chart$aligned_prime_implicants[[2]][["conservative"]]), 2L)
  expect_identical(unname(chart$aligned_prime_implicants[[2]][["intermediate"]]), 3L)
  expect_identical(unname(chart$aligned_prime_implicants[[2]][["parsimonious"]]), 4L)
  expect_true(any(chart$labels$label == "PI 1"))

  expect_true(any(chart$states$state == "absent"))
  expect_true(any(chart$text$label == "0.96"))
  expect_true(any(grepl("case1", chart$plain_text$label, fixed = TRUE)))
})

test_that("sol.chart merges solution-fit body cells", {
  chart <- .sol_chart_prepare(sol_chart_fixture(), digits = 2)
  solution_rows <- chart$row_structure$row[chart$row_structure$kind == "solution_fit"]
  solution_grid <- chart$grid[chart$grid$row %in% solution_rows, , drop = FALSE]

  expect_identical(sum(solution_grid$row == solution_rows[[1L]]), 2L)
  expect_true(any(is.na(solution_grid$col)))
  expect_true(any(solution_grid$width == chart$n_prime_implicants))
})

test_that("sol.chart requires an intermediate branch when several are present", {
  x <- sol_chart_fixture(branches = c("C1P1", "C1P2"))

  expect_error(
    .sol_chart_prepare(x),
    "`intermediate_branch` must be supplied",
    fixed = TRUE
  )

  chart <- .sol_chart_prepare(x, intermediate_branch = "C1P2")
  expect_length(chart$aligned_prime_implicants, 2L)
})

test_that("sol.chart requires a model when several are present", {
  x <- rbind(
    sol_chart_fixture(),
    transform(sol_chart_fixture(), Model = 2L)
  )

  expect_error(
    .sol_chart_prepare(x),
    "`model` must be supplied when more than one model is present.",
    fixed = TRUE
  )

  chart <- .sol_chart_prepare(x, model = 2)
  expect_true(all(chart$terms$Model == 2L))
})

test_that("sol.chart returns a ggplot object", {
  skip_if_not_installed("ggplot2")

  p <- sol.chart(sol_chart_fixture(), legend = TRUE)
  expect_s3_class(p, "ggplot")
})

.reduced_error_message <- function(e) {
  if (!inherits(e, "error")) return(NA_character_)
  tryCatch(conditionMessage(e), error = function(...) as.character(e))
}

.reduced_extract_standard_fit <- function(res, solution_type, which_M, fit_measures) {
  if (length(fit_measures) == 0L) {
    return(list(
      fit = NULL,
      fit_missing = FALSE,
      fit_available = character(0),
      fit_missing_measures = character(0),
      meta = list(n_fit_rows = NA_integer_)
    ))
  }

  ic <- res$IC
  tbl <- if (is.list(ic) && "sol.incl.cov" %in% names(ic)) ic$sol.incl.cov else NULL

  if (is.null(tbl) || !is.data.frame(tbl) || nrow(tbl) == 0L) {
    return(list(
      fit = NULL,
      fit_missing = TRUE,
      fit_available = character(0),
      fit_missing_measures = fit_measures,
      meta = list(n_fit_rows = 0L)
    ))
  }

  available <- intersect(fit_measures, colnames(tbl))
  missing_measures <- setdiff(fit_measures, available)

  if (nrow(tbl) < which_M) {
    return(list(
      fit = NULL,
      fit_missing = TRUE,
      fit_available = available,
      fit_missing_measures = fit_measures,
      meta = list(n_fit_rows = nrow(tbl))
    ))
  }

  vals <- suppressWarnings(as.numeric(tbl[which_M, available, drop = TRUE]))
  names(vals) <- paste0(solution_type, ".", available)

  list(
    fit = vals,
    fit_missing = FALSE,
    fit_available = available,
    fit_missing_measures = missing_measures,
    meta = list(n_fit_rows = nrow(tbl))
  )
}

.reduced_extract_intermediate_fit <- function(res, solution_type, which_M, i_mode, fit_measures) {
  if (length(fit_measures) == 0L) {
    return(list(
      fit = NULL,
      fit_missing = FALSE,
      fit_available = character(0),
      fit_missing_measures = character(0),
      meta = list(n_i = NA_integer_, n_fit_rows = NA_integer_)
    ))
  }

  i.sol <- res$i.sol

  if (is.null(i.sol) || length(i.sol) == 0L) {
    return(list(
      fit = NULL,
      fit_missing = TRUE,
      fit_available = character(0),
      fit_missing_measures = fit_measures,
      meta = list(n_i = 0L, n_fit_rows = 0L)
    ))
  }

  nm <- .normalize_i_solution_names(i.sol)

  if (i_mode == "C1P1") {
    idx <- which(nm == "C1P1")
    if (length(idx) == 0L) {
      return(list(
        fit = NULL,
        fit_missing = TRUE,
        fit_available = character(0),
        fit_missing_measures = fit_measures,
        meta = list(n_i = length(i.sol), n_fit_rows = 0L)
      ))
    }
    i.sol <- i.sol[idx[1L]]
    nm <- nm[idx[1L]]
  }

  out <- numeric(0)
  missing_any <- FALSE
  avail_any <- character(0)
  fit_rows <- integer(0)

  for (k in seq_along(i.sol)) {
    el <- i.sol[[k]]
    ic <- if (is.list(el) && "IC" %in% names(el)) el$IC else NULL
    tbl <- if (is.list(ic) && "sol.incl.cov" %in% names(ic)) ic$sol.incl.cov else NULL

    if (is.null(tbl) || !is.data.frame(tbl) || nrow(tbl) == 0L) {
      missing_any <- TRUE
      fit_rows <- c(fit_rows, 0L)
      next
    }

    fit_rows <- c(fit_rows, nrow(tbl))
    available <- intersect(fit_measures, colnames(tbl))
    avail_any <- union(avail_any, available)

    if (nrow(tbl) < which_M) {
      missing_any <- TRUE
    } else {
      vals <- suppressWarnings(as.numeric(tbl[which_M, available, drop = TRUE]))
      names(vals) <- paste0(solution_type, ".", nm[k], ".", available)
      out <- c(out, vals)
    }
  }

  list(
    fit = if (length(out) == 0L) NULL else out,
    fit_missing = missing_any && length(out) == 0L,
    fit_available = avail_any,
    fit_missing_measures = setdiff(fit_measures, avail_any),
    meta = list(n_i = length(i.sol), n_fit_rows = fit_rows)
  )
}

.reduced_extract_solution_type_fit <- function(res, solution_type, which_M, i_mode, fit_measures) {
  if (solution_type %in% c("conservative", "parsimonious")) {
    out <- .reduced_extract_standard_fit(
      res = res,
      solution_type = solution_type,
      which_M = which_M,
      fit_measures = fit_measures
    )
    out$solution_type <- solution_type
    return(out)
  }

  out <- .reduced_extract_intermediate_fit(
    res,
    solution_type = solution_type,
    which_M = which_M,
    i_mode = i_mode,
    fit_measures = fit_measures
  )
  out$solution_type <- solution_type
  out
}

.reduced_as_df <- function(x) {
  if (is.data.frame(x)) return(x)
  as.data.frame(x, stringsAsFactors = FALSE)
}

.reduced_validate_calib_spec <- function(calib_spec, analysis_vars, raw.data, data) {
  if (!is.character(analysis_vars) || length(analysis_vars) < 2L) {
    stop("When `calib = \"recompute\"`, `analysis_vars` must contain the outcome and at least one condition.")
  }

  raw_df <- .reduced_as_df(raw.data)
  data_df <- .reduced_as_df(data)

  for (nm in analysis_vars) {
    if (!nm %in% colnames(data_df)) {
      stop(sprintf("Calibrated set '%s' is not a column in `data`.", nm))
    }
  }

  findTh_specs <- stats::setNames(vector("list", length(analysis_vars)), analysis_vars)
  calib_spec_shared <- calib_spec

  if (!is.null(calib_spec_shared) && is.list(calib_spec_shared) && !is.data.frame(calib_spec_shared)) {
    for (nm in names(calib_spec_shared)) {
      spec <- calib_spec_shared[[nm]]
      if (!is.list(spec) || is.data.frame(spec)) {
        next
      }

      if ("findTh" %in% names(spec)) {
        if (!is.null(spec$findTh) && !is.list(spec$findTh)) {
          stop(sprintf("`calib_spec[['%s']]$findTh` must be NULL or a list of `QCA::findTh()` arguments.", nm))
        }
        if (nm %in% analysis_vars) {
          findTh_specs[[nm]] <- spec$findTh
        }
        spec$findTh <- NULL
        calib_spec_shared[[nm]] <- spec
      }
    }
  }

  outcome <- analysis_vars[[1L]]
  conditions <- analysis_vars[-1L]

  calib_spec_shared <- .normalize_calib_specs(
    conditions = conditions,
    outcome = outcome,
    calib_spec = calib_spec_shared,
    test.outcome = TRUE,
    caller = "reduced-run recalibration"
  )
  calib_spec_shared <- calib_spec_shared[analysis_vars]

  raw_sources <- vapply(calib_spec_shared, function(x) x$raw, character(1))
  missing_raw <- setdiff(raw_sources, colnames(raw_df))
  if (length(missing_raw) > 0L) {
    stop(
      "All raw columns referenced by the calibration specification must exist in `raw.data`. Missing: ",
      paste(missing_raw, collapse = ", "),
      "."
    )
  }

  for (nm in analysis_vars) {
    calib_spec_shared[[nm]]$findTh <- findTh_specs[[nm]]
  }

  calib_spec_shared
}

.reduced_thresholds_from_findTh <- function(raw_x, spec, set_name) {
  ft_args <- c(list(x = raw_x), spec$findTh)
  thresholds_used <- tryCatch(
    suppressWarnings(do.call(QCA::findTh, ft_args)),
    error = function(e) e
  )

  if (inherits(thresholds_used, "error")) {
    return(thresholds_used)
  }

  tryCatch(
    .normalize_calib_thresholds(
      thresholds = thresholds_used,
      type_fc = spec$type,
      method = spec$method,
      condition = set_name,
      caller = "reduced-run recalibration"
    ),
    error = function(e) e
  )
}

.reduced_calibrate_from_spec <- function(raw_x, spec, thresholds) {
  args <- list(
    x = raw_x,
    type = spec$qca_type,
    thresholds = unname(thresholds)
  )

  if (identical(spec$type, "f")) {
    args$method <- spec$method
  }

  do.call(QCA::calibrate, c(args, spec$calibrate))
}

.reduced_recalibrate_dataset <- function(data_step, raw_step, analysis_vars, calib_spec) {
  data_out <- .reduced_as_df(data_step)
  raw_step <- .reduced_as_df(raw_step)

  realized_thresholds <- stats::setNames(vector("list", length(analysis_vars)), analysis_vars)
  realized_calls <- stats::setNames(vector("list", length(analysis_vars)), analysis_vars)

  for (nm in analysis_vars) {
    spec <- calib_spec[[nm]]
    raw_x <- raw_step[[spec$raw]]
    thresholds_used <- spec$thresholds

    if (!is.null(spec$findTh)) {
      thresholds_used <- .reduced_thresholds_from_findTh(
        raw_x = raw_x,
        spec = spec,
        set_name = nm
      )

      if (inherits(thresholds_used, "error")) {
        return(list(
          error = thresholds_used,
          error_source = paste0("findTh:", nm),
          error_message = .reduced_error_message(thresholds_used),
          data = NULL,
          thresholds = realized_thresholds,
          calls = realized_calls
        ))
      }
    }

    cal_res <- tryCatch(
      suppressWarnings(
        .reduced_calibrate_from_spec(
          raw_x = raw_x,
          spec = spec,
          thresholds = thresholds_used
        )
      ),
      error = function(e) e
    )

    if (inherits(cal_res, "error")) {
      return(list(
        error = cal_res,
        error_source = paste0("calibrate:", nm),
        error_message = .reduced_error_message(cal_res),
        data = NULL,
        thresholds = realized_thresholds,
        calls = realized_calls
      ))
    }

    data_out[[nm]] <- cal_res
    realized_thresholds[[nm]] <- thresholds_used
    realized_calls[[nm]] <- list(
      raw = spec$raw,
      type = spec$type,
      method = spec$method,
      findTh = spec$findTh,
      thresholds = thresholds_used,
      calibrate = spec$calibrate
    )
  }

  list(
    error = NULL,
    error_source = NA_character_,
    error_message = NA_character_,
    data = data_out,
    thresholds = realized_thresholds,
    calls = realized_calls
  )
}

.reduced_blank_payload <- function() {
  list(
    res = list(conservative = NULL, parsimonious = NULL, intermediate = NULL),
    sig = list(conservative = NULL, parsimonious = NULL, intermediate = NULL),
    selected_solution_missing = c(conservative = NA, parsimonious = NA, intermediate = NA),
    fit = list(conservative = NULL, parsimonious = NULL, intermediate = NULL),
    fit_missing = c(conservative = NA, parsimonious = NA, intermediate = NA),
    meta = list(conservative = NULL, parsimonious = NULL, intermediate = NULL),
    solution_type_status = c(conservative = NA_character_, parsimonious = NA_character_, intermediate = NA_character_),
    solution_type_error_source = c(conservative = NA_character_, parsimonious = NA_character_, intermediate = NA_character_),
    solution_type_error_message = c(conservative = NA_character_, parsimonious = NA_character_, intermediate = NA_character_)
  )
}

.reduced_make_run_result <- function(
    status,
    tt = NULL,
    exclude_used = NULL,
    calibration = NULL,
    analysis_data = NULL,
    include_analysis_data = FALSE,
    payload = NULL,
    error_source = NA_character_,
    error_message = NA_character_
) {
  if (is.null(payload)) payload <- .reduced_blank_payload()

  head <- list(
    status = status,
    tt = tt,
    exclude_used = exclude_used,
    calibration = calibration
  )

  if (isTRUE(include_analysis_data)) {
    head$analysis_data <- analysis_data
  }

  c(
    head,
    payload,
    list(
      error_source = error_source,
      error_message = error_message
    )
  )
}

.reduced_payload_status <- function(payload, monitored_solutions) {
  solution_type_status <- payload$solution_type_status[monitored_solutions]

  if (any(solution_type_status == "ok", na.rm = TRUE)) {
    return(list(
      status = "ok",
      error_source = NA_character_,
      error_message = NA_character_
    ))
  }

  first_bad <- which(!is.na(solution_type_status) & nzchar(solution_type_status))[1L]

  if (is.na(first_bad)) {
    return(list(
      status = "run_error",
      error_source = NA_character_,
      error_message = NA_character_
    ))
  }

  solution_type <- monitored_solutions[first_bad]

  list(
    status = unname(payload$solution_type_status[[solution_type]]),
    error_source = unname(payload$solution_type_error_source[[solution_type]]),
    error_message = unname(payload$solution_type_error_message[[solution_type]])
  )
}

.reduced_compute_exclude <- function(tt_obj, exclude_mode, exclude_recompute, exclude_static) {
  .qcaert_compute_exclude(
    tt_obj = tt_obj,
    exclude_mode = exclude_mode,
    exclude_recompute = exclude_recompute,
    exclude_static = exclude_static
  )
}

.reduced_run_solution_type <- function(
    tt_obj,
    solution_type,
    ex,
    dots_min,
    dir.exp,
    which_M,
    i_mode,
    fit_measures
) {
  args <- c(
    list(
      input = tt_obj,
      include = if (solution_type == "conservative") "" else "?"
    ),
    dots_min
  )

  if (solution_type == "intermediate") {
    args$dir.exp <- dir.exp
  }

  if (solution_type %in% c("parsimonious", "intermediate") && !is.null(ex)) {
    args$exclude <- ex
  }

  res <- tryCatch(
    suppressWarnings(do.call(QCA::minimize, args)),
    error = function(e) e
  )

  if (inherits(res, "error")) {
    return(list(
      error = res,
      res = NULL,
      sig = NULL,
      selected_solution_missing = NA,
      fit = NULL,
      fit_missing = NA,
      meta = NULL
    ))
  }

  sig_info <- .extract_solution_type_sig(
    res,
    solution_type = solution_type,
    which_M = which_M,
    i_mode = i_mode
  )

  fit_info <- .reduced_extract_solution_type_fit(
    res,
    solution_type = solution_type,
    which_M = which_M,
    i_mode = i_mode,
    fit_measures = fit_measures
  )

  list(
    error = NULL,
    res = res,
    sig = sig_info$sig,
    selected_solution_missing = isTRUE(sig_info$selected_solution_missing),
    fit = fit_info$fit,
    fit_missing = isTRUE(fit_info$fit_missing),
    meta = list(
      solution = sig_info$meta,
      fit = fit_info$meta,
      fit_available = fit_info$fit_available,
      fit_missing_measures = fit_info$fit_missing_measures
    )
  )
}

.reduced_run_once <- function(
    data_step,
    raw_step = NULL,
    calib,
    recalibrate_dataset,
    build_truth_table,
    monitored_solutions,
    exclude_mode,
    exclude_recompute,
    exclude_static,
    dots_min,
    dir.exp,
    which_M,
    i_mode,
    fit_measures,
    include_analysis_data = FALSE
) {
  analysis_data <- .reduced_as_df(data_step)
  calibration_info <- list(
    mode = calib,
    status = if (calib == "fixed") "fixed" else "ok",
    thresholds_used = NULL,
    calls_used = NULL
  )

  if (calib == "recompute") {
    rc <- recalibrate_dataset(data_step = data_step, raw_step = raw_step)

    if (!is.null(rc$error)) {
      calibration_info$status <- "error"
      calibration_info$thresholds_used <- rc$thresholds
      calibration_info$calls_used <- rc$calls

      return(.reduced_make_run_result(
        status = "calibration_error",
        tt = NULL,
        exclude_used = NULL,
        calibration = calibration_info,
        analysis_data = NULL,
        include_analysis_data = include_analysis_data,
        error_source = rc$error_source,
        error_message = rc$error_message
      ))
    }

    analysis_data <- rc$data
    calibration_info$status <- "ok"
    calibration_info$thresholds_used <- rc$thresholds
    calibration_info$calls_used <- rc$calls
  }

  tt_obj <- tryCatch(
    suppressWarnings(build_truth_table(analysis_data)),
    error = function(e) e
  )

  if (inherits(tt_obj, "error")) {
    return(.reduced_make_run_result(
      status = "truth_table_error",
      tt = NULL,
      exclude_used = NULL,
      calibration = calibration_info,
      analysis_data = analysis_data,
      include_analysis_data = include_analysis_data,
      error_source = "truthTable",
      error_message = .reduced_error_message(tt_obj)
    ))
  }

  need_exclude <- any(monitored_solutions %in% c("parsimonious", "intermediate")) &&
    exclude_mode != "none"

  ex <- NULL
  ex_error <- NULL
  if (need_exclude) {
    ex <- .reduced_compute_exclude(
      tt_obj = tt_obj,
      exclude_mode = exclude_mode,
      exclude_recompute = exclude_recompute,
      exclude_static = exclude_static
    )

    if (inherits(ex, "error")) {
      ex_error <- ex
      ex <- NULL
    }
  }

  payload <- .reduced_blank_payload()

  for (solution_type in monitored_solutions) {
    if (solution_type %in% c("parsimonious", "intermediate") && inherits(ex_error, "error")) {
      payload$solution_type_status[[solution_type]] <- "exclude_error"
      payload$solution_type_error_source[[solution_type]] <- "exclude"
      payload$solution_type_error_message[[solution_type]] <- .reduced_error_message(ex_error)
      next
    }

    rr <- .reduced_run_solution_type(
      tt_obj = tt_obj,
      solution_type = solution_type,
      ex = ex,
      dots_min = dots_min,
      dir.exp = dir.exp,
      which_M = which_M,
      i_mode = i_mode,
      fit_measures = fit_measures
    )

    if (!is.null(rr$error)) {
      payload$solution_type_status[[solution_type]] <- "minimize_error"
      payload$solution_type_error_source[[solution_type]] <- solution_type
      payload$solution_type_error_message[[solution_type]] <- .reduced_error_message(rr$error)
      next
    }

    payload$res[[solution_type]] <- rr$res
    payload$sig[[solution_type]] <- rr$sig
    payload$selected_solution_missing[[solution_type]] <- isTRUE(rr$selected_solution_missing)
    payload$fit[[solution_type]] <- rr$fit
    payload$fit_missing[[solution_type]] <- isTRUE(rr$fit_missing)
    payload$meta[[solution_type]] <- rr$meta
    payload$solution_type_status[[solution_type]] <- "ok"
    payload$solution_type_error_source[[solution_type]] <- NA_character_
    payload$solution_type_error_message[[solution_type]] <- NA_character_
  }

  run_status <- .reduced_payload_status(payload, monitored_solutions)

  .reduced_make_run_result(
    status = run_status$status,
    tt = tt_obj,
    exclude_used = ex,
    calibration = calibration_info,
    analysis_data = analysis_data,
    include_analysis_data = include_analysis_data,
    payload = payload,
    error_source = run_status$error_source,
    error_message = run_status$error_message
  )
}

.reduced_comparable_solution_types <- function(base, cur, monitored_solutions) {
  if (is.null(base$solution_type_status) || is.null(cur$solution_type_status)) {
    if (!identical(base$status, "ok") || !identical(cur$status, "ok")) {
      return(character(0))
    }
    return(monitored_solutions)
  }

  base_ok <- base$solution_type_status[monitored_solutions] == "ok"
  cur_ok <- cur$solution_type_status[monitored_solutions] == "ok"
  base_ok[is.na(base_ok)] <- FALSE
  cur_ok[is.na(cur_ok)] <- FALSE
  monitored_solutions[base_ok & cur_ok]
}

.reduced_baseline_solution_types <- function(base, monitored_solutions, fit_measures = character(0), require_fit = FALSE) {
  if (is.null(base$solution_type_status)) {
    if (!identical(base$status, "ok")) {
      return(character(0))
    }
    status_ok <- rep(TRUE, length(monitored_solutions))
  } else {
    status_ok <- base$solution_type_status[monitored_solutions] == "ok"
    status_ok[is.na(status_ok)] <- FALSE
  }

  selected_solution_ok <- !(base$selected_solution_missing[monitored_solutions] %in% TRUE)
  selected_solution_ok[is.na(selected_solution_ok)] <- FALSE

  out <- monitored_solutions[status_ok & selected_solution_ok]

  if (isTRUE(require_fit) && length(fit_measures) > 0L) {
    fit_ok <- !(base$fit_missing[out] %in% TRUE)
    fit_ok[is.na(fit_ok)] <- FALSE
    out <- out[fit_ok]
  }

  out
}

.reduced_no_comparable_status <- function(cur, candidate_solution_types) {
  if (!identical(cur$status, "ok")) {
    return(list(
      status = .reduced_stop_reason_from_status(cur$status),
      error_source = cur$error_source,
      error_message = cur$error_message
    ))
  }

  if (!is.null(cur$solution_type_status)) {
    bad <- candidate_solution_types[cur$solution_type_status[candidate_solution_types] != "ok"]
    bad <- bad[!is.na(bad)]

    if (length(bad) > 0L) {
      solution_type <- bad[1L]
      return(list(
        status = .reduced_stop_reason_from_status(cur$solution_type_status[[solution_type]]),
        error_source = cur$solution_type_error_source[[solution_type]],
        error_message = cur$solution_type_error_message[[solution_type]]
      ))
    }
  }

  list(
    status = "no_comparable_solution",
    error_source = NA_character_,
    error_message = NA_character_
  )
}

.reduced_solution_change_info <- function(base, cur, monitored_solutions) {
  comparable_solutions <- .reduced_comparable_solution_types(
    base = base,
    cur = cur,
    monitored_solutions = monitored_solutions
  )

  if (length(comparable_solutions) == 0L) {
    return(list(
      changed = FALSE,
      changed_types = NA_character_,
      change_kind = NA_character_
    ))
  }

  changed_flags <- vapply(
    comparable_solutions,
    function(solution_type) .sig_changed(base$sig[[solution_type]], cur$sig[[solution_type]]),
    logical(1)
  )

  changed <- any(changed_flags)

  if (!changed) {
    return(list(
      changed = FALSE,
      changed_types = NA_character_,
      change_kind = NA_character_
    ))
  }

  changed_types <- paste(comparable_solutions[changed_flags], collapse = ",")
  change_kind <- paste(
    vapply(
      comparable_solutions[changed_flags],
      function(solution_type) {
        paste0(solution_type, ":", .change_kind_sig(base$sig[[solution_type]], cur$sig[[solution_type]]))
      },
      character(1)
    ),
    collapse = ","
  )

  list(
    changed = TRUE,
    changed_types = changed_types,
    change_kind = change_kind
  )
}

.reduced_fit_change_info <- function(base, cur, monitored_solutions, fit_tol, fit_measures) {
  if (length(fit_measures) == 0L) {
    return(list(
      changed = FALSE,
      changed_types = NA_character_,
      changed_measures = character(0),
      delta = numeric(0),
      n_changed = 0L,
      max_abs_delta = NA_real_
    ))
  }

  comparable_solutions <- .reduced_comparable_solution_types(
    base = base,
    cur = cur,
    monitored_solutions = monitored_solutions
  )

  if (length(comparable_solutions) == 0L) {
    return(list(
      changed = FALSE,
      changed_types = NA_character_,
      changed_measures = character(0),
      delta = numeric(0),
      n_changed = 0L,
      max_abs_delta = NA_real_
    ))
  }

  changed_flags <- vapply(
    comparable_solutions,
    function(solution_type) .fit_changed(base$fit[[solution_type]], cur$fit[[solution_type]], tol = fit_tol),
    logical(1)
  )

  changed <- any(changed_flags)

  if (!changed) {
    return(list(
      changed = FALSE,
      changed_types = NA_character_,
      changed_measures = character(0),
      delta = numeric(0),
      n_changed = 0L,
      max_abs_delta = 0
    ))
  }

  changed_types <- paste(comparable_solutions[changed_flags], collapse = ",")

  changed_measures <- unlist(
    lapply(
      comparable_solutions[changed_flags],
      function(solution_type) .fit_change_names(base$fit[[solution_type]], cur$fit[[solution_type]], tol = fit_tol)
    ),
    use.names = FALSE
  )

  delta_vec <- unlist(
    lapply(
      comparable_solutions[changed_flags],
      function(solution_type) {
        keep_names <- .fit_change_names(base$fit[[solution_type]], cur$fit[[solution_type]], tol = fit_tol)
        dv <- .fit_delta_vec(base$fit[[solution_type]], cur$fit[[solution_type]])
        dv[names(dv) %in% keep_names]
      }
    ),
    use.names = TRUE
  )

  list(
    changed = TRUE,
    changed_types = changed_types,
    changed_measures = sort(unique(changed_measures)),
    delta = delta_vec,
    n_changed = length(unique(changed_measures)),
    max_abs_delta = .fit_abs_max(delta_vec)
  )
}

.reduced_stop_reason_from_status <- function(status, baseline = FALSE) {
  out <- switch(
    status,
    calibration_error = "calibration_error",
    truth_table_error = "truth_table_build_error",
    exclude_error = "exclude_recompute_error",
    minimize_error = "requested_minimize_error",
    status
  )

  if (isTRUE(baseline)) {
    out <- paste0("baseline_", out)
  }

  out
}

.reduced_resolve_case_labels <- function(data, case_labels = NULL) {
  n <- nrow(data)

  full_labels <- if (!is.null(case_labels)) {
    if (!is.character(case_labels) || length(case_labels) != n) {
      stop("`case_labels` must be NULL or a character vector of length `nrow(data)`.")
    }
    as.character(case_labels)
  } else if (!is.null(rownames(data)) && length(rownames(data)) == n) {
    as.character(rownames(data))
  } else {
    as.character(seq_len(n))
  }

  data.frame(
    row_index = seq_len(n),
    case_label = full_labels,
    stringsAsFactors = FALSE
  )
}

.reduced_resolve_case_reference <- function(cases, data, case_labels = NULL) {
  n <- nrow(data)
  all_cases <- .reduced_resolve_case_labels(data = data, case_labels = case_labels)
  full_labels <- all_cases$case_label

  if (is.null(cases)) {
    idx <- seq_len(n)
  } else if (is.numeric(cases)) {
    idx <- suppressWarnings(as.integer(cases))
    if (any(is.na(idx)) || any(idx < 1L | idx > n)) {
      stop("Numeric `cases` must be valid row indices.")
    }
  } else if (is.character(cases)) {
    idx <- match(cases, full_labels)
    if (any(is.na(idx))) {
      stop("Character `cases` contain labels not found in the available case labels / row names.")
    }
  } else {
    stop("`cases` must be NULL, a numeric vector of row indices, or a character vector of case labels.")
  }

  idx <- unique(idx)

  data.frame(
    row_index = idx,
    case_label = full_labels[idx],
    stringsAsFactors = FALSE
  )
}

test_that("loo.test keeps conservative comparable when all-solution exclude recomputation fails", {
  skip_if_not_installed("QCA")

  fixture <- qcaert_fixture_direct6()

  out <- suppressWarnings(loo.test(
    data = fixture$calib,
    outcome = fixture$outcome,
    conditions = fixture$conditions,
    cases = 1,
    incl.cut = 0.75,
    n.cut = 1,
    solution = "all",
    dir.exp = fixture$dir.exp,
    exclude_recompute = list(type = 2, definitely_not_a_findRows_arg = TRUE),
    fit_measures = NULL,
    progress = FALSE
  ))

  expect_identical(out$baseline$status, "ok")
  expect_identical(out$baseline$solution_type_status[["conservative"]], "ok")
  expect_identical(out$baseline$solution_type_status[["parsimonious"]], "exclude_error")
  expect_identical(out$baseline$solution_type_status[["intermediate"]], "exclude_error")
  expect_identical(out$results$status, "ok")
})

test_that("subsample.test keeps conservative comparable when all-solution exclude recomputation fails", {
  skip_if_not_installed("QCA")

  fixture <- qcaert_fixture_direct6()

  out <- suppressWarnings(subsample.test(
    data = fixture$calib,
    outcome = fixture$outcome,
    conditions = fixture$conditions,
    sample_n = 8,
    reps = 1,
    seed = 202,
    incl.cut = 0.75,
    n.cut = 1,
    solution = "all",
    dir.exp = fixture$dir.exp,
    exclude_recompute = list(type = 2, definitely_not_a_findRows_arg = TRUE),
    fit_measures = NULL,
    progress = FALSE
  ))

  expect_identical(out$baseline$status, "ok")
  expect_identical(out$baseline$solution_type_status[["conservative"]], "ok")
  expect_identical(out$baseline$solution_type_status[["parsimonious"]], "exclude_error")
  expect_identical(out$baseline$solution_type_status[["intermediate"]], "exclude_error")
  expect_identical(out$by_run[[1L]]$reduced$status, "ok")
  expect_identical(out$by_run[[1L]]$reduced$solution_type_status[["conservative"]], "ok")
  expect_identical(out$by_run[[1L]]$reduced$solution_type_status[["parsimonious"]], "exclude_error")
  expect_identical(out$by_run[[1L]]$reduced$solution_type_status[["intermediate"]], "exclude_error")
  expect_identical(out$results$status, "ok")
})

test_that("baseline solution_type selection keeps valid solution_types when another baseline solution_type is unusable", {
  baseline <- list(
    status = "ok",
    selected_solution_missing = c(conservative = FALSE, parsimonious = TRUE, intermediate = FALSE),
    fit_missing = c(conservative = FALSE, parsimonious = FALSE, intermediate = TRUE),
    solution_type_status = c(conservative = "ok", parsimonious = "ok", intermediate = "ok")
  )

  monitored <- c("conservative", "parsimonious", "intermediate")

  expect_identical(
    .reduced_baseline_solution_types(baseline, monitored),
    c("conservative", "intermediate")
  )
  expect_identical(
    .reduced_baseline_solution_types(
      baseline,
      monitored,
      fit_measures = "inclS",
      require_fit = TRUE
    ),
    "conservative"
  )
})

test_that("reduced-run recalibration accepts the common calib_spec structure", {
  skip_if_not_installed("QCA")

  fixture <- qcaert_fixture_outcome_calibration()
  analysis_vars <- c(fixture$outcome, fixture$conditions)

  spec <- .reduced_validate_calib_spec(
    calib_spec = fixture$calib_spec,
    analysis_vars = analysis_vars,
    raw.data = fixture$raw,
    data = fixture$calib
  )

  expect_identical(names(spec), analysis_vars)
  expect_identical(spec$Y$qca_type, "fuzzy")
  expect_identical(spec$A$method, "direct")
  expect_equal(unname(spec$A$thresholds), c(10, 20, 30, 70, 80, 90))
  expect_null(spec$A$findTh)

  rebuilt <- .reduced_recalibrate_dataset(
    data_step = fixture$calib[-1, , drop = FALSE],
    raw_step = fixture$raw[-1, , drop = FALSE],
    analysis_vars = analysis_vars,
    calib_spec = spec
  )

  expect_null(rebuilt$error)
  expect_true(all(analysis_vars %in% names(rebuilt$data)))
  expect_identical(names(rebuilt$thresholds), analysis_vars)
  expect_true(all(vapply(rebuilt$thresholds, length, integer(1)) > 0L))
  expect_identical(rebuilt$calls$A$method, "direct")
})

test_that("reduced-run recalibration preserves optional findTh recomputation", {
  skip_if_not_installed("QCA")

  fixture <- qcaert_fixture_outcome_calibration()
  analysis_vars <- c(fixture$outcome, fixture$conditions)
  calib_spec <- fixture$calib_spec
  calib_spec$A$findTh <- list(groups = 4)

  spec <- .reduced_validate_calib_spec(
    calib_spec = calib_spec,
    analysis_vars = analysis_vars,
    raw.data = fixture$raw,
    data = fixture$calib
  )

  expect_identical(spec$A$findTh, list(groups = 4))

  rebuilt <- .reduced_recalibrate_dataset(
    data_step = fixture$calib[-1, , drop = FALSE],
    raw_step = fixture$raw[-1, , drop = FALSE],
    analysis_vars = analysis_vars,
    calib_spec = spec
  )

  expect_null(rebuilt$error)
  expect_identical(rebuilt$calls$A$findTh, list(groups = 4))
  expect_true(length(rebuilt$thresholds$A) %in% c(3L, 6L))
})

test_that("loo.test and subsample.test use the common calib_spec structure when recalibrating", {
  skip_if_not_installed("QCA")

  fixture <- qcaert_fixture_outcome_calibration()

  loo_out <- suppressWarnings(loo.test(
    data = fixture$calib,
    outcome = fixture$outcome,
    conditions = fixture$conditions,
    cases = 1L,
    calib = "recompute",
    raw.data = fixture$raw,
    calib_spec = fixture$calib_spec,
    incl.cut = 0.75,
    n.cut = 1,
    solution = "conservative",
    progress = FALSE
  ))

  expect_s3_class(loo_out, "loo_test")
  expect_identical(loo_out$settings$calib, "recompute")
  expect_identical(loo_out$baseline$calibration$status, "ok")

  subsample_out <- suppressWarnings(subsample.test(
    data = fixture$calib,
    outcome = fixture$outcome,
    conditions = fixture$conditions,
    calib = "recompute",
    raw.data = fixture$raw,
    calib_spec = fixture$calib_spec,
    sample_n = 8,
    reps = 1,
    seed = 123,
    incl.cut = 0.75,
    n.cut = 1,
    solution = "conservative",
    progress = FALSE
  ))

  expect_s3_class(subsample_out, "subsample_test")
  expect_identical(subsample_out$settings$calib, "recompute")
  expect_identical(subsample_out$baseline$calibration$status, "ok")
})

.sanitize_terms <- function(x) {
  x <- as.character(x)
  x <- x[!is.na(x)]
  x <- trimws(x)
  x <- x[nzchar(x)]
  x <- gsub("\\s+", "", x)
  sort(unique(x))
}

.prime_labels <- function(primes) {
  if (is.null(primes)) return(character(0))

  rn <- rownames(primes)
  if (!is.null(rn) && length(rn) > 0L) {
    return(as.character(rn))
  }

  if (is.matrix(primes) || is.data.frame(primes)) {
    return(as.character(seq_len(nrow(primes))))
  }

  if (is.list(primes) && !is.null(names(primes)) && length(names(primes)) > 0L) {
    return(as.character(names(primes)))
  }

  character(0)
}

.solution_terms_from_primes <- function(sol_item, primes) {
  labels <- .prime_labels(primes)

  rec <- function(x) {
    if (is.null(x)) return(character(0))

    if (is.data.frame(x)) {
      x <- as.matrix(x)
    }

    if (is.matrix(x)) {
      rn <- rownames(x)
      if (!is.null(rn) && nrow(x) > 0L) {
        keep <- rep(TRUE, nrow(x))

        if (is.numeric(x) || is.integer(x)) {
          keep <- rowSums(abs(x), na.rm = TRUE) > 0
        } else if (is.logical(x)) {
          keep <- rowSums(x, na.rm = TRUE) > 0
        }

        return(.sanitize_terms(rn[keep]))
      }

      return(rec(as.list(x)))
    }

    if (is.logical(x)) {
      if (length(labels) > 0L && length(x) == length(labels)) {
        return(.sanitize_terms(labels[which(x)]))
      }
      return(character(0))
    }

    if (is.numeric(x)) {
      xi <- suppressWarnings(as.integer(x))
      if (
        length(labels) > 0L &&
        length(xi) > 0L &&
        all(!is.na(xi)) &&
        all(abs(x - xi) < sqrt(.Machine$double.eps)) &&
        all(xi >= 1L & xi <= length(labels))
      ) {
        return(.sanitize_terms(labels[xi]))
      }
      return(.sanitize_terms(x))
    }

    if (is.character(x)) {
      xs <- .sanitize_terms(x)
      if (length(labels) > 0L && all(xs %in% labels)) {
        return(xs)
      }
      return(xs)
    }

    if (is.list(x)) {
      return(.sanitize_terms(unlist(lapply(x, rec), use.names = FALSE)))
    }

    .sanitize_terms(x)
  }

  out <- rec(sol_item)
  if (length(out) == 0L) "<EMPTY>" else out
}

.extract_selected_m_standard_sig <- function(res, which_M) {
  sol_all <- if (!is.null(res) && !inherits(res, "error")) res$solution else NULL
  primes <- if (!is.null(res) && !inherits(res, "error")) res$primes else NULL

  nM <- if (!is.null(sol_all)) length(sol_all) else 0L
  hasM <- !is.null(sol_all) && nM >= which_M
  sig <- if (hasM) .solution_terms_from_primes(sol_all[[which_M]], primes) else "<M_MISSING>"

  list(
    sig = sig,
    n_M = as.integer(nM),
    has_M = isTRUE(hasM)
  )
}

.extract_selected_m_intermediate_sig <- function(i.sol, which_M, i_mode = "all") {
  if (is.null(i.sol)) {
    return(list(sig = "<I_SOL_MISSING>", missing_M = TRUE, n_M_min = 0L))
  }

  nm <- .normalize_i_solution_names(i.sol)

  if (i_mode == "C1P1") {
    idx <- which(nm == "C1P1")
    if (length(idx) == 0L) {
      return(list(sig = "<C1P1_MISSING>", missing_M = TRUE, n_M_min = 0L))
    }
    i.sol <- i.sol[idx[1L]]
    nm <- nm[idx[1L]]
  }

  out <- character(0)
  missing_M <- FALSE
  n_M_min <- Inf

  for (k in seq_along(i.sol)) {
    el <- i.sol[[k]]
    sol_all <- if (is.list(el) && "solution" %in% names(el)) el$solution else NULL
    primes <- if (is.list(el) && "primes" %in% names(el)) el$primes else NULL

    nM <- if (!is.null(sol_all)) length(sol_all) else 0L
    n_M_min <- min(n_M_min, nM)

    if (is.null(sol_all) || nM < which_M) {
      missing_M <- TRUE
      sigs <- "<M_MISSING>"
    } else {
      sigs <- .solution_terms_from_primes(sol_all[[which_M]], primes)
      sigs <- paste(sigs, collapse = "+")
    }

    out <- c(out, paste0(nm[k], ":", sigs))
  }

  if (!is.finite(n_M_min)) n_M_min <- 0L

  list(
    sig = sort(unique(out)),
    missing_M = missing_M,
    n_M_min = as.integer(n_M_min)
  )
}

.extract_standard_sig <- function(res, which_M) {
  sol_all <- res$solution
  primes <- res$primes

  nM <- if (!is.null(sol_all)) length(sol_all) else 0L

  if (nM < which_M) {
    return(list(
      sig = "<M_MISSING>",
      selected_solution_missing = TRUE,
      meta = list(nM = nM)
    ))
  }

  sig <- paste(.solution_terms_from_primes(sol_all[[which_M]], primes), collapse = "+")

  list(
    sig = sig,
    selected_solution_missing = FALSE,
    meta = list(nM = nM)
  )
}

.extract_intermediate_sig <- function(res, which_M, i_mode) {
  i.sol <- res$i.sol

  if (is.null(i.sol) || length(i.sol) == 0L) {
    return(list(
      sig = "<I_SOL_MISSING>",
      selected_solution_missing = TRUE,
      meta = list(n_i = 0L)
    ))
  }

  nm <- .normalize_i_solution_names(i.sol)

  if (i_mode == "C1P1") {
    idx <- which(nm == "C1P1")
    if (length(idx) == 0L) {
      return(list(
        sig = "<C1P1_MISSING>",
        selected_solution_missing = TRUE,
        meta = list(n_i = length(i.sol))
      ))
    }
    i.sol <- i.sol[idx[1L]]
    nm <- nm[idx[1L]]
  }

  out <- character(0)
  missing_any <- FALSE

  for (k in seq_along(i.sol)) {
    el <- i.sol[[k]]
    sol_all <- if (is.list(el) && "solution" %in% names(el)) el$solution else NULL
    primes <- if (is.list(el) && "primes" %in% names(el)) el$primes else NULL

    nM <- if (!is.null(sol_all)) length(sol_all) else 0L

    if (nM < which_M) {
      out <- c(out, paste0(nm[k], ":<M_MISSING>"))
      missing_any <- TRUE
    } else {
      out <- c(
        out,
        paste0(
          nm[k], ":",
          paste(.solution_terms_from_primes(sol_all[[which_M]], primes), collapse = "+")
        )
      )
    }
  }

  list(
    sig = sort(unique(out)),
    selected_solution_missing = missing_any,
    meta = list(n_i = length(i.sol))
  )
}

.extract_solution_type_sig <- function(res, solution_type, which_M, i_mode) {
  if (solution_type %in% c("conservative", "parsimonious")) {
    out <- .extract_standard_sig(res, which_M = which_M)
    out$solution_type <- solution_type
    return(out)
  }

  out <- .extract_intermediate_sig(
    res,
    which_M = which_M,
    i_mode = i_mode
  )
  out$solution_type <- solution_type
  out
}

.sig_changed <- function(base, cur) {
  if (is.null(base) && is.null(cur)) return(FALSE)
  if (xor(is.null(base), is.null(cur))) return(TRUE)
  !setequal(base, cur)
}

.change_kind_sig <- function(base, cur) {
  if (!.sig_changed(base, cur)) return(NA_character_)

  base_missing <- !is.null(base) && any(grepl("<.*MISSING>|<NO_SOLUTIONS>", base))
  cur_missing <- !is.null(cur) && any(grepl("<.*MISSING>|<NO_SOLUTIONS>", cur))

  if (!base_missing && cur_missing) return("selected_solution_missing")
  if (base_missing && !cur_missing) return("selected_solution_appeared")
  if (base_missing && cur_missing) return("mixed_change")
  "formula_changed"
}

.change_kind_selected_m_sig <- function(base, cur) {
  if (!.sig_changed(base, cur)) return(NA_character_)

  base_missing <- !is.null(base) && any(grepl("<M_MISSING>", base, fixed = TRUE))
  cur_missing <- !is.null(cur) && any(grepl("<M_MISSING>", cur, fixed = TRUE))

  if (!base_missing && cur_missing) return("selected_M_disappeared")
  if (base_missing && !cur_missing) return("selected_M_appeared")
  if (base_missing && cur_missing) return("mixed_change")
  "formula_changed"
}

#' Alternative-set robustness test for QCA solutions
#'
#' Repeatedly generates alternative analyses by jointly varying calibration
#' anchors, the raw consistency threshold, and the frequency cutoff for truth
#' table rows, then reruns the QCA analysis for each generated specification.
#' For each specification, the function records whether the solution under
#' evaluation changes, whether parameters of fit change, which conditions or
#' outcome were recalibrated, and whether the analysis failed to produce a
#' solution.
#'
#' @param raw.data A data frame object containing the raw, uncalibrated
#'   condition columns, and the raw outcome column when `test.outcome = TRUE`.
#' @param calib.data A data frame object containing the calibrated outcome
#'   and calibrated condition columns used for the baseline analysis.
#' @param outcome Name of the outcome column in `calib.data`. This must be a
#'   single non-empty character string.
#' @param conditions Character vector of calibrated condition names in
#'   `calib.data`.
#' @param calib_spec Named list of calibration specifications, one per
#'   condition, plus one for the outcome when `test.outcome = TRUE`. Each entry
#'   must contain `raw`, `type`, and `thresholds`, may contain `method`, and may
#'   contain `calibrate`, a named list of additional [QCA::calibrate()]
#'   arguments for that set.
#' @param test.conditions Subset of `conditions` to perturb when generating
#'   random alternative calibration draws. May be `NULL` only when
#'   `test.outcome = TRUE`, in which case only the outcome calibration is
#'   perturbed.
#' @param test.outcome Logical; if `TRUE`, also perturb the outcome calibration.
#'   The outcome must be represented in `calib_spec`, and must not be included
#'   in `conditions`.
#' @param anchors_to_test Character vector of anchors eligible for perturbation,
#'   or `NULL` to use all anchors implied by each tested set's calibration
#'   method. Crisp sets use `"T"`. Fuzzy direct three-threshold sets use
#'   `"E"`, `"C"`, and `"I"`. Fuzzy indirect sets use `"T1"`,
#'   `"T2"`, and so on. Fuzzy direct six-threshold sets use `"E1"`,
#'   `"C1"`, `"I1"`, `"I2"`, `"C2"`, and `"E2"`.
#' @param solution Solution type to monitor. Accepted values are `"all"`,
#'   `"con"` or `"conservative"`, `"par"` or `"parsimonious"`, and `"int"` or
#'   `"intermediate"`.
#' @param include Optional minimization include setting. Currently, this
#'   argument accepts only `NULL`, `""`, or `"?"`.
#' @param which_M Positive integer giving which solution alternative to use
#'   when minimization returns multiple models.
#' @param unit_step Numeric step size used to move calibration thresholds when
#'   building candidate alternative draws. Supply either one value applied to
#'   all calibrated sets or one value per condition, plus the outcome when
#'   `test.outcome = TRUE`. If `NULL`, qcaERT computes scale-aware steps from
#'   each set's threshold spacing.
#' @param unit_step_divisor Positive number used to compute `unit_step`
#'   automatically when `unit_step = NULL`. The default is `10`.
#' @param calib_max_steps Maximum number of threshold moves away from the
#'   baseline used when building alternative calibration candidates.
#' @param incl.cut Baseline inclusion cutoff passed to [QCA::truthTable()].
#' @param incl_step Positive step size used to build the candidate grid of
#'   inclusion-cutoff values around `incl.cut`.
#' @param incl_max_steps Maximum number of stepwise moves away from the
#'   baseline `incl.cut` used to build the candidate grid.
#' @param n.cut Baseline truth table frequency cutoff passed to
#'   [QCA::truthTable()].
#' @param ncut_step Positive integer step size used to build the candidate grid
#'   of frequency-cutoff values around `n.cut`.
#' @param ncut_max_steps Maximum number of stepwise moves away from the
#'   baseline `n.cut` used to build the candidate grid.
#' @param dir.exp Directional expectations used when the monitored solution is
#'   intermediate.
#' @param i_mode Character string controlling intermediate-solution selection.
#'   Accepted values are `"all"` and `"C1P1"`.
#' @param exclude_mode Character string controlling how excluded rows are
#'   handled for parsimonious and intermediate minimization. `"recompute"`
#'   recalculates exclusions from each drawn truth table using
#'   `exclude_recompute`, `"static"` reuses `exclude_static`, and `"none"`
#'   does not use exclusions.
#' @param exclude_recompute Named list of arguments passed to [QCA::findRows()]
#'   when `exclude_mode = "recompute"`.
#' @param exclude_static Already computed exclusion object reused when
#'   `exclude_mode = "static"`.
#' @param n_draws Number of random alternative draws to run.
#' @param fit_tol Non-negative tolerance used when deciding whether fit values
#'   changed.
#' @param seed Optional integer seed passed to [set.seed()] before generating
#'   the random draws.
#' @param progress Logical; if `TRUE` and the session is interactive, show a
#'   text progress bar.
#' @param verbose Logical; if `TRUE`, print a completion message after each
#'   draw.
#' @param x An `altset_test` object returned by [altset.test()].
#' @param row.names Logical; passed to [print.data.frame()] by
#'   [print.altset_test()].
#' @param ... Additional arguments routed through the QCA workflow. Arguments
#'   matching [QCA::truthTable()] are forwarded to truth table construction;
#'   remaining minimization arguments are filtered and forwarded to
#'   [QCA::minimize()]. The function also looks in `...` for `include`,
#'   `dir.exp`, or `direxp` if those arguments were not supplied explicitly.
#'   In [print.altset_test()], `...` is passed to [print.data.frame()]. In
#'   [as.data.frame.altset_test()], `...` is ignored.
#'
#' @returns An object of class `altset_test` with the following components:
#'   \describe{
#'     \item{`summary`}{A list with summary values named `n_draws`,
#'     `n_same_solution`, `n_fit_compared`, `n_same_fit`,
#'     `score_solution`, `score_fit`, `score_total`,
#'     `score_solution_by_solution_type`, and `score_fit_by_solution_type`.}
#'     \item{`baseline`}{A list containing the baseline analysis, baseline draw
#'     metadata, exclusion information, selected solution terms used for
#'     comparison, fit information, and status information.}
#'     \item{`diagnostics`}{A detailed data frame with one row per draw. It
#'     includes the sampled `incl.cut` and `n.cut`, run status,
#'     solution-change information, fit-change information, changed-set
#'     information, and error information when a draw fails.}
#'     \item{`results`}{A compact data frame with the columns
#'     `draw`, `incl.cut`, `n.cut`, `status`, `n_changed_sets`,
#'     `changed_sets`, `changed_roles`, `solution_change`, `fit_changed_types`,
#'     `n_fit_deltas`, and `max_abs_fit_delta`.}
#'     \item{`by_draw`}{A named list with one entry per draw. Each entry stores
#'     the full run result, solution-comparison information, fit-comparison
#'     information, and draw-level fit details.}
#'     \item{`settings`}{A list containing the analysis settings used to build
#'     the result object.}
#'   }
#'
#'   `print.altset_test()` prints a concise summary, the `results` table, and
#'   solution-type-specific match-rate tables. `as.data.frame.altset_test()` returns
#'   the `results` table.
#'
#' @details
#' The function first runs a baseline analysis using `calib.data`,
#' `incl.cut`, and `n.cut`. It then generates `n_draws` random alternative
#' draws.
#'
#' Each draw samples:
#' \itemize{
#'   \item one admissible inclusion cutoff from the grid defined by
#'   `incl.cut`, `incl_step`, and `incl_max_steps`,
#'   \item one admissible frequency cutoff from the grid defined by `n.cut`,
#'   `ncut_step`, and `ncut_max_steps`, and
#'   \item one admissible alternative calibration specification sampled from
#'   `calib_spec` by moving method-specific anchors for `test.conditions`, and
#'   for the outcome when `test.outcome = TRUE`, by integer multiples of each
#'   set's `unit_step`.
#' }
#'
#' The function rejects a sampled draw if it is identical to the baseline on
#' all three dimensions and resamples until it gets a non-baseline draw or
#' reaches the internal retry limit.
#'
#' The shared solution-control, exclusion, calibration-specification, and
#' returned-object conventions are described in `?qcaERT_conventions`.
#'
#' @examples
#' \donttest{
#' library(QCA)
#' data(LR)
#'
#' conditions <- c("DEV", "URB", "LIT", "IND", "STB")
#' outcome <- "SURV"
#' dir_exp <- rep("1", length(conditions))
#'
#' thresholds <- list(
#'   DEV = findTh(LR$DEV, groups = 7),
#'   URB = findTh(LR$URB, groups = 4),
#'   LIT = findTh(LR$LIT, groups = 4),
#'   IND = findTh(LR$IND, groups = 4),
#'   STB = findTh(LR$STB, groups = 4),
#'   SURV = findTh(LR$SURV, groups = 4)
#' )
#'
#' dat <- LR
#' dat$DEV <- calibrate(LR$DEV, type = "fuzzy", thresholds = thresholds$DEV)
#' dat$URB <- calibrate(LR$URB, type = "fuzzy", thresholds = thresholds$URB)
#' dat$LIT <- calibrate(LR$LIT, type = "fuzzy", thresholds = thresholds$LIT)
#' dat$IND <- calibrate(LR$IND, type = "fuzzy", thresholds = thresholds$IND)
#' dat$STB <- calibrate(LR$STB, type = "fuzzy", thresholds = thresholds$STB)
#' dat$SURV <- calibrate(LR$SURV, type = "fuzzy", thresholds = thresholds$SURV)
#'
#' calib_spec <- list(
#'   DEV = list(raw = "DEV", type = "fuzzy", method = "direct", thresholds = thresholds$DEV),
#'   URB = list(raw = "URB", type = "fuzzy", method = "direct", thresholds = thresholds$URB),
#'   LIT = list(raw = "LIT", type = "fuzzy", method = "direct", thresholds = thresholds$LIT),
#'   IND = list(raw = "IND", type = "fuzzy", method = "direct", thresholds = thresholds$IND),
#'   STB = list(raw = "STB", type = "fuzzy", method = "direct", thresholds = thresholds$STB)
#' )
#'
#' calib_spec_outcome <- calib_spec
#' calib_spec_outcome$SURV <- list(
#'   raw = "SURV",
#'   type = "fuzzy",
#'   method = "direct",
#'   thresholds = thresholds$SURV
#' )
#'
#' # Common use: sample alternative calibration, incl.cut, and n.cut settings.
#' altset_out <- altset.test(
#'   raw.data = LR,
#'   calib.data = dat,
#'   outcome = outcome,
#'   conditions = conditions,
#'   calib_spec = calib_spec,
#'   test.conditions = c("DEV", "URB"),
#'   unit_step = NULL,
#'   unit_step_divisor = 10,
#'   calib_max_steps = 5,
#'   incl.cut = 0.8,
#'   incl_step = 0.02,
#'   incl_max_steps = 5,
#'   n.cut = 1,
#'   ncut_step = 1,
#'   ncut_max_steps = 2,
#'   n_draws = 50,
#'   seed = 123,
#'   solution = "all",
#'   dir.exp = dir_exp,
#'   progress = TRUE
#' )
#'
#' altset_out
#' as.data.frame(altset_out)
#' altset_out$summary
#'
#' # Special case: include the calibrated outcome among sampled calibration
#' # perturbations. The outcome stays out of conditions.
#' altset_outcome <- altset.test(
#'   raw.data = LR,
#'   calib.data = dat,
#'   outcome = outcome,
#'   conditions = conditions,
#'   calib_spec = calib_spec_outcome,
#'   test.conditions = c("DEV", "URB"),
#'   test.outcome = TRUE,
#'   unit_step = NULL,
#'   unit_step_divisor = 10,
#'   calib_max_steps = 5,
#'   incl.cut = 0.8,
#'   incl_step = 0.02,
#'   incl_max_steps = 5,
#'   n.cut = 1,
#'   ncut_step = 1,
#'   ncut_max_steps = 2,
#'   n_draws = 50,
#'   seed = 456,
#'   solution = "all",
#'   dir.exp = dir_exp,
#'   progress = TRUE
#' )
#'
#' as.data.frame(altset_outcome)
#' }
#'
#' @seealso [calib.test()], [incl.test()], [ncut.test()], [loo.test()],
#'   [subsample.test()], [theory.test()], [cluster.test()], [sol.df()]
#' @export
altset.test <- function(
    raw.data,
    calib.data,
    outcome,
    conditions,
    calib_spec,
    test.conditions = conditions,
    test.outcome = FALSE,
    anchors_to_test = NULL,
    solution = "all",
    include = NULL,
    which_M = 1,
    unit_step = NULL,
    unit_step_divisor = 10,
    calib_max_steps = 20,
    incl.cut = 1,
    incl_step = 0.01,
    incl_max_steps = 20,
    n.cut = 1,
    ncut_step = 1,
    ncut_max_steps = 20,
    dir.exp = NULL,
    i_mode = c("all", "C1P1"),
    exclude_mode = c("recompute", "static", "none"),
    exclude_recompute = list(type = 2),
    exclude_static = NULL,
    n_draws = 100,
    fit_tol = 1e-6,
    seed = NULL,
    progress = TRUE,
    verbose = FALSE,
    ...
) {
  .require_qca()

  selection_controls <- .normalize_i_mode(i_mode)
  i_mode <- selection_controls$i_mode
  exclude_mode <- match.arg(exclude_mode)

  altset <- match.call(expand.dots = FALSE)
  dots_raw <- list(...)
  caller_env <- parent.frame()
  .reject_exclusion_controls_in_dots(dots_raw, "altset.test")
  .reject_calibration_inputs_in_dots(dots_raw, "altset.test")

  # basic checks --
  if (is.null(dim(raw.data)) || is.null(nrow(raw.data)) || nrow(raw.data) < 1L) {
    stop("`raw.data` must be a non-empty data frame object with at least one row.")
  }
  if (is.null(dim(calib.data)) || is.null(nrow(calib.data)) || nrow(calib.data) < 1L) {
    stop("`calib.data` must be a non-empty data frame object with at least one row.")
  }
  if (nrow(raw.data) != nrow(calib.data)) {
    stop("`raw.data` and `calib.data` must have the same number of rows.")
  }
  if (!is.character(outcome) || length(outcome) != 1L || !nzchar(outcome)) {
    stop("`outcome` must be a single non-empty character string.")
  }
  if (!outcome %in% colnames(calib.data)) {
    stop("`outcome` must exist in `calib.data`.")
  }
  if (!is.character(conditions) || length(conditions) < 1L) {
    stop("`conditions` must be a non-empty character vector.")
  }
  if (!all(conditions %in% colnames(calib.data))) {
    stop("All `conditions` must exist in `calib.data`.")
  }
  .validate_outcome_conditions_distinct(outcome, conditions)

  test.outcome <- .normalize_test_outcome(test.outcome)
  test.conditions <- .normalize_test_conditions(
    test.conditions = test.conditions,
    conditions = conditions,
    test.outcome = test.outcome
  )
  if (!is.numeric(incl.cut) || length(incl.cut) != 1L || !is.finite(incl.cut) ||
      incl.cut < 0 || incl.cut > 1) {
    stop("`incl.cut` must be a single finite number in [0, 1].")
  }
  if (!is.numeric(incl_step) || length(incl_step) != 1L || !is.finite(incl_step) || incl_step <= 0) {
    stop("`incl_step` must be a single finite number > 0.")
  }
  if (!is.numeric(fit_tol) || length(fit_tol) != 1L || !is.finite(fit_tol) || fit_tol < 0) {
    stop("`fit_tol` must be a single finite number >= 0.")
  }
  n_draws <- .as_integerish_scalar(n_draws, "n_draws", min = 1L)

  n.cut <- .as_integerish_scalar(n.cut, "n.cut", min = 1L)
  ncut_step <- .as_integerish_scalar(ncut_step, "ncut_step", min = 1L)
  ncut_max_steps <- .as_integerish_scalar(ncut_max_steps, "ncut_max_steps", min = 1L)
  incl_max_steps <- .as_integerish_scalar(incl_max_steps, "incl_max_steps", min = 1L)
  calib_max_steps <- .as_integerish_scalar(calib_max_steps, "calib_max_steps", min = 1L)

  # minimize-style conveniences from calib.test -
  if (!is.null(altset$outcome)) {
    expr <- altset$outcome
    val <- .safe_eval_expr(expr, envir = caller_env)
    if (!inherits(val, "error") && is.character(val) && length(val) == 1) {
      outcome <- val
    } else {
      outcome <- .expr_to_chrvec(expr)
      if (length(outcome) != 1) stop("outcome must resolve to a single outcome name.")
    }
  }

  if (!is.null(altset$conditions)) {
    expr <- altset$conditions
    val <- .safe_eval_expr(expr, envir = caller_env)
    if (!inherits(val, "error") && is.character(val) && length(val) >= 1) {
      conditions <- val
    } else {
      conditions <- .expr_to_chrvec(expr)
    }
  }

  if (!is.null(altset[["dir.exp"]])) {
    expr <- altset[["dir.exp"]]
    val <- .safe_eval_expr(expr, envir = caller_env)
    if (!inherits(val, "error")) {
      dir.exp <- val
    } else {
      dir.exp <- .expr_to_chrvec(expr)
    }
  }

  dir.exp <- .normalize_dir_exp_generic(dir.exp, conditions, endpoint_phrase = "`conditions`")

  if (is.null(dir.exp)) {
    de <- .dot_get(dots_raw, "dir.exp")
    if (is.null(de)) de <- .dot_get(dots_raw, "direxp")
    if (!is.null(de)) dir.exp <- de
  }

  dir.exp <- .normalize_dir_exp_generic(dir.exp, conditions, endpoint_phrase = "`conditions`")

  if (is.null(include)) {
    inc <- .dot_get(dots_raw, "include")
    if (!is.null(inc)) include <- inc
  }

  solution_controls <- .resolve_solution_controls(
    solution = solution,
    include = include,
    dir.exp = dir.exp,
    caller = "altset.test",
    style = "plain"
  )

  solution <- solution_controls$solution
  include <- solution_controls$include
  monitored_solutions <- solution_controls$monitored

  .validate_exclusion_controls(
    exclude_mode = exclude_mode,
    exclude_recompute = exclude_recompute,
    exclude_static = exclude_static,
    exclude_recompute_supplied = !is.null(altset$exclude_recompute),
    exclude_static_supplied = !is.null(altset$exclude_static),
    monitored_solutions = monitored_solutions,
    style = "plain"
  )

  .err_msg <- function(e) {
    if (!inherits(e, "error")) return(NA_character_)
    tryCatch(conditionMessage(e), error = function(...) as.character(e))
  }

  # fit helpers aligned to selected solution keys -
  .extract_standard_fit_map <- function(res, which_M) {
    sol_all <- if (!is.null(res) && !inherits(res, "error")) res$solution else NULL
    primes <- if (!is.null(res) && !inherits(res, "error")) res$primes else NULL
    ic_obj <- if (!is.null(res) && !inherits(res, "error")) res$IC else NULL

    nM <- if (!is.null(sol_all)) length(sol_all) else 0L
    out <- list()

    if (nM < which_M) return(out)

    key <- paste(.solution_terms_from_primes(sol_all[[which_M]], primes), collapse = "+")
    vec <- .fit_vec_from_ic(ic_obj, which_M)
    if (!is.null(vec)) out[[key]] <- vec

    out
  }

  .extract_intermediate_fit_map <- function(res, which_M, i_mode) {
    i.sol <- if (!is.null(res) && !inherits(res, "error")) res$i.sol else NULL
    if (is.null(i.sol) || length(i.sol) == 0L) return(list())

    nm <- .normalize_i_solution_names(i.sol)

    if (i_mode == "C1P1") {
      idx <- which(nm == "C1P1")
      if (length(idx) == 0L) return(list())
      i.sol <- i.sol[idx[1L]]
      nm <- nm[idx[1L]]
    }

    out <- list()

    for (k in seq_along(i.sol)) {
      el <- i.sol[[k]]
      sol_all <- if (is.list(el) && "solution" %in% names(el)) el$solution else NULL
      primes <- if (is.list(el) && "primes" %in% names(el)) el$primes else NULL
      ic_obj <- if (is.list(el) && "IC" %in% names(el)) el$IC else NULL

      nM <- if (!is.null(sol_all)) length(sol_all) else 0L

      if (nM < which_M) next

      key <- paste0(
        nm[k], ":",
        paste(.solution_terms_from_primes(sol_all[[which_M]], primes), collapse = "+")
      )
      vec <- .fit_vec_from_ic(ic_obj, which_M)
      if (!is.null(vec)) out[[key]] <- vec
    }

    out
  }

  .extract_solution_type_fit <- function(res, solution_type, which_M, i_mode) {
    if (solution_type %in% c("conservative", "parsimonious")) {
      return(list(
        raw = if (!is.null(res) && !inherits(res, "error")) res$IC else NULL,
        map = .extract_standard_fit_map(res, which_M = which_M)
        ))
    }
    list(
      raw = if (!is.null(res) && !inherits(res, "error")) lapply(res$i.sol, function(x) x$IC) else NULL,
      map = .extract_intermediate_fit_map(res, which_M = which_M, i_mode = i_mode)
      )
  }

  .compare_solution_type_fit <- function(base, cur, solution_type, tol) {
    base_map <- base$fit[[solution_type]]$map
    cur_map <- cur$fit[[solution_type]]$map

    .compare_fit_maps_hybrid(
      base_map = base_map,
      cur_map = cur_map,
      solution_type = solution_type,
      tol = tol
    )
  }

  # QCA run helpers -
  .build_truth_table <- function(data_step, incl_cutoff, n_cutoff) {
    args <- c(
      list(
        data = data_step,
        outcome = outcome,
        incl.cut = incl_cutoff,
        n.cut = n_cutoff
      ),
      dots_tt
    )

    if (!is.null(conditions)) {
      args$conditions <- conditions
    }

    do.call(QCA::truthTable, args)
  }

  .compute_exclude <- function(tt_obj) {
    .qcaert_compute_exclude(
      tt_obj = tt_obj,
      exclude_mode = exclude_mode,
      exclude_recompute = exclude_recompute,
      exclude_static = exclude_static
    )
  }

  .run_solution_type <- function(tt_obj, solution_type, ex, dots_filtered) {
    args <- c(
      list(
        input = tt_obj,
        include = if (solution_type == "conservative") "" else "?"
      ),
      dots_filtered
    )

    if (solution_type == "intermediate") {
      args$dir.exp <- dir.exp
    }

    if (solution_type %in% c("parsimonious", "intermediate") && !is.null(ex)) {
      args$exclude <- ex
    }

    res <- tryCatch(
      suppressWarnings(do.call(QCA::minimize, args)),
      error = function(e) e
    )

    if (inherits(res, "error")) {
      return(list(
        error = res,
        res = NULL,
        sig = NULL,
        fit = NULL,
        selected_solution_missing = NA,
        meta = NULL
      ))
    }

    sig_info <- .extract_solution_type_sig(
      res,
      solution_type = solution_type,
      which_M = which_M,
      i_mode = i_mode
    )

    fit_info <- .extract_solution_type_fit(
      res,
      solution_type = solution_type,
      which_M = which_M,
      i_mode = i_mode
    )

    list(
      error = NULL,
      res = res,
      sig = sig_info$sig,
      fit = fit_info,
      selected_solution_missing = isTRUE(sig_info$selected_solution_missing),
      meta = sig_info$meta
    )
  }

  .empty_solution_type_list <- function() {
    list(conservative = NULL, parsimonious = NULL, intermediate = NULL)
  }

  .empty_solution_type_status <- function() {
    c(conservative = NA_character_, parsimonious = NA_character_, intermediate = NA_character_)
  }

  .run_status_from_solution_types <- function(solution_type_status, solution_type_error_source, solution_type_error_message, monitored_solutions) {
    status_vec <- solution_type_status[monitored_solutions]

    if (any(status_vec == "ok", na.rm = TRUE)) {
      return(list(
        status = "ok",
        error_source = NA_character_,
        error_message = NA_character_
      ))
    }

    first_bad <- which(!is.na(status_vec) & nzchar(status_vec))[1L]

    if (is.na(first_bad)) {
      return(list(
        status = "run_error",
        error_source = NA_character_,
        error_message = NA_character_
      ))
    }

    solution_type <- monitored_solutions[first_bad]

    list(
      status = unname(solution_type_status[[solution_type]]),
      error_source = unname(solution_type_error_source[[solution_type]]),
      error_message = unname(solution_type_error_message[[solution_type]])
    )
  }

  .baseline_solution_types <- function(base, monitored_solutions) {
    if (is.null(base$solution_type_status)) {
      status_ok <- rep(identical(base$status, "ok"), length(monitored_solutions))
    } else {
      status_ok <- base$solution_type_status[monitored_solutions] == "ok"
      status_ok[is.na(status_ok)] <- FALSE
    }

    selected_solution_ok <- !(base$selected_solution_missing[monitored_solutions] %in% TRUE)
    selected_solution_ok[is.na(selected_solution_ok)] <- FALSE

    monitored_solutions[status_ok & selected_solution_ok]
  }

  .comparable_solution_types <- function(base, cur, monitored_solutions) {
    if (is.null(base$solution_type_status) || is.null(cur$solution_type_status)) {
      if (!identical(base$status, "ok") || !identical(cur$status, "ok")) {
        return(character(0))
      }
      return(monitored_solutions)
    }

    base_ok <- base$solution_type_status[monitored_solutions] == "ok"
    cur_ok <- cur$solution_type_status[monitored_solutions] == "ok"
    base_ok[is.na(base_ok)] <- FALSE
    cur_ok[is.na(cur_ok)] <- FALSE

    monitored_solutions[base_ok & cur_ok]
  }

  .no_comparable_status <- function(cur, candidate_solution_types) {
    if (!identical(cur$status, "ok")) {
      return(list(
        status = cur$status,
        error_source = cur$error_source,
        error_message = cur$error_message
      ))
    }

    if (!is.null(cur$solution_type_status)) {
      bad <- candidate_solution_types[cur$solution_type_status[candidate_solution_types] != "ok"]
      bad <- bad[!is.na(bad)]

      if (length(bad) > 0L) {
        solution_type <- bad[1L]
        return(list(
          status = cur$solution_type_status[[solution_type]],
          error_source = cur$solution_type_error_source[[solution_type]],
          error_message = cur$solution_type_error_message[[solution_type]]
        ))
      }
    }

    list(
      status = "no_comparable_solution",
      error_source = NA_character_,
      error_message = NA_character_
    )
  }

  .run_once <- function(data_step, incl_cutoff, n_cutoff, draw_meta) {
    tt_obj <- tryCatch(
      suppressWarnings(.build_truth_table(data_step, incl_cutoff = incl_cutoff, n_cutoff = n_cutoff)),
      error = function(e) e
    )

    if (inherits(tt_obj, "error")) {
      solution_type_status <- .empty_solution_type_status()
      solution_type_status[monitored_solutions] <- "truth_table_error"
      solution_type_error_source <- .empty_solution_type_status()
      solution_type_error_source[monitored_solutions] <- "truthTable"
      solution_type_error_message <- .empty_solution_type_status()
      solution_type_error_message[monitored_solutions] <- .err_msg(tt_obj)

      return(list(
        status = "truth_table_error",
        data = data_step,
        tt = NULL,
        exclude_used = NULL,
        res = .empty_solution_type_list(),
        sig = .empty_solution_type_list(),
        fit = .empty_solution_type_list(),
        selected_solution_missing = c(conservative = NA, parsimonious = NA, intermediate = NA),
        meta = .empty_solution_type_list(),
        solution_type_status = solution_type_status,
        solution_type_error_source = solution_type_error_source,
        solution_type_error_message = solution_type_error_message,
        error_source = "truthTable",
        error_message = .err_msg(tt_obj),
        draw_meta = draw_meta
      ))
    }

    need_exclude <- any(monitored_solutions %in% c("parsimonious", "intermediate"))

    ex <- NULL
    ex_error <- NULL
    if (need_exclude) {
      ex <- .compute_exclude(tt_obj)

      if (inherits(ex, "error")) {
        ex_error <- ex
        ex <- NULL
      }
    }

    res_out <- .empty_solution_type_list()
    sig_out <- .empty_solution_type_list()
    fit_out <- .empty_solution_type_list()
    selected_solution_out <- c(conservative = NA, parsimonious = NA, intermediate = NA)
    meta_out <- .empty_solution_type_list()
    solution_type_status <- .empty_solution_type_status()
    solution_type_error_source <- .empty_solution_type_status()
    solution_type_error_message <- .empty_solution_type_status()

    for (solution_type in monitored_solutions) {
      if (solution_type %in% c("parsimonious", "intermediate") && inherits(ex_error, "error")) {
        solution_type_status[[solution_type]] <- "exclude_error"
        solution_type_error_source[[solution_type]] <- "exclude"
        solution_type_error_message[[solution_type]] <- .err_msg(ex_error)
        next
      }

      rr <- .run_solution_type(tt_obj, solution_type = solution_type, ex = ex, dots_filtered = dots_filtered)

      if (!is.null(rr$error)) {
        solution_type_status[[solution_type]] <- "minimize_error"
        solution_type_error_source[[solution_type]] <- solution_type
        solution_type_error_message[[solution_type]] <- .err_msg(rr$error)
        next
      }

      res_out[[solution_type]] <- rr$res
      sig_out[[solution_type]] <- rr$sig
      fit_out[[solution_type]] <- rr$fit
      selected_solution_out[[solution_type]] <- isTRUE(rr$selected_solution_missing)
      meta_out[[solution_type]] <- rr$meta
      solution_type_status[[solution_type]] <- "ok"
      solution_type_error_source[[solution_type]] <- NA_character_
      solution_type_error_message[[solution_type]] <- NA_character_
    }

    run_status <- .run_status_from_solution_types(
      solution_type_status = solution_type_status,
      solution_type_error_source = solution_type_error_source,
      solution_type_error_message = solution_type_error_message,
      monitored_solutions = monitored_solutions
    )

    list(
      status = run_status$status,
      data = data_step,
      tt = tt_obj,
      exclude_used = ex,
      res = res_out,
      sig = sig_out,
      fit = fit_out,
      selected_solution_missing = selected_solution_out,
      meta = meta_out,
      solution_type_status = solution_type_status,
      solution_type_error_source = solution_type_error_source,
      solution_type_error_message = solution_type_error_message,
      error_source = run_status$error_source,
      error_message = run_status$error_message,
      draw_meta = draw_meta
    )
  }

  .solution_change_info <- function(base, cur, monitored_solutions) {
    changed_flags <- vapply(
      monitored_solutions,
      function(solution_type) .sig_changed(base$sig[[solution_type]], cur$sig[[solution_type]]),
      logical(1)
    )

    changed <- any(changed_flags)
    changed_by_solution_type <- setNames(as.list(changed_flags), monitored_solutions)

    if (!changed) {
      return(list(
        changed = FALSE,
        changed_overall = FALSE,
        changed_types = NA_character_,
        change_kind = NA_character_,
        changed_by_solution_type = changed_by_solution_type
      ))
    }

    changed_types <- paste(monitored_solutions[changed_flags], collapse = ",")
    change_kind <- paste(
      vapply(
        monitored_solutions[changed_flags],
        function(solution_type) {
          paste0(solution_type, ":", .change_kind_sig(base$sig[[solution_type]], cur$sig[[solution_type]]))
        },
        character(1)
      ),
      collapse = ","
    )

    list(
      changed = TRUE,
      changed_overall = TRUE,
      changed_types = changed_types,
      change_kind = change_kind,
      changed_by_solution_type = changed_by_solution_type
    )
  }

  # materialize settings -
  which_M <- .coerce_which_M(which_M)
  calib_context <- .prepare_calib_context(
    conditions = conditions,
    outcome = outcome,
    calib_spec = calib_spec,
    test.outcome = test.outcome,
    raw.data = raw.data,
    unit_step = unit_step,
    unit_step_divisor = unit_step_divisor,
    anchors_to_test = anchors_to_test,
    caller = "altset.test"
  )
  calib_specs <- calib_context$calib_specs
  raw_targets <- calib_context$raw_targets
  type_fc <- calib_context$type_targets
  thr_list <- calib_context$thresholds_targets
  unit_step_vec <- calib_context$unit_step_targets
  anchors_to_test <- calib_context$anchors_to_test
  dots_split <- .split_truth_table_minimize_dots(dots_raw)
  dots_tt <- dots_split$tt
  dots_filtered <- dots_split$min

  tested_sets <- c(test.conditions, if (test.outcome) outcome else character(0))
  tested_roles <- c(rep("condition", length(test.conditions)), if (test.outcome) "outcome" else character(0))
  names(tested_roles) <- tested_sets

  incl_grid <- incl.cut + seq.int(-incl_max_steps, incl_max_steps) * incl_step
  incl_grid <- sort(unique(incl_grid[incl_grid >= 0 & incl_grid <= 1]))
  if (length(incl_grid) == 0L) {
    stop("No admissible `incl.cut` values were generated from `incl_step` and `incl_max_steps`.")
  }

  ncut_grid <- n.cut + seq.int(-ncut_max_steps, ncut_max_steps) * ncut_step
  ncut_grid <- as.integer(sort(unique(ncut_grid[ncut_grid >= 1L & ncut_grid <= nrow(calib.data)])))
  if (length(ncut_grid) == 0L) {
    stop("No admissible `n.cut` values were generated from `ncut_step` and `ncut_max_steps`.")
  }

  candidate_map <- Map(
    function(set_name, role) {
      .altset_build_calib_candidate(
        set_name = set_name,
        role = role,
        calib_specs = calib_specs,
        raw.data = raw.data,
        unit_step_vec = unit_step_vec,
        anchors_to_test = anchors_to_test,
        calib_max_steps = calib_max_steps
      )
    },
    tested_sets,
    tested_roles
  )
  names(candidate_map) <- tested_sets

  if (!is.null(seed)) set.seed(seed)

  # baseline -
  baseline_draw_meta <- list(
    draw = 0L,
    incl.cut = incl.cut,
    n.cut = n.cut,
    calibration = lapply(names(thr_list), function(nm) {
      list(
        set = nm,
        role = if (nm %in% conditions) "condition" else "outcome",
        raw = raw_targets[[nm]],
        type = type_fc[[nm]],
        method = calib_specs[[nm]]$method,
        thresholds_baseline = thr_list[[nm]],
        thresholds_draw = thr_list[[nm]],
        delta_steps = .calib_zero_delta_steps(calib_specs[[nm]]),
        changed = FALSE
      )
    })
  )
  names(baseline_draw_meta$calibration) <- names(thr_list)

  baseline <- .run_once(
    data_step = calib.data,
    incl_cutoff = incl.cut,
    n_cutoff = n.cut,
    draw_meta = baseline_draw_meta
  )

  baseline_solution_types <- .baseline_solution_types(
    base = baseline,
    monitored_solutions = monitored_solutions
  )

  if (length(baseline_solution_types) == 0L) {
    stop(
      "Baseline model is not valid for comparison. ",
      if (baseline$status != "ok") {
        paste0("Status: ", baseline$status, "; source: ", baseline$error_source, "; message: ", baseline$error_message)
      } else {
        "The selected baseline solution is missing."
      }
    )
  }

  # draws -
  progress_state <- .new_qcaert_progress(total = n_draws, progress = progress)
  on.exit(progress_state$close(), add = TRUE)

  diag_rows <- vector("list", n_draws)
  by_draw <- vector("list", n_draws)

  for (b in seq_len(n_draws)) {
    tries <- 0L

    repeat {
      tries <- tries + 1L

      incl_b <- sample(incl_grid, size = 1L)
      ncut_b <- sample(ncut_grid, size = 1L)
      calib_b <- .altset_sample_calibration_draw(
        candidate_map = candidate_map,
        calib.data = calib.data,
        raw.data = raw.data,
        thr_list = thr_list
      )

      any_change <- !isTRUE(all.equal(incl_b, incl.cut)) ||
        !identical(ncut_b, n.cut) ||
        isTRUE(calib_b$changed)

      if (any_change) break
      if (tries >= 1000L) {
        stop("Could not generate a non-baseline random draw after 1000 attempts.")
      }
    }

    draw_meta <- list(
      draw = b,
      incl.cut = incl_b,
      n.cut = ncut_b,
      calibration = calib_b$changes
    )

    cur <- .run_once(
      data_step = calib_b$data,
      incl_cutoff = incl_b,
      n_cutoff = ncut_b,
      draw_meta = draw_meta
    )

    comparable_solution_types <- .comparable_solution_types(
      base = baseline,
      cur = cur,
      monitored_solutions = baseline_solution_types
    )
    status_ok <- identical(cur$status, "ok") && length(comparable_solution_types) > 0L
    status_info <- if (status_ok) {
      list(
        status = "ok",
        error_source = NA_character_,
        error_message = NA_character_
      )
    } else {
      .no_comparable_status(
        cur = cur,
        candidate_solution_types = baseline_solution_types
      )
    }

    solution_cmp <- if (status_ok) {
      .solution_change_info(baseline, cur, monitored_solutions = comparable_solution_types)
    } else {
      list(
        changed = NA,
        changed_overall = NA,
        changed_types = NA_character_,
        change_kind = NA_character_,
        changed_by_solution_type = setNames(as.list(rep(NA, length(monitored_solutions))), monitored_solutions)
      )
    }

    fit_delta_by_solution_type <- setNames(as.list(rep(NA_real_, length(monitored_solutions))), monitored_solutions)
    fit_compared_by_solution_type <- setNames(as.list(rep(FALSE, length(monitored_solutions))), monitored_solutions)
    fit_keys_by_solution_type <- setNames(vector("list", length(monitored_solutions)), monitored_solutions)
    fit_details <- setNames(vector("list", length(monitored_solutions)), monitored_solutions)
    fit_changed_names_by_solution_type <- setNames(vector("list", length(monitored_solutions)), monitored_solutions)
    fit_delta_vec_by_solution_type <- setNames(vector("list", length(monitored_solutions)), monitored_solutions)

    if (status_ok) {
      for (solution_type in comparable_solution_types) {
        cmp <- .compare_solution_type_fit(baseline, cur, solution_type = solution_type, tol = fit_tol)
        fit_delta_by_solution_type[[solution_type]] <- cmp$max_abs_delta
        fit_compared_by_solution_type[[solution_type]] <- isTRUE(cmp$fit_compared)
        fit_keys_by_solution_type[[solution_type]] <- cmp$matched_keys
        fit_details[[solution_type]] <- cmp$details
        fit_changed_names_by_solution_type[[solution_type]] <- cmp$changed_names
        fit_delta_vec_by_solution_type[[solution_type]] <- cmp$delta
      }
    }

    compared_solution_types <- monitored_solutions[unlist(fit_compared_by_solution_type)]

    fit_changed_solution_types <- compared_solution_types[
      vapply(
        compared_solution_types,
        function(solution_type) length(fit_changed_names_by_solution_type[[solution_type]]) > 0L,
        logical(1)
      )
    ]

    fit_changed_measures <- if (length(compared_solution_types) > 0L) {
      sort(unique(unlist(fit_changed_names_by_solution_type[compared_solution_types], use.names = FALSE)))
    } else {
      character(0)
    }

    fit_delta <- if (length(compared_solution_types) > 0L) {
      unlist(fit_delta_vec_by_solution_type[compared_solution_types], use.names = TRUE)
    } else {
      numeric(0)
    }

    solution_changed <- if (status_ok) isTRUE(solution_cmp$changed_overall) else NA
    fit_changed <- if (status_ok && length(compared_solution_types) > 0L) length(fit_changed_measures) > 0L else NA
    changed_any <- if (status_ok) isTRUE(solution_changed) || isTRUE(fit_changed) else NA
    fit_changed_types <- if (status_ok && length(fit_changed_solution_types) > 0L) paste(fit_changed_solution_types, collapse = ",") else NA_character_
    n_fit_deltas <- if (status_ok) as.integer(length(fit_changed_measures)) else NA_integer_
    max_abs_fit_delta <- if (status_ok) .fit_abs_max(fit_delta) else NA_real_

    calib_changed <- Filter(function(x) isTRUE(x$changed), draw_meta$calibration)
    calib_changed_sets <- names(calib_changed)
    calib_changed_roles <- unique(vapply(calib_changed, function(x) x$role, character(1)))
    if (length(calib_changed_sets) == 0L) calib_changed_sets <- NA_character_
    if (length(calib_changed_roles) == 0L) calib_changed_roles <- NA_character_

    diag_rows[[b]] <- data.frame(
      draw = b,
      incl.cut = incl_b,
      n.cut = ncut_b,
      status = status_info$status,
      changed = changed_any,
      solution_changed = solution_changed,
      changed_types = solution_cmp$changed_types,
      change_kind = solution_cmp$change_kind,
      fit_compared = length(compared_solution_types) > 0L,
      fit_changed = fit_changed,
      fit_changed_types = fit_changed_types,
      n_fit_deltas = n_fit_deltas,
      max_abs_fit_delta = max_abs_fit_delta,
      error_source = status_info$error_source,
      error_message = status_info$error_message,
      changed_sets = paste(calib_changed_sets, collapse = ","),
      changed_roles = paste(calib_changed_roles, collapse = ","),
      n_changed_sets = sum(vapply(draw_meta$calibration, function(x) isTRUE(x$changed), logical(1))),
      stringsAsFactors = FALSE
    )
    diag_rows[[b]]$exclude_used <- I(list(cur$exclude_used))
    diag_rows[[b]]$calibration <- I(list(draw_meta$calibration))

    for (solution_type in monitored_solutions) {
      diag_rows[[b]][[paste0("status_", solution_type)]] <- cur$solution_type_status[[solution_type]]
      diag_rows[[b]][[paste0("error_source_", solution_type)]] <- cur$solution_type_error_source[[solution_type]]
      diag_rows[[b]][[paste0("error_message_", solution_type)]] <- cur$solution_type_error_message[[solution_type]]
      diag_rows[[b]][[paste0("solution_changed_", solution_type)]] <- if (status_ok && (solution_type %in% comparable_solution_types)) isTRUE(solution_cmp$changed_by_solution_type[[solution_type]]) else NA
      diag_rows[[b]][[paste0("fit_compared_", solution_type)]] <- isTRUE(fit_compared_by_solution_type[[solution_type]])
      diag_rows[[b]][[paste0("fit_changed_", solution_type)]] <- if (status_ok && isTRUE(fit_compared_by_solution_type[[solution_type]])) length(fit_changed_names_by_solution_type[[solution_type]]) > 0L else NA
      diag_rows[[b]][[paste0("max_abs_fit_delta_", solution_type)]] <- fit_delta_by_solution_type[[solution_type]]
      diag_rows[[b]][[paste0("matched_fit_keys_", solution_type)]] <- paste(fit_keys_by_solution_type[[solution_type]], collapse = ",")
    }

    by_draw[[b]] <- list(
      result = cur,
      solution_comparison = solution_cmp,
      comparable_solutions = comparable_solution_types,
      solution_changed = solution_changed,
      fit_changed = fit_changed,
      fit_changed_types = fit_changed_types,
      fit_changed_measures = fit_changed_measures,
      fit_delta = fit_delta,
      n_fit_deltas = n_fit_deltas,
      max_abs_fit_delta = max_abs_fit_delta,
      fit_compared_by_solution_type = fit_compared_by_solution_type,
      fit_changed_by_solution_type = setNames(
        lapply(
          monitored_solutions,
          function(solution_type) {
            if (!status_ok || !(solution_type %in% comparable_solution_types) || !isTRUE(fit_compared_by_solution_type[[solution_type]])) return(NA)
            length(fit_changed_names_by_solution_type[[solution_type]]) > 0L
          }
        ),
        monitored_solutions
      ),
      max_abs_fit_delta_by_solution_type = fit_delta_by_solution_type,
      matched_fit_keys_by_solution_type = fit_keys_by_solution_type,
      fit_details = fit_details
    )

    progress_state$set(b)
    if (isTRUE(verbose)) message("Completed draw ", b, "/", n_draws)
  }

  diagnostics <- .bind_rows_result(diag_rows)

  # summaries --
  solution_same_vec <- !diagnostics$solution_changed
  solution_same_vec[is.na(solution_same_vec)] <- FALSE

  score_solution <- mean(solution_same_vec)
  n_same_solution <- sum(solution_same_vec, na.rm = TRUE)

  n_fit_compared <- sum(diagnostics$fit_compared, na.rm = TRUE)
  n_fit_den <- sum(solution_same_vec, na.rm = TRUE)
  if (n_fit_den > 0L) {
    fit_same_equal <- !diagnostics$fit_changed[solution_same_vec]
    fit_same_equal[is.na(fit_same_equal)] <- FALSE
    n_same_fit <- sum(fit_same_equal)
    score_fit <- n_same_fit / n_fit_den
  } else {
    n_same_fit <- 0L
    score_fit <- NA_real_
  }

  score_total <- if (is.na(score_fit)) 0 else score_solution * score_fit

  score_solution_by_solution_type <- setNames(
    vapply(
      monitored_solutions,
      function(solution_type) {
        changed_vec <- diagnostics[[paste0("solution_changed_", solution_type)]]
        if (all(is.na(changed_vec))) return(NA_real_)
        mean(!changed_vec[!is.na(changed_vec)])
      },
      numeric(1)
    ),
    monitored_solutions
  )

  score_fit_by_solution_type <- setNames(
    vapply(
      monitored_solutions,
      function(solution_type) {
        idx <- diagnostics[[paste0("fit_compared_", solution_type)]]
        if (!any(idx, na.rm = TRUE)) return(NA_real_)
        vec <- !diagnostics[[paste0("fit_changed_", solution_type)]][idx]
        vec[is.na(vec)] <- FALSE
        mean(vec)
      },
      numeric(1)
    ),
    monitored_solutions
  )

  .new_result_object(
    "altset_test",
    diagnostics = diagnostics,
    results = .make_altset_results(diagnostics),
    summary = list(
      n_draws = n_draws,
      n_same_solution = n_same_solution,
      n_fit_compared = n_fit_compared,
      n_same_fit = n_same_fit,
      score_solution = score_solution,
      score_fit = score_fit,
      score_total = score_total,
      score_solution_by_solution_type = score_solution_by_solution_type,
      score_fit_by_solution_type = score_fit_by_solution_type
    ),
    baseline = list(
      result = baseline,
      draw_meta = baseline_draw_meta,
      status = baseline$status,
      exclude_used = baseline$exclude_used,
      sig = baseline$sig,
      fit = baseline$fit,
      selected_solution_missing = baseline$selected_solution_missing,
      meta = baseline$meta,
      solution_type_status = baseline$solution_type_status,
      solution_type_error_source = baseline$solution_type_error_source,
      solution_type_error_message = baseline$solution_type_error_message,
      comparable_solutions = baseline_solution_types
    ),
    by_draw = by_draw,
    settings = list(
      outcome = outcome,
      conditions = conditions,
      test.conditions = test.conditions,
      test.outcome = test.outcome,
      calib_spec = calib_specs,
      anchors_to_test = anchors_to_test,
      solution = solution,
      monitored_solutions = monitored_solutions,
      include = include,
      which_M = which_M,
      unit_step = unit_step,
      unit_step_divisor = unit_step_divisor,
      calib_max_steps = calib_max_steps,
      incl.cut = incl.cut,
      incl_step = incl_step,
      incl_max_steps = incl_max_steps,
      n.cut = n.cut,
      ncut_step = ncut_step,
      ncut_max_steps = ncut_max_steps,
      dir.exp = dir.exp,
      i_mode = i_mode,
      exclude_mode = exclude_mode,
      exclude_recompute = exclude_recompute,
      exclude_static = exclude_static,
      n_draws = n_draws,
      fit_tol = fit_tol,
      seed = seed
    )
  )
}

#' Print an `altset_test` object
#'
#' @rdname altset.test
#' @export
print.altset_test <- function(x, row.names = FALSE, ...) {
  results <- x$results
  settings <- x$settings
  summary <- x$summary

  .print_qcaert_heading("altset_test", "alternative-set robustness", settings)
  if (!is.null(settings$n_draws)) {
    cat("Draws: ", settings$n_draws, "\n", sep = "")
  }
  if (!is.null(settings$incl.cut)) {
    cat("Baseline incl.cut: ", format(settings$incl.cut, trim = TRUE), "\n", sep = "")
  }
  if (!is.null(settings$n.cut)) {
    cat("Baseline n.cut: ", format(settings$n.cut, trim = TRUE), "\n", sep = "")
  }
  if (!is.null(settings$fit_tol)) {
    cat("Fit tolerance: ", format(settings$fit_tol, trim = TRUE), "\n", sep = "")
  }
  if (!is.null(settings$test.conditions) && length(settings$test.conditions) > 0L) {
    cat("Test conditions: ", paste(settings$test.conditions, collapse = ", "), "\n", sep = "")
  }
  if (isTRUE(settings$test.outcome)) {
    cat("Test outcome: ", settings$outcome, "\n", sep = "")
  }

  .print_qcaert_table(results, "Result", row.names = row.names, ...)

  cat("\nSummary\n")
  if (!is.null(summary$n_draws)) {
    cat(" Draws: ", summary$n_draws, "\n", sep = "")
  }
  if (is.data.frame(results) && nrow(results) > 0L) {
    n_solution_changed <- .count_present(results$solution_change)
    n_fit_changed <- .count_present(results$fit_changed_types)

    cat(" Draws with any solution change: ", n_solution_changed, "\n", sep = "")
    cat(" Draws with any fit change: ", n_fit_changed, "\n", sep = "")
  }
  if (!is.null(summary$n_fit_compared)) {
    cat(" Draws with fit comparison: ", summary$n_fit_compared, "\n", sep = "")
  }

  if (!is.null(summary$score_solution_by_solution_type) && length(summary$score_solution_by_solution_type) > 0L) {
    solution_table <- data.frame(
      solution_type = vapply(names(summary$score_solution_by_solution_type), .compact_solution_type_names, character(1)),
      match_rate = as.numeric(summary$score_solution_by_solution_type),
      stringsAsFactors = FALSE
    )
    .print_qcaert_table(solution_table, "Solution match rate by solution type", row.names = FALSE)
  }

  if (!is.null(summary$score_fit_by_solution_type) && length(summary$score_fit_by_solution_type) > 0L) {
    fit_table <- data.frame(
      solution_type = vapply(names(summary$score_fit_by_solution_type), .compact_solution_type_names, character(1)),
      match_rate = as.numeric(summary$score_fit_by_solution_type),
      stringsAsFactors = FALSE
    )
    if (any(!is.na(fit_table$match_rate))) {
      .print_qcaert_table(fit_table, "Fit match rate by solution type", row.names = FALSE)
    }
  }

  invisible(x)
}

#' Return the main results table from an `altset_test` object
#'
#' @rdname altset.test
#' @export
as.data.frame.altset_test <- function(x, ...) {
  .as.data.frame_results(x, ...)
}

test_that("altset.test can perturb only the outcome calibration", {
  skip_if_not_installed("QCA")

  fixture <- qcaert_fixture_outcome_calibration()
  out <- suppressWarnings(altset.test(
    raw.data = fixture$raw,
    calib.data = fixture$calib,
    outcome = fixture$outcome,
    conditions = fixture$conditions,
    calib_spec = fixture$calib_spec,
    test.conditions = NULL,
    test.outcome = TRUE,
    anchors_to_test = c("E1", "C1", "I1"),
    unit_step = 1,
    calib_max_steps = 1,
    incl.cut = 0.75,
    incl_step = 0.1,
    incl_max_steps = 1,
    n.cut = 1,
    ncut_step = 1,
    ncut_max_steps = 1,
    n_draws = 2,
    seed = 101,
    solution = "conservative",
    progress = FALSE
  ))

  expect_qcaert_result_structure(
    out,
    class = "altset_test",
    diagnostics = c("draw", "status", "changed_sets", "changed_roles", "n_changed_sets"),
    results = c("draw", "status", "changed_sets", "changed_roles", "n_changed_sets"),
    settings = c("outcome", "conditions", "test.conditions", "test.outcome", "calib_spec"),
    top = c("summary", "baseline", "by_draw", "settings")
  )

  expect_identical(out$settings$test.conditions, character(0))
  expect_true(out$settings$test.outcome)

  draw_sets <- lapply(out$diagnostics$calibration, names)
  expect_true(all(vapply(draw_sets, identical, logical(1), fixture$outcome)))

  draw_roles <- unlist(lapply(out$diagnostics$calibration, function(x) {
    vapply(x, function(y) y$role, character(1))
  }), use.names = FALSE)
  expect_identical(unique(draw_roles), "outcome")
})

test_that("altset.test can perturb selected conditions and the outcome together", {
  skip_if_not_installed("QCA")

  fixture <- qcaert_fixture_outcome_calibration()
  out <- suppressWarnings(altset.test(
    raw.data = fixture$raw,
    calib.data = fixture$calib,
    outcome = fixture$outcome,
    conditions = fixture$conditions,
    calib_spec = fixture$calib_spec,
    test.conditions = "A",
    test.outcome = TRUE,
    anchors_to_test = c("E1", "C1", "I1"),
    unit_step = 1,
    calib_max_steps = 1,
    incl.cut = 0.75,
    incl_step = 0.1,
    incl_max_steps = 1,
    n.cut = 1,
    ncut_step = 1,
    ncut_max_steps = 1,
    n_draws = 2,
    seed = 202,
    solution = "conservative",
    progress = FALSE
  ))

  draw_sets <- lapply(out$diagnostics$calibration, names)
  expect_true(all(vapply(draw_sets, identical, logical(1), c("A", fixture$outcome))))

  draw_roles <- lapply(out$diagnostics$calibration, function(x) {
    vapply(x, function(y) y$role, character(1))
  })
  expect_true(all(vapply(draw_roles, identical, logical(1), c(A = "condition", Y = "outcome"))))
})

test_that("altset.test requires outcome calibration specs for outcome testing", {
  skip_if_not_installed("QCA")

  fixture <- qcaert_fixture_direct6()

  expect_error(
    altset.test(
      raw.data = fixture$raw,
      calib.data = fixture$calib,
      outcome = fixture$outcome,
      conditions = fixture$conditions,
      calib_spec = fixture$calib_spec,
      test.conditions = NULL,
      test.outcome = TRUE,
      unit_step = 1,
      calib_max_steps = 1,
      incl.cut = 0.75,
      incl_step = 0.1,
      incl_max_steps = 1,
      n.cut = 1,
      ncut_step = 1,
      ncut_max_steps = 1,
      n_draws = 2,
      seed = 303,
      solution = "conservative",
      progress = FALSE
    ),
    "`calib_spec` must contain exactly one entry for each condition and the outcome when `test.outcome = TRUE`.",
    fixed = TRUE
  )
})

test_that("altset.test keeps conservative draws comparable when all-solution exclude recomputation fails", {
  skip_if_not_installed("QCA")

  fixture <- qcaert_fixture_direct6()

  out <- suppressWarnings(altset.test(
    raw.data = fixture$raw,
    calib.data = fixture$calib,
    outcome = fixture$outcome,
    conditions = fixture$conditions,
    calib_spec = fixture$calib_spec,
    test.conditions = fixture$test.conditions,
    anchors_to_test = c("E1", "C1", "I1"),
    unit_step = 1,
    calib_max_steps = 1,
    incl.cut = 0.75,
    incl_step = 0.1,
    incl_max_steps = 1,
    n.cut = 1,
    ncut_step = 1,
    ncut_max_steps = 1,
    n_draws = 2,
    seed = 101,
    solution = "all",
    dir.exp = fixture$dir.exp,
    exclude_recompute = list(type = 2, definitely_not_a_findRows_arg = TRUE),
    progress = FALSE
  ))

  expect_identical(out$baseline$solution_type_status[["conservative"]], "ok")
  expect_identical(out$baseline$solution_type_status[["parsimonious"]], "exclude_error")
  expect_identical(out$baseline$solution_type_status[["intermediate"]], "exclude_error")
  expect_identical(out$baseline$comparable_solutions, "conservative")

  expect_true(all(out$results$status == "ok"))
  expect_true(all(out$diagnostics$status_conservative == "ok"))
  expect_true(all(out$diagnostics$status_parsimonious == "exclude_error"))
  expect_true(all(out$diagnostics$status_intermediate == "exclude_error"))
  expect_true(all(is.na(out$diagnostics$solution_changed_parsimonious)))
  expect_true(all(is.na(out$diagnostics$solution_changed_intermediate)))
  expect_true(is.na(out$summary$score_solution_by_solution_type[["parsimonious"]]))
  expect_true(is.na(out$summary$score_solution_by_solution_type[["intermediate"]]))
})

.qca_solution_objects <- function(conservative, intermediate, parsimonious) {
  out <- list()
  if (!is.null(conservative)) out$conservative <- conservative
  if (!is.null(parsimonious)) out$parsimonious <- parsimonious
  if (!is.null(intermediate)) out$intermediate <- intermediate
  out
}

.qca_requested_solutions <- function(solution, objects) {
  if (solution == "all") {
    return(names(objects))
  }

  if (!solution %in% names(objects)) {
    stop("Requested `solution = \"", solution, "\"` but no `", solution, "` object was supplied.")
  }

  solution
}

.qca_solution_label <- function(x) {
  switch(
    x,
    conservative = "Conservative",
    parsimonious = "Parsimonious",
    intermediate = "Intermediate",
    x
  )
}

.qca_select_model_position <- function(n_models, which_M, context) {
  if (!is.numeric(n_models) || length(n_models) != 1L || is.na(n_models) || n_models < 1L) {
    stop("No models are available for ", context, ".")
  }

  if (which_M > n_models) {
    stop("`which_M = ", which_M, "` exceeds the number of available models for ", context, " (", n_models, ").")
  }

  as.integer(which_M)
}

.qca_select_ic_model <- function(ic, which_M, context) {
  if (!is.null(ic[["individual"]])) {
    m <- .qca_select_model_position(
      n_models = length(ic[["individual"]]),
      which_M = which_M,
      context = context
    )

    return(list(ic = ic[["individual"]][[m]], model = m))
  }

  list(ic = ic, model = NA_integer_)
}

.qca_canonicalize_names <- function(x) {
  out <- trimws(as.character(x))
  gsub("\\s+", "", out)
}

.qca_pims_by_terms <- function(pims, terms) {
  if (is.null(pims) || !is.data.frame(pims)) {
    return(NULL)
  }

  cn <- colnames(pims)
  if (is.null(cn) || length(cn) == 0L) {
    return(NULL)
  }

  canon_cn <- .qca_canonicalize_names(cn)
  canon_terms <- .qca_canonicalize_names(terms)
  idx <- match(canon_terms, canon_cn)

  if (any(is.na(idx))) {
    return(NULL)
  }

  out <- pims[, idx, drop = FALSE]
  colnames(out) <- terms
  out
}

.qca_cases_from_incl_cov <- function(incl_cov, pims, include_cases = TRUE, case_cut = 0.5) {
  if (!isTRUE(include_cases)) {
    return(rep(NA_character_, nrow(incl_cov)))
  }

  if ("cases" %in% colnames(incl_cov)) {
    cases_col <- incl_cov[["cases"]]

    if (is.list(cases_col)) {
      return(vapply(
        cases_col,
        function(x) {
          if (length(x) == 0L) return(NA_character_)
          paste(as.character(x), collapse = ", ")
        },
        character(1)
      ))
    }

    return(as.character(cases_col))
  }

  if (is.null(pims) || !is.data.frame(pims)) {
    return(rep(NA_character_, nrow(incl_cov)))
  }

  rn <- rownames(incl_cov)
  if (is.null(rn) || length(rn) == 0L) {
    return(rep(NA_character_, nrow(incl_cov)))
  }

  keep <- intersect(rn, colnames(pims))
  if (length(keep) == 0L) {
    return(rep(NA_character_, nrow(incl_cov)))
  }

  case_names <- rownames(pims)

  vapply(
    rn,
    function(pi) {
      if (!pi %in% colnames(pims)) return(NA_character_)
      x <- pims[[pi]]
      cases <- case_names[which(x >= case_cut)]
      if (length(cases) == 0L) NA_character_ else paste(cases, collapse = ", ")
    },
    character(1)
  )
}

.qca_selected_solution_missing <- function(
    solution_type,
    target_key,
    signature,
    status,
    error_source,
    error_message,
    meta = NULL
) {
  list(
    solution_type = solution_type,
    target_key = target_key,
    signature = signature,
    terms = character(0),
    solution_membership = NULL,
    term_memberships = NULL,
    status = status,
    error_source = error_source,
    error_message = error_message,
    meta = meta
  )
}

.qca_extract_standard_targets <- function(res, solution_type, which_M) {
  sol_all <- res$solution
  primes <- res$primes
  pims <- res$pims
  nM <- if (!is.null(sol_all)) length(sol_all) else 0L

  if (nM < which_M) {
    sig <- "<M_MISSING>"
    return(list(.qca_selected_solution_missing(
      solution_type = solution_type,
      target_key = paste0(solution_type, ":", sig),
      signature = sig,
      status = "baseline_selected_solution_missing",
      error_source = solution_type,
      error_message = "The selected baseline solution is missing.",
      meta = list(nM = nM)
    )))
  }

  terms <- .solution_terms_from_primes(sol_all[[which_M]], primes)
  sig <- paste(terms, collapse = "+")
  tm <- .qca_pims_by_terms(pims, terms)

  if (is.null(tm)) {
    return(list(list(
      solution_type = solution_type,
      target_key = paste0(solution_type, ":", sig),
      signature = sig,
      terms = terms,
      solution_membership = NULL,
      term_memberships = NULL,
      status = "input_error",
      error_source = "pims",
      error_message = "Could not align term memberships with the selected baseline solution.",
      meta = list(nM = nM)
    )))
  }

  sm <- if (ncol(tm) == 1L) as.numeric(tm[[1L]]) else apply(tm, 1L, max)
  list(list(
    solution_type = solution_type,
    target_key = paste0(solution_type, ":", sig),
    signature = sig,
    terms = terms,
    solution_membership = sm,
    term_memberships = tm,
    status = "ok",
    error_source = NA_character_,
    error_message = NA_character_,
    meta = list(nM = nM)
  ))
}

.qca_extract_intermediate_targets <- function(res, solution_type, which_M, i_mode) {
  i.sol <- res$i.sol
  out <- list()

  if (is.null(i.sol) || length(i.sol) == 0L) {
    return(list(.qca_selected_solution_missing(
      solution_type = solution_type,
      target_key = "<I_SOL_MISSING>",
      signature = "<I_SOL_MISSING>",
      status = "baseline_selected_solution_missing",
      error_source = solution_type,
      error_message = "No intermediate baseline solutions are available.",
      meta = list(n_i = 0L)
    )))
  }

  nm <- .normalize_i_solution_names(i.sol)

  if (i_mode == "C1P1") {
    idx <- which(nm == "C1P1")
    if (length(idx) == 0L) {
      return(list(.qca_selected_solution_missing(
        solution_type = solution_type,
        target_key = "<C1P1_MISSING>",
        signature = "<C1P1_MISSING>",
        status = "baseline_selected_solution_missing",
        error_source = solution_type,
        error_message = "The requested intermediate family `C1P1` is missing.",
        meta = list(n_i = length(i.sol))
      )))
    }

    i.sol <- i.sol[idx[1L]]
    nm <- nm[idx[1L]]
  }

  for (k in seq_along(i.sol)) {
    el <- i.sol[[k]]
    sol_all <- if (is.list(el) && "solution" %in% names(el)) el$solution else NULL
    primes <- if (is.list(el) && "primes" %in% names(el)) el$primes else NULL
    pims <- if (is.list(el) && "pims" %in% names(el)) el$pims else NULL
    nM <- if (!is.null(sol_all)) length(sol_all) else 0L

    if (nM < which_M) {
      out[[length(out) + 1L]] <- .qca_selected_solution_missing(
        solution_type = solution_type,
        target_key = paste0(nm[k], ":<M_MISSING>"),
        signature = paste0(nm[k], ":<M_MISSING>"),
        status = "baseline_selected_solution_missing",
        error_source = solution_type,
        error_message = "The selected intermediate baseline solution is missing.",
        meta = list(n_i = length(i.sol), nM = nM)
      )
      next
    }

    terms <- .solution_terms_from_primes(sol_all[[which_M]], primes)
    sig <- paste0(nm[k], ":", paste(terms, collapse = "+"))
    tm <- .qca_pims_by_terms(pims, terms)

    if (is.null(tm)) {
      out[[length(out) + 1L]] <- list(
        solution_type = solution_type,
        target_key = sig,
        signature = sig,
        terms = terms,
        solution_membership = NULL,
        term_memberships = NULL,
        status = "input_error",
        error_source = "pims",
        error_message = "Could not align intermediate term memberships with the selected baseline solution.",
        meta = list(n_i = length(i.sol), nM = nM)
      )
      next
    }

    sm <- if (ncol(tm) == 1L) as.numeric(tm[[1L]]) else apply(tm, 1L, max)
    out[[length(out) + 1L]] <- list(
      solution_type = solution_type,
      target_key = sig,
      signature = sig,
      terms = terms,
      solution_membership = sm,
      term_memberships = tm,
      status = "ok",
      error_source = NA_character_,
      error_message = NA_character_,
      meta = list(n_i = length(i.sol), nM = nM)
    )
  }

  out
}

.qca_extract_targets_for_solution_type <- function(res, solution_type, which_M, i_mode) {
  if (is.null(res)) {
    return(list(.qca_selected_solution_missing(
      solution_type = solution_type,
      target_key = paste0("<", toupper(substr(solution_type, 1L, 3L)), "_MISSING>"),
      signature = paste0("<", toupper(substr(solution_type, 1L, 3L)), "_MISSING>"),
      status = "baseline_selected_solution_missing",
      error_source = solution_type,
      error_message = "This solution_type is not available.",
      meta = NULL
    )))
  }

  if (solution_type %in% c("conservative", "parsimonious")) {
    return(.qca_extract_standard_targets(
      res = res,
      solution_type = solution_type,
      which_M = which_M
    ))
  }

  .qca_extract_intermediate_targets(
    res = res,
    solution_type = solution_type,
    which_M = which_M,
    i_mode = i_mode
  )
}

test_that("QCA solution helpers follow data-frame pims", {
  pims <- data.frame(
    "A * B" = c(1, 0.4, 0.8),
    check.names = FALSE,
    row.names = c("case1", "case2", "case3")
  )

  out <- .qca_pims_by_terms(pims, "A*B")
  expect_identical(names(out), "A*B")
  expect_equal(out[["A*B"]], c(1, 0.4, 0.8))

  expect_null(.qca_pims_by_terms(as.matrix(pims), "A*B"))
})

test_that("sol.df case reconstruction uses QCA-style pims data frames", {
  incl_cov <- data.frame(
    inclS = 0.9,
    PRI = 0.8,
    covS = 0.7,
    covU = 0.6,
    row.names = "A*B"
  )
  pims <- data.frame(
    "A*B" = c(1, 0.4, 0.8),
    check.names = FALSE,
    row.names = c("case1", "case2", "case3")
  )

  expect_identical(
    unname(.qca_cases_from_incl_cov(incl_cov, pims, include_cases = TRUE)),
    "case1, case3"
  )
})

test_that("QCA minimize provides data-frame pims for solution consumers", {
  skip_if_not_installed("QCA")

  dat <- qcaert_schema_calib3()
  tt <- suppressWarnings(QCA::truthTable(
    dat,
    outcome = "Y",
    conditions = c("A", "B"),
    incl.cut = 0.75,
    n.cut = 1
  ))
  sol <- suppressWarnings(QCA::minimize(tt, include = ""))

  expect_true(is.data.frame(sol$pims))

  sol_table <- sol.df(conservative = sol, solution = "conservative")
  expect_true("Cases" %in% names(sol_table))
  expect_true(nrow(sol_table) >= 1L)
})

test_that("cluster.test does not duplicate one-term solutions as term components", {
  skip_if_not_installed("QCA")

  fixture <- qcaert_fixture_cluster()
  out <- suppressWarnings(cluster.test(
    data = fixture$data,
    tt = fixture$truth_table,
    cluster_id = fixture$cluster_id,
    unit_id = fixture$unit_id,
    solution = "all",
    dir.exp = fixture$dir.exp,
    progress = FALSE
  ))

  single_configurations <- out$results$overview$configuration[out$results$overview$components == 1L]
  clusters <- out$results$clusters[out$results$clusters$configuration %in% single_configurations, , drop = FALSE]

  expect_true(nrow(clusters) > 0L)
  expect_identical(unique(clusters$component), "solution")
  expect_false(any(clusters$component == clusters$solution_expression))

  if (!is.null(out$results$units)) {
    units <- out$results$units[out$results$units$configuration %in% single_configurations, , drop = FALSE]
    expect_identical(unique(units$component), "solution")
    expect_false(any(units$component == units$solution_expression))
  }
})

#' qcaERT: Enhanced robustness tests for QCA
#'
#' qcaERT extends the usual Qualitative Comparative Analysis (QCA) with tools
#' for assessing how stable and robust a QCA analysis is. It can test alternative
#' calibration thresholds, inclusion cutoffs, truth table frequency cutoffs,
#' case deletion, subsampling, alternative solutions, cluster-specific
#' analyses and different theoretical configurations. It also offers a compact way to organize and visualize QCA
#' solutions.
#' The package is built around the QCA package (Dusa 2019) workflow using
#' [QCA::calibrate()], [QCA::truthTable()], [QCA::minimize()], and
#' [QCA::findRows()]. It is not intended to cover every possible QCA variant
#' or calibration transformation.
#'
#' @section Start Here:
#' If you know the robustness concern but not the function name, start with
#' `?qcaERT_tests`, which maps qcaERT's tools.
#'
#' A typical R workflow for sufficient analysis using qcaERT is:
#' \enumerate{
#'   \item calibrate data with [QCA::calibrate()],
#'   \item build a truth table with [QCA::truthTable()],
#'   \item minimize with [QCA::minimize()],
#'   \item run one or more qcaERT robustness tests, and
#'   \item inspect `print(x)`, `as.data.frame(x)`, `x$diagnostics`, and, where
#'   available, `plot(x)`.
#' }
#'
#' @section Function Family:
#' The main qcaERT functions are:
#' \itemize{
#'   \item [calib.test()] for calibration-threshold robustness.
#'   \item [incl.test()] for truth table inclusion-cutoff robustness.
#'   \item [ncut.test()] for truth table frequency-cutoff robustness.
#'   \item [loo.test()] for leave-one-out case influence.
#'   \item [subsample.test()] for repeated subsample stability (advanced).
#'   \item [altset.test()] for sampled alternative analysis settings that can
#'   combine calibration, inclusion-cutoff, and frequency-cutoff perturbations.
#'   \item [theory.test()] for comparing theoretically motivated configurations
#'   using the same outcome, truth-table cutoffs, solution type,
#'   exclusion handling, and model-selection settings.
#'   \item [cluster.test()] for cluster, group, or repeated-unit heterogeneity.
#'   \item [sol.df()] for converting QCA minimization objects into compact
#'   solution tables.
#'   \item [sol.chart()] for rendering `sol.df()` tables as solution charts.
#' }
#'
#' @section Returned Objects:
#' Most robustness functions return R objects with a common structure:
#' \itemize{
#'   \item `diagnostics`: the detailed table, useful for inspection and
#'   troubleshooting.
#'   \item `results`: a cleaner table.
#'   \item `settings`: the analysis settings used to create the result.
#'   \item supporting components such as `baseline`, `bounds`, `by_direction`,
#'   `by_case`, `by_run`, `by_draw`, or `summary`, depending on the test.
#' }
#'
#' `print()` gives a concise summary. `as.data.frame()` returns the main clean
#' table. `cluster.test()` and `theory.test()` use structured `results` lists:
#' `cluster.test()` returns `overview`, `clusters`, and `units`, while
#' `theory.test()` returns `models`, `pairwise`, and `solutions`.
#'
#' @section Common Conventions:
#' The family-wide argument and output conventions are described in
#' `?qcaERT_conventions`. It explains solution controls such as `solution`,
#' `include`, `dir.exp`, `which_M`, and `i_mode`; exclusion handling and its
#' function-specific exceptions; calibration specifications; and the common
#' returned-object structure.
#'
#' @section Plotting:
#' Plotting requires `ggplot2`. Current plot methods are
#' available for [calib.test()], [incl.test()], and [theory.test()] results.
#' [sol.chart()] provides a visual presentation for [sol.df()] tables. See
#' `?qcaERT_plots`.
#'
#' @section Learn More:
#' \itemize{
#'   \item `?qcaERT_tests` for choosing the right robustness test.
#'   \item `?qcaERT_conventions` for common argument and output conventions.
#'   \item `?qcaERT_plots` for plotting `calib_test`, `incl_test`, and
#'   `theory_test` objects, and for charting `sol.df()` tables.
#'   \item `vignette("qcaERT-overview", package = "qcaERT")` for a guided
#'   introduction.
#'   \item `vignette("qcaERT-result-objects", package = "qcaERT")` for the
#'   common returned-object structure.
#'   \item `vignette("qcaERT-calibration", package = "qcaERT")` for
#'   calibration specifications, scale-aware perturbations, and alternative
#'   sets.
#'   \item `news(package = "qcaERT")` for development changes.
#'   \item `citation(package = "qcaERT")` for citation information.
#' }
#'
#' @section Recommended citation:
#' Marisguia, B. A. H. (2026). qcaERT: Enhanced Robustness Tests for
#' Qualitative Comparative Analysis. R package version 0.1.0.
#'
#' Please also cite the QCA package and the methodological sources relevant
#' to the analysis.
#'
#' @references
#' Dusa, Adrian. 2019. *QCA with R. A Comprehensive Resource*. Cham:
#' Springer. <https://adriandusa.com/research/books/2019-QCA/>.
#'
#' Ragin, C. C. 2014. *The Comparative Method: Moving Beyond Qualitative and
#' Quantitative Strategies*. Oakland, California: University of California
#' Press.
#'
#' @author Breno A. H. Marisguia

#' @aliases qcaERT-package
#' @importFrom stats setNames xtabs
"_PACKAGE"

if (getRversion() >= "2.15.1") {
  utils::globalVariables(c("component"))
}

.bind_rows_result <- function(rows) {
  out <- do.call(rbind, rows)
  rownames(out) <- NULL
  out
}

.bind_rows_result_sorted <- function(rows, order_by) {
  out <- do.call(rbind, rows)
  out <- out[order(out[[order_by]]), , drop = FALSE]
  rownames(out) <- NULL
  out
}

.new_result_object <- function(class, ...) {
  structure(
    list(...),
    class = class
  )
}

.print_diagnostics_result <- function(x, row.names = FALSE, ...) {
  print.data.frame(x$diagnostics, row.names = row.names, ...)
  invisible(x)
}

.as.data.frame_diagnostics_result <- function(x, ...) {
  x$diagnostics
}

.as.data.frame_results <- function(x, ...) {
  x$results
}

.as.data.frame_cluster_overview <- function(x, ...) {
  x$results$overview
}

.qcaert_progress_messages <- function() {
  c(
    "This might take a while... maybe go grab a coffee?",
    "This might take a while... perfect moment to stretch.",
    "This might take a while... please resist the urge to stare directly at the progress bar.",
    "This might take a while... the Boolean gods are being consulted.",
    "This might take a while... multi-value support is currently a counterfactual.",
    "This might take a while... the minimization elves are hard at work.",
    "This might take a while... the truth is out there... in a table, apparently.",
    "This might take a while... maybe hydrate before the next robustness check.",
    "This might take a while... Santa only brings presents to good counterfactuals.",
    "This might take a while... causal complexity refuses to be rushed.",
    "This might take a while... please enjoy this brief intermission of set-theoretic suspense.",
    "This might take a while... the remainders are a bit illogical today.",
    "This might take a while... another day, another battle with limited diversity.",
    "This might take a while... your computer is discovering causal complexity the hard way.",
    "This might take a while... limited diversity has entered the chat.",
    "This might take a while... 'less is more', said the parsimonious solution.",
    "This might take a while... one of the remainders is having an identity crisis.",
    "This might take a while... the truth table has some truths to work through.",
    "This might take a while... 'am I not sufficient for you?' - said the necessary condition.",
    "This might take a while... proportional reduction in impatience is not guaranteed.",
    "This might take a while... inclusion and coverage are in a complicated relationship.",
    "This might take a while... some contradictory configurations need to calm down first.",
    "This might take a while... the process of logical minimization doesn't minimize anxiety.",
    "This might take a while... equifinality, asymmetry and conjunctural causation walk into a bar...",
    "This might take a while... insufficient but necessary conditions are unionizing.",
    "This might take a while... the conditions are becoming jointly sufficient, one insecurity at a time."
  )
}

.new_qcaert_progress <- function(total, progress = TRUE) {
  pb <- NULL
  done <- 0L
  total <- as.integer(total)
  if (length(total) != 1L || is.na(total)) {
    total <- 0L
  }

  if (isTRUE(progress) && interactive() && is.finite(total) && total > 0L) {
    cat(sample(.qcaert_progress_messages(), size = 1L), "\n")
    pb <- utils::txtProgressBar(min = 0, max = total, style = 3)
  }

  list(
    tick = function(step = 1L) {
      if (!is.null(pb)) {
        done <<- min(total, done + as.integer(step))
        utils::setTxtProgressBar(pb, done)
      }
      invisible(done)
    },
    set = function(value) {
      if (!is.null(pb)) {
        done <<- max(0L, min(total, as.integer(value)))
        utils::setTxtProgressBar(pb, done)
      }
      invisible(done)
    },
    close = function() {
      if (!is.null(pb)) {
        try(close(pb), silent = TRUE)
        pb <<- NULL
      }
      invisible(NULL)
    }
  )
}

.print_qcaert_solution_settings <- function(settings) {
  if (!is.null(settings$solution)) {
    cat("Solution: ", settings$solution, "\n", sep = "")
  }
  if (!is.null(settings$monitored_solutions)) {
    cat("Monitored solutions: ", paste(settings$monitored_solutions, collapse = ", "), "\n", sep = "")
  }
  if (!is.null(settings$which_M)) {
    cat("which_M: ", settings$which_M, "\n", sep = "")
  }
  if (!is.null(settings$i_mode) &&
      !is.null(settings$monitored_solutions) &&
      any(settings$monitored_solutions %in% "intermediate")) {
    cat("i_mode: ", settings$i_mode, "\n", sep = "")
  }
}

.print_qcaert_heading <- function(class, label, settings) {
  cat("<", class, ">\n", sep = "")
  cat("Outcome test: ", label, "\n", sep = "")
  .print_qcaert_solution_settings(settings)
}

.print_qcaert_table <- function(x, title, row.names = FALSE, ...) {
  cat("\n", title, "\n", sep = "")
  print(x, row.names = row.names, ...)
}

.title_case_word <- function(x) {
  paste0(toupper(substr(x, 1L, 1L)), substring(x, 2L))
}

.print_boundary_summary <- function(x) {
  cat("\nSummary\n")
  if (!is.null(x$bounds) && is.matrix(x$bounds)) {
    intervals <- data.frame(
      solution_type = vapply(colnames(x$bounds), .compact_solution_type_names, character(1)),
      lower = unname(x$bounds["Lower", ]),
      upper = unname(x$bounds["Upper", ]),
      width = unname(x$bounds["Upper", ] - x$bounds["Lower", ]),
      stringsAsFactors = FALSE
    )
    .print_qcaert_table(intervals, "Stable intervals", row.names = FALSE)
  } else {
    lower <- if (!is.null(x$bounds)) unname(x$bounds[["Lower"]]) else NA_real_
    upper <- if (!is.null(x$bounds)) unname(x$bounds[["Upper"]]) else NA_real_

    if (!is.na(lower) && !is.na(upper)) {
      cat(" Stable interval: [", format(lower, trim = TRUE), ", ", format(upper, trim = TRUE), "]\n", sep = "")
      cat(" Total width: ", format(upper - lower, trim = TRUE), "\n", sep = "")
    }
  }
}

.count_present <- function(x) {
  if (is.null(x)) {
    return(0L)
  }
  sum(!is.na(x), na.rm = TRUE)
}

.require_result_columns <- function(diagnostics, needed, label) {
  if (!is.data.frame(diagnostics)) {
    stop("Cannot build ", label, " `results`; `diagnostics` must be a data frame.")
  }

  missing_cols <- setdiff(needed, names(diagnostics))
  if (length(missing_cols) > 0L) {
    stop(
      "Cannot build ", label, " `results`; missing columns: ",
      paste(missing_cols, collapse = ", ")
    )
  }

  invisible(TRUE)
}

.select_result_columns <- function(diagnostics, from, to = from, label) {
  .require_result_columns(diagnostics, from, label)

  out <- diagnostics[, from, drop = FALSE]
  names(out) <- to
  out
}

.finish_result_table <- function(out, columns = NULL) {
  if (!is.null(columns)) {
    out <- out[, columns, drop = FALSE]
  }

  rownames(out) <- NULL
  out
}

.result_reason_vector <- function(stop_reason, change_kind) {
  vapply(
    seq_along(stop_reason),
    function(i) .compact_result_reason(stop_reason[i], change_kind[i]),
    character(1)
  )
}

.result_solution_change_vector <- function(change_kind) {
  vapply(
    seq_along(change_kind),
    function(i) {
      if (is.na(change_kind[i])) {
        return(NA_character_)
      }
      .compact_change_kind(change_kind[i])
    },
    character(1)
  )
}

.result_fit_changed_types_vector <- function(fit_changed_types) {
  vapply(
    seq_along(fit_changed_types),
    function(i) {
      if (is.na(fit_changed_types[i])) {
        return(NA_character_)
      }
      .compact_solution_type_names(fit_changed_types[i])
    },
    character(1)
  )
}

.add_solution_and_fit_columns <- function(out) {
  out$solution_change <- .result_solution_change_vector(out$change_kind)
  out$fit_changed_types <- .result_fit_changed_types_vector(out$fit_changed_types)
  out
}

.result_solution_type_order <- function(solution_types) {
  solution_types <- as.character(solution_types)
  solution_types <- solution_types[!is.na(solution_types)]
  solution_types <- unique(solution_types[nzchar(solution_types)])
  canonical_order <- c("conservative", "parsimonious", "intermediate")
  c(canonical_order[canonical_order %in% solution_types], setdiff(solution_types, canonical_order))
}

.result_solution_type_prefix <- function(solution_type) {
  switch(
    solution_type,
    conservative = "con",
    parsimonious = "par",
    intermediate = "int",
    gsub("[^A-Za-z0-9]+", "_", tolower(solution_type))
  )
}

.boundary_bounds <- function(diagnostics, value_name) {
  value_col <- paste0(value_name, "_last_safe")

  if ("solution_type" %in% names(diagnostics)) {
    solution_types <- .result_solution_type_order(diagnostics$solution_type)
    out <- matrix(
      NA_real_,
      nrow = 2L,
      ncol = length(solution_types),
      dimnames = list(c("Lower", "Upper"), solution_types)
    )

    for (solution_type in solution_types) {
      for (direction in c("lower", "upper")) {
        dd <- diagnostics[
          diagnostics$solution_type == solution_type & diagnostics$direction == direction,
          ,
          drop = FALSE
        ]
        if (nrow(dd) > 0L) {
          out[.title_case_word(direction), solution_type] <- dd[[value_col]][1L]
        }
      }
    }

    return(out)
  }

  c(
    Lower = diagnostics[[value_col]][diagnostics$direction == "lower"][1],
    Upper = diagnostics[[value_col]][diagnostics$direction == "upper"][1]
  )
}

.compact_change_kind <- function(x) {
  if (is.null(x) || length(x) == 0L || all(is.na(x))) {
    return(NA_character_)
  }

  x <- as.character(x)
  x <- x[!is.na(x)]
  x <- trimws(x)
  x <- x[nzchar(x)]

  if (length(x) == 0L) {
    return(NA_character_)
  }

  x <- gsub("\\s+", "", x)
  x <- gsub("conservative:", "CON:", x, fixed = TRUE)
  x <- gsub("parsimonious:", "PAR:", x, fixed = TRUE)
  x <- gsub("intermediate:", "INT:", x, fixed = TRUE)
  x <- gsub(",", " | ", x, fixed = TRUE)

  paste(x, collapse = " | ")
}

.compact_result_reason <- function(stop_reason, change_kind) {
  ck <- .compact_change_kind(change_kind)

  if (!is.na(ck) && nzchar(ck)) {
    return(ck)
  }

  if (is.null(stop_reason) || length(stop_reason) == 0L || all(is.na(stop_reason))) {
    return(NA_character_)
  }

  sr <- as.character(stop_reason[[1L]])
  sr <- trimws(sr)

  if (!nzchar(sr)) {
    return(NA_character_)
  }

  sr
}

.make_boundary_results <- function(diagnostics, value_name, result_shape = c("wide", "long")) {
  result_shape <- match.arg(result_shape)

  if ("solution_type" %in% names(diagnostics)) {
    needed <- c(
      "direction",
      "solution_type",
      paste0(value_name, "_start"),
      paste0(value_name, "_last_safe"),
      paste0(value_name, "_first_failing"),
      "number_of_steps",
      "total_delta",
      "stop_reason",
      "change_kind"
    )
    .require_result_columns(diagnostics, needed, "boundary-test")

    if (identical(result_shape, "long")) {
      out <- .select_result_columns(
        diagnostics,
        from = needed,
        to = c(
          "direction",
          "solution_type",
          "start",
          "last_safe",
          "first_failing",
          "steps",
          "total_delta",
          "stop_reason",
          "change_kind"
        ),
        label = "boundary-test"
      )
      out$reason <- .result_reason_vector(out$stop_reason, out$change_kind)

      out$solution_type <- factor(as.character(out$solution_type), levels = .result_solution_type_order(out$solution_type))
      out$direction <- factor(as.character(out$direction), levels = c("lower", "upper"))
      out <- out[order(out$solution_type, out$direction), , drop = FALSE]
      out$solution_type <- as.character(out$solution_type)
      out$direction <- as.character(out$direction)

      return(.finish_result_table(out, c(
        "solution_type",
        "direction",
        "start",
        "last_safe",
        "first_failing",
        "steps",
        "total_delta",
        "reason"
      )))
    }

    directions <- unique(as.character(diagnostics$direction))
    solution_types <- .result_solution_type_order(diagnostics$solution_type)
    rows <- lapply(
      directions,
      function(direction) {
        dd_direction <- diagnostics[diagnostics$direction == direction, , drop = FALSE]
        row <- data.frame(
          direction = direction,
          start = dd_direction[[paste0(value_name, "_start")]][1L],
          stringsAsFactors = FALSE
        )

        for (solution_type in solution_types) {
          dd <- dd_direction[dd_direction$solution_type == solution_type, , drop = FALSE]
          prefix <- .result_solution_type_prefix(solution_type)
          if (nrow(dd) == 0L) {
            row[[paste0(prefix, "_last_safe")]] <- NA_real_
            row[[paste0(prefix, "_first_failing")]] <- NA_real_
            row[[paste0(prefix, "_steps")]] <- NA_integer_
            row[[paste0(prefix, "_total_delta")]] <- NA_real_
            row[[paste0(prefix, "_reason")]] <- NA_character_
          } else {
            reason <- .compact_result_reason(dd$stop_reason[1L], dd$change_kind[1L])
            row[[paste0(prefix, "_last_safe")]] <- dd[[paste0(value_name, "_last_safe")]][1L]
            row[[paste0(prefix, "_first_failing")]] <- dd[[paste0(value_name, "_first_failing")]][1L]
            row[[paste0(prefix, "_steps")]] <- dd$number_of_steps[1L]
            row[[paste0(prefix, "_total_delta")]] <- dd$total_delta[1L]
            row[[paste0(prefix, "_reason")]] <- reason
          }
        }

        row
      }
    )

    return(.finish_result_table(.bind_rows_result(rows)))
  }

  needed <- c(
    "direction",
    paste0(value_name, "_start"),
    paste0(value_name, "_last_safe"),
    paste0(value_name, "_first_failing"),
    "number_of_steps",
    "total_delta",
    "stop_reason",
    "change_kind"
  )

  out <- .select_result_columns(
    diagnostics,
    from = needed,
    to = c(
      "direction",
      "start",
      "last_safe",
      "first_failing",
      "steps",
      "total_delta",
      "stop_reason",
      "change_kind"
    ),
    label = "boundary-test"
  )

  out$reason <- .result_reason_vector(out$stop_reason, out$change_kind)

  .finish_result_table(out, c(
    "direction",
    "start",
    "last_safe",
    "first_failing",
    "steps",
    "total_delta",
    "reason"
  ))
}

.make_calib_results <- function(diagnostics, result_shape = c("wide", "long")) {
  result_shape <- match.arg(result_shape)

  if ("solution_type" %in% names(diagnostics)) {
    needed <- c(
      "set",
      "role",
      "raw",
      "type",
      "method",
      "anchor",
      "direction",
      "solution_type",
      "start_value",
      "last_safe_value",
      "failing_value",
      "step_unit_used",
      "number_of_steps",
      "total_delta_units",
      "delta_as_pct_of_raw_range",
      "stop_reason",
      "change_kind"
    )
    .require_result_columns(diagnostics, needed, "calibration-test")

    if (identical(result_shape, "long")) {
      out <- .select_result_columns(
        diagnostics,
        from = needed,
        to = c(
          "set",
          "role",
          "raw",
          "type",
          "method",
          "anchor",
          "direction",
          "solution_type",
          "start",
          "last_safe",
          "first_failing",
          "step_unit",
          "steps",
          "total_delta",
          "pct_raw_range",
          "stop_reason",
          "change_kind"
        ),
        label = "calibration-test"
      )
      out$reason <- .result_reason_vector(out$stop_reason, out$change_kind)
      out$.row_order <- seq_len(nrow(out))

      out$solution_type <- factor(as.character(out$solution_type), levels = .result_solution_type_order(out$solution_type))
      out$direction <- factor(as.character(out$direction), levels = c("lower", "upper"))
      out <- out[order(out$solution_type, out$.row_order), , drop = FALSE]
      out$solution_type <- as.character(out$solution_type)
      out$direction <- as.character(out$direction)

      return(.finish_result_table(out, c(
        "solution_type",
        "set",
        "role",
        "raw",
        "type",
        "method",
        "anchor",
        "direction",
        "start",
        "last_safe",
        "first_failing",
        "step_unit",
        "steps",
        "total_delta",
        "pct_raw_range",
        "reason"
      )))
    }

    key_cols <- c("set", "role", "raw", "type", "method", "anchor", "direction")
    keys <- unique(diagnostics[, key_cols, drop = FALSE])
    solution_types <- .result_solution_type_order(diagnostics$solution_type)

    rows <- lapply(
      seq_len(nrow(keys)),
      function(i) {
        key <- keys[i, , drop = FALSE]
        keep <- rep(TRUE, nrow(diagnostics))
        for (col in key_cols) {
          keep <- keep & diagnostics[[col]] == key[[col]]
        }

        dd_key <- diagnostics[keep, , drop = FALSE]
        row <- key
        row$start <- dd_key$start_value[1L]
        row$step_unit <- dd_key$step_unit_used[1L]

        for (solution_type in solution_types) {
          dd <- dd_key[dd_key$solution_type == solution_type, , drop = FALSE]
          prefix <- .result_solution_type_prefix(solution_type)
          if (nrow(dd) == 0L) {
            row[[paste0(prefix, "_last_safe")]] <- NA_real_
            row[[paste0(prefix, "_first_failing")]] <- NA_real_
            row[[paste0(prefix, "_steps")]] <- NA_integer_
            row[[paste0(prefix, "_total_delta")]] <- NA_real_
            row[[paste0(prefix, "_pct_raw_range")]] <- NA_real_
            row[[paste0(prefix, "_reason")]] <- NA_character_
          } else {
            reason <- .compact_result_reason(dd$stop_reason[1L], dd$change_kind[1L])
            row[[paste0(prefix, "_last_safe")]] <- dd$last_safe_value[1L]
            row[[paste0(prefix, "_first_failing")]] <- dd$failing_value[1L]
            row[[paste0(prefix, "_steps")]] <- dd$number_of_steps[1L]
            row[[paste0(prefix, "_total_delta")]] <- dd$total_delta_units[1L]
            row[[paste0(prefix, "_pct_raw_range")]] <- dd$delta_as_pct_of_raw_range[1L]
            row[[paste0(prefix, "_reason")]] <- reason
          }
        }

        row
      }
    )

    return(.finish_result_table(.bind_rows_result(rows)))
  }

  needed <- c(
    "set",
    "role",
    "raw",
    "type",
    "method",
    "anchor",
    "direction",
    "start_value",
    "last_safe_value",
    "failing_value",
    "step_unit_used",
    "number_of_steps",
    "total_delta_units",
    "delta_as_pct_of_raw_range",
    "stop_reason",
    "change_kind"
  )

  out <- .select_result_columns(
    diagnostics,
    from = needed,
    to = c(
      "set",
      "role",
      "raw",
      "type",
      "method",
      "anchor",
      "direction",
      "start",
      "last_safe",
      "first_failing",
      "step_unit",
      "steps",
      "total_delta",
      "pct_raw_range",
      "stop_reason",
      "change_kind"
    ),
    label = "calibration-test"
  )

  out$reason <- .result_reason_vector(out$stop_reason, out$change_kind)

  .finish_result_table(out, c(
    "set",
    "role",
    "raw",
    "type",
    "method",
    "anchor",
    "direction",
    "start",
    "last_safe",
    "first_failing",
    "step_unit",
    "steps",
    "total_delta",
    "pct_raw_range",
    "reason"
  ))
}

.compact_solution_type_names <- function(x) {
  if (is.null(x) || length(x) == 0L || all(is.na(x))) {
    return(NA_character_)
  }

  x <- as.character(x)
  x <- x[!is.na(x)]
  x <- trimws(x)
  x <- x[nzchar(x)]

  if (length(x) == 0L) {
    return(NA_character_)
  }

  x <- gsub("\\s+", "", x)
  x <- gsub("conservative", "CON", x, fixed = TRUE)
  x <- gsub("parsimonious", "PAR", x, fixed = TRUE)
  x <- gsub("intermediate", "INT", x, fixed = TRUE)
  x <- gsub(",", " | ", x, fixed = TRUE)

  paste(x, collapse = " | ")
}

.make_loo_results <- function(diagnostics) {
  needed <- c(
    "row_index",
    "case_label",
    "status",
    "change_kind",
    "fit_changed_types",
    "n_fit_deltas",
    "max_abs_fit_delta"
  )

  out <- .select_result_columns(diagnostics, needed, label = "leave-one-out")
  out <- .add_solution_and_fit_columns(out)

  .finish_result_table(out, c(
    "row_index",
    "case_label",
    "status",
    "solution_change",
    "fit_changed_types",
    "n_fit_deltas",
    "max_abs_fit_delta"
  ))
}

.make_subsample_results <- function(diagnostics) {
  needed <- c(
    "rep",
    "n_sample",
    "n_holdout",
    "status",
    "change_kind",
    "fit_changed_types",
    "n_fit_deltas",
    "max_abs_fit_delta",
    "exact_match_baseline",
    "term_jaccard_baseline"
  )

  out <- .select_result_columns(diagnostics, needed, label = "subsample-test")
  out <- .add_solution_and_fit_columns(out)

  .finish_result_table(out, c(
    "rep",
    "n_sample",
    "n_holdout",
    "status",
    "exact_match_baseline",
    "term_jaccard_baseline",
    "solution_change",
    "fit_changed_types",
    "n_fit_deltas",
    "max_abs_fit_delta"
  ))
}

.make_altset_results <- function(diagnostics) {
  needed <- c(
    "draw",
    "incl.cut",
    "n.cut",
    "status",
    "change_kind",
    "fit_changed_types",
    "n_fit_deltas",
    "max_abs_fit_delta",
    "changed_sets",
    "changed_roles",
    "n_changed_sets"
  )

  out <- .select_result_columns(diagnostics, needed, label = "altset-test")
  out <- .add_solution_and_fit_columns(out)

  out$changed_sets <- as.character(out$changed_sets)
  out$changed_sets[out$changed_sets %in% c("NA", "")] <- NA_character_
  out$changed_sets <- ifelse(
    is.na(out$changed_sets),
    NA_character_,
    gsub(",", ", ", out$changed_sets, fixed = TRUE)
  )

  out$changed_roles <- as.character(out$changed_roles)
  out$changed_roles[out$changed_roles %in% c("NA", "")] <- NA_character_
  out$changed_roles <- ifelse(
    is.na(out$changed_roles),
    NA_character_,
    gsub(",", ", ", out$changed_roles, fixed = TRUE)
  )

  .finish_result_table(out, c(
    "draw",
    "incl.cut",
    "n.cut",
    "status",
    "n_changed_sets",
    "changed_sets",
    "changed_roles",
    "solution_change",
    "fit_changed_types",
    "n_fit_deltas",
    "max_abs_fit_delta"
  ))
}

.make_cluster_results <- function(diagnostics, by_cluster, by_unit) {
  needed <- c(
    "solution_type",
    "configuration_key",
    "component_count",
    "status",
    "pooled_consistency",
    "pooled_coverage",
    "max_abs_delta_consistency",
    "max_abs_delta_coverage",
    "worst_cluster_consistency_id",
    "worst_cluster_coverage_id",
    "within_available",
    "n_units_repeated"
  )

  .require_result_columns(diagnostics, needed, "cluster-test")
  overview <- diagnostics[, needed, drop = FALSE]
  overview$solution_type <- vapply(overview$solution_type, .compact_solution_type_names, character(1))

  names(overview) <- c(
    "solution_type",
    "configuration",
    "components",
    "status",
    "pooled_consistency",
    "pooled_coverage",
    "max_abs_cluster_delta_consistency",
    "max_abs_cluster_delta_coverage",
    "worst_cluster_consistency_id",
    "worst_cluster_coverage_id",
    "within_available",
    "n_units_repeated"
  )

  overview <- .finish_result_table(overview)

  clusters <- NULL
  if (!is.null(by_cluster) && length(by_cluster) > 0L) {
    cluster_rows <- lapply(by_cluster, function(x) {
      d <- x$clusters
      if (is.null(d) || !is.data.frame(d) || nrow(d) == 0L) {
        return(NULL)
      }

      data.frame(
        solution_type = .compact_solution_type_names(x$solution_type),
        configuration = x$configuration_key,
        solution_expression = x$solution_expression,
        component = as.character(d$component),
        cluster_id = as.character(d$cluster_id),
        cluster_size = as.integer(d$cluster_size),
        consistency = as.numeric(d$consistency),
        coverage = as.numeric(d$coverage),
        delta_consistency = as.numeric(d$delta_consistency),
        delta_coverage = as.numeric(d$delta_coverage),
        stringsAsFactors = FALSE
      )
    })

    cluster_rows <- Filter(Negate(is.null), cluster_rows)
    if (length(cluster_rows) > 0L) {
      clusters <- do.call(rbind, cluster_rows)
      rownames(clusters) <- NULL
    }
  }

  units <- NULL
  if (!is.null(by_unit) && length(by_unit) > 0L) {
    unit_rows <- lapply(by_unit, function(x) {
      d <- x$units
      if (is.null(d) || !is.data.frame(d) || nrow(d) == 0L) {
        return(NULL)
      }

      data.frame(
        solution_type = .compact_solution_type_names(x$solution_type),
        configuration = x$configuration_key,
        solution_expression = x$solution_expression,
        component = as.character(d$component),
        unit_id = as.character(d$unit_id),
        n_clusters = as.integer(d$n_clusters),
        consistency = as.numeric(d$consistency),
        coverage = as.numeric(d$coverage),
        delta_consistency = as.numeric(d$delta_consistency),
        delta_coverage = as.numeric(d$delta_coverage),
        stringsAsFactors = FALSE
      )
    })

    unit_rows <- Filter(Negate(is.null), unit_rows)
    if (length(unit_rows) > 0L) {
      units <- do.call(rbind, unit_rows)
      rownames(units) <- NULL
    }
  }

  list(
    overview = overview,
    clusters = clusters,
    units = units
  )
}

test_that("result constructors compact shared solution and fit columns", {
  diagnostics <- data.frame(
    row_index = 1L,
    case_label = "case 1",
    status = "ok",
    change_kind = "conservative:A,parsimonious:B",
    fit_changed_types = "conservative,intermediate",
    n_fit_deltas = 2L,
    max_abs_fit_delta = 0.25,
    stringsAsFactors = FALSE
  )

  out <- .make_loo_results(diagnostics)

  expect_identical(
    names(out),
    c(
      "row_index",
      "case_label",
      "status",
      "solution_change",
      "fit_changed_types",
      "n_fit_deltas",
      "max_abs_fit_delta"
    )
  )
  expect_identical(out$solution_change, "CON:A | PAR:B")
  expect_identical(out$fit_changed_types, "CON | INT")
  expect_identical(rownames(out), as.character(seq_len(nrow(out))))
})

test_that("result constructors report missing diagnostic columns consistently", {
  diagnostics <- data.frame(
    row_index = 1L,
    case_label = "case 1",
    status = "ok",
    change_kind = NA_character_,
    n_fit_deltas = 0L,
    max_abs_fit_delta = 0,
    stringsAsFactors = FALSE
  )

  expect_error(
    .make_loo_results(diagnostics),
    "Cannot build leave-one-out `results`; missing columns: fit_changed_types",
    fixed = TRUE
  )
})

test_that("boundary results use change kind before stop reason", {
  diagnostics <- data.frame(
    direction = "lower",
    incl.cut_start = 0.8,
    incl.cut_last_safe = 0.7,
    incl.cut_first_failing = 0.6,
    number_of_steps = 2L,
    total_delta = -0.2,
    stop_reason = "search_boundary_lower",
    change_kind = "conservative:A",
    stringsAsFactors = FALSE
  )

  out <- .make_boundary_results(diagnostics, "incl.cut")

  expect_identical(
    names(out),
    c("direction", "start", "last_safe", "first_failing", "steps", "total_delta", "reason")
  )
  expect_identical(out$reason, "CON:A")
})

test_that("all-solution boundary results support wide and long layouts", {
  diagnostics <- data.frame(
    direction = rep(c("lower", "upper"), each = 2L),
    solution_type = rep(c("conservative", "parsimonious"), times = 2L),
    incl.cut_start = 0.8,
    incl.cut_last_safe = c(0.7, 0.8, 0.9, 0.85),
    incl.cut_first_failing = c(0.6, 0.7, 1, 0.9),
    number_of_steps = c(1L, 0L, 1L, 1L),
    total_delta = c(-0.1, 0, 0.1, 0.05),
    stop_reason = c("search_boundary_lower", "search_boundary_lower", "search_boundary_upper", "search_boundary_upper"),
    change_kind = c("conservative:A", NA, NA, "parsimonious:B"),
    stringsAsFactors = FALSE
  )

  wide <- .make_boundary_results(diagnostics, "incl.cut")
  long <- .make_boundary_results(diagnostics, "incl.cut", result_shape = "long")

  expect_false("reason" %in% names(wide))
  expect_true(all(c("con_reason", "par_reason") %in% names(wide)))
  expect_identical(
    names(long),
    c("solution_type", "direction", "start", "last_safe", "first_failing", "steps", "total_delta", "reason")
  )
  expect_identical(long$solution_type, c("conservative", "conservative", "parsimonious", "parsimonious"))
  expect_identical(long$direction, c("lower", "upper", "lower", "upper"))
})

test_that("all-solution calibration results support wide and long layouts", {
  diagnostics <- data.frame(
    set = rep("A", 4L),
    role = rep("condition", 4L),
    raw = rep("A_raw", 4L),
    type = rep("fuzzy", 4L),
    method = rep("direct", 4L),
    anchor = rep(c("E", "C"), each = 2L),
    direction = rep("lower", 4L),
    solution_type = rep(c("conservative", "parsimonious"), times = 2L),
    start_value = c(20, 20, 40, 40),
    last_safe_value = c(19, 20, 39, 38),
    failing_value = c(18, 19, 38, 37),
    step_unit_used = 1,
    number_of_steps = c(1L, 0L, 1L, 2L),
    total_delta_units = c(-1, 0, -1, -2),
    delta_as_pct_of_raw_range = c(-0.01, 0, -0.01, -0.02),
    stop_reason = "search_boundary_lower",
    change_kind = c("conservative:A", NA, NA, "parsimonious:B"),
    stringsAsFactors = FALSE
  )

  wide <- .make_calib_results(diagnostics)
  long <- .make_calib_results(diagnostics, result_shape = "long")

  expect_false("reason" %in% names(wide))
  expect_true(all(c("con_reason", "par_reason") %in% names(wide)))
  expect_identical(
    names(long),
    c(
      "solution_type",
      "set",
      "role",
      "raw",
      "type",
      "method",
      "anchor",
      "direction",
      "start",
      "last_safe",
      "first_failing",
      "step_unit",
      "steps",
      "total_delta",
      "pct_raw_range",
      "reason"
    )
  )
  expect_identical(long$solution_type, c("conservative", "conservative", "parsimonious", "parsimonious"))
  expect_identical(long$anchor, c("E", "C", "E", "C"))
})

.normalize_i_mode <- function(i_mode = NULL) {
  if (is.null(i_mode)) {
    return(list())
  }

  i_mode <- match.arg(i_mode, c("all", "C1P1"))

  list(
    i_mode = i_mode
  )
}

.normalize_i_solution_names <- function(i.sol) {
  nm <- names(i.sol)

  if (is.null(nm) || any(nm == "")) {
    nm <- paste0("I", seq_along(i.sol))
  }

  nm
}

.select_single_model_position <- function(n_models, which_M, context) {
  if (n_models < 1L) return(integer(0))

  if (which_M > n_models) {
    stop("Requested `which_M = ", which_M, "` but only ", n_models, " model(s) are available for ", context, ".")
  }

  which_M
}

.resolve_branches <- function(obj, i_mode, branches) {
  if (is.null(obj[["i.sol"]]) || length(obj[["i.sol"]]) == 0L) {
    stop("The supplied intermediate object has no `i.sol` component.")
  }

  available <- names(obj[["i.sol"]])

  if (is.null(available) || any(available == "")) {
    stop("The supplied intermediate object has unnamed `i.sol` branches.")
  }

  if (!is.null(branches)) {
    if (!is.character(branches) || length(branches) < 1L) {
      stop("`branches` must be NULL or a non-empty character vector.")
    }

    missing_branches <- setdiff(branches, available)

    if (length(missing_branches) > 0L) {
      stop("Requested branch(es) not found in `i.sol`: ", paste(missing_branches, collapse = ", "))
    }

    return(branches)
  }

  if (i_mode == "C1P1") {
    if (!"C1P1" %in% available) {
      stop("`i_mode = \"C1P1\"` requested, but no `C1P1` branch exists in `i.sol`.")
    }

    return("C1P1")
  }

  available
}

#' Subsample robustness test for QCA solutions
#'
#' Draws repeated subsamples without replacement, reruns the QCA analysis on
#' each subsample, and records whether the monitored solution changes, whether
#' selected fit measures change, and whether a subsample run ends with an
#' error. The function can sample without stratification, stratify by the
#' baseline outcome, or stratify by a grouping column. This is a stringent,
#' punishing robustness check and is most useful when sample-composition
#' sensitivity is substantively important.
#'
#' @param data A data frame object containing the outcome and condition
#'   columns used in the QCA analysis. `data` must have at least
#'   three rows.
#' @param outcome Name of the outcome. This must be a single
#'   non-empty character string.
#' @param conditions Optional character vector of condition names. If `NULL`,
#'   the condition set is left to [QCA::truthTable()]. When
#'   `calib = "recompute"`, `conditions` must be supplied explicitly.
#' @param calib Calibration handling for subsample runs. `"fixed"` uses the
#'   calibrated values already present in `data`. `"recompute"` recalibrates
#'   the outcome and conditions within each subsample using `raw.data` and
#'   `calib_spec`.
#' @param raw.data Raw-data frame object used when `calib = "recompute"`.
#'   It must have the same number of rows as `data`.
#' @param calib_spec Calibration specification used when
#'   `calib = "recompute"`. This must be a named list keyed by
#'   `c(outcome, conditions)`. Each entry must describe the raw source column,
#'   the set type, and the calibration inputs used to rebuild the calibrated
#'   set. Use the same `calib_spec` structure described for [calib.test()]. An
#'   entry may additionally contain `findTh`, a named list of [QCA::findTh()]
#'   arguments; when supplied, thresholds are re-estimated within each
#'   subsample instead of reusing the baseline `thresholds`.
#' @param sample_n Integer subsample size. Supply exactly one of `sample_n` or
#'   `sample_prop`.
#' @param sample_prop Proportion used to determine the realized subsample size.
#'   Supply exactly one of `sample_n` or `sample_prop`. The realized sample
#'   size is `round(sample_prop * nrow(data))`.
#' @param reps Number of subsample replications.
#' @param stratify Stratification mode. `"none"` samples from the full dataset
#'   without stratification, `"outcome"` stratifies on the baseline analysis
#'   outcome, and `"user"` stratifies on `strata`.
#' @param strata Optional stratification column used when
#'   `stratify = "user"`. It must have length `nrow(data)` and contain no
#'   missing values.
#' @param seed Optional integer seed passed to [set.seed()] before drawing the
#'   subsamples.
#' @param case_labels Optional character vector of case labels with length
#'   `nrow(data)`. If `NULL`, the function uses `rownames(data)` when
#'   available; otherwise it uses row numbers converted to character.
#' @param incl.cut Inclusion cutoff passed to [QCA::truthTable()]. This must
#'   be a single finite number in `[0, 1]`.
#' @param n.cut Truth table frequency cutoff passed to [QCA::truthTable()].
#'   This must be an integer-like value of at least `1`.
#' @param solution Solution type to monitor. Accepted values are `"all"`,
#'   `"con"` or `"conservative"`, `"par"` or `"parsimonious"`, and `"int"` or
#'   `"intermediate"`.
#' @param include Optional minimization include setting. Currently, this
#'   argument accepts only `NULL`, `""`, or `"?"`.
#' @param dir.exp Directional expectations used when the monitored solution is
#'   intermediate.
#' @param which_M Positive integer giving which solution alternative to use
#'   when minimization returns multiple models.
#' @param i_mode Character string controlling intermediate-solution selection.
#'   Accepted values are `"all"` and `"C1P1"`.
#' @param exclude_mode Character string controlling how excluded rows are
#'   handled for parsimonious and intermediate minimization. `"recompute"`
#'   recalculates exclusions from each subsample truth table, `"static"`
#'   reuses `exclude_static`, and `"none"` does not use exclusions.
#' @param exclude_recompute Named list of arguments passed to [QCA::findRows()]
#'   when `exclude_mode = "recompute"`.
#' @param exclude_static Already computed exclusion object reused when
#'   `exclude_mode = "static"`.
#' @param fit_measures Character vector of fit-measure names to compare between
#'   the baseline run and each subsample run. Use `NULL` to disable fit
#'   comparison.
#' @param fit_tol Non-negative tolerance used when deciding whether fit values
#'   changed.
#' @param progress Logical; if `TRUE` and the session is interactive, show a
#'   text progress bar.
#' @param x A `subsample_test` object returned by [subsample.test()].
#' @param row.names Logical; passed to [print.data.frame()] by
#'   [print.subsample_test()].
#' @param ... Additional arguments. In [subsample.test()], named arguments
#'   matching [QCA::truthTable()] formals are passed to [QCA::truthTable()],
#'   and the remaining named arguments are forwarded to [QCA::minimize()]
#'   after removing names reserved by `subsample.test()`. The function
#'   also looks in `...` for `include`, `dir.exp`, or `direxp` if those
#'   arguments were not supplied explicitly. In [print.subsample_test()],
#'   `...` is passed to [print.data.frame()]. In
#'   [as.data.frame.subsample_test()], `...` is ignored.
#'
#' @returns An object of class `subsample_test` with the following components:
#'   \describe{
#'     \item{`diagnostics`}{A detailed data frame with one row per subsample
#'     run. It records the replication number, subsample size, holdout size,
#'     run status, solution-change classification, fit-change classification,
#'     exact-match status relative to the baseline solution, term-set Jaccard
#'     similarity to the baseline solution, and error information when a run
#'     fails.}
#'     \item{`results`}{A compact data frame with the columns
#'     `rep`, `n_sample`, `n_holdout`, `status`, `exact_match_baseline`,
#'     `term_jaccard_baseline`, `solution_change`, `fit_changed_types`,
#'     `n_fit_deltas`, and `max_abs_fit_delta`.}
#'     \item{`summary`}{A list with summary objects named `exact_solution`,
#'     `term_stability`, `fit_stability`, `similarity`, and `calibration`.}
#'     \item{`baseline`}{A list containing the baseline analysis built from the
#'     full dataset, including the truth table, minimization output,
#'     selected solution terms used for comparison, fit values, exclusion
#'     information, and calibration information.}
#'     \item{`by_run`}{A named list with one entry per subsample run. Each
#'     entry stores the sampled and held-out cases, run status, baseline and
#'     subsample analyses, change summaries, fit deltas, exclusion
#'     information, calibration information, and any error information for
#'     that run.}
#'     \item{`settings`}{A list containing the analysis settings used to build
#'     the result object.}
#'   }
#'
#'   `print.subsample_test()` prints a concise summary and the `results`
#'   table. `as.data.frame.subsample_test()` returns the `results` table.
#'
#' @details
#' The function first runs the baseline analysis on the full dataset. It then
#' draws `reps` subsamples without replacement and reruns the analysis on each
#' subsample.
#'
#' Exactly one of `sample_n` or `sample_prop` must be supplied. The realized
#' subsample size must be at least `2` and strictly smaller than
#' `nrow(data)`.
#'
#' When `stratify = "outcome"`, the function stratifies on the baseline
#' analysis outcome and therefore requires that outcome to be crisp/binary
#' `0/1`. When `stratify = "user"`, the function stratifies on the supplied
#' `strata` vector.
#'
#' When `calib = "fixed"`, the subsample runs use the calibrated values
#' already present in `data`. When `calib = "recompute"`, the function rebuilds
#' the calibrated outcome and condition columns within each subsample using
#' [QCA::findTh()] and [QCA::calibrate()] as specified in `calib_spec`.
#'
#' The shared solution-control, exclusion, and returned-object conventions are
#' described in `?qcaERT_conventions`.
#'
#' @examples
#' \donttest{
#' library(QCA)
#' data(LR)
#'
#' conditions <- c("DEV", "URB", "LIT", "IND", "STB")
#' outcome <- "SURV"
#' dir_exp <- rep("1", length(conditions))
#'
#' thresholds <- list(
#'   DEV = findTh(LR$DEV, groups = 7),
#'   URB = findTh(LR$URB, groups = 4),
#'   LIT = findTh(LR$LIT, groups = 4),
#'   IND = findTh(LR$IND, groups = 4),
#'   STB = findTh(LR$STB, groups = 4),
#'   SURV = findTh(LR$SURV, groups = 4)
#' )
#'
#' dat <- LR
#' dat$DEV <- calibrate(LR$DEV, type = "fuzzy", thresholds = thresholds$DEV)
#' dat$URB <- calibrate(LR$URB, type = "fuzzy", thresholds = thresholds$URB)
#' dat$LIT <- calibrate(LR$LIT, type = "fuzzy", thresholds = thresholds$LIT)
#' dat$IND <- calibrate(LR$IND, type = "fuzzy", thresholds = thresholds$IND)
#' dat$STB <- calibrate(LR$STB, type = "fuzzy", thresholds = thresholds$STB)
#' dat$SURV <- calibrate(LR$SURV, type = "fuzzy", thresholds = thresholds$SURV)
#'
#' out <- subsample.test(
#'   data = dat,
#'   outcome = outcome,
#'   conditions = conditions,
#'   calib = "fixed",
#'   sample_prop = 0.8,
#'   reps = 50,
#'   seed = 123,
#'   case_labels = rownames(dat),
#'   incl.cut = 0.8,
#'   n.cut = 1,
#'   solution = "all",
#'   dir.exp = dir_exp,
#'   fit_measures = c("inclS", "PRI", "covS"),
#'   progress = TRUE
#' )
#'
#' out
#' as.data.frame(out)
#' }
#'
#' @seealso [calib.test()], [incl.test()], [ncut.test()], [loo.test()],
#'   [altset.test()], [theory.test()], [cluster.test()], [sol.df()]
#' @export
subsample.test <- function(
    data,
    outcome,
    conditions = NULL,
    calib = c("fixed", "recompute"),
    raw.data = NULL,
    calib_spec = NULL,
    sample_n = NULL,
    sample_prop = NULL,
    reps = 100,
    stratify = c("none", "outcome", "user"),
    strata = NULL,
    seed = NULL,
    case_labels = NULL,
    incl.cut = 1,
    n.cut = 1,
    solution = "all",
    include = NULL,
    dir.exp = NULL,
    which_M = 1,
    i_mode = c("all", "C1P1"),
    exclude_mode = c("recompute", "static", "none"),
    exclude_recompute = list(type = 2),
    exclude_static = NULL,
    fit_measures = c("inclS", "PRI", "covS"),
    fit_tol = 0,
    progress = TRUE,
    ...
) {
  .require_qca()

  selection_controls <- .normalize_i_mode(i_mode)
  i_mode <- selection_controls$i_mode
  exclude_mode <- match.arg(exclude_mode)
  stratify <- match.arg(stratify)
  calib <- match.arg(calib)

  if (is.null(dim(data)) || is.null(nrow(data)) || nrow(data) < 3L) {
    stop("`data` must be a data frame object with at least 3 rows.")
  }

  if (!is.character(outcome) || length(outcome) != 1L || !nzchar(outcome)) {
    stop("`outcome` must be a single non-empty character string.")
  }

  if (!is.null(conditions)) {
    if (!is.character(conditions) || length(conditions) < 1L) {
      stop("`conditions` must be NULL or a non-empty character vector.")
    }
  }
  .validate_outcome_conditions_distinct(outcome, conditions)

  if (!is.numeric(reps) || length(reps) != 1L || !is.finite(reps) || reps < 1) {
    stop("`reps` must be a single integer >= 1.")
  }
  reps <- as.integer(reps)

  if (!is.null(seed)) {
    if (!is.numeric(seed) || length(seed) != 1L || !is.finite(seed)) {
      stop("`seed` must be NULL or a single finite numeric value.")
    }
    seed <- as.integer(seed)
  }

  if (xor(is.null(sample_n), is.null(sample_prop)) == FALSE) {
    stop("Supply exactly one of `sample_n` or `sample_prop`.")
  }

  n_total <- nrow(data)

  if (!is.null(sample_n)) {
    if (!is.numeric(sample_n) || length(sample_n) != 1L || !is.finite(sample_n)) {
      stop("`sample_n` must be a single finite number.")
    }
    sample_n <- as.integer(sample_n)
  } else {
    if (!is.numeric(sample_prop) || length(sample_prop) != 1L || !is.finite(sample_prop) ||
        sample_prop <= 0 || sample_prop >= 1) {
      stop("`sample_prop` must be a single finite number in (0, 1).")
    }
    sample_n <- as.integer(round(sample_prop * n_total))
  }

  if (sample_n < 2L || sample_n >= n_total) {
    stop("The realized sample size must be at least 2 and strictly smaller than `nrow(data)`.")
  }

  sample_prop_realized <- sample_n / n_total

  if (!is.numeric(incl.cut) || length(incl.cut) != 1L || !is.finite(incl.cut) ||
      incl.cut < 0 || incl.cut > 1) {
    stop("`incl.cut` must be a single finite number in [0, 1].")
  }

  n.cut <- .as_integerish_scalar(n.cut, "n.cut", min = 1L)

  which_M <- .as_integerish_scalar(which_M, "which_M", min = 1L)

  if (!is.numeric(fit_tol) || length(fit_tol) != 1L || !is.finite(fit_tol) || fit_tol < 0) {
    stop("`fit_tol` must be a single finite number >= 0.")
  }

  if (is.null(fit_measures)) {
    fit_measures <- character(0)
  } else {
    if (!is.character(fit_measures)) {
      stop("`fit_measures` must be NULL or a character vector.")
    }
    fit_measures <- unique(trimws(fit_measures))
    fit_measures <- fit_measures[nzchar(fit_measures)]
  }

  mc <- match.call(expand.dots = FALSE)
  dots_raw <- list(...)
  .reject_exclusion_controls_in_dots(dots_raw, "subsample.test")
  .reject_calibration_inputs_in_dots(dots_raw, "subsample.test")

  if (is.null(dir.exp)) {
    de <- .dot_get(dots_raw, "dir.exp")
    if (is.null(de)) de <- .dot_get(dots_raw, "direxp")
    if (!is.null(de)) dir.exp <- de
  }

  if (is.null(include)) {
    inc <- .dot_get(dots_raw, "include")
    if (!is.null(inc)) include <- inc
  }

  solution_controls <- .resolve_solution_controls(
    solution = solution,
    include = include,
    dir.exp = dir.exp,
    caller = "subsample.test",
    style = "std"
  )

  solution <- solution_controls$solution
  include <- solution_controls$include
  monitored_solutions <- solution_controls$monitored

  .validate_exclusion_controls(
    exclude_mode = exclude_mode,
    exclude_recompute = exclude_recompute,
    exclude_static = exclude_static,
    exclude_recompute_supplied = !is.null(mc$exclude_recompute),
    exclude_static_supplied = !is.null(mc$exclude_static),
    monitored_solutions = monitored_solutions,
    style = "std"
  )

  dots_split <- .split_truth_table_minimize_dots(dots_raw)
  dots_tt <- dots_split$tt
  dots_min <- dots_split$min

  analysis_vars <- NULL
  raw_data_work <- NULL

  if (calib == "recompute") {
    if (is.null(conditions)) {
      stop("When `calib = \"recompute\"`, `conditions` must be supplied explicitly.")
    }

    analysis_vars <- c(outcome, conditions)

    if (is.null(raw.data)) {
      stop("When `calib = \"recompute\"`, `raw.data` must be supplied.")
    }
    if (is.null(dim(raw.data)) || is.null(nrow(raw.data)) || nrow(raw.data) != nrow(data)) {
      stop("When `calib = \"recompute\"`, `raw.data` must be a data frame object with the same number of rows as `data`.")
    }

    raw_data_work <- .reduced_as_df(raw.data)
    calib_spec <- .reduced_validate_calib_spec(
      calib_spec = calib_spec,
      analysis_vars = analysis_vars,
      raw.data = raw_data_work,
      data = data
    )
  } else {
    analysis_vars <- unique(c(outcome, conditions))
  }

  .build_truth_table <- function(data_step) {
    args <- c(
      list(
        data = data_step,
        outcome = outcome,
        incl.cut = incl.cut,
        n.cut = n.cut
      ),
      dots_tt
    )

    if (!is.null(conditions)) {
      args$conditions <- conditions
    }

    do.call(QCA::truthTable, args)
  }

  .run_once <- function(data_step, raw_step = NULL) {
    .reduced_run_once(
      data_step = data_step,
      raw_step = raw_step,
      calib = calib,
      recalibrate_dataset = function(data_step, raw_step) {
        .reduced_recalibrate_dataset(
          data_step = data_step,
          raw_step = raw_step,
          analysis_vars = analysis_vars,
          calib_spec = calib_spec
        )
      },
      build_truth_table = .build_truth_table,
      monitored_solutions = monitored_solutions,
      exclude_mode = exclude_mode,
      exclude_recompute = exclude_recompute,
      exclude_static = exclude_static,
      dots_min = dots_min,
      dir.exp = dir.exp,
      which_M = which_M,
      i_mode = i_mode,
      fit_measures = fit_measures,
      include_analysis_data = TRUE
    )
  }

  .is_binary01 <- function(x) {
    x <- unique(x[!is.na(x)])
    length(x) > 0L && all(x %in% c(0, 1))
  }

  .alloc_by_strata <- function(group_sizes, n_total) {
    if (sum(group_sizes) < n_total) {
      stop("Internal error: requested sample size exceeds total available size.")
    }

    props <- group_sizes / sum(group_sizes)
    raw_alloc <- props * n_total
    alloc <- pmin(group_sizes, floor(raw_alloc))

    remainder <- n_total - sum(alloc)
    if (remainder > 0L) {
      frac <- raw_alloc - floor(raw_alloc)
      capacity <- group_sizes - alloc

      ord <- order(frac, decreasing = TRUE)
      for (i in ord) {
        if (remainder <= 0L) break
        if (capacity[i] > 0L) {
          alloc[i] <- alloc[i] + 1L
          capacity[i] <- capacity[i] - 1L
          remainder <- remainder - 1L
        }
      }
    }

    if (sum(alloc) != n_total) {
      capacity <- group_sizes - alloc
      ord <- order(capacity, decreasing = TRUE)
      for (i in ord) {
        if (sum(alloc) >= n_total) break
        if (capacity[i] > 0L) {
          take <- min(capacity[i], n_total - sum(alloc))
          alloc[i] <- alloc[i] + take
        }
      }
    }

    if (sum(alloc) != n_total) {
      stop("Could not allocate the requested sample size across strata.")
    }

    alloc
  }

  .sig_term_set <- function(sig) {
    if (is.null(sig)) return(character(0))
    x <- as.character(sig)
    x <- x[!is.na(x)]
    x <- trimws(x)
    x <- x[nzchar(x)]
    if (length(x) == 0L) return(character(0))

    x <- gsub("^.*:", "", x)
    out <- unlist(strsplit(x, "\\+", fixed = FALSE), use.names = FALSE)
    out <- .sanitize_terms(out)
    out[!grepl("^<.*>$", out)]
  }

  .run_term_set <- function(run_obj, monitored_solutions) {
    out <- character(0)
    for (solution_type in monitored_solutions) {
      out <- c(out, .sig_term_set(run_obj$sig[[solution_type]]))
    }
    sort(unique(out))
  }

  .solution_type_term_set <- function(run_obj, solution_type) {
    sort(unique(.sig_term_set(run_obj$sig[[solution_type]])))
  }

  .jaccard <- function(a, b) {
    a <- sort(unique(a))
    b <- sort(unique(b))
    if (length(a) == 0L && length(b) == 0L) return(1)
    u <- union(a, b)
    if (length(u) == 0L) return(NA_real_)
    length(intersect(a, b)) / length(u)
  }

  case_ref <- .reduced_resolve_case_labels(data = data, case_labels = case_labels)

  baseline <- .run_once(
    data_step = data,
    raw_step = if (calib == "recompute") raw_data_work else NULL
  )

  baseline_solution_types <- .reduced_baseline_solution_types(
    base = baseline,
    monitored_solutions = monitored_solutions
  )
  baseline_fit_solution_types <- .reduced_baseline_solution_types(
    base = baseline,
    monitored_solutions = monitored_solutions,
    fit_measures = fit_measures,
    require_fit = TRUE
  )
  baseline_missing <- baseline$status == "ok" && length(baseline_solution_types) == 0L
  baseline_invalid <- baseline$status != "ok" || length(baseline_solution_types) == 0L

  if (baseline_invalid) {
    diagnostics <- data.frame(
      rep = seq_len(reps),
      status = if (baseline$status != "ok") paste0("baseline_", baseline$status) else if (baseline_missing) "baseline_selected_solution_missing" else "baseline_no_comparable_solution",
      n_sample = rep(sample_n, reps),
      n_holdout = rep(n_total - sample_n, reps),
      changed = NA,
      solution_changed = NA,
      changed_types = NA_character_,
      change_kind = NA_character_,
      fit_changed = NA,
      fit_changed_types = NA_character_,
      n_fit_deltas = NA_integer_,
      max_abs_fit_delta = NA_real_,
      exact_match_baseline = NA,
      term_jaccard_baseline = NA_real_,
      error_source = baseline$error_source,
      error_message = baseline$error_message,
      stringsAsFactors = FALSE
    )

    return(.new_result_object(
      "subsample_test",
      diagnostics = diagnostics,
      results = .make_subsample_results(diagnostics),
      summary = list(),
      baseline = baseline,
      by_run = setNames(vector("list", reps), paste0("R", seq_len(reps))),
      settings = list(
        outcome = outcome,
        conditions = conditions,
        calib = calib,
        sample_n = sample_n,
        sample_prop_requested = sample_prop,
        sample_prop_realized = sample_prop_realized,
        reps = reps,
        stratify = stratify,
        incl.cut = incl.cut,
        n.cut = n.cut,
        solution = solution,
        monitored_solutions = monitored_solutions,
        include = include,
        dir.exp = dir.exp,
        which_M = which_M,
        i_mode = i_mode,
        exclude_mode = exclude_mode,
        exclude_recompute = exclude_recompute,
        exclude_static = exclude_static,
        fit_measures = fit_measures,
        fit_tol = fit_tol,
        seed = seed
      )
    ))
  }

  baseline_analysis_data <- baseline$analysis_data

  if (stratify == "outcome") {
    yb <- baseline_analysis_data[[outcome]]
    if (!.is_binary01(yb)) {
      stop("`stratify = \"outcome\"` requires the baseline analysis outcome to be crisp/binary 0/1.")
    }
    strata_use <- yb
  } else if (stratify == "user") {
    if (is.null(strata)) {
      stop("`stratify = \"user\"` requires `strata`.")
    }
    if (length(strata) != n_total) {
      stop("`strata` must have length `nrow(data)`.")
    }
    strata_use <- as.character(strata)
  } else {
    strata_use <- rep("all", n_total)
  }

  strata_use <- as.character(strata_use)
  if (any(is.na(strata_use))) {
    stop("`strata` / baseline outcome strata contain missing values.")
  }

  strata_levels <- unique(strata_use)
  strata_index <- split(seq_len(n_total), strata_use)
  strata_sizes <- vapply(strata_index, length, integer(1))
  sample_alloc <- .alloc_by_strata(group_sizes = strata_sizes, n_total = sample_n)
  names(sample_alloc) <- names(strata_sizes)

  sampling_plan <- vector("list", reps)

  if (!is.null(seed)) set.seed(seed)

  for (r in seq_len(reps)) {
    idx_sample <- integer(0)
    realized_counts <- integer(0)
    names(realized_counts) <- character(0)

    for (lev in names(strata_index)) {
      idx_pool <- strata_index[[lev]]
      n_take <- sample_alloc[[lev]]
      if (n_take > 0L) {
        draw <- sample(idx_pool, size = n_take, replace = FALSE)
        idx_sample <- c(idx_sample, draw)
        realized_counts <- c(realized_counts, n_take)
        names(realized_counts)[length(realized_counts)] <- lev
      }
    }

    idx_sample <- sort(idx_sample)
    idx_holdout <- setdiff(seq_len(n_total), idx_sample)

    sampling_plan[[r]] <- list(
      rep = r,
      sample_index = idx_sample,
      holdout_index = idx_holdout,
      sample_labels = case_ref$case_label[idx_sample],
      holdout_labels = case_ref$case_label[idx_holdout],
      strata_counts = realized_counts
    )
  }

  progress_state <- .new_qcaert_progress(total = reps, progress = progress)
  on.exit(progress_state$close(), add = TRUE)
  .bump_pb <- progress_state$tick

  diag_rows <- vector("list", reps)
  by_run <- vector("list", reps)
  names(by_run) <- paste0("R", seq_len(reps))

  baseline_term_set_all <- .run_term_set(baseline, baseline_solution_types)

  for (r in seq_len(reps)) {
    plan_r <- sampling_plan[[r]]
    idx_sample <- plan_r$sample_index
    idx_holdout <- plan_r$holdout_index

    data_sample <- data[idx_sample, , drop = FALSE]
    raw_sample <- if (calib == "recompute") raw_data_work[idx_sample, , drop = FALSE] else NULL

    cur <- .run_once(
      data_step = data_sample,
      raw_step = raw_sample
    )

    comparable_solution_types <- .reduced_comparable_solution_types(
      base = baseline,
      cur = cur,
      monitored_solutions = baseline_solution_types
    )

    if (cur$status != "ok" || length(comparable_solution_types) == 0L) {
      status_info <- .reduced_no_comparable_status(
        cur = cur,
        candidate_solution_types = baseline_solution_types
      )

      diag_rows[[r]] <- data.frame(
        rep = r,
        status = status_info$status,
        n_sample = length(idx_sample),
        n_holdout = length(idx_holdout),
        changed = NA,
        solution_changed = NA,
        changed_types = NA_character_,
        change_kind = NA_character_,
        fit_changed = NA,
        fit_changed_types = NA_character_,
        n_fit_deltas = NA_integer_,
        max_abs_fit_delta = NA_real_,
        exact_match_baseline = NA,
        term_jaccard_baseline = NA_real_,
        error_source = status_info$error_source,
        error_message = status_info$error_message,
        stringsAsFactors = FALSE
      )

      by_run[[r]] <- list(
        rep = r,
        sample_index = idx_sample,
        sample_labels = plan_r$sample_labels,
        holdout_index = idx_holdout,
        holdout_labels = plan_r$holdout_labels,
        strata_counts = plan_r$strata_counts,
        status = status_info$status,
        baseline = baseline,
        reduced = cur,
        solution_changed = NA,
        changed_types = NA_character_,
        change_kind = NA_character_,
        fit_changed = NA,
        fit_changed_types = NA_character_,
        fit_changed_measures = character(0),
        fit_delta = numeric(0),
        n_fit_deltas = NA_integer_,
        max_abs_fit_delta = NA_real_,
        exact_match_baseline = NA,
        term_jaccard_baseline = NA_real_,
        exclude_baseline = baseline$exclude_used,
        exclude_reduced = cur$exclude_used,
        calibration_baseline = baseline$calibration,
        calibration_run = cur$calibration,
        baseline_fit = baseline$fit,
        reduced_fit = cur$fit,
        error_source = status_info$error_source,
        error_message = status_info$error_message
      )

      .bump_pb()
      next
    }

    sol_info <- .reduced_solution_change_info(
      base = baseline,
      cur = cur,
      monitored_solutions = baseline_solution_types
    )

    fit_info <- .reduced_fit_change_info(
      base = baseline,
      cur = cur,
      monitored_solutions = baseline_fit_solution_types,
      fit_tol = fit_tol,
      fit_measures = fit_measures
    )

    exact_match_baseline <- !isTRUE(sol_info$changed)
    term_set_cur <- .run_term_set(cur, baseline_solution_types)
    term_jaccard_baseline <- .jaccard(baseline_term_set_all, term_set_cur)
    changed_any <- isTRUE(sol_info$changed) || isTRUE(fit_info$changed)

    diag_rows[[r]] <- data.frame(
      rep = r,
      status = "ok",
      n_sample = length(idx_sample),
      n_holdout = length(idx_holdout),
      changed = changed_any,
      solution_changed = isTRUE(sol_info$changed),
      changed_types = sol_info$changed_types,
      change_kind = sol_info$change_kind,
      fit_changed = isTRUE(fit_info$changed),
      fit_changed_types = fit_info$changed_types,
      n_fit_deltas = as.integer(fit_info$n_changed),
      max_abs_fit_delta = fit_info$max_abs_delta,
      exact_match_baseline = exact_match_baseline,
      term_jaccard_baseline = term_jaccard_baseline,
      error_source = NA_character_,
      error_message = NA_character_,
      stringsAsFactors = FALSE
    )

    by_run[[r]] <- list(
      rep = r,
      sample_index = idx_sample,
      sample_labels = plan_r$sample_labels,
      holdout_index = idx_holdout,
      holdout_labels = plan_r$holdout_labels,
      strata_counts = plan_r$strata_counts,
      status = "ok",
      baseline = baseline,
      reduced = cur,
      solution_changed = isTRUE(sol_info$changed),
      changed_types = sol_info$changed_types,
      change_kind = sol_info$change_kind,
      fit_changed = isTRUE(fit_info$changed),
      fit_changed_types = fit_info$changed_types,
      fit_changed_measures = fit_info$changed_measures,
      fit_delta = fit_info$delta,
      n_fit_deltas = as.integer(fit_info$n_changed),
      max_abs_fit_delta = fit_info$max_abs_delta,
      exact_match_baseline = exact_match_baseline,
      term_jaccard_baseline = term_jaccard_baseline,
      exclude_baseline = baseline$exclude_used,
      exclude_reduced = cur$exclude_used,
      calibration_baseline = baseline$calibration,
      calibration_run = cur$calibration,
      baseline_fit = baseline$fit,
      reduced_fit = cur$fit,
      error_source = NA_character_,
      error_message = NA_character_
    )

    .bump_pb()
  }

  diagnostics <- .bind_rows_result_sorted(diag_rows, order_by = "rep")

  .sig_to_key <- function(sig) {
    if (is.null(sig)) return(NA_character_)
    if (length(sig) == 1L) return(as.character(sig))
    paste(sort(as.character(sig)), collapse = " || ")
  }

  .build_exact_solution_summary <- function(by_run, baseline, monitored_solutions) {
    out <- vector("list", length(monitored_solutions))
    names(out) <- monitored_solutions

    for (solution_type in monitored_solutions) {
      ok_runs <- by_run[vapply(
        by_run,
        function(x) identical(x$status, "ok") && identical(x$reduced$solution_type_status[[solution_type]], "ok"),
        logical(1)
      )]
      baseline_sig <- baseline$sig[[solution_type]]
      baseline_key <- .sig_to_key(baseline_sig)

      keys <- vapply(
        ok_runs,
        function(x) .sig_to_key(x$reduced$sig[[solution_type]]),
        character(1)
      )

      freq <- sort(table(keys), decreasing = TRUE)

      out[[solution_type]] <- list(
        baseline_solution_expression = baseline_sig,
        baseline_key = baseline_key,
        n_successful = length(ok_runs),
        n_exact = sum(keys == baseline_key, na.rm = TRUE),
        prop_exact = if (length(ok_runs) == 0L) NA_real_ else mean(keys == baseline_key, na.rm = TRUE),
        frequencies = freq
      )
    }

    out
  }

  .build_term_stability_summary <- function(by_run, baseline, monitored_solutions) {
    out <- vector("list", length(monitored_solutions))
    names(out) <- monitored_solutions

    for (solution_type in monitored_solutions) {
      ok_runs <- by_run[vapply(
        by_run,
        function(x) identical(x$status, "ok") && identical(x$reduced$solution_type_status[[solution_type]], "ok"),
        logical(1)
      )]
      baseline_terms <- .solution_type_term_set(baseline, solution_type)

      run_terms <- lapply(ok_runs, function(x) .solution_type_term_set(x$reduced, solution_type))
      all_terms <- sort(unique(c(baseline_terms, unlist(run_terms, use.names = FALSE))))

      if (length(all_terms) == 0L) {
        out[[solution_type]] <- data.frame(
          term = character(0),
          baseline = logical(0),
          appearance_rate = numeric(0),
          stringsAsFactors = FALSE
        )
      } else {
        appearance <- vapply(
          all_terms,
          function(term) {
            if (length(run_terms) == 0L) return(NA_real_)
            mean(vapply(run_terms, function(x) term %in% x, logical(1)))
          },
          numeric(1)
        )

        out[[solution_type]] <- data.frame(
          term = all_terms,
          baseline = all_terms %in% baseline_terms,
          appearance_rate = appearance,
          stringsAsFactors = FALSE
        )
      }
    }

    out
  }

  .build_fit_summary <- function(by_run, baseline, monitored_solutions) {
    if (length(fit_measures) == 0L) return(list())

    out <- vector("list", length(monitored_solutions))
    names(out) <- monitored_solutions

    for (solution_type in monitored_solutions) {
      ok_runs <- by_run[vapply(
        by_run,
        function(x) identical(x$status, "ok") && identical(x$reduced$solution_type_status[[solution_type]], "ok"),
        logical(1)
      )]
      base_fit <- baseline$fit[[solution_type]]
      fit_names <- sort(unique(c(names(base_fit), unlist(lapply(ok_runs, function(x) names(x$reduced$fit[[solution_type]])), use.names = FALSE))))

      if (length(fit_names) == 0L) {
        out[[solution_type]] <- data.frame(
          measure = character(0),
          baseline = numeric(0),
          median = numeric(0),
          q25 = numeric(0),
          q75 = numeric(0),
          median_delta = numeric(0),
          q25_delta = numeric(0),
          q75_delta = numeric(0),
          prop_changed = numeric(0),
          stringsAsFactors = FALSE
        )
        next
      }

      rows <- lapply(
        fit_names,
        function(nm) {
          vals <- vapply(ok_runs, function(x) {
            v <- .fit_get(x$reduced$fit[[solution_type]], nm)
            if (is.null(v)) NA_real_ else as.numeric(v)
          }, numeric(1))

          base_v <- .fit_get(base_fit, nm)
          if (is.null(base_v)) base_v <- NA_real_
          deltas <- vals - base_v

          data.frame(
            measure = nm,
            baseline = as.numeric(base_v),
            median = if (all(is.na(vals))) NA_real_ else stats::median(vals, na.rm = TRUE),
            q25 = if (all(is.na(vals))) NA_real_ else as.numeric(stats::quantile(vals, probs = 0.25, na.rm = TRUE, names = FALSE)),
            q75 = if (all(is.na(vals))) NA_real_ else as.numeric(stats::quantile(vals, probs = 0.75, na.rm = TRUE, names = FALSE)),
            median_delta = if (all(is.na(deltas))) NA_real_ else stats::median(deltas, na.rm = TRUE),
            q25_delta = if (all(is.na(deltas))) NA_real_ else as.numeric(stats::quantile(deltas, probs = 0.25, na.rm = TRUE, names = FALSE)),
            q75_delta = if (all(is.na(deltas))) NA_real_ else as.numeric(stats::quantile(deltas, probs = 0.75, na.rm = TRUE, names = FALSE)),
            prop_changed = if (length(vals) == 0L) NA_real_ else mean(abs(deltas) > fit_tol, na.rm = TRUE),
            stringsAsFactors = FALSE
          )
        }
      )

      out[[solution_type]] <- do.call(rbind, rows)
      rownames(out[[solution_type]]) <- NULL
    }

    out
  }

  .build_similarity_summary <- function(by_run, baseline, monitored_solutions) {
    rows <- lapply(
      monitored_solutions,
      function(solution_type) {
        ok_runs <- by_run[vapply(
          by_run,
          function(x) identical(x$status, "ok") && identical(x$reduced$solution_type_status[[solution_type]], "ok"),
          logical(1)
        )]
        base_terms <- .solution_type_term_set(baseline, solution_type)
        run_term_sets <- lapply(ok_runs, function(x) .solution_type_term_set(x$reduced, solution_type))

        baseline_j <- vapply(run_term_sets, function(x) .jaccard(base_terms, x), numeric(1))

        pairwise_avg <- NA_real_
        if (length(run_term_sets) >= 2L) {
          combs <- utils::combn(seq_along(run_term_sets), 2L)
          pair_vals <- apply(
            combs,
            2L,
            function(ii) .jaccard(run_term_sets[[ii[1L]]], run_term_sets[[ii[2L]]])
          )
          pairwise_avg <- mean(pair_vals, na.rm = TRUE)
        }

        data.frame(
          solution_type = solution_type,
          n_successful = length(ok_runs),
          mean_baseline_jaccard = if (length(baseline_j) == 0L) NA_real_ else mean(baseline_j, na.rm = TRUE),
          median_baseline_jaccard = if (length(baseline_j) == 0L) NA_real_ else stats::median(baseline_j, na.rm = TRUE),
          avg_pairwise_jaccard = pairwise_avg,
          stringsAsFactors = FALSE
        )
      }
    )

    out <- do.call(rbind, rows)
    rownames(out) <- NULL
    out
  }

  .build_calibration_summary <- function(by_run, baseline, analysis_vars) {
    if (calib != "recompute") return(NULL)

    ok_runs <- by_run[vapply(by_run, function(x) identical(x$status, "ok"), logical(1))]
    rows <- list()

    for (nm in analysis_vars) {
      base_th <- baseline$calibration$thresholds_used[[nm]]
      run_th_list <- lapply(ok_runs, function(x) x$calibration_run$thresholds_used[[nm]])

      all_lengths <- c(length(base_th), vapply(run_th_list, length, integer(1)))
      kmax <- max(all_lengths, 0L)
      if (kmax == 0L) next

      th_names <- names(base_th)
      if (is.null(th_names) || length(th_names) != length(base_th) || any(!nzchar(th_names))) {
        th_names <- paste0("T", seq_len(max(length(base_th), kmax)))
      }
      if (length(th_names) < kmax) {
        th_names <- c(th_names, paste0("T", seq(from = length(th_names) + 1L, to = kmax)))
      }

      for (k in seq_len(kmax)) {
        vals <- vapply(
          run_th_list,
          function(x) {
            if (length(x) < k) return(NA_real_)
            as.numeric(x[[k]])
          },
          numeric(1)
        )

        base_v <- if (length(base_th) >= k) as.numeric(base_th[[k]]) else NA_real_

        rows[[length(rows) + 1L]] <- data.frame(
          variable = nm,
          threshold = th_names[k],
          baseline = base_v,
          median = if (all(is.na(vals))) NA_real_ else stats::median(vals, na.rm = TRUE),
          q25 = if (all(is.na(vals))) NA_real_ else as.numeric(stats::quantile(vals, probs = 0.25, na.rm = TRUE, names = FALSE)),
          q75 = if (all(is.na(vals))) NA_real_ else as.numeric(stats::quantile(vals, probs = 0.75, na.rm = TRUE, names = FALSE)),
          prop_diff_baseline = if (length(vals) == 0L || is.na(base_v)) NA_real_ else mean(abs(vals - base_v) > 0, na.rm = TRUE),
          n_successful = length(ok_runs),
          stringsAsFactors = FALSE
        )
      }
    }

    if (length(rows) == 0L) return(NULL)
    out <- do.call(rbind, rows)
    rownames(out) <- NULL
    out
  }

  summary <- list(
    exact_solution = .build_exact_solution_summary(by_run, baseline, baseline_solution_types),
    term_stability = .build_term_stability_summary(by_run, baseline, baseline_solution_types),
    fit_stability = .build_fit_summary(by_run, baseline, baseline_fit_solution_types),
    similarity = .build_similarity_summary(by_run, baseline, baseline_solution_types),
    calibration = .build_calibration_summary(by_run, baseline, analysis_vars)
  )

  .new_result_object(
    "subsample_test",
    diagnostics = diagnostics,
    results = .make_subsample_results(diagnostics),
    summary = summary,
    baseline = baseline,
    by_run = by_run,
    settings = list(
      outcome = outcome,
      conditions = conditions,
      calib = calib,
      sample_n = sample_n,
      sample_prop_requested = sample_prop,
      sample_prop_realized = sample_prop_realized,
      reps = reps,
      stratify = stratify,
      incl.cut = incl.cut,
      n.cut = n.cut,
      solution = solution,
      monitored_solutions = monitored_solutions,
      include = include,
      dir.exp = dir.exp,
      which_M = which_M,
      i_mode = i_mode,
      exclude_mode = exclude_mode,
      exclude_recompute = exclude_recompute,
      exclude_static = exclude_static,
      fit_measures = fit_measures,
      fit_tol = fit_tol,
      seed = seed
    )
  )
}

#' Print a `subsample_test` object
#'
#' @rdname subsample.test
#' @export
print.subsample_test <- function(x, row.names = FALSE, ...) {
  results <- x$results
  settings <- x$settings
  summary <- x$summary

  .print_qcaert_heading("subsample_test", "subsample robustness", settings)
  if (!is.null(settings$reps)) {
    cat("Replications: ", settings$reps, "\n", sep = "")
  }
  if (!is.null(settings$sample_n)) {
    cat("Sample size: ", settings$sample_n, "\n", sep = "")
  }
  if (!is.null(settings$sample_prop_realized)) {
    cat("Sample proportion: ", format(settings$sample_prop_realized, trim = TRUE), "\n", sep = "")
  }
  if (!is.null(settings$stratify)) {
    cat("Stratification: ", settings$stratify, "\n", sep = "")
  }
  if (!is.null(settings$fit_measures) && length(settings$fit_measures) > 0L) {
    cat("Fit measures: ", paste(settings$fit_measures, collapse = ", "), "\n", sep = "")
  }

  .print_qcaert_table(results, "Result", row.names = row.names, ...)

  cat("\nSummary\n")
  if (is.data.frame(results) && nrow(results) > 0L) {
    n_ok <- sum(results$status == "ok", na.rm = TRUE)
    n_solution_changed <- .count_present(results$solution_change)
    n_fit_changed <- .count_present(results$fit_changed_types)
    n_errors <- sum(!is.na(results$status) & results$status != "ok", na.rm = TRUE)

    cat(" Successful runs: ", n_ok, "/", nrow(results), "\n", sep = "")
    cat(" Runs with any solution change: ", n_solution_changed, "\n", sep = "")
    cat(" Runs with any fit change: ", n_fit_changed, "\n", sep = "")
    cat(" Non-ok runs: ", n_errors, "\n", sep = "")

    if (any(!is.na(results$term_jaccard_baseline))) {
      cat(
        " Mean combined term-set Jaccard vs baseline: ",
        format(mean(results$term_jaccard_baseline, na.rm = TRUE), trim = TRUE),
        "\n",
        sep = ""
      )
    }
  }

  if (!is.null(summary$exact_solution) && length(summary$exact_solution) > 0L) {
    exact_rows <- lapply(
      names(summary$exact_solution),
      function(solution_type) {
        sx <- summary$exact_solution[[solution_type]]
        data.frame(
          solution_type = .compact_solution_type_names(solution_type),
          exact = as.integer(sx$n_exact),
          successful = as.integer(sx$n_successful),
          proportion = as.numeric(sx$prop_exact),
          stringsAsFactors = FALSE
        )
      }
    )

    exact_table <- do.call(rbind, exact_rows)
    rownames(exact_table) <- NULL
    .print_qcaert_table(exact_table, "Exact baseline matches by solution type", row.names = FALSE)
  }

  invisible(x)
}

#' Return the main results table from a `subsample_test` object
#'
#' @rdname subsample.test
#' @export
as.data.frame.subsample_test <- function(x, ...) {
  .as.data.frame_results(x, ...)
}

#' @rdname qcaERT_plots
#' @name qcaERT_plots
#' @method plot incl_test
#' @export
plot.incl_test <- function(
    x,
    type = c("interval", "trace"),
    directions = c("lower", "upper"),
    stop_reason = NULL,
    changed_types = NULL,
    solution_type = NULL,
    solution = NULL,
    monitored_solutions = NULL,
    i_mode = NULL,
    direction = NULL,
    show_stop = TRUE,
    legend = TRUE,
    theme = .plot_theme(),
    ...
) {
  .plot_require_ggplot2()

  type <- match.arg(type)
  dots_plot <- list(...)
  .plot_reject_solution_args(solution, monitored_solutions, "plot.incl_test")
  .plot_reject_solution_dots(dots_plot, "plot.incl_test")

  if (is.null(x$diagnostics) || !is.data.frame(x$diagnostics)) {
    stop("x must contain a data frame in x$diagnostics.", call. = FALSE)
  }

  required <- c(
    "direction",
    "solution",
    "monitored_solutions",
    "i_mode",
    "incl.cut_start",
    "incl.cut_last_safe",
    "incl.cut_first_failing",
    "stop_reason",
    "changed_types"
  )
  missing <- setdiff(required, names(x$diagnostics))
  if (length(missing)) {
    stop(
      "x$diagnostics is missing required column(s): ",
      paste(missing, collapse = ", "),
      call. = FALSE
    )
  }

  df0 <- x$diagnostics

  normalize_i_mode <- function(value) {
    if (is.null(value)) return(NULL)

    value <- trimws(as.character(value))
    value <- value[nzchar(value)]
    if (!length(value)) return(character(0))

    out <- character(length(value))
    for (i in seq_along(value)) {
      mode <- tolower(value[i])
      out[i] <- if (mode == "all") {
        "all"
      } else if (mode == "c1p1") {
        "C1P1"
      } else {
        stop("`i_mode` must contain only 'all' and/or 'C1P1'.", call. = FALSE)
      }
    }

    unique(out)
  }

  stop_no_rows <- function() {
    lines <- c("No rows left to plot after filtering.")

    if (!is.null(directions)) lines <- c(lines, paste0("Requested directions: ", paste(directions, collapse = ", ")))
    if (!is.null(stop_reason)) lines <- c(lines, paste0("Requested stop_reason: ", paste(stop_reason, collapse = ", ")))
    if (!is.null(changed_types)) lines <- c(lines, paste0("Requested changed_types: ", paste(changed_types, collapse = ", ")))
    if (!is.null(solution_type)) lines <- c(lines, paste0("Requested solution_type: ", paste(solution_type, collapse = ", ")))
    if (!is.null(i_mode)) lines <- c(lines, paste0("Requested i_mode: ", paste(i_mode, collapse = ", ")))

    for (column in c("direction", "solution_type", "stop_reason", "changed_types", "i_mode")) {
      available <- sort(table(.plot_as_chr0(df0[[column]])), decreasing = TRUE)
      available <- available[nzchar(names(available))]
      label <- paste0("Available ", column, ": ")
      if (length(available)) {
        lines <- c(lines, paste0(label, .plot_fmt_counts(available)))
      } else {
        lines <- c(lines, paste0(label, "<none>"))
      }
    }

    stop(paste(lines, collapse = "\n"), call. = FALSE)
  }

  directions <- .plot_normalize_directions(directions)
  selected_solution <- .plot_select_solution_type(df0, solution_type)
  df0 <- selected_solution$data
  df <- df0
  solution_type <- selected_solution$solution_type
  i_mode <- normalize_i_mode(i_mode)

  if (!is.null(directions)) df <- df[df$direction %in% directions, , drop = FALSE]
  if (!is.null(stop_reason)) df <- df[df$stop_reason %in% stop_reason, , drop = FALSE]
  if (!is.null(changed_types)) {
    keep <- .plot_match_solution_tokens(df$changed_types, changed_types, arg = "changed_types")
    df <- df[keep, , drop = FALSE]
  }
  if (!is.null(i_mode)) df <- df[df$i_mode %in% i_mode, , drop = FALSE]

  if (!nrow(df)) stop_no_rows()

  if (type == "trace") {
    if (is.null(direction)) {
      if (length(directions) == 1L) {
        direction <- directions
      } else {
        stop("For type = 'trace', supply direction = 'lower' or direction = 'upper'.", call. = FALSE)
      }
    }

    direction <- .plot_normalize_directions(direction, arg = "direction")
    if (length(direction) != 1L) {
      stop("For type = 'trace', `direction` must be a single value.", call. = FALSE)
    }
    if (!direction %in% df$direction) stop_no_rows()

    if (is.null(x$by_direction) || is.null(x$by_direction[[direction]])) {
      stop("direction not found in x$by_direction.", call. = FALSE)
    }

    path <- x$by_direction[[direction]]
    row <- df[df$direction == direction, , drop = FALSE][1L, , drop = FALSE]
    if (!is.null(path$by_solution_type)) {
      row_solution_type <- as.character(row$solution_type[1L])
      if (!is.null(path$by_solution_type[[row_solution_type]])) {
        path <- path$by_solution_type[[row_solution_type]]
      }
    }

    trace <- path$trace
    if (is.null(trace) || !is.data.frame(trace)) {
      stop("Trace not found for this direction.", call. = FALSE)
    }

    trace_required <- c("step", "incl.cut", "changed", "status")
    missing_trace <- setdiff(trace_required, names(trace))
    if (length(missing_trace)) {
      stop(
        "Trace is missing required column(s): ",
        paste(missing_trace, collapse = ", "),
        call. = FALSE
      )
    }

    trace$safe <- ifelse(is.na(trace$changed), "Unknown", ifelse(trace$changed, "No", "Yes"))
    trace$safe <- factor(trace$safe, levels = c("Yes", "No", "Unknown"))

    start_value <- row$incl.cut_start
    failing_value <- row$incl.cut_first_failing

    fail_point <- NULL
    if (is.finite(failing_value) && !is.na(failing_value)) {
      next_step <- if (nrow(trace) == 0L) 1L else max(trace$step, na.rm = TRUE) + 1L
      fail_point <- data.frame(
        step = next_step,
        incl.cut = failing_value,
        safe = factor("No", levels = levels(trace$safe))
      )
    }

    line_df <- trace[, c("step", "incl.cut"), drop = FALSE]
    if (!is.null(fail_point)) {
      present <- nrow(line_df) && any(line_df$step == fail_point$step & line_df$incl.cut == fail_point$incl.cut)
      if (!present) line_df <- rbind(line_df, fail_point[, c("step", "incl.cut"), drop = FALSE])
      line_df <- line_df[order(line_df$step), , drop = FALSE]
    }

    p <- ggplot2::ggplot() +
      ggplot2::geom_line(
        data = line_df,
        ggplot2::aes(x = .data[["step"]], y = .data[["incl.cut"]]),
        ...
      ) +
      ggplot2::geom_point(
        data = trace,
        ggplot2::aes(
          x = .data[["step"]],
          y = .data[["incl.cut"]],
          shape = .data[["safe"]],
          colour = .data[["safe"]]
        ),
        size = 4,
        alpha = 0.8,
        ...
      ) +
      ggplot2::scale_shape_manual(name = "Preserved", values = c(Yes = 19, No = 15, Unknown = 17)) +
      ggplot2::scale_colour_manual(name = "Preserved", values = .plot_preserved_colors()) +
      ggplot2::labs(
        title = paste0("Inclusion-cutoff trace: ", direction),
        x = "Step",
        y = "Inclusion cutoff"
      ) +
      theme

    if (!is.null(fail_point)) {
      p <- p + ggplot2::geom_point(
        data = fail_point,
        ggplot2::aes(
          x = .data[["step"]],
          y = .data[["incl.cut"]],
          shape = .data[["safe"]],
          colour = .data[["safe"]]
        ),
        inherit.aes = FALSE,
        size = 4,
        alpha = 0.8
      )
    }
    if (show_stop && is.finite(start_value) && !is.na(start_value)) {
      p <- p + ggplot2::geom_hline(yintercept = start_value, linetype = "dashed")
    }
    if (show_stop && is.finite(failing_value) && !is.na(failing_value)) {
      p <- p + ggplot2::geom_hline(yintercept = failing_value, linetype = "dotdash")
    }
    if (!legend) p <- p + ggplot2::theme(legend.position = "none")

    return(p)
  }

  lower <- df[df$direction == "lower", , drop = FALSE]
  upper <- df[df$direction == "upper", , drop = FALSE]

  lower_safe <- if (nrow(lower)) lower$incl.cut_last_safe[1L] else NA_real_
  upper_safe <- if (nrow(upper)) upper$incl.cut_last_safe[1L] else NA_real_
  start_value <- if (nrow(lower)) lower$incl.cut_start[1L] else NA_real_
  if (!is.finite(start_value) || is.na(start_value)) {
    if (nrow(upper)) start_value <- upper$incl.cut_start[1L]
  }

  points <- data.frame(
    point = c("Baseline", "Lower last safe", "Upper last safe"),
    value = c(start_value, lower_safe, upper_safe),
    band = "Robustness interval",
    stringsAsFactors = FALSE
  )
  points <- points[is.finite(points$value) & !is.na(points$value), , drop = FALSE]
  if (!nrow(points)) {
    stop("No finite values to plot after filtering.", call. = FALSE)
  }

  points$point <- factor(points$point, levels = c("Baseline", "Lower last safe", "Upper last safe"))
  points$band <- factor(points$band, levels = "Robustness interval")

  segment <- NULL
  if (is.finite(lower_safe) && !is.na(lower_safe) && is.finite(upper_safe) && !is.na(upper_safe)) {
    segment <- data.frame(
      x = lower_safe,
      xend = upper_safe,
      y = factor("Robustness interval", levels = "Robustness interval"),
      yend = factor("Robustness interval", levels = "Robustness interval")
    )
  }

  dots_point <- list(...)
  pt_size <- if (!is.null(dots_point$size)) dots_point$size else 3
  pt_alpha <- if (!is.null(dots_point$alpha)) dots_point$alpha else 0.8

  dots_point$size <- NULL
  dots_point$alpha <- NULL
  dots_point$mapping <- NULL
  dots_point$data <- NULL
  dots_point$inherit.aes <- NULL

  point_layer <- do.call(
    ggplot2::geom_point,
    c(
      list(
        mapping = ggplot2::aes(
          x = .data[["value"]],
          y = .data[["band"]],
          shape = .data[["point"]],
          colour = .data[["point"]],
          fill = .data[["point"]]
        ),
        size = pt_size,
        alpha = pt_alpha
      ),
      dots_point
    )
  )

  p <- ggplot2::ggplot(points)

  if (!is.null(segment)) {
    p <- p + ggplot2::geom_segment(
      data = segment,
      ggplot2::aes(
        x = .data[["x"]],
        xend = .data[["xend"]],
        y = .data[["y"]],
        yend = .data[["yend"]]
      ),
      inherit.aes = FALSE,
      linewidth = 0.8
    )
  }

  p <- p +
    point_layer +
    ggplot2::scale_shape_manual(name = "Point", values = c("Baseline" = 16, "Lower last safe" = 25, "Upper last safe" = 24)) +
    ggplot2::scale_colour_manual(name = "Point", values = .plot_boundary_colors()) +
    ggplot2::scale_fill_manual(name = "Point", values = .plot_boundary_colors()) +
    ggplot2::coord_cartesian(xlim = c(0, 1)) +
    ggplot2::labs(x = "Inclusion cutoff", y = NULL, title = NULL) +
    theme

  if (show_stop && is.finite(start_value) && !is.na(start_value)) {
    p <- p + ggplot2::geom_vline(xintercept = start_value, linetype = "dashed")
  }
  if (!legend) p <- p + ggplot2::theme(legend.position = "none")

  p
}

.solution_message <- function(style, plain, std) {
  style <- match.arg(style, c("std", "plain"))
  if (style == "std") std else plain
}

.normalize_solution <- function(solution, style = c("std", "plain")) {
  style <- match.arg(style)
  msg <- .solution_message(
    style,
    plain = "solution must be one of 'all', 'con'/'conservative', 'par'/'parsimonious', or 'int'/'intermediate'.",
    std = "`solution` must be one of 'all', 'con'/'conservative', 'par'/'parsimonious', or 'int'/'intermediate'."
  )

  if (!is.character(solution) || length(solution) != 1L || is.na(solution)) {
    stop(msg)
  }

  s <- tolower(trimws(solution))

  if (s == "all") return("all")
  if (s %in% c("con", "conservative")) return("conservative")
  if (s %in% c("par", "parsimonious")) return("parsimonious")
  if (s %in% c("int", "intermediate")) return("intermediate")

  stop(msg)
}

.normalize_solution_std <- function(solution) {
  .normalize_solution(solution, style = "std")
}

.normalize_include <- function(include, caller, style = c("std", "plain")) {
  style <- match.arg(style)
  if (is.null(include)) return(NULL)

  msg <- .solution_message(
    style,
    plain = "include must be NULL, '', or '?'.",
    std = "`include` must be NULL, '', or '?'."
  )
  if (!is.character(include) || length(include) != 1L || is.na(include)) {
    stop(msg)
  }

  include <- trimws(include)

  if (!include %in% c("", "?")) {
    stop(.solution_message(
      style,
      plain = sprintf("For %s, include must be NULL, '', or '?'.", caller),
      std = sprintf("For `%s`, `include` must be NULL, '', or '?'.", caller)
    ))
  }

  include
}

.split_dir_exp_string_generic <- function(x, conditions, endpoint_phrase) {
  x <- trimws(as.character(x))
  if (!nzchar(x)) return(character(0))

  if (grepl(":", x, fixed = TRUE)) {
    des <- trimws(unlist(strsplit(x, split = ":", fixed = TRUE), use.names = FALSE))
    des <- des[nzchar(des)]

    if (length(des) != 2L || !all(des %in% conditions)) {
      stop(
        sprintf(
          "dir.exp contains an invalid condition sequence. When using 'A:B', both endpoints must be condition names in %s.",
          endpoint_phrase
        )
      )
    }

    return(conditions[seq(match(des[1], conditions), match(des[2], conditions))])
  }

  out <- trimws(unlist(strsplit(x, "[[:space:],]+"), use.names = FALSE))
  out[nzchar(out)]
}

.normalize_dir_exp_generic <- function(dir.exp, conditions, endpoint_phrase) {
  if (is.null(dir.exp)) return(NULL)

  if (is.character(dir.exp) && length(dir.exp) == 1L && identical(dir.exp, "character(0)")) {
    return(NULL)
  }

  if (is.character(dir.exp) && length(dir.exp) == 0L) {
    return(NULL)
  }

  if (is.data.frame(dir.exp)) {
    dir.exp <- as.matrix(dir.exp)
  }

  if (is.matrix(dir.exp)) {
    if (ncol(dir.exp) != length(conditions)) {
      stop(
        sprintf(
          "dir.exp must have exactly one column per condition: expected %d column(s) for %d condition(s), got %d.",
          length(conditions), length(conditions), ncol(dir.exp)
        )
      )
    }
    return(dir.exp)
  }

  if (is.character(dir.exp)) {
    if (length(dir.exp) == 1L) {
      dir.exp <- .split_dir_exp_string_generic(
        x = dir.exp,
        conditions = conditions,
        endpoint_phrase = endpoint_phrase
      )
    } else {
      dir.exp <- trimws(as.character(dir.exp))
      dir.exp <- dir.exp[nzchar(dir.exp)]
    }

    if (length(dir.exp) == 0L) {
      return(NULL)
    }

    if (length(dir.exp) != length(conditions)) {
      stop(
        sprintf(
          "dir.exp must specify exactly one expectation per condition: expected %d value(s) for %d condition(s), got %d.",
          length(conditions), length(conditions), length(dir.exp)
        )
      )
    }

    return(dir.exp)
  }

  stop("dir.exp must be NULL, a character vector, a single minimize-style character string, or a matrix/data.frame with one column per condition.")
}

.resolve_solution_controls <- function(solution, include, dir.exp, caller, style = c("std", "plain")) {
  style <- match.arg(style)
  solution <- .normalize_solution(solution, style = style)
  include <- .normalize_include(include, caller, style = style)

  if (solution == "all") {
    if (!is.null(include)) {
      stop(.solution_message(
        style,
        plain = "When solution = 'all', do not supply include.",
        std = "When `solution = \"all\"`, do not supply `include`."
      ))
    }

    monitored <- c("conservative", "parsimonious")
    if (!is.null(dir.exp)) {
      monitored <- c(monitored, "intermediate")
    }

    return(list(
      solution = solution,
      include = NULL,
      monitored = monitored
    ))
  }

  if (solution == "conservative") {
    if (!is.null(include) && !identical(include, "")) {
      stop(.solution_message(
        style,
        plain = "When solution = 'conservative'/'con', include must be ''.",
        std = "When `solution = \"conservative\"`/`\"con\"`, `include` must be ''."
      ))
    }
    if (!is.null(dir.exp)) {
      stop(.solution_message(
        style,
        plain = "When solution = 'conservative'/'con', dir.exp must be NULL.",
        std = "When `solution = \"conservative\"`/`\"con\"`, `dir.exp` must be NULL."
      ))
    }

    return(list(
      solution = solution,
      include = "",
      monitored = "conservative"
    ))
  }

  if (solution == "parsimonious") {
    if (!is.null(include) && !identical(include, "?")) {
      stop(.solution_message(
        style,
        plain = "When solution = 'parsimonious'/'par', include must be '?'.",
        std = "When `solution = \"parsimonious\"`/`\"par\"`, `include` must be '?'."
      ))
    }
    if (!is.null(dir.exp)) {
      stop(.solution_message(
        style,
        plain = "When solution = 'parsimonious'/'par', dir.exp must be NULL.",
        std = "When `solution = \"parsimonious\"`/`\"par\"`, `dir.exp` must be NULL."
      ))
    }

    return(list(
      solution = solution,
      include = "?",
      monitored = "parsimonious"
    ))
  }

  if (!is.null(include) && !identical(include, "?")) {
    stop(.solution_message(
      style,
      plain = "When solution = 'intermediate'/'int', include must be '?'.",
      std = "When `solution = \"intermediate\"`/`\"int\"`, `include` must be '?'."
    ))
  }
  if (is.null(dir.exp)) {
    stop(.solution_message(
      style,
      plain = "When solution = 'intermediate'/'int', dir.exp must be provided.",
      std = "When `solution = \"intermediate\"`/`\"int\"`, `dir.exp` must be provided."
    ))
  }

  list(
    solution = solution,
    include = "?",
    monitored = "intermediate"
  )
}

test_that("cluster solution = all follows family rules without dir.exp", {
  out <- qcaERT:::.resolve_solution_controls(
    solution = "all",
    include = NULL,
    dir.exp = NULL,
    caller = "cluster.test",
    style = "std"
  )

  expect_identical(out$solution, "all")
  expect_null(out$include)
  expect_identical(out$monitored, c("conservative", "parsimonious"))
})

test_that("cluster solution = all includes intermediate only when dir.exp is supplied", {
  out <- qcaERT:::.resolve_solution_controls(
    solution = "all",
    include = NULL,
    dir.exp = c(1, 1),
    caller = "cluster.test",
    style = "std"
  )

  expect_identical(out$solution, "all")
  expect_null(out$include)
  expect_identical(
    out$monitored,
    c("conservative", "parsimonious", "intermediate")
  )
})

test_that("cluster solution = all now rejects supplied include", {
  expect_error(
    qcaERT:::.resolve_solution_controls(
      solution = "all",
      include = "?",
      dir.exp = c(1, 1),
      caller = "cluster.test",
      style = "std"
    ),
    "When `solution = \"all\"`, do not supply `include`.",
    fixed = TRUE
  )
})

test_that("cluster solution = NULL is now rejected", {
  expect_error(
    qcaERT:::.resolve_solution_controls(
      solution = NULL,
      include = NULL,
      dir.exp = NULL,
      caller = "cluster.test",
      style = "std"
    ),
    "`solution` must be one of 'all', 'con'/'conservative', 'par'/'parsimonious', or 'int'/'intermediate'.",
    fixed = TRUE
  )
})

test_that("cluster.test now defaults to solution = all", {
  expect_identical(
    formals(cluster.test)$solution,
    "all"
  )
})

.normalize_calib_type <- function(type, name = "type") {
  if (!is.character(type) || length(type) != 1L || is.na(type)) {
    stop("`", name, "` must be 'f', 'fuzzy', 'c', or 'crisp'.")
  }

  type <- tolower(trimws(type))
  if (type %in% c("f", "fuzzy")) return("f")
  if (type %in% c("c", "crisp")) return("c")

  stop("`", name, "` must be 'f', 'fuzzy', 'c', or 'crisp'.")
}

.qca_calib_type <- function(type_fc) {
  if (identical(type_fc, "c")) "crisp" else "fuzzy"
}

.normalize_calib_method <- function(method, type_fc, caller) {
  if (identical(type_fc, "c")) return("direct")
  if (is.null(method)) method <- "direct"
  if (!is.character(method) || length(method) != 1L || is.na(method)) {
    stop("`method` in calibration specifications must be 'direct' or 'indirect'.")
  }

  method <- tolower(trimws(method))
  if (!method %in% c("direct", "indirect")) {
    stop(
      caller,
      " supports threshold perturbation for QCA::calibrate() methods ",
      "'direct' and 'indirect'."
    )
  }

  method
}

.as_numeric_thresholds <- function(x, label) {
  nm <- names(x)
  out <- suppressWarnings(as.numeric(x))
  if (length(out) != length(x) || anyNA(out) || any(!is.finite(out))) {
    stop(label, " must contain finite numeric thresholds.")
  }
  names(out) <- nm
  out
}

.named_thresholds_are <- function(nms, expected) {
  if (is.null(nms) || any(!nzchar(nms))) return(FALSE)
  nms <- tolower(nms)
  length(nms) == length(expected) &&
    length(unique(nms)) == length(expected) &&
    all(expected %in% nms)
}

.normalize_calib_thresholds <- function(thresholds, type_fc, method, condition, caller) {
  label <- paste0("`calib_spec[['", condition, "']]$thresholds`")
  if (is.null(thresholds)) {
    stop(label, " must be supplied.")
  }

  thresholds <- .as_numeric_thresholds(thresholds, label)
  nms <- names(thresholds)

  if (identical(type_fc, "c")) {
    if (length(thresholds) != 1L) {
      stop(label, " must contain one threshold for crisp calibration.")
    }
    names(thresholds) <- "T"
    return(thresholds)
  }

  if (identical(method, "direct")) {
    if (!length(thresholds) %in% c(3L, 6L)) {
      stop(label, " must contain three or six thresholds for fuzzy direct calibration.")
    }

    if (length(thresholds) == 3L) {
      expected <- c("e", "c", "i")
      if (.named_thresholds_are(nms, expected)) {
        thresholds <- thresholds[match(expected, tolower(nms))]
      } else if (!is.null(nms) && any(nzchar(nms))) {
        stop(label, " names must be E, C, and I for fuzzy direct three-threshold calibration.")
      }
      names(thresholds) <- c("E", "C", "I")
    } else {
      expected <- c("e1", "c1", "i1", "i2", "c2", "e2")
      if (.named_thresholds_are(nms, expected)) {
        thresholds <- thresholds[match(expected, tolower(nms))]
      } else if (!is.null(nms) && any(nzchar(nms))) {
        stop(
          label,
          " names must be E1, C1, I1, I2, C2, and E2 ",
          "for fuzzy direct six-threshold calibration."
        )
      }
      names(thresholds) <- c("E1", "C1", "I1", "I2", "C2", "E2")
    }
  } else {
    if (length(thresholds) < 1L) {
      stop(label, " must contain at least one threshold for fuzzy indirect calibration.")
    }
    thresholds <- sort(unname(thresholds))
    names(thresholds) <- paste0("T", seq_along(thresholds))
  }

  if (!.valid_calib_thresholds(thresholds, type_fc, method)) {
    stop(label, " are not ordered consistently with ", caller, " calibration method '", method, "'.")
  }

  thresholds
}

.valid_calib_thresholds <- function(thresholds, type_fc, method) {
  thresholds <- as.numeric(thresholds)
  if (any(!is.finite(thresholds))) return(FALSE)

  if (identical(type_fc, "c")) {
    return(length(thresholds) == 1L)
  }

  if (identical(method, "indirect")) {
    return(length(thresholds) >= 1L && all(diff(thresholds) > 0))
  }

  if (length(thresholds) == 3L) {
    return(
      (thresholds[1] < thresholds[2] && thresholds[2] < thresholds[3]) ||
        (thresholds[3] < thresholds[2] && thresholds[2] < thresholds[1])
    )
  }

  if (length(thresholds) == 6L) {
    e1 <- thresholds[1]
    c1 <- thresholds[2]
    i1 <- thresholds[3]
    i2 <- thresholds[4]
    c2 <- thresholds[5]
    e2 <- thresholds[6]

    increasing <- e1 < c1 && c1 < i1 && i1 <= i2 && i2 < c2 && c2 < e2
    decreasing <- i1 < c1 && c1 < e1 && e1 <= e2 && e2 < c2 && c2 < i2
    return(increasing || decreasing)
  }

  FALSE
}

.normalize_calibrate_options <- function(calibrate, caller) {
  if (is.null(calibrate)) return(list())
  if (!is.list(calibrate)) {
    stop("`calibrate` must be a named list of additional QCA::calibrate() arguments.")
  }
  if (length(calibrate) == 0L) return(calibrate)

  nms <- names(calibrate)
  if (is.null(nms) || any(!nzchar(nms))) {
    stop("`calibrate` must be a named list.")
  }

  reserved <- c("x", "type", "thresholds")
  used <- intersect(nms, reserved)
  if (length(used) > 0L) {
    stop(
      "`calibrate` must not contain ",
      paste0("`", used, "`", collapse = ", "),
      "; ",
      caller,
      " supplies those arguments from the calibration specification."
    )
  }

  calibrate
}

.normalize_test_outcome <- function(test.outcome) {
  if (!is.logical(test.outcome) || length(test.outcome) != 1L || is.na(test.outcome)) {
    stop("`test.outcome` must be a single TRUE/FALSE value.")
  }
  isTRUE(test.outcome)
}

.normalize_test_conditions <- function(test.conditions, conditions, test.outcome = FALSE) {
  test.outcome <- .normalize_test_outcome(test.outcome)

  if (is.null(test.conditions)) {
    if (!test.outcome) {
      stop("`test.conditions` can be NULL only when `test.outcome = TRUE`.")
    }
    return(character(0))
  }

  if (!is.character(test.conditions) || length(test.conditions) < 1L) {
    stop("`test.conditions` must be NULL or a non-empty character vector.")
  }

  test.conditions <- unique(trimws(test.conditions))
  test.conditions <- test.conditions[nzchar(test.conditions)]
  if (length(test.conditions) < 1L) {
    stop("`test.conditions` must include at least one non-empty condition name, or be NULL when `test.outcome = TRUE`.")
  }
  if (!all(test.conditions %in% conditions)) {
    stop("All `test.conditions` must be contained in `conditions`.")
  }

  test.conditions
}

.calib_spec_targets <- function(conditions, outcome = NULL, test.outcome = FALSE) {
  test.outcome <- .normalize_test_outcome(test.outcome)

  if (!test.outcome) return(conditions)
  if (is.null(outcome) || !is.character(outcome) || length(outcome) != 1L || !nzchar(outcome)) {
    stop("`outcome` must be supplied when `test.outcome = TRUE`.")
  }

  .validate_outcome_conditions_distinct(outcome, conditions)
  c(conditions, outcome)
}

.normalize_calib_specs <- function(
    conditions,
    outcome = NULL,
    calib_spec,
    test.outcome = FALSE,
    caller = "qcaERT"
) {
  if (!is.character(conditions) || length(conditions) < 1L) {
    stop("`conditions` must be a non-empty character vector.")
  }

  test.outcome <- .normalize_test_outcome(test.outcome)
  spec_targets <- .calib_spec_targets(
    conditions = conditions,
    outcome = outcome,
    test.outcome = test.outcome
  )
  if (is.null(calib_spec)) {
    stop("`calib_spec` must be supplied.")
  }
  if (!is.list(calib_spec)) {
    stop("`calib_spec` must be a named list.")
  }
  if (is.null(names(calib_spec)) || any(!nzchar(names(calib_spec)))) {
    if (test.outcome) {
      stop("`calib_spec` must be named by `conditions` plus `outcome` when `test.outcome = TRUE`.")
    }
    stop("`calib_spec` must be named by `conditions`.")
  }
  if (!setequal(names(calib_spec), spec_targets) || length(calib_spec) != length(spec_targets)) {
    if (test.outcome) {
      stop("`calib_spec` must contain exactly one entry for each condition and the outcome when `test.outcome = TRUE`.")
    }
    stop("`calib_spec` must contain exactly one entry for each condition.")
  }

  calib_spec <- calib_spec[spec_targets]
  out <- setNames(vector("list", length(spec_targets)), spec_targets)

  for (cond in spec_targets) {
    spec <- calib_spec[[cond]]
    if (!is.list(spec)) {
      stop("`calib_spec[['", cond, "']]` must be a list.")
    }
    if (is.null(spec$raw) || !is.character(spec$raw) || length(spec$raw) != 1L || !nzchar(spec$raw)) {
      stop("`calib_spec[['", cond, "']]$raw` must be a single non-empty character string.")
    }
    if (is.null(spec$type)) {
      stop("`calib_spec[['", cond, "']]$type` must be supplied.")
    }

    type_fc <- .normalize_calib_type(spec$type, paste0("calib_spec[['", cond, "']]$type"))
    cal_extra <- if (is.null(spec$calibrate)) list() else spec$calibrate
    cal_extra <- .normalize_calibrate_options(cal_extra, caller)
    method <- spec$method
    if (is.null(method) && "method" %in% names(cal_extra)) {
      method <- cal_extra$method
      cal_extra$method <- NULL
    }
    method <- .normalize_calib_method(method, type_fc, caller)

    out[[cond]] <- list(
      raw = spec$raw,
      type = type_fc,
      qca_type = .qca_calib_type(type_fc),
      method = method,
      thresholds = .normalize_calib_thresholds(spec$thresholds, type_fc, method, cond, caller),
      calibrate = cal_extra
    )
  }

  out
}

.calib_anchor_labels <- function(spec) {
  names(spec$thresholds)
}

.normalize_calib_anchors_to_test <- function(anchors_to_test) {
  if (is.null(anchors_to_test)) return(NULL)
  if (!is.character(anchors_to_test) || length(anchors_to_test) < 1L) {
    stop("`anchors_to_test` must be NULL or a non-empty character vector.")
  }
  toupper(trimws(anchors_to_test))
}

.effective_calib_anchors <- function(spec, anchors_to_test) {
  anchors <- .calib_anchor_labels(spec)
  if (is.null(anchors_to_test)) return(anchors)
  intersect(anchors_to_test, anchors)
}

.calib_anchor_value <- function(spec, anchor) {
  anchor <- toupper(anchor)
  idx <- match(anchor, .calib_anchor_labels(spec))
  if (is.na(idx)) {
    stop("Unknown calibration anchor '", anchor, "'.")
  }
  unname(spec$thresholds[idx])
}

.replace_calib_anchor <- function(spec, anchor, value, validate = TRUE) {
  anchor <- toupper(anchor)
  idx <- match(anchor, .calib_anchor_labels(spec))
  if (is.na(idx)) {
    stop("Unknown calibration anchor '", anchor, "'.")
  }
  spec$thresholds[idx] <- value

  if (isTRUE(validate) && !.valid_calib_thresholds(spec$thresholds, spec$type, spec$method)) {
    stop("Replacing calibration anchor '", anchor, "' produced invalid thresholds.")
  }

  spec
}

.is_calib_anchor_feasible <- function(spec, anchor, value, direction, xmin, xmax) {
  direction <- match.arg(direction, c("lower", "upper"))
  if (!is.finite(value)) return(FALSE)
  if (is.finite(xmin) && value < xmin) return(FALSE)
  if (is.finite(xmax) && value > xmax) return(FALSE)

  spec_try <- .replace_calib_anchor(spec, anchor, value, validate = FALSE)
  .valid_calib_thresholds(spec_try$thresholds, spec_try$type, spec_try$method)
}

.calib_zero_delta_steps <- function(spec) {
  stats::setNames(rep(0L, length(.calib_anchor_labels(spec))), .calib_anchor_labels(spec))
}

.calibrate_from_spec <- function(raw_x, spec) {
  args <- list(
    x = raw_x,
    type = spec$qca_type,
    thresholds = unname(spec$thresholds)
  )

  if (identical(spec$type, "f")) {
    args$method <- spec$method
  }

  do.call(QCA::calibrate, c(args, spec$calibrate))
}

.prepare_calib_context <- function(
    conditions,
    outcome = NULL,
    calib_spec,
    test.outcome = FALSE,
    raw.data,
    unit_step,
    unit_step_divisor,
    anchors_to_test,
    caller
) {
  calib_specs <- .normalize_calib_specs(
    conditions = conditions,
    outcome = outcome,
    calib_spec = calib_spec,
    test.outcome = test.outcome,
    caller = caller
  )

  raw_sources <- vapply(calib_specs, function(x) x$raw, character(1))
  if (!all(raw_sources %in% colnames(raw.data))) {
    stop("All raw columns referenced by the calibration specification must exist in `raw.data`.")
  }

  spec_targets <- names(calib_specs)
  condition_specs <- calib_specs[conditions]
  outcome_spec <- if (isTRUE(test.outcome)) calib_specs[[outcome]] else NULL

  unit_step_targets <- .coerce_unit_step(
    unit_step = unit_step,
    conds = spec_targets,
    raw.data = raw.data,
    unit_step_divisor = unit_step_divisor,
    calib_specs = calib_specs
  )

  list(
    calib_specs = calib_specs,
    condition_specs = condition_specs,
    outcome_spec = outcome_spec,
    spec_targets = spec_targets,
    raw_sources = raw_sources[conditions],
    raw_targets = raw_sources,
    type_fc = vapply(condition_specs, function(x) x$type, character(1)),
    type_targets = vapply(calib_specs, function(x) x$type, character(1)),
    thresholds = lapply(condition_specs, function(x) x$thresholds),
    thresholds_targets = lapply(calib_specs, function(x) x$thresholds),
    unit_step = unit_step_targets[conditions],
    unit_step_targets = unit_step_targets,
    anchors_to_test = .normalize_calib_anchors_to_test(anchors_to_test)
  )
}

.calib_apply_specs_to_data <- function(data, raw.data, calib_specs, conditions) {
  data_out <- data

  for (cond in conditions) {
    spec <- calib_specs[[cond]]
    data_out[[cond]] <- .calibrate_from_spec(
      raw_x = raw.data[[spec$raw]],
      spec = spec
    )
  }

  data_out
}

.calib_thresholds_in_range <- function(spec, xmin, xmax) {
  vals <- as.numeric(spec$thresholds)
  all(is.finite(vals)) && all(vals >= xmin) && all(vals <= xmax)
}

.altset_build_calib_candidate <- function(
    set_name,
    role,
    calib_specs,
    raw.data,
    unit_step_vec,
    anchors_to_test,
    calib_max_steps
) {
  spec0 <- calib_specs[[set_name]]
  raw_name <- spec0$raw
  t0 <- spec0$thresholds
  step_i <- unit_step_vec[[set_name]]

  anchors_all <- .calib_anchor_labels(spec0)
  anchors_eff <- .effective_calib_anchors(spec0, anchors_to_test)

  if (length(anchors_eff) == 0L) {
    stop(
      "`anchors_to_test` did not match any eligible calibration anchors for set '",
      set_name,
      "'. Eligible anchors are: ",
      paste(anchors_all, collapse = ", "),
      "."
    )
  }

  x_raw <- raw.data[[raw_name]]
  xmin <- suppressWarnings(min(x_raw, na.rm = TRUE))
  xmax <- suppressWarnings(max(x_raw, na.rm = TRUE))

  if (!is.finite(xmin) || !is.finite(xmax)) {
    stop("Raw column '", raw_name, "' must have finite min and max.")
  }

  if (!.calib_thresholds_in_range(spec0, xmin, xmax)) {
    stop("Baseline thresholds for set '", set_name, "' must lie within the raw-data range.")
  }

  delta_ranges <- stats::setNames(vector("list", length(anchors_all)), anchors_all)
  for (anchor in anchors_all) {
    if (!anchor %in% anchors_eff) {
      delta_ranges[[anchor]] <- 0L
      next
    }

    baseline_value <- .calib_anchor_value(spec0, anchor)
    lower_by_range <- ceiling((xmin - baseline_value) / step_i)
    upper_by_range <- floor((xmax - baseline_value) / step_i)
    lower <- max(-calib_max_steps, lower_by_range)
    upper <- min(calib_max_steps, upper_by_range)

    if (!is.finite(lower) || !is.finite(upper) || lower > upper) {
      stop("No admissible threshold moves were found for anchor '", anchor, "' in set '", set_name, "'.")
    }

    delta_ranges[[anchor]] <- seq.int(as.integer(lower), as.integer(upper))
  }

  list(
    set = set_name,
    role = role,
    raw = raw_name,
    type = spec0$type,
    method = spec0$method,
    baseline = t0,
    anchors = anchors_all,
    sampled_anchors = anchors_eff,
    delta_ranges = delta_ranges,
    unit_step = step_i,
    xmin = xmin,
    xmax = xmax,
    spec = spec0,
    max_attempts = max(1000L, 200L * length(anchors_eff))
  )
}

.altset_sample_calib_candidate <- function(info) {
  for (attempt in seq_len(info$max_attempts)) {
    delta_steps <- stats::setNames(integer(length(info$anchors)), info$anchors)

    for (anchor in info$sampled_anchors) {
      choices <- info$delta_ranges[[anchor]]
      delta_steps[anchor] <- if (length(choices) == 1L) choices else sample(choices, size = 1L)
    }

    spec_new <- info$spec
    for (anchor in info$anchors) {
      delta <- delta_steps[anchor]
      if (delta == 0L) next
      next_value <- .calib_anchor_value(info$spec, anchor) + delta * info$unit_step
      spec_new <- .replace_calib_anchor(spec_new, anchor, next_value, validate = FALSE)
    }

    if (
      .calib_thresholds_in_range(spec_new, info$xmin, info$xmax) &&
      .valid_calib_thresholds(spec_new$thresholds, spec_new$type, spec_new$method)
    ) {
      return(list(
        spec = spec_new,
        delta_steps = delta_steps,
        changed = any(delta_steps != 0L)
      ))
    }
  }

  stop(
    "Could not generate an admissible random calibration candidate for set '",
    info$set,
    "' after ",
    info$max_attempts,
    " attempts. Consider reducing `calib_max_steps`, reducing `unit_step`, ",
    "or narrowing `anchors_to_test`."
  )
}

.altset_sample_calibration_draw <- function(candidate_map, calib.data, raw.data, thr_list) {
  data_step <- calib.data
  thresholds_step <- thr_list
  calib_changes <- list()
  any_changed <- FALSE

  for (set_name in names(candidate_map)) {
    info <- candidate_map[[set_name]]
    sampled <- .altset_sample_calib_candidate(info)

    thr_new <- sampled$spec$thresholds
    thr_old <- info$baseline
    delta_steps <- sampled$delta_steps
    changed <- sampled$changed

    thresholds_step[[set_name]] <- thr_new
    data_step[[set_name]] <- .calibrate_from_spec(raw.data[[info$raw]], sampled$spec)
    calib_changes[[set_name]] <- list(
      set = info$set,
      role = info$role,
      raw = info$raw,
      type = info$type,
      method = info$method,
      thresholds_baseline = thr_old,
      thresholds_draw = thr_new,
      delta_steps = delta_steps,
      changed = changed
    )
    any_changed <- any_changed || changed
  }

  list(
    data = data_step,
    thresholds = thresholds_step,
    changes = calib_changes,
    changed = any_changed
  )
}

test_that("test.conditions handling allows outcome-only calibration tests", {
  expect_identical(
    .normalize_test_conditions(NULL, c("A", "B"), test.outcome = TRUE),
    character(0)
  )

  expect_identical(
    .normalize_test_conditions(c(" A ", "A", "B"), c("A", "B"), test.outcome = FALSE),
    c("A", "B")
  )

  expect_error(
    .normalize_test_conditions(NULL, c("A", "B"), test.outcome = FALSE),
    "`test.conditions` can be NULL only when `test.outcome = TRUE`.",
    fixed = TRUE
  )

  expect_error(
    .normalize_test_conditions("Y", c("A", "B"), test.outcome = TRUE),
    "All `test.conditions` must be contained in `conditions`.",
    fixed = TRUE
  )
})

test_that("calibration specs can include the outcome only when requested", {
  spec <- qcaert_schema_calib6_spec()
  spec$Y <- list(
    raw = "Y_raw",
    type = "fuzzy",
    method = "direct",
    thresholds = c(10, 20, 30, 70, 80, 90),
    calibrate = list(logistic = FALSE)
  )

  normalized <- .normalize_calib_specs(
    conditions = c("A", "B"),
    outcome = "Y",
    calib_spec = spec,
    test.outcome = TRUE
  )

  expect_identical(names(normalized), c("A", "B", "Y"))
  expect_identical(normalized$Y$raw, "Y_raw")
  expect_identical(normalized$Y$method, "direct")

  expect_error(
    .normalize_calib_specs(
      conditions = c("A", "B"),
      outcome = "Y",
      calib_spec = qcaert_schema_calib6_spec(),
      test.outcome = TRUE
    ),
    "`calib_spec` must contain exactly one entry for each condition and the outcome when `test.outcome = TRUE`.",
    fixed = TRUE
  )

  expect_error(
    .normalize_calib_specs(
      conditions = c("A", "B"),
      outcome = "Y",
      calib_spec = spec
    ),
    "`calib_spec` must contain exactly one entry for each condition.",
    fixed = TRUE
  )

  expect_error(
    .normalize_calib_specs(
      conditions = c("A", "B"),
      outcome = "Y",
      calib_spec = NULL,
      test.outcome = FALSE
    ),
    "`calib_spec` must be supplied.",
    fixed = TRUE
  )
})

test_that("calibration context separates condition specs from outcome spec", {
  raw <- qcaert_schema_raw()
  raw$Y_raw <- raw$A_raw
  spec <- qcaert_schema_calib6_spec()
  spec$Y <- list(
    raw = "Y_raw",
    type = "fuzzy",
    method = "direct",
    thresholds = c(10, 20, 30, 70, 80, 90),
    calibrate = list(logistic = FALSE)
  )

  context <- .prepare_calib_context(
    conditions = c("A", "B"),
    outcome = "Y",
    calib_spec = spec,
    test.outcome = TRUE,
    raw.data = raw,
    unit_step = 1,
    unit_step_divisor = NULL,
    anchors_to_test = NULL,
    caller = "test"
  )

  expect_identical(names(context$calib_specs), c("A", "B", "Y"))
  expect_identical(names(context$condition_specs), c("A", "B"))
  expect_identical(context$outcome_spec$raw, "Y_raw")
  expect_identical(names(context$thresholds), c("A", "B"))
  expect_identical(names(context$thresholds_targets), c("A", "B", "Y"))
})

#' Calibration-threshold robustness test for QCA solutions
#'
#' Perturbs calibration thresholds for selected conditions, and optionally the
#' outcome, one anchor and one direction at a time, and checks when the
#' monitored QCA solution changes.
#' For each tested path, the function records the starting threshold, the last
#' value that preserved the baseline solution, the first value that changed the
#' solution or triggered an error, and the reason the path stopped.
#'
#' @param raw.data A data frame object containing the raw, uncalibrated
#'   condition columns.
#' @param calib.data A data frame object containing the calibrated outcome
#'   and calibrated condition columns used for the analysis.
#' @param outcome Name of the outcome column in `calib.data`. This can be
#'   supplied as a character string or as an unquoted name.
#' @param conditions Character vector of calibrated condition names in
#'   `calib.data`. This can be supplied as a character vector or as unquoted
#'   names.
#' @param calib_spec Named list of calibration specifications. When
#'   `test.outcome = FALSE`, it must contain one entry per condition. When
#'   `test.outcome = TRUE`, it must contain one entry per condition plus one
#'   entry named by `outcome`. Each entry must contain `raw`, `type`, and
#'   `thresholds`, may contain `method`, and may contain `calibrate`, a named
#'   list of additional [QCA::calibrate()] arguments.
#' @param test.conditions Subset of `conditions` to perturb. Use `NULL` only
#'   when `test.outcome = TRUE` and no condition calibration should be tested.
#' @param test.outcome Logical; if `TRUE`, also perturb the outcome
#'   calibration while keeping the outcome out of `conditions`.
#' @param anchors_to_test Character vector of anchors to test, or `NULL` to
#'   test all anchors implied by each set's calibration method. Crisp
#'   sets use `"T"`. Fuzzy direct three-threshold sets use `"E"`,
#'   `"C"`, and `"I"`. Fuzzy direct six-threshold sets use `"E1"`,
#'   `"C1"`, `"I1"`, `"I2"`, `"C2"`, and `"E2"`. Fuzzy indirect sets
#'   use `"T1"`, `"T2"`, and so on.
#' @param solution Solution type to monitor. Accepted values are `"all"`,
#'   `"con"` or `"conservative"`, `"par"` or `"parsimonious"`, and `"int"` or
#'   `"intermediate"`.
#' @param include Optional minimization include setting. Currently, this
#'   argument accepts only `NULL`, `""`, or `"?"`.
#' @param which_M Positive integer giving which solution alternative to use
#'   when minimization returns multiple models.
#' @param unit_step Numeric step size used to move thresholds. Supply either
#'   one value applied to all tested sets, one value per condition when
#'   `test.outcome = FALSE`, or one value per `c(conditions, outcome)` when
#'   `test.outcome = TRUE`. If `NULL`, qcaERT computes scale-aware steps from
#'   each set's threshold spacing.
#' @param unit_step_divisor Positive number used to compute `unit_step`
#'   automatically when `unit_step = NULL`. The default is `10`.
#' @param max_steps Maximum number of upward or downward threshold moves to
#'   attempt for each tested anchor.
#' @param incl.cut Inclusion cutoff passed to [QCA::truthTable()].
#' @param n.cut Frequency cutoff passed to [QCA::truthTable()].
#' @param dir.exp Directional expectations used when the monitored solution is
#'   intermediate.
#' @param i_mode Character string controlling intermediate-solution selection.
#'   Accepted values are `"all"` and `"C1P1"`.
#' @param exclude_mode Character string controlling how excluded rows are
#'   handled for parsimonious and intermediate minimization. `"recompute"`
#'   recalculates exclusions from each perturbed truth table using
#'   `exclude_recompute`, `"static"` reuses `exclude_static`, and `"none"`
#'   does not use exclusions.
#' @param exclude_recompute Named list of arguments passed to [QCA::findRows()]
#'   when `exclude_mode = "recompute"`.
#' @param exclude_static Already computed exclusion object reused when
#'   `exclude_mode = "static"`.
#' @param result_shape Layout of the clean `results` table when
#'   `solution = "all"`. `"wide"` keeps one row per tested
#'   set-anchor-direction path with solution-type-specific columns such as
#'   `con_last_safe` and `par_reason`. `"long"` returns one row per tested
#'   set-anchor-direction path and solution type, with a `solution_type` column.
#'   Single-solution-type calls keep the compact single-solution-type layout.
#' @param progress Logical; if `TRUE` and the session is interactive, show a
#'   text progress bar.
#' @param x A `calib_test` object returned by [calib.test()].
#' @param row.names Logical; passed to [print.data.frame()] by
#'   [print.calib_test()].
#' @param ... Additional arguments routed through the QCA workflow. Arguments
#'   matching [QCA::truthTable()] are forwarded to truth table construction;
#'   remaining minimization arguments are filtered and forwarded to
#'   [QCA::minimize()]. The function also looks in `...` for `include`,
#'   `dir.exp`, or `direxp` if those arguments were not supplied explicitly.
#'   In [print.calib_test()], `...` is passed to [print.data.frame()]. In
#'   [as.data.frame.calib_test()], `...` is ignored.
#'
#' @returns An object of class `calib_test` with the following components:
#'   \describe{
#'     \item{`diagnostics`}{A detailed data frame with one row per tested
#'     anchor-direction path. When `solution = "all"`, each monitored solution
#'     type is searched independently, so diagnostics has one row per tested
#'     anchor, direction, and solution type. It includes the starting threshold, the
#'     last safe value, the first failing value, the number of successful
#'     steps, stop reason, and change classification.}
#'     \item{`results`}{A compact data frame with the columns `set`, `role`,
#'     `raw`, `type`, `method`, `anchor`, `direction`, `start`,
#'     `last_safe`, `first_failing`, `step_unit`, `steps`, `total_delta`,
#'     `pct_raw_range`, and `reason`. When `solution = "all"` and
#'     `result_shape = "wide"`, the row unit remains set-anchor-direction,
#'     but the boundary and reason columns are solution-type-specific, using prefixes
#'     `con_`, `par_`, and, when available, `int_`. When
#'     `result_shape = "long"`, `results` has one row per tested path and
#'     solution type.}
#'     \item{`bounds`}{A named list of compact lower/upper bound matrices, one
#'     per tested set. When `solution = "all"`, each tested set contains a
#'     named list of solution-type-specific lower/upper matrices.}
#'     \item{`baseline`}{A list containing the baseline analysis status,
#'     selected solution terms used for comparison, metadata, and
#'     solution-type-specific baseline objects.}
#'     \item{`by_set`}{A named list containing per-set step results, including
#'     path-level traces.}
#'     \item{`settings`}{A list containing the analysis settings used to build
#'     the result object.}
#'   }
#'
#'   `print.calib_test()` prints a concise summary and a compact display table.
#'   In that printed table, `set` and `role` are omitted and the raw source
#'   column is labelled `condition`. `as.data.frame.calib_test()` returns the
#'   stored `results` table.
#'   If `ggplot2` is installed, `plot.calib_test()` provides interval, heatmap,
#'   and trace views; see `?qcaERT_plots`.
#'
#' @details
#' The function rebuilds calibrated condition columns, and the outcome when
#' `test.outcome = TRUE`, from `raw.data` using `calib_spec`. It then compares
#' perturbed analyses against the baseline solution under the selected solution
#' type.
#'
#' When `solution = "all"`, conservative, parsimonious, and, when requested,
#' intermediate paths are searched independently for each tested set, anchor,
#' and direction.
#'
#' Each tested path starts from the supplied threshold for one method-specific
#' anchor of one set and moves in one direction at a time (`"lower"` or
#' `"upper"`), using `unit_step` until one of four things happens:
#' \itemize{
#'   \item the monitored solution changes,
#'   \item minimization fails,
#'   \item the raw-data range or anchor ordering blocks the next move, or
#'   \item `max_steps` is reached.
#' }
#'
#' The shared solution-control, exclusion, calibration-specification, and
#' returned-object conventions are described in `?qcaERT_conventions`.
#'
#' @examples
#' \donttest{
#' library(QCA)
#' library(ggplot2)
#' data(LR)
#'
#' conditions <- c("DEV", "URB", "LIT", "IND", "STB")
#' outcome <- "SURV"
#' dir_exp <- rep("1", length(conditions))
#'
#' thresholds <- list(
#'   DEV = findTh(LR$DEV, groups = 7),
#'   URB = findTh(LR$URB, groups = 4),
#'   LIT = findTh(LR$LIT, groups = 4),
#'   IND = findTh(LR$IND, groups = 4),
#'   STB = findTh(LR$STB, groups = 4),
#'   SURV = findTh(LR$SURV, groups = 4)
#' )
#'
#' dat <- LR
#' dat$DEV <- calibrate(LR$DEV, type = "fuzzy", thresholds = thresholds$DEV)
#' dat$URB <- calibrate(LR$URB, type = "fuzzy", thresholds = thresholds$URB)
#' dat$LIT <- calibrate(LR$LIT, type = "fuzzy", thresholds = thresholds$LIT)
#' dat$IND <- calibrate(LR$IND, type = "fuzzy", thresholds = thresholds$IND)
#' dat$STB <- calibrate(LR$STB, type = "fuzzy", thresholds = thresholds$STB)
#' dat$SURV <- calibrate(LR$SURV, type = "fuzzy", thresholds = thresholds$SURV)
#'
#' calib_spec <- list(
#'   DEV = list(raw = "DEV", type = "fuzzy", method = "direct", thresholds = thresholds$DEV),
#'   URB = list(raw = "URB", type = "fuzzy", method = "direct", thresholds = thresholds$URB),
#'   LIT = list(raw = "LIT", type = "fuzzy", method = "direct", thresholds = thresholds$LIT),
#'   IND = list(raw = "IND", type = "fuzzy", method = "direct", thresholds = thresholds$IND),
#'   STB = list(raw = "STB", type = "fuzzy", method = "direct", thresholds = thresholds$STB)
#' )
#'
#' calib_spec_outcome <- calib_spec
#' calib_spec_outcome$SURV <- list(
#'   raw = "SURV",
#'   type = "fuzzy",
#'   method = "direct",
#'   thresholds = thresholds$SURV
#' )
#'
#' # Common use: test selected calibrated conditions with scale-aware steps.
#' calib_out <- calib.test(
#'   raw.data = LR,
#'   calib.data = dat,
#'   outcome = outcome,
#'   conditions = conditions,
#'   calib_spec = calib_spec,
#'   test.conditions = c("DEV", "URB"),
#'   unit_step = NULL,
#'   unit_step_divisor = 10,
#'   max_steps = 5,
#'   incl.cut = 0.8,
#'   n.cut = 1,
#'   solution = "all",
#'   dir.exp = dir_exp,
#'   progress = TRUE
#' )
#'
#' calib_out
#' as.data.frame(calib_out)
#' calib_out$bounds
#' calib_out$diagnostics
#'
#' plot(calib_out, solution_type = "conservative")
#' plot(calib_out, solution_type = "conservative", type = "heatmap")
#'
#' # Special case: test the calibrated outcome without putting it in
#' # conditions.
#' outcome_out <- calib.test(
#'   raw.data = LR,
#'   calib.data = dat,
#'   outcome = outcome,
#'   conditions = conditions,
#'   calib_spec = calib_spec_outcome,
#'   test.conditions = NULL,
#'   test.outcome = TRUE,
#'   unit_step = NULL,
#'   unit_step_divisor = 10,
#'   max_steps = 5,
#'   incl.cut = 0.8,
#'   n.cut = 1,
#'   solution = "all",
#'   dir.exp = dir_exp,
#'   progress = TRUE
#' )
#'
#' outcome_out
#' as.data.frame(outcome_out)
#' plot(outcome_out, solution_type = "conservative")
#'
#' # Special case: focus on selected anchors in a six-threshold direct
#' # calibration. For DEV, anchors are E1, C1, I1, I2, C2, and E2.
#' dev_anchor_out <- calib.test(
#'   raw.data = LR,
#'   calib.data = dat,
#'   outcome = outcome,
#'   conditions = conditions,
#'   calib_spec = calib_spec,
#'   test.conditions = "DEV",
#'   anchors_to_test = c("E1", "C1", "I1"),
#'   unit_step = NULL,
#'   unit_step_divisor = 10,
#'   max_steps = 5,
#'   incl.cut = 0.8,
#'   n.cut = 1,
#'   solution = "all",
#'   dir.exp = dir_exp,
#'   progress = TRUE
#' )
#'
#' as.data.frame(dev_anchor_out)
#'
#' plot(
#'   dev_anchor_out,
#'   solution_type = "conservative",
#'   type = "trace",
#'   set = "DEV",
#'   anchor = "E1",
#'   direction = "lower"
#' )
#'
#' # Special case: indirect calibration. Indirect anchors are positional:
#' # T1, T2, and so on.
#' thresholds_indirect <- thresholds
#' thresholds_indirect$DEV <- findTh(LR$DEV, groups = 4)
#'
#' dat_indirect <- dat
#' dat_indirect$DEV <- calibrate(
#'   LR$DEV,
#'   type = "fuzzy",
#'   method = "indirect",
#'   thresholds = thresholds_indirect$DEV
#' )
#'
#' calib_spec_indirect <- calib_spec
#' calib_spec_indirect$DEV <- list(
#'   raw = "DEV",
#'   type = "fuzzy",
#'   method = "indirect",
#'   thresholds = thresholds_indirect$DEV
#' )
#'
#' indirect_out <- calib.test(
#'   raw.data = LR,
#'   calib.data = dat_indirect,
#'   outcome = outcome,
#'   conditions = conditions,
#'   calib_spec = calib_spec_indirect,
#'   test.conditions = "DEV",
#'   anchors_to_test = c("T1", "T2"),
#'   unit_step = NULL,
#'   unit_step_divisor = 10,
#'   max_steps = 5,
#'   incl.cut = 0.8,
#'   n.cut = 1,
#'   solution = "all",
#'   dir.exp = dir_exp,
#'   progress = TRUE
#' )
#'
#' as.data.frame(indirect_out)
#' }
#'
#' @seealso [qcaERT_plots], [incl.test()], [ncut.test()], [loo.test()],
#'   [subsample.test()], [altset.test()], [theory.test()], [cluster.test()],
#'   [sol.df()]
#' @export
calib.test <- function(
    raw.data,
    calib.data,
    outcome,
    conditions,
    calib_spec,
    test.conditions = conditions,
    test.outcome = FALSE,
    anchors_to_test = NULL,
    solution = "all",
    include = NULL,
    which_M = 1,
    unit_step = NULL,
    unit_step_divisor = 10,
    max_steps = 20,
    incl.cut = 1,
    n.cut = 1,
    dir.exp = NULL,
    i_mode = c("all", "C1P1"),
    exclude_mode = c("recompute", "static", "none"),
    exclude_recompute = list(type = 2),
    exclude_static = NULL,
    result_shape = c("wide", "long"),
    progress = TRUE,
    ...
) {

  .require_qca()

  mc <- match.call(expand.dots = FALSE)
  dots_raw <- list(...)
  caller_env <- parent.frame()

  selection_controls <- .normalize_i_mode(i_mode)
  i_mode <- selection_controls$i_mode
  exclude_mode <- match.arg(exclude_mode)
  result_shape <- match.arg(result_shape)
  .reject_exclusion_controls_in_dots(dots_raw, "calib.test")
  .reject_calibration_inputs_in_dots(dots_raw, "calib.test")
  dots_split <- .split_truth_table_minimize_dots(dots_raw)
  dots_tt <- dots_split$tt
  dots_min <- dots_split$min

  # capture minimize-style conveniences for outcome/conditions/dir.exp ----
  if (!is.null(mc$outcome)) {
    expr <- mc$outcome
    val <- .safe_eval_expr(expr, envir = caller_env)
    if (!inherits(val, "error") && is.character(val) && length(val) == 1) {
      outcome <- val
    } else {
      outcome <- .expr_to_chrvec(expr)
      if (length(outcome) != 1) stop("outcome must resolve to a single outcome name.")
    }
  }

  if (!is.null(mc$conditions)) {
    expr <- mc$conditions
    val <- .safe_eval_expr(expr, envir = caller_env)
    if (!inherits(val, "error") && is.character(val) && length(val) >= 1) {
      conditions <- val
    } else {
      conditions <- .expr_to_chrvec(expr)
    }
  }

  if (!is.null(mc[["dir.exp"]])) {
    expr <- mc[["dir.exp"]]
    val <- .safe_eval_expr(expr, envir = caller_env)
    if (!inherits(val, "error")) {
      dir.exp <- val
    } else {
      dir.exp <- .expr_to_chrvec(expr)
    }
  }

  dir.exp <- .normalize_dir_exp_generic(dir.exp, conditions, endpoint_phrase = "`conditions`")

  if (is.null(dir.exp)) {
    de <- .dot_get(dots_raw, "dir.exp")
    if (is.null(de)) de <- .dot_get(dots_raw, "direxp")
    if (!is.null(de)) dir.exp <- de
  }
  dir.exp <- .normalize_dir_exp_generic(dir.exp, conditions, endpoint_phrase = "`conditions`")

  if (is.null(include)) {
    inc <- .dot_get(dots_raw, "include")
    if (!is.null(inc)) include <- inc
  }

  solution_controls <- .resolve_solution_controls(
    solution = solution,
    include = include,
    dir.exp = dir.exp,
    caller = "calib.test",
    style = "plain"
  )

  solution <- solution_controls$solution
  include <- solution_controls$include
  monitored_solutions <- solution_controls$monitored

  .validate_exclusion_controls(
    exclude_mode = exclude_mode,
    exclude_recompute = exclude_recompute,
    exclude_static = exclude_static,
    exclude_recompute_supplied = !is.null(mc$exclude_recompute),
    exclude_static_supplied = !is.null(mc$exclude_static),
    monitored_solutions = monitored_solutions,
    style = "plain"
  )

  .err_msg <- function(e) {
    if (!inherits(e, "error")) return(NA_character_)
    tryCatch(conditionMessage(e), error = function(...) as.character(e))
  }

  .stage_stop_reason <- function(source) {
    switch(
      source,
      truthTable = "truth_table_build_error",
      exclude = "exclude_recompute_error",
      conservative = "requested_minimize_error",
      parsimonious = "requested_minimize_error",
      intermediate = "requested_minimize_error",
      parsimonious_seed = "requested_minimize_error",
      "run_error"
    )
  }

  .ok_run_info <- function() {
    list(error = FALSE, source = NA_character_, message = NA_character_, reason = NA_character_)
  }

  .stage_error_info <- function(source, e) {
    list(
      error = TRUE,
      source = source,
      message = .err_msg(e),
      reason = .stage_stop_reason(source)
    )
  }

  .selected_solution_missing_info <- function(solution_type, meta) {
    list(
      error = TRUE,
      source = solution_type,
      message = paste0(
        "Requested `which_M = ",
        meta$which_M,
        "` is not available for the ",
        solution_type,
        " baseline solution."
      ),
      reason = "selected_solution_missing"
    )
  }

  .solution_type_selected_solution_missing <- function(run, solution_type) {
    if (!is.list(run) || is.null(run$meta)) return(FALSE)

    if (solution_type == "conservative") {
      return(identical(run$meta$con_hasM, FALSE))
    }

    if (solution_type == "parsimonious") {
      return(identical(run$meta$par_hasM, FALSE))
    }

    if (solution_type == "intermediate") {
      return(isTRUE(run$meta$int_missingM))
    }

    FALSE
  }

  .build_truth_table <- function(data_step) {
    tt_args <- c(
      list(
        data = data_step,
        outcome = outcome,
        incl.cut = incl.cut,
        n.cut = n.cut
      ),
      dots_tt
    )

    if (!is.null(conditions)) {
      tt_args$conditions <- conditions
    }

    tryCatch(
      suppressWarnings(do.call(QCA::truthTable, tt_args)),
      error = function(e) e
    )
  }

  .minimize_from_tt <- function(tt_obj, include_val, dir.exp_val = NULL, exclude_val = NULL, dots_filtered) {
    args <- c(
      list(
        input = tt_obj,
        include = include_val
      ),
      dots_filtered
    )

    if (!is.null(dir.exp_val)) args$dir.exp <- dir.exp_val
    if (!is.null(exclude_val)) args$exclude <- exclude_val

    tryCatch(
      suppressWarnings(do.call(QCA::minimize, args)),
      error = function(e) e
    )
  }

  .minimize_call <- function(data_step, include_val, dir.exp_val = NULL, exclude_val = NULL, dots_filtered) {
    tt_obj <- .build_truth_table(data_step)

    if (inherits(tt_obj, "error")) {
      return(tt_obj)
    }

    .minimize_from_tt(
      tt_obj = tt_obj,
      include_val = include_val,
      dir.exp_val = dir.exp_val,
      exclude_val = exclude_val,
      dots_filtered = dots_filtered
    )
  }

  .compute_exclude <- function(tt_obj) {
    .qcaert_compute_exclude(
      tt_obj = tt_obj,
      exclude_mode = exclude_mode,
      exclude_recompute = exclude_recompute,
      exclude_static = exclude_static
    )
  }

  .run_selected_solution <- function(data_step, solution, which_M, dots_filtered) {
    sig <- list(conservative = NULL, parsimonious = NULL, intermediate = NULL)
    meta <- list(
      which_M = which_M,
      con_nM = NA_integer_,
      par_nM = NA_integer_,
      int_nM_min = NA_integer_,
      con_hasM = NA,
      par_hasM = NA,
      int_missingM = NA
    )
    errs <- list(
      truthTable = NULL,
      conservative = NULL,
      parsimonious = NULL,
      exclude = NULL,
      intermediate = NULL,
      seed = NULL
    )

    r_con <- NULL
    r_par <- NULL
    r_int <- NULL
    r_par0 <- NULL
    exclude_used <- NULL

    tt_obj <- .build_truth_table(data_step)

    if (inherits(tt_obj, "error")) {
      errs$truthTable <- tt_obj
      if (solution == "conservative") {
        r_con <- tt_obj
      } else if (solution == "parsimonious") {
        r_par <- tt_obj
      } else {
        r_int <- tt_obj
      }

      return(list(
        r_con = r_con, r_par = r_par, r_int = r_int,
        exclude_used = exclude_used, sig = sig, meta = meta, errs = errs, r_par0 = r_par0
      ))
    }

    if (solution == "conservative") {
      r_con <- .minimize_from_tt(
        tt_obj = tt_obj,
        include_val = "",
        dots_filtered = dots_filtered
      )

      if (inherits(r_con, "error")) {
        errs$conservative <- r_con
      } else {
        con_info <- .extract_selected_m_standard_sig(r_con, which_M = which_M)
        sig$conservative <- con_info$sig
        meta$con_nM <- as.integer(con_info$n_M)
        meta$con_hasM <- isTRUE(con_info$has_M)
      }

      return(list(
        r_con = r_con, r_par = r_par, r_int = r_int,
        exclude_used = exclude_used, sig = sig, meta = meta, errs = errs, r_par0 = r_par0
      ))
    }

    if (solution == "parsimonious") {
      ex <- .compute_exclude(tt_obj)
      exclude_used <- ex

      if (inherits(ex, "error")) {
        errs$exclude <- ex
        r_par <- ex
      } else {
        r_par <- .minimize_from_tt(
          tt_obj = tt_obj,
          include_val = "?",
          exclude_val = ex,
          dots_filtered = dots_filtered
        )

        if (inherits(r_par, "error")) {
          errs$parsimonious <- r_par
        }
      }

      if (!inherits(r_par, "error")) {
        par_info <- .extract_selected_m_standard_sig(r_par, which_M = which_M)
        sig$parsimonious <- par_info$sig
        meta$par_nM <- as.integer(par_info$n_M)
        meta$par_hasM <- isTRUE(par_info$has_M)
      }

      return(list(
        r_con = r_con, r_par = r_par, r_int = r_int,
        exclude_used = exclude_used, sig = sig, meta = meta, errs = errs, r_par0 = r_par0
      ))
    }

    ex <- .compute_exclude(tt_obj)
    exclude_used <- ex

    if (inherits(ex, "error")) {
      errs$exclude <- ex
      r_int <- ex
    } else {
      r_int <- .minimize_from_tt(
        tt_obj = tt_obj,
        include_val = "?",
        dir.exp_val = dir.exp,
        exclude_val = ex,
        dots_filtered = dots_filtered
      )

      if (inherits(r_int, "error")) {
        errs$intermediate <- r_int
      }
    }

    if (!inherits(r_int, "error")) {
      tmp <- .extract_selected_m_intermediate_sig(r_int$i.sol, which_M = which_M, i_mode = i_mode)
      sig$intermediate <- tmp$sig
      meta$int_nM_min <- as.integer(tmp$n_M_min)
      meta$int_missingM <- isTRUE(tmp$missing_M)
    }

    list(
      r_con = r_con, r_par = r_par, r_int = r_int,
      exclude_used = exclude_used, sig = sig, meta = meta, errs = errs, r_par0 = r_par0
    )
  }

  .baseline_error_info <- function(base, solution, dir.exp) {
    if (!is.null(base$errs$truthTable)) {
      return(.stage_error_info("truthTable", base$errs$truthTable))
    }

    if (solution == "all") {
      baseline_error <- inherits(base$r_con, "error") ||
        inherits(base$r_par, "error") ||
        (!is.null(dir.exp) && (is.null(base$r_int) || inherits(base$r_int, "error")))

      if (!is.null(base$errs$conservative)) {
        return(.stage_error_info("conservative", base$errs$conservative))
      }
      if (!is.null(base$errs$parsimonious)) {
        return(.stage_error_info("parsimonious", base$errs$parsimonious))
      }
      if (!is.null(base$errs$exclude)) {
        return(.stage_error_info("exclude", base$errs$exclude))
      }
      if (!is.null(base$errs$intermediate)) {
        return(.stage_error_info("intermediate", base$errs$intermediate))
      }

      for (solution_type in c("conservative", "parsimonious", "intermediate")) {
        if (.solution_type_selected_solution_missing(base, solution_type)) {
          return(.selected_solution_missing_info(solution_type, base$meta))
        }
      }

      if (!baseline_error) {
        return(.ok_run_info())
      }

      return(list(error = TRUE, source = "unknown", message = NA_character_, reason = "run_error"))
    }

    if (solution == "conservative") {
      baseline_error <- inherits(base$r_con, "error")
      if (baseline_error) {
        return(.stage_error_info("conservative", base$errs$conservative))
      }
      if (.solution_type_selected_solution_missing(base, "conservative")) {
        return(.selected_solution_missing_info("conservative", base$meta))
      }
      return(.ok_run_info())
    }

    if (solution == "parsimonious") {
      baseline_error <- inherits(base$r_par, "error")
      if (!is.null(base$errs$exclude)) {
        return(.stage_error_info("exclude", base$errs$exclude))
      }
      if (baseline_error) {
        return(.stage_error_info("parsimonious", base$errs$parsimonious))
      }
      if (.solution_type_selected_solution_missing(base, "parsimonious")) {
        return(.selected_solution_missing_info("parsimonious", base$meta))
      }
      return(.ok_run_info())
    }

    baseline_error <- inherits(base$r_int, "error")
    if (!is.null(base$errs$seed)) {
      return(.stage_error_info("parsimonious_seed", base$errs$seed))
    }
    if (!is.null(base$errs$exclude)) {
      return(.stage_error_info("exclude", base$errs$exclude))
    }
    if (baseline_error) {
      return(.stage_error_info("intermediate", base$errs$intermediate))
    }
    if (.solution_type_selected_solution_missing(base, "intermediate")) {
      return(.selected_solution_missing_info("intermediate", base$meta))
    }
    .ok_run_info()
  }

  .step_error_info <- function(cur, solution, dir.exp) {
    if (!is.null(cur$errs$truthTable)) {
      return(.stage_error_info("truthTable", cur$errs$truthTable))
    }

    if (solution == "all") {
      any_err <- inherits(cur$r_con, "error") ||
        inherits(cur$r_par, "error") ||
        (!is.null(dir.exp) && !is.null(cur$r_int) && inherits(cur$r_int, "error"))

      if (!any_err) {
        return(.ok_run_info())
      }

      if (!is.null(cur$errs$conservative)) {
        return(.stage_error_info("conservative", cur$errs$conservative))
      }
      if (!is.null(cur$errs$parsimonious)) {
        return(.stage_error_info("parsimonious", cur$errs$parsimonious))
      }
      if (!is.null(cur$errs$exclude)) {
        return(.stage_error_info("exclude", cur$errs$exclude))
      }
      if (!is.null(cur$errs$intermediate)) {
        return(.stage_error_info("intermediate", cur$errs$intermediate))
      }

      return(list(error = TRUE, source = "unknown", message = NA_character_, reason = "run_error"))
    }

    if (solution == "conservative") {
      any_err <- inherits(cur$r_con, "error")
      if (!any_err) return(.ok_run_info())
      return(.stage_error_info("conservative", cur$errs$conservative))
    }

    if (solution == "parsimonious") {
      any_err <- inherits(cur$r_par, "error")
      if (!any_err) return(.ok_run_info())
      if (!is.null(cur$errs$exclude)) {
        return(.stage_error_info("exclude", cur$errs$exclude))
      }
      return(.stage_error_info("parsimonious", cur$errs$parsimonious))
    }

    any_err <- inherits(cur$r_int, "error")
    if (!any_err) return(.ok_run_info())
    if (!is.null(cur$errs$seed)) {
      return(.stage_error_info("parsimonious_seed", cur$errs$seed))
    }
    if (!is.null(cur$errs$exclude)) {
      return(.stage_error_info("exclude", cur$errs$exclude))
    }
    .stage_error_info("intermediate", cur$errs$intermediate)
  }

  .solution_change_info <- function(base, cur, solution, dir.exp) {
    if (solution == "all") {
      ch_con <- .sig_changed(base$sig$conservative, cur$sig$conservative)
      ch_par <- .sig_changed(base$sig$parsimonious, cur$sig$parsimonious)
      ch_int <- (!is.null(dir.exp) && .sig_changed(base$sig$intermediate, cur$sig$intermediate))

      changed <- ch_con || ch_par || ch_int

      changed_types <- if (changed) {
        paste(
          c(
            if (ch_con) "conservative" else NULL,
            if (ch_par) "parsimonious" else NULL,
            if (ch_int) "intermediate" else NULL
          ),
          collapse = ","
        )
      } else {
        NA_character_
      }

      change_kind <- if (changed) {
        paste(
          c(
            if (ch_con) paste0("conservative:", .change_kind_selected_m_sig(base$sig$conservative, cur$sig$conservative)) else NULL,
            if (ch_par) paste0("parsimonious:", .change_kind_selected_m_sig(base$sig$parsimonious, cur$sig$parsimonious)) else NULL,
            if (ch_int) paste0("intermediate:", .change_kind_selected_m_sig(base$sig$intermediate, cur$sig$intermediate)) else NULL
          ),
          collapse = ","
        )
      } else {
        NA_character_
      }

      return(list(
        changed = changed,
        changed_types = changed_types,
        change_kind = change_kind
      ))
    }

    if (solution == "conservative") {
      changed <- .sig_changed(base$sig$conservative, cur$sig$conservative)
      return(list(
        changed = changed,
        changed_types = if (changed) "conservative" else NA_character_,
        change_kind = if (changed) paste0("conservative:", .change_kind_selected_m_sig(base$sig$conservative, cur$sig$conservative)) else NA_character_
      ))
    }

    if (solution == "parsimonious") {
      changed <- .sig_changed(base$sig$parsimonious, cur$sig$parsimonious)
      return(list(
        changed = changed,
        changed_types = if (changed) "parsimonious" else NA_character_,
        change_kind = if (changed) paste0("parsimonious:", .change_kind_selected_m_sig(base$sig$parsimonious, cur$sig$parsimonious)) else NA_character_
      ))
    }

    changed <- .sig_changed(base$sig$intermediate, cur$sig$intermediate)
    list(
      changed = changed,
      changed_types = if (changed) "intermediate" else NA_character_,
      change_kind = if (changed) paste0("intermediate:", .change_kind_selected_m_sig(base$sig$intermediate, cur$sig$intermediate)) else NA_character_
    )
  }

  # validate/align ----
  if (is.null(dim(raw.data)) || is.null(nrow(raw.data)) || nrow(raw.data) < 1L) {
    stop("`raw.data` must be a non-empty data frame with at least one row.")
  }

  if (is.null(dim(calib.data)) || is.null(nrow(calib.data)) || nrow(calib.data) < 1L) {
    stop("`calib.data` must be a non-empty data frame with at least one row.")
  }

  if (nrow(raw.data) != nrow(calib.data)) {
    stop("`raw.data` and `calib.data` must have the same number of rows.")
  }

  if (!is.character(outcome) || length(outcome) != 1L || !nzchar(outcome)) {
    stop("`outcome` must be a single non-empty character string.")
  }

  if (!outcome %in% colnames(calib.data)) {
    stop("`outcome` must exist in `calib.data`.")
  }

  if (!is.character(conditions) || length(conditions) < 1L) {
    stop("`conditions` must be a non-empty character vector.")
  }

  if (!all(conditions %in% colnames(calib.data))) {
    stop("All `conditions` must exist in `calib.data`.")
  }

  .validate_outcome_conditions_distinct(outcome, conditions)

  test.outcome <- .normalize_test_outcome(test.outcome)
  test.conditions <- .normalize_test_conditions(
    test.conditions = test.conditions,
    conditions = conditions,
    test.outcome = test.outcome
  )

  calib_context <- .prepare_calib_context(
    conditions = conditions,
    outcome = outcome,
    calib_spec = calib_spec,
    test.outcome = test.outcome,
    raw.data = raw.data,
    unit_step = unit_step,
    unit_step_divisor = unit_step_divisor,
    anchors_to_test = anchors_to_test,
    caller = "calib.test"
  )
  calib_specs <- calib_context$calib_specs
  type_fc <- calib_context$type_fc
  thr_list <- calib_context$thresholds
  unit_step_vec <- calib_context$unit_step_targets
  anchors_to_test <- calib_context$anchors_to_test
  which_M <- .coerce_which_M(which_M)

  tested_sets <- c(test.conditions, if (test.outcome) outcome else character(0))
  tested_roles <- c(rep("condition", length(test.conditions)), if (test.outcome) "outcome" else character(0))
  names(tested_roles) <- tested_sets

  dots_filtered <- dots_min

  # baseline data / baseline model ----
  baseline_data <- .calib_apply_specs_to_data(
    data = calib.data,
    raw.data = raw.data,
    calib_specs = calib_specs,
    conditions = calib_context$spec_targets
  )

  if (solution == "all") {
    baseline_by_solution_type <- setNames(
      lapply(
        monitored_solutions,
        function(solution_type) {
          .run_selected_solution(
            baseline_data,
            solution = solution_type,
            which_M = which_M,
            dots_filtered = dots_filtered
          )
        }
      ),
      monitored_solutions
    )

    baseline_model <- list(
      by_solution_type = baseline_by_solution_type,
      sig = lapply(baseline_by_solution_type, function(x) x$sig),
      meta = lapply(baseline_by_solution_type, function(x) x$meta),
      errs = lapply(baseline_by_solution_type, function(x) x$errs)
    )

    baseline_err_by_solution_type <- setNames(
      lapply(
        monitored_solutions,
        function(solution_type) .baseline_error_info(baseline_by_solution_type[[solution_type]], solution = solution_type, dir.exp = dir.exp)
      ),
      monitored_solutions
    )

    baseline_stop_reason <- NA_character_
    baseline_err_source <- NA_character_
    baseline_err_message <- NA_character_

    baseline_nM_con <- if ("conservative" %in% monitored_solutions) {
      baseline_by_solution_type$conservative$meta$con_nM
    } else {
      NA_integer_
    }
    baseline_nM_par <- if ("parsimonious" %in% monitored_solutions) {
      baseline_by_solution_type$parsimonious$meta$par_nM
    } else {
      NA_integer_
    }
    baseline_nM_int <- if ("intermediate" %in% monitored_solutions) {
      baseline_by_solution_type$intermediate$meta$int_nM_min
    } else {
      NA_integer_
    }

    baseline_error_flags <- vapply(
      baseline_err_by_solution_type,
      function(x) isTRUE(x$error),
      logical(1)
    )
    baseline_status <- if (!any(baseline_error_flags)) {
      "ok"
    } else if (all(baseline_error_flags)) {
      "baseline_error"
    } else {
      "partial"
    }
    first_baseline_error <- which(baseline_error_flags)[1L]
    baseline_error_source <- if (is.na(first_baseline_error)) {
      NA_character_
    } else {
      baseline_err_by_solution_type[[first_baseline_error]]$source
    }
    baseline_error_message <- if (is.na(first_baseline_error)) {
      NA_character_
    } else {
      baseline_err_by_solution_type[[first_baseline_error]]$message
    }
  } else {
    baseline_model <- .run_selected_solution(
      baseline_data,
      solution = solution,
      which_M = which_M,
      dots_filtered = dots_filtered
    )

    baseline_err <- .baseline_error_info(baseline_model, solution = solution, dir.exp = dir.exp)
    baseline_error <- isTRUE(baseline_err$error)
    baseline_stop_reason <- if (baseline_error) paste0("baseline_", baseline_err$reason) else NA_character_

    baseline_nM_con <- baseline_model$meta$con_nM
    baseline_nM_par <- baseline_model$meta$par_nM
    baseline_nM_int <- baseline_model$meta$int_nM_min

    baseline_err_source <- baseline_err$source
    baseline_err_message <- baseline_err$message

    baseline_by_solution_type <- setNames(list(baseline_model), solution)
    baseline_status <- if (baseline_error) "baseline_error" else "ok"
    baseline_error_source <- baseline_err_source
    baseline_error_message <- baseline_err_message
  }

  baseline <- list(
    status = baseline_status,
    solution = solution,
    monitored_solutions = monitored_solutions,
    by_solution_type = baseline_by_solution_type,
    sig = setNames(
      lapply(
        names(baseline_by_solution_type),
        function(solution_type) baseline_by_solution_type[[solution_type]]$sig[[solution_type]]
      ),
      names(baseline_by_solution_type)
    ),
    meta = setNames(
      lapply(
        names(baseline_by_solution_type),
        function(solution_type) baseline_by_solution_type[[solution_type]]$meta
      ),
      names(baseline_by_solution_type)
    ),
    errs = setNames(
      lapply(
        names(baseline_by_solution_type),
        function(solution_type) baseline_by_solution_type[[solution_type]]$errs
      ),
      names(baseline_by_solution_type)
    ),
    error_source = baseline_error_source,
    error_message = baseline_error_message
  )

  # outputs ----
  diag_rows <- list()
  by_set <- list()

  count_paths_one <- function(set_name) {
    n_paths <- length(.effective_calib_anchors(calib_specs[[set_name]], anchors_to_test)) * 2L
    if (solution == "all") {
      n_paths <- n_paths * length(monitored_solutions)
    }
    n_paths
  }
  total_paths <- sum(vapply(tested_sets, count_paths_one, integer(1)))
  progress_state <- .new_qcaert_progress(total = total_paths, progress = progress)
  on.exit(progress_state$close(), add = TRUE)
  .bump_pb <- progress_state$tick

  # main loop ----
  for (set_name in tested_sets) {

    role <- tested_roles[[set_name]]
    spec0 <- calib_specs[[set_name]]
    raw_name <- spec0$raw
    t0 <- spec0$thresholds
    t_fc <- spec0$type
    step_i <- unit_step_vec[[set_name]]

    anchors_eff <- .effective_calib_anchors(spec0, anchors_to_test)
    if (length(anchors_eff) == 0L) {
      stop(
        "`anchors_to_test` did not match any eligible calibration anchors for set '",
        set_name,
        "'. Eligible anchors are: ",
        paste(.calib_anchor_labels(spec0), collapse = ", "),
        "."
      )
    }

    x_raw <- raw.data[[raw_name]]
    xmin <- suppressWarnings(min(x_raw, na.rm = TRUE))
    xmax <- suppressWarnings(max(x_raw, na.rm = TRUE))
    x_range <- xmax - xmin

    set_out <- list(
      set = set_name,
      role = role,
      raw = raw_name,
      type = t_fc,
      method = spec0$method,
      thresholds0 = t0,
      xmin = xmin, xmax = xmax,
      unit_step_used = step_i,
      solution = solution,
      monitored_solutions = monitored_solutions,
      baseline = baseline
    )

    if (!is.na(baseline_stop_reason)) {
      for (a in anchors_eff) for (d in c("lower", "upper")) {
        diag_rows[[length(diag_rows) + 1]] <- data.frame(
          set = set_name,
          role = role,
          raw = raw_name,
          type = t_fc,
          method = spec0$method,
          anchor = a,
          direction = d,
          solution = solution,
          monitored_solutions = paste(monitored_solutions, collapse = ","),
          which_M_tested = which_M,
          n_M_conservative_baseline = baseline_nM_con,
          n_M_parsimonious_baseline = baseline_nM_par,
          n_M_intermediate_min_baseline = baseline_nM_int,
          step_unit_used = step_i,
          start_value = NA_real_,
          last_safe_value = NA_real_,
          failing_value = NA_real_,
          number_of_steps = NA_integer_,
          total_delta_units = NA_real_,
          delta_as_pct_of_raw_range = NA_real_,
          stop_reason = baseline_stop_reason,
          changed_types = NA_character_,
          change_kind = NA_character_,
          error_source = baseline_err_source,
          error_message = baseline_err_message,
          stringsAsFactors = FALSE
        )
        .bump_pb()
      }
      by_set[[set_name]] <- set_out
      next
    }

    step_results <- list()

    search_path <- function(anchor, direction, solution_type = NULL) {
      start_value <- .calib_anchor_value(spec0, anchor)
      solution_i <- if (is.null(solution_type)) solution else solution_type
      monitored_i <- if (is.null(solution_type)) monitored_solutions else solution_type
      baseline_i <- if (is.null(solution_type)) baseline_model else baseline_by_solution_type[[solution_type]]
      baseline_err_i <- if (is.null(solution_type)) {
        list(
          error = !is.na(baseline_stop_reason),
          source = baseline_err_source,
          message = baseline_err_message
        )
      } else {
        baseline_err_by_solution_type[[solution_type]]
      }

      cur_value <- start_value
      steps_done <- 0L
      stop_reason <- NA_character_
      failing_value <- NA_real_
      changed_types <- NA_character_
      change_kind <- NA_character_
      error_source <- NA_character_
      error_message <- NA_character_

      trace <- .empty_change_trace("value", value_type = "numeric")

      if (isTRUE(baseline_err_i$error)) {
        stop_reason <- paste0("baseline_", baseline_err_i$reason)
        error_source <- baseline_err_i$source
        error_message <- baseline_err_i$message

        row <- data.frame(
          set = set_name,
          role = role,
          raw = raw_name,
          type = t_fc,
          method = spec0$method,
          anchor = anchor,
          direction = direction,
          solution = solution,
          monitored_solutions = paste(monitored_i, collapse = ","),
          which_M_tested = which_M,
          n_M_conservative_baseline = baseline_nM_con,
          n_M_parsimonious_baseline = baseline_nM_par,
          n_M_intermediate_min_baseline = baseline_nM_int,
          step_unit_used = step_i,
          start_value = start_value,
          last_safe_value = NA_real_,
          failing_value = NA_real_,
          number_of_steps = NA_integer_,
          total_delta_units = NA_real_,
          delta_as_pct_of_raw_range = NA_real_,
          stop_reason = stop_reason,
          changed_types = NA_character_,
          change_kind = NA_character_,
          error_source = error_source,
          error_message = error_message,
          stringsAsFactors = FALSE
        )

        if (!is.null(solution_type)) {
          row$solution_type <- solution_type
        }

        path <- list(
          anchor = anchor,
          direction = direction,
          solution = solution,
          monitored_solutions = monitored_i,
          solution_type = solution_type,
          method = spec0$method,
          step_unit_used = step_i,
          start_value = start_value,
          last_safe_value = NA_real_,
          failing_value = NA_real_,
          steps_done = NA_integer_,
          stop_reason = stop_reason,
          changed_types = NA_character_,
          change_kind = NA_character_,
          error_source = error_source,
          error_message = error_message,
          trace = trace
        )

        return(list(row = row, path = path))
      }

      for (k in seq_len(max_steps)) {
        next_value <- if (direction == "lower") cur_value - step_i else cur_value + step_i

        if (!.is_calib_anchor_feasible(spec0, anchor, next_value, direction, xmin, xmax)) {
          stop_reason <- "feasibility_boundary"
          failing_value <- next_value
          break
        }

        spec_step <- .replace_calib_anchor(spec0, anchor, next_value)

        data_step <- baseline_data
        data_step[[set_name]] <- .calibrate_from_spec(x_raw, spec_step)

        cur <- .run_selected_solution(
          data_step,
          solution = solution_i,
          which_M = which_M,
          dots_filtered = dots_filtered
        )

        cur_err <- .step_error_info(cur, solution = solution_i, dir.exp = dir.exp)

        if (isTRUE(cur_err$error)) {
          stop_reason <- cur_err$reason
          failing_value <- next_value
          error_source <- cur_err$source
          error_message <- cur_err$message

          trace <- .append_change_trace(
            trace = trace,
            step = k,
            value_col = "value",
            value = next_value,
            changed = NA,
            status = stop_reason,
            change_kind = NA_character_
          )
          break
        }

        change_info <- .solution_change_info(
          baseline_i,
          cur,
          solution = solution_i,
          dir.exp = dir.exp
        )
        changed <- isTRUE(change_info$changed)

        trace <- .append_change_trace(
          trace = trace,
          step = k,
          value_col = "value",
          value = next_value,
          changed = changed,
          status = "ok",
          change_kind = if (changed) change_info$change_kind else NA_character_
        )

        if (changed) {
          stop_reason <- "solution_change"
          failing_value <- next_value
          changed_types <- change_info$changed_types
          change_kind <- change_info$change_kind
          break
        }

        cur_value <- next_value
        steps_done <- steps_done + 1L
      }

      if (is.na(stop_reason)) stop_reason <- "run_budget_exhausted"

      last_safe_value <- cur_value
      total_delta <- last_safe_value - start_value
      pct_delta <- if (is.finite(x_range) && x_range > 0) (abs(total_delta) / x_range) * 100 else NA_real_

      row <- data.frame(
        set = set_name,
        role = role,
        raw = raw_name,
        type = t_fc,
        method = spec0$method,
        anchor = anchor,
        direction = direction,
        solution = solution,
        monitored_solutions = paste(monitored_i, collapse = ","),
        which_M_tested = which_M,
        n_M_conservative_baseline = baseline_nM_con,
        n_M_parsimonious_baseline = baseline_nM_par,
        n_M_intermediate_min_baseline = baseline_nM_int,
        step_unit_used = step_i,
        start_value = start_value,
        last_safe_value = last_safe_value,
        failing_value = failing_value,
        number_of_steps = steps_done,
        total_delta_units = total_delta,
        delta_as_pct_of_raw_range = pct_delta,
        stop_reason = stop_reason,
        changed_types = changed_types,
        change_kind = change_kind,
        error_source = error_source,
        error_message = error_message,
        stringsAsFactors = FALSE
      )

      if (!is.null(solution_type)) {
        row$solution_type <- solution_type
      }

      path <- list(
        anchor = anchor,
        direction = direction,
        solution = solution,
        monitored_solutions = monitored_i,
        solution_type = solution_type,
        method = spec0$method,
        step_unit_used = step_i,
        start_value = start_value,
        last_safe_value = last_safe_value,
        failing_value = failing_value,
        steps_done = steps_done,
        stop_reason = stop_reason,
        changed_types = changed_types,
        change_kind = change_kind,
        error_source = error_source,
        error_message = error_message,
        trace = trace
      )

      list(row = row, path = path)
    }

    for (anchor in anchors_eff) {
      for (direction in c("lower", "upper")) {
        step_key <- paste(anchor, direction, sep = "_")

        if (solution == "all") {
          paths <- list()
          for (solution_type in monitored_solutions) {
            searched <- search_path(anchor, direction, solution_type = solution_type)
            diag_rows[[length(diag_rows) + 1]] <- searched$row
            paths[[solution_type]] <- searched$path
            .bump_pb()
          }

          first_path <- paths[[monitored_solutions[1L]]]
          step_results[[step_key]] <- c(
            list(
              anchor = anchor,
              direction = direction,
              solution = solution,
              monitored_solutions = monitored_solutions,
              by_solution_type = paths
            ),
            first_path[setdiff(names(first_path), c("anchor", "direction", "solution", "monitored_solutions"))]
          )
        } else {
          searched <- search_path(anchor, direction)
          diag_rows[[length(diag_rows) + 1]] <- searched$row
          step_results[[step_key]] <- searched$path
          .bump_pb()
        }
      }
    }

    set_out$steps <- step_results
    by_set[[set_name]] <- set_out
  }

  diagnostics <- .bind_rows_result(diag_rows)

  compact_bounds <- function(diag_df, set_name) {
    dd <- diag_df[diag_df$set == set_name, , drop = FALSE]
    if (nrow(dd) == 0) return(NULL)

    if ("solution_type" %in% names(dd)) {
      solution_types <- .result_solution_type_order(dd$solution_type)
      out <- setNames(vector("list", length(solution_types)), solution_types)

      for (solution_type in solution_types) {
        dd_solution_type <- dd[dd$solution_type == solution_type, , drop = FALSE]
        anchors <- sort(unique(dd_solution_type$anchor))
        mat <- matrix(
          NA_real_,
          nrow = 2,
          ncol = length(anchors),
          dimnames = list(c("Lower", "Upper"), anchors)
        )

        for (a in anchors) {
          lo <- dd_solution_type[dd_solution_type$anchor == a & dd_solution_type$direction == "lower", , drop = FALSE]
          up <- dd_solution_type[dd_solution_type$anchor == a & dd_solution_type$direction == "upper", , drop = FALSE]
          if (nrow(lo) == 1) mat["Lower", a] <- lo$last_safe_value
          if (nrow(up) == 1) mat["Upper", a] <- up$last_safe_value
        }

        out[[solution_type]] <- mat
      }

      return(out)
    }

    anchors <- sort(unique(dd$anchor))
    out <- matrix(NA_real_, nrow = 2, ncol = length(anchors),
                  dimnames = list(c("Lower", "Upper"), anchors))
    for (a in anchors) {
      lo <- dd[dd$anchor == a & dd$direction == "lower", , drop = FALSE]
      up <- dd[dd$anchor == a & dd$direction == "upper", , drop = FALSE]
      if (nrow(lo) == 1) out["Lower", a] <- lo$last_safe_value
      if (nrow(up) == 1) out["Upper", a] <- up$last_safe_value
    }
    out
  }

  bounds_by_set <- lapply(names(by_set), function(nm) compact_bounds(diagnostics, nm))
  names(bounds_by_set) <- names(by_set)

  .new_result_object(
    "calib_test",
    diagnostics = diagnostics,
    results = .make_calib_results(diagnostics, result_shape = result_shape),
    bounds = bounds_by_set,
    baseline = baseline,
    by_set = by_set,
    settings = list(
      outcome = outcome,
      conditions = conditions,
      test.conditions = test.conditions,
      test.outcome = test.outcome,
      calib_spec = calib_specs,
      anchors_to_test = anchors_to_test,
      solution = solution,
      monitored_solutions = monitored_solutions,
      include = include,
      which_M = which_M,
      unit_step = unit_step,
      unit_step_divisor = unit_step_divisor,
      max_steps = max_steps,
      incl.cut = incl.cut,
      n.cut = n.cut,
      dir.exp = dir.exp,
      i_mode = i_mode,
      exclude_mode = exclude_mode,
      exclude_recompute = exclude_recompute,
      exclude_static = exclude_static,
      result_shape = result_shape
      )
  )
}

#' Print a `calib_test` object
#'
#' @rdname calib.test
#' @export
print.calib_test <- function(x, row.names = FALSE, ...) {
  results <- x$results
  display_results <- .calib_results_for_print(results)
  settings <- x$settings

  .print_qcaert_heading("calib_test", "calibration robustness", settings)
  if (!is.null(settings$test.conditions) && length(settings$test.conditions) > 0L) {
    cat("Test conditions: ", paste(settings$test.conditions, collapse = ", "), "\n", sep = "")
  }
  if (isTRUE(settings$test.outcome)) {
    cat("Test outcome: ", settings$outcome, "\n", sep = "")
  }
  if (!is.null(settings$anchors_to_test)) {
    cat("Anchors tested: ", paste(settings$anchors_to_test, collapse = ", "), "\n", sep = "")
  }
  if (!is.null(settings$max_steps)) {
    cat("Max steps: ", settings$max_steps, "\n", sep = "")
  }

  .print_qcaert_table(display_results, "Result", row.names = row.names, ...)

  cat("\nSummary\n")
  if (is.data.frame(display_results) && nrow(display_results) > 0L) {
    cat(" Rows: ", nrow(display_results), "\n", sep = "")
    if ("condition" %in% names(display_results)) {
      cat(" Conditions covered: ", paste(unique(display_results$condition), collapse = ", "), "\n", sep = "")
    }
    cat(" Bounds are available in x$bounds.\n", sep = "")
  }

  invisible(x)
}

.calib_results_for_print <- function(results) {
  if (!is.data.frame(results)) {
    return(results)
  }

  out <- results
  out <- out[, setdiff(names(out), c("set", "role")), drop = FALSE]
  raw_idx <- match("raw", names(out))
  if (!is.na(raw_idx)) {
    names(out)[raw_idx] <- "condition"
  }

  out
}

#' Return the main results table from a `calib_test` object
#'
#' @rdname calib.test
#' @export
as.data.frame.calib_test <- function(x, ...) {
  .as.data.frame_results(x, ...)
}

test_that("calib.test can test only the outcome calibration", {
  skip_if_not_installed("QCA")

  fixture <- qcaert_fixture_outcome_calibration()
  out <- suppressWarnings(calib.test(
    raw.data = fixture$raw,
    calib.data = fixture$calib,
    outcome = fixture$outcome,
    conditions = fixture$conditions,
    calib_spec = fixture$calib_spec,
    test.conditions = NULL,
    test.outcome = TRUE,
    unit_step = 1,
    max_steps = 1,
    incl.cut = 0.75,
    n.cut = 1,
    solution = "conservative",
    progress = FALSE
  ))

  expect_qcaert_result_structure(
    out,
    class = "calib_test",
    diagnostics = c("set", "role", "method", "anchor", "direction", "change_kind"),
    results = c("set", "role", "method", "anchor", "direction", "start", "reason"),
    settings = c("outcome", "conditions", "test.conditions", "test.outcome", "calib_spec"),
    top = c("bounds", "baseline", "by_set", "settings")
  )
  expect_identical(unique(out$results$set), "Y")
  expect_identical(unique(out$results$role), "outcome")
  expect_identical(names(out$by_set), "Y")
  expect_identical(out$settings$test.conditions, character(0))
  expect_true(out$settings$test.outcome)
})

test_that("calib.test can test selected conditions and the outcome together", {
  skip_if_not_installed("QCA")

  fixture <- qcaert_fixture_outcome_calibration()
  out <- suppressWarnings(calib.test(
    raw.data = fixture$raw,
    calib.data = fixture$calib,
    outcome = fixture$outcome,
    conditions = fixture$conditions,
    calib_spec = fixture$calib_spec,
    test.conditions = "A",
    test.outcome = TRUE,
    unit_step = 1,
    max_steps = 1,
    incl.cut = 0.75,
    n.cut = 1,
    solution = "conservative",
    progress = FALSE
  ))

  expect_identical(unique(out$results$set), c("A", "Y"))
  expect_identical(unique(out$results$role), c("condition", "outcome"))
  expect_identical(names(out$by_set), c("A", "Y"))
  expect_identical(nrow(out$results), 24L)
})

test_that("calib.test requires outcome calibration specs for outcome testing", {
  skip_if_not_installed("QCA")

  fixture <- qcaert_fixture_direct6()

  expect_error(
    calib.test(
      raw.data = fixture$raw,
      calib.data = fixture$calib,
      outcome = fixture$outcome,
      conditions = fixture$conditions,
      calib_spec = fixture$calib_spec,
      test.conditions = NULL,
      test.outcome = TRUE,
      unit_step = 1,
      max_steps = 1,
      incl.cut = 0.75,
      n.cut = 1,
      solution = "conservative",
      progress = FALSE
    ),
    "`calib_spec` must contain exactly one entry for each condition and the outcome when `test.outcome = TRUE`.",
    fixed = TRUE
  )
})

test_that("LOO print counts standardized fit-change results", {
  x <- structure(
    list(
      results = data.frame(
        row_index = 1:2,
        case_label = c("case 1", "case 2"),
        status = c("ok", "ok"),
        solution_change = c(NA_character_, "CON: A"),
        fit_changed_types = c("CON", NA_character_),
        n_fit_deltas = c(1L, 0L),
        max_abs_fit_delta = c(0.1, 0),
        stringsAsFactors = FALSE
      ),
      settings = list(
        solution = "conservative",
        monitored_solutions = "conservative",
        which_M = 1L,
        n_cases_tested = 2L,
        fit_measures = "inclS"
      )
    ),
    class = "loo_test"
  )

  out <- capture.output(print(x))

  expect_true(any(grepl("Solution changes: 1", out, fixed = TRUE)))
  expect_true(any(grepl("Fit changes: 1", out, fixed = TRUE)))
})

test_that("boundary print summaries do not repeat stop reasons", {
  results <- data.frame(
    direction = c("lower", "upper"),
    start = c(0.8, 0.8),
    last_safe = c(0.75, 0.85),
    first_failing = c(0.7, 0.9),
    steps = c(1L, 1L),
    total_delta = c(-0.05, 0.05),
    reason = c("CON:formula_changed", "PAR:formula_changed"),
    stringsAsFactors = FALSE
  )
  x <- structure(
    list(
      results = results,
      bounds = c(Lower = 0.75, Upper = 0.85),
      settings = list(
        solution = "conservative",
        monitored_solutions = "conservative",
        which_M = 1L,
        step = 0.05,
        max_steps = 2L
      )
    ),
    class = "incl_test"
  )

  out <- capture.output(print(x))

  expect_true(any(grepl("Stable interval", out, fixed = TRUE)))
  expect_false(any(grepl("search stopped because", out, fixed = TRUE)))
})

test_that("all-solution long boundary print summaries do not repeat solution_type stop reasons", {
  results <- data.frame(
    solution_type = rep(c("conservative", "parsimonious", "intermediate"), each = 2L),
    direction = rep(c("lower", "upper"), times = 3L),
    start = 0.8,
    last_safe = c(0.75, 0.85, 0.8, 0.85, 0.75, 0.8),
    first_failing = c(0.7, 0.9, 0.75, 0.9, 0.7, 0.85),
    steps = c(1L, 1L, 0L, 1L, 1L, 0L),
    total_delta = c(-0.05, 0.05, 0, 0.05, -0.05, 0),
    reason = c(
      "CON:formula_changed",
      "CON:formula_changed",
      "PAR:formula_changed",
      "PAR:formula_changed",
      "INT:formula_changed",
      "INT:formula_changed"
    ),
    stringsAsFactors = FALSE
  )
  bounds <- matrix(
    c(0.75, 0.85, 0.8, 0.85, 0.75, 0.8),
    nrow = 2L,
    dimnames = list(c("Lower", "Upper"), c("conservative", "parsimonious", "intermediate"))
  )
  x <- structure(
    list(
      results = results,
      bounds = bounds,
      settings = list(
        solution = "all",
        monitored_solutions = c("conservative", "parsimonious", "intermediate"),
        which_M = 1L,
        i_mode = "all",
        step = 0.05,
        max_steps = 2L
      )
    ),
    class = "ncut_test"
  )

  out <- capture.output(print(x))

  expect_true(any(grepl("Stable intervals", out, fixed = TRUE)))
  expect_false(any(grepl("CON Lower search stopped because", out, fixed = TRUE)))
  expect_false(any(grepl("search stopped because", out, fixed = TRUE)))
})

test_that("subsample print labels aggregate rows and solution-type-specific exact matches clearly", {
  x <- structure(
    list(
      results = data.frame(
        rep = 1:2,
        n_sample = c(8L, 8L),
        n_holdout = c(2L, 2L),
        status = c("ok", "ok"),
        exact_match_baseline = c(TRUE, FALSE),
        term_jaccard_baseline = c(1, 0.5),
        solution_change = c(NA_character_, "CON: formula_changed"),
        fit_changed_types = c(NA_character_, "CON"),
        n_fit_deltas = c(0L, 1L),
        max_abs_fit_delta = c(0, 0.1),
        stringsAsFactors = FALSE
      ),
      summary = list(
        exact_solution = list(
          conservative = list(
            n_exact = 1L,
            n_successful = 2L,
            prop_exact = 0.5
          )
        )
      ),
      settings = list(
        solution = "conservative",
        monitored_solutions = "conservative",
        reps = 2L,
        sample_n = 8L
      )
    ),
    class = "subsample_test"
  )

  out <- capture.output(print(x))

  expect_true(any(grepl("Runs with any solution change: 1", out, fixed = TRUE)))
  expect_true(any(grepl("Runs with any fit change: 1", out, fixed = TRUE)))
  expect_true(any(grepl("Exact baseline matches by solution type", out, fixed = TRUE)))
})

test_that("altset print foregrounds solution-type-specific match rates", {
  x <- structure(
    list(
      results = data.frame(
        draw = 1:2,
        incl.cut = c(0.75, 0.8),
        n.cut = c(1L, 1L),
        status = c("ok", "ok"),
        n_changed_sets = c(1L, 1L),
        changed_sets = c("A", "A"),
        changed_roles = c("condition", "condition"),
        solution_change = c(NA_character_, "CON: formula_changed"),
        fit_changed_types = c(NA_character_, "CON"),
        n_fit_deltas = c(0L, 1L),
        max_abs_fit_delta = c(0, 0.1),
        stringsAsFactors = FALSE
      ),
      summary = list(
        n_draws = 2L,
        n_fit_compared = 2L,
        score_solution_by_solution_type = c(conservative = 0.5),
        score_fit_by_solution_type = c(conservative = 0.5)
      ),
      settings = list(
        solution = "conservative",
        monitored_solutions = "conservative",
        n_draws = 2L,
        incl.cut = 0.75,
        n.cut = 1,
        fit_tol = 1e-6,
        test.conditions = "A",
        test.outcome = FALSE
      )
    ),
    class = "altset_test"
  )

  out <- capture.output(print(x))

  expect_true(any(grepl("Draws with any solution change: 1", out, fixed = TRUE)))
  expect_true(any(grepl("Solution match rate by solution type", out, fixed = TRUE)))
  expect_false(any(grepl("Solution stability score", out, fixed = TRUE)))
  expect_false(any(grepl("Total stability score", out, fixed = TRUE)))
})

test_that("as.data.frame returns the standardized user-facing table", {
  results <- data.frame(id = 1L)
  classes <- c(
    "altset_test",
    "calib_test",
    "incl_test",
    "loo_test",
    "ncut_test",
    "subsample_test"
  )

  for (class in classes) {
    x <- structure(list(results = results), class = class)
    expect_identical(as.data.frame(x), results)
  }

  cluster <- structure(
    list(results = list(overview = results, clusters = NULL, units = NULL)),
    class = "cluster_test"
  )
  expect_identical(as.data.frame(cluster), results)
})

test_that("calib print uses compact condition-first display table", {
  results <- data.frame(
    set = c("A", "Y"),
    role = c("condition", "outcome"),
    raw = c("A_raw", "Y_raw"),
    type = c("fuzzy", "fuzzy"),
    method = c("direct", "direct"),
    anchor = c("E1", "E1"),
    direction = c("lower", "lower"),
    start = c(10, 20),
    last_safe = c(9, 19),
    first_failing = c(8, 18),
    step_unit = c(1, 1),
    steps = c(1L, 1L),
    total_delta = c(-1, -1),
    pct_raw_range = c(-0.1, -0.1),
    reason = c("formula_changed", "formula_changed"),
    stringsAsFactors = FALSE
  )
  x <- structure(
    list(
      results = results,
      bounds = list(),
      settings = list(
        outcome = "Y",
        solution = "conservative",
        monitored_solutions = "conservative",
        which_M = 1L,
        test.conditions = "A",
        test.outcome = TRUE,
        max_steps = 2L
      )
    ),
    class = "calib_test"
  )

  display <- qcaERT:::.calib_results_for_print(results)
  expect_false("set" %in% names(display))
  expect_false("role" %in% names(display))
  expect_false("raw" %in% names(display))
  expect_true("condition" %in% names(display))
  expect_identical(display$condition, c("A_raw", "Y_raw"))
  expect_identical(as.data.frame(x), results)

  out <- capture.output(print(x))
  expect_true(any(grepl("condition", out, fixed = TRUE)))
  expect_false(any(grepl("\\brole\\b", out)))
  expect_false(any(grepl("\\braw\\b", out)))
  expect_true(any(grepl("Conditions covered: A_raw, Y_raw", out, fixed = TRUE)))
})

test_that("cluster print uses compact overview and single-configuration details", {
  overview <- data.frame(
    solution_type = "INT",
    configuration = "C1P1:~DEV*URB*LIT*IND*~STB",
    components = 2L,
    status = "ok",
    pooled_consistency = 0.8716332,
    pooled_coverage = 0.3921098,
    max_abs_cluster_delta_consistency = 0.3308799,
    max_abs_cluster_delta_coverage = 0.02347898,
    worst_cluster_consistency_id = "low development",
    worst_cluster_coverage_id = "low development",
    within_available = FALSE,
    n_units_repeated = 0L,
    stringsAsFactors = FALSE
  )
  clusters <- data.frame(
    solution_type = c("INT", "INT"),
    configuration = overview$configuration,
    solution_expression = overview$configuration,
    component = "solution",
    cluster_id = c("high development", "low development"),
    cluster_size = c(9L, 9L),
    consistency = c(0.9637, 0.5408),
    coverage = c(0.3961, 0.3686),
    delta_consistency = c(0.0921, -0.3309),
    delta_coverage = c(0.0040, -0.0235),
    stringsAsFactors = FALSE
  )
  x <- structure(
    list(
      results = list(overview = overview, clusters = clusters, units = NULL),
      settings = list(
        solution = "intermediate",
        monitored_solutions = "intermediate",
        which_M = 1L,
        i_mode = "all",
        cluster_id = "DEV_group",
        unit_id = "case_id",
        necessity = FALSE
      )
    ),
    class = "cluster_test"
  )

  out <- capture.output(print(x))

  expect_true(any(grepl("pooled_cons", out, fixed = TRUE)))
  expect_true(any(grepl("worst_consistency_cluster", out, fixed = TRUE)))
  expect_true(any(grepl("Cluster details", out, fixed = TRUE)))
  expect_true(any(grepl("+0.092", out, fixed = TRUE)))
  expect_true(any(grepl("Within-unit diagnostics", out, fixed = TRUE)))
  expect_true(any(grepl("Not available: 0 units are observed in more than one cluster.", out, fixed = TRUE)))
  expect_true(any(grepl("Detailed tables: x$results$overview, x$results$clusters", out, fixed = TRUE)))
  expect_false(any(grepl("Detailed cluster table", out, fixed = TRUE)))
})

test_that("incl_test plot methods return ggplot objects", {
  skip_if_not_installed("QCA")
  skip_if_not_installed("ggplot2")

  fixture <- qcaert_fixture_direct6()
  out <- suppressWarnings(incl.test(
    data = fixture$calib,
    outcome = fixture$outcome,
    conditions = fixture$conditions,
    incl.cut = 0.75,
    step = 0.1,
    max_steps = 1,
    n.cut = 1,
    solution = "all",
    dir.exp = fixture$dir.exp,
    progress = FALSE
  ))

  expect_error(
    plot(out),
    "`solution_type` must be supplied when more than one solution type is present.",
    fixed = TRUE
  )
  expect_error(
    plot(out, solution = "conservative"),
    "Use `solution_type`",
    fixed = TRUE
  )
  expect_error(
    plot(out, solution_type = "all"),
    "`solution_type` must use",
    fixed = TRUE
  )
  expect_s3_class(plot(out, solution_type = "conservative"), "ggplot")
  expect_s3_class(plot(out, solution_type = "conservative", type = "trace", direction = "lower"), "ggplot")
  expect_error(
    plot(out, solution_type = "all"),
    "`solution_type` must use",
    fixed = TRUE
  )
})

test_that("calib_test plot methods return ggplot objects for direct-six anchors", {
  skip_if_not_installed("QCA")
  skip_if_not_installed("ggplot2")

  fixture <- qcaert_fixture_direct6()
  out <- suppressWarnings(calib.test(
    raw.data = fixture$raw,
    calib.data = fixture$calib,
    outcome = fixture$outcome,
    conditions = fixture$conditions,
    calib_spec = fixture$calib_spec,
    test.conditions = fixture$test.conditions,
    unit_step = 1,
    max_steps = 1,
    incl.cut = 0.75,
    n.cut = 1,
    solution = "all",
    dir.exp = fixture$dir.exp,
    progress = FALSE
  ))

  expect_error(
    plot(out),
    "`solution_type` must be supplied when more than one solution type is present.",
    fixed = TRUE
  )
  expect_s3_class(plot(out, solution_type = "conservative"), "ggplot")
  expect_s3_class(plot(out, solution_type = "conservative", type = "heatmap"), "ggplot")
  expect_s3_class(
    plot(out, solution_type = "conservative", type = "trace", set = "A", anchor = "E1", direction = "lower"),
    "ggplot"
  )
  expect_error(
    plot(out, solution = "all"),
    "Use `solution_type`",
    fixed = TRUE
  )
})

test_that("calib_test plot methods return ggplot objects for indirect anchors", {
  skip_if_not_installed("QCA")
  skip_if_not_installed("ggplot2")

  fixture <- qcaert_fixture_indirect()
  out <- suppressWarnings(calib.test(
    raw.data = fixture$raw,
    calib.data = fixture$calib,
    outcome = fixture$outcome,
    conditions = fixture$conditions,
    calib_spec = fixture$calib_spec,
    test.conditions = fixture$test.conditions,
    unit_step = 1,
    max_steps = 1,
    incl.cut = 0.75,
    n.cut = 1,
    solution = "conservative",
    progress = FALSE
  ))

  expect_s3_class(plot(out), "ggplot")
  expect_s3_class(plot(out, type = "heatmap", cell = "anchor"), "ggplot")
  expect_s3_class(
    plot(out, type = "trace", set = "A", anchor = "T1", direction = "upper"),
    "ggplot"
  )
})

test_that("theory_test plot selects one solution_type and returns a ggplot object", {
  skip_if_not_installed("QCA")
  skip_if_not_installed("ggplot2")

  fixture <- qcaert_fixture_lr()
  theories <- list(
    development = c("DEV", "URB", "LIT"),
    industrial = c("DEV", "URB", "IND"),
    broad = c("DEV", "URB", "LIT", "IND", "STB")
  )
  dir_exp <- list(
    development = c("1", "1", "1"),
    industrial = c("1", "1", "1"),
    broad = c("1", "1", "1", "1", "1")
  )

  out <- theory.test(
    data = fixture$calib,
    outcome = fixture$outcome,
    theories = theories,
    incl.cut = 0.8,
    n.cut = 1,
    solution = "all",
    dir.exp = dir_exp,
    progress = FALSE
  )

  expect_error(
    plot(out),
    "`solution_type` must be supplied when more than one solution type is present.",
    fixed = TRUE
  )
  expect_s3_class(plot(out, solution_type = "conservative"), "ggplot")
  expect_s3_class(plot(out, solution_type = "conservative", label_line = FALSE), "ggplot")
  expect_s3_class(plot(out, solution_type = "par"), "ggplot")
  expect_s3_class(plot(out, solution_type = "intermediate"), "ggplot")
  expect_error(
    plot(out, solution_type = "conservative", intermediate_branch = "C1P1"),
    "`intermediate_branch` is only used when solution_type = 'intermediate'.",
    fixed = TRUE
  )
})

if (getRversion() >= "2.15.1") {
  utils::globalVariables(".data")
}

.plot_require_ggplot2 <- function() {
  if (!requireNamespace("ggplot2", quietly = TRUE)) {
    stop("Package 'ggplot2' is required to plot qcaERT results.", call. = FALSE)
  }

  invisible(TRUE)
}

.plot_colors <- function() {
  c(
    blue = "#0072B2",
    sky = "#56B4E9",
    green = "#009E73",
    yellow = "#F0E442",
    orange = "#E69F00",
    vermillion = "#D55E00",
    purple = "#CC79A7",
    grey = "#7A7A7A",
    light_grey = "#EAEAEA",
    dark = "#222222"
  )
}

.plot_theme <- function(base_size = 12) {
  ggplot2::theme_bw(base_size = base_size) +
    ggplot2::theme(
      panel.grid.minor = ggplot2::element_blank(),
      panel.border = ggplot2::element_rect(colour = .plot_colors()[["grey"]], fill = NA),
      strip.background = ggplot2::element_rect(fill = .plot_colors()[["light_grey"]], colour = .plot_colors()[["grey"]]),
      legend.key = ggplot2::element_blank(),
      legend.title = ggplot2::element_text(size = base_size),
      plot.title = ggplot2::element_text(face = "plain")
    )
}

.plot_preserved_colors <- function() {
  c(
    Yes = .plot_colors()[["blue"]],
    No = .plot_colors()[["vermillion"]],
    Unknown = .plot_colors()[["grey"]]
  )
}

.plot_boundary_colors <- function() {
  c(
    "Baseline" = .plot_colors()[["grey"]],
    "Lower last safe" = .plot_colors()[["orange"]],
    "Upper last safe" = .plot_colors()[["blue"]]
  )
}

.plot_theory_colors <- function(theories) {
  theories <- as.character(theories)
  base <- .plot_colors()[c("blue", "orange", "green", "purple", "sky", "vermillion", "yellow", "dark")]
  stats::setNames(rep(base, length.out = length(theories)), theories)
}

.plot_normalize_directions <- function(value, arg = "directions") {
  if (is.null(value)) return(NULL)

  value <- tolower(trimws(as.character(value)))
  value <- value[nzchar(value)]
  if (!length(value)) return(character(0))

  if (!all(value %in% c("lower", "upper"))) {
    stop(sprintf("`%s` must contain only 'lower' and/or 'upper'.", arg), call. = FALSE)
  }

  unique(value)
}

.plot_as_chr0 <- function(x) {
  x <- as.character(x)
  x[is.na(x)] <- ""
  x
}

.plot_split_tokens <- function(x, normalize_solution = FALSE, allow_all = FALSE) {
  x <- .plot_as_chr0(x)
  x <- x[nzchar(x)]
  if (!length(x)) return(character(0))

  tokens <- unlist(strsplit(x, ",", fixed = TRUE), use.names = FALSE)
  tokens <- trimws(tokens)
  tokens <- tokens[nzchar(tokens)]
  if (!length(tokens)) return(character(0))

  if (normalize_solution) {
    tokens <- vapply(tokens, .normalize_solution_std, character(1))
    if (!allow_all && any(tokens == "all")) {
      stop("Solution-token filters must use 'con'/'conservative', 'par'/'parsimonious', or 'int'/'intermediate'.", call. = FALSE)
    }
  }

  sort(unique(tokens))
}

.plot_solution_filter <- function(x, allow_all = TRUE, arg = "solution") {
  if (is.null(x)) return(NULL)

  x <- trimws(as.character(x))
  x <- x[nzchar(x)]
  if (!length(x)) return(character(0))

  out <- vapply(x, .normalize_solution_std, character(1))
  if (!allow_all && any(out == "all")) {
    stop(sprintf("`%s` must use 'con'/'conservative', 'par'/'parsimonious', or 'int'/'intermediate'.", arg), call. = FALSE)
  }

  sort(unique(out))
}

.plot_reject_solution_dots <- function(dots, caller) {
  if (!length(dots)) return(invisible(TRUE))
  nms <- names(dots)
  if (is.null(nms)) return(invisible(TRUE))

  blocked <- intersect(nms, c("solution", "monitored_solutions"))
  if (length(blocked)) {
    stop(
      "`", caller, "()` does not accept `", blocked[[1L]],
      "`. Use `solution_type` to select conservative, intermediate, or parsimonious plots.",
      call. = FALSE
    )
  }

  invisible(TRUE)
}

.plot_reject_solution_args <- function(solution, monitored_solutions, caller) {
  if (!is.null(solution)) {
    stop(
      "`", caller, "()` does not accept `solution`. Use `solution_type` to select conservative, intermediate, or parsimonious plots.",
      call. = FALSE
    )
  }
  if (!is.null(monitored_solutions)) {
    stop(
      "`", caller, "()` does not accept `monitored_solutions`. Use `solution_type` to select conservative, intermediate, or parsimonious plots.",
      call. = FALSE
    )
  }

  invisible(TRUE)
}

.plot_row_solution_type <- function(row) {
  if ("solution_type" %in% names(row)) {
    value <- .plot_as_chr0(row$solution_type[[1L]])
    if (nzchar(value)) return(.normalize_solution_std(value))
  }

  if ("monitored_solutions" %in% names(row)) {
    tokens <- .plot_split_tokens(row$monitored_solutions[[1L]], normalize_solution = TRUE, allow_all = FALSE)
    if (length(tokens) == 1L) return(tokens[[1L]])
  }

  if ("solution" %in% names(row)) {
    value <- .plot_as_chr0(row$solution[[1L]])
    if (nzchar(value) && !identical(value, "all")) return(.normalize_solution_std(value))
  }

  NA_character_
}

.plot_ensure_solution_type <- function(df) {
  df$solution_type <- vapply(
    seq_len(nrow(df)),
    function(i) .plot_row_solution_type(df[i, , drop = FALSE]),
    character(1)
  )
  df
}

.plot_select_solution_type <- function(df, solution_type) {
  df <- .plot_ensure_solution_type(df)
  available <- .result_solution_type_order(df$solution_type)

  if (!length(available)) {
    stop("No solution types are available to plot.", call. = FALSE)
  }

  if (is.null(solution_type)) {
    if (length(available) == 1L) {
      return(list(data = df[df$solution_type == available[[1L]], , drop = FALSE], solution_type = available[[1L]]))
    }
    stop(
      "`solution_type` must be supplied when more than one solution type is present. Available solution types: ",
      paste(available, collapse = ", "),
      call. = FALSE
    )
  }

  solution_type <- .plot_solution_filter(solution_type, allow_all = FALSE, arg = "solution_type")
  if (length(solution_type) != 1L) {
    stop("`solution_type` must be a single solution type.", call. = FALSE)
  }
  if (!solution_type %in% available) {
    stop(
      "`solution_type` is not available in this object. Available solution types: ",
      paste(available, collapse = ", "),
      call. = FALSE
    )
  }

  list(data = df[df$solution_type == solution_type, , drop = FALSE], solution_type = solution_type)
}

.plot_match_solution_tokens <- function(values, requested, arg = "solution") {
  requested <- .plot_solution_filter(requested, allow_all = FALSE, arg = arg)
  values <- .plot_as_chr0(values)
  keep <- rep(FALSE, length(values))

  for (i in seq_along(values)) {
    tokens <- .plot_split_tokens(values[i], normalize_solution = FALSE)
    keep[i] <- any(tokens %in% requested)
  }

  keep[is.na(keep)] <- FALSE
  keep
}

.plot_fmt_counts <- function(tab) {
  names <- names(tab)
  paste0(names, " (", as.integer(tab), ")", collapse = ", ")
}

#' Leave-one-out robustness test for QCA solutions
#'
#' Deletes one case at a time, reruns the QCA analysis, and records whether the
#' monitored solution changes, if selected fit measures change, and
#' whether the reduced-data run ends with an error. The tested cases can be
#' identified by row index or by case label.
#'
#' @param data A data frame object containing the outcome and condition
#'   columns used in the QCA analysis. `data` must have at least
#'   two rows.
#' @param outcome Name of the outcome. This must be a single
#'   non-empty character string.
#' @param conditions Optional character vector of condition names. If `NULL`,
#'   the condition set is left to [QCA::truthTable()]. When
#'   `calib = "recompute"`, `conditions` must be supplied explicitly.
#' @param cases Cases to test. Use `NULL` to test all rows, a numeric vector of
#'   row indices, or a character vector of case labels.
#' @param case_labels Optional character vector of case labels with length
#'   `nrow(data)`. If `NULL`, the function uses `rownames(data)` when
#'   available; otherwise it uses row numbers converted to character.
#' @param calib Calibration handling for reduced-data runs. `"fixed"` uses the
#'   calibrated values already present in `data`. `"recompute"` recalibrates
#'   the outcome and conditions after deleting each case, using `raw.data` and
#'   `calib_spec`.
#' @param raw.data Raw-data frame object used when `calib = "recompute"`.
#'   It must have the same number of rows as `data`.
#' @param calib_spec Calibration specification used when
#'   `calib = "recompute"`. This must be a named list keyed by the analysis
#'   sets `c(outcome, conditions)`. Each entry must describe the raw source
#'   column, the set type, and the calibration inputs used to rebuild the
#'   calibrated set. Use the same `calib_spec` structure described for
#'   [calib.test()]. An entry may additionally contain `findTh`, a named list
#'   of [QCA::findTh()] arguments; when supplied, thresholds are re-estimated
#'   after each case deletion instead of reusing the baseline `thresholds`.
#' @param incl.cut Inclusion cutoff passed to [QCA::truthTable()]. This must
#'   be a single finite number in `[0, 1]`.
#' @param n.cut Truth table frequency cutoff passed to [QCA::truthTable()].
#'   This must be an integer-like value of at least `1`.
#' @param solution Solution type to monitor. Accepted values are `"all"`,
#'   `"con"` or `"conservative"`, `"par"` or `"parsimonious"`, and `"int"` or
#'   `"intermediate"`.
#' @param include Optional minimization include setting. Currently, this
#'   argument accepts only `NULL`, `""`, or `"?"`.
#' @param dir.exp Directional expectations used when the monitored solution is
#'   intermediate.
#' @param which_M Positive integer giving which solution alternative to use
#'   when minimization returns multiple models.
#' @param i_mode Character string controlling intermediate-solution selection.
#'   Accepted values are `"all"` and `"C1P1"`.
#' @param exclude_mode Character string controlling how excluded rows are
#'   handled for parsimonious and intermediate minimization. `"recompute"`
#'   recalculates exclusions from each reduced truth table, `"static"` reuses
#'   `exclude_static`, and `"none"` does not use exclusions.
#' @param exclude_recompute Named list of arguments passed to [QCA::findRows()]
#'   when `exclude_mode = "recompute"`.
#' @param exclude_static Already computed exclusion object reused when
#'   `exclude_mode = "static"`.
#' @param fit_measures Character vector of fit-measure names to compare between
#'   the baseline run and each reduced-data run. Use `NULL` to disable fit
#'   comparison.
#' @param fit_tol Non-negative tolerance used when deciding whether fit values
#'   changed.
#' @param progress Logical; if `TRUE` and the session is interactive, show a
#'   text progress bar.
#' @param x A `loo_test` object returned by [loo.test()].
#' @param row.names Logical; passed to [print.data.frame()] by
#'   [print.loo_test()].
#' @param ... Additional arguments. In [loo.test()], named arguments matching
#'   [QCA::truthTable()] formals are passed to [QCA::truthTable()], and the
#'   remaining named arguments are forwarded to [QCA::minimize()] after
#'   removing names reserved by `loo.test()`. The function also
#'   looks in `...` for `include`, `dir.exp`, or `direxp` if those arguments
#'   were not supplied explicitly. In [print.loo_test()], `...` is passed to
#'   [print.data.frame()]. In [as.data.frame.loo_test()], `...` is ignored.
#'
#' @returns An object of class `loo_test` with the following components:
#'   \describe{
#'     \item{`diagnostics`}{A detailed data frame with one row per tested case.
#'     It records the tested row index, case label, run status, solution-change
#'     classification, fit-change classification, the number of changed fit
#'     measures, the largest absolute fit change, and error information when a
#'     reduced-data run fails.}
#'     \item{`results`}{A compact data frame with the columns
#'     `row_index`, `case_label`, `status`, `solution_change`,
#'     `fit_changed_types`, `n_fit_deltas`, and `max_abs_fit_delta`.}
#'     \item{`baseline`}{A list containing the baseline analysis built from the
#'     full dataset, including the truth table, minimization output, selected
#'     solution terms used for comparison, fit values, exclusion information,
#'     and calibration information.}
#'     \item{`by_case`}{A named list with one entry per tested case. Each entry
#'     stores the baseline run, the reduced-data run when available, change
#'     summaries, fit deltas, exclusion information, calibration information,
#'     and any error information for that case.}
#'     \item{`settings`}{A list containing the analysis settings used to build
#'     the result object.}
#'   }
#'
#'   `print.loo_test()` prints a concise summary and the `results` table.
#'   `as.data.frame.loo_test()` returns the `results` table.
#'
#' @details
#' The function first runs the baseline analysis on the full dataset, then
#' removes one selected case at a time and reruns the analysis on the reduced
#' data.
#'
#' When `calib = "fixed"`, the reduced-data runs use the calibrated values
#' already present in `data`. When `calib = "recompute"`, the function rebuilds
#' the calibrated outcome and condition columns after each case deletion
#' using [QCA::findTh()] and [QCA::calibrate()] as specified in `calib_spec`.
#'
#' For each tested case, the function records whether the monitored solution
#' changed, whether monitored fit measures changed beyond `fit_tol`, and
#' whether the reduced-data run stopped with a calibration, truth table,
#' exclusion, or minimization error.
#'
#' The shared solution-control, exclusion, and returned-object conventions are
#' described in `?qcaERT_conventions`.
#'
#' @examples
#' \donttest{
#' library(QCA)
#' data(LR)
#'
#' conditions <- c("DEV", "URB", "LIT", "IND", "STB")
#' outcome <- "SURV"
#' dir_exp <- rep("1", length(conditions))
#'
#' thresholds <- list(
#'   DEV = findTh(LR$DEV, groups = 7),
#'   URB = findTh(LR$URB, groups = 4),
#'   LIT = findTh(LR$LIT, groups = 4),
#'   IND = findTh(LR$IND, groups = 4),
#'   STB = findTh(LR$STB, groups = 4),
#'   SURV = findTh(LR$SURV, groups = 4)
#' )
#'
#' dat <- LR
#' dat$DEV <- calibrate(LR$DEV, type = "fuzzy", thresholds = thresholds$DEV)
#' dat$URB <- calibrate(LR$URB, type = "fuzzy", thresholds = thresholds$URB)
#' dat$LIT <- calibrate(LR$LIT, type = "fuzzy", thresholds = thresholds$LIT)
#' dat$IND <- calibrate(LR$IND, type = "fuzzy", thresholds = thresholds$IND)
#' dat$STB <- calibrate(LR$STB, type = "fuzzy", thresholds = thresholds$STB)
#' dat$SURV <- calibrate(LR$SURV, type = "fuzzy", thresholds = thresholds$SURV)
#'
#' out <- loo.test(
#'   data = dat,
#'   outcome = outcome,
#'   conditions = conditions,
#'   cases = seq_len(nrow(dat)),
#'   case_labels = rownames(dat),
#'   calib = "fixed",
#'   incl.cut = 0.8,
#'   n.cut = 1,
#'   solution = "all",
#'   dir.exp = dir_exp,
#'   fit_measures = c("inclS", "PRI", "covS"),
#'   progress = TRUE
#' )
#'
#' out
#' as.data.frame(out)
#' }
#'
#' @seealso [calib.test()], [incl.test()], [ncut.test()], [subsample.test()],
#'   [altset.test()], [theory.test()], [cluster.test()], [sol.df()]
#' @export
loo.test <- function(
    data,
    outcome,
    conditions = NULL,
    cases = NULL,
    case_labels = NULL,
    calib = c("fixed", "recompute"),
    raw.data = NULL,
    calib_spec = NULL,
    incl.cut = 1,
    n.cut = 1,
    solution = "all",
    include = NULL,
    dir.exp = NULL,
    which_M = 1,
    i_mode = c("all", "C1P1"),
    exclude_mode = c("recompute", "static", "none"),
    exclude_recompute = list(type = 2),
    exclude_static = NULL,
    fit_measures = c("inclS", "PRI", "covS"),
    fit_tol = 0,
    progress = TRUE,
    ...
) {
  .require_qca()

  selection_controls <- .normalize_i_mode(i_mode)
  i_mode <- selection_controls$i_mode
  exclude_mode <- match.arg(exclude_mode)
  calib <- match.arg(calib)

  if (is.null(dim(data)) || is.null(nrow(data)) || nrow(data) < 2L) {
    stop("`data` must be a data frame object with at least 2 rows.")
  }

  if (!is.character(outcome) || length(outcome) != 1L || !nzchar(outcome)) {
    stop("`outcome` must be a single non-empty character string.")
  }

  if (!is.null(conditions)) {
    if (!is.character(conditions) || length(conditions) < 1L) {
      stop("`conditions` must be NULL or a non-empty character vector.")
    }
  }
  .validate_outcome_conditions_distinct(outcome, conditions)

  if (!is.numeric(incl.cut) || length(incl.cut) != 1L || !is.finite(incl.cut) ||
      incl.cut < 0 || incl.cut > 1) {
    stop("`incl.cut` must be a single finite number in [0, 1].")
  }

  n.cut <- .as_integerish_scalar(n.cut, "n.cut", min = 1L)

  which_M <- .as_integerish_scalar(which_M, "which_M", min = 1L)

  if (!is.numeric(fit_tol) || length(fit_tol) != 1L || !is.finite(fit_tol) || fit_tol < 0) {
    stop("`fit_tol` must be a single finite number >= 0.")
  }

  if (is.null(fit_measures)) {
    fit_measures <- character(0)
  } else {
    if (!is.character(fit_measures)) {
      stop("`fit_measures` must be NULL or a character vector.")
    }
    fit_measures <- unique(trimws(fit_measures))
    fit_measures <- fit_measures[nzchar(fit_measures)]
  }

  mc <- match.call(expand.dots = FALSE)
  dots_raw <- list(...)
  .reject_exclusion_controls_in_dots(dots_raw, "loo.test")
  .reject_calibration_inputs_in_dots(dots_raw, "loo.test")

  if (is.null(dir.exp)) {
    de <- .dot_get(dots_raw, "dir.exp")
    if (is.null(de)) de <- .dot_get(dots_raw, "direxp")
    if (!is.null(de)) dir.exp <- de
  }

  if (is.null(include)) {
    inc <- .dot_get(dots_raw, "include")
    if (!is.null(inc)) include <- inc
  }

  solution_controls <- .resolve_solution_controls(
    solution = solution,
    include = include,
    dir.exp = dir.exp,
    caller = "loo.test",
    style = "std"
  )

  solution <- solution_controls$solution
  include <- solution_controls$include
  monitored_solutions <- solution_controls$monitored

  .validate_exclusion_controls(
    exclude_mode = exclude_mode,
    exclude_recompute = exclude_recompute,
    exclude_static = exclude_static,
    exclude_recompute_supplied = !is.null(mc$exclude_recompute),
    exclude_static_supplied = !is.null(mc$exclude_static),
    monitored_solutions = monitored_solutions,
    style = "std"
  )

  dots_split <- .split_truth_table_minimize_dots(dots_raw)
  dots_tt <- dots_split$tt
  dots_min <- dots_split$min

  analysis_vars <- NULL
  raw_data_work <- NULL

  if (calib == "recompute") {
    if (is.null(conditions)) {
      stop("When `calib = \"recompute\"`, `conditions` must be supplied explicitly.")
    }

    analysis_vars <- c(outcome, conditions)

    if (is.null(raw.data)) {
      stop("When `calib = \"recompute\"`, `raw.data` must be supplied.")
    }
    if (is.null(dim(raw.data)) || is.null(nrow(raw.data)) || nrow(raw.data) != nrow(data)) {
      stop("When `calib = \"recompute\"`, `raw.data` must be a data frame object with the same number of rows as `data`.")
    }

    raw_data_work <- .reduced_as_df(raw.data)
    calib_spec <- .reduced_validate_calib_spec(
      calib_spec = calib_spec,
      analysis_vars = analysis_vars,
      raw.data = raw_data_work,
      data = data
    )
  }

  .build_truth_table <- function(data_step) {
    args <- c(
      list(
        data = data_step,
        outcome = outcome,
        incl.cut = incl.cut,
        n.cut = n.cut
      ),
      dots_tt
    )

    if (!is.null(conditions)) {
      args$conditions <- conditions
    }

    do.call(QCA::truthTable, args)
  }

  .run_once <- function(data_step, raw_step = NULL) {
    .reduced_run_once(
      data_step = data_step,
      raw_step = raw_step,
      calib = calib,
      recalibrate_dataset = function(data_step, raw_step) {
        .reduced_recalibrate_dataset(
          data_step = data_step,
          raw_step = raw_step,
          analysis_vars = analysis_vars,
          calib_spec = calib_spec
        )
      },
      build_truth_table = .build_truth_table,
      monitored_solutions = monitored_solutions,
      exclude_mode = exclude_mode,
      exclude_recompute = exclude_recompute,
      exclude_static = exclude_static,
      dots_min = dots_min,
      dir.exp = dir.exp,
      which_M = which_M,
      i_mode = i_mode,
      fit_measures = fit_measures,
      include_analysis_data = FALSE
    )
  }

  case_ref <- .reduced_resolve_case_reference(cases = cases, data = data, case_labels = case_labels)

  total_cases <- nrow(case_ref)

  progress_state <- .new_qcaert_progress(total = total_cases, progress = progress)
  on.exit(progress_state$close(), add = TRUE)
  .bump_pb <- progress_state$tick

  baseline <- .run_once(
    data_step = data,
    raw_step = if (calib == "recompute") raw_data_work else NULL
  )

  baseline_solution_types <- .reduced_baseline_solution_types(
    base = baseline,
    monitored_solutions = monitored_solutions
  )
  baseline_fit_solution_types <- .reduced_baseline_solution_types(
    base = baseline,
    monitored_solutions = monitored_solutions,
    fit_measures = fit_measures,
    require_fit = TRUE
  )
  baseline_missing <- baseline$status == "ok" && length(baseline_solution_types) == 0L
  baseline_invalid <- baseline$status != "ok" || length(baseline_solution_types) == 0L

  diag_rows <- list()
  by_case <- list()

  for (ii in seq_len(nrow(case_ref))) {
    row_idx <- case_ref$row_index[ii]
    case_label_i <- case_ref$case_label[ii]
    key <- paste0(row_idx, ":", case_label_i)

    if (baseline_invalid) {
      status <- if (baseline$status != "ok") {
        .reduced_stop_reason_from_status(baseline$status, baseline = TRUE)
      } else if (baseline_missing) {
        "baseline_selected_solution_missing"
      } else {
        "baseline_no_comparable_solution"
      }

      diag_rows[[length(diag_rows) + 1L]] <- data.frame(
        row_index = row_idx,
        case_label = case_label_i,
        status = status,
        changed = NA,
        solution_changed = NA,
        changed_types = NA_character_,
        change_kind = NA_character_,
        fit_changed = NA,
        fit_changed_types = NA_character_,
        n_fit_deltas = NA_integer_,
        max_abs_fit_delta = NA_real_,
        error_source = baseline$error_source,
        error_message = baseline$error_message,
        stringsAsFactors = FALSE
      )

      by_case[[key]] <- list(
        row_index = row_idx,
        case_label = case_label_i,
        status = status,
        baseline = baseline,
        reduced = NULL,
        solution_changed = NA,
        changed_types = NA_character_,
        change_kind = NA_character_,
        fit_changed = NA,
        fit_changed_types = NA_character_,
        fit_changed_measures = character(0),
        fit_delta = numeric(0),
        n_fit_deltas = NA_integer_,
        max_abs_fit_delta = NA_real_,
        exclude_baseline = baseline$exclude_used,
        exclude_reduced = NULL,
        calibration_baseline = baseline$calibration,
        calibration_reduced = NULL,
        baseline_fit = baseline$fit,
        reduced_fit = NULL,
        error_source = baseline$error_source,
        error_message = baseline$error_message
      )

      .bump_pb()
      next
    }

    data_reduced <- data[-row_idx, , drop = FALSE]
    raw_reduced <- if (calib == "recompute") raw_data_work[-row_idx, , drop = FALSE] else NULL

    cur <- .run_once(
      data_step = data_reduced,
      raw_step = raw_reduced
    )

    comparable_solution_types <- .reduced_comparable_solution_types(
      base = baseline,
      cur = cur,
      monitored_solutions = baseline_solution_types
    )

    if (cur$status != "ok" || length(comparable_solution_types) == 0L) {
      status_info <- .reduced_no_comparable_status(
        cur = cur,
        candidate_solution_types = baseline_solution_types
      )
      status <- status_info$status

      diag_rows[[length(diag_rows) + 1L]] <- data.frame(
        row_index = row_idx,
        case_label = case_label_i,
        status = status,
        changed = NA,
        solution_changed = NA,
        changed_types = NA_character_,
        change_kind = NA_character_,
        fit_changed = NA,
        fit_changed_types = NA_character_,
        n_fit_deltas = NA_integer_,
        max_abs_fit_delta = NA_real_,
        error_source = status_info$error_source,
        error_message = status_info$error_message,
        stringsAsFactors = FALSE
      )

      by_case[[key]] <- list(
        row_index = row_idx,
        case_label = case_label_i,
        status = status,
        baseline = baseline,
        reduced = cur,
        solution_changed = NA,
        changed_types = NA_character_,
        change_kind = NA_character_,
        fit_changed = NA,
        fit_changed_types = NA_character_,
        fit_changed_measures = character(0),
        fit_delta = numeric(0),
        n_fit_deltas = NA_integer_,
        max_abs_fit_delta = NA_real_,
        exclude_baseline = baseline$exclude_used,
        exclude_reduced = cur$exclude_used,
        calibration_baseline = baseline$calibration,
        calibration_reduced = cur$calibration,
        baseline_fit = baseline$fit,
        reduced_fit = NULL,
        error_source = status_info$error_source,
        error_message = status_info$error_message
      )

      .bump_pb()
      next
    }

    sol_info <- .reduced_solution_change_info(
      base = baseline,
      cur = cur,
      monitored_solutions = baseline_solution_types
    )

    fit_info <- .reduced_fit_change_info(
      base = baseline,
      cur = cur,
      monitored_solutions = baseline_fit_solution_types,
      fit_tol = fit_tol,
      fit_measures = fit_measures
    )

    changed_any <- isTRUE(sol_info$changed) || isTRUE(fit_info$changed)

    diag_rows[[length(diag_rows) + 1L]] <- data.frame(
      row_index = row_idx,
      case_label = case_label_i,
      status = "ok",
      changed = changed_any,
      solution_changed = isTRUE(sol_info$changed),
      changed_types = sol_info$changed_types,
      change_kind = sol_info$change_kind,
      fit_changed = isTRUE(fit_info$changed),
      fit_changed_types = fit_info$changed_types,
      n_fit_deltas = as.integer(fit_info$n_changed),
      max_abs_fit_delta = fit_info$max_abs_delta,
      error_source = NA_character_,
      error_message = NA_character_,
      stringsAsFactors = FALSE
    )

    by_case[[key]] <- list(
      row_index = row_idx,
      case_label = case_label_i,
      status = "ok",
      baseline = baseline,
      reduced = cur,
      solution_changed = isTRUE(sol_info$changed),
      changed_types = sol_info$changed_types,
      change_kind = sol_info$change_kind,
      fit_changed = isTRUE(fit_info$changed),
      fit_changed_types = fit_info$changed_types,
      fit_changed_measures = fit_info$changed_measures,
      fit_delta = fit_info$delta,
      n_fit_deltas = as.integer(fit_info$n_changed),
      max_abs_fit_delta = fit_info$max_abs_delta,
      exclude_baseline = baseline$exclude_used,
      exclude_reduced = cur$exclude_used,
      calibration_baseline = baseline$calibration,
      calibration_reduced = cur$calibration,
      baseline_fit = baseline$fit,
      reduced_fit = cur$fit,
      error_source = NA_character_,
      error_message = NA_character_
    )

    .bump_pb()
  }

  diagnostics <- .bind_rows_result_sorted(diag_rows, order_by = "row_index")

  .new_result_object(
    "loo_test",
    diagnostics = diagnostics,
    results = .make_loo_results(diagnostics),
    baseline = baseline,
    by_case = by_case,
    settings = list(
      outcome = outcome,
      conditions = conditions,
      cases = cases,
      case_labels = case_labels,
      calib = calib,
      n_cases_tested = nrow(case_ref),
      incl.cut = incl.cut,
      n.cut = n.cut,
      solution = solution,
      monitored_solutions = monitored_solutions,
      include = include,
      dir.exp = dir.exp,
      which_M = which_M,
      i_mode = i_mode,
      exclude_mode = exclude_mode,
      exclude_recompute = exclude_recompute,
      exclude_static = exclude_static,
      fit_measures = fit_measures,
      fit_tol = fit_tol
    )
  )
}

#' Print a `loo_test` object
#'
#' @rdname loo.test
#' @export
print.loo_test <- function(x, row.names = FALSE, ...) {
  results <- x$results
  settings <- x$settings

  .print_qcaert_heading("loo_test", "leave-one-out robustness", settings)
  if (!is.null(settings$n_cases_tested)) {
    cat("Cases tested: ", settings$n_cases_tested, "\n", sep = "")
  }
  if (!is.null(settings$fit_measures) && length(settings$fit_measures) > 0L) {
    cat("Fit measures: ", paste(settings$fit_measures, collapse = ", "), "\n", sep = "")
  }

  .print_qcaert_table(results, "Result", row.names = row.names, ...)

  cat("\nSummary\n")
  if (is.data.frame(results) && nrow(results) > 0L) {
    n_ok <- sum(results$status == "ok", na.rm = TRUE)
    n_solution_changed <- .count_present(results$solution_change)
    n_fit_changed <- .count_present(results$fit_changed_types)
    n_errors <- sum(!is.na(results$status) & results$status != "ok", na.rm = TRUE)

    cat(" Successful runs: ", n_ok, "/", nrow(results), "\n", sep = "")
    cat(" Solution changes: ", n_solution_changed, "\n", sep = "")
    cat(" Fit changes: ", n_fit_changed, "\n", sep = "")
    cat(" Non-ok runs: ", n_errors, "\n", sep = "")

    if (n_solution_changed > 0L) {
      labs <- results$case_label[!is.na(results$solution_change)]
      cat(" Cases with solution changes: ", paste(labs, collapse = ", "), "\n", sep = "")
    }

    if (n_errors > 0L) {
      labs <- results$case_label[!is.na(results$status) & results$status != "ok"]
      cat(" Non-ok cases: ", paste(labs, collapse = ", "), "\n", sep = "")
    }
  }

  invisible(x)
}

#' Return the main results table from a `loo_test` object
#'
#' @rdname loo.test
#' @export
as.data.frame.loo_test <- function(x, ...) {
  .as.data.frame_results(x, ...)
}

#' Choose a qcaERT robustness tool
#'
#' This page maps common QCA robustness tests to their respective functions
#'
#' @section Calibration:
#' Use [calib.test()] when you want to know whether a QCA solution is sensitive
#' to calibration thresholds. It perturbs one calibration anchor at a time and
#' reports how far each threshold can move before the monitored solution
#' changes or the search reaches a boundary.
#'
#' You can also use [altset.test()] when you want calibration perturbations to
#' be combined with other perturbations in sampled alternative analysis settings,
#' such as alternative inclusion cutoffs or frequency cutoffs.
#'
#' @section Inclusion or n.cut:
#' Use [incl.test()] to check the inclusion (consistency) cutoff for
#' sufficiency. It searches below and above the baseline `incl.cut` and records
#' the last safe cutoff and first failing cutoff in each direction.
#'
#' Use [ncut.test()] when the concern is the minimum number of cases under which
#' a truth table row is declared as a remainder. It searches lower and upper
#' `n.cut` values and records when the monitored solution changes or the
#' feasible boundary is reached.
#'
#' @section Cases:
#' Use [loo.test()] when you suspect individual cases may drive the solution.
#' It removes selected cases one at a time and compares each reduced analysis
#' with the baseline.
#'
#' Use [subsample.test()] when you want repeated partial-sample checks. It draws
#' subsamples, rebuilds the QCA analysis, and summarizes how often the baseline
#' solution is preserved across runs. This is a stringent, punishing robustness
#' check and is most useful when sample-composition sensitivity is substantively
#' important.
#'
#' @section Groups or clusters:
#' Use [cluster.test()] when you expect heterogeneity across clusters, groups,
#' or repeated units. It compares cluster-specific consistency, coverage, and
#' related fit patterns against the overall truth table.
#'
#' @section Theory Specifications:
#' Use [theory.test()] when you want to compare several theoretically motivated
#' condition sets under the same outcome, truth-table cutoffs, solution type,
#' exclusion handling, and settings for `which_M` and `i_mode`.
#'
#' @section Reporting QCA Solutions:
#' Use [sol.df()] when you want QCA minimization objects turned into a compact,
#' data frame. Use [sol.chart()] when you want that table
#' rendered as a visual chart of prime implicants.
#'
#' @section How To Read The Results:
#' Most qcaERT robustness functions return `diagnostics`, `results`, and
#' `settings`. `diagnostics` is the detailed table; `results` is the clean table
#' returned by `as.data.frame()`. `print()` shows a concise summary. The shared
#' output structure is described in `?qcaERT_conventions`.
#'
#' For [calib.test()], [incl.test()], and [theory.test()] objects, `plot()`
#' methods provide optional visual inspection helpers when `ggplot2` is
#' installed. [sol.chart()] provides the corresponding visual presentation for
#' [sol.df()] tables. See `?qcaERT_plots`.
#'
#' @examples
#' \donttest{
#' ?qcaERT_tests
#' ?qcaERT_conventions
#' ?qcaERT_plots
#' }
#'
#' @name qcaERT_tests
NULL

qcaert_truthy_env <- function(name) {
  tolower(Sys.getenv(name, unset = "false")) %in% c("1", "true", "yes", "on")
}

qcaert_stress_enabled <- function() {
  qcaert_truthy_env("QCAERT_STRESS")
}

skip_qcaert_stress <- function() {
  skip_if_not(
    qcaert_stress_enabled(),
    "Set QCAERT_STRESS=true to run optional stress tests."
  )
}

.normalize_theory_names <- function(names, label) {
  if (is.null(names)) {
    stop("`", label, "` must be named.")
  }

  names <- trimws(as.character(names))
  if (any(is.na(names)) || any(!nzchar(names))) {
    stop("`", label, "` names must be non-empty character strings.")
  }
  if (anyDuplicated(names) > 0L) {
    stop("`", label, "` names must be unique.")
  }

  names
}

.normalize_theories <- function(theories, data, outcome) {
  if (!is.list(theories) || is.data.frame(theories) || length(theories) < 2L) {
    stop("`theories` must be a named list with at least two theory-specific condition sets.")
  }

  theory_names <- .normalize_theory_names(names(theories), "theories")
  names(theories) <- theory_names

  out <- lapply(theory_names, function(theory) {
    conditions <- theories[[theory]]
    if (!is.character(conditions) || length(conditions) < 1L) {
      stop("Each entry in `theories` must be a non-empty character vector of condition names.")
    }

    conditions <- trimws(as.character(conditions))
    if (any(is.na(conditions)) || any(!nzchar(conditions))) {
      stop("Condition names in `theories` must be non-empty character strings.")
    }
    if (anyDuplicated(conditions) > 0L) {
      stop("Theory `", theory, "` contains duplicate condition names.")
    }
    if (outcome %in% conditions) {
      stop("Theory `", theory, "` includes the outcome as a condition; the outcome is handled separately.")
    }

    missing <- setdiff(conditions, colnames(data))
    if (length(missing) > 0L) {
      stop(
        "Theory `", theory, "` contains condition(s) not found in `data`: ",
        paste(missing, collapse = ", ")
      )
    }

    conditions
  })

  names(out) <- theory_names
  out
}

.normalize_theory_dir_exp_one <- function(dir_exp, theory, conditions) {
  if (is.character(dir_exp) && length(dir_exp) > 1L) {
    nms <- names(dir_exp)
    if (!is.null(nms) && all(nzchar(nms))) {
      nms <- trimws(as.character(nms))
      if (!setequal(nms, conditions)) {
        stop(
          "`dir.exp[[\"", theory, "\"]]` names must match that theory's conditions."
        )
      }
      dir_exp <- unname(dir_exp[match(conditions, nms)])
    }
  }

  .normalize_dir_exp_generic(
    dir_exp,
    conditions = conditions,
    endpoint_phrase = paste0("the conditions for theory `", theory, "`")
  )
}

.normalize_theory_dir_exp <- function(dir.exp, theories, solution) {
  if (is.null(dir.exp)) {
    if (identical(solution, "intermediate")) {
      stop("When `solution = \"intermediate\"`/`\"int\"`, `dir.exp` must be provided as a named list with one entry per theory.")
    }
    return(NULL)
  }

  if (!is.list(dir.exp) || is.data.frame(dir.exp)) {
    stop("For `theory.test()`, `dir.exp` must be NULL or a named list with one entry per theory.")
  }

  dir_names <- .normalize_theory_names(names(dir.exp), "dir.exp")
  theory_names <- names(theories)
  missing <- setdiff(theory_names, dir_names)
  extra <- setdiff(dir_names, theory_names)
  if (length(missing) > 0L || length(extra) > 0L) {
    msg <- "`dir.exp` names must match `theories` names."
    if (length(missing) > 0L) {
      msg <- paste0(msg, " Missing: ", paste(missing, collapse = ", "), ".")
    }
    if (length(extra) > 0L) {
      msg <- paste0(msg, " Unknown: ", paste(extra, collapse = ", "), ".")
    }
    stop(msg)
  }

  out <- lapply(theory_names, function(theory) {
    .normalize_theory_dir_exp_one(dir.exp[[theory]], theory, theories[[theory]])
  })
  names(out) <- theory_names
  out
}

.validate_theory_static_exclude <- function(exclude_mode, exclude_static, theories, monitored_solutions) {
  if (!identical(exclude_mode, "static")) {
    return(invisible(TRUE))
  }

  if (!any(monitored_solutions %in% c("parsimonious", "intermediate"))) {
    return(invisible(TRUE))
  }

  if (is.null(exclude_static) || !is.list(exclude_static) || is.data.frame(exclude_static)) {
    stop("For `theory.test()`, `exclude_mode = \"static\"` requires `exclude_static` as a named list with one entry per theory.")
  }

  exclude_names <- .normalize_theory_names(names(exclude_static), "exclude_static")
  theory_names <- names(theories)
  missing <- setdiff(theory_names, exclude_names)
  extra <- setdiff(exclude_names, theory_names)
  if (length(missing) > 0L || length(extra) > 0L) {
    msg <- "`exclude_static` names must match `theories` names when `exclude_mode = \"static\"`."
    if (length(missing) > 0L) {
      msg <- paste0(msg, " Missing: ", paste(missing, collapse = ", "), ".")
    }
    if (length(extra) > 0L) {
      msg <- paste0(msg, " Unknown: ", paste(extra, collapse = ", "), ".")
    }
    stop(msg)
  }

  invisible(TRUE)
}

.empty_theory_diagnostics <- function() {
  data.frame(
    theory = character(0),
    solution_type = character(0),
    status = character(0),
    n_conditions = integer(0),
    conditions = character(0),
    n_tt_rows = integer(0),
    n_observed_rows = integer(0),
    n_remainders = integer(0),
    n_excluded = integer(0),
    selected_solution_missing = logical(0),
    error_source = character(0),
    error_message = character(0),
    stringsAsFactors = FALSE
  )
}

.empty_theory_models <- function() {
  data.frame(
    theory = character(0),
    solution_type = character(0),
    intermediate_branch = character(0),
    status = character(0),
    n_conditions = integer(0),
    n_tt_rows = integer(0),
    n_observed_rows = integer(0),
    n_remainders = integer(0),
    n_excluded = integer(0),
    n_models = integer(0),
    selected_model = integer(0),
    n_terms = integer(0),
    inclS = numeric(0),
    PRI = numeric(0),
    covS = numeric(0),
    stringsAsFactors = FALSE
  )
}

.empty_theory_pairwise <- function() {
  data.frame(
    solution_type = character(0),
    intermediate_branch = character(0),
    theory_1 = character(0),
    theory_2 = character(0),
    delta_inclS = numeric(0),
    delta_PRI = numeric(0),
    delta_covS = numeric(0),
    membership_jaccard = numeric(0),
    mean_abs_membership_delta = numeric(0),
    stringsAsFactors = FALSE
  )
}

.empty_theory_solutions <- function() {
  data.frame(
    theory = character(0),
    solution_type = character(0),
    model = integer(0),
    intermediate_branch = character(0),
    prime_implicant = character(0),
    inclS = numeric(0),
    PRI = numeric(0),
    covS = numeric(0),
    stringsAsFactors = FALSE
  )
}

.empty_theory_results <- function() {
  list(
    models = .empty_theory_models(),
    pairwise = .empty_theory_pairwise(),
    solutions = .empty_theory_solutions()
  )
}

.theory_tt_stats <- function(tt_obj) {
  empty <- list(
    n_tt_rows = NA_integer_,
    n_observed_rows = NA_integer_,
    n_remainders = NA_integer_
  )

  if (is.null(tt_obj) || is.null(tt_obj$tt) || !is.data.frame(tt_obj$tt)) {
    return(empty)
  }

  n_values <- if ("n" %in% names(tt_obj$tt)) {
    suppressWarnings(as.numeric(tt_obj$tt$n))
  } else {
    rep(NA_real_, nrow(tt_obj$tt))
  }

  list(
    n_tt_rows = as.integer(nrow(tt_obj$tt)),
    n_observed_rows = as.integer(sum(!is.na(n_values) & n_values > 0)),
    n_remainders = as.integer(sum(!is.na(n_values) & n_values == 0))
  )
}

.theory_n_excluded <- function(exclude_used) {
  if (is.null(exclude_used)) {
    return(0L)
  }
  as.integer(length(exclude_used))
}

.theory_payload_status <- function(payload, monitored_solutions) {
  solution_type_status <- payload$solution_type_status[monitored_solutions]
  if (all(!is.na(solution_type_status) & solution_type_status == "ok")) {
    return(list(
      status = "ok",
      error_source = NA_character_,
      error_message = NA_character_
    ))
  }

  if (any(solution_type_status == "ok", na.rm = TRUE)) {
    first_bad <- which(!is.na(solution_type_status) & solution_type_status != "ok")[1L]
    solution_type <- monitored_solutions[first_bad]
    return(list(
      status = "partial",
      error_source = unname(payload$solution_type_error_source[[solution_type]]),
      error_message = unname(payload$solution_type_error_message[[solution_type]])
    ))
  }

  first_bad <- which(!is.na(solution_type_status) & nzchar(solution_type_status))[1L]
  if (is.na(first_bad)) {
    return(list(
      status = "run_error",
      error_source = NA_character_,
      error_message = NA_character_
    ))
  }

  solution_type <- monitored_solutions[first_bad]
  list(
    status = unname(payload$solution_type_status[[solution_type]]),
    error_source = unname(payload$solution_type_error_source[[solution_type]]),
    error_message = unname(payload$solution_type_error_message[[solution_type]])
  )
}

.theory_make_run_result <- function(
    theory,
    conditions,
    dir.exp,
    status,
    tt = NULL,
    exclude_used = NULL,
    payload = NULL,
    error_source = NA_character_,
    error_message = NA_character_
) {
  if (is.null(payload)) {
    payload <- .reduced_blank_payload()
  }

  c(
    list(
      theory = theory,
      conditions = conditions,
      dir.exp = dir.exp,
      status = status,
      tt = tt,
      exclude_used = exclude_used
    ),
    payload,
    list(
      error_source = error_source,
      error_message = error_message
    )
  )
}

.theory_build_truth_table <- function(data, outcome, conditions, incl.cut, n.cut, dots_tt) {
  args <- c(
    list(
      data = data,
      outcome = outcome,
      conditions = conditions,
      incl.cut = incl.cut,
      n.cut = n.cut
    ),
    dots_tt
  )

  tryCatch(
    suppressWarnings(do.call(QCA::truthTable, args)),
    error = function(e) e
  )
}

.theory_run_one <- function(
    theory,
    data,
    outcome,
    conditions,
    dir.exp,
    incl.cut,
    n.cut,
    monitored_solutions,
    exclude_mode,
    exclude_recompute,
    exclude_static,
    dots_tt,
    dots_min,
    which_M,
    i_mode
) {
  tt_obj <- .theory_build_truth_table(
    data = data,
    outcome = outcome,
    conditions = conditions,
    incl.cut = incl.cut,
    n.cut = n.cut,
    dots_tt = dots_tt
  )

  if (inherits(tt_obj, "error")) {
    payload <- .reduced_blank_payload()
    payload$solution_type_status[monitored_solutions] <- "truth_table_error"
    payload$solution_type_error_source[monitored_solutions] <- "truthTable"
    payload$solution_type_error_message[monitored_solutions] <- .reduced_error_message(tt_obj)

    return(.theory_make_run_result(
      theory = theory,
      conditions = conditions,
      dir.exp = dir.exp,
      status = "truth_table_error",
      tt = NULL,
      exclude_used = NULL,
      payload = payload,
      error_source = "truthTable",
      error_message = .reduced_error_message(tt_obj)
    ))
  }

  need_exclude <- any(monitored_solutions %in% c("parsimonious", "intermediate")) &&
    exclude_mode != "none"

  exclude_i <- if (identical(exclude_mode, "static") && !is.null(exclude_static)) {
    exclude_static[[theory]]
  } else {
    exclude_static
  }

  ex <- NULL
  ex_error <- NULL
  if (need_exclude) {
    ex <- .reduced_compute_exclude(
      tt_obj = tt_obj,
      exclude_mode = exclude_mode,
      exclude_recompute = exclude_recompute,
      exclude_static = exclude_i
    )

    if (inherits(ex, "error")) {
      ex_error <- ex
      ex <- NULL
    }
  }

  payload <- .reduced_blank_payload()

  for (solution_type in monitored_solutions) {
    if (solution_type %in% c("parsimonious", "intermediate") && inherits(ex_error, "error")) {
      payload$solution_type_status[[solution_type]] <- "exclude_error"
      payload$solution_type_error_source[[solution_type]] <- "exclude"
      payload$solution_type_error_message[[solution_type]] <- .reduced_error_message(ex_error)
      next
    }

    rr <- .reduced_run_solution_type(
      tt_obj = tt_obj,
      solution_type = solution_type,
      ex = ex,
      dots_min = dots_min,
      dir.exp = dir.exp,
      which_M = which_M,
      i_mode = i_mode,
      fit_measures = character(0)
    )

    if (!is.null(rr$error)) {
      payload$solution_type_status[[solution_type]] <- "minimize_error"
      payload$solution_type_error_source[[solution_type]] <- solution_type
      payload$solution_type_error_message[[solution_type]] <- .reduced_error_message(rr$error)
      next
    }

    payload$res[[solution_type]] <- rr$res
    payload$sig[[solution_type]] <- rr$sig
    payload$selected_solution_missing[[solution_type]] <- isTRUE(rr$selected_solution_missing)
    payload$fit[[solution_type]] <- rr$fit
    payload$fit_missing[[solution_type]] <- isTRUE(rr$fit_missing)
    payload$meta[[solution_type]] <- rr$meta
    payload$solution_type_status[[solution_type]] <- "ok"
    payload$solution_type_error_source[[solution_type]] <- NA_character_
    payload$solution_type_error_message[[solution_type]] <- NA_character_
  }

  run_status <- .theory_payload_status(payload, monitored_solutions)

  .theory_make_run_result(
    theory = theory,
    conditions = conditions,
    dir.exp = dir.exp,
    status = run_status$status,
    tt = tt_obj,
    exclude_used = ex,
    payload = payload,
    error_source = run_status$error_source,
    error_message = run_status$error_message
  )
}

.theory_diagnostics_from_runs <- function(by_theory, monitored_solutions) {
  rows <- list()
  for (theory in names(by_theory)) {
    run <- by_theory[[theory]]
    stats <- .theory_tt_stats(run$tt)
    n_excluded <- .theory_n_excluded(run$exclude_used)

    for (solution_type in monitored_solutions) {
      status <- run$solution_type_status[[solution_type]]
      if (is.null(status) || is.na(status)) {
        status <- run$status
      }

      rows[[length(rows) + 1L]] <- data.frame(
        theory = theory,
        solution_type = solution_type,
        status = status,
        n_conditions = length(run$conditions),
        conditions = paste(run$conditions, collapse = ","),
        n_tt_rows = stats$n_tt_rows,
        n_observed_rows = stats$n_observed_rows,
        n_remainders = stats$n_remainders,
        n_excluded = if (solution_type %in% c("parsimonious", "intermediate")) n_excluded else 0L,
        selected_solution_missing = run$selected_solution_missing[[solution_type]] %in% TRUE,
        error_source = run$solution_type_error_source[[solution_type]],
        error_message = run$solution_type_error_message[[solution_type]],
        stringsAsFactors = FALSE
      )
    }
  }

  if (length(rows) == 0L) {
    return(.empty_theory_diagnostics())
  }

  .bind_rows_result(rows)
}

.theory_fit_value <- function(fit_table, row, column) {
  if (is.null(fit_table) || !is.data.frame(fit_table) || nrow(fit_table) < row ||
      !column %in% colnames(fit_table)) {
    return(NA_real_)
  }

  suppressWarnings(as.numeric(fit_table[[column]][row]))
}

.theory_selected_ic <- function(ic, which_M, has_model) {
  if (!isTRUE(has_model) || is.null(ic)) {
    return(NULL)
  }

  if (!is.null(ic[["individual"]])) {
    if (length(ic[["individual"]]) < which_M) {
      return(NULL)
    }
    return(ic[["individual"]][[which_M]])
  }

  ic
}

.theory_selected_fit_row <- function(ic, which_M, has_model) {
  if (!isTRUE(has_model) || is.null(ic)) {
    return(NA_integer_)
  }

  if (!is.null(ic[["individual"]])) {
    return(1L)
  }

  as.integer(which_M)
}

.theory_selected_terms <- function(sol_all, primes, which_M) {
  n_models <- if (!is.null(sol_all)) length(sol_all) else 0L
  if (n_models < which_M) {
    return(NULL)
  }

  terms <- .solution_terms_from_primes(sol_all[[which_M]], primes)
  if (identical(terms, "<EMPTY>")) {
    return(character(0))
  }
  terms
}

.theory_standard_model_info <- function(res, solution_type, which_M) {
  sol_all <- if (!is.null(res) && !inherits(res, "error")) res$solution else NULL
  primes <- if (!is.null(res) && !inherits(res, "error")) res$primes else NULL
  ic <- if (!is.null(res) && !inherits(res, "error")) res$IC else NULL

  n_models <- if (!is.null(sol_all)) length(sol_all) else 0L
  has_model <- n_models >= which_M
  terms <- .theory_selected_terms(sol_all, primes, which_M)
  selected_ic <- .theory_selected_ic(
    ic = ic,
    which_M = which_M,
    has_model = has_model
  )
  fit_row <- .theory_selected_fit_row(ic, which_M = which_M, has_model = has_model)
  fit_table <- if (!is.null(selected_ic)) selected_ic$sol.incl.cov else NULL

  data.frame(
    solution_type = solution_type,
    intermediate_branch = NA_character_,
    n_models = as.integer(n_models),
    selected_model = if (has_model) as.integer(which_M) else NA_integer_,
    n_terms = if (has_model) as.integer(length(terms)) else NA_integer_,
    inclS = if (has_model) .theory_fit_value(fit_table, fit_row, "inclS") else NA_real_,
    PRI = if (has_model) .theory_fit_value(fit_table, fit_row, "PRI") else NA_real_,
    covS = if (has_model) .theory_fit_value(fit_table, fit_row, "covS") else NA_real_,
    stringsAsFactors = FALSE
  )
}

.theory_intermediate_model_info <- function(res, which_M, i_mode) {
  i.sol <- if (!is.null(res) && !inherits(res, "error")) res$i.sol else NULL

  if (is.null(i.sol) || length(i.sol) == 0L) {
    return(data.frame(
      solution_type = "intermediate",
      intermediate_branch = NA_character_,
      n_models = 0L,
      selected_model = NA_integer_,
      n_terms = NA_integer_,
      inclS = NA_real_,
      PRI = NA_real_,
      covS = NA_real_,
      stringsAsFactors = FALSE
    ))
  }

  branch_names <- .normalize_i_solution_names(i.sol)
  if (i_mode == "C1P1") {
    idx <- which(branch_names == "C1P1")
    if (length(idx) == 0L) {
      return(data.frame(
        solution_type = "intermediate",
        intermediate_branch = "C1P1",
        n_models = 0L,
        selected_model = NA_integer_,
        n_terms = NA_integer_,
        inclS = NA_real_,
        PRI = NA_real_,
        covS = NA_real_,
        stringsAsFactors = FALSE
      ))
    }

    i.sol <- i.sol[idx[1L]]
    branch_names <- branch_names[idx[1L]]
  }

  rows <- lapply(seq_along(i.sol), function(i) {
    branch <- i.sol[[i]]
    sol_all <- if (is.list(branch) && "solution" %in% names(branch)) branch$solution else NULL
    primes <- if (is.list(branch) && "primes" %in% names(branch)) branch$primes else NULL

    n_models <- if (!is.null(sol_all)) length(sol_all) else 0L
    has_model <- n_models >= which_M
    terms <- .theory_selected_terms(sol_all, primes, which_M)
    ic <- if (is.list(branch) && "IC" %in% names(branch)) branch$IC else NULL
    selected_ic <- .theory_selected_ic(
      ic = ic,
      which_M = which_M,
      has_model = has_model
    )
    fit_row <- .theory_selected_fit_row(ic, which_M = which_M, has_model = has_model)
    fit_table <- if (!is.null(selected_ic)) selected_ic$sol.incl.cov else NULL

    data.frame(
      solution_type = "intermediate",
      intermediate_branch = branch_names[i],
      n_models = as.integer(n_models),
      selected_model = if (has_model) as.integer(which_M) else NA_integer_,
      n_terms = if (has_model) as.integer(length(terms)) else NA_integer_,
      inclS = if (has_model) .theory_fit_value(fit_table, fit_row, "inclS") else NA_real_,
      PRI = if (has_model) .theory_fit_value(fit_table, fit_row, "PRI") else NA_real_,
      covS = if (has_model) .theory_fit_value(fit_table, fit_row, "covS") else NA_real_,
      stringsAsFactors = FALSE
    )
  })

  .bind_rows_result(rows)
}

.theory_term_fit_value <- function(incl_cov, term, column) {
  if (is.null(incl_cov) || !is.data.frame(incl_cov) ||
      !column %in% colnames(incl_cov) || nrow(incl_cov) == 0L) {
    return(NA_real_)
  }

  rn <- rownames(incl_cov)
  if (is.null(rn) || length(rn) == 0L) {
    return(NA_real_)
  }

  idx <- match(.qca_canonicalize_names(term), .qca_canonicalize_names(rn))
  if (is.na(idx)) {
    return(NA_real_)
  }

  suppressWarnings(as.numeric(incl_cov[[column]][idx]))
}

.theory_solution_rows <- function(theory, solution_type, model, intermediate_branch, terms, incl_cov) {
  if (length(terms) == 0L) {
    return(.empty_theory_solutions())
  }

  data.frame(
    theory = rep(theory, length(terms)),
    solution_type = rep(solution_type, length(terms)),
    model = rep(as.integer(model), length(terms)),
    intermediate_branch = rep(intermediate_branch, length(terms)),
    prime_implicant = as.character(terms),
    inclS = vapply(terms, .theory_term_fit_value, numeric(1), incl_cov = incl_cov, column = "inclS"),
    PRI = vapply(terms, .theory_term_fit_value, numeric(1), incl_cov = incl_cov, column = "PRI"),
    covS = vapply(terms, .theory_term_fit_value, numeric(1), incl_cov = incl_cov, column = "covS"),
    stringsAsFactors = FALSE
  )
}

.theory_standard_solution_rows <- function(theory, res, solution_type, which_M) {
  if (is.null(res) || inherits(res, "error")) {
    return(.empty_theory_solutions())
  }

  sol_all <- res$solution
  primes <- res$primes
  n_models <- if (!is.null(sol_all)) length(sol_all) else 0L
  has_model <- n_models >= which_M
  terms <- .theory_selected_terms(sol_all, primes, which_M)
  selected_ic <- .theory_selected_ic(
    ic = res$IC,
    which_M = which_M,
    has_model = has_model
  )
  incl_cov <- if (!is.null(selected_ic)) selected_ic$incl.cov else NULL

  if (!isTRUE(has_model) || is.null(terms)) {
    return(.empty_theory_solutions())
  }

  .theory_solution_rows(
    theory = theory,
    solution_type = solution_type,
    model = which_M,
    intermediate_branch = NA_character_,
    terms = terms,
    incl_cov = incl_cov
  )
}

.theory_intermediate_solution_rows <- function(theory, res, which_M, i_mode) {
  i.sol <- if (!is.null(res) && !inherits(res, "error")) res$i.sol else NULL
  if (is.null(i.sol) || length(i.sol) == 0L) {
    return(.empty_theory_solutions())
  }

  branch_names <- .normalize_i_solution_names(i.sol)
  if (i_mode == "C1P1") {
    idx <- which(branch_names == "C1P1")
    if (length(idx) == 0L) {
      return(.empty_theory_solutions())
    }
    i.sol <- i.sol[idx[1L]]
    branch_names <- branch_names[idx[1L]]
  }

  rows <- list()
  for (i in seq_along(i.sol)) {
    branch <- i.sol[[i]]
    sol_all <- if (is.list(branch) && "solution" %in% names(branch)) branch$solution else NULL
    primes <- if (is.list(branch) && "primes" %in% names(branch)) branch$primes else NULL
    n_models <- if (!is.null(sol_all)) length(sol_all) else 0L
    has_model <- n_models >= which_M
    terms <- .theory_selected_terms(sol_all, primes, which_M)
    selected_ic <- .theory_selected_ic(
      ic = if (is.list(branch) && "IC" %in% names(branch)) branch$IC else NULL,
      which_M = which_M,
      has_model = has_model
    )
    incl_cov <- if (!is.null(selected_ic)) selected_ic$incl.cov else NULL

    if (!isTRUE(has_model) || is.null(terms)) {
      next
    }

    rows[[length(rows) + 1L]] <- .theory_solution_rows(
      theory = theory,
      solution_type = "intermediate",
      model = which_M,
      intermediate_branch = branch_names[i],
      terms = terms,
      incl_cov = incl_cov
    )
  }

  if (length(rows) == 0L) {
    return(.empty_theory_solutions())
  }

  .bind_rows_result(rows)
}

.theory_solution_type_solution_rows <- function(theory, run, solution_type, which_M, i_mode) {
  if (!identical(run$solution_type_status[[solution_type]], "ok")) {
    return(.empty_theory_solutions())
  }

  if (solution_type %in% c("conservative", "parsimonious")) {
    return(.theory_standard_solution_rows(
      theory = theory,
      res = run$res[[solution_type]],
      solution_type = solution_type,
      which_M = which_M
    ))
  }

  .theory_intermediate_solution_rows(
    theory = theory,
    res = run$res[[solution_type]],
    which_M = which_M,
    i_mode = i_mode
  )
}

.theory_solutions_from_runs <- function(by_theory, monitored_solutions, which_M, i_mode) {
  rows <- list()

  for (theory in names(by_theory)) {
    run <- by_theory[[theory]]

    for (solution_type in monitored_solutions) {
      current <- .theory_solution_type_solution_rows(
        theory = theory,
        run = run,
        solution_type = solution_type,
        which_M = which_M,
        i_mode = i_mode
      )

      if (nrow(current) > 0L) {
        rows[[length(rows) + 1L]] <- current
      }
    }
  }

  if (length(rows) == 0L) {
    return(.empty_theory_solutions())
  }

  .bind_rows_result(rows)
}

.theory_target_branch <- function(target, solution_type) {
  if (!identical(solution_type, "intermediate")) {
    return(NA_character_)
  }

  sig <- as.character(target$signature)
  if (length(sig) != 1L || is.na(sig) || !grepl(":", sig, fixed = TRUE)) {
    return(NA_character_)
  }

  sub(":.*$", "", sig)
}

.theory_target_membership <- function(target) {
  membership <- target$solution_membership
  if (is.null(membership)) {
    return(NULL)
  }

  out <- suppressWarnings(as.numeric(membership))
  if (length(out) == 0L) {
    return(NULL)
  }

  term_memberships <- target$term_memberships
  rn <- if (!is.null(term_memberships) && is.data.frame(term_memberships)) {
    rownames(term_memberships)
  } else {
    NULL
  }

  if (!is.null(rn) && length(rn) == length(out)) {
    names(out) <- rn
  }

  out
}

.theory_targets_from_runs <- function(by_theory, monitored_solutions, which_M, i_mode) {
  rows <- list()

  for (theory in names(by_theory)) {
    run <- by_theory[[theory]]

    for (solution_type in monitored_solutions) {
      if (!identical(run$solution_type_status[[solution_type]], "ok")) {
        next
      }

      targets <- .qca_extract_targets_for_solution_type(
        res = run$res[[solution_type]],
        solution_type = solution_type,
        which_M = which_M,
        i_mode = i_mode
      )

      for (target in targets) {
        membership <- .theory_target_membership(target)
        if (!identical(target$status, "ok") || is.null(membership)) {
          next
        }

        rows[[length(rows) + 1L]] <- list(
          theory = theory,
          solution_type = solution_type,
          intermediate_branch = .theory_target_branch(target, solution_type),
          membership = membership
        )
      }
    }
  }

  rows
}

.theory_same_branch <- function(x, y) {
  if (is.na(x) && is.na(y)) {
    return(TRUE)
  }
  identical(as.character(x), as.character(y))
}

.theory_find_target <- function(targets, theory, solution_type, intermediate_branch) {
  for (target in targets) {
    if (identical(target$theory, theory) &&
        identical(target$solution_type, solution_type) &&
        .theory_same_branch(target$intermediate_branch, intermediate_branch)) {
      return(target)
    }
  }

  NULL
}

.theory_align_memberships <- function(a, b) {
  a <- suppressWarnings(as.numeric(a))
  b <- suppressWarnings(as.numeric(b))
  names_a <- names(a)
  names_b <- names(b)

  if (!is.null(names_a) && !is.null(names_b) &&
      length(names_a) == length(a) && length(names_b) == length(b) &&
      all(nzchar(names_a)) && all(nzchar(names_b))) {
    common <- intersect(names_a, names_b)
    if (length(common) == 0L) {
      return(NULL)
    }

    out <- list(
      a = unname(a[match(common, names_a)]),
      b = unname(b[match(common, names_b)])
    )
  } else {
    if (length(a) != length(b)) {
      return(NULL)
    }

    out <- list(a = unname(a), b = unname(b))
  }

  keep <- !is.na(out$a) & !is.na(out$b)
  if (!any(keep)) {
    return(NULL)
  }

  list(a = out$a[keep], b = out$b[keep])
}

.theory_fuzzy_jaccard <- function(a, b) {
  aligned <- .theory_align_memberships(a, b)
  if (is.null(aligned)) {
    return(NA_real_)
  }

  denom <- sum(pmax(aligned$a, aligned$b), na.rm = TRUE)
  if (denom == 0) {
    return(1)
  }

  sum(pmin(aligned$a, aligned$b), na.rm = TRUE) / denom
}

.theory_mean_abs_membership_delta <- function(a, b) {
  aligned <- .theory_align_memberships(a, b)
  if (is.null(aligned)) {
    return(NA_real_)
  }

  mean(abs(aligned$b - aligned$a))
}

.theory_model_fit_row <- function(models, theory, solution_type, intermediate_branch) {
  keep <- models$theory == theory & models$solution_type == solution_type

  if ("intermediate_branch" %in% names(models)) {
    if (is.na(intermediate_branch)) {
      keep <- keep & is.na(models$intermediate_branch)
    } else {
      keep <- keep & !is.na(models$intermediate_branch) &
        models$intermediate_branch == intermediate_branch
    }
  }

  idx <- which(keep)
  if (length(idx) == 0L) {
    return(NULL)
  }

  models[idx[1L], , drop = FALSE]
}

.theory_fit_delta <- function(models, theory_1, theory_2, solution_type, intermediate_branch, column) {
  row_1 <- .theory_model_fit_row(models, theory_1, solution_type, intermediate_branch)
  row_2 <- .theory_model_fit_row(models, theory_2, solution_type, intermediate_branch)

  if (is.null(row_1) || is.null(row_2) || !column %in% names(row_1) || !column %in% names(row_2)) {
    return(NA_real_)
  }

  suppressWarnings(as.numeric(row_2[[column]]) - as.numeric(row_1[[column]]))
}

.theory_pairwise_branches <- function(targets, solution_type) {
  branches <- vapply(
    targets,
    function(x) {
      if (!identical(x$solution_type, solution_type)) {
        return(NA_character_)
      }
      x$intermediate_branch
    },
    character(1)
  )

  branches <- unique(branches)
  if (!identical(solution_type, "intermediate")) {
    return(NA_character_)
  }

  branches[!is.na(branches)]
}

.theory_pairwise_from_runs <- function(by_theory, models, monitored_solutions, which_M, i_mode) {
  targets <- .theory_targets_from_runs(
    by_theory = by_theory,
    monitored_solutions = monitored_solutions,
    which_M = which_M,
    i_mode = i_mode
  )

  theory_names <- names(by_theory)
  if (length(targets) == 0L || length(theory_names) < 2L) {
    return(.empty_theory_pairwise())
  }

  pairs <- utils::combn(theory_names, 2L, simplify = FALSE)
  rows <- list()

  for (solution_type in monitored_solutions) {
    branches <- .theory_pairwise_branches(targets, solution_type)
    if (length(branches) == 0L) {
      next
    }

    for (branch in branches) {
      for (pair in pairs) {
        theory_1 <- pair[1L]
        theory_2 <- pair[2L]
        target_1 <- .theory_find_target(targets, theory_1, solution_type, branch)
        target_2 <- .theory_find_target(targets, theory_2, solution_type, branch)

        if (is.null(target_1) || is.null(target_2)) {
          next
        }

        rows[[length(rows) + 1L]] <- data.frame(
          solution_type = solution_type,
          intermediate_branch = branch,
          theory_1 = theory_1,
          theory_2 = theory_2,
          delta_inclS = .theory_fit_delta(models, theory_1, theory_2, solution_type, branch, "inclS"),
          delta_PRI = .theory_fit_delta(models, theory_1, theory_2, solution_type, branch, "PRI"),
          delta_covS = .theory_fit_delta(models, theory_1, theory_2, solution_type, branch, "covS"),
          membership_jaccard = .theory_fuzzy_jaccard(target_1$membership, target_2$membership),
          mean_abs_membership_delta = .theory_mean_abs_membership_delta(target_1$membership, target_2$membership),
          stringsAsFactors = FALSE
        )
      }
    }
  }

  if (length(rows) == 0L) {
    return(.empty_theory_pairwise())
  }

  .bind_rows_result(rows)
}

.theory_solution_type_model_info <- function(run, solution_type, which_M, i_mode) {
  if (solution_type %in% c("conservative", "parsimonious")) {
    return(.theory_standard_model_info(
      res = run$res[[solution_type]],
      solution_type = solution_type,
      which_M = which_M
    ))
  }

  .theory_intermediate_model_info(
    res = run$res[[solution_type]],
    which_M = which_M,
    i_mode = i_mode
  )
}

.theory_model_status <- function(run, solution_type, selected_model) {
  status <- run$solution_type_status[[solution_type]]
  if (is.null(status) || is.na(status)) {
    status <- run$status
  }

  if (identical(status, "ok") && is.na(selected_model)) {
    return("selected_solution_missing")
  }

  status
}

.theory_models_from_runs <- function(by_theory, monitored_solutions, which_M, i_mode) {
  rows <- list()

  for (theory in names(by_theory)) {
    run <- by_theory[[theory]]
    stats <- .theory_tt_stats(run$tt)
    n_excluded <- .theory_n_excluded(run$exclude_used)

    for (solution_type in monitored_solutions) {
      model_info <- .theory_solution_type_model_info(
        run = run,
        solution_type = solution_type,
        which_M = which_M,
        i_mode = i_mode
      )

      for (i in seq_len(nrow(model_info))) {
        rows[[length(rows) + 1L]] <- data.frame(
          theory = theory,
          solution_type = model_info$solution_type[i],
          intermediate_branch = model_info$intermediate_branch[i],
          status = .theory_model_status(run, solution_type, model_info$selected_model[i]),
          n_conditions = length(run$conditions),
          n_tt_rows = stats$n_tt_rows,
          n_observed_rows = stats$n_observed_rows,
          n_remainders = stats$n_remainders,
          n_excluded = if (solution_type %in% c("parsimonious", "intermediate")) n_excluded else 0L,
          n_models = model_info$n_models[i],
          selected_model = model_info$selected_model[i],
          n_terms = model_info$n_terms[i],
          inclS = model_info$inclS[i],
          PRI = model_info$PRI[i],
          covS = model_info$covS[i],
          stringsAsFactors = FALSE
        )
      }
    }
  }

  if (length(rows) == 0L) {
    return(.empty_theory_models())
  }

  .bind_rows_result(rows)
}

.theory_deparse_one <- function(x) {
  if (is.null(x)) {
    return("NULL")
  }
  paste(utils::capture.output(dput(x)), collapse = "")
}

.theory_n_rows <- function(x) {
  if (!is.data.frame(x)) {
    return(0L)
  }
  as.integer(nrow(x))
}

.theory_print_solutions <- function(solutions, row.names = FALSE, ...) {
  display <- solutions
  if ("model" %in% names(display) && all(display$model == 1L, na.rm = TRUE)) {
    display$model <- NULL
  }
  if ("intermediate_branch" %in% names(display) && all(is.na(display$intermediate_branch))) {
    display$intermediate_branch <- NULL
  }

  fit_cols <- intersect(c("inclS", "PRI", "covS"), names(display))
  for (col in fit_cols) {
    display[[col]] <- round(display[[col]], 3L)
  }

  print.data.frame(display, row.names = row.names, ...)
}

#' Theory-specification robustness for QCA models
#'
#' Compare several theoretically oriented QCA condition sets while holding the
#' outcome, truth-table cutoffs, solution type, exclusion handling, and
#' model-selection settings constant. Each theory supplies its own condition
#' set.
#'
#' The function builds a separate truth table for each theory, recomputes
#' exclusions separately when requested, and runs the monitored solution types
#' under the same analytic settings. The raw per-theory QCA objects are stored
#' in `by_theory`. `results$models` gives the clean model-level comparison
#' table, and `results$solutions` gives the selected prime implicants or
#' solution terms by theory and solution type. `results$pairwise` compares selected
#' solution memberships across theories using fuzzy Jaccard similarity and
#' mean absolute membership differences.
#'
#' @param data A non-empty data frame containing the calibrated outcome and
#'   calibrated condition columns.
#' @param outcome Name of the calibrated outcome column in `data`.
#' @param theories Named list of theory-specific condition sets. Each entry
#'   must be a non-empty character vector of condition names in `data`.
#'   Conditions may be shared across theories, but no theory may contain the
#'   outcome as a condition.
#' @param incl.cut Inclusion cutoff to be used for every theory-specific truth
#'   table.
#' @param n.cut Frequency cutoff to be used for every theory-specific truth
#'   table.
#' @param solution Solution type to compare. Accepted values are `"all"`,
#'   `"con"` or `"conservative"`, `"par"` or `"parsimonious"`, and `"int"` or
#'   `"intermediate"`.
#' @param include Optional minimization include setting. Currently, this
#'   argument accepts only `NULL`, `""`, or `"?"`.
#' @param dir.exp Theory-specific directional expectations. Use `NULL` when no
#'   intermediate solutions are monitored. Otherwise supply a named list with
#'   one entry per theory. Each entry may be an ordered character vector aligned
#'   to that theory's conditions or a named character vector whose names match
#'   that theory's conditions.
#' @param which_M Positive integer giving which solution alternative to use
#'   when minimization returns multiple models.
#' @param i_mode Character string controlling intermediate-solution selection.
#'   Accepted values are `"C1P1"` and `"all"`. The default is `"C1P1"` because
#'   comparative theory testing needs a single comparable intermediate branch
#'   per theory by default.
#' @param exclude_mode Character string controlling how excluded rows are
#'   handled for parsimonious and intermediate minimization. `"recompute"`
#'   recalculates exclusions separately for each theory-specific truth table,
#'   `"static"` reuses theory-specific static exclusions supplied through
#'   `exclude_static`, and `"none"` does not use exclusions.
#' @param exclude_recompute Named list of arguments passed to [QCA::findRows()]
#'   when `exclude_mode = "recompute"`.
#' @param exclude_static Static exclusions used when `exclude_mode = "static"`.
#'   For
#'   `theory.test()`, this must be a named list with one entry per theory when
#'   parsimonious or intermediate solutions are monitored.
#' @param progress Logical; if `TRUE`, show a text progress bar during
#'   per-theory QCA runs in interactive sessions.
#' @param x A `theory_test` object returned by [theory.test()].
#' @param row.names Logical; if `TRUE`, print row names in the selected
#'   solutions table printed by [print.theory_test()].
#' @param ... Additional arguments split between [QCA::truthTable()] and
#'   [QCA::minimize()]. The function also looks in `...` for `include`,
#'   `dir.exp`, or `direxp` if those arguments were not supplied explicitly. In
#'   [print.theory_test()], `...` is passed to [print.data.frame()]. In
#'   [as.data.frame.theory_test()], `...` is ignored.
#'
#' @returns An object of class `theory_test` with the following components:
#'   \describe{
#'     \item{`diagnostics`}{A detailed internal diagnostics table with one row
#'     per theory and monitored solution type.}
#'     \item{`results`}{A named list with `models`, `pairwise`, and
#'     `solutions` tables. `models` is populated with model-level diagnostics
#'     and fit values; `solutions` is populated with selected solution terms
#'     and term-level fit values; `pairwise` is populated with pairwise theory
#'     comparisons by solution type.}
#'     \item{`by_theory`}{A named list containing the theory-specific condition
#'     sets, directional expectations, truth tables, exclusions, minimization
#'     objects, selected solution terms used for comparison, and current run
#'     status.}
#'     \item{`settings`}{A list containing the analysis settings used in the
#'     call.}
#'   }
#'
#'   `print.theory_test()` prints a compact theory-specification summary and
#'   the selected solutions table. `as.data.frame.theory_test()` returns
#'   `results$models`.
#'
#' @examples
#' \donttest{
#' library(QCA)
#' data(LR)
#'
#' outcome <- "SURV"
#'
#' thresholds <- list(
#'   DEV = findTh(LR$DEV, groups = 7),
#'   URB = findTh(LR$URB, groups = 4),
#'   LIT = findTh(LR$LIT, groups = 4),
#'   IND = findTh(LR$IND, groups = 4),
#'   STB = findTh(LR$STB, groups = 4),
#'   SURV = findTh(LR$SURV, groups = 4)
#' )
#'
#' dat <- LR
#' dat$DEV <- calibrate(LR$DEV, type = "fuzzy", thresholds = thresholds$DEV)
#' dat$URB <- calibrate(LR$URB, type = "fuzzy", thresholds = thresholds$URB)
#' dat$LIT <- calibrate(LR$LIT, type = "fuzzy", thresholds = thresholds$LIT)
#' dat$IND <- calibrate(LR$IND, type = "fuzzy", thresholds = thresholds$IND)
#' dat$STB <- calibrate(LR$STB, type = "fuzzy", thresholds = thresholds$STB)
#' dat$SURV <- calibrate(LR$SURV, type = "fuzzy", thresholds = thresholds$SURV)
#'
#' theories <- list(
#'   development = c("DEV", "URB", "LIT"),
#'   industrial = c("DEV", "URB", "IND"),
#'   broad = c("DEV", "URB", "LIT", "IND", "STB")
#' )
#'
#' dir_exp <- list(
#'   development = c("1", "1", "1"),
#'   industrial = c("1", "1", "1"),
#'   broad = c("1", "1", "1", "1", "1")
#' )
#'
#' theory_out <- theory.test(
#'   data = dat,
#'   outcome = outcome,
#'   theories = theories,
#'   incl.cut = 0.8,
#'   n.cut = 1,
#'   solution = "all",
#'   dir.exp = dir_exp,
#'   progress = TRUE
#' )
#'
#' theory_out
#' as.data.frame(theory_out)
#' theory_out$results$solutions
#' theory_out$results$pairwise
#' }
#'
#' @seealso [calib.test()], [incl.test()], [ncut.test()], [loo.test()],
#'   [subsample.test()], [altset.test()], [cluster.test()], [sol.df()]
#' @export
theory.test <- function(
    data,
    outcome,
    theories,
    incl.cut = 1,
    n.cut = 1,
    solution = "all",
    include = NULL,
    dir.exp = NULL,
    which_M = 1,
    i_mode = "C1P1",
    exclude_mode = c("recompute", "static", "none"),
    exclude_recompute = list(type = 2),
    exclude_static = NULL,
    progress = TRUE,
    ...
) {
  .require_qca()

  selection_controls <- .normalize_i_mode(i_mode)
  i_mode <- selection_controls$i_mode
  exclude_mode <- match.arg(exclude_mode)
  mc <- match.call(expand.dots = FALSE)
  dots_raw <- list(...)
  .reject_exclusion_controls_in_dots(dots_raw, "theory.test")

  if (is.null(dim(data)) || is.null(nrow(data)) || nrow(data) < 1L) {
    stop("`data` must be a non-empty data frame object with at least one row.")
  }

  if (!is.character(outcome) || length(outcome) != 1L || is.na(outcome) || !nzchar(outcome)) {
    stop("`outcome` must be a single non-empty character string.")
  }
  outcome <- trimws(outcome)
  if (!outcome %in% colnames(data)) {
    stop("`outcome` must name a column in `data`.")
  }

  theories <- .normalize_theories(theories, data = data, outcome = outcome)

  if (!is.numeric(incl.cut) || length(incl.cut) != 1L || !is.finite(incl.cut) ||
      incl.cut < 0 || incl.cut > 1) {
    stop("`incl.cut` must be a single finite number in [0, 1].")
  }

  n.cut <- .as_integerish_scalar(n.cut, "n.cut", min = 1L)
  which_M <- .coerce_which_M(which_M)

  if (!is.logical(progress) || length(progress) != 1L || is.na(progress)) {
    stop("`progress` must be `TRUE` or `FALSE`.")
  }

  if (is.null(include)) {
    inc <- .dot_get(dots_raw, "include")
    if (!is.null(inc)) include <- inc
  }

  if (is.null(dir.exp)) {
    de <- .dot_get(dots_raw, "dir.exp")
    if (is.null(de)) de <- .dot_get(dots_raw, "direxp")
    if (!is.null(de)) dir.exp <- de
  }

  solution_norm <- .normalize_solution_std(solution)
  dir.exp <- .normalize_theory_dir_exp(dir.exp, theories, solution = solution_norm)

  solution_controls <- .resolve_solution_controls(
    solution = solution_norm,
    include = include,
    dir.exp = dir.exp,
    caller = "theory.test",
    style = "std"
  )

  solution <- solution_controls$solution
  include <- solution_controls$include
  monitored_solutions <- solution_controls$monitored

  .validate_exclusion_controls(
    exclude_mode = exclude_mode,
    exclude_recompute = exclude_recompute,
    exclude_static = exclude_static,
    exclude_recompute_supplied = !is.null(mc$exclude_recompute),
    exclude_static_supplied = !is.null(mc$exclude_static),
    monitored_solutions = monitored_solutions,
    style = "std"
  )
  .validate_theory_static_exclude(
    exclude_mode = exclude_mode,
    exclude_static = exclude_static,
    theories = theories,
    monitored_solutions = monitored_solutions
  )

  dots_split <- .split_truth_table_minimize_dots(dots_raw)
  progress_state <- .new_qcaert_progress(total = length(theories), progress = progress)
  on.exit(progress_state$close(), add = TRUE)

  by_theory <- lapply(names(theories), function(theory) {
    run <- .theory_run_one(
      theory = theory,
      data = data,
      outcome = outcome,
      conditions = theories[[theory]],
      dir.exp = if (is.null(dir.exp)) NULL else dir.exp[[theory]],
      incl.cut = incl.cut,
      n.cut = n.cut,
      monitored_solutions = monitored_solutions,
      exclude_mode = exclude_mode,
      exclude_recompute = exclude_recompute,
      exclude_static = exclude_static,
      dots_tt = dots_split$tt,
      dots_min = dots_split$min,
      which_M = which_M,
      i_mode = i_mode
    )
    progress_state$tick()
    run
  })
  names(by_theory) <- names(theories)

  diagnostics <- .theory_diagnostics_from_runs(
    by_theory = by_theory,
    monitored_solutions = monitored_solutions
  )
  models <- .theory_models_from_runs(
    by_theory = by_theory,
    monitored_solutions = monitored_solutions,
    which_M = which_M,
    i_mode = i_mode
  )
  solutions <- .theory_solutions_from_runs(
    by_theory = by_theory,
    monitored_solutions = monitored_solutions,
    which_M = which_M,
    i_mode = i_mode
  )
  pairwise <- .theory_pairwise_from_runs(
    by_theory = by_theory,
    models = models,
    monitored_solutions = monitored_solutions,
    which_M = which_M,
    i_mode = i_mode
  )
  results <- .empty_theory_results()
  results$models <- models
  results$pairwise <- pairwise
  results$solutions <- solutions

  .new_result_object(
    "theory_test",
    diagnostics = diagnostics,
    results = results,
    by_theory = by_theory,
    settings = list(
      outcome = outcome,
      theories = theories,
      incl.cut = incl.cut,
      n.cut = n.cut,
      solution = solution,
      monitored_solutions = monitored_solutions,
      include = include,
      dir.exp = dir.exp,
      which_M = which_M,
      i_mode = i_mode,
      exclude_mode = exclude_mode,
      exclude_recompute = exclude_recompute,
      exclude_static = exclude_static,
      progress = progress
    )
  )
}

#' Print a `theory_test` object
#'
#' @rdname theory.test
#' @export
print.theory_test <- function(x, row.names = FALSE, ...) {
  settings <- x$settings

  cat("<theory_test>\n")
  cat("qcaERT theory-specification test\n\n")
  cat("Outcome: ", settings$outcome, "\n", sep = "")

  theories <- settings$theories
  cat("Theories tested: ", length(theories), "\n", sep = "")
  for (theory in names(theories)) {
    cat("  ", theory, ": ", paste(theories[[theory]], collapse = ", "), "\n", sep = "")
  }

  cat("\nSolution: ", settings$solution, "\n", sep = "")
  cat("Monitored solutions: ", paste(settings$monitored_solutions, collapse = ", "), "\n", sep = "")
  cat("which_M: ", settings$which_M, "\n", sep = "")
  if (any(settings$monitored_solutions %in% "intermediate")) {
    cat("i_mode: ", settings$i_mode, "\n", sep = "")
  }

  cat("\nTruth-table settings:\n")
  cat("  incl.cut: ", format(settings$incl.cut, trim = TRUE), "\n", sep = "")
  cat("  n.cut: ", settings$n.cut, "\n", sep = "")

  cat("\nRun status:\n")
  for (theory in names(x$by_theory)) {
    cat("  ", theory, ": ", x$by_theory[[theory]]$status, "\n", sep = "")
  }

  if (.theory_n_rows(x$results$solutions) > 0L) {
    cat("\nSolutions\n")
    .theory_print_solutions(x$results$solutions, row.names = row.names, ...)
  }

  cat("\nTables:\n")
  cat("  as.data.frame(x) or x$results$models: model-level diagnostics (", .theory_n_rows(x$results$models), " rows)\n", sep = "")
  cat("  x$results$solutions: extracted solution terms (", .theory_n_rows(x$results$solutions), " rows)\n", sep = "")
  cat("  x$results$pairwise: pairwise theory comparisons (", .theory_n_rows(x$results$pairwise), " rows)\n", sep = "")
  cat("  x$diagnostics: raw per-theory solution_type diagnostics (", .theory_n_rows(x$diagnostics), " rows)\n", sep = "")
  cat("  x$by_theory: truth tables, exclusions, and minimization objects\n", sep = "")

  invisible(x)
}

#' Return the model-level table from a `theory_test` object
#'
#' @rdname theory.test
#' @export
as.data.frame.theory_test <- function(x, ...) {
  x$results$models
}

theory_test_fixture <- function() {
  data.frame(
    Y = c(0, 0.33, 0.67, 1),
    A = c(1, 0.8, 0.2, 0),
    B = c(0.9, 0.7, 0.4, 0.1),
    C = c(0.2, 0.6, 0.8, 1),
    D = c(0.1, 0.4, 0.7, 0.9)
  )
}

test_that("theory.test empty table helpers define the public schemas", {
  diagnostics <- .empty_theory_diagnostics()
  models <- .empty_theory_models()
  pairwise <- .empty_theory_pairwise()
  solutions <- .empty_theory_solutions()
  results <- .empty_theory_results()

  expect_identical(
    names(diagnostics),
    c(
      "theory",
      "solution_type",
      "status",
      "n_conditions",
      "conditions",
      "n_tt_rows",
      "n_observed_rows",
      "n_remainders",
      "n_excluded",
      "selected_solution_missing",
      "error_source",
      "error_message"
    )
  )
  expect_identical(
    names(models),
    c(
      "theory",
      "solution_type",
      "intermediate_branch",
      "status",
      "n_conditions",
      "n_tt_rows",
      "n_observed_rows",
      "n_remainders",
      "n_excluded",
      "n_models",
      "selected_model",
      "n_terms",
      "inclS",
      "PRI",
      "covS"
    )
  )
  expect_identical(
    names(pairwise),
    c(
      "solution_type",
      "intermediate_branch",
      "theory_1",
      "theory_2",
      "delta_inclS",
      "delta_PRI",
      "delta_covS",
      "membership_jaccard",
      "mean_abs_membership_delta"
    )
  )
  expect_identical(
    names(solutions),
    c("theory", "solution_type", "model", "intermediate_branch", "prime_implicant", "inclS", "PRI", "covS")
  )
  expect_identical(names(results), c("models", "pairwise", "solutions"))
  expect_identical(results$models, models)
  expect_identical(results$pairwise, pairwise)
  expect_identical(results$solutions, solutions)

  expect_identical(nrow(diagnostics), 0L)
  expect_identical(nrow(models), 0L)
  expect_identical(nrow(pairwise), 0L)
  expect_identical(nrow(solutions), 0L)
  expect_type(diagnostics$selected_solution_missing, "logical")
  expect_type(models$inclS, "double")
  expect_type(pairwise$membership_jaccard, "double")
  expect_type(solutions$model, "integer")
})

test_that("theory.test lightweight normalizers preserve theory-specific settings", {
  dat <- theory_test_fixture()
  theories <- .normalize_theories(
    theories = list(
      institutional = c("A", "B"),
      structural = c("B", "C", "D")
    ),
    data = dat,
    outcome = "Y"
  )

  expect_identical(
    theories,
    list(
      institutional = c("A", "B"),
      structural = c("B", "C", "D")
    )
  )

  dir_exp <- .normalize_theory_dir_exp(
    dir.exp = list(
      institutional = c(B = "0", A = "1"),
      structural = c(D = "1", B = "0", C = "1")
    ),
    theories = theories,
    solution = "all"
  )

  expect_identical(dir_exp$institutional, c("1", "0"))
  expect_identical(dir_exp$structural, c("0", "1", "1"))
  expect_null(.normalize_theory_dir_exp(NULL, theories, solution = "conservative"))

  expect_error(
    .normalize_theories(
      theories = list(one = c("A", "Y"), two = c("B", "C")),
      data = dat,
      outcome = "Y"
    ),
    "Theory `one` includes the outcome as a condition",
    fixed = TRUE
  )
  expect_error(
    .normalize_theory_dir_exp(
      dir.exp = list(institutional = c("1", "0")),
      theories = theories,
      solution = "all"
    ),
    "Missing: structural.",
    fixed = TRUE
  )
  expect_error(
    .normalize_theory_dir_exp(NULL, theories, solution = "intermediate"),
    "When `solution = \"intermediate\"`/`\"int\"`, `dir.exp` must be provided",
    fixed = TRUE
  )
})

test_that("theory.test LR integration fills models, solutions, and pairwise tables", {
  skip_if_not_installed("QCA")

  fixture <- qcaert_fixture_lr()
  theories <- list(
    development = c("DEV", "URB", "LIT"),
    industrial = c("DEV", "URB", "IND"),
    broad = c("DEV", "URB", "LIT", "IND", "STB")
  )
  dir_exp <- list(
    development = c("1", "1", "1"),
    industrial = c("1", "1", "1"),
    broad = c("1", "1", "1", "1", "1")
  )

  out <- theory.test(
    data = fixture$calib,
    outcome = fixture$outcome,
    theories = theories,
    incl.cut = 0.8,
    n.cut = 1,
    solution = "all",
    dir.exp = dir_exp,
    progress = FALSE
  )

  expect_s3_class(out, "theory_test")
  expect_identical(out$settings$theories, theories)
  expect_identical(out$settings$monitored_solutions, c("conservative", "parsimonious", "intermediate"))
  expect_identical(as.data.frame(out), out$results$models)
  expect_true(all(vapply(out$by_theory, `[[`, character(1), "status") == "ok"))

  expect_identical(nrow(out$diagnostics), 9L)
  expect_identical(nrow(out$results$models), 9L)
  expect_identical(nrow(out$results$solutions), 9L)
  expect_identical(nrow(out$results$pairwise), 9L)
  expect_true(all(out$diagnostics$status == "ok"))
  expect_true(all(out$results$models$status == "ok"))

  expect_identical(
    out$results$models[, c("theory", "solution_type")],
    data.frame(
      theory = rep(names(theories), each = 3L),
      solution_type = rep(c("conservative", "parsimonious", "intermediate"), times = 3L),
      stringsAsFactors = FALSE
    )
  )
  expect_identical(out$results$models$n_conditions, rep(c(3L, 3L, 5L), each = 3L))
  expect_identical(out$results$models$n_tt_rows, rep(c(8L, 8L, 32L), each = 3L))
  expect_identical(out$results$models$n_observed_rows, rep(c(5L, 6L, 11L), each = 3L))
  expect_true(all(out$results$models$n_terms >= 1L))
  expect_true(all(out$results$models$inclS >= 0.8))
  expect_true(all(out$results$pairwise$membership_jaccard >= 0))
  expect_true(all(out$results$pairwise$membership_jaccard <= 1))
  expect_true(all(out$results$pairwise$mean_abs_membership_delta >= 0))
  expect_setequal(
    out$results$solutions$prime_implicant,
    c("~DEV*URB*LIT", "~DEV*URB*IND", "~DEV*URB*LIT*IND*~STB", "~DEV*URB*LIT*~STB")
  )
})

test_that("theory.test returns family object structure with per-theory runs", {
  skip_if_not_installed("QCA")

  dat <- theory_test_fixture()
  theories <- list(
    institutional = c("A", "B"),
    structural = c("B", "C", "D")
  )
  dir_exp <- list(
    institutional = c(B = "0", A = "1"),
    structural = c(D = "1", B = "0", C = "1")
  )

  out <- theory.test(
    data = dat,
    outcome = "Y",
    theories = theories,
    incl.cut = 0.8,
    n.cut = 1,
    solution = "all",
    dir.exp = dir_exp,
    i_mode = "C1P1",
    progress = FALSE
  )

  expect_s3_class(out, "theory_test")
  expect_identical(names(out), c("diagnostics", "results", "by_theory", "settings"))
  expect_identical(names(out$results), c("models", "pairwise", "solutions"))
  expect_identical(as.data.frame(out), out$results$models)
  expect_identical(out$settings$monitored_solutions, c("conservative", "parsimonious", "intermediate"))
  expect_identical(out$settings$theories, theories)
  expect_identical(out$settings$dir.exp$institutional, c("1", "0"))
  expect_identical(out$settings$dir.exp$structural, c("0", "1", "1"))
  expect_identical(out$by_theory$institutional$status, "ok")
  expect_s3_class(out$by_theory$institutional$tt, "QCA_tt")
  expect_s3_class(out$by_theory$institutional$res$conservative, "QCA_min")
  expect_s3_class(out$by_theory$institutional$res$parsimonious, "QCA_min")
  expect_s3_class(out$by_theory$institutional$res$intermediate, "QCA_min")
  expect_identical(
    out$diagnostics[, c("theory", "solution_type", "status")],
    data.frame(
      theory = rep(names(theories), each = 3L),
      solution_type = rep(c("conservative", "parsimonious", "intermediate"), times = 2L),
      status = rep("ok", 6L),
      stringsAsFactors = FALSE
    )
  )
  expect_identical(out$diagnostics$n_tt_rows, c(4L, 4L, 4L, 8L, 8L, 8L))
  expect_identical(out$diagnostics$n_observed_rows, c(2L, 2L, 2L, 3L, 3L, 3L))
  expect_identical(out$diagnostics$n_remainders, c(2L, 2L, 2L, 5L, 5L, 5L))

  models <- out$results$models
  expect_identical(
    names(models),
    c(
      "theory",
      "solution_type",
      "intermediate_branch",
      "status",
      "n_conditions",
      "n_tt_rows",
      "n_observed_rows",
      "n_remainders",
      "n_excluded",
      "n_models",
      "selected_model",
      "n_terms",
      "inclS",
      "PRI",
      "covS"
    )
  )
  expect_identical(as.data.frame(out), models)
  expect_identical(models[, c("theory", "solution_type", "status")], out$diagnostics[, c("theory", "solution_type", "status")])
  expect_identical(models$n_conditions, out$diagnostics$n_conditions)
  expect_identical(models$n_tt_rows, out$diagnostics$n_tt_rows)
  expect_identical(models$n_observed_rows, out$diagnostics$n_observed_rows)
  expect_identical(models$n_remainders, out$diagnostics$n_remainders)
  expect_identical(models$n_excluded, out$diagnostics$n_excluded)
  expect_identical(models$n_models, rep(1L, 6L))
  expect_identical(models$selected_model, rep(1L, 6L))
  expect_identical(models$n_terms, rep(1L, 6L))
  expect_identical(models$intermediate_branch, c(NA, NA, "C1P1", NA, NA, "C1P1"))
  expect_equal(models$inclS, c(1, 1, 1, 0.9473684, 0.9473684, 0.9473684), tolerance = 1e-6)
  expect_equal(models$PRI, c(1, 1, 1, 0.9212598, 0.9212598, 0.9212598), tolerance = 1e-6)
  expect_equal(models$covS, c(0.85, 0.85, 0.85, 0.9, 0.9, 0.9), tolerance = 1e-12)

  solutions <- out$results$solutions
  expect_identical(
    names(solutions),
    c("theory", "solution_type", "model", "intermediate_branch", "prime_implicant", "inclS", "PRI", "covS")
  )
  expect_identical(nrow(solutions), sum(models$n_terms))
  expect_identical(
    solutions[, c("theory", "solution_type", "model", "intermediate_branch", "prime_implicant")],
    data.frame(
      theory = rep(names(theories), each = 3L),
      solution_type = rep(c("conservative", "parsimonious", "intermediate"), times = 2L),
      model = rep(1L, 6L),
      intermediate_branch = c(NA, NA, "C1P1", NA, NA, "C1P1"),
      prime_implicant = c("~A*~B", "~A*~B", "~A*~B", "~B*C*D", "~B*D", "~B*C*D"),
      stringsAsFactors = FALSE
    )
  )
  expect_equal(solutions$inclS, models$inclS, tolerance = 1e-6)
  expect_equal(solutions$PRI, models$PRI, tolerance = 1e-6)
  expect_equal(solutions$covS, models$covS, tolerance = 1e-12)

  pairwise <- out$results$pairwise
  expect_identical(
    names(pairwise),
    c(
      "solution_type",
      "intermediate_branch",
      "theory_1",
      "theory_2",
      "delta_inclS",
      "delta_PRI",
      "delta_covS",
      "membership_jaccard",
      "mean_abs_membership_delta"
    )
  )
  expect_identical(pairwise$solution_type, c("conservative", "parsimonious", "intermediate"))
  expect_identical(pairwise$intermediate_branch, c(NA, NA, "C1P1"))
  expect_identical(pairwise$theory_1, rep("institutional", 3L))
  expect_identical(pairwise$theory_2, rep("structural", 3L))
  expect_equal(pairwise$delta_inclS, rep(-0.05263158, 3L), tolerance = 1e-6)
  expect_equal(pairwise$delta_PRI, rep(-0.07874016, 3L), tolerance = 1e-6)
  expect_equal(pairwise$delta_covS, rep(0.05, 3L), tolerance = 1e-12)
  expect_equal(pairwise$membership_jaccard, rep(0.8947368, 3L), tolerance = 1e-6)
  expect_equal(pairwise$mean_abs_membership_delta, rep(0.05, 3L), tolerance = 1e-12)
})

test_that("theory.test keeps conservative-only settings without dir.exp", {
  skip_if_not_installed("QCA")

  dat <- theory_test_fixture()

  out <- theory.test(
    data = dat,
    outcome = "Y",
    theories = list(
      one = c("A", "B"),
      two = c("B", "C")
    ),
    incl.cut = 0.8,
    n.cut = 1,
    solution = "conservative",
    progress = FALSE
  )

  expect_identical(out$settings$solution, "conservative")
  expect_identical(out$settings$include, "")
  expect_identical(out$settings$monitored_solutions, "conservative")
  expect_null(out$settings$dir.exp)
  expect_identical(out$by_theory$one$status, "ok")
  expect_identical(out$by_theory$two$status, "ok")
  expect_identical(out$diagnostics$solution_type, c("conservative", "conservative"))
  expect_true(all(out$diagnostics$status == "ok"))
  expect_identical(out$results$models$solution_type, c("conservative", "conservative"))
  expect_true(all(out$results$models$status == "ok"))
  expect_true(all(out$results$models$n_models >= 1L))
  expect_identical(out$results$solutions$solution_type, c("conservative", "conservative"))
  expect_true(all(!is.na(out$results$solutions$prime_implicant)))
  expect_identical(out$results$pairwise$solution_type, "conservative")
  expect_true(is.na(out$results$pairwise$intermediate_branch))
  expect_equal(out$results$pairwise$membership_jaccard, 0.8947368, tolerance = 1e-6)
})

test_that("theory.test pairwise comparisons expand all theory pairs by solution type", {
  skip_if_not_installed("QCA")

  dat <- theory_test_fixture()
  out <- theory.test(
    data = dat,
    outcome = "Y",
    theories = list(
      institutional = c("A", "B"),
      structural = c("B", "C", "D"),
      behavioral = c("A", "C")
    ),
    incl.cut = 0.8,
    n.cut = 1,
    solution = "all",
    dir.exp = list(
      institutional = c(B = "0", A = "1"),
      structural = c(D = "1", B = "0", C = "1"),
      behavioral = c(A = "1", C = "1")
    ),
    progress = FALSE
  )

  pairwise <- out$results$pairwise

  expect_identical(nrow(pairwise), 9L)
  solution_type_counts <- table(pairwise$solution_type)
  expect_identical(
    as.integer(solution_type_counts[c("conservative", "parsimonious", "intermediate")]),
    c(3L, 3L, 3L)
  )
  expect_setequal(
    paste(pairwise$theory_1, pairwise$theory_2, sep = " / "),
    c(
      "institutional / structural",
      "institutional / behavioral",
      "structural / behavioral"
    )
  )
  expect_true(all(pairwise$membership_jaccard >= 0 & pairwise$membership_jaccard <= 1))
  expect_true(all(pairwise$mean_abs_membership_delta >= 0))
})

test_that("theory.test validates theory condition sets strictly", {
  dat <- theory_test_fixture()

  expect_error(
    theory.test(
      data = dat,
      outcome = "Y",
      theories = list(one = c("A", "Y"), two = c("B", "C")),
      progress = FALSE
    ),
    "Theory `one` includes the outcome as a condition",
    fixed = TRUE
  )

  expect_error(
    theory.test(
      data = dat,
      outcome = "Y",
      theories = list(one = c("A", "missing"), two = c("B", "C")),
      progress = FALSE
    ),
    "Theory `one` contains condition(s) not found in `data`: missing",
    fixed = TRUE
  )

  expect_error(
    theory.test(
      data = dat,
      outcome = "Y",
      theories = list(one = c("A", "A"), two = c("B", "C")),
      progress = FALSE
    ),
    "Theory `one` contains duplicate condition names.",
    fixed = TRUE
  )
})

test_that("theory.test validates theory-specific dir.exp", {
  dat <- theory_test_fixture()
  theories <- list(one = c("A", "B"), two = c("B", "C"))

  expect_error(
    theory.test(
      data = dat,
      outcome = "Y",
      theories = theories,
      solution = "intermediate",
      progress = FALSE
    ),
    "When `solution = \"intermediate\"`/`\"int\"`, `dir.exp` must be provided as a named list",
    fixed = TRUE
  )

  expect_error(
    theory.test(
      data = dat,
      outcome = "Y",
      theories = theories,
      dir.exp = list(one = c("1", "0")),
      progress = FALSE
    ),
    "Missing: two.",
    fixed = TRUE
  )

  expect_error(
    theory.test(
      data = dat,
      outcome = "Y",
      theories = theories,
      dir.exp = list(one = c(A = "1", C = "0"), two = c(B = "1", C = "0")),
      progress = FALSE
    ),
    "`dir.exp[[\"one\"]]` names must match that theory's conditions.",
    fixed = TRUE
  )
})

test_that("theory.test validates static theory-specific exclusions", {
  dat <- theory_test_fixture()
  theories <- list(one = c("A", "B"), two = c("B", "C"))
  dir_exp <- list(one = c("1", "0"), two = c("1", "1"))

  expect_error(
    theory.test(
      data = dat,
      outcome = "Y",
      theories = theories,
      dir.exp = dir_exp,
      exclude_mode = "static",
      exclude_static = list(one = 1:2),
      progress = FALSE
    ),
    "Missing: two.",
    fixed = TRUE
  )
})

test_that("theory.test diagnostics preserve solution-type-specific exclude errors", {
  skip_if_not_installed("QCA")

  dat <- theory_test_fixture()
  dir_exp <- list(
    one = c(A = "1", B = "0"),
    two = c(B = "0", C = "1")
  )

  out <- theory.test(
    data = dat,
    outcome = "Y",
    theories = list(one = c("A", "B"), two = c("B", "C")),
    incl.cut = 0.8,
    n.cut = 1,
    solution = "all",
    dir.exp = dir_exp,
    exclude_recompute = list(type = 2, unknown_argument = TRUE),
    progress = FALSE
  )

  expect_identical(out$by_theory$one$status, "partial")
  expect_identical(out$by_theory$two$status, "partial")
  expect_identical(
    out$diagnostics$status,
    rep(c("ok", "exclude_error", "exclude_error"), times = 2L)
  )
  expect_identical(
    out$results$models$status,
    rep(c("ok", "exclude_error", "exclude_error"), times = 2L)
  )
  expect_true(all(is.na(out$results$models$inclS[out$results$models$status == "exclude_error"])))
  expect_true(all(is.na(out$results$models$selected_model[out$results$models$status == "exclude_error"])))
  expect_identical(out$results$solutions$solution_type, c("conservative", "conservative"))
  expect_false(any(out$results$solutions$solution_type %in% c("parsimonious", "intermediate")))
  expect_identical(out$results$pairwise$solution_type, "conservative")
  expect_false(any(out$results$pairwise$solution_type %in% c("parsimonious", "intermediate")))
  expect_true(all(out$diagnostics$error_source[out$diagnostics$status == "exclude_error"] == "exclude"))
  expect_true(all(is.na(out$diagnostics$error_source[out$diagnostics$status == "ok"])))
})

test_that("theory.test print summarizes run structure", {
  skip_if_not_installed("QCA")

  dat <- theory_test_fixture()
  out <- theory.test(
    data = dat,
    outcome = "Y",
    theories = list(one = c("A", "B"), two = c("B", "C")),
    incl.cut = 0.8,
    n.cut = 1,
    solution = "all",
    progress = FALSE
  )

  printed <- capture.output(print(out))

  expect_true(any(grepl("<theory_test>", printed, fixed = TRUE)))
  expect_true(any(grepl("qcaERT theory-specification test", printed, fixed = TRUE)))
  expect_true(any(grepl("Theories tested: 2", printed, fixed = TRUE)))
  expect_true(any(grepl("Solution: all", printed, fixed = TRUE)))
  expect_true(any(grepl("Monitored solutions: conservative, parsimonious", printed, fixed = TRUE)))
  expect_false(any(grepl("Exclusion handling:", printed, fixed = TRUE)))
  expect_true(any(grepl("one: ok", printed, fixed = TRUE)))
  expect_true(any(grepl("Solutions", printed, fixed = TRUE)))
  expect_true(any(grepl("prime_implicant", printed, fixed = TRUE)))
  expect_true(any(grepl("Tables:", printed, fixed = TRUE)))
  expect_true(any(grepl("x$results$models: model-level diagnostics (4 rows)", printed, fixed = TRUE)))
  expect_true(any(grepl("x$results$solutions: extracted solution terms (4 rows)", printed, fixed = TRUE)))
  expect_true(any(grepl("x$results$pairwise: pairwise theory comparisons (2 rows)", printed, fixed = TRUE)))
  expect_true(any(grepl("x$diagnostics: raw per-theory solution_type diagnostics (4 rows)", printed, fixed = TRUE)))
  expect_true(any(grepl("x$results$pairwise", printed, fixed = TRUE)))
})

.as_integerish_scalar <- function(x, name, min = NULL) {
  if (!is.numeric(x) || length(x) != 1L || !is.finite(x)) {
    stop(sprintf("`%s` must be a single finite integer-like number.", name))
  }
  xi <- suppressWarnings(as.integer(x))
  if (is.na(xi) || abs(x - xi) > sqrt(.Machine$double.eps)) {
    stop(sprintf("`%s` must be integer-like.", name))
  }
  if (!is.null(min) && xi < min) {
    stop(sprintf("`%s` must be >= %s.", name, min))
  }
  xi
}

.coerce_which_M <- function(which_M) {
  .as_integerish_scalar(which_M, "which_M", min = 1L)
}

.coerce_unit_step <- function(
    unit_step,
    conds,
    raw.data,
    unit_step_divisor,
    calib_specs
) {
  if (!is.null(unit_step)) {
    if (!is.numeric(unit_step)) stop("unit_step must be numeric (scalar or vector aligned to conditions), or NULL to compute automatically.")
    if (length(unit_step) == 1L) {
      if (!is.finite(unit_step) || unit_step <= 0) stop("Scalar unit_step must be a finite number > 0.")
      out <- rep(as.numeric(unit_step), length(conds))
      names(out) <- conds
      return(out)
    }
    if (length(unit_step) != length(conds)) {
      stop("If unit_step is a vector, it must have length equal to the number of calibrated sets being stepped and be aligned to that order.")
    }
    if (any(!is.finite(unit_step)) || any(unit_step <= 0)) {
      stop("All per-set unit_step values must be finite numbers > 0.")
    }
    out <- as.numeric(unit_step)
    names(out) <- conds
    return(out)
  }

  if (is.null(unit_step_divisor)) {
    stop("If unit_step is NULL, unit_step_divisor must be provided.")
  }
  if (!is.numeric(unit_step_divisor) || length(unit_step_divisor) != 1L ||
      !is.finite(unit_step_divisor) || unit_step_divisor <= 0) {
    stop("unit_step_divisor must be a single finite number > 0.")
  }

  out <- numeric(length(conds))
  names(out) <- conds

  for (i in seq_along(conds)) {
    cond_name <- conds[i]
    spec <- calib_specs[[cond_name]]
    raw_name <- spec$raw
    thr <- as.numeric(spec$thresholds)
    type_i <- spec$type

    x_raw <- raw.data[[raw_name]]
    xmin <- suppressWarnings(min(x_raw, na.rm = TRUE))
    xmax <- suppressWarnings(max(x_raw, na.rm = TRUE))

    if (!is.finite(xmin) || !is.finite(xmax)) {
      stop("Cannot compute automatic unit_step: raw column '", raw_name, "' must have finite min and max.")
    }

    if (type_i == "f") {
      if (length(thr) < 1L || any(!is.finite(thr))) {
        stop("Automatic unit_step for fuzzy sets requires finite thresholds for set '", cond_name, "'.")
      }
      gaps <- diff(sort(unique(thr)))
      if (length(gaps) < 1L || any(gaps <= 0)) {
        stop("Automatic unit_step for fuzzy sets requires distinct finite thresholds for set '", cond_name, "'.")
      }
      base_span <- min(gaps)
    } else {
      if (length(thr) != 1L || !is.finite(thr)) {
        stop("Automatic unit_step for crisp sets requires one finite threshold for set '", cond_name, "'.")
      }
      if (thr < xmin || thr > xmax) {
        stop("Automatic unit_step for crisp set '", cond_name, "' requires the threshold to lie within the raw-data range.")
      }
      base_span <- min(thr - xmin, xmax - thr)
    }

    if (!is.finite(base_span) || base_span <= 0) {
      stop("Automatic unit_step produced a non-positive base span for set '", cond_name, "'.")
    }

    out[i] <- base_span / unit_step_divisor
  }

  out
}

.validate_qca_min <- function(obj, name) {
  if (is.null(obj)) return(invisible(TRUE))
  if (!inherits(obj, "QCA_min")) {
    stop("`", name, "` must be a `QCA_min` object or NULL.")
  }
  invisible(TRUE)
}

.validate_outcome_conditions_distinct <- function(outcome, conditions) {
  if (is.null(outcome) || length(outcome) < 1L) return(invisible(TRUE))
  if (is.null(conditions)) return(invisible(TRUE))
  if (any(outcome %in% conditions)) {
    stop("`outcome` must not be included in `conditions`; the outcome is handled separately from the causal conditions.")
  }
  invisible(TRUE)
}

test_that("integerish n.cut validator now rejects decimals", {
  expect_error(
    qcaERT:::.as_integerish_scalar(1.5, "n.cut", min = 1L),
    "`n.cut` must be integer-like.",
    fixed = TRUE
  )
})

test_that("integerish n.cut validator now rejects zero", {
  expect_error(
    qcaERT:::.as_integerish_scalar(0, "n.cut", min = 1L),
    "`n.cut` must be >= 1.",
    fixed = TRUE
  )
})

test_that("integerish n.cut validator accepts whole numbers", {
  expect_identical(
    qcaERT:::.as_integerish_scalar(1, "n.cut", min = 1L),
    1L
  )

  expect_identical(
    qcaERT:::.as_integerish_scalar(2.0, "n.cut", min = 1L),
    2L
  )
})

test_that("coerce_which_M now rejects non-integer numerics", {
  expect_error(
    qcaERT:::.coerce_which_M(1.9),
    "`which_M` must be integer-like.",
    fixed = TRUE
  )
})

test_that("coerce_which_M still accepts integer-like input", {
  expect_identical(
    qcaERT:::.coerce_which_M(1),
    1L
  )

  expect_identical(
    qcaERT:::.coerce_which_M(1.0),
    1L
  )
})

test_that("track has been removed from public function arguments", {
  expect_false("track" %in% names(formals(altset.test)))
  expect_false("track" %in% names(formals(cluster.test)))
  expect_false("track" %in% names(formals(incl.test)))
  expect_false("track" %in% names(formals(loo.test)))
  expect_false("track" %in% names(formals(ncut.test)))
  expect_false("track" %in% names(formals(sol.df)))
  expect_false("track" %in% names(formals(subsample.test)))
})

test_that("old track helpers are gone", {
  ns <- asNamespace("qcaERT")

  expect_false(exists(".normalize_track_i_mode", envir = ns, inherits = FALSE))
  expect_false(exists(".select_model_positions", envir = ns, inherits = FALSE))
})

test_that("no-op verbose has been removed from quiet function arguments", {
  quiet <- list(
    incl.test,
    ncut.test,
    calib.test,
    loo.test,
    subsample.test,
    theory.test
  )

  for (fn in quiet) {
    expect_false("verbose" %in% names(formals(fn)))
  }

  expect_true("verbose" %in% names(formals(altset.test)))
  expect_true("verbose" %in% names(formals(sol.df)))
})

tiny_qca <- data.frame(
  Y = c(1, 1, 0, 0),
  A = c(1, 1, 0, 0),
  B = c(1, 0, 1, 0)
)

qcaert_schema_raw <- function() {
  data.frame(
    A_raw = c(0, 5, 10, 25, 35, 50, 65, 75, 90, 100),
    B_raw = c(0, 100, 0, 100, 0, 100, 0, 100, 0, 100)
  )
}

qcaert_schema_calib3 <- function(raw = qcaert_schema_raw()) {
  A <- QCA::calibrate(raw$A_raw, type = "fuzzy", thresholds = c(20, 45, 75))
  data.frame(
    A = A,
    B = QCA::calibrate(raw$B_raw, type = "crisp", thresholds = 50),
    Y = as.integer(A > 0.5)
  )
}

qcaert_schema_calib6 <- function(raw = qcaert_schema_raw()) {
  A <- QCA::calibrate(
    raw$A_raw,
    type = "fuzzy",
    method = "direct",
    thresholds = c(10, 20, 30, 70, 80, 90),
    logistic = FALSE
  )

  data.frame(
    A = A,
    B = QCA::calibrate(raw$B_raw, type = "crisp", thresholds = 50),
    Y = as.integer(A > 0.5),
    cluster = rep(c("C1", "C2"), each = 5),
    unit = paste0("U", seq_len(nrow(raw)))
  )
}

qcaert_schema_indirect <- function(raw = qcaert_schema_raw()) {
  A <- QCA::calibrate(
    raw$A_raw,
    type = "fuzzy",
    method = "indirect",
    thresholds = c(25, 50, 75)
  )

  data.frame(
    A = A,
    B = QCA::calibrate(raw$B_raw, type = "crisp", thresholds = 50),
    Y = as.integer(A > 0.5)
  )
}

qcaert_schema_calib6_spec <- function() {
  list(
    A = list(
      raw = "A_raw",
      type = "fuzzy",
      method = "direct",
      thresholds = c(10, 20, 30, 70, 80, 90),
      calibrate = list(logistic = FALSE)
    ),
    B = list(
      raw = "B_raw",
      type = "crisp",
      thresholds = 50
    )
  )
}

qcaert_schema_indirect_spec <- function() {
  list(
    A = list(
      raw = "A_raw",
      type = "fuzzy",
      method = "indirect",
      thresholds = c(25, 50, 75)
    ),
    B = list(
      raw = "B_raw",
      type = "crisp",
      thresholds = 50
    )
  )
}

qcaert_truth_table <- function(data, outcome, conditions, incl.cut = 0.75, n.cut = 1, ...) {
  skip_if_not_installed("QCA")

  args <- c(
    list(
      data = data,
      outcome = outcome,
      conditions = conditions,
      incl.cut = incl.cut,
      n.cut = n.cut
    ),
    list(...)
  )

  suppressWarnings(do.call(QCA::truthTable, args))
}

qcaert_fixture_direct6 <- function() {
  skip_if_not_installed("QCA")

  raw <- qcaert_schema_raw()
  calib <- qcaert_schema_calib6(raw)

  list(
    raw = raw,
    calib = calib,
    outcome = "Y",
    conditions = c("A", "B"),
    test.conditions = "A",
    calib_spec = qcaert_schema_calib6_spec(),
    dir.exp = c("1", "1")
  )
}

qcaert_fixture_outcome_calibration <- function() {
  skip_if_not_installed("QCA")

  raw <- qcaert_schema_raw()
  raw$Y_raw <- raw$A_raw

  calib <- qcaert_schema_calib6(raw)
  calib$Y <- QCA::calibrate(
    raw$Y_raw,
    type = "fuzzy",
    method = "direct",
    thresholds = c(10, 20, 30, 70, 80, 90),
    logistic = FALSE
  )

  calib_spec <- qcaert_schema_calib6_spec()
  calib_spec$Y <- list(
    raw = "Y_raw",
    type = "fuzzy",
    method = "direct",
    thresholds = c(10, 20, 30, 70, 80, 90),
    calibrate = list(logistic = FALSE)
  )

  list(
    raw = raw,
    calib = calib,
    outcome = "Y",
    conditions = c("A", "B"),
    calib_spec = calib_spec,
    dir.exp = c("1", "1")
  )
}

qcaert_fixture_indirect <- function() {
  skip_if_not_installed("QCA")

  raw <- qcaert_schema_raw()
  calib <- qcaert_schema_indirect(raw)

  list(
    raw = raw,
    calib = calib,
    outcome = "Y",
    conditions = c("A", "B"),
    test.conditions = "A",
    calib_spec = qcaert_schema_indirect_spec(),
    dir.exp = c("1", "1")
  )
}

qcaert_fixture_cluster <- function() {
  skip_if_not_installed("QCA")

  direct <- qcaert_fixture_direct6()
  data <- direct$calib
  data$cluster <- rep(c("C1", "C2"), each = 5)
  data$unit <- rep(paste0("U", seq_len(5)), times = 2)

  tt <- qcaert_truth_table(
    data = data,
    outcome = direct$outcome,
    conditions = direct$conditions,
    incl.cut = 0.75,
    n.cut = 1
  )

  list(
    data = data,
    truth_table = tt,
    outcome = direct$outcome,
    conditions = direct$conditions,
    cluster_id = "cluster",
    unit_id = "unit",
    dir.exp = direct$dir.exp
  )
}

qcaert_fixture_lr <- function() {
  skip_if_not_installed("QCA")

  data_env <- new.env(parent = emptyenv())
  utils::data("LR", package = "QCA", envir = data_env)
  raw <- data_env$LR

  conditions <- c("DEV", "URB", "LIT", "IND", "STB")
  outcome <- "SURV"
  variables <- c(conditions, outcome)
  threshold_groups <- c(DEV = 7, URB = 4, LIT = 4, IND = 4, STB = 4, SURV = 4)
  thresholds <- lapply(variables, function(variable) {
    QCA::findTh(raw[[variable]], groups = threshold_groups[[variable]])
  })
  names(thresholds) <- variables

  calib <- raw
  for (variable in variables) {
    calib[[variable]] <- QCA::calibrate(
      raw[[variable]],
      type = "fuzzy",
      thresholds = thresholds[[variable]]
    )
  }

  calib_spec <- lapply(conditions, function(condition) {
    list(
      raw = condition,
      type = "fuzzy",
      method = "direct",
      thresholds = thresholds[[condition]]
    )
  })
  names(calib_spec) <- conditions

  list(
    raw = raw,
    calib = calib,
    thresholds = thresholds,
    calib_spec = calib_spec,
    outcome = outcome,
    conditions = conditions,
    test.conditions = conditions,
    dir.exp = rep("1", length(conditions))
  )
}

qcaert_fixture_lr_solutions <- function() {
  skip_if_not_installed("QCA")

  fixture <- qcaert_fixture_lr()
  tt <- qcaert_truth_table(
    data = fixture$calib,
    outcome = fixture$outcome,
    conditions = fixture$conditions,
    incl.cut = 0.8,
    n.cut = 1,
    complete = TRUE,
    show.cases = TRUE,
    sort.by = c("incl", "n")
  )
  enhanced <- QCA::findRows(tt, type = 2)

  list(
    fixture = fixture,
    truth_table = tt,
    conservative = QCA::minimize(
      tt,
      details = TRUE,
      show.cases = FALSE,
      use.tilde = FALSE
    ),
    intermediate = QCA::minimize(
      tt,
      details = TRUE,
      show.cases = FALSE,
      use.tilde = FALSE,
      dir.exp = fixture$dir.exp,
      include = "?",
      exclude = enhanced
    ),
    parsimonious = QCA::minimize(
      tt,
      details = TRUE,
      show.cases = FALSE,
      use.tilde = FALSE,
      include = "?"
    )
  )
}

expect_qcaert_schema <- function(obj, class, top, diagnostics, results, settings = NULL) {
  expect_identical(class(obj), class)
  expect_identical(names(obj), top)
  expect_identical(names(obj$diagnostics), diagnostics)

  if (is.data.frame(obj$results)) {
    expect_identical(names(obj$results), results)
  } else {
    expect_identical(names(obj$results), results)
  }

  expect_identical(names(as.data.frame(obj)), if (is.data.frame(obj$results)) results else names(obj$results$overview))

  if (!is.null(settings)) {
    expect_identical(names(obj$settings), settings)
  }
}

test_that("shared direct-six and indirect fixtures are coherent", {
  skip_if_not_installed("QCA")

  direct <- qcaert_fixture_direct6()
  indirect <- qcaert_fixture_indirect()

  expect_identical(direct$outcome, "Y")
  expect_identical(direct$conditions, c("A", "B"))
  expect_identical(direct$calib_spec$A$method, "direct")
  expect_identical(length(direct$calib_spec$A$thresholds), 6L)
  expect_true(all(direct$calib$A >= 0 & direct$calib$A <= 1))

  expect_identical(indirect$outcome, "Y")
  expect_identical(indirect$conditions, c("A", "B"))
  expect_identical(indirect$calib_spec$A$method, "indirect")
  expect_identical(length(indirect$calib_spec$A$thresholds), 3L)
  expect_true(all(indirect$calib$A >= 0 & indirect$calib$A <= 1))
})

test_that("shared cluster fixture has truth table and repeated units", {
  skip_if_not_installed("QCA")

  cluster <- qcaert_fixture_cluster()

  expect_s3_class(cluster$truth_table, "QCA_tt")
  expect_identical(cluster$cluster_id, "cluster")
  expect_identical(cluster$unit_id, "unit")
  expect_true(any(duplicated(cluster$data$unit)))
  expect_setequal(unique(cluster$data$cluster), c("C1", "C2"))
})

test_that("shared LR fixture builds evaluated QCA solution objects", {
  skip_if_not_installed("QCA")

  lr <- qcaert_fixture_lr_solutions()

  expect_identical(lr$fixture$outcome, "SURV")
  expect_identical(lr$fixture$conditions, c("DEV", "URB", "LIT", "IND", "STB"))
  expect_identical(length(lr$fixture$thresholds$DEV), 6L)
  expect_identical(length(lr$fixture$thresholds$URB), 3L)
  expect_s3_class(lr$truth_table, "QCA_tt")
  expect_s3_class(lr$conservative, "QCA_min")
  expect_s3_class(lr$intermediate, "QCA_min")
  expect_s3_class(lr$parsimonious, "QCA_min")
})

qcaert_regular_result_classes <- c(
  "altset_test",
  "calib_test",
  "incl_test",
  "loo_test",
  "ncut_test",
  "subsample_test"
)

qcaert_expect_clean_names <- function(names, label = "object") {
  expect_false(
    is.null(names),
    info = paste(label, "must have names")
  )
  expect_false(
    anyNA(names),
    info = paste(label, "must not have missing names")
  )
  expect_false(
    any(names == ""),
    info = paste(label, "must not have empty names")
  )
  expect_false(
    anyDuplicated(names) > 0L,
    info = paste(label, "must not have duplicated names")
  )
  expect_false(
    any(grepl("\\.[0-9]+$", names)),
    info = paste(label, "must not contain repaired duplicate names")
  )

  invisible(names)
}

qcaert_expect_table <- function(x, label = "table", required = NULL, exact = NULL, allow_null = FALSE) {
  if (allow_null && is.null(x)) {
    return(invisible(x))
  }

  expect_true(is.data.frame(x), info = paste(label, "must be a data frame"))
  qcaert_expect_clean_names(names(x), label)

  if (!is.null(required)) {
    missing <- setdiff(required, names(x))
    expect_true(
      length(missing) == 0L,
      info = paste(label, "is missing columns:", paste(missing, collapse = ", "))
    )
  }

  if (!is.null(exact)) {
    expect_identical(names(x), exact)
  }

  invisible(x)
}

qcaert_expect_print_ok <- function(x) {
  expect_error(capture.output(print(x)), NA)
  invisible(x)
}

qcaert_expect_settings <- function(settings, required = NULL) {
  expect_type(settings, "list")

  if (!is.null(names(settings))) {
    qcaert_expect_clean_names(names(settings), "settings")
  }

  if (!is.null(required)) {
    missing <- setdiff(required, names(settings))
    expect_true(
      length(missing) == 0L,
      info = paste("settings is missing fields:", paste(missing, collapse = ", "))
    )
  }

  invisible(settings)
}

expect_qcaert_result_structure <- function(
  x,
  class = NULL,
  diagnostics = NULL,
  results = NULL,
  settings = NULL,
  top = NULL,
  print = TRUE
) {
  if (!is.null(class)) {
    expect_s3_class(x, class)
    expect_false(
      identical(class, "cluster_test"),
      info = "Use expect_qcaert_cluster_structure() for cluster_test objects."
    )
  }

  expect_type(x, "list")
  qcaert_expect_clean_names(names(x), "result object")

  required_top <- unique(c("diagnostics", "results", top))
  missing_top <- setdiff(required_top, names(x))
  expect_true(
    length(missing_top) == 0L,
    info = paste("result object is missing components:", paste(missing_top, collapse = ", "))
  )

  qcaert_expect_table(x$diagnostics, "diagnostics", required = diagnostics)
  qcaert_expect_table(x$results, "results", required = results)
  expect_identical(as.data.frame(x), x$results)

  if ("settings" %in% names(x)) {
    qcaert_expect_settings(x$settings, required = settings)
  } else if (!is.null(settings)) {
    expect_true(FALSE, info = "result object is missing settings")
  }

  if (print) {
    qcaert_expect_print_ok(x)
  }

  invisible(x)
}

expect_qcaert_cluster_structure <- function(
  x,
  diagnostics = NULL,
  overview = NULL,
  clusters = NULL,
  units = NULL,
  settings = NULL,
  top = NULL,
  print = TRUE
) {
  expect_s3_class(x, "cluster_test")
  expect_type(x, "list")
  qcaert_expect_clean_names(names(x), "cluster result object")

  required_top <- unique(c("diagnostics", "results", top))
  missing_top <- setdiff(required_top, names(x))
  expect_true(
    length(missing_top) == 0L,
    info = paste("cluster result object is missing components:", paste(missing_top, collapse = ", "))
  )

  qcaert_expect_table(x$diagnostics, "diagnostics", required = diagnostics)

  expect_type(x$results, "list")
  expect_identical(names(x$results), c("overview", "clusters", "units"))
  qcaert_expect_table(x$results$overview, "results$overview", required = overview)
  qcaert_expect_table(x$results$clusters, "results$clusters", required = clusters, allow_null = TRUE)
  qcaert_expect_table(x$results$units, "results$units", required = units, allow_null = TRUE)
  expect_identical(as.data.frame(x), x$results$overview)

  if ("settings" %in% names(x)) {
    qcaert_expect_settings(x$settings, required = settings)
  } else if (!is.null(settings)) {
    expect_true(FALSE, info = "cluster result object is missing settings")
  }

  if (print) {
    qcaert_expect_print_ok(x)
  }

  invisible(x)
}

---
title: "Getting started with qcaERT"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Getting started with qcaERT}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = FALSE
)
```

qcaERT adds robustness checks to a normal QCA workflow. It is meant for the
moment after calibration, truth table construction, and minimization, when the
main question becomes:

> How much does this solution depend on my thresholds, cutoffs, cases, sample,
> or grouping structure?

This vignette gives a short path through the package. For a map from robustness
questions to functions, see `?qcaERT_tests`.

## Start from a QCA analysis

The usual workflow still begins with `QCA`. In this example, the `LR` data are
calibrated, minimized, and then passed to qcaERT.

```{r setup-lr}
library(QCA)
library(qcaERT)

data(LR)

conditions <- c("DEV", "URB", "LIT", "IND", "STB")
outcome <- "SURV"
dir_exp <- rep("1", length(conditions))

thresholds <- list(
  DEV = findTh(LR$DEV, groups = 7),
  URB = findTh(LR$URB, groups = 4),
  LIT = findTh(LR$LIT, groups = 4),
  IND = findTh(LR$IND, groups = 4),
  STB = findTh(LR$STB, groups = 4),
  SURV = findTh(LR$SURV, groups = 4)
)

dat <- LR
dat$DEV <- calibrate(LR$DEV, type = "fuzzy", thresholds = thresholds$DEV)
dat$URB <- calibrate(LR$URB, type = "fuzzy", thresholds = thresholds$URB)
dat$LIT <- calibrate(LR$LIT, type = "fuzzy", thresholds = thresholds$LIT)
dat$IND <- calibrate(LR$IND, type = "fuzzy", thresholds = thresholds$IND)
dat$STB <- calibrate(LR$STB, type = "fuzzy", thresholds = thresholds$STB)
dat$SURV <- calibrate(LR$SURV, type = "fuzzy", thresholds = thresholds$SURV)

tt <- truthTable(
  data = dat,
  outcome = outcome,
  conditions = conditions,
  incl.cut = 0.8,
  n.cut = 1
)

sol <- minimize(tt, include = "?", dir.exp = dir_exp)
sol.df(intermediate = sol, solution = "intermediate")
```

`sol.df()` is not a robustness test. It is a compact table builder for QCA
minimization output.

## Choose a robustness test

Each qcaERT function answers a different robustness question.

| Question | Function |
| --- | --- |
| Do calibration thresholds matter? | `calib.test()` |
| Does the inclusion cutoff matter? | `incl.test()` |
| Does the frequency cutoff matter? | `ncut.test()` |
| Do individual cases drive the solution? | `loo.test()` |
| Is the solution stable across subsamples? | `subsample.test()` |
| What if several analysis choices vary together? | `altset.test()` |
| Do theoretically motivated condition sets lead to different solutions? | `theory.test()` |
| Do results differ across groups or clusters? | `cluster.test()` |
| How do I turn QCA solutions into a table? | `sol.df()` |

## Test inclusion and frequency cutoffs

`incl.test()` and `ncut.test()` are the simplest robustness tests. They move one
truth table cutoff below and above the baseline value.

```{r incl-ncut}
incl_out <- incl.test(
  data = dat,
  outcome = outcome,
  conditions = conditions,
  incl.cut = 0.8,
  n.cut = 1,
  step = 0.02,
  max_steps = 5,
  solution = "all",
  dir.exp = dir_exp,
  progress = TRUE
)

incl_out
as.data.frame(incl_out)
incl_out$diagnostics

ncut_out <- ncut.test(
  data = dat,
  outcome = outcome,
  conditions = conditions,
  n.cut = 1,
  incl.cut = 0.8,
  step = 1,
  max_steps = 3,
  solution = "all",
  dir.exp = dir_exp,
  progress = TRUE
)

ncut_out
as.data.frame(ncut_out)
```

The printed object is the quick view. `as.data.frame()` returns the clean table.
`diagnostics` contains the more detailed internal table.

## Test calibration thresholds

`calib.test()` perturbs one calibration anchor at a time. Use `calib_spec` as
the calibration specification: it records the raw column, calibration type, method,
and thresholds in one place. A good default is to let qcaERT compute the
perturbation size from the threshold spacing instead of choosing a raw number
by hand.

```{r calib}
calib_spec <- list(
  DEV = list(raw = "DEV", type = "fuzzy", thresholds = thresholds$DEV),
  URB = list(raw = "URB", type = "fuzzy", thresholds = thresholds$URB),
  LIT = list(raw = "LIT", type = "fuzzy", thresholds = thresholds$LIT),
  IND = list(raw = "IND", type = "fuzzy", thresholds = thresholds$IND),
  STB = list(raw = "STB", type = "fuzzy", thresholds = thresholds$STB)
)

calib_spec_outcome <- calib_spec
calib_spec_outcome$SURV <- list(
  raw = "SURV",
  type = "fuzzy",
  thresholds = thresholds$SURV
)

calib_out <- calib.test(
  raw.data = LR,
  calib.data = dat,
  outcome = outcome,
  conditions = conditions,
  calib_spec = calib_spec,
  test.conditions = c("DEV", "URB"),
  unit_step = NULL,
  unit_step_divisor = 10,
  max_steps = 5,
  incl.cut = 0.8,
  n.cut = 1,
  solution = "all",
  dir.exp = dir_exp,
  progress = TRUE
)

calib_out
as.data.frame(calib_out)
calib_out$bounds
```

`conditions` names the full condition set used in the QCA model.
`test.conditions` names the subset whose calibration thresholds should be
perturbed. If `test.conditions` is not supplied, qcaERT tests every condition in
`conditions`.

Outcome calibration is requested explicitly. The outcome is not added to
`conditions`; instead, use `test.outcome = TRUE` and include the outcome in
`calib_spec`.

```{r calib-outcome}
calib_outcome <- calib.test(
  raw.data = LR,
  calib.data = dat,
  outcome = outcome,
  conditions = conditions,
  calib_spec = calib_spec_outcome,
  test.conditions = NULL,
  test.outcome = TRUE,
  unit_step = NULL,
  unit_step_divisor = 10,
  max_steps = 5,
  incl.cut = 0.8,
  n.cut = 1,
  solution = "all",
  dir.exp = dir_exp,
  progress = TRUE
)
```

## Test cases and samples

`loo.test()` removes cases one at a time. `subsample.test()` runs repeated
subsamples.

```{r cases-samples}
loo_out <- loo.test(
  data = dat,
  outcome = outcome,
  conditions = conditions,
  cases = 1:5,
  incl.cut = 0.8,
  n.cut = 1,
  solution = "all",
  dir.exp = dir_exp,
  progress = TRUE
)

loo_out
as.data.frame(loo_out)

subsample_out <- subsample.test(
  data = dat,
  outcome = outcome,
  conditions = conditions,
  sample_prop = 0.8,
  reps = 25,
  seed = 123,
  incl.cut = 0.8,
  n.cut = 1,
  solution = "all",
  dir.exp = dir_exp,
  progress = TRUE
)

subsample_out
subsample_out$summary
```

## Test alternative analysis sets

`altset.test()` samples alternative analysis settings. It can vary calibration
thresholds, `incl.cut`, and `n.cut` in the same run.

```{r altsets}
altset_out <- altset.test(
  raw.data = LR,
  calib.data = dat,
  outcome = outcome,
  conditions = conditions,
  calib_spec = calib_spec,
  test.conditions = c("DEV", "URB"),
  unit_step = NULL,
  unit_step_divisor = 10,
  calib_max_steps = 5,
  incl.cut = 0.8,
  incl_step = 0.02,
  incl_max_steps = 5,
  n.cut = 1,
  ncut_step = 1,
  ncut_max_steps = 2,
  n_draws = 50,
  seed = 123,
  solution = "all",
  dir.exp = dir_exp,
  progress = TRUE
)

altset_out
as.data.frame(altset_out)
altset_out$summary
```

Use `calib.test()` when you want directional threshold bounds. Use
`altset.test()` when you want sampled compound perturbations.

## Compare theory specifications

`theory.test()` compares several theoretically motivated condition sets under
the same outcome, truth-table cutoffs, solution setting, exclusion handling, and
the same settings for `which_M` and `i_mode`. Each theory gets its own truth
table; exclusions are recomputed separately when `exclude_mode = "recompute"`.

```{r theory}
theories <- list(
  development = c("DEV", "URB", "LIT"),
  industrial = c("DEV", "URB", "IND"),
  broad = c("DEV", "URB", "LIT", "IND", "STB")
)

dir_exp_theories <- list(
  development = c("1", "1", "1"),
  industrial = c("1", "1", "1"),
  broad = c("1", "1", "1", "1", "1")
)

theory_out <- theory.test(
  data = dat,
  outcome = outcome,
  theories = theories,
  incl.cut = 0.8,
  n.cut = 1,
  solution = "all",
  dir.exp = dir_exp_theories,
  progress = TRUE
)

theory_out
as.data.frame(theory_out)
theory_out$results$solutions
theory_out$results$pairwise
```

The model table compares fit and truth-table diagnostics by theory and solution_type.
The solutions table extracts the selected terms. The pairwise table compares
selected solution memberships across theories.

## Test clusters

`cluster.test()` starts from an existing truth table and compares selected
configurations across groups.

```{r clusters}
cluster_data <- dat
cluster_data$region <- ifelse(seq_len(nrow(cluster_data)) %% 2 == 0, "A", "B")
cluster_data$unit <- rownames(cluster_data)

cluster_out <- cluster.test(
  data = cluster_data,
  tt = tt,
  cluster_id = "region",
  unit_id = "unit",
  solution = "all",
  dir.exp = dir_exp,
  progress = TRUE
)

cluster_out
as.data.frame(cluster_out)
cluster_out$results$clusters
cluster_out$results$units
```

`cluster.test()` is one of the structured-result exceptions in the package:
`results` is a list with `overview`, `clusters`, and `units`.
The other structured-result exception is `theory.test()`, whose `results`
component contains `models`, `solutions`, and `pairwise`.

## Plot selected results

Plotting is optional and requires `ggplot2`.

```{r plots}
plot(incl_out, solution_type = "conservative")
plot(calib_out, solution_type = "conservative")
plot(calib_out, solution_type = "conservative", type = "heatmap")
plot(calib_out, solution_type = "conservative", type = "trace", set = "DEV", anchor = "E1", direction = "lower")
plot(theory_out, solution_type = "conservative")
```

For plotting details, see `?qcaERT_plots`.

## Next

Read:

- `vignette("qcaERT-result-objects", package = "qcaERT")` for the common
  returned-object structure.
- `vignette("qcaERT-calibration", package = "qcaERT")` for calibration
  specifications, scale-aware threshold moves, and alternative sets.
- `?qcaERT_conventions` for the package-wide argument and output conventions.

---
title: "Reading qcaERT result objects"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Reading qcaERT result objects}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = FALSE
)
```

qcaERT functions are designed as siblings. The details differ by test, but the
main returned-object structure is intentionally shared.

The code fragments in this vignette use object names such as `incl_out`,
`calib_out`, and `cluster_out`. The getting-started vignette shows how to
create those objects in a normal QCA workflow.

Most robustness objects contain:

- `results`: the clean table
- `diagnostics`: the detailed table for inspection and troubleshooting
- `settings`: the arguments and derived settings used in the run
- supporting components such as `baseline`, `bounds`, `by_direction`,
  `by_case`, `by_run`, `by_draw`, or `summary`

Main summaries are:

```{r core-rule}
print(x)
as.data.frame(x)
x$results
x$diagnostics
```

`print(x)` gives a concise summary. `as.data.frame(x)` returns the clean table.
`x$diagnostics` is where to look when you need to understand a specific run,
case, draw, or stopping reason.

## Boundary tests

`incl.test()`, `ncut.test()`, and `calib.test()` are boundary tests. They move a
quantity away from the baseline until either the monitored solution changes, an
error occurs, or the search reaches a limit.

```{r boundary-layout}
as.data.frame(incl_out)
incl_out$diagnostics
incl_out$bounds

as.data.frame(ncut_out)
ncut_out$diagnostics
ncut_out$bounds

as.data.frame(calib_out)
calib_out$diagnostics
calib_out$bounds
```

The clean table answers the practical question: how far did the value move
before the result stopped being stable?

The diagnostics table preserves the path-level detail.

## Case and sample tests

`loo.test()` gives one row per removed case. `subsample.test()` gives one row
per subsample run.

```{r case-sample-layout}
as.data.frame(loo_out)
loo_out$diagnostics
loo_out$by_case

as.data.frame(subsample_out)
subsample_out$diagnostics
subsample_out$by_run
subsample_out$summary
```

Use the clean table to find cases or runs that changed the solution. Use the
supporting list components when you need the full run object for a specific
case or replication.

## Alternative-set tests

`altset.test()` gives one row per random draw. The result object also stores the
draw-level run details and aggregate scores.

```{r altset-layout}
as.data.frame(altset_out)
altset_out$diagnostics
altset_out$by_draw
altset_out$summary
```

The clean table is for scanning. The summary component is for stability scores
across all draws. When calibration is part of the sampled draw, `changed_sets`
names the perturbed calibrated sets and `changed_roles` records whether those
sets were conditions, the outcome, or both.

## Cluster tests

`cluster.test()` and `theory.test()` are the deliberate structured-result
exceptions. For `cluster.test()`, `results` is a list:

- `overview`
- `clusters`
- `units`

```{r cluster-layout}
cluster_out$results$overview
cluster_out$results$clusters
cluster_out$results$units

as.data.frame(cluster_out)
cluster_out$diagnostics
```

For `cluster.test()`, `as.data.frame(cluster_out)` returns
`cluster_out$results$overview`.

## Theory tests

`theory.test()` is also structured because comparative theory testing has three
natural clean tables:

- `models`
- `solutions`
- `pairwise`

```{r theory-layout}
as.data.frame(theory_out)
theory_out$results$models
theory_out$results$solutions
theory_out$results$pairwise
theory_out$diagnostics
theory_out$by_theory
```

For `theory.test()`, `as.data.frame(theory_out)` returns
`theory_out$results$models`. Inspect `results$solutions` when the question is
which terms were selected, and inspect `results$pairwise` when the question is
how far selected solution memberships differ across theories.

## Solution tables

`sol.df()` returns a regular data frame directly.

```{r sol-df-layout}
tab <- sol.df(intermediate = sol, solution = "intermediate")
tab
is.data.frame(tab)
```

Use `sol.df()` when you want to report or inspect QCA minimization output in a
compact table.

## What to inspect first

In normal use, inspect objects in this order:

1. Print the object.
2. Call `as.data.frame(x)` for the clean table.
3. Look at `x$summary` if the function has one.
4. Open `x$diagnostics` only when you need the full detail.
5. For `cluster.test()`, inspect `x$results$clusters` and `x$results$units`
   after checking the overview.
6. For `theory.test()`, inspect `x$results$solutions` and
   `x$results$pairwise` after checking the model table.

The same conventions are described formally in `?qcaERT_conventions`.

---
title: "Calibration robustness and alternative sets"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Calibration robustness and alternative sets}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = FALSE
)
```

Calibration robustness needs a little more structure than the cutoff and case
tests. qcaERT has to know which raw columns created the calibrated
conditions, which calibration method was used, and which thresholds may be
perturbed.

This vignette focuses on `calib.test()` and `altset.test()`.

The examples below use the `LR` data from the `QCA` package.

```{r setup-lr}
library(QCA)
library(qcaERT)

data(LR)

conditions <- c("DEV", "URB", "LIT", "IND", "STB")
outcome <- "SURV"
dir_exp <- rep("1", length(conditions))

thresholds <- list(
  DEV = findTh(LR$DEV, groups = 7),
  URB = findTh(LR$URB, groups = 4),
  LIT = findTh(LR$LIT, groups = 4),
  IND = findTh(LR$IND, groups = 4),
  STB = findTh(LR$STB, groups = 4),
  SURV = findTh(LR$SURV, groups = 4)
)

dat <- LR
dat$DEV <- calibrate(LR$DEV, type = "fuzzy", thresholds = thresholds$DEV)
dat$URB <- calibrate(LR$URB, type = "fuzzy", thresholds = thresholds$URB)
dat$LIT <- calibrate(LR$LIT, type = "fuzzy", thresholds = thresholds$LIT)
dat$IND <- calibrate(LR$IND, type = "fuzzy", thresholds = thresholds$IND)
dat$STB <- calibrate(LR$STB, type = "fuzzy", thresholds = thresholds$STB)
dat$SURV <- calibrate(LR$SURV, type = "fuzzy", thresholds = thresholds$SURV)
```

## Use calibration specifications

Calibration-family functions use `calib_spec`: a named list with one entry per
set you want qcaERT to be able to recalibrate. This structure keeps the raw
source, calibration type, calibration method, thresholds, and extra
`QCA::calibrate()` arguments together.

```{r calib-spec}
calib_spec <- list(
  DEV = list(
    raw = "DEV",
    type = "fuzzy",
    method = "direct",
    thresholds = thresholds$DEV
  ),
  URB = list(
    raw = "URB",
    type = "fuzzy",
    thresholds = thresholds$URB
  ),
  LIT = list(
    raw = "LIT",
    type = "fuzzy",
    thresholds = thresholds$LIT
  ),
  IND = list(
    raw = "IND",
    type = "fuzzy",
    thresholds = thresholds$IND
  ),
  STB = list(
    raw = "STB",
    type = "fuzzy",
    thresholds = thresholds$STB
  )
)

calib_spec_outcome <- calib_spec
calib_spec_outcome$SURV <- list(
  raw = "SURV",
  type = "fuzzy",
  thresholds = thresholds$SURV
)
```

Each entry contains:

- `raw`: the raw column name in `raw.data`
- `type`: `"crisp"` or `"fuzzy"`
- `thresholds`: the baseline calibration thresholds
- `method`: optional, passed to `QCA::calibrate()`
- `calibrate`: optional list of extra `QCA::calibrate()` arguments

For condition-only runs, `calib_spec` must contain exactly one entry for every
condition in `conditions`. When `test.outcome = TRUE`, it must contain the
conditions plus one entry named by `outcome`. This keeps the recalibration
structure the same across the calibration-family functions.

## Conditions and test.conditions

`conditions` is the full condition set in the QCA model.

`test.conditions` is the subset whose thresholds should be perturbed.

```{r conditions}
calib.test(
  raw.data = LR,
  calib.data = dat,
  outcome = "SURV",
  conditions = c("DEV", "URB", "LIT", "IND", "STB"),
  calib_spec = calib_spec,
  test.conditions = c("DEV", "URB"),
  unit_step = NULL,
  unit_step_divisor = 10,
  max_steps = 5,
  incl.cut = 0.8,
  n.cut = 1,
  solution = "all",
  dir.exp = rep("1", 5),
  progress = TRUE
)
```

If `test.conditions` is not supplied, qcaERT tests all conditions in
`conditions`.

## Outcome calibration

The outcome is not a condition and should not be placed in `conditions`.
To test outcome calibration, keep the QCA model unchanged and request the
outcome explicitly with `test.outcome = TRUE`.

Use `test.conditions = NULL` when you want to test only the outcome
calibration.

```{r outcome-calibration}
calib_outcome <- calib.test(
  raw.data = LR,
  calib.data = dat,
  outcome = outcome,
  conditions = conditions,
  calib_spec = calib_spec_outcome,
  test.conditions = NULL,
  test.outcome = TRUE,
  unit_step = NULL,
  unit_step_divisor = 10,
  max_steps = 5,
  incl.cut = 0.8,
  n.cut = 1,
  solution = "all",
  dir.exp = dir_exp,
  progress = TRUE
)

calib_outcome
as.data.frame(calib_outcome)
```

## Anchors by calibration type

qcaERT uses anchor names that match the calibration method.

| Calibration type | Thresholds | Anchor names |
| --- | --- | --- |
| Crisp | one threshold | `T` |
| Fuzzy direct, three thresholds | `c(E, C, I)` | `E`, `C`, `I` |
| Fuzzy direct, six thresholds | `c(E1, C1, I1, I2, C2, E2)` | `E1`, `C1`, `I1`, `I2`, `C2`, `E2` |
| Fuzzy indirect | ordered cutpoints | `T1`, `T2`, ... |

By default, `anchors_to_test = NULL` means qcaERT uses all anchors implied by
each tested condition.

```{r anchors}
calib.test(
  raw.data = LR,
  calib.data = dat,
  outcome = "SURV",
  conditions = conditions,
  calib_spec = calib_spec,
  test.conditions = "DEV",
  anchors_to_test = c("E1", "C1", "I1", "I2", "C2", "E2"),
  unit_step = NULL,
  unit_step_divisor = 10,
  max_steps = 5,
  incl.cut = 0.8,
  n.cut = 1,
  solution = "all",
  dir.exp = dir_exp,
  progress = TRUE
)
```

## Scale-aware perturbations

Raw columns can live on very different scales. A one-unit move may be tiny
for an economic measure and huge for an index. For that reason, qcaERT can
compute the threshold step automatically.

```{r scale-aware}
calib_out <- calib.test(
  raw.data = LR,
  calib.data = dat,
  outcome = outcome,
  conditions = conditions,
  calib_spec = calib_spec,
  test.conditions = c("DEV", "URB"),
  unit_step = NULL,
  unit_step_divisor = 10,
  max_steps = 5,
  incl.cut = 0.8,
  n.cut = 1,
  solution = "all",
  dir.exp = dir_exp,
  progress = TRUE
)
```

With `unit_step = NULL`, qcaERT derives a step from the threshold geometry for
each condition. `unit_step_divisor = 10` means the function uses roughly
one-tenth of the relevant threshold spacing as the move size.

Use a manually supplied `unit_step` only when you have a substantive reason to
define the perturbation size yourself.

## Direct, indirect, and six-threshold calibration

qcaERT supports the calibration-perturbation cases used in the ordinary QCA
workflow:

- crisp calibration
- fuzzy direct calibration with three thresholds
- fuzzy direct calibration with six thresholds
- fuzzy indirect calibration with ordered cutpoints

For indirect calibration, qcaERT treats thresholds as ordered cutpoints. Anchor
names are positional: `T1`, `T2`, and so on.

This scope is deliberate. `calib.test()` and `altset.test()` perturb explicit
calibration thresholds; they are not general multi-value calibration robustness
tools and they do not rediscover thresholds during the perturbation search.

```{r indirect}
calib_spec_indirect <- list(
  A = list(
    raw = "A_raw",
    type = "fuzzy",
    method = "indirect",
    thresholds = c(10, 20, 30, 40)
  )
)
```

The same output convention applies: `results` is the clean table,
`diagnostics` keeps the path detail, and `bounds` summarizes the tested
threshold bounds.

## From calibration bounds to alternative sets

`calib.test()` asks: how far can one anchor move in one direction before the
monitored solution changes?

`altset.test()` asks: what happens across sampled alternative analysis settings?

```{r altset}
altset_out <- altset.test(
  raw.data = LR,
  calib.data = dat,
  outcome = outcome,
  conditions = conditions,
  calib_spec = calib_spec,
  test.conditions = c("DEV", "URB"),
  unit_step = NULL,
  unit_step_divisor = 10,
  calib_max_steps = 5,
  incl.cut = 0.8,
  incl_step = 0.02,
  incl_max_steps = 5,
  n.cut = 1,
  ncut_step = 1,
  ncut_max_steps = 2,
  n_draws = 50,
  seed = 123,
  solution = "all",
  dir.exp = dir_exp,
  progress = TRUE
)

altset_out
as.data.frame(altset_out)
altset_out$summary
```

The same calibration specification is used in both functions. This is intentional:
the meaning of `conditions`, `test.conditions`, and `test.outcome` should not
change from one sibling function to another.

```{r altset-outcome}
altset_outcome <- altset.test(
  raw.data = LR,
  calib.data = dat,
  outcome = outcome,
  conditions = conditions,
  calib_spec = calib_spec_outcome,
  test.conditions = c("DEV", "URB"),
  test.outcome = TRUE,
  unit_step = NULL,
  unit_step_divisor = 10,
  calib_max_steps = 5,
  incl.cut = 0.8,
  incl_step = 0.02,
  incl_max_steps = 5,
  n.cut = 1,
  ncut_step = 1,
  ncut_max_steps = 2,
  n_draws = 50,
  seed = 456,
  solution = "all",
  dir.exp = dir_exp,
  progress = TRUE
)

as.data.frame(altset_outcome)
```

## Plotting calibration results

If `ggplot2` is installed, `plot.calib_test()` can show interval, heatmap, and
trace views.

```{r calib-plots}
plot(calib_out, solution_type = "conservative")
plot(calib_out, solution_type = "conservative", type = "heatmap")
plot(calib_out, solution_type = "conservative", type = "trace", set = "DEV", anchor = "E1", direction = "lower")
```

Trace plots require an existing condition-anchor-direction path. If the
requested path is unavailable, the error message lists the available paths.

See `?qcaERT_plots` for the plotting API and `?qcaERT_conventions` for the
package-wide conventions.

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sol.chart.R
\name{sol.chart}
\alias{sol.chart}
\title{Draw a chart from a sol.df table}
\usage{
sol.chart(
  x,
  conditions = NULL,
  solution_types = c("conservative", "intermediate", "parsimonious"),
  model = NULL,
  intermediate_branch = NULL,
  colors = c(conservative = "#222222", intermediate = "#E69F00", parsimonious =
    "#0072B2"),
  show_cases = TRUE,
  show_pi_fit = TRUE,
  show_solution_fit = TRUE,
  digits = 2,
  title = NULL,
  note = TRUE,
  legend = FALSE,
  point_size = 3.2,
  text_size = 3.3,
  theme = .plot_theme()
)
}
\arguments{
\item{x}{A data frame returned by \code{\link[=sol.df]{sol.df()}}.}

\item{conditions}{Optional character vector giving the condition order. If
\code{NULL}, conditions are inferred from \code{Prime_Implicants} in order of first
appearance.}

\item{solution_types}{Solution types to display, in plotting order. Accepted
values follow the common qcaERT solution-type conventions, excluding \code{"all"}.
\code{"complex"} is accepted as an alias for \code{"conservative"} for display use.}

\item{model}{Optional positive integer selecting which model to display when
\code{x} contains more than one model.}

\item{intermediate_branch}{Optional intermediate branch to display when \code{x}
contains more than one intermediate branch.}

\item{colors}{Named character vector with colors for \code{"conservative"},
\code{"intermediate"}, and \code{"parsimonious"}.}

\item{show_cases}{Logical; if \code{TRUE}, include a cases row. Cases are taken
from the most detailed available configuration in each column.}

\item{show_pi_fit}{Logical; if \code{TRUE}, include prime-implicant consistency,
raw coverage, unique coverage, and PRI rows.}

\item{show_solution_fit}{Logical; if \code{TRUE}, include solution-level
consistency, coverage, and PRI rows.}

\item{digits}{Non-negative integer used to format fit statistics.}

\item{title}{Optional plot title. If \code{NULL}, a default title is used.}

\item{note}{Logical; if \code{TRUE}, include a caption explaining symbols and
colors.}

\item{legend}{Logical; if \code{TRUE}, include ggplot legends for solution type
and condition state.}

\item{point_size}{Size of presence/absence symbols.}

\item{text_size}{Size of table text.}

\item{theme}{A ggplot2 theme object added to the chart.}
}
\value{
A ggplot object.
}
\description{
Turns the solution table returned by \code{\link[=sol.df]{sol.df()}} into a visual chart
of sufficient configurations. The chart uses one column per aligned prime
implicant and shows condition presence, condition absence, prime implicant
fit, solution fit, and optionally cases.
}
\details{
\code{sol.chart()} is a visual display for \code{\link[=sol.df]{sol.df()}} tables. It does not extract
solutions from QCA minimization objects. Use \code{\link[=sol.df]{sol.df()}} first, then pass the
resulting table to \code{sol.chart()}.

Prime implicant columns are aligned according to the following hierarchy:
conservative/complex, intermediate, and parsimonious. When several
detailed configurations simplify into the same intermediate or parsimonious
prime implicant, the simpler prime implicant is repeated across the relevant
columns so that the detailed empirical configurations remain visible.

Filled circles denote condition presence. Open circles denote condition
absence. Empty cells denote conditions that are irrelevant to that prime
implicant at the displayed solution type.
}
\examples{
library(QCA)
data(LC)

conditions <- c("DEV", "URB", "LIT", "IND", "STB")
dir_exp <- rep("1", length(conditions))

tt <- truthTable(
  data = LC,
  outcome = "SURV",
  conditions = conditions,
  incl.cut = 0.8,
  n.cut = 1
)

enhanced <- findRows(tt, type = 2)

con <- minimize(
  input = tt,
  include = "",
  details = TRUE,
  show.cases = FALSE
)

par <- minimize(
  input = tt,
  include = "?",
  exclude = enhanced,
  details = TRUE,
  show.cases = FALSE
)

int <- minimize(
  input = tt,
  include = "?",
  dir.exp = dir_exp,
  exclude = enhanced,
  details = TRUE,
  show.cases = FALSE
)

solution_table <- sol.df(
  conservative = con,
  parsimonious = par,
  intermediate = int,
  solution = "all",
  which_M = 1,
  i_mode = "C1P1",
  include_cases = FALSE,
  digits = 2
)

if (requireNamespace("ggplot2", quietly = TRUE)) {
  library(ggplot2)

  sol.chart(
    solution_table,
    conditions = conditions,
    solution_types = c("conservative", "intermediate", "parsimonious"),
    show_cases = FALSE
  )
}

\donttest{
library(QCA)
library(ggplot2)
data(LR)

conditions <- c("DEV", "URB", "LIT", "IND", "STB")
outcome <- "SURV"
dir_exp <- rep("1", length(conditions))

thresholds <- list(
  DEV = findTh(LR$DEV, groups = 7),
  URB = findTh(LR$URB, groups = 4),
  LIT = findTh(LR$LIT, groups = 4),
  IND = findTh(LR$IND, groups = 4),
  STB = findTh(LR$STB, groups = 4),
  SURV = findTh(LR$SURV, groups = 4)
)

dat <- LR
dat$DEV <- calibrate(LR$DEV, type = "fuzzy", thresholds = thresholds$DEV)
dat$URB <- calibrate(LR$URB, type = "fuzzy", thresholds = thresholds$URB)
dat$LIT <- calibrate(LR$LIT, type = "fuzzy", thresholds = thresholds$LIT)
dat$IND <- calibrate(LR$IND, type = "fuzzy", thresholds = thresholds$IND)
dat$STB <- calibrate(LR$STB, type = "fuzzy", thresholds = thresholds$STB)
dat$SURV <- calibrate(LR$SURV, type = "fuzzy", thresholds = thresholds$SURV)

tt <- truthTable(
  data = dat,
  outcome = outcome,
  conditions = conditions,
  incl.cut = 0.8,
  n.cut = 1,
  complete = TRUE,
  show.cases = TRUE
)
enhanced <- findRows(tt, type = 2)

con <- minimize(tt, include = "", details = TRUE, show.cases = FALSE)
par <- minimize(tt, include = "?", exclude = enhanced, details = TRUE, show.cases = FALSE)
int <- minimize(
  tt,
  include = "?",
  dir.exp = dir_exp,
  exclude = enhanced,
  details = TRUE,
  show.cases = FALSE
)

solution_table <- sol.df(
  conservative = con,
  intermediate = int,
  parsimonious = par,
  solution = "all",
  which_M = 1,
  i_mode = "C1P1",
  include_cases = TRUE,
  digits = 2
)

sol.chart(solution_table)
}

}
\seealso{
\code{\link[=sol.df]{sol.df()}}, \link{qcaERT_plots}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/loo.test.R
\name{loo.test}
\alias{loo.test}
\alias{print.loo_test}
\alias{as.data.frame.loo_test}
\title{Leave-one-out robustness test for QCA solutions}
\usage{
loo.test(
  data,
  outcome,
  conditions = NULL,
  cases = NULL,
  case_labels = NULL,
  calib = c("fixed", "recompute"),
  raw.data = NULL,
  calib_spec = NULL,
  incl.cut = 1,
  n.cut = 1,
  solution = "all",
  include = NULL,
  dir.exp = NULL,
  which_M = 1,
  i_mode = c("all", "C1P1"),
  exclude_mode = c("recompute", "static", "none"),
  exclude_recompute = list(type = 2),
  exclude_static = NULL,
  fit_measures = c("inclS", "PRI", "covS"),
  fit_tol = 0,
  progress = TRUE,
  ...
)

\method{print}{loo_test}(x, row.names = FALSE, ...)

\method{as.data.frame}{loo_test}(x, ...)
}
\arguments{
\item{data}{A data frame object containing the outcome and condition
columns used in the QCA analysis. \code{data} must have at least
two rows.}

\item{outcome}{Name of the outcome. This must be a single
non-empty character string.}

\item{conditions}{Optional character vector of condition names. If \code{NULL},
the condition set is left to \code{\link[QCA:truthTable]{QCA::truthTable()}}. When
\code{calib = "recompute"}, \code{conditions} must be supplied explicitly.}

\item{cases}{Cases to test. Use \code{NULL} to test all rows, a numeric vector of
row indices, or a character vector of case labels.}

\item{case_labels}{Optional character vector of case labels with length
\code{nrow(data)}. If \code{NULL}, the function uses \code{rownames(data)} when
available; otherwise it uses row numbers converted to character.}

\item{calib}{Calibration handling for reduced-data runs. \code{"fixed"} uses the
calibrated values already present in \code{data}. \code{"recompute"} recalibrates
the outcome and conditions after deleting each case, using \code{raw.data} and
\code{calib_spec}.}

\item{raw.data}{Raw-data frame object used when \code{calib = "recompute"}.
It must have the same number of rows as \code{data}.}

\item{calib_spec}{Calibration specification used when
\code{calib = "recompute"}. This must be a named list keyed by the analysis
sets \code{c(outcome, conditions)}. Each entry must describe the raw source
column, the set type, and the calibration inputs used to rebuild the
calibrated set. Use the same \code{calib_spec} structure described for
\code{\link[=calib.test]{calib.test()}}. An entry may additionally contain \code{findTh}, a named list
of \code{\link[QCA:findTh]{QCA::findTh()}} arguments; when supplied, thresholds are re-estimated
after each case deletion instead of reusing the baseline \code{thresholds}.}

\item{incl.cut}{Inclusion cutoff passed to \code{\link[QCA:truthTable]{QCA::truthTable()}}. This must
be a single finite number in \verb{[0, 1]}.}

\item{n.cut}{Truth table frequency cutoff passed to \code{\link[QCA:truthTable]{QCA::truthTable()}}.
This must be an integer-like value of at least \code{1}.}

\item{solution}{Solution type to monitor. Accepted values are \code{"all"},
\code{"con"} or \code{"conservative"}, \code{"par"} or \code{"parsimonious"}, and \code{"int"} or
\code{"intermediate"}.}

\item{include}{Optional minimization include setting. Currently, this
argument accepts only \code{NULL}, \code{""}, or \code{"?"}.}

\item{dir.exp}{Directional expectations used when the monitored solution is
intermediate.}

\item{which_M}{Positive integer giving which solution alternative to use
when minimization returns multiple models.}

\item{i_mode}{Character string controlling intermediate-solution selection.
Accepted values are \code{"all"} and \code{"C1P1"}.}

\item{exclude_mode}{Character string controlling how excluded rows are
handled for parsimonious and intermediate minimization. \code{"recompute"}
recalculates exclusions from each reduced truth table, \code{"static"} reuses
\code{exclude_static}, and \code{"none"} does not use exclusions.}

\item{exclude_recompute}{Named list of arguments passed to \code{\link[QCA:findRows]{QCA::findRows()}}
when \code{exclude_mode = "recompute"}.}

\item{exclude_static}{Already computed exclusion object reused when
\code{exclude_mode = "static"}.}

\item{fit_measures}{Character vector of fit-measure names to compare between
the baseline run and each reduced-data run. Use \code{NULL} to disable fit
comparison.}

\item{fit_tol}{Non-negative tolerance used when deciding whether fit values
changed.}

\item{progress}{Logical; if \code{TRUE} and the session is interactive, show a
text progress bar.}

\item{...}{Additional arguments. In \code{\link[=loo.test]{loo.test()}}, named arguments matching
\code{\link[QCA:truthTable]{QCA::truthTable()}} formals are passed to \code{\link[QCA:truthTable]{QCA::truthTable()}}, and the
remaining named arguments are forwarded to \code{\link[QCA:minimize]{QCA::minimize()}} after
removing names reserved by \code{loo.test()}. The function also
looks in \code{...} for \code{include}, \code{dir.exp}, or \code{direxp} if those arguments
were not supplied explicitly. In \code{\link[=print.loo_test]{print.loo_test()}}, \code{...} is passed to
\code{\link[=print.data.frame]{print.data.frame()}}. In \code{\link[=as.data.frame.loo_test]{as.data.frame.loo_test()}}, \code{...} is ignored.}

\item{x}{A \code{loo_test} object returned by \code{\link[=loo.test]{loo.test()}}.}

\item{row.names}{Logical; passed to \code{\link[=print.data.frame]{print.data.frame()}} by
\code{\link[=print.loo_test]{print.loo_test()}}.}
}
\value{
An object of class \code{loo_test} with the following components:
\describe{
\item{\code{diagnostics}}{A detailed data frame with one row per tested case.
It records the tested row index, case label, run status, solution-change
classification, fit-change classification, the number of changed fit
measures, the largest absolute fit change, and error information when a
reduced-data run fails.}
\item{\code{results}}{A compact data frame with the columns
\code{row_index}, \code{case_label}, \code{status}, \code{solution_change},
\code{fit_changed_types}, \code{n_fit_deltas}, and \code{max_abs_fit_delta}.}
\item{\code{baseline}}{A list containing the baseline analysis built from the
full dataset, including the truth table, minimization output, selected
solution terms used for comparison, fit values, exclusion information,
and calibration information.}
\item{\code{by_case}}{A named list with one entry per tested case. Each entry
stores the baseline run, the reduced-data run when available, change
summaries, fit deltas, exclusion information, calibration information,
and any error information for that case.}
\item{\code{settings}}{A list containing the analysis settings used to build
the result object.}
}

\code{print.loo_test()} prints a concise summary and the \code{results} table.
\code{as.data.frame.loo_test()} returns the \code{results} table.
}
\description{
Deletes one case at a time, reruns the QCA analysis, and records whether the
monitored solution changes, if selected fit measures change, and
whether the reduced-data run ends with an error. The tested cases can be
identified by row index or by case label.
}
\details{
The function first runs the baseline analysis on the full dataset, then
removes one selected case at a time and reruns the analysis on the reduced
data.

When \code{calib = "fixed"}, the reduced-data runs use the calibrated values
already present in \code{data}. When \code{calib = "recompute"}, the function rebuilds
the calibrated outcome and condition columns after each case deletion
using \code{\link[QCA:findTh]{QCA::findTh()}} and \code{\link[QCA:calibrate]{QCA::calibrate()}} as specified in \code{calib_spec}.

For each tested case, the function records whether the monitored solution
changed, whether monitored fit measures changed beyond \code{fit_tol}, and
whether the reduced-data run stopped with a calibration, truth table,
exclusion, or minimization error.

The shared solution-control, exclusion, and returned-object conventions are
described in \code{?qcaERT_conventions}.
}
\examples{
\donttest{
library(QCA)
data(LR)

conditions <- c("DEV", "URB", "LIT", "IND", "STB")
outcome <- "SURV"
dir_exp <- rep("1", length(conditions))

thresholds <- list(
  DEV = findTh(LR$DEV, groups = 7),
  URB = findTh(LR$URB, groups = 4),
  LIT = findTh(LR$LIT, groups = 4),
  IND = findTh(LR$IND, groups = 4),
  STB = findTh(LR$STB, groups = 4),
  SURV = findTh(LR$SURV, groups = 4)
)

dat <- LR
dat$DEV <- calibrate(LR$DEV, type = "fuzzy", thresholds = thresholds$DEV)
dat$URB <- calibrate(LR$URB, type = "fuzzy", thresholds = thresholds$URB)
dat$LIT <- calibrate(LR$LIT, type = "fuzzy", thresholds = thresholds$LIT)
dat$IND <- calibrate(LR$IND, type = "fuzzy", thresholds = thresholds$IND)
dat$STB <- calibrate(LR$STB, type = "fuzzy", thresholds = thresholds$STB)
dat$SURV <- calibrate(LR$SURV, type = "fuzzy", thresholds = thresholds$SURV)

out <- loo.test(
  data = dat,
  outcome = outcome,
  conditions = conditions,
  cases = seq_len(nrow(dat)),
  case_labels = rownames(dat),
  calib = "fixed",
  incl.cut = 0.8,
  n.cut = 1,
  solution = "all",
  dir.exp = dir_exp,
  fit_measures = c("inclS", "PRI", "covS"),
  progress = TRUE
)

out
as.data.frame(out)
}

}
\seealso{
\code{\link[=calib.test]{calib.test()}}, \code{\link[=incl.test]{incl.test()}}, \code{\link[=ncut.test]{ncut.test()}}, \code{\link[=subsample.test]{subsample.test()}},
\code{\link[=altset.test]{altset.test()}}, \code{\link[=theory.test]{theory.test()}}, \code{\link[=cluster.test]{cluster.test()}}, \code{\link[=sol.df]{sol.df()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cluster.test.R
\name{cluster.test}
\alias{cluster.test}
\alias{print.cluster_test}
\alias{as.data.frame.cluster_test}
\title{Cluster heterogeneity diagnostics for QCA configurations}
\usage{
cluster.test(
  data,
  tt,
  cluster_id,
  unit_id = NULL,
  solution = "all",
  include = NULL,
  dir.exp = NULL,
  exclude = NULL,
  which_M = 1,
  i_mode = c("all", "C1P1"),
  necessity = FALSE,
  progress = TRUE,
  ...
)

\method{print}{cluster_test}(x, row.names = FALSE, ...)

\method{as.data.frame}{cluster_test}(x, ...)
}
\arguments{
\item{data}{A data frame object containing the columns used to build
\code{tt} plus the cluster and, optionally, unit identifiers. \code{data}
must have the same number of rows as \code{tt$recoded.data}.}

\item{tt}{A truth table object of class \code{"QCA_tt"}, typically created with
\code{\link[QCA:truthTable]{QCA::truthTable()}}.}

\item{cluster_id}{Name of the column in \code{data} identifying clusters. This
must be a single non-empty character string, and the column must contain
at least two distinct non-missing cluster values.}

\item{unit_id}{Optional name of the column in \code{data} identifying units that
may appear in more than one cluster. If \code{NULL}, units are treated as row
positions and within-unit diagnostics are not available across clusters.}

\item{solution}{Solution type to evaluate. Accepted values are \code{"all"},
\code{"con"} or \code{"conservative"}, \code{"par"} or \code{"parsimonious"}, and \code{"int"} or
\code{"intermediate"}.}

\item{include}{Optional minimization include setting. Currently, this
argument accepts only \code{NULL}, \code{""}, or \code{"?"}.}

\item{dir.exp}{Directional expectations used when the monitored solution is
intermediate. When \code{solution = "all"}, supplying \code{dir.exp} adds
intermediate solutions to the monitored set.}

\item{exclude}{Optional exclusion specification passed to \code{\link[QCA:minimize]{QCA::minimize()}}
for parsimonious and intermediate minimization.}

\item{which_M}{Positive integer giving which solution alternative to use
when minimization returns multiple models.}

\item{i_mode}{Character string controlling intermediate-solution selection.
Accepted values are \code{"all"} and \code{"C1P1"}.}

\item{necessity}{Logical; if \code{TRUE}, compute fit as necessity consistency
and coverage. If \code{FALSE}, compute fit as sufficiency consistency and
coverage.}

\item{progress}{Logical; if \code{TRUE} and the session is interactive, show a
text progress bar while minimizing the monitored solution types.}

\item{...}{Additional arguments. In \code{\link[=cluster.test]{cluster.test()}}, these are filtered and
forwarded to \code{\link[QCA:minimize]{QCA::minimize()}}. The function also looks in \code{...} for
\code{include}, \code{dir.exp} or \code{direxp}, and \code{exclude} or \code{omit} if those
arguments were not supplied explicitly. In \code{\link[=print.cluster_test]{print.cluster_test()}}, \code{...}
is passed to \code{\link[=print.data.frame]{print.data.frame()}}. In
\code{\link[=as.data.frame.cluster_test]{as.data.frame.cluster_test()}}, \code{...} is ignored.}

\item{x}{A \code{cluster_test} object returned by \code{\link[=cluster.test]{cluster.test()}}.}

\item{row.names}{Logical; passed to \code{\link[=print.data.frame]{print.data.frame()}} by
\code{\link[=print.cluster_test]{print.cluster_test()}}.}
}
\value{
An object of class \code{cluster_test} with the following components:
\describe{
\item{\code{diagnostics}}{A detailed data frame with one row per selected
configuration. It records the solution type, configuration key, component count,
status, pooled consistency and coverage, maximum and mean absolute
cluster-level deltas, the clusters with the worst consistency and
coverage values, whether within-unit diagnostics were available, the
number of repeated units, and error information when a configuration could not
be evaluated.}
\item{\code{results}}{A named list with three tables:
\describe{
\item{\code{overview}}{One row per selected configuration, summarizing pooled fit,
maximum cluster deltas, worst clusters, and within-unit
availability.}
\item{\code{clusters}}{One row per configuration-component-cluster combination,
giving cluster-specific consistency, coverage, and deltas from the
pooled configuration values. The whole configuration is stored as
\code{component = "solution"}; separate term-level rows are included only
for multi-term solutions.}
\item{\code{units}}{One row per configuration-component-unit combination for
units observed in more than one cluster, giving within-unit
consistency, coverage, and deltas from the pooled configuration values.
Separate term-level rows are included only for multi-term solutions.}
}}
\item{\code{baseline}}{A list containing the input truth table, minimization
results by solution type, selected solution terms used for comparison,
metadata, and the extracted configuration definitions used for the
heterogeneity diagnostics.}
\item{\code{by_cluster}}{A named list of detailed cluster-level diagnostics
for each selected configuration.}
\item{\code{by_unit}}{A named list of detailed unit-level diagnostics for
each selected configuration when repeated units are available, or \code{NULL}
otherwise.}
\item{\code{settings}}{A list containing the analysis settings used to build
the result object.}
}

\code{print.cluster_test()} prints a compact overview and, when the object has
one displayed configuration, cluster and within-unit details.
\code{as.data.frame.cluster_test()} returns \code{results$overview}.
}
\description{
Evaluates how the consistency and coverage of selected QCA configurations
vary across clusters and, when repeated units are available, across units
observed in more than one cluster. The function starts from an existing
truth table, minimizes it under the requested solution type, extracts the
selected baseline configurations, and then compares pooled fit values with
cluster-specific and within-unit fit values.
}
\details{
The function recovers the outcome membership from \code{tt$recoded.data}, runs
minimization for the requested solution type or solution types, extracts the
selected baseline configurations, and computes pooled fit values for each
configuration and configuration component.

Cluster-level diagnostics compare the pooled fit of each selected
configuration with the fit obtained inside each cluster defined by
\code{cluster_id}.

If \code{unit_id} identifies units that appear in more than one cluster, the
function also computes within-unit diagnostics by comparing the same unit
across clusters.

Configuration extraction uses the data-frame \code{pims} component produced by
\code{\link[QCA:minimize]{QCA::minimize()}}. The shared solution-control, QCA solution-object, and
returned-object conventions are described in \code{?qcaERT_conventions}.
}
\examples{
\donttest{
library(QCA)
data(LR)

conditions <- c("DEV", "URB", "LIT", "IND", "STB")
outcome <- "SURV"
dir_exp <- rep("1", length(conditions))

thresholds <- list(
  DEV = findTh(LR$DEV, groups = 7),
  URB = findTh(LR$URB, groups = 4),
  LIT = findTh(LR$LIT, groups = 4),
  IND = findTh(LR$IND, groups = 4),
  STB = findTh(LR$STB, groups = 4),
  SURV = findTh(LR$SURV, groups = 4)
)

dat <- LR
dat$DEV <- calibrate(LR$DEV, type = "fuzzy", thresholds = thresholds$DEV)
dat$URB <- calibrate(LR$URB, type = "fuzzy", thresholds = thresholds$URB)
dat$LIT <- calibrate(LR$LIT, type = "fuzzy", thresholds = thresholds$LIT)
dat$IND <- calibrate(LR$IND, type = "fuzzy", thresholds = thresholds$IND)
dat$STB <- calibrate(LR$STB, type = "fuzzy", thresholds = thresholds$STB)
dat$SURV <- calibrate(LR$SURV, type = "fuzzy", thresholds = thresholds$SURV)
dat$development_group <- ifelse(
  LR$DEV >= median(LR$DEV, na.rm = TRUE),
  "higher development",
  "lower development"
)
dat$case_id <- rownames(dat)

tt <- truthTable(
  data = dat,
  outcome = outcome,
  conditions = conditions,
  incl.cut = 0.8,
  n.cut = 1,
  complete = TRUE,
  show.cases = TRUE
)
enhanced <- findRows(tt, type = 2)

out <- cluster.test(
  data = dat,
  tt = tt,
  cluster_id = "development_group",
  unit_id = "case_id",
  solution = "intermediate",
  dir.exp = dir_exp,
  exclude = enhanced,
  which_M = 1,
  necessity = FALSE,
  progress = TRUE
)

out
as.data.frame(out)
out$results$clusters
out$results$units
}

}
\seealso{
\code{\link[=calib.test]{calib.test()}}, \code{\link[=incl.test]{incl.test()}}, \code{\link[=ncut.test]{ncut.test()}}, \code{\link[=loo.test]{loo.test()}},
\code{\link[=subsample.test]{subsample.test()}}, \code{\link[=altset.test]{altset.test()}}, \code{\link[=theory.test]{theory.test()}}, \code{\link[=sol.df]{sol.df()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/calib.test.R
\name{calib.test}
\alias{calib.test}
\alias{print.calib_test}
\alias{as.data.frame.calib_test}
\title{Calibration-threshold robustness test for QCA solutions}
\usage{
calib.test(
  raw.data,
  calib.data,
  outcome,
  conditions,
  calib_spec,
  test.conditions = conditions,
  test.outcome = FALSE,
  anchors_to_test = NULL,
  solution = "all",
  include = NULL,
  which_M = 1,
  unit_step = NULL,
  unit_step_divisor = 10,
  max_steps = 20,
  incl.cut = 1,
  n.cut = 1,
  dir.exp = NULL,
  i_mode = c("all", "C1P1"),
  exclude_mode = c("recompute", "static", "none"),
  exclude_recompute = list(type = 2),
  exclude_static = NULL,
  result_shape = c("wide", "long"),
  progress = TRUE,
  ...
)

\method{print}{calib_test}(x, row.names = FALSE, ...)

\method{as.data.frame}{calib_test}(x, ...)
}
\arguments{
\item{raw.data}{A data frame object containing the raw, uncalibrated
condition columns.}

\item{calib.data}{A data frame object containing the calibrated outcome
and calibrated condition columns used for the analysis.}

\item{outcome}{Name of the outcome column in \code{calib.data}. This can be
supplied as a character string or as an unquoted name.}

\item{conditions}{Character vector of calibrated condition names in
\code{calib.data}. This can be supplied as a character vector or as unquoted
names.}

\item{calib_spec}{Named list of calibration specifications. When
\code{test.outcome = FALSE}, it must contain one entry per condition. When
\code{test.outcome = TRUE}, it must contain one entry per condition plus one
entry named by \code{outcome}. Each entry must contain \code{raw}, \code{type}, and
\code{thresholds}, may contain \code{method}, and may contain \code{calibrate}, a named
list of additional \code{\link[QCA:calibrate]{QCA::calibrate()}} arguments.}

\item{test.conditions}{Subset of \code{conditions} to perturb. Use \code{NULL} only
when \code{test.outcome = TRUE} and no condition calibration should be tested.}

\item{test.outcome}{Logical; if \code{TRUE}, also perturb the outcome
calibration while keeping the outcome out of \code{conditions}.}

\item{anchors_to_test}{Character vector of anchors to test, or \code{NULL} to
test all anchors implied by each set's calibration method. Crisp
sets use \code{"T"}. Fuzzy direct three-threshold sets use \code{"E"},
\code{"C"}, and \code{"I"}. Fuzzy direct six-threshold sets use \code{"E1"},
\code{"C1"}, \code{"I1"}, \code{"I2"}, \code{"C2"}, and \code{"E2"}. Fuzzy indirect sets
use \code{"T1"}, \code{"T2"}, and so on.}

\item{solution}{Solution type to monitor. Accepted values are \code{"all"},
\code{"con"} or \code{"conservative"}, \code{"par"} or \code{"parsimonious"}, and \code{"int"} or
\code{"intermediate"}.}

\item{include}{Optional minimization include setting. Currently, this
argument accepts only \code{NULL}, \code{""}, or \code{"?"}.}

\item{which_M}{Positive integer giving which solution alternative to use
when minimization returns multiple models.}

\item{unit_step}{Numeric step size used to move thresholds. Supply either
one value applied to all tested sets, one value per condition when
\code{test.outcome = FALSE}, or one value per \code{c(conditions, outcome)} when
\code{test.outcome = TRUE}. If \code{NULL}, qcaERT computes scale-aware steps from
each set's threshold spacing.}

\item{unit_step_divisor}{Positive number used to compute \code{unit_step}
automatically when \code{unit_step = NULL}. The default is \code{10}.}

\item{max_steps}{Maximum number of upward or downward threshold moves to
attempt for each tested anchor.}

\item{incl.cut}{Inclusion cutoff passed to \code{\link[QCA:truthTable]{QCA::truthTable()}}.}

\item{n.cut}{Frequency cutoff passed to \code{\link[QCA:truthTable]{QCA::truthTable()}}.}

\item{dir.exp}{Directional expectations used when the monitored solution is
intermediate.}

\item{i_mode}{Character string controlling intermediate-solution selection.
Accepted values are \code{"all"} and \code{"C1P1"}.}

\item{exclude_mode}{Character string controlling how excluded rows are
handled for parsimonious and intermediate minimization. \code{"recompute"}
recalculates exclusions from each perturbed truth table using
\code{exclude_recompute}, \code{"static"} reuses \code{exclude_static}, and \code{"none"}
does not use exclusions.}

\item{exclude_recompute}{Named list of arguments passed to \code{\link[QCA:findRows]{QCA::findRows()}}
when \code{exclude_mode = "recompute"}.}

\item{exclude_static}{Already computed exclusion object reused when
\code{exclude_mode = "static"}.}

\item{result_shape}{Layout of the clean \code{results} table when
\code{solution = "all"}. \code{"wide"} keeps one row per tested
set-anchor-direction path with solution-type-specific columns such as
\code{con_last_safe} and \code{par_reason}. \code{"long"} returns one row per tested
set-anchor-direction path and solution type, with a \code{solution_type} column.
Single-solution-type calls keep the compact single-solution-type layout.}

\item{progress}{Logical; if \code{TRUE} and the session is interactive, show a
text progress bar.}

\item{...}{Additional arguments routed through the QCA workflow. Arguments
matching \code{\link[QCA:truthTable]{QCA::truthTable()}} are forwarded to truth table construction;
remaining minimization arguments are filtered and forwarded to
\code{\link[QCA:minimize]{QCA::minimize()}}. The function also looks in \code{...} for \code{include},
\code{dir.exp}, or \code{direxp} if those arguments were not supplied explicitly.
In \code{\link[=print.calib_test]{print.calib_test()}}, \code{...} is passed to \code{\link[=print.data.frame]{print.data.frame()}}. In
\code{\link[=as.data.frame.calib_test]{as.data.frame.calib_test()}}, \code{...} is ignored.}

\item{x}{A \code{calib_test} object returned by \code{\link[=calib.test]{calib.test()}}.}

\item{row.names}{Logical; passed to \code{\link[=print.data.frame]{print.data.frame()}} by
\code{\link[=print.calib_test]{print.calib_test()}}.}
}
\value{
An object of class \code{calib_test} with the following components:
\describe{
\item{\code{diagnostics}}{A detailed data frame with one row per tested
anchor-direction path. When \code{solution = "all"}, each monitored solution
type is searched independently, so diagnostics has one row per tested
anchor, direction, and solution type. It includes the starting threshold, the
last safe value, the first failing value, the number of successful
steps, stop reason, and change classification.}
\item{\code{results}}{A compact data frame with the columns \code{set}, \code{role},
\code{raw}, \code{type}, \code{method}, \code{anchor}, \code{direction}, \code{start},
\code{last_safe}, \code{first_failing}, \code{step_unit}, \code{steps}, \code{total_delta},
\code{pct_raw_range}, and \code{reason}. When \code{solution = "all"} and
\code{result_shape = "wide"}, the row unit remains set-anchor-direction,
but the boundary and reason columns are solution-type-specific, using prefixes
\code{con_}, \code{par_}, and, when available, \code{int_}. When
\code{result_shape = "long"}, \code{results} has one row per tested path and
solution type.}
\item{\code{bounds}}{A named list of compact lower/upper bound matrices, one
per tested set. When \code{solution = "all"}, each tested set contains a
named list of solution-type-specific lower/upper matrices.}
\item{\code{baseline}}{A list containing the baseline analysis status,
selected solution terms used for comparison, metadata, and
solution-type-specific baseline objects.}
\item{\code{by_set}}{A named list containing per-set step results, including
path-level traces.}
\item{\code{settings}}{A list containing the analysis settings used to build
the result object.}
}

\code{print.calib_test()} prints a concise summary and a compact display table.
In that printed table, \code{set} and \code{role} are omitted and the raw source
column is labelled \code{condition}. \code{as.data.frame.calib_test()} returns the
stored \code{results} table.
If \code{ggplot2} is installed, \code{plot.calib_test()} provides interval, heatmap,
and trace views; see \code{?qcaERT_plots}.
}
\description{
Perturbs calibration thresholds for selected conditions, and optionally the
outcome, one anchor and one direction at a time, and checks when the
monitored QCA solution changes.
For each tested path, the function records the starting threshold, the last
value that preserved the baseline solution, the first value that changed the
solution or triggered an error, and the reason the path stopped.
}
\details{
The function rebuilds calibrated condition columns, and the outcome when
\code{test.outcome = TRUE}, from \code{raw.data} using \code{calib_spec}. It then compares
perturbed analyses against the baseline solution under the selected solution
type.

When \code{solution = "all"}, conservative, parsimonious, and, when requested,
intermediate paths are searched independently for each tested set, anchor,
and direction.

Each tested path starts from the supplied threshold for one method-specific
anchor of one set and moves in one direction at a time (\code{"lower"} or
\code{"upper"}), using \code{unit_step} until one of four things happens:
\itemize{
\item the monitored solution changes,
\item minimization fails,
\item the raw-data range or anchor ordering blocks the next move, or
\item \code{max_steps} is reached.
}

The shared solution-control, exclusion, calibration-specification, and
returned-object conventions are described in \code{?qcaERT_conventions}.
}
\examples{
\donttest{
library(QCA)
library(ggplot2)
data(LR)

conditions <- c("DEV", "URB", "LIT", "IND", "STB")
outcome <- "SURV"
dir_exp <- rep("1", length(conditions))

thresholds <- list(
  DEV = findTh(LR$DEV, groups = 7),
  URB = findTh(LR$URB, groups = 4),
  LIT = findTh(LR$LIT, groups = 4),
  IND = findTh(LR$IND, groups = 4),
  STB = findTh(LR$STB, groups = 4),
  SURV = findTh(LR$SURV, groups = 4)
)

dat <- LR
dat$DEV <- calibrate(LR$DEV, type = "fuzzy", thresholds = thresholds$DEV)
dat$URB <- calibrate(LR$URB, type = "fuzzy", thresholds = thresholds$URB)
dat$LIT <- calibrate(LR$LIT, type = "fuzzy", thresholds = thresholds$LIT)
dat$IND <- calibrate(LR$IND, type = "fuzzy", thresholds = thresholds$IND)
dat$STB <- calibrate(LR$STB, type = "fuzzy", thresholds = thresholds$STB)
dat$SURV <- calibrate(LR$SURV, type = "fuzzy", thresholds = thresholds$SURV)

calib_spec <- list(
  DEV = list(raw = "DEV", type = "fuzzy", method = "direct", thresholds = thresholds$DEV),
  URB = list(raw = "URB", type = "fuzzy", method = "direct", thresholds = thresholds$URB),
  LIT = list(raw = "LIT", type = "fuzzy", method = "direct", thresholds = thresholds$LIT),
  IND = list(raw = "IND", type = "fuzzy", method = "direct", thresholds = thresholds$IND),
  STB = list(raw = "STB", type = "fuzzy", method = "direct", thresholds = thresholds$STB)
)

calib_spec_outcome <- calib_spec
calib_spec_outcome$SURV <- list(
  raw = "SURV",
  type = "fuzzy",
  method = "direct",
  thresholds = thresholds$SURV
)

# Common use: test selected calibrated conditions with scale-aware steps.
calib_out <- calib.test(
  raw.data = LR,
  calib.data = dat,
  outcome = outcome,
  conditions = conditions,
  calib_spec = calib_spec,
  test.conditions = c("DEV", "URB"),
  unit_step = NULL,
  unit_step_divisor = 10,
  max_steps = 5,
  incl.cut = 0.8,
  n.cut = 1,
  solution = "all",
  dir.exp = dir_exp,
  progress = TRUE
)

calib_out
as.data.frame(calib_out)
calib_out$bounds
calib_out$diagnostics

plot(calib_out, solution_type = "conservative")
plot(calib_out, solution_type = "conservative", type = "heatmap")

# Special case: test the calibrated outcome without putting it in
# conditions.
outcome_out <- calib.test(
  raw.data = LR,
  calib.data = dat,
  outcome = outcome,
  conditions = conditions,
  calib_spec = calib_spec_outcome,
  test.conditions = NULL,
  test.outcome = TRUE,
  unit_step = NULL,
  unit_step_divisor = 10,
  max_steps = 5,
  incl.cut = 0.8,
  n.cut = 1,
  solution = "all",
  dir.exp = dir_exp,
  progress = TRUE
)

outcome_out
as.data.frame(outcome_out)
plot(outcome_out, solution_type = "conservative")

# Special case: focus on selected anchors in a six-threshold direct
# calibration. For DEV, anchors are E1, C1, I1, I2, C2, and E2.
dev_anchor_out <- calib.test(
  raw.data = LR,
  calib.data = dat,
  outcome = outcome,
  conditions = conditions,
  calib_spec = calib_spec,
  test.conditions = "DEV",
  anchors_to_test = c("E1", "C1", "I1"),
  unit_step = NULL,
  unit_step_divisor = 10,
  max_steps = 5,
  incl.cut = 0.8,
  n.cut = 1,
  solution = "all",
  dir.exp = dir_exp,
  progress = TRUE
)

as.data.frame(dev_anchor_out)

plot(
  dev_anchor_out,
  solution_type = "conservative",
  type = "trace",
  set = "DEV",
  anchor = "E1",
  direction = "lower"
)

# Special case: indirect calibration. Indirect anchors are positional:
# T1, T2, and so on.
thresholds_indirect <- thresholds
thresholds_indirect$DEV <- findTh(LR$DEV, groups = 4)

dat_indirect <- dat
dat_indirect$DEV <- calibrate(
  LR$DEV,
  type = "fuzzy",
  method = "indirect",
  thresholds = thresholds_indirect$DEV
)

calib_spec_indirect <- calib_spec
calib_spec_indirect$DEV <- list(
  raw = "DEV",
  type = "fuzzy",
  method = "indirect",
  thresholds = thresholds_indirect$DEV
)

indirect_out <- calib.test(
  raw.data = LR,
  calib.data = dat_indirect,
  outcome = outcome,
  conditions = conditions,
  calib_spec = calib_spec_indirect,
  test.conditions = "DEV",
  anchors_to_test = c("T1", "T2"),
  unit_step = NULL,
  unit_step_divisor = 10,
  max_steps = 5,
  incl.cut = 0.8,
  n.cut = 1,
  solution = "all",
  dir.exp = dir_exp,
  progress = TRUE
)

as.data.frame(indirect_out)
}

}
\seealso{
\link{qcaERT_plots}, \code{\link[=incl.test]{incl.test()}}, \code{\link[=ncut.test]{ncut.test()}}, \code{\link[=loo.test]{loo.test()}},
\code{\link[=subsample.test]{subsample.test()}}, \code{\link[=altset.test]{altset.test()}}, \code{\link[=theory.test]{theory.test()}}, \code{\link[=cluster.test]{cluster.test()}},
\code{\link[=sol.df]{sol.df()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/qcaERT-package.R
\docType{package}
\name{qcaERT-package}
\alias{qcaERT}
\alias{qcaERT-package}
\title{qcaERT: Enhanced robustness tests for QCA}
\description{
qcaERT extends the usual Qualitative Comparative Analysis (QCA) with tools
for assessing how stable and robust a QCA analysis is. It can test alternative
calibration thresholds, inclusion cutoffs, truth table frequency cutoffs,
case deletion, subsampling, alternative solutions, cluster-specific
analyses and different theoretical configurations. It also offers a compact way to organize and visualize QCA
solutions.
The package is built around the QCA package (Dusa 2019) workflow using
\code{\link[QCA:calibrate]{QCA::calibrate()}}, \code{\link[QCA:truthTable]{QCA::truthTable()}}, \code{\link[QCA:minimize]{QCA::minimize()}}, and
\code{\link[QCA:findRows]{QCA::findRows()}}. It is not intended to cover every possible QCA variant
or calibration transformation.
}
\section{Start Here}{

If you know the robustness concern but not the function name, start with
\code{?qcaERT_tests}, which maps qcaERT's tools.

A typical R workflow for sufficient analysis using qcaERT is:
\enumerate{
\item calibrate data with \code{\link[QCA:calibrate]{QCA::calibrate()}},
\item build a truth table with \code{\link[QCA:truthTable]{QCA::truthTable()}},
\item minimize with \code{\link[QCA:minimize]{QCA::minimize()}},
\item run one or more qcaERT robustness tests, and
\item inspect \code{print(x)}, \code{as.data.frame(x)}, \code{x$diagnostics}, and, where
available, \code{plot(x)}.
}
}

\section{Function Family}{

The main qcaERT functions are:
\itemize{
\item \code{\link[=calib.test]{calib.test()}} for calibration-threshold robustness.
\item \code{\link[=incl.test]{incl.test()}} for truth table inclusion-cutoff robustness.
\item \code{\link[=ncut.test]{ncut.test()}} for truth table frequency-cutoff robustness.
\item \code{\link[=loo.test]{loo.test()}} for leave-one-out case influence.
\item \code{\link[=subsample.test]{subsample.test()}} for repeated subsample stability (advanced).
\item \code{\link[=altset.test]{altset.test()}} for sampled alternative analysis settings that can
combine calibration, inclusion-cutoff, and frequency-cutoff perturbations.
\item \code{\link[=theory.test]{theory.test()}} for comparing theoretically motivated configurations
using the same outcome, truth-table cutoffs, solution type,
exclusion handling, and model-selection settings.
\item \code{\link[=cluster.test]{cluster.test()}} for cluster, group, or repeated-unit heterogeneity.
\item \code{\link[=sol.df]{sol.df()}} for converting QCA minimization objects into compact
solution tables.
\item \code{\link[=sol.chart]{sol.chart()}} for rendering \code{sol.df()} tables as solution charts.
}
}

\section{Returned Objects}{

Most robustness functions return R objects with a common structure:
\itemize{
\item \code{diagnostics}: the detailed table, useful for inspection and
troubleshooting.
\item \code{results}: a cleaner table.
\item \code{settings}: the analysis settings used to create the result.
\item supporting components such as \code{baseline}, \code{bounds}, \code{by_direction},
\code{by_case}, \code{by_run}, \code{by_draw}, or \code{summary}, depending on the test.
}

\code{print()} gives a concise summary. \code{as.data.frame()} returns the main clean
table. \code{cluster.test()} and \code{theory.test()} use structured \code{results} lists:
\code{cluster.test()} returns \code{overview}, \code{clusters}, and \code{units}, while
\code{theory.test()} returns \code{models}, \code{pairwise}, and \code{solutions}.
}

\section{Common Conventions}{

The family-wide argument and output conventions are described in
\code{?qcaERT_conventions}. It explains solution controls such as \code{solution},
\code{include}, \code{dir.exp}, \code{which_M}, and \code{i_mode}; exclusion handling and its
function-specific exceptions; calibration specifications; and the common
returned-object structure.
}

\section{Plotting}{

Plotting requires \code{ggplot2}. Current plot methods are
available for \code{\link[=calib.test]{calib.test()}}, \code{\link[=incl.test]{incl.test()}}, and \code{\link[=theory.test]{theory.test()}} results.
\code{\link[=sol.chart]{sol.chart()}} provides a visual presentation for \code{\link[=sol.df]{sol.df()}} tables. See
\code{?qcaERT_plots}.
}

\section{Learn More}{

\itemize{
\item \code{?qcaERT_tests} for choosing the right robustness test.
\item \code{?qcaERT_conventions} for common argument and output conventions.
\item \code{?qcaERT_plots} for plotting \code{calib_test}, \code{incl_test}, and
\code{theory_test} objects, and for charting \code{sol.df()} tables.
\item \code{vignette("qcaERT-overview", package = "qcaERT")} for a guided
introduction.
\item \code{vignette("qcaERT-result-objects", package = "qcaERT")} for the
common returned-object structure.
\item \code{vignette("qcaERT-calibration", package = "qcaERT")} for
calibration specifications, scale-aware perturbations, and alternative
sets.
\item \code{news(package = "qcaERT")} for development changes.
\item \code{citation(package = "qcaERT")} for citation information.
}
}

\section{Recommended citation}{

Marisguia, B. A. H. (2026). qcaERT: Enhanced Robustness Tests for
Qualitative Comparative Analysis. R package version 0.1.0.

Please also cite the QCA package and the methodological sources relevant
to the analysis.
}

\references{
Dusa, Adrian. 2019. \emph{QCA with R. A Comprehensive Resource}. Cham:
Springer. \url{https://adriandusa.com/research/books/2019-QCA/}.

Ragin, C. C. 2014. \emph{The Comparative Method: Moving Beyond Qualitative and
Quantitative Strategies}. Oakland, California: University of California
Press.
}
\author{
Breno A. H. Marisguia
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/altset.test.R
\name{altset.test}
\alias{altset.test}
\alias{print.altset_test}
\alias{as.data.frame.altset_test}
\title{Alternative-set robustness test for QCA solutions}
\usage{
altset.test(
  raw.data,
  calib.data,
  outcome,
  conditions,
  calib_spec,
  test.conditions = conditions,
  test.outcome = FALSE,
  anchors_to_test = NULL,
  solution = "all",
  include = NULL,
  which_M = 1,
  unit_step = NULL,
  unit_step_divisor = 10,
  calib_max_steps = 20,
  incl.cut = 1,
  incl_step = 0.01,
  incl_max_steps = 20,
  n.cut = 1,
  ncut_step = 1,
  ncut_max_steps = 20,
  dir.exp = NULL,
  i_mode = c("all", "C1P1"),
  exclude_mode = c("recompute", "static", "none"),
  exclude_recompute = list(type = 2),
  exclude_static = NULL,
  n_draws = 100,
  fit_tol = 1e-06,
  seed = NULL,
  progress = TRUE,
  verbose = FALSE,
  ...
)

\method{print}{altset_test}(x, row.names = FALSE, ...)

\method{as.data.frame}{altset_test}(x, ...)
}
\arguments{
\item{raw.data}{A data frame object containing the raw, uncalibrated
condition columns, and the raw outcome column when \code{test.outcome = TRUE}.}

\item{calib.data}{A data frame object containing the calibrated outcome
and calibrated condition columns used for the baseline analysis.}

\item{outcome}{Name of the outcome column in \code{calib.data}. This must be a
single non-empty character string.}

\item{conditions}{Character vector of calibrated condition names in
\code{calib.data}.}

\item{calib_spec}{Named list of calibration specifications, one per
condition, plus one for the outcome when \code{test.outcome = TRUE}. Each entry
must contain \code{raw}, \code{type}, and \code{thresholds}, may contain \code{method}, and may
contain \code{calibrate}, a named list of additional \code{\link[QCA:calibrate]{QCA::calibrate()}}
arguments for that set.}

\item{test.conditions}{Subset of \code{conditions} to perturb when generating
random alternative calibration draws. May be \code{NULL} only when
\code{test.outcome = TRUE}, in which case only the outcome calibration is
perturbed.}

\item{test.outcome}{Logical; if \code{TRUE}, also perturb the outcome calibration.
The outcome must be represented in \code{calib_spec}, and must not be included
in \code{conditions}.}

\item{anchors_to_test}{Character vector of anchors eligible for perturbation,
or \code{NULL} to use all anchors implied by each tested set's calibration
method. Crisp sets use \code{"T"}. Fuzzy direct three-threshold sets use
\code{"E"}, \code{"C"}, and \code{"I"}. Fuzzy indirect sets use \code{"T1"},
\code{"T2"}, and so on. Fuzzy direct six-threshold sets use \code{"E1"},
\code{"C1"}, \code{"I1"}, \code{"I2"}, \code{"C2"}, and \code{"E2"}.}

\item{solution}{Solution type to monitor. Accepted values are \code{"all"},
\code{"con"} or \code{"conservative"}, \code{"par"} or \code{"parsimonious"}, and \code{"int"} or
\code{"intermediate"}.}

\item{include}{Optional minimization include setting. Currently, this
argument accepts only \code{NULL}, \code{""}, or \code{"?"}.}

\item{which_M}{Positive integer giving which solution alternative to use
when minimization returns multiple models.}

\item{unit_step}{Numeric step size used to move calibration thresholds when
building candidate alternative draws. Supply either one value applied to
all calibrated sets or one value per condition, plus the outcome when
\code{test.outcome = TRUE}. If \code{NULL}, qcaERT computes scale-aware steps from
each set's threshold spacing.}

\item{unit_step_divisor}{Positive number used to compute \code{unit_step}
automatically when \code{unit_step = NULL}. The default is \code{10}.}

\item{calib_max_steps}{Maximum number of threshold moves away from the
baseline used when building alternative calibration candidates.}

\item{incl.cut}{Baseline inclusion cutoff passed to \code{\link[QCA:truthTable]{QCA::truthTable()}}.}

\item{incl_step}{Positive step size used to build the candidate grid of
inclusion-cutoff values around \code{incl.cut}.}

\item{incl_max_steps}{Maximum number of stepwise moves away from the
baseline \code{incl.cut} used to build the candidate grid.}

\item{n.cut}{Baseline truth table frequency cutoff passed to
\code{\link[QCA:truthTable]{QCA::truthTable()}}.}

\item{ncut_step}{Positive integer step size used to build the candidate grid
of frequency-cutoff values around \code{n.cut}.}

\item{ncut_max_steps}{Maximum number of stepwise moves away from the
baseline \code{n.cut} used to build the candidate grid.}

\item{dir.exp}{Directional expectations used when the monitored solution is
intermediate.}

\item{i_mode}{Character string controlling intermediate-solution selection.
Accepted values are \code{"all"} and \code{"C1P1"}.}

\item{exclude_mode}{Character string controlling how excluded rows are
handled for parsimonious and intermediate minimization. \code{"recompute"}
recalculates exclusions from each drawn truth table using
\code{exclude_recompute}, \code{"static"} reuses \code{exclude_static}, and \code{"none"}
does not use exclusions.}

\item{exclude_recompute}{Named list of arguments passed to \code{\link[QCA:findRows]{QCA::findRows()}}
when \code{exclude_mode = "recompute"}.}

\item{exclude_static}{Already computed exclusion object reused when
\code{exclude_mode = "static"}.}

\item{n_draws}{Number of random alternative draws to run.}

\item{fit_tol}{Non-negative tolerance used when deciding whether fit values
changed.}

\item{seed}{Optional integer seed passed to \code{\link[=set.seed]{set.seed()}} before generating
the random draws.}

\item{progress}{Logical; if \code{TRUE} and the session is interactive, show a
text progress bar.}

\item{verbose}{Logical; if \code{TRUE}, print a completion message after each
draw.}

\item{...}{Additional arguments routed through the QCA workflow. Arguments
matching \code{\link[QCA:truthTable]{QCA::truthTable()}} are forwarded to truth table construction;
remaining minimization arguments are filtered and forwarded to
\code{\link[QCA:minimize]{QCA::minimize()}}. The function also looks in \code{...} for \code{include},
\code{dir.exp}, or \code{direxp} if those arguments were not supplied explicitly.
In \code{\link[=print.altset_test]{print.altset_test()}}, \code{...} is passed to \code{\link[=print.data.frame]{print.data.frame()}}. In
\code{\link[=as.data.frame.altset_test]{as.data.frame.altset_test()}}, \code{...} is ignored.}

\item{x}{An \code{altset_test} object returned by \code{\link[=altset.test]{altset.test()}}.}

\item{row.names}{Logical; passed to \code{\link[=print.data.frame]{print.data.frame()}} by
\code{\link[=print.altset_test]{print.altset_test()}}.}
}
\value{
An object of class \code{altset_test} with the following components:
\describe{
\item{\code{summary}}{A list with summary values named \code{n_draws},
\code{n_same_solution}, \code{n_fit_compared}, \code{n_same_fit},
\code{score_solution}, \code{score_fit}, \code{score_total},
\code{score_solution_by_solution_type}, and \code{score_fit_by_solution_type}.}
\item{\code{baseline}}{A list containing the baseline analysis, baseline draw
metadata, exclusion information, selected solution terms used for
comparison, fit information, and status information.}
\item{\code{diagnostics}}{A detailed data frame with one row per draw. It
includes the sampled \code{incl.cut} and \code{n.cut}, run status,
solution-change information, fit-change information, changed-set
information, and error information when a draw fails.}
\item{\code{results}}{A compact data frame with the columns
\code{draw}, \code{incl.cut}, \code{n.cut}, \code{status}, \code{n_changed_sets},
\code{changed_sets}, \code{changed_roles}, \code{solution_change}, \code{fit_changed_types},
\code{n_fit_deltas}, and \code{max_abs_fit_delta}.}
\item{\code{by_draw}}{A named list with one entry per draw. Each entry stores
the full run result, solution-comparison information, fit-comparison
information, and draw-level fit details.}
\item{\code{settings}}{A list containing the analysis settings used to build
the result object.}
}

\code{print.altset_test()} prints a concise summary, the \code{results} table, and
solution-type-specific match-rate tables. \code{as.data.frame.altset_test()} returns
the \code{results} table.
}
\description{
Repeatedly generates alternative analyses by jointly varying calibration
anchors, the raw consistency threshold, and the frequency cutoff for truth
table rows, then reruns the QCA analysis for each generated specification.
For each specification, the function records whether the solution under
evaluation changes, whether parameters of fit change, which conditions or
outcome were recalibrated, and whether the analysis failed to produce a
solution.
}
\details{
The function first runs a baseline analysis using \code{calib.data},
\code{incl.cut}, and \code{n.cut}. It then generates \code{n_draws} random alternative
draws.

Each draw samples:
\itemize{
\item one admissible inclusion cutoff from the grid defined by
\code{incl.cut}, \code{incl_step}, and \code{incl_max_steps},
\item one admissible frequency cutoff from the grid defined by \code{n.cut},
\code{ncut_step}, and \code{ncut_max_steps}, and
\item one admissible alternative calibration specification sampled from
\code{calib_spec} by moving method-specific anchors for \code{test.conditions}, and
for the outcome when \code{test.outcome = TRUE}, by integer multiples of each
set's \code{unit_step}.
}

The function rejects a sampled draw if it is identical to the baseline on
all three dimensions and resamples until it gets a non-baseline draw or
reaches the internal retry limit.

The shared solution-control, exclusion, calibration-specification, and
returned-object conventions are described in \code{?qcaERT_conventions}.
}
\examples{
\donttest{
library(QCA)
data(LR)

conditions <- c("DEV", "URB", "LIT", "IND", "STB")
outcome <- "SURV"
dir_exp <- rep("1", length(conditions))

thresholds <- list(
  DEV = findTh(LR$DEV, groups = 7),
  URB = findTh(LR$URB, groups = 4),
  LIT = findTh(LR$LIT, groups = 4),
  IND = findTh(LR$IND, groups = 4),
  STB = findTh(LR$STB, groups = 4),
  SURV = findTh(LR$SURV, groups = 4)
)

dat <- LR
dat$DEV <- calibrate(LR$DEV, type = "fuzzy", thresholds = thresholds$DEV)
dat$URB <- calibrate(LR$URB, type = "fuzzy", thresholds = thresholds$URB)
dat$LIT <- calibrate(LR$LIT, type = "fuzzy", thresholds = thresholds$LIT)
dat$IND <- calibrate(LR$IND, type = "fuzzy", thresholds = thresholds$IND)
dat$STB <- calibrate(LR$STB, type = "fuzzy", thresholds = thresholds$STB)
dat$SURV <- calibrate(LR$SURV, type = "fuzzy", thresholds = thresholds$SURV)

calib_spec <- list(
  DEV = list(raw = "DEV", type = "fuzzy", method = "direct", thresholds = thresholds$DEV),
  URB = list(raw = "URB", type = "fuzzy", method = "direct", thresholds = thresholds$URB),
  LIT = list(raw = "LIT", type = "fuzzy", method = "direct", thresholds = thresholds$LIT),
  IND = list(raw = "IND", type = "fuzzy", method = "direct", thresholds = thresholds$IND),
  STB = list(raw = "STB", type = "fuzzy", method = "direct", thresholds = thresholds$STB)
)

calib_spec_outcome <- calib_spec
calib_spec_outcome$SURV <- list(
  raw = "SURV",
  type = "fuzzy",
  method = "direct",
  thresholds = thresholds$SURV
)

# Common use: sample alternative calibration, incl.cut, and n.cut settings.
altset_out <- altset.test(
  raw.data = LR,
  calib.data = dat,
  outcome = outcome,
  conditions = conditions,
  calib_spec = calib_spec,
  test.conditions = c("DEV", "URB"),
  unit_step = NULL,
  unit_step_divisor = 10,
  calib_max_steps = 5,
  incl.cut = 0.8,
  incl_step = 0.02,
  incl_max_steps = 5,
  n.cut = 1,
  ncut_step = 1,
  ncut_max_steps = 2,
  n_draws = 50,
  seed = 123,
  solution = "all",
  dir.exp = dir_exp,
  progress = TRUE
)

altset_out
as.data.frame(altset_out)
altset_out$summary

# Special case: include the calibrated outcome among sampled calibration
# perturbations. The outcome stays out of conditions.
altset_outcome <- altset.test(
  raw.data = LR,
  calib.data = dat,
  outcome = outcome,
  conditions = conditions,
  calib_spec = calib_spec_outcome,
  test.conditions = c("DEV", "URB"),
  test.outcome = TRUE,
  unit_step = NULL,
  unit_step_divisor = 10,
  calib_max_steps = 5,
  incl.cut = 0.8,
  incl_step = 0.02,
  incl_max_steps = 5,
  n.cut = 1,
  ncut_step = 1,
  ncut_max_steps = 2,
  n_draws = 50,
  seed = 456,
  solution = "all",
  dir.exp = dir_exp,
  progress = TRUE
)

as.data.frame(altset_outcome)
}

}
\seealso{
\code{\link[=calib.test]{calib.test()}}, \code{\link[=incl.test]{incl.test()}}, \code{\link[=ncut.test]{ncut.test()}}, \code{\link[=loo.test]{loo.test()}},
\code{\link[=subsample.test]{subsample.test()}}, \code{\link[=theory.test]{theory.test()}}, \code{\link[=cluster.test]{cluster.test()}}, \code{\link[=sol.df]{sol.df()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/qcaERT-conventions.R
\name{qcaERT_conventions}
\alias{qcaERT_conventions}
\title{qcaERT argument and output conventions}
\description{
The qcaERT robustness functions are siblings. They share a common public
structure for solution controls, exclusion handling, returned objects,
printing, and data-frame coercion. Individual function pages describe the
function-specific workflow; this page describes the package-wide conventions
that should be read consistently across the function family.
}
\section{Solution Controls}{

Most functions accept \code{solution}, \code{include}, \code{dir.exp}, \code{which_M}, and
\code{i_mode}.

\code{solution} may be \code{"all"}, \code{"con"}/\code{"conservative"},
\code{"par"}/\code{"parsimonious"}, or \code{"int"}/\code{"intermediate"}. For single-solution-type
calls, \code{include} is optional and is resolved from \code{solution}: conservative
uses \code{include = ""}, while parsimonious and intermediate use
\code{include = "?"}. Intermediate solutions require \code{dir.exp}.

When \code{solution = "all"}, do not supply \code{include}. The monitored set contains
conservative and parsimonious solutions by default. If \code{dir.exp} is supplied,
the intermediate solution is also monitored. In stepwise boundary searches
such as \code{\link[=incl.test]{incl.test()}}, \code{\link[=ncut.test]{ncut.test()}}, and \code{\link[=calib.test]{calib.test()}}, these solution
types are searched independently; one solution type changing does not stop
the search for another solution type. When a baseline or step fails in one
monitored solution type, inspect the solution-type-specific rows in
\code{diagnostics}, \code{results}, or supporting \verb{by_*} components before interpreting
an all-solution run as a single combined failure.

\code{which_M} selects the model or solution alternative to use when QCA
minimization returns multiple alternatives. \code{i_mode} controls which
intermediate branches are included; accepted values are \code{"all"} and
\code{"C1P1"}. Most robustness functions default to allowing all intermediate
branches. \code{theory.test()} defaults to \code{"C1P1"} because
comparative theory testing needs one comparable intermediate branch per
theory by default.
}

\section{Exclusions}{

Functions that can monitor parsimonious or intermediate solutions use a
common exclusion convention. \code{exclude_mode = "recompute"} recalculates excluded
rows for each perturbed, reduced, or sampled truth table using
\code{\link[QCA:findRows]{QCA::findRows()}} and \code{exclude_recompute}. \code{exclude_mode = "static"}
reuses the supplied \code{exclude_static} object. \code{exclude_mode = "none"} avoids
exclusion handling.

The family default for recomputation is
\code{exclude_recompute = list(type = 2)}. \code{\link[=incl.test]{incl.test()}}, \code{\link[=ncut.test]{ncut.test()}},
\code{\link[=calib.test]{calib.test()}}, \code{\link[=loo.test]{loo.test()}}, \code{\link[=subsample.test]{subsample.test()}}, \code{\link[=altset.test]{altset.test()}}, and
\code{\link[=theory.test]{theory.test()}} share this \code{exclude_mode}, \code{exclude_recompute}, and
\code{exclude_static} convention. \code{cluster.test()} starts from an existing truth
table and therefore accepts \code{exclude} directly instead of exposing
\code{exclude_mode}.
}

\section{Calibration Specifications}{

Calibration-family functions use \code{calib_spec} for calibration information.
A \code{calib_spec} entry is named by calibrated set, contains the raw source name
in \code{raw}, the calibration \code{type}, and \code{thresholds}, and may contain \code{method}
and a \code{calibrate} list forwarded to \code{\link[QCA:calibrate]{QCA::calibrate()}}.

For condition-only calibration tests, \code{calib_spec} is named by
\code{conditions}. When \code{calib.test()} or \code{altset.test()} is called with
\code{test.outcome = TRUE}, \code{calib_spec} is named by \code{c(conditions, outcome)}.
The outcome must not be included in
\code{conditions}; outcome calibration testing is requested explicitly with
\code{test.outcome = TRUE}.

Crisp conditions use one threshold. Fuzzy direct calibration supports three
thresholds in \code{c(E, C, I)} order or six thresholds in
\code{c(E1, C1, I1, I2, C2, E2)} order. Fuzzy indirect calibration treats its
thresholds as ordered cutpoints.

These calibration-perturbation routines are designed for QCA workflows using
crisp and fuzzy sets. They are not multi-value calibration tools.
In \code{\link[=loo.test]{loo.test()}} and \code{\link[=subsample.test]{subsample.test()}} with \code{calib = "recompute"}, a
\code{calib_spec} entry may additionally contain \code{findTh}, a named list of
\code{\link[QCA:findTh]{QCA::findTh()}} arguments, to re-estimate thresholds on each reduced or
subsampled raw dataset. Without \code{findTh}, the baseline \code{thresholds} stored in
\code{calib_spec} are reused.
}

\section{Result Objects}{

Most robustness functions return an S3 object with \code{diagnostics}, \code{results},
and \code{settings}, plus supporting components such as \code{baseline}, \code{bounds},
\code{by_direction}, \code{by_case}, \code{by_run}, \code{by_draw}, or \code{summary}.

\code{diagnostics} is the detailed internal table. \code{results} is the clean
table. The \code{print()} method shows a concise summary plus the
results, and \code{as.data.frame()} returns the clean \code{results} table.

For \code{\link[=incl.test]{incl.test()}}, \code{\link[=ncut.test]{ncut.test()}}, and \code{\link[=calib.test]{calib.test()}}, \code{result_shape}
controls the layout of the clean \code{results} table when \code{solution = "all"}.
The default \code{"wide"} layout keeps one row per tested path and uses
solution-type-prefixed columns such as \code{con_last_safe} and \code{par_reason}. The
\code{"long"} layout uses one row per tested path and solution type, with a
\code{solution_type} column. The raw \code{diagnostics} table is unchanged.

\code{cluster.test()} and \code{theory.test()} use structured \code{results} lists because
their clean output has more than one natural table. \code{cluster.test()} returns
\code{overview}, \code{clusters}, and \code{units}; \code{as.data.frame()} returns
\code{results$overview}. \code{theory.test()} returns \code{models}, \code{pairwise}, and
\code{solutions}; \code{as.data.frame()} returns \code{results$models}.
}

\section{Plotting}{

Plotting requires \code{ggplot2}. \code{plot()} methods are provided
for \code{incl_test}, \code{calib_test}, and \code{theory_test} objects. Interval and trace
views map directly to stored boundary-search diagnostics; theory plots map
selected theory-specific models into consistency/coverage space.
\code{\link[=sol.chart]{sol.chart()}} renders \code{\link[=sol.df]{sol.df()}} tables as solution charts. See
\code{?qcaERT_plots} and \code{?sol.chart}.
}

\section{QCA Solution Objects}{

\code{cluster.test()} and \code{\link[=sol.df]{sol.df()}} consume QCA minimization objects. \code{\link[=sol.df]{sol.df()}}
extracts a compact data frame, and \code{\link[=sol.chart]{sol.chart()}} gives a visual display of
that extracted table. Case and solution-expression reconstruction use the
data-frame \code{pims} component produced by \code{\link[QCA:minimize]{QCA::minimize()}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/qcaERT-tests.R
\name{qcaERT_tests}
\alias{qcaERT_tests}
\title{Choose a qcaERT robustness tool}
\description{
This page maps common QCA robustness tests to their respective functions
}
\section{Calibration}{

Use \code{\link[=calib.test]{calib.test()}} when you want to know whether a QCA solution is sensitive
to calibration thresholds. It perturbs one calibration anchor at a time and
reports how far each threshold can move before the monitored solution
changes or the search reaches a boundary.

You can also use \code{\link[=altset.test]{altset.test()}} when you want calibration perturbations to
be combined with other perturbations in sampled alternative analysis settings,
such as alternative inclusion cutoffs or frequency cutoffs.
}

\section{Inclusion or n.cut}{

Use \code{\link[=incl.test]{incl.test()}} to check the inclusion (consistency) cutoff for
sufficiency. It searches below and above the baseline \code{incl.cut} and records
the last safe cutoff and first failing cutoff in each direction.

Use \code{\link[=ncut.test]{ncut.test()}} when the concern is the minimum number of cases under which
a truth table row is declared as a remainder. It searches lower and upper
\code{n.cut} values and records when the monitored solution changes or the
feasible boundary is reached.
}

\section{Cases}{

Use \code{\link[=loo.test]{loo.test()}} when you suspect individual cases may drive the solution.
It removes selected cases one at a time and compares each reduced analysis
with the baseline.

Use \code{\link[=subsample.test]{subsample.test()}} when you want repeated partial-sample checks. It draws
subsamples, rebuilds the QCA analysis, and summarizes how often the baseline
solution is preserved across runs. This is a stringent, punishing robustness
check and is most useful when sample-composition sensitivity is substantively
important.
}

\section{Groups or clusters}{

Use \code{\link[=cluster.test]{cluster.test()}} when you expect heterogeneity across clusters, groups,
or repeated units. It compares cluster-specific consistency, coverage, and
related fit patterns against the overall truth table.
}

\section{Theory Specifications}{

Use \code{\link[=theory.test]{theory.test()}} when you want to compare several theoretically motivated
condition sets under the same outcome, truth-table cutoffs, solution type,
exclusion handling, and settings for \code{which_M} and \code{i_mode}.
}

\section{Reporting QCA Solutions}{

Use \code{\link[=sol.df]{sol.df()}} when you want QCA minimization objects turned into a compact,
data frame. Use \code{\link[=sol.chart]{sol.chart()}} when you want that table
rendered as a visual chart of prime implicants.
}

\section{How To Read The Results}{

Most qcaERT robustness functions return \code{diagnostics}, \code{results}, and
\code{settings}. \code{diagnostics} is the detailed table; \code{results} is the clean table
returned by \code{as.data.frame()}. \code{print()} shows a concise summary. The shared
output structure is described in \code{?qcaERT_conventions}.

For \code{\link[=calib.test]{calib.test()}}, \code{\link[=incl.test]{incl.test()}}, and \code{\link[=theory.test]{theory.test()}} objects, \code{plot()}
methods provide optional visual inspection helpers when \code{ggplot2} is
installed. \code{\link[=sol.chart]{sol.chart()}} provides the corresponding visual presentation for
\code{\link[=sol.df]{sol.df()}} tables. See \code{?qcaERT_plots}.
}

\examples{
\donttest{
?qcaERT_tests
?qcaERT_conventions
?qcaERT_plots
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/theory.test.R
\name{theory.test}
\alias{theory.test}
\alias{print.theory_test}
\alias{as.data.frame.theory_test}
\title{Theory-specification robustness for QCA models}
\usage{
theory.test(
  data,
  outcome,
  theories,
  incl.cut = 1,
  n.cut = 1,
  solution = "all",
  include = NULL,
  dir.exp = NULL,
  which_M = 1,
  i_mode = "C1P1",
  exclude_mode = c("recompute", "static", "none"),
  exclude_recompute = list(type = 2),
  exclude_static = NULL,
  progress = TRUE,
  ...
)

\method{print}{theory_test}(x, row.names = FALSE, ...)

\method{as.data.frame}{theory_test}(x, ...)
}
\arguments{
\item{data}{A non-empty data frame containing the calibrated outcome and
calibrated condition columns.}

\item{outcome}{Name of the calibrated outcome column in \code{data}.}

\item{theories}{Named list of theory-specific condition sets. Each entry
must be a non-empty character vector of condition names in \code{data}.
Conditions may be shared across theories, but no theory may contain the
outcome as a condition.}

\item{incl.cut}{Inclusion cutoff to be used for every theory-specific truth
table.}

\item{n.cut}{Frequency cutoff to be used for every theory-specific truth
table.}

\item{solution}{Solution type to compare. Accepted values are \code{"all"},
\code{"con"} or \code{"conservative"}, \code{"par"} or \code{"parsimonious"}, and \code{"int"} or
\code{"intermediate"}.}

\item{include}{Optional minimization include setting. Currently, this
argument accepts only \code{NULL}, \code{""}, or \code{"?"}.}

\item{dir.exp}{Theory-specific directional expectations. Use \code{NULL} when no
intermediate solutions are monitored. Otherwise supply a named list with
one entry per theory. Each entry may be an ordered character vector aligned
to that theory's conditions or a named character vector whose names match
that theory's conditions.}

\item{which_M}{Positive integer giving which solution alternative to use
when minimization returns multiple models.}

\item{i_mode}{Character string controlling intermediate-solution selection.
Accepted values are \code{"C1P1"} and \code{"all"}. The default is \code{"C1P1"} because
comparative theory testing needs a single comparable intermediate branch
per theory by default.}

\item{exclude_mode}{Character string controlling how excluded rows are
handled for parsimonious and intermediate minimization. \code{"recompute"}
recalculates exclusions separately for each theory-specific truth table,
\code{"static"} reuses theory-specific static exclusions supplied through
\code{exclude_static}, and \code{"none"} does not use exclusions.}

\item{exclude_recompute}{Named list of arguments passed to \code{\link[QCA:findRows]{QCA::findRows()}}
when \code{exclude_mode = "recompute"}.}

\item{exclude_static}{Static exclusions used when \code{exclude_mode = "static"}.
For
\code{theory.test()}, this must be a named list with one entry per theory when
parsimonious or intermediate solutions are monitored.}

\item{progress}{Logical; if \code{TRUE}, show a text progress bar during
per-theory QCA runs in interactive sessions.}

\item{...}{Additional arguments split between \code{\link[QCA:truthTable]{QCA::truthTable()}} and
\code{\link[QCA:minimize]{QCA::minimize()}}. The function also looks in \code{...} for \code{include},
\code{dir.exp}, or \code{direxp} if those arguments were not supplied explicitly. In
\code{\link[=print.theory_test]{print.theory_test()}}, \code{...} is passed to \code{\link[=print.data.frame]{print.data.frame()}}. In
\code{\link[=as.data.frame.theory_test]{as.data.frame.theory_test()}}, \code{...} is ignored.}

\item{x}{A \code{theory_test} object returned by \code{\link[=theory.test]{theory.test()}}.}

\item{row.names}{Logical; if \code{TRUE}, print row names in the selected
solutions table printed by \code{\link[=print.theory_test]{print.theory_test()}}.}
}
\value{
An object of class \code{theory_test} with the following components:
\describe{
\item{\code{diagnostics}}{A detailed internal diagnostics table with one row
per theory and monitored solution type.}
\item{\code{results}}{A named list with \code{models}, \code{pairwise}, and
\code{solutions} tables. \code{models} is populated with model-level diagnostics
and fit values; \code{solutions} is populated with selected solution terms
and term-level fit values; \code{pairwise} is populated with pairwise theory
comparisons by solution type.}
\item{\code{by_theory}}{A named list containing the theory-specific condition
sets, directional expectations, truth tables, exclusions, minimization
objects, selected solution terms used for comparison, and current run
status.}
\item{\code{settings}}{A list containing the analysis settings used in the
call.}
}

\code{print.theory_test()} prints a compact theory-specification summary and
the selected solutions table. \code{as.data.frame.theory_test()} returns
\code{results$models}.
}
\description{
Compare several theoretically oriented QCA condition sets while holding the
outcome, truth-table cutoffs, solution type, exclusion handling, and
model-selection settings constant. Each theory supplies its own condition
set.
}
\details{
The function builds a separate truth table for each theory, recomputes
exclusions separately when requested, and runs the monitored solution types
under the same analytic settings. The raw per-theory QCA objects are stored
in \code{by_theory}. \code{results$models} gives the clean model-level comparison
table, and \code{results$solutions} gives the selected prime implicants or
solution terms by theory and solution type. \code{results$pairwise} compares selected
solution memberships across theories using fuzzy Jaccard similarity and
mean absolute membership differences.
}
\examples{
\donttest{
library(QCA)
data(LR)

outcome <- "SURV"

thresholds <- list(
  DEV = findTh(LR$DEV, groups = 7),
  URB = findTh(LR$URB, groups = 4),
  LIT = findTh(LR$LIT, groups = 4),
  IND = findTh(LR$IND, groups = 4),
  STB = findTh(LR$STB, groups = 4),
  SURV = findTh(LR$SURV, groups = 4)
)

dat <- LR
dat$DEV <- calibrate(LR$DEV, type = "fuzzy", thresholds = thresholds$DEV)
dat$URB <- calibrate(LR$URB, type = "fuzzy", thresholds = thresholds$URB)
dat$LIT <- calibrate(LR$LIT, type = "fuzzy", thresholds = thresholds$LIT)
dat$IND <- calibrate(LR$IND, type = "fuzzy", thresholds = thresholds$IND)
dat$STB <- calibrate(LR$STB, type = "fuzzy", thresholds = thresholds$STB)
dat$SURV <- calibrate(LR$SURV, type = "fuzzy", thresholds = thresholds$SURV)

theories <- list(
  development = c("DEV", "URB", "LIT"),
  industrial = c("DEV", "URB", "IND"),
  broad = c("DEV", "URB", "LIT", "IND", "STB")
)

dir_exp <- list(
  development = c("1", "1", "1"),
  industrial = c("1", "1", "1"),
  broad = c("1", "1", "1", "1", "1")
)

theory_out <- theory.test(
  data = dat,
  outcome = outcome,
  theories = theories,
  incl.cut = 0.8,
  n.cut = 1,
  solution = "all",
  dir.exp = dir_exp,
  progress = TRUE
)

theory_out
as.data.frame(theory_out)
theory_out$results$solutions
theory_out$results$pairwise
}

}
\seealso{
\code{\link[=calib.test]{calib.test()}}, \code{\link[=incl.test]{incl.test()}}, \code{\link[=ncut.test]{ncut.test()}}, \code{\link[=loo.test]{loo.test()}},
\code{\link[=subsample.test]{subsample.test()}}, \code{\link[=altset.test]{altset.test()}}, \code{\link[=cluster.test]{cluster.test()}}, \code{\link[=sol.df]{sol.df()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/incl.test.R
\name{incl.test}
\alias{incl.test}
\alias{print.incl_test}
\alias{as.data.frame.incl_test}
\title{Inclusion-cutoff robustness test for QCA solutions}
\usage{
incl.test(
  data,
  outcome,
  conditions = NULL,
  incl.cut = 1,
  step = 0.01,
  max_steps = 20,
  n.cut = 1,
  solution = "all",
  include = NULL,
  dir.exp = NULL,
  which_M = 1,
  i_mode = c("all", "C1P1"),
  exclude_mode = c("recompute", "static", "none"),
  exclude_recompute = list(type = 2),
  exclude_static = NULL,
  result_shape = c("wide", "long"),
  progress = TRUE,
  ...
)

\method{print}{incl_test}(x, row.names = FALSE, ...)

\method{as.data.frame}{incl_test}(x, ...)
}
\arguments{
\item{data}{A non-empty data frame object containing the outcome and
condition columns used in the QCA analysis.}

\item{outcome}{Name of the outcome. This must be a single
non-empty character string.}

\item{conditions}{Optional character vector of condition names. If \code{NULL},
the condition set is left to \code{\link[QCA:truthTable]{QCA::truthTable()}}.}

\item{incl.cut}{Baseline inclusion cutoff passed to \code{\link[QCA:truthTable]{QCA::truthTable()}}.
This must be a single finite number in \verb{[0, 1]}.}

\item{step}{Positive numeric step size used to move \code{incl.cut} downward and
upward from the baseline value.}

\item{max_steps}{Maximum number of stepwise moves to attempt in each
direction.}

\item{n.cut}{Frequency cutoff passed to \code{\link[QCA:truthTable]{QCA::truthTable()}}.}

\item{solution}{Solution type to monitor. Accepted values are \code{"all"},
\code{"con"} or \code{"conservative"}, \code{"par"} or \code{"parsimonious"}, and \code{"int"} or
\code{"intermediate"}.}

\item{include}{Optional minimization include setting. Currently, this
argument accepts only \code{NULL}, \code{""}, or \code{"?"}.}

\item{dir.exp}{Directional expectations used when the monitored solution is
intermediate.}

\item{which_M}{Positive integer giving which solution alternative to use
when minimization returns multiple models.}

\item{i_mode}{Character string controlling intermediate-solution selection.
Accepted values are \code{"all"} and \code{"C1P1"}.}

\item{exclude_mode}{Character string controlling how excluded rows are
handled for parsimonious and intermediate minimization. \code{"recompute"}
recalculates exclusions from each truth table, \code{"static"} reuses
\code{exclude_static}, and \code{"none"} does not use exclusions.}

\item{exclude_recompute}{Named list of arguments passed to \code{\link[QCA:findRows]{QCA::findRows()}}
when \code{exclude_mode = "recompute"}.}

\item{exclude_static}{Already computed exclusion object reused when
\code{exclude_mode = "static"}.}

\item{result_shape}{Layout of the clean \code{results} table when
\code{solution = "all"}. \code{"wide"} keeps one row per direction with
solution-type-specific columns such as \code{con_last_safe} and \code{par_reason}.
\code{"long"} returns one row per direction and solution type, with a \code{solution_type}
column. Single-solution-type calls keep the compact single-solution-type
layout.}

\item{progress}{Logical; if \code{TRUE} and the session is interactive, show a
text progress bar.}

\item{...}{Additional arguments routed through the QCA workflow. Arguments
matching \code{\link[QCA:truthTable]{QCA::truthTable()}} are forwarded to truth table construction;
remaining minimization arguments are filtered and forwarded to
\code{\link[QCA:minimize]{QCA::minimize()}}. The function also looks in \code{...} for \code{include},
\code{dir.exp}, or \code{direxp} if those arguments were not supplied explicitly.
In \code{\link[=print.incl_test]{print.incl_test()}}, \code{...} is passed to \code{\link[=print.data.frame]{print.data.frame()}}. In
\code{\link[=as.data.frame.incl_test]{as.data.frame.incl_test()}}, \code{...} is ignored.}

\item{x}{An \code{incl_test} object returned by \code{\link[=incl.test]{incl.test()}}.}

\item{row.names}{Logical; passed to \code{\link[=print.data.frame]{print.data.frame()}} by
\code{\link[=print.incl_test]{print.incl_test()}}.}
}
\value{
An object of class \code{incl_test} with the following components:
\describe{
\item{\code{diagnostics}}{A detailed data frame with one row for the lower
search and one row for the upper search. When \code{solution = "all"}, each
monitored solution type is searched independently, so diagnostics has
one row per direction and solution type. It includes the baseline
\code{incl.cut}, the last safe value, the first failing value, the number of
successful steps, stop reason, change classification, and exclusion
information for the baseline, last safe, and first failing runs.}
\item{\code{results}}{A compact data frame with the columns
\code{direction}, \code{start}, \code{last_safe}, \code{first_failing}, \code{steps},
\code{total_delta}, and \code{reason}. When \code{solution = "all"} and
\code{result_shape = "wide"}, the row unit remains \code{direction}, but the
boundary and reason columns are solution-type-specific, using prefixes \code{con_},
\code{par_}, and, when available, \code{int_}. When \code{result_shape = "long"},
\code{results} has one row per direction and solution type.}
\item{\code{bounds}}{A named numeric vector with \code{Lower} and \code{Upper}
elements taken from the last safe values in each search direction. When
\code{solution = "all"}, this is a lower/upper matrix with one column per
monitored solution type.}
\item{\code{baseline}}{A list containing the baseline truth table,
minimization results, selected solution terms used for comparison,
exclusion set used, and
status information.}
\item{\code{by_direction}}{A named list with one entry for the lower search
and one for the upper search, each containing the search trace and
stopping information.}
\item{\code{settings}}{A list containing the analysis settings used to build
the result object.}
}

\code{print.incl_test()} prints a concise summary and the \code{results} table.
\code{as.data.frame.incl_test()} returns the \code{results} table.
If \code{ggplot2} is installed, \code{plot.incl_test()} provides interval and trace
views; see \code{?qcaERT_plots}.
}
\description{
Perturbs the truth table inclusion cutoff used in the analysis and checks
when the monitored QCA solution changes. Starting from a baseline
\code{incl.cut}, the function searches downward and upward in fixed steps and
records the last value that preserved the baseline solution, the first value
that changed the solution or triggered an error, and the reason the search
stopped in each direction.
}
\details{
The baseline analysis is built with \code{\link[QCA:truthTable]{QCA::truthTable()}} using the supplied
\code{incl.cut} and \code{n.cut}, followed by minimization under the requested
solution type. The function then tests lower and upper values of
\code{incl.cut} one step at a time. When \code{solution = "all"}, conservative,
parsimonious, and, when requested, intermediate paths are searched
independently.

A directional search stops when one of the following occurs:
\itemize{
\item the next tested value falls outside \verb{[0, 1]},
\item truth table construction fails,
\item exclusion recomputation fails,
\item minimization fails, or
\item the monitored solution changes relative to the baseline.
}

The shared solution-control, exclusion, and returned-object conventions are
described in \code{?qcaERT_conventions}.
}
\examples{
\donttest{
library(QCA)
data(LR)

conditions <- c("DEV", "URB", "LIT", "IND", "STB")
outcome <- "SURV"
dir_exp <- rep("1", length(conditions))

thresholds <- list(
  DEV = findTh(LR$DEV, groups = 7),
  URB = findTh(LR$URB, groups = 4),
  LIT = findTh(LR$LIT, groups = 4),
  IND = findTh(LR$IND, groups = 4),
  STB = findTh(LR$STB, groups = 4),
  SURV = findTh(LR$SURV, groups = 4)
)

dat <- LR
dat$DEV <- calibrate(LR$DEV, type = "fuzzy", thresholds = thresholds$DEV)
dat$URB <- calibrate(LR$URB, type = "fuzzy", thresholds = thresholds$URB)
dat$LIT <- calibrate(LR$LIT, type = "fuzzy", thresholds = thresholds$LIT)
dat$IND <- calibrate(LR$IND, type = "fuzzy", thresholds = thresholds$IND)
dat$STB <- calibrate(LR$STB, type = "fuzzy", thresholds = thresholds$STB)
dat$SURV <- calibrate(LR$SURV, type = "fuzzy", thresholds = thresholds$SURV)

out <- incl.test(
  data = dat,
  outcome = outcome,
  conditions = conditions,
  incl.cut = 0.8,
  step = 0.05,
  max_steps = 5,
  n.cut = 1,
  solution = "all",
  dir.exp = dir_exp,
  result_shape = "long",
  progress = TRUE
)

out
as.data.frame(out)
plot(out, solution_type = "conservative")
}

}
\seealso{
\link{qcaERT_plots}, \code{\link[=calib.test]{calib.test()}}, \code{\link[=ncut.test]{ncut.test()}}, \code{\link[=loo.test]{loo.test()}},
\code{\link[=subsample.test]{subsample.test()}}, \code{\link[=altset.test]{altset.test()}}, \code{\link[=theory.test]{theory.test()}}, \code{\link[=cluster.test]{cluster.test()}},
\code{\link[=sol.df]{sol.df()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sol.df.R
\name{sol.df}
\alias{sol.df}
\title{Build a compact solution table from QCA minimization output}
\usage{
sol.df(
  conservative = NULL,
  intermediate = NULL,
  parsimonious = NULL,
  solution = "all",
  which_M = 1,
  i_mode = c("all", "C1P1"),
  branches = NULL,
  include_cases = TRUE,
  pi_incl_cut = NULL,
  digits = NULL,
  na_string = "-",
  verbose = FALSE
)
}
\arguments{
\item{conservative}{Optional conservative minimization result of class
\code{"QCA_min"}.}

\item{intermediate}{Optional intermediate minimization result of class
\code{"QCA_min"}.}

\item{parsimonious}{Optional parsimonious minimization result of class
\code{"QCA_min"}.}

\item{solution}{Which solution type to extract. Accepted values are
\code{"all"}, \code{"con"} or \code{"conservative"}, \code{"par"} or \code{"parsimonious"}, and
\code{"int"} or \code{"intermediate"}. When \code{solution = "all"}, the function
extracts every supplied object among \code{conservative}, \code{parsimonious}, and
\code{intermediate}.}

\item{which_M}{Positive integer giving which model to extract when a
supplied minimization object contains multiple models.}

\item{i_mode}{Character string controlling which intermediate branches to
extract when \code{intermediate} is supplied and \code{branches = NULL}. Accepted
values are \code{"all"} and \code{"C1P1"}.}

\item{branches}{Optional character vector of intermediate branch names to
extract from \code{intermediate$i.sol}. When supplied, this overrides \code{i_mode}.}

\item{include_cases}{Logical; if \code{TRUE}, include a \code{Cases} column using the
case information stored in the minimization object when available. Case
reconstruction expects the data-frame \code{pims} component produced by
\code{\link[QCA:minimize]{QCA::minimize()}}.}

\item{pi_incl_cut}{Optional lower cutoff applied to the prime-implicant
consistency column. Rows with \code{Consistency_PI < pi_incl_cut} are removed.}

\item{digits}{Optional non-negative integer used to round numeric columns in
the returned table.}

\item{na_string}{Single character string used to replace missing values in
non-numeric output fields. The default is \code{"-"}.}

\item{verbose}{Logical; if \code{TRUE}, print a short progress message while
processing each requested solution type.}
}
\value{
A data frame with one row per extracted prime implicant and the
following columns:
\describe{
\item{\code{Solution}}{Solution type label: \code{"Conservative"},
\code{"Parsimonious"}, or \code{"Intermediate"}.}
\item{\code{Model}}{Selected model number when model-specific output is
available.}
\item{\code{Intermediate_CnPn}}{Intermediate branch label when rows come from
\code{intermediate$i.sol}.}
\item{\code{Prime_Implicants}}{Prime implicant name.}
\item{\code{Consistency_PI}}{Prime-implicant consistency (\code{inclS}).}
\item{\code{PRI_PI}}{Prime-implicant PRI.}
\item{\code{Raw_Coverage_PI}}{Prime-implicant raw coverage (\code{covS}).}
\item{\code{Unique_Coverage_PI}}{Prime-implicant unique coverage (\code{covU}).}
\item{\code{Solution_Consistency}}{Solution-level consistency.}
\item{\code{Solution_PRI}}{Solution-level PRI.}
\item{\code{Solution_Coverage}}{Solution-level coverage.}
\item{\code{Cases}}{Case labels or case strings when available and requested.}
}

The function returns a regular data frame, not a custom result object.
}
\description{
Extracts prime implicants, fit statistics, model identifiers, intermediate
branch labels, and optional case membership strings from supplied
\code{\link[QCA:minimize]{QCA::minimize()}} result objects and returns them as one combined data
frame.
}
\details{
Supply at least one of \code{conservative}, \code{parsimonious}, or \code{intermediate}.

For conservative and parsimonious solutions, the function extracts rows from
the \code{IC} component of the supplied minimization object. For intermediate
solutions, it extracts rows from the selected \code{i.sol} branches.

If a supplied minimization object contains multiple models in
\code{IC$individual}, \code{which_M} selects the model position to extract. If the
requested model does not exist, the function stops with an error.

When \code{include_cases = TRUE}, the function first uses any \code{cases} column
already stored in the relevant \code{incl.cov} table. If that is unavailable, it
tries to reconstruct case strings from the corresponding \code{pims} data frame
stored by \code{\link[QCA:minimize]{QCA::minimize()}}.

The QCA solution-object conventions are described in \code{?qcaERT_conventions}.
}
\examples{
library(QCA)
data(LC)

conditions <- c("DEV", "URB", "LIT", "IND", "STB")
dir_exp <- rep("1", length(conditions))

tt <- truthTable(
  data = LC,
  outcome = "SURV",
  conditions = conditions,
  incl.cut = 0.8,
  n.cut = 1
)

enhanced <- findRows(tt, type = 2)

con <- minimize(
  input = tt,
  include = "",
  details = TRUE,
  show.cases = FALSE
)

par <- minimize(
  input = tt,
  include = "?",
  exclude = enhanced,
  details = TRUE,
  show.cases = FALSE
)

int <- minimize(
  input = tt,
  include = "?",
  dir.exp = dir_exp,
  exclude = enhanced,
  details = TRUE,
  show.cases = FALSE
)

sol.df(
  conservative = con,
  parsimonious = par,
  intermediate = int,
  solution = "all",
  which_M = 1,
  i_mode = "C1P1",
  include_cases = FALSE,
  digits = 2
)

\donttest{
library(QCA)
data(LR)

conditions <- c("DEV", "URB", "LIT", "IND", "STB")
outcome <- "SURV"
dir_exp <- rep("1", length(conditions))

thresholds <- list(
  DEV = findTh(LR$DEV, groups = 7),
  URB = findTh(LR$URB, groups = 4),
  LIT = findTh(LR$LIT, groups = 4),
  IND = findTh(LR$IND, groups = 4),
  STB = findTh(LR$STB, groups = 4),
  SURV = findTh(LR$SURV, groups = 4)
)

dat <- LR
dat$DEV <- calibrate(LR$DEV, type = "fuzzy", thresholds = thresholds$DEV)
dat$URB <- calibrate(LR$URB, type = "fuzzy", thresholds = thresholds$URB)
dat$LIT <- calibrate(LR$LIT, type = "fuzzy", thresholds = thresholds$LIT)
dat$IND <- calibrate(LR$IND, type = "fuzzy", thresholds = thresholds$IND)
dat$STB <- calibrate(LR$STB, type = "fuzzy", thresholds = thresholds$STB)
dat$SURV <- calibrate(LR$SURV, type = "fuzzy", thresholds = thresholds$SURV)

tt <- truthTable(
  data = dat,
  outcome = outcome,
  conditions = conditions,
  incl.cut = 0.8,
  n.cut = 1,
  complete = TRUE,
  show.cases = TRUE
)
enhanced <- findRows(tt, type = 2)

con <- minimize(tt, include = "", details = TRUE, show.cases = FALSE)
par <- minimize(tt, include = "?", exclude = enhanced, details = TRUE, show.cases = FALSE)
int <- minimize(
  tt,
  include = "?",
  dir.exp = dir_exp,
  exclude = enhanced,
  details = TRUE,
  show.cases = FALSE
)

sol.df(
  conservative = con,
  parsimonious = par,
  intermediate = int,
  solution = "all",
  which_M = 1,
  i_mode = "C1P1",
  include_cases = TRUE,
  digits = 3
)
}

}
\seealso{
\code{\link[=sol.chart]{sol.chart()}}, \code{\link[=calib.test]{calib.test()}}, \code{\link[=incl.test]{incl.test()}}, \code{\link[=ncut.test]{ncut.test()}},
\code{\link[=loo.test]{loo.test()}}, \code{\link[=subsample.test]{subsample.test()}}, \code{\link[=altset.test]{altset.test()}}, \code{\link[=theory.test]{theory.test()}},
\code{\link[=cluster.test]{cluster.test()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot.calib.R, R/plot.incl.R, R/plot.theory.R,
%   R/qcaERT-plots.R
\name{qcaERT_plots}
\alias{qcaERT_plots}
\alias{plot.calib_test}
\alias{plot.incl_test}
\alias{plot.theory_test}
\title{Plot and chart qcaERT results}
\usage{
\method{plot}{calib_test}(
  x,
  type = c("interval", "heatmap", "trace"),
  sets = NULL,
  roles = NULL,
  anchors = NULL,
  directions = c("lower", "upper"),
  stop_reason = NULL,
  changed_types = NULL,
  solution_type = NULL,
  solution = NULL,
  monitored_solutions = NULL,
  metric = c("raw", "pct", "steps"),
  value = c("delta", "last_safe", "failing"),
  abs_delta = FALSE,
  cell = c("anchor_direction", "anchor"),
  show_text = FALSE,
  set = NULL,
  anchor = NULL,
  direction = NULL,
  show_stop = TRUE,
  order_sets = c("input", "most_sensitive", "least_sensitive"),
  legend = TRUE,
  theme = .plot_theme(),
  ...
)

\method{plot}{incl_test}(
  x,
  type = c("interval", "trace"),
  directions = c("lower", "upper"),
  stop_reason = NULL,
  changed_types = NULL,
  solution_type = NULL,
  solution = NULL,
  monitored_solutions = NULL,
  i_mode = NULL,
  direction = NULL,
  show_stop = TRUE,
  legend = TRUE,
  theme = .plot_theme(),
  ...
)

\method{plot}{theory_test}(
  x,
  solution_type = NULL,
  intermediate_branch = NULL,
  show_labels = TRUE,
  label_line = TRUE,
  label_line_alpha = 0.45,
  label_line_width = 0.35,
  point_size = 3.5,
  text_size = 3.5,
  label_nudge_x = 0.008,
  label_nudge_y = 0.006,
  legend = TRUE,
  theme = .plot_theme(),
  ...
)
}
\arguments{
\item{x}{An \code{incl_test} object returned by \code{\link[=incl.test]{incl.test()}}, a \code{calib_test}
object returned by \code{\link[=calib.test]{calib.test()}}, or a \code{theory_test} object returned by
\code{\link[=theory.test]{theory.test()}}.}

\item{type}{Plot type. \code{incl_test} supports \code{"interval"} and \code{"trace"}.
\code{calib_test} supports \code{"interval"}, \code{"heatmap"}, and \code{"trace"}.
\code{theory_test} has one consistency-coverage plot and does not use \code{type}.}

\item{sets}{For \code{calib_test}, optional character vector selecting calibrated
sets.}

\item{roles}{For \code{calib_test}, optional character vector selecting set
roles. Supported values are \code{"condition"} and \code{"outcome"}.}

\item{anchors}{For \code{calib_test}, optional character vector selecting
calibration anchors. Supported anchors are taken from the result object and
may include crisp (\code{"T"}), fuzzy direct three-threshold (\code{"E"}, \code{"C"},
\code{"I"}), fuzzy direct six-threshold (\code{"E1"}, \code{"C1"}, \code{"I1"}, \code{"I2"},
\code{"C2"}, \code{"E2"}), or fuzzy indirect (\code{"T1"}, \code{"T2"}, ...) anchors.}

\item{directions}{Character vector selecting lower and/or upper searches.}

\item{stop_reason}{Optional character vector used to filter diagnostic rows
by stop reason before plotting.}

\item{changed_types}{Optional solution-type filter applied to comma-coded
\code{changed_types} diagnostic values. Accepted tokens follow the common
qcaERT solution-type conventions: \code{"con"}/\code{"conservative"},
\code{"par"}/\code{"parsimonious"}, and \code{"int"}/\code{"intermediate"}.}

\item{solution_type}{Solution type to plot. Required when more than one
solution type is present. Accepted values follow the common qcaERT
solution-type conventions, excluding \code{"all"}.}

\item{solution}{Reserved. Plot methods do not accept \code{solution}; use
\code{solution_type}.}

\item{monitored_solutions}{Reserved. Plot methods do not accept
\code{monitored_solutions}; use \code{solution_type}.}

\item{metric}{For \code{calib_test} heatmaps, the quantity used for the fill
color: \code{"raw"}, \code{"pct"}, or \code{"steps"}.}

\item{value}{For \code{calib_test} heatmaps with \code{metric = "raw"}, the raw
threshold value to show: \code{"delta"}, \code{"last_safe"}, or \code{"failing"}.}

\item{abs_delta}{Logical. If \code{TRUE}, heatmaps using raw or percentage deltas
display absolute values.}

\item{cell}{For \code{calib_test} heatmaps, whether cells are split by
anchor-direction path (\code{"anchor_direction"}) or by anchor only
(\code{"anchor"}).}

\item{show_text}{Logical. If \code{TRUE}, add rounded heatmap values as text.}

\item{set}{For \code{calib_test} trace plots, the single calibrated set to plot.}

\item{anchor}{For \code{calib_test} trace plots, the single calibration anchor to
plot.}

\item{direction}{For \code{type = "trace"}, the single search direction to plot:
\code{"lower"} or \code{"upper"}.}

\item{show_stop}{Logical. If \code{TRUE}, draw reference lines for baseline and,
when present, first-failing values.}

\item{order_sets}{For \code{calib_test}, set ordering in interval and heatmap
views. \code{"input"} preserves result order; \code{"most_sensitive"} and
\code{"least_sensitive"} order by the smallest solution-changing percentage
delta when available.}

\item{legend}{Logical. If \code{FALSE}, hide the plot legend.}

\item{theme}{A ggplot2 theme object added to the plot.}

\item{...}{Additional graphical arguments forwarded to the primary ggplot2
geometry.}

\item{i_mode}{For \code{incl_test}, optional filter applied to the \code{i_mode}
diagnostic column. Accepted values are \code{"all"} and \code{"C1P1"}.}

\item{intermediate_branch}{For \code{theory_test} intermediate plots, the
intermediate branch to plot when more than one branch is present.}

\item{show_labels}{For \code{theory_test}, logical. If \code{TRUE}, print theory names
next to their consistency/coverage points.}

\item{label_line}{For \code{theory_test}, logical. If \code{TRUE}, draw a line from
each consistency/coverage point to its theory-name label.}

\item{label_line_alpha}{For \code{theory_test}, transparency of label connector
lines.}

\item{label_line_width}{For \code{theory_test}, width of label connector lines.}

\item{point_size}{For \code{theory_test}, point size.}

\item{text_size}{For \code{theory_test}, theory-name label size.}

\item{label_nudge_x}{For \code{theory_test}, horizontal nudge for theory-name
labels.}

\item{label_nudge_y}{For \code{theory_test}, vertical nudge for theory-name
labels.}
}
\value{
A ggplot object.
}
\description{
Plot methods provide visual inspection helpers for qcaERT result objects.
They are simplified views of the existing result object: the
detailed \code{diagnostics} table supplies the interval and heatmap views, while
the path-level trace components supply trace views. \code{\link[=sol.chart]{sol.chart()}} provides a
matching visual display for \code{\link[=sol.df]{sol.df()}} solution tables.
}
\details{
Plotting is optional. The qcaERT analysis functions do not require
\code{ggplot2}, but these methods do.
}
\section{Plot Types}{

For \code{incl_test}, \code{"interval"} shows the baseline inclusion cutoff and the
lower/upper last-safe values. \code{"trace"} shows the stepwise path for one
search direction.

For \code{calib_test}, \code{"interval"} shows baseline and lower/upper last-safe
threshold values by set and anchor. \code{"heatmap"} summarizes sensitivity
across anchor paths. \code{"trace"} shows the stepwise path for one set,
anchor, and direction.

For \code{theory_test}, the plot compares theory-specific selected models in
consistency/coverage space for one selected solution type. Theory names are shown
near the points, and the legend pairs each theory with its selected solution
terms.

For \code{\link[=sol.df]{sol.df()}} tables, \code{\link[=sol.chart]{sol.chart()}} draws a table-like chart of sufficient
configurations, using filled and open circles for condition presence and
absence.
}

\section{Family Consistency}{

These methods follow the common qcaERT output conventions described in
\code{?qcaERT_conventions}: plots read from \code{diagnostics} and path-level supporting
components, while \code{print()} and \code{as.data.frame()} keep their existing
concise and tabular behavior. \code{\link[=sol.chart]{sol.chart()}} follows the same division of
labor by reading from the already-extracted \code{\link[=sol.df]{sol.df()}} table.
}

\examples{
\donttest{
library(QCA)
library(ggplot2)

data(LR)

conditions <- c("DEV", "URB", "LIT", "IND", "STB")
outcome <- "SURV"
dir_exp <- rep("1", length(conditions))

thresholds <- list(
  DEV = findTh(LR$DEV, groups = 7),
  URB = findTh(LR$URB, groups = 4),
  LIT = findTh(LR$LIT, groups = 4),
  IND = findTh(LR$IND, groups = 4),
  STB = findTh(LR$STB, groups = 4),
  SURV = findTh(LR$SURV, groups = 4)
)

dat <- LR
dat$DEV <- calibrate(LR$DEV, type = "fuzzy", thresholds = thresholds$DEV)
dat$URB <- calibrate(LR$URB, type = "fuzzy", thresholds = thresholds$URB)
dat$LIT <- calibrate(LR$LIT, type = "fuzzy", thresholds = thresholds$LIT)
dat$IND <- calibrate(LR$IND, type = "fuzzy", thresholds = thresholds$IND)
dat$STB <- calibrate(LR$STB, type = "fuzzy", thresholds = thresholds$STB)
dat$SURV <- calibrate(LR$SURV, type = "fuzzy", thresholds = thresholds$SURV)

incl_out <- incl.test(
  data = dat,
  outcome = outcome,
  conditions = conditions,
  incl.cut = 0.8,
  step = 0.05,
  max_steps = 5,
  n.cut = 1,
  solution = "all",
  dir.exp = dir_exp,
  progress = TRUE
)

calib_spec <- list(
  DEV = list(raw = "DEV", type = "fuzzy", method = "direct", thresholds = thresholds$DEV),
  URB = list(raw = "URB", type = "fuzzy", method = "direct", thresholds = thresholds$URB),
  LIT = list(raw = "LIT", type = "fuzzy", method = "direct", thresholds = thresholds$LIT),
  IND = list(raw = "IND", type = "fuzzy", method = "direct", thresholds = thresholds$IND),
  STB = list(raw = "STB", type = "fuzzy", method = "direct", thresholds = thresholds$STB)
)

calib_out <- calib.test(
  raw.data = LR,
  calib.data = dat,
  outcome = outcome,
  conditions = conditions,
  calib_spec = calib_spec,
  test.conditions = c("DEV", "URB"),
  unit_step = NULL,
  unit_step_divisor = 10,
  max_steps = 5,
  incl.cut = 0.8,
  n.cut = 1,
  solution = "all",
  dir.exp = dir_exp,
  progress = TRUE
)

theories <- list(
  development = c("DEV", "URB", "LIT"),
  industrial = c("DEV", "URB", "IND"),
  broad = c("DEV", "URB", "LIT", "IND", "STB")
)

theory_out <- theory.test(
  data = dat,
  outcome = outcome,
  theories = theories,
  incl.cut = 0.8,
  n.cut = 1,
  solution = "all",
  dir.exp = list(
    development = c("1", "1", "1"),
    industrial = c("1", "1", "1"),
    broad = c("1", "1", "1", "1", "1")
  ),
  progress = TRUE
)

plot(incl_out, solution_type = "conservative")
plot(incl_out, solution_type = "conservative", type = "trace", direction = "lower")
plot(calib_out, solution_type = "conservative")
plot(calib_out, solution_type = "conservative", type = "heatmap")
plot(
  calib_out,
  solution_type = "conservative",
  type = "trace",
  set = "DEV",
  anchor = "E1",
  direction = "lower"
)
plot(theory_out, solution_type = "conservative")
}

}
